#include "func.mcr"
#include "graph.h"
#include "graphops.h"
#define TRACE 1

#if TRACE
#include <stdio.h>
#define Trace(x,y) if (graph_trace)\
   {printf x ;y}
#define Trace2(x,y) if (graph_trace >= 2)\
   {printf x ;y}
#define Trace3(x,y) if (graph_trace >= 3)\
   {printf x ;y}
#else
#define Trace(x,y)
#define Trace2(x,y)
#endif

Nodes sel1_node_nodes PL2(Node, x1, Nodes, x2)
{  
   return mk_one_nodes(x1);
}                           

int graph_trace = 0;

/* exported routines from graphops.c */

extern Graph insert1_graph
             PL3( Graph, g, Node, n, Node, n1 )
{
   Graph g1; Nodes succ;
   
   LET g1 = IF is_in_graph(n1,g)
            THEN g
            ELSE union1_graph(g,n1,mk_empty_nodes())  IN
   LET succ = sel_graph(g1,n)                          IN
      return
         ovwrt1_graph(g1,n,union1_nodes(succ,n1));
}         


extern Labeled_graph insert1_labeled_graph
             PL4( Labeled_graph, lg, Node, n, Node, n1, Currency, c) 
{
   Graph g0,g1; Labels l0,l1; Cost c0,c1,c2; 
   
   Trace3(("\ninsert1_labeled_graph\ngraph: "),pr_labeled_graph(lg);)
   Trace2(("nodes n and n1: "),{pr_node(n);pr_node(n1);})
   Trace2(("cost : "),pr_currency(c);); 
   
   LET g0 = s_gr(lg)                         IN
   LET l0 = s_la(lg)                         IN
   LET g1 = insert1_graph(g0,n,n1)           IN
   LET c0 = sel_labels(l0,n) IN
   LET c1 = mk_cost(heuristic(n1),s_g(c0)+c) IN
   LET l1 = IF is_in_dom_labels(n1,l0,&c2)
            THEN ovwrt1_labels(l0,n1,min_cost(c1,c2))
            ELSE union1_labels(l0,n1,c1)     IN
      return
         mk_labeled_graph(g1,l1);
}

extern Labeled_graph insert_labeled_graph
             PL3( Labeled_graph, lg, Node, n, Nodes, succ )
{
   Node n1; Nodes succ1; Currency c1; Labeled_graph lg1;
      
   if (is_empty_nodes(succ))
   {
      del1_nodes(succ);
      return lg;
   }
   else
      LET n1 = sel_nodes(succ)                      IN
      LET succ1 = sub1_nodes(succ,n1)               IN
      LET c1    = cost1_node(n,n1)                  IN
      LET lg1   = insert1_labeled_graph(lg,n,n1,c1) IN
         return
            insert_labeled_graph(lg1,n,succ1);
}



extern Node select_best_node PL3(Nodes, succ, Labels, l, Node, n0)
{
  Node n1,n2;
  
  if (is_empty_nodes(succ))
  {
     del1_nodes(succ);
     return n0;
  }
  LET n1 = sel_nodes(succ)                       IN
  LET n2 = IF is_better_cost(sel_labels(l,n1),
                             sel_labels(l,n0))
           THEN n1
           ELSE n0                               IN
     return
        select_best_node(sub1_nodes(succ,n1),l,n2);          
}

extern Node sel_best_node PL2(Nodes, succ, Labels, l)
{  
   Node n;
   
   LET n = sel_nodes(succ) IN
   return select_best_node(sub1_nodes(succ,n),l,n);
}
             
extern Goal generate_labeled_graph
             PL2( Labeled_graph, lg, Nodes, open_nodes )
{
   Graph g0; Labels l0; Node n; 
   Nodes open_nodes1, open_nodes2, succ;
   Labeled_graph lg2;
   
   static calls = 0;
   Trace(("\ngenerate_labeled_graph: %d. call\n ",++calls),{})
   Trace3(("graph: "),pr_labeled_graph(lg);)
   Trace(("open nodes: "),pr_nodes(open_nodes);)

   if (is_empty_nodes(open_nodes))
      return 
         mk_goal(mk_empty_nodes(),lg);
   
   LET g0 = s_gr(lg)                                   IN
   LET l0 = s_la(lg)                                   IN
   LET n  = sel_best_node(copy1_nodes(open_nodes),l0)  IN
      Trace(("currently best node: "),pr_node(n););
      
      if (is_solution_node(n))
         return 
            mk_goal(mk_one_nodes(n),lg);
      LET open_nodes1 = sub1_nodes(open_nodes,n)        IN
      LET succ        = generate_succ_nodes(n)          IN
      LET open_nodes2 = 
             union_nodes(open_nodes1,
                         diff_nodes(copy1_nodes(succ),
                                    dom_graph(g0)))     IN
      LET lg2         = insert_labeled_graph(lg,n,succ) IN
         return
            generate_labeled_graph(lg2,open_nodes2);                                
}
          
extern Goal solve_labeled_graph
             PL1( Labeled_graph, initlg )
{
   Trace(("\ninitial graph:\n")
        ,pr_labeled_graph(initlg);
        )

   return
      generate_labeled_graph(initlg
                            ,copy1_nodes(dom_graph(s_gr(initlg)))
                            );
}       


extern Currency min_currency PL2(Currency, x1, Currency, x2)
{
   return 
      IF x1 < x2 THEN x1 ELSE x2;
}
             
extern Cost min_cost PL2( Cost, x1, Cost, x2 )
{
   return
      mk_cost(min_currency(s_h(x1),s_h(x2)),
              min_currency(s_g(x1),s_g(x2)));
}
       
extern Bool is_better_cost PL2( Cost, x1, Cost, x2 )
{
   Currency s1,s2;
   
   LET s1 = s_g(x1) + s_h(x1) IN
   LET s2 = s_g(x2) + s_h(x2) IN
      return (s1 < s2);
}
       
