/* main program for water mug problem */

#include <stdio.h>
#include "func.mcr"
#include "graph.h"
#include "graphops.h"
#include "mugops.h"


int main PL2(int, argc, char **, argv)
{
   if (argc <= 1)
   {  
      /* usage */
      char *progname = "mug";
      if (argv[0] && *argv[0])
         progname= argv[0];
      fprintf(stderr,"usage : %s [ -h<n> -c<m> -t ]\n",progname);
      fprintf(stderr,"  -h<n>  : selection of heuristic\n%s%s%s%s",
                     "   <n>=0 : distance from solution in one mug\n",
                     "   <n>=1 : distance from solution in mug4\n",
                     "   <n>=2 : distance from solution in mug3\n",
                     "   <n>=3 : constant 0");
      fprintf(stderr,"  -c<m>  : selection of cost function\n%s%s%s%s",
                     "   <m>=0 : water costs\n",
                     "   <m>=1 : cost = path length\n",
                     "   <m>=2 : cost = 0\n",
                     "   <m>=3 : cost = water + path length\n");
      fprintf(stderr,"  -t     : trace\n");
      exit(1);
   }
   {
      /* process options */
      int i = 1;
      while (i < argc && *argv[i]=='-')
      { 
         char *opt = argv[i++];
         switch (*(opt+1))
         {  case 'h': select_heuristic_function = *(opt+2) - '0';
                      break;
            case 'c': select_cost_function = *(opt+2) - '0';
                      break;
            case 't': graph_trace = 1;
                      if (*(opt+2) >= '0')
                         graph_trace = *(opt+2) - '0';
                      break;
            default : fprintf(stderr,"unknown option %s ignored\n",opt);
         }      
      }
   }
   {
     printf("\nThe Water Mug Problem\n");
     printf("=====================\n\n");
     printf("Given 2 mugs of size 4 and 3 pints\n");
     printf("How do we get 2 pints into the large mug?\n");
     printf("The solution is constructed by constructing\n");
     printf("a search graph. This is done by the A*-Algorithm\n");
     printf("which tries to generate as few nodes as possible.\n");
     printf("The algorithm needs a heuritic function and a cost function\n");
     printf("for selecting the best node for extending the graph\n");
     printf("various heuristic and cost functions can be selected via command line options\n\n");
   }

   {  /* process graph construction */
   
      Goal g;
      
      printf("start searching for a solution\n");
      printf("==============================\n\n");
      g = solve_labeled_graph(init_labeled_graph());

      printf("the solution and the final graph are\n");
      printf("====================================\n\n");
      pr_goal(g);
   }

   return 0;
}
