#include "func.mcr"
#include "graph.h"
#include "graphops.h"
#include "mugops.h"

/* water mug problem :
   given two mugs with size 4 litre
                  and  size 3 litre
   How can we fill 2 litres of water in the 4 litre mug ?
 */


/* the initial graph */
extern Labeled_graph init_labeled_graph PL0()
{
   Node n0; Labeled_graph g0;

   LET n0 = mk_node(0,0) IN
   LET g0 =
      mk_labeled_graph
         (mk_one_graph(n0,mk_empty_nodes())
         ,mk_one_labels(n0,mk_cost(heuristic(n0),0))
         )
   IN 
      return g0;
}

/* the solution */
extern Bool is_solution_node PL1( Node, n )
{
   return s_m4(n) == 2;
}
       

/* the production system  */
extern Nodes generate_succ_nodes
             PL1 ( Node, n )
{
   Mug4 x; Mug3 y;
   
   LET x = s_m4(n) IN
   LET y = s_m3(n) IN
   return
   union_nodes(IF x<4 THEN mk_one_nodes(mk_node(4,y))
                      ELSE mk_empty_nodes(),
   union_nodes(IF y<3 THEN mk_one_nodes(mk_node(x,3))
                      ELSE mk_empty_nodes(),

   union_nodes(IF x>0 THEN mk_one_nodes(mk_node(0,y))
                      ELSE mk_empty_nodes(),

   union_nodes(IF y>0 THEN mk_one_nodes(mk_node(x,0))
                      ELSE mk_empty_nodes(),

   union_nodes(IF x+y>=4 && y>0
                      THEN mk_one_nodes(mk_node(4,x+y-4))
                      ELSE mk_empty_nodes(),

   union_nodes(IF x+y>=3 && x>0
                      THEN mk_one_nodes(mk_node(x+y-3,3))
                      ELSE mk_empty_nodes(),

   union_nodes(IF x+y<=4 && y>0
                      THEN mk_one_nodes(mk_node(x+y,0))
                      ELSE mk_empty_nodes(),

               IF x+y<=3 && x>0 
                      THEN mk_one_nodes(mk_node(0,x+y))
                      ELSE mk_empty_nodes()
              )))))));
}

 
 
static sel_cost PL3(Node, n, Node, n1, unsigned, cf)           
{
   Currency c,c1;
   switch (cf)
   { 
      case  0 : LET c  = s_m4(n ) + s_m3(n ) IN
                LET c1 = s_m4(n1) + s_m3(n1) IN
                   return
                      IF c1 > c THEN c1-c ELSE 0;
      case  1 : return 1;
      case  2 : return 0;
      case  3 : return sel_cost(n,n1,0) + sel_cost(n,n1,1);
      default : return sel_cost(n,n1,0);
      
   }
}

unsigned select_cost_function = 0;

extern Currency cost1_node
             PL2( Node, n, Node, n1 )
{
    return sel_cost(n,n1,select_cost_function);
} 

static Currency sel_heuristic
             PL2(Node, n, unsigned, hf)
{
   Mug4 m4; Mug3 m3;
   
   LET m4 = s_m4(n) IN
   LET m3 = s_m3(n) IN
   switch (hf)
   {
      case  1 : return IF m4>=2 THEN m4-2 ELSE 2-m4;
      case  2 : return IF m3>=2 THEN m3-2 ELSE 2-m3;
      case  3 : return 0;
      case  0 : return min_currency(sel_heuristic(n,1),
                                    sel_heuristic(n,2));
      default : return sel_heuristic(n,0);
   }   

}             

unsigned select_heuristic_function = 0;

extern Currency heuristic
             PL1(Node, n)
{
   return sel_heuristic(n,select_heuristic_function);
}             

