// Testprogramm TYPGEN=Bag, IMPL=automem
//
// Version 1.2, 17/11/93, Thorsten Tews
// ===========================================================================

#include "bag_vdm.hpp"
#include "bag_test.def"

// Beschreibung des Testmakros TEST:
//    - Parameter e : Ausdruck, der, falls Anweisungen ok, TRUE ergeben muss
//    - Parameter c1: Variable ve darf (YES) bzw. darf nicht (NO) durch den
//                    Ausdruck e veraendert werden
//    - Parameter c2: Variable v1 darf (YES) bzw. darf nicht (NO) durch den
//                    Ausdruck e veraendert werden
//    - Parameter c3: Variable v2 darf (YES) bzw. darf nicht (NO) durch den
//                    Ausdruck e veraendert werden

#define YES FALSE
#define NO  TRUE
#define TEST(e,c1,c2,c3) {\
                   BAG ve = BAG(empty), v1 = BAG(ELEM,2), v2 = BAG(ELEM);\
                   unsigned long err_code = 0;\
                   if ((e) != true) err_code |= 1;\
                   if (c1 && !(ve.is_eq(BAG(empty))))  err_code |= 2;\
                   if (c2 && !(v1.is_eq(BAG(ELEM,2)))) err_code |= 4;\
                   if (c3 && !(v2.is_eq(BAG(ELEM))))   err_code |= 8;\
                   if (err_code != 0)\
                      cout << " ### TEST FAILED: " #e ", error code: " << err_code << " ###\n";\
                   else\
                      cout << " Test passed : " #e "\n";\
                 }

// ---------------------------------------------------------------------------

void perform_test(void)
{
   cout << "Used objects in tests:\n" ;
   cout << "   ve ... the empty bag {}\n";
   cout << "   v1 ... the bag {ELEM,ELEM}\n";
   cout << "   v2 ... the bag {ELEM}\n\n";
   
   TEST(ve.is_empty(),NO,NO,NO)   // Praedikate, nicht zerstoerend !
   TEST(!v1.is_empty(),NO,NO,NO)   
   TEST(v1.is_eq(v1),NO,NO,NO)
   TEST(!ve.is_eq(v1),NO,NO,NO)
   TEST(v1.is_in(ELEM),NO,NO,NO)
   TEST(!ve.is_in(ELEM),NO,NO,NO)
   TEST(v2.is_subbag(v1),NO,NO,NO)
   TEST(!v2.is_subbag(ve),NO,NO,NO)

   TEST(v1.card() == 2,NO,NO,NO)   // Attribute, nicht zerstoerend !
   TEST(ve.card() == 0,NO,NO,NO)   
   TEST(v1.occurs(ELEM) == 2,NO,NO,NO)
   TEST(ve.occurs(ELEM) == 0,NO,NO,NO)
   
   TEST(v1.sel() == ELEM,NO,NO,NO)  // Selektor, nicht zerstoerend !  

   TEST((v1 - ELEM) == v2,NO,NO,NO)   // Kombinatoren, z.T. zerstoerend !
   TEST((v1 -= ELEM) == v2,NO,YES,NO)
   TEST(v1.sub1all(ELEM) == ve,NO,YES,NO)   
   TEST(v1.sub1n(ELEM,1) == v2,NO,YES,NO)   
   TEST((v2 | ELEM) == v1,NO,NO,NO)   
   TEST((v2 |= ELEM) == v1,NO,NO,YES)   
   TEST((ve.plus1n(ELEM,2)) == v1,YES,NO,NO)   
   TEST((v2 | v2) == v1,NO,NO,NO)   
   TEST((v2 |= v2) == v1,NO,NO,YES)   
   TEST((v1 - v2) == v2,NO,NO,NO)   
   TEST((v1 -= v2) == v2,NO,YES,NO)      
   TEST((v1 & v2) == v2,NO,NO,NO)   
   TEST((v1 &= v2) == v2,NO,YES,NO)   
   
   {
   TUPLE ts;
   TEST((ts = v1.cvtotupl()
        ,BAG(ELEM,2) == v1.cvfromtupl(ts)
        ),NO,NO,NO) // cv(to|from)tupl
   }
   cout << "\n--- Please check the following test results: ---\n\n";
   BAG ve = BAG(empty), v1 = BAG(ELEM,2), v2 = BAG(ELEM);
   cout << " fprint    : \n   ve = " << ve << "   v1 = " << v1 << "   v2 = " << v2;   
   cout << " v1.hash() : " << v1.hash() << "\n";
   cout << " finput    : \n   Please enter a value  v1 = ";
   cin  >> v1;
   cout << "   Value read by finput  v1 = " << v1;   
}

// ---------------------------------------------------------------------------
 
main()
{
   cout << "\n\n=== Test for TYPGEN: Bag, IMPL: Automem (17/11/93) ===\n\n";
   
   perform_test();
   profile_bag();
   profile_nset();
   profile_tuple();

   return 0;
}   
