// Testprogramm TYPGEN=Set/IMPL=automem
//
// Version 1.4, 17/11/93, Thorsten Tews
// ===========================================================================

#include "set_vdm.hpp"
#include "set_test.def"
#include "iostream.h"


// Beschreibung des Testmakros TEST:
//    - Parameter e : Ausdruck, der falls Anweisungen ok, TRUE ergeben mu
//    - Parameter c1: Variable ve darf (YES) bzw. darf nicht (NO) durch den
//                    Ausdruck e veraendert werden
//    - Parameter c2: Variable vo1 darf (YES) bzw. darf nicht (NO) durch den
//                    Ausdruck e veraendert werden
//    - Parameter c3: Variable vo2 darf (YES) bzw. darf nicht (NO) durch den
//                    Ausdruck e veraendert werden

#define YES FALSE
#define NO  TRUE
#define TEST(e,c1,c2,c3) {\
                   SDOMAIN ve=empty, vo1=(E_ONE), vo2=vo1;\
                   unsigned long err_code = 0;\
                   if ((e) != true) err_code |= 1;\
                   if (c1 && !(ve.is_eq(SDOMAIN(empty)))) err_code |= 2;\
                   if (c2 && !(vo1.is_eq(E_ONE))) err_code |= 4;\
                   if (c3 && !(vo2.is_eq(E_ONE))) err_code |= 8;\
                   if (err_code != 0)\
                      cout << " ### TEST FAILED: " #e ", error code : " << err_code << " ###\n";\
                   else\
                      cout << " Test passed : " #e "\n";\
                 }

// ---------------------------------------------------------------------------

void perform_test(void)
{
   cout << "Used objects in tests:\n" ;
   cout << "   ve  ... empty object\n";
   cout << "   vo1 ... object containing one element (E_ONE)\n";
   cout << "   vo2 ... object containing one element (E_ONE)\n\n";
 
   TEST(vo1.is_eq(vo2),NO,NO,NO)     // Praedikate, nicht zerstrend
   TEST(ve.is_empty(),NO,NO,NO)
   TEST(vo1.is_in(E_ONE),NO,NO,NO)
   TEST(vo1.is_subset(vo2),NO,NO,NO)
   
   TEST(vo1.card()==1,NO,NO,NO)
   TEST(E_ONE == vo1.sel(),NO,NO,NO)
   
   TEST((vo1-E_ONE).is_empty(),NO,NO,NO)      // sub1, nicht zerstoerend
   TEST((vo1-vo2).is_empty(),NO,NO,NO)        // diff,   -"-
   TEST((ve | vo1) == vo2,NO,NO,NO)           // union,  -"-
   TEST((ve | E_ONE) == vo1,NO,NO,NO)         // union1, -"-
   TEST((ve & vo1).is_empty(),NO,NO,NO)       // inters, -"-
   
   TEST((vo1 -= (E_ONE)).is_empty(),NO,YES,NO)  // sub1, zerstoerend
   TEST((vo1 -= vo2).is_empty(),NO,YES,NO)      // diff,   -"-
   TEST((ve |= vo1) == vo2,YES,NO,NO)           // union,  -"-
   TEST((ve |= E_ONE) == vo1,YES,NO,NO)         // union1, -"-
   TEST((vo1 &= ve).is_empty(),NO,YES,NO)       // inters, -"-
   
   {
   TUPLE ts;
   TEST((ts = vo1.cvtotupl()
        ,vo2 == vo1.cvfromtupl(ts)
        ),NO,NO,NO) // cv(to|from)tupl
   }

   cout << "\n--- Please check the following test results: ---\n\n";
   SDOMAIN vo1=E_ONE;

   cout << " fpr(vo1)   : " << vo1;
   cout << " vo1.hash() : " << vo1.hash() << "\n";
   cout << " finput     : \n Please enter a value, v = ";
   SDOMAIN v;
   cin  >> v;
   cout << " Value read by finput: v = " << v;
}

// ---------------------------------------------------------------------------
 
main()
{
   cout << "\n\n=== Test for TYPGEN:Set, IMPL:automem (17/11/93) ===\n\n";
   
   perform_test();
   
   profile_sdomain();
   profile_nset();
   profile_tuple();

}   
