// Testprogramm TYPGEN=Union, IMPL=automem
//
// Version 1.1, 17/11/93, Thorsten Tews
// ===========================================================================

#include "union_vdm.hpp"
#include "union_test.def"

// Beschreibung des Testmakros TEST:
//    - Parameter e  : Ausdruck, der, falls Anweisungen ok, TRUE ergeben muss
//    - Parameter c1 : Variable v1 darf (YES) bzw. darf nicht (NO) durch den
//                     Ausdruck e veraendert werden
//    - Parameter c11: Variable v11 darf (YES) bzw. darf nicht (NO) durch den
//                     Ausdruck e veraendert werden
//    - Parameter c2 : Variable v2 darf (YES) bzw. darf nicht (NO) durch den
//                     Ausdruck e veraendert werden

#define YES FALSE
#define NO  TRUE
#define TEST(e,c1,c11,c2) {\
   UNION v1 = UNION(ELEM1), v11 = UNION(ELEM11), v2 = UNION(ELEM2);\
   unsigned long err_code = 0;\
   if ((e) != true) err_code |= 1;\
   if (c1  && !(v1.is_eq(UNION(ELEM1)))) err_code |= 2;\
   if (c11 && !(v11.is_eq(UNION(ELEM11)))) err_code |= 4;\
   if (c2  && !(v2.is_eq(UNION(ELEM2)))) err_code |= 8;\
   if (err_code != 0)\
      cout << " ### TEST FAILED: " #e ", error code = " << err_code << " ###\n";\
   else\
      cout << " Test passed : " #e "\n";\
}

// ---------------------------------------------------------------------------

void perform_test(void)
{
   cout << "Used objects in tests:\n" ;
   cout << "   v1  ... the union mk_value1(ELEM1)\n"; 
   cout << "   v11 ... the union mk_value1(ELEM11) (ELEM11>ELEM1)\n";
   cout << "   v2  ... the union mk_value2(ELEM2)\n\n";
   
   TEST(v1 == v1,NO,NO,NO)      // Praedikate, nicht zerstoerend
   TEST(v1 != v2,NO,NO,NO)  
#if !defined(NO_GREATER)  
   TEST(v11 > v1,NO,NO,NO)  
   TEST(v2 > v1,NO,NO,NO)  
   TEST(v11 >= v1,NO,NO,NO)  
   TEST(v2 >= v1,NO,NO,NO)  
#endif   
   TEST(v1.is_value1(),NO,NO,NO)  
   TEST(v2.is_value2(),NO,NO,NO)  
  
   TEST(v1.value1() == ELEM1,NO,NO,NO)       // Attribute, nicht zerstoerend
   TEST(v2.value2() == ELEM2,NO,NO,NO)  
   TEST(v1.discr() == d_value1_union,NO,NO,NO)  
   TEST(v2.discr() == d_value2_union,NO,NO,NO)  
   
   TEST(v1.ovwrt_value2(ELEM2) == v2,YES,NO,NO)    // Modifikation, z.T. zerst.
   TEST(v2.ovwrt_value1(ELEM1) == v1,NO,NO,YES)  
   TEST(v1.chg_value1(ELEM11) == v11,YES,NO,NO)  
   TEST(v2.chg_value2(ELEM22) == UNION(ELEM22),NO,NO,YES)  
   TEST(v1.subst_value1(ELEM11) == v11,NO,NO,NO)  
   TEST(v2.subst_value2(ELEM22) == UNION(ELEM22),NO,NO,NO)  
   
   {
   TUPLE ts;
   TEST((ts = v1.cvtotupl()
        ,UNION(ELEM1) == v1.cvfromtupl(ts)
        ),NO,NO,NO) // cv(to|from)tupl
   }

   UNION v1 = UNION(ELEM1), v11 = UNION(ELEM11), v2 = UNION(ELEM2);   
   cout << "\n--- Please check the following test results: ---\n\n"; 
   cout << " fpr       :\n   v1  = " << v1 << "   v11 = " << v11 << "   v2  = " << v2;
   cout << " v1.hash() : " << v1.hash() << "\n";
   cout << " finput    : \n   Please enter a value  v1 = ";
   cin  >> v1;
   cout << "   Value read by finput  v1 = " << v1; 
    
}

// ---------------------------------------------------------------------------
 
main()
{
   cout << "\n\n=== Test for TYPGEN: Union, IMPL: Automem (10/11/93) ===\n\n";
   
   perform_test();
   profile_union();
   profile_nset();
   return 0;
}   
