/* $Id: btree_auto.cc,v 1.2 1994/06/12 16:35:41 vdm Exp $
* Testprogramm zur Performance-Analyse:
*   - Realisierung eines Binaeren Baumes mit Records und Optionalen
*
* Version 1.0 - 28/11/93
* ========================================================================= 
*/

#include "btree_vdm.hpp"
#include <time.h> 
#include <stdlib.h>

// ---------------------------------------------------------------------------

Bool exists(BTREE b, NAT0 key)
{
   while (!b.is_nil())
   {
      NODE node = b.s();
      NAT0 info = node.info();
      
      if (key == info) return TRUE;
      if (key > info)
         b = node.right();
      else
         b = node.left();
   }
   return FALSE;
}

BTREE insert(BTREE b, NAT0 key)
{
   if (b.is_nil()) return BTREE(NODE(nil,nil,key));
 
   NODE node = b.s();
   NAT0 info = node.info();
    
   if (key == info) return b;
   if (key > info)
      return BTREE(node.subst_right(insert(node.right(),key)));
   else
      return BTREE(node.subst_left(insert(node.left(),key)));
}

BTREE convert_from_set(NSET nset)
{
   BTREE b = BTREE(nil);
   
   while (!nset.is_empty())
   {
      NAT0 info = nset.sel();
      nset -= info;
      b = insert(b,info);
   }
   return b;
   
}

// ---------------------------------------------------------------------------

void btree_test(void)
{
   // Einen Nset mit 2n+1 Elementen erzeugen
   srand(42424242);
   NSET nset = NSET(empty);
   for (Nat0 i=1; i<=50; i++)
   {
      nset |= NAT0((Nat0)(rand() & 0xFF));
      nset |= NAT0((Nat0)(512 + (rand() & 0xFF)));
   }
   nset |= NAT0(256);
      
   // Daraus eines Btree machen
   BTREE btree = convert_from_set(nset);   
   
   // Prfen, ob alle Elemente aus dem Set auch im Baum vorhanden sind
   while (!nset.is_empty())
   {
      NAT0 info = nset.sel();
      nset -= info;
      if (!exists(btree,info))
         cout << "Fehler: Element " << info << " nicht gefunden\n";
   }
}

// ---------------------------------------------------------------------------

main()
{
//   cout << "\n\n=== Realisierung eines binaeren Baumes (automem) ===\n\n";
   
   clock_t start_clocks, end_clocks;
   start_clocks = clock();
   for (int i = 0; i<100; i++) btree_test();   
   end_clocks = clock();
   cout << " " << end_clocks - start_clocks << " ";
  
//   profile_node();
//   profile_nset();
}
