/* $Id: btree_man.cc,v 1.2 1994/06/12 16:35:42 vdm Exp $
* Testprogramm zur Performance-Analyse:
*   - Realisierung eines Binaeren Baumes mit Records und Optionalen
*   - Manuelle Version
* Version 1.0 - 28/11/93
* ========================================================================= 
*/

#include "btree_mvdm.hpp"
#include <time.h> 
#include <stdlib.h>

// ---------------------------------------------------------------------------

Bool exists(BTREE b, NAT0 key)
{
   while (!b.is_nil())
   {
      NODE node = b.s();
      NAT0 info = node.info();
      
      if (key == info) return TRUE;
      if (key > info)
         b = node.right();
      else
         b = node.left();
   }
   return FALSE;
}

// der Originalbaum wird in dieser Operation zerstoert !
BTREE insert(BTREE b, NAT0 key)
{
   if (b.is_nil()) return BTREE(NODE(nil,nil,key));
 
   NODE node = b.s();
   NAT0 info = node.info();
    
   if (key == info) return b;
   NODE n2;
   if (key > info)
      n2 = node.subst_right(insert(node.right(),key));
   else
      n2 = node.subst_left(insert(node.left(),key));
   node.del1();  // subst macht nur ein copy1 !!!!!
   return BTREE(n2);
}

// diese Operation zerstoert den Parameter nset !
BTREE convert_from_set(NSET nset)
{
   BTREE b = BTREE(nil);
   
   while (!nset.is_empty())
   {
      NAT0 info = nset.sel();
      nset = nset - info;
      b = insert(b,info); // Kopie von info nicht notwendig, da elemenaterer Typ
      // nicht notwendig, da hier elementarer Typ fuer info: info.del();        
   }
   return b;
   
}

// ---------------------------------------------------------------------------

void btree_test(void)
{
   // Einen Nset mit 2n+1 Elementen erzeugen
   srand(42424242);
   NSET nset = NSET(empty);
   for (Nat0 i=1; i<=50; i++)
   {
      nset = nset | NAT0((Nat0)(rand() & 0xFF));
      nset = nset | NAT0((Nat0)(512 + (rand() & 0xFF)));
   }
   nset = nset | NAT0(256);
      
   // Daraus eines Btree machen
   NSET s = nset.copy1();
   BTREE btree = convert_from_set(s); // s wird hier geloescht !
   
   // Prfen, ob alle Elemente aus dem Set auch im Baum vorhanden sind
   while (!nset.is_empty())
   {
      NAT0 info = nset.sel();
      nset = nset - info;
      if (!exists(btree,info))
         cout << "Fehler: Element " << info << " nicht gefunden\n";
      // nicht notwendig, da hier elementarer Typ fuer info: info.del();        
   }
   
   nset.del();
   btree.del();
}

// ---------------------------------------------------------------------------

main()
{
//   cout << "\n\n=== Realisierung eines binaeren Baumes (automem) ===\n\n";
   
   clock_t start_clocks, end_clocks;
   start_clocks = clock();
   for (int i = 0; i<100; i++) btree_test();   
   end_clocks = clock();
   cout << " " << end_clocks - start_clocks << " ";
  
//   profile_node();
//   profile_nset();
}
