/* $Id: cp_man2.cc,v 1.1 1994/02/24 11:37:41 vdm Exp $
* Testprogramm zur Performance-Analyse:
*   - Berechnung der Coursest Partioning einer Menge von Mengen
*     (tail-recursion in Schleife verwandelt, keine Optimierung)
*   - manuelle Speicherverwaltung
*
* Version 1.0 - 24/11/93
* ========================================================================= 
*/

#include "cp_mvdm.hpp"
#include <time.h>

// Achtung: Parameter b wird in dieser Routine mit del1() zerstoert !!!
//          -> Die Elemente der Liste b werden auch noch in der aufrufenden
//             Routine 'benutzt' ! Sie duerfen also NICHT geloescht werden !
// 
ASET find_partner(BSET& b, ASET& a)
{
   while (!b.is_empty())
   {
      ASET ret_val = b.sel();
      b = b - ret_val;
      ASET tmp = a.copy1();
      tmp = tmp & (ret_val.copy1()); // unbounded copy destroyed by inters
      if (!tmp.is_empty())
      {
         tmp.del1();
         b.del1();
         return ret_val;
      }
      tmp.del1();
   }
   b.del1();
   return ASET(empty);
}

// der Parameter b wird in dieser Routine zerstoert ! 
BSET coursest_part(BSET& b)
{
   ASET x,y;  

   while (1)     
   {
      Bool found = FALSE;

      BSET tmp_b1 = b.copy1();
      while (!found && !(tmp_b1.is_empty()))
      {
         x = tmp_b1.sel();
         tmp_b1 = tmp_b1 -x ;
         BSET tmp_b2 = tmp_b1.copy1();
         y = find_partner(tmp_b2,x);   // tmp_b2 wird voellig geloescht !
         found = !y.is_empty();
      }
      tmp_b1.del1();
   
      if (!found) break;
   
      b = b - x;
      b = b - y;
      ASET tmp_x = x.copy1(), tmp_y = y.copy1();
      b = b | (tmp_x & tmp_y);  // tmp_x,tmp_y werden beide impl. geloescht !
      tmp_x = x.copy1();
      x = x - y;
      y = y - tmp_x;
      tmp_x.del1();
   
      if (!x.is_empty()) b = b | x;
      if (!y.is_empty()) b = b | y;
   }
   return b;
}

void run_cp(void)
{
    ASET a = NAT0(1); a = a | NAT0(3);
    a = a | NAT0(5);  a = a | NAT0(7);
    a = a | NAT0(9);  a = a | NAT0(11);
    a = a | NAT0(13); a = a | NAT0(15);
    BSET b = BSET(a);
    
    a = NAT0(2);      a = a | NAT0(4);
    a = a | NAT0(6);  a = a | NAT0(8);
    a = a | NAT0(10); a = a | NAT0(12);
    a = a | NAT0(14); a = a | NAT0(16);
    b = b | a;

    a = NAT0(1);     a = a | NAT0(2);
    a = a | NAT0(3); a = a | NAT0(4);
    a = a | NAT0(5); a = a | NAT0(6);
    a = a | NAT0(7); a = a | NAT0(8);
    b = b | a;
    
    a = NAT0(2);     a = a | NAT0(11);
    a = a | NAT0(3); a = a | NAT0(13);
    a = a | NAT0(5); a = a | NAT0(17);
    a = a | NAT0(7); a = a | NAT0(19);
    b = b | a;    
  
//    cout << "Ergebnis : " <<  coursest_part(b);
    b = coursest_part(b);
    b.del();
}

// ---------------------------------------------------------------------------

main()
{
//   cout << "\n\n=== Berechnung der Coursest Partioning (manuell) ===\n\n";

  clock_t start_clocks, end_clocks;
   
   start_clocks = clock();
   for (int i=0; i<1500; i++)
      run_cp();
   end_clocks = clock();
   cout << " " << end_clocks - start_clocks << " ";      

//	profile_aset();
// only for automem implementation: profile_rep_aset();
//	profile_bset();
// only for automem implementation:	profile_rep_bset();   
}
