/* $Id: prof_auto.cc,v 1.2 1994/06/12 16:35:45 vdm Exp $
* Testprogramm zur Performance-Analyse:
*   - Teilrealsierung eines Profilers
*   - automatische Speicherverwaltung
*
* Format der Eingabedatei:   1.  Call-Info = {Line}
*                            2.  Line      = C-String C-String
*
* Version 1.0 - 28/11/93
* ========================================================================= 
*/

#include "prof_vdm.hpp"
#include <time.h>
#include <stdlib.h>
#include <fstream.h>

// ---------------------------------------------------------------------------

class TProfiler {
   private: 
      TCALLS calls;
      TPROCS procs;
      void insert(STR caller, STR callee);
      NAT0 cnt_calls(TPROCID pid);
      void fatal_error(char const * msg);

   public:
      TProfiler() {}
      ~TProfiler() {}   
      void read_calling_info(char const * filename);
      void write_frequency_info(char const * filename);
};

void TProfiler::fatal_error(char const * msg)
{
   cerr << "\nTProfile - FATAL ERROR : " << msg << "\n";
   exit(EXIT_FAILURE);
}

void TProfiler::insert(STR caller, STR callee)
{
   TPROCID caller_id, callee_id;
   
   if (!procs.is_in_rng(caller,caller_id))
   {
      caller_id = TPROCID(new_val); procs.plus1(caller_id,caller);
   }
   if (!procs.is_in_rng(callee,callee_id))
   {
      callee_id = TPROCID(new_val); procs.plus1(callee_id,callee);
   }   
   
   TCALLED called_procs;
   if (!calls.is_in_dom(caller_id,called_procs))
      calls.plus1(caller_id,TCALLED(callee_id,1));
   else
   {
      NAT1 cnt;
      if (!called_procs.is_in_dom(callee_id,cnt))
         called_procs.plus1(callee_id,1);
      else
         called_procs.ovwrt1(callee_id,cnt+1);
      calls.ovwrt1(caller_id,called_procs);   
   }   
}

void TProfiler::read_calling_info(char const * filename)
{
   char caller_c[255], callee_c[255];
   
   calls = TCALLS(empty);
   procs = TPROCS(empty);

   ifstream in_file(filename);
   if (!in_file) fatal_error("Can't create input file");
   
   while (!in_file.eof())
   {
      in_file >> caller_c >> callee_c;
      if (!in_file.eof()) insert(STR(caller_c),STR(callee_c));
   }
}

NAT0 TProfiler::cnt_calls(TPROCID pid)
{
   NAT0 total_cnt = 0; // whow !
   NAT1 cnt;
   
   // Hier waere ein Iterator sinnvoll, man koennte dann die Zerstoerung
   // der Map the_calls vermeiden !
   TCALLS lcalls = calls;
   while (!lcalls.is_empty())
   {
      TPROCID p = lcalls.sel_dom();
      TCALLED called_procs = lcalls.sel(p);
      if (called_procs.is_in_dom(pid,cnt)) total_cnt = total_cnt + cnt;
      lcalls -= p;
   }
   
   return total_cnt;
}

void TProfiler::write_frequency_info(char const * filename)
{
   ofstream out_file(filename);
   if (!out_file) fatal_error("Can't create output file");
   
   TPROCS lprocs = procs;
   // Hier waere ein Iterator sinnvoll, man koennte dann die Zerstoerung
   // der Map procs vermeiden !   
   while (!lprocs.is_empty())
   {
      TPROCID pid = lprocs.sel_dom();
      out_file << "Procedure " << lprocs.sel(pid) << " ... is called ";
      out_file << cnt_calls(pid) << " time(s)\n";
      lprocs -= pid;
   }
}

// ---------------------------------------------------------------------------

void test_profiler(void)
{
   TProfiler my_prof;
   my_prof.read_calling_info("call_info.prof");
   my_prof.write_frequency_info("freq_info.prof");
}

// ---------------------------------------------------------------------------

main()
{
/*
   cout << "\n\n=== Ein einfacher Profiler (automem) ===\n\n";
   
   profile_tprocs();
   profile_tcalls();
   profile_tcalled();
*/

   clock_t start_clock, end_clock; 
   start_clock=clock();
   for (int i=0; i<100; i++) test_profiler();
   end_clock = clock();
   cout << end_clock - start_clock << " ";

   return 0;
}
