// $Id: spec2_test.cc,v 1.1 1994/02/24 11:37:48 vdm Exp $
// Testprogram: spec2_test
//
// Aufgabe : Ueberpruefen, ob Implementation mit Spezifikation uebereinstimmt
//           Hierzu wird der State 'Program' aus den aktuellen Strukturen
//           im Speicher ermittelt und ausgegeben und kann mit den 
//           erwarteten Werten verglichen werden.
//
// Hinweise : - Die eigenen Test sollten innerhalb der Funktion test_proc()
//              stattfinden              
//            - Ausgabe des State zu beliebigen Zeitpunkt mit 
//              show_state_program
//            - Fuer den Token Var wird der Zeiger this des entsprechenden
//              Objektes verwendet; zur besseren Lesbarkeit wird der 
//              Wert mod 10000 ausgegeben !. Dies funktioniert unter Linux 
//              V0.99.9/g++ 2.3.3 hervorragend (sonst alle Objekte im 3 oder 
//              4 Gigabyte), mu fuer andere System aber evtl. geaendert werden !
//            - Fuer den Token Handle wird die Adresse des Handle_Struct im 
//              Heap verwendet
//
// Version : 2.2/07-02-94
// ===========================================================================


#include "spec2.hpp"

// --- globale Variablen/Konstanten: -----------------------------------------

#define Var    Nat0
#define Handle Nat0
#define VAR    NAT0
#define HANDLE NAT0

VARIABLES the_variables = empty;

// ---------------------------------------------------------------------------
 
class CHECKED_ADT {
public: 
   virtual VAR get_the_var(void)         = 0;
   virtual HANDLE get_the_handle(void)   = 0;
   virtual ADTS get_the_local_adts(void) = 0;
};

// ---------------------------------------------------------------------------

void show_state_program(void)
{
   DATEN     data = DATEN(empty);
   ADTS      adts = ADTS(empty);
   VARIABLES temp = the_variables;

	while (!temp.is_empty())
	{
	   NAT0 a_var = temp.sel();
	   CHECKED_ADT * ptr_var = (CHECKED_ADT *) Nat0(a_var);
	   data.plus1( Nat0(a_var) % 10000, ptr_var->get_the_handle());
	   if (ptr_var->get_the_handle() != 0 /* nil */ )
	      adts |= ptr_var->get_the_local_adts();
	      
	   temp -= a_var;
	}   
	cout << "State 'program' is : \n";
	cout << "\tdata = " << data << "\tadts = " << adts;
	cout << "\n";
}

// ---------------------------------------------------------------------------

class CHECKED_NBAG : public CHECKED_ADT, public NBAG {
public:
   CHECKED_NBAG() // constructor NBAG called implicitly
      { the_variables.plus1(VAR( (Var)this ));}
   CHECKED_NBAG(NILETC& x) : NBAG(x) 
      { the_variables.plus1(VAR( (Var)this ));}
   CHECKED_NBAG(NAT0& x) : NBAG(x) 
      { the_variables.plus1(VAR( (Var)this ));}
   CHECKED_NBAG(const CHECKED_NBAG& x) : NBAG(x)
      { the_variables.plus1(VAR( (Var)this ));}   
   ~CHECKED_NBAG()
      { the_variables.sub1(VAR( (Var)this ));}

   CHECKED_NBAG& operator=(const NBAG& x1)
   { 
      check_handle_nbag(x1.data); 
      (x1.data)->f_rc++;
      if (data!=0) del_nbag(data);
      data = x1.data;
      return *this;
   }
      
   virtual VAR get_the_var(void) { return VAR( (Var)this );}   
   virtual HANDLE get_the_handle(void) { return NAT0(get_handle());}
   virtual ADTS get_the_local_adts(void) { return get_gc_info(); }
};

class CHECKED_BBAG : public CHECKED_ADT, public BBAG{
public:
   CHECKED_BBAG()  // constructor BBAG called implicitly 
      { the_variables.plus1(VAR( (Var)this ));}
   CHECKED_BBAG(NILETC& x) : BBAG(x) 
      { the_variables.plus1(VAR( (Var)this ));}
   CHECKED_BBAG(CHECKED_NBAG& x) : BBAG(x) 
      { the_variables.plus1(VAR( (Var)this ));}
   CHECKED_BBAG(const CHECKED_BBAG& x) : BBAG(x)
      { the_variables.plus1(VAR( (Var)this ));}   
   ~CHECKED_BBAG()
      { the_variables.sub1(VAR( (Var)this ) );}

   CHECKED_BBAG& operator=(const BBAG& x1)
   { 
      check_handle_bbag(x1.data); 
      (x1.data)->f_rc++;
      if (data!=0) del_bbag(data);
      data = x1.data;
      return *this;
   }

   virtual VAR get_the_var(void) { return VAR( (Var)this );}   
   virtual HANDLE get_the_handle(void) { return NAT0(get_handle());}
   virtual ADTS get_the_local_adts(void) { return get_gc_info(); }
};

// --- Testfunktionen: -----------------------------------------------------

void test_prog(void)
{
   cout << "=== Konstruktor:  n1.NBAG(42)  entspr.  constructor(n1,{})\n";
   CHECKED_NBAG n1 = CHECKED_NBAG(NAT0(42)); 
   // Erwartetes Ergebnis: 
   // data = [n1 -> h1] adts = [h1 -> (1,{})] 
   cout << "Neues Objekt erzeugt n1 = " << Nat0(n1.get_the_var()) %10000 << "\n";
   show_state_program();  
   
   cout << "=== Konstruktor:  n2.NBAG(4242)  entspr.  constructor(n2,{})\n";
   CHECKED_NBAG n2 = CHECKED_NBAG(NAT0(4242));
   cout << "Neues Objekt erzeugt n2 = " << Nat0(n2.get_the_var()) %10000 << "\n";
   cout << "Erwartetes Ergebnis:\n";
   cout << "\tdata = [n1 -> h1, n2 -> h2]\n";
   cout << "\tadts = [h1 -> (1,{}), h2 -> (1,{})]\n";
   show_state_program();  
   
   {
   	cout << "=== Konstruktor: s1.BBAG(n1) entspr. constructor(s1,{n1})\n";
   	CHECKED_BBAG s1 = CHECKED_BBAG(n1);
   	cout << "Neues Objekt erzeugt s1 = " << Nat0(s1.get_the_var()) %10000 << "\n";
   	cout << "Erwartetes Ergebnis:\n";
   	cout << "\tdata = [n1 -> h1, n2 -> h2, s1 -> h3]\n";
   	cout << "\tadts = [h1 -> (2,{}), h2 -> (1,{}), h3-> (1,{h1})]\n";
      show_state_program();  
   	{
   		cout << "=== Kopierkonstruktor s2.BBAG(const& s1) entspr. copy-constr(s2,s1) \n";
   		CHECKED_BBAG s2 = s1;
	      // data = [n1->h1, n2->h2, s1->h3, s2->h3]
	      // adts = [h1->(2,{}), h2->(1,{}), h3->(2,{h1})]
			cout << "Neues Objekt erzeugt s2 = " << Nat0(s2.get_the_var()) %10000 << "\n";
			cout << "Erwartetes Ergebnis:\n";
   		cout << "\tdata = [n1 -> h1, n2 -> h2, s1 -> h3, s2 -> h3]\n";
   		cout << "\tadts = [h1 -> (2,{}), h2 -> (1,{}), h3-> (2,{h1})]\n";
      	show_state_program();  	      
   		
   		cout << "=== Attribut   s2.card() entspr. attribut({s2})\n";
   		s2.card();
   		cout << "Erwartetes Ergebnis:\n";
   		cout << "\tdata = [n1 -> h1, n2 -> h2, s1 -> h3, s2 -> h3]\n";
   		cout << "\tadts = [h1 -> (2,{}), h2 -> (1,{}), h3-> (2,{h1})]\n";
      	show_state_program();  

			cout << "=== Selektor   s2.sel() entspr. selector(temp,s2), destructor(temp)\n";
   		s2.sel();
   		cout << "Erwartetes Ergebnis:\n";
   		cout << "\tdata = [n1 -> h1, n2 -> h2, s1 -> h3, s2 -> h3]\n";
   		cout << "\tadts = [h1 -> (2,{}), h2 -> (1,{}), h3-> (2,{h1})]\n";
      	show_state_program();  		

   		cout << "===  Kombinator 1: update_D, einfuegen, cond_clone notwendig\n";
   		cout << "     s1.plus1(n2) entspr. update(s1,{n2},union1-bag-op)\n";
   		//   union1-bag-op(hdl,elems,adts) = (elems,{}) (nur fuer diesen Spezialfall!)
   		s1.plus1(n2);
   		cout << "Erwartetes Ergebnis:\n";
   		cout << "\tdata = [n1->h1, n2->h2, s1->h4, s2->h3]\n";
   		cout << "\tadts = [h1->(3,{}), h2->(2,{}), h3->(1,{h1}), h4->(1,{h1,h2})]\n";
      	show_state_program();  	      

   		cout << "=== Kombinator 2: update_D, entfernen, kein cond_clone notwendig\n";
   		cout << "    s1.sub1(n2)  entspr. update(s1,{n2},sub1-bag-op)\n";
   		//                      sub1-bag-op(hdl,elems,adts) = ({},elems) (nur fuer diesen Spezialfall !)
   		s1.sub1(n2);
   		cout << "Erwartetes Ergebnis:\n";
   		cout << "\tdata = [n1->h1, n2->h2, s1->h4, s2->h3]\n";
   		cout << "\tadts = [h1->(3,{}), h2->(1,{}), h3->(1,{h1}), h4->(1,{h1})]\n";
      	show_state_program();  	      
   		
   		cout << "=== Kombinator 3: update_DD, kein cond_clone notwendig\n";
   		cout << "    s1.plus(s1)  entspr. update(s1,{s1},union-bag-op)\n";
   		//  union-bag-op(hdl,elems,adts) = (union-bag {adts(x).elems | x e elems},{})
    		//     ( nur fuer diesen Spezialfall !)
   		s1.plus(s1);
   		cout << "Erwartetes Ergebnis:\n";
   		cout << "\tdata = [n1->h1, n2->h2, s1->h4, s2->h3]\n";
   		cout << "\tadts = [h1->(4,{}), h2->(1,{}), h3->(1,{h1}), h4->(1,{h1,h1})]\n";
      	show_state_program();  	      
   		
   		{
   			cout << "=== Kopierkonstruktor: s3.BBAG(const& s2) entspr. copy-constr(s3,s2)\n";
   			CHECKED_BBAG s3 = s2;
		      cout << "Neues Objekt erzeugt s3 = " << Nat0(s3.get_the_var()) %10000 << "\n";
		      cout << "Erwartetes Ergebnis:\n";
	   		cout << "\tdata = [n1->h1, n2->h2, s1->h4, s2->h3, s3->h3]\n";
   			cout << "\tadts = [h1->(4,{}), h2->(1,{}), h3->(2,{h1}), h4->(1,{h1,h1})]\n";
	      	show_state_program();  
	      
   			cout << "=== Zuweisung: s3 = s1 enstpr. assign(s3,s1)\n";
   			s3 = s1;
   			cout << "Erwartetes Ergebnis:\n";
	   		cout << "\tdata = [n1->h1, n2->h2, s1->h4, s2->h3, s3->h4]\n";
   			cout << "\tadts = [h1->(4,{}), h2->(1,{}), h3->(1,{h1}), h4->(2,{h1,h1})]\n";
   			show_state_program();  
   			
   		}	cout << "=== Destructor s3.~BBAG() entspr. destructor(s3) \n";
            cout << "Erwartetes Ergebnis:\n";
	   		cout << "\tdata = [n1->h1, n2->h2, s1->h4, s2->h3]\n";
   			cout << "\tadts = [h1->(4,{}), h2->(1,{}), h3->(1,{h1}), h4->(1,{h1,h1})]\n";  		
   			// Erwartetes Ergebnis:
		      // data = [n1->h1, n2->h2, s1->h4, s2->h3]
		      // adts = [h1->(4,{}), h2->(1,{}), h3->(1,{h1}), h4->(1,{h1,h1})]
		      show_state_program();  

   	} 	cout << "=== Destructor s2.~BBAG() entspr. destructor(s2)\n";
			cout << "Erwartetes Ergebnis:\n";
	   	cout << "\tdata = [n1->h1, n2->h2, s1->h4]\n";
   		cout << "\tadts = [h1->(3,{}), h2->(1,{}), h4->(1,{h1,h1})]\n";   	
		   show_state_program();  

   } 	cout << "=== Destructor s1.~BBAG() entspr. destructor(s1
   )\n";
   	cout << "Erwartetes Ergebnis:\n";
	  	cout << "\tdata = [n1->h1, n2->h2]\n";
   	cout << "\tadts = [h1->(1,{}), h2->(1,{})]\n";
	   show_state_program();  
}

// -------------------------------------------------------------------------

main()
{
   cout << "(Init) "; show_state_program();  

	test_prog();

   cout << "(End) "; show_state_program();  

   profile_adts();
   profile_handles();
   profile_adt();
   profile_daten();
   
   return 0;
}
