
//                                                             Thorsten Tews
//                                                             V1.3 (17/11/93)
// ===========================================================================

#include "auto_vdm.hpp"

void test_operator(void)
{
   cout << "--- Test der Operatoren (Infix,Zuweisung) ---\n\n";
   
   BSET b1 = empty, b2 = empty;
   for (Nat0 i=1; i<=10; i++)
   {
      b1 |= ASET(NAT0(i));
      b2 |= ASET(NAT0(i+5));
   }
   
   cout << "1. Nicht zerstrende Operatoren:\n";
   cout << "  b1 vorher: " << b1;
   cout << "  b2 vorher: " << b2;
   cout << "  Differenzmenge    b1 - b2 = " << (b1-b2);
   cout << "  Schnittmenge      b1 & b2 = " << (b1 & b2);
   cout << "  Vereinigungsmenge b1 | b2 = " << (b1 | b2);
   cout << "  Menge   (b1-b2) | (b2-b1) = " << ((b1-b2)|(b2-b1));
   cout << "  b1 nachher: " << b1;
   cout << "  b2 nachher: " << b2;
   
   cout << "\n2. Zerstrende Operationen:\n";
   cout << "  Vereinigungsmenge b1 |= b2 = " << (b1 |= b2);
   cout << "  Differenzmenge    b1 -= b2 = " << (b1 -= b2);
   cout << "  Schnittmenge      b1 &= b2 = " << (b1 &= b2);
   cout << "  b1 nachher: " << b1;
   cout << "  b2 nachher: " << b2;   
}

// ---------------------------------------------------------------------------

void test_selector(void)
{
   cout << "\n\n--- Selektoren ---\n\n";
    
   ASET a  = NAT0(1); a |= NAT0(2); a |= NAT0(3);
   BSET b1 = a;
   a  = NAT0(4); a |= NAT0(5); a |= NAT0(6);
   b1 |= a;
   a  = NAT0(7); a |= NAT0(8); a |= NAT0(9);
   b1 |= a;   
    
   cout << "b1 : " << b1;
   BSET b2 = empty;

   while (!b1.is_empty())
   {
      b2 |= b1.sel();
      b1 -= b1.sel();
      cout << "#b1 : " << long(b1.card()) << " #b2 : " << long(b2.card()) << "\n";
   }
   cout << "b1 : " << b1;
   cout << "b2 : " << b2;
   b1 |= b2;
   cout << "b1 : " << b1;
   cout << "b2 : " << b2;   
}   

// ---------

void proc_destroy(BSET b)
{
   while (!b.is_empty())
   {
      cout << "Removing : " << b.sel();
      b -= b.sel();
   }
}

void proc_harmless(BSET b)
{
   cout << "Laenge des Parameters b: " << long(b.card()) << "\n";
}

BSET func(void)
{
   ASET a = NAT0(12345678);
   BSET b = BSET(a);
   
   return b;
}

void test_functions(void)
{
   cout << "\n\n--- Prozeduren und Funktionen ---\n\n";
   
   ASET a  = NAT0(1); a |= NAT0(2); a |= NAT0(3);
   BSET b1 = a;
   
   a  = NAT0(4); a |= NAT0(5); a |= NAT0(6);
   b1 |= a;
   
   a  = NAT0(7); a |= NAT0(8); a |= NAT0(9);
   b1 |= a;   
   
   cout << "b1 vorher: " << b1;
   proc_destroy(b1);
   cout << "b1 nach Funktion proc_destroy  : " << b1;
   proc_harmless(b1);
   cout << "b1 nach Funktion proc_harmless : " << b1;   
   
   BSET b2 = func();
   cout << " b2 = func() -> b2 : " << b2;   
   
   proc_destroy(func()); // What's going on here ?
}

ASET find_partner(BSET b, ASET a)
{
   if (b.is_empty()) return ASET(empty);
   
   ASET ret_val = b.sel();
   b -= ret_val;
   if (!(ret_val & a).is_empty()) 
      return ret_val;
   return find_partner(b,a);
}

BSET coursest_part(BSET S)
{
   Bool found = FALSE;
   ASET x,y;
   
   BSET tmp = S;       
   while (!found && !tmp.is_empty())
   {
      x = tmp.sel();
      tmp -= x;
      y = find_partner(tmp,x);
      found = !y.is_empty();
   }
   if (!found) return S;
   
   S -= x;
   S -= y;
   S |= (x & y); 
   if (!(x-y).is_empty())  S |= (x-y); 
   if (!(y-x).is_empty())  S |= (y-x);
   
   return coursest_part(S);
}

void test_usage(void)
{
    cout << "\n\n--- Coursest Partioning ---\n\n";
    
    ASET a = NAT0(1); a |= NAT0(3);
    a |= NAT0(5); a |= NAT0(7);
    a |= NAT0(9); a |= NAT0(11);
    a |= NAT0(13);a |= NAT0(15);
    BSET b = BSET(a);
    
    a = NAT0(2);  a |= NAT0(4);
    a |= NAT0(6);  a |= NAT0(8);
    a |= NAT0(10); a |= NAT0(12);
    a |= NAT0(14); a |= NAT0(16);
    b |= a;
    
    a = NAT0(1); a |= NAT0(2);
    a |= NAT0(3); a |= NAT0(4);
    a |= NAT0(5); a |= NAT0(6);
    a |= NAT0(7); a |= NAT0(8);
    b |= (a);
    
    a = NAT0(2);  a |= NAT0(11);
    a |= NAT0(3); a |= NAT0(13);
    a |= NAT0(5); a |= NAT0(17);
    a |= NAT0(7); a |= NAT0(19);
    b |= a;    
    
    cout << "Menge b vor Zerlegung: " << b;
    cout << "Ergebnis : " << coursest_part(b);
    cout << "Menge b nach Zerlegung: " << b;
}


/* void test_functional(void)
{
    cout << "\n\n--- Verwendung von Funktionalen ---\n\n";

    ASET a = NAT0(1); a |= NAT0(3);
    a |= NAT0(5); a |= NAT0(7);
    a |= NAT0(9); a |= NAT0(11);
    a |= NAT0(13);a |= NAT0(15);
    BSET b1 = BSET(a);
    
    a = NAT0(2);  a |= NAT0(4);
    a |= NAT0(6);  a |= NAT0(8);
    a |= NAT0(10); a |= NAT0(12);
    a |= NAT0(14); a |= NAT0(16);
    b1 |= a;
    
    a = NAT0(1); a |= NAT0(2);
    a |= NAT0(3); a |= NAT0(4);
    a |= NAT0(5); a |= NAT0(6);
    a |= NAT0(7); a |= NAT0(8);
    BSET b2 = BSET(a);
    
    a = NAT0(2);  a |= NAT0(11);
    a |= NAT0(3); a |= NAT0(13);
    a |= NAT0(5); a |= NAT0(17);
    a |= NAT0(7); a |= NAT0(19);
    b2 |= a;     
    
    cout << "b1 vorher : " << b1;
    cout << "b2 vorher : " << b2;
    
    cout << "... calling b1.my_union(b1) ...\n"; b1.my_union(b1);
    cout << "... calling b1.my_union(b2) ...\n"; b1.my_union(b2);    
    
    cout << "b1 nachher : " << b1;
    cout << "b2 nachher : " << b2;
    
    cout << "Berechnung der distributiven Vereinigung von b1:\n";
    ASET r = b1.distr_union() ;
    cout << r;
    
    cout << "Summer der Elemente in der oberen Menge: " << long(r.sum()) << "\n";
} */

void test_mixing(void)
{
   cout << "\n\n--- Mischen von normalen und automem-Implementationen ---\n\n";
   
   SSET s1 = empty;
   s1 |= STR("Eins"); s1 |= STR("Zwei"); s1 |= STR("Drei"); s1 |= STR("Vier");
   SSET s2 = empty;
   s2 |= STR("Fuenf"); s2 |= STR("Sechs"); s2 |= STR("Drei"); s2 |= STR("Vier");   
   
   cout << "  s1 vorher: " << s1;
   cout << "  s2 vorher: " << s2;
   cout << "  Differenzmenge    s1 - s2 = " << (s1-s2);
   cout << "  Vereinigungsmenge s1 | s2 = " << (s1 | s2);
   cout << "  Menge   (s1-s2) | (s2-s1) = " << ((s1-s2)|(s2-s1));
   cout << "  s1 nachher: " << s1;
   cout << "  s2 nachher: " << s2;
}

// ---------------------------------------------------------------------------

void test_finput(void)
{
   cout << "\n\n--- Einlesen von Werten mit finput ---\n\n";
   
   ASET a1,a2;
   BSET b1,b2;
   
   cout << "Einlesen von zwei Nat0-set       : ";
   cin >>  a1 >> a2;
   cout << "Einlesen von zwei (Nat0-set)-set : ";
   cin >> b1 >> b2;
   cout << "\nEingelesene Werte:\n a1 = " << a1 << " a2 = " << a2;
   cout << " b1 = " << b1 << " b2 = " << b2;
   
   cout << "\nUeberschreiben von b1 mit neuem Wert: b1 = ";
   cin >> b1;
   cout << "neuer Wert b1 = " << b1;
}

// ---------------------------------------------------------------------------

main()
{
   cout << "\n\n=== Testprogramm automem-Konzept, Version 1.3 (17/11/93) ===\n\n";

	test_operator();
	test_selector();
	test_functions();
	test_usage();
/* 	test_functional(); */
	test_mixing();
   test_finput();

	profile_aset();
	profile_rep_aset();   
	profile_bset();
	profile_rep_bset();
	profile_sset();
	profile_rep_sset();   
}