// Testprogramm TYPGEN=Map, IMPL=automem
//
// Version 1.2, 17/11/93, Thorsten Tews
// ===========================================================================

#include "map_vdm.hpp"
#include "map_test.def"

// Beschreibung des Testmakros TEST:
//    - Parameter e : Ausdruck, der, falls Anweisungen ok, TRUE ergeben muss
//    - Parameter c1: Variable ve darf (YES) bzw. darf nicht (NO) durch den
//                    Ausdruck e veraendert werden
//    - Parameter c2: Variable v1 darf (YES) bzw. darf nicht (NO) durch den
//                    Ausdruck e veraendert werden

#define YES FALSE
#define NO  TRUE
#define TEST(e,c1,c2) {\
   MAP ve = MAP(empty), v1 = MAP(ELEM1,ELEM2), vtmp;\
   unsigned long err_code = 0;\
   if ((e) != TRUE) err_code |= 1;\
   if (c1 && !(ve.is_eq(MAP(empty))))  err_code |= 2;\
   if (c2 && !(v1.is_eq(MAP(ELEM1,ELEM2)))) err_code |= 4;\
   if (err_code != 0)\
      cout << " ### TEST FAILED: " #e ", error code: " << err_code << " ###\n";\
   else\
      cout << " Test passed : " #e "\n";\
}


// Beschreibung des Testmakros TEST_IS_IN_XXX:
// Spezieller Test fuer die is_in_xxx-Operationen wg. vorhandenem Referenz-
// parameter
//    - Parameter name : Name der zu testenden Operation
//    - Parameter p1   : in der Map gesuchter Wert
//    - Parameter t2   : Typ des zweiten (Referenz-)Parameters der Operation
//    - Parameter e2   : erwarteter Wert fuer zweiten Parameter
//
// Zum Test wird die Map [ELEM1 :-> ELEM2] verwendet !

#define TEST_IS_IN_XXX(name,p1,t2,e2) {\
   MAP v1 = MAP(ELEM1,ELEM2);\
   t2  ref_param;\
   unsigned long err_code = 0;\
   if (!v1.name(p1,ref_param)) err_code |= 1;\
   if (!(ref_param==e2)) err_code |= 2;\
   if (v1 != MAP(ELEM1,ELEM2)) err_code |= 4;\
   if (err_code != 0)\
      cout << " ### TEST FAILED: Operation " #name ", error code : " << err_code << "\n";\
   else\
      cout << " Test passed : v1." #name "(" #p1 ",r), r == " #e2 "\n";\
}   


// ---------------------------------------------------------------------------

void perform_test(void)
{
   cout << "Used objects in tests:\n" ;
   cout << "   ve ... the empty map []\n";
   cout << "   v1 ... the map [ELEM1 :-> ELEM2]\n\n";
   
   TEST(ve.is_empty(),NO,NO)        // Praedikate, nicht zerstoerend
   TEST(!v1.is_empty(),NO,NO)
   TEST(v1 == v1,NO,NO)
   TEST(!(v1 == ve),NO,NO)
   TEST(v1.is_in(ELEM1),NO,NO)
   TEST(ve.is_submap(v1),NO,NO)
   TEST(!v1.is_submap(ve),NO,NO)
   
   TEST(v1.max_dom() == ELEM1,NO,NO)  // Attribute, nicht zerstoerend
   TEST(v1.min_dom() == ELEM1,NO,NO)
   TEST(v1.card() == 1,NO,NO)
   TEST(ve.card() == 0,NO,NO)
   
   TEST(v1.sel(ELEM1) == ELEM2,NO,NO) // Selektoren, nicht zerstoerend
   TEST(v1.sel_dom() == ELEM1,NO,NO)
   
   TEST((v1 - ELEM1) == ve,NO,NO)     // Kombinatoren, z.T. zerstoerend
   TEST((v1 -= ELEM1) == ve,NO,YES)
   TEST((v1 - v1) == ve,NO,NO)
   TEST((v1 -= v1) == ve,NO,YES)
   TEST(ve.plus1(ELEM1,ELEM2) == v1,YES,NO)
   TEST((ve | v1) == v1,NO,NO)
   TEST((ve |= v1) == v1,YES,NO)
   TEST(ve.ovwrt1(ELEM1,ELEM2) == v1,YES,NO)
   TEST((ve + v1) == v1,NO,NO)
   TEST((ve += v1) == v1,YES,NO)
   TEST((v1 & ve) == ve,NO,NO)
   TEST((v1 &= ve) == ve,NO,YES)

	TEST_IS_IN_XXX(is_in_dom,ELEM1,Cpp_TYPE2,ELEM2)   
	TEST_IS_IN_XXX(is_in_min,ELEM1,Cpp_TYPE1,ELEM1)   
	TEST_IS_IN_XXX(is_in_max,ELEM1,Cpp_TYPE1,ELEM1)   
	TEST_IS_IN_XXX(is_in_rng,ELEM2,Cpp_TYPE1,ELEM1)   
	
	TEST(MAP(ELEM1,ELEM2) == vtmp.cvfromtupl(v1.cvtotupl()),NO,NO)  // cv(to|from)tupl
	
	// Test der fwrite/fread-Operationen :
#ifdef FILE_TEST
   MAP vwrite = MAP(ELEM1,ELEM2);
   vwrite.writef(TESTDIR);
   MAP vread = MAP(READF(TESTDIR));
   if (v1 == vread)
      cout << " Test passed : Operations fwrite/fread\n";
   else
      cout << " ### TEST FAILED: Operations fwrite/fread ###\n";
#endif	


   {
     MAP ve = MAP(empty);
     MAP v1 = MAP(ELEM1,ELEM2);

     cout << "\n--- Please check the following test results: ---\n\n";
     cout << " fprint    : \n   ve = " << ve << "   v1 = " << v1;   
     cout << " v1.hash() : " << v1.hash() << "\n";
     cout << " finput    : \n   Please enter a value  v1 = ";
     cin  >> v1;
     cout << "   Value read by finput  v1 = " << v1;   
   }
}

// ---------------------------------------------------------------------------
 
main()
{
   cout << "\n\n=== Test for TYPGEN: Map, IMPL: Automem (17/11/93) ===\n\n";
   
   perform_test();
   profile_map();
   profile_nset();

   return 0;
}   
