// Testprogramm TYPGEN=Tree, IMPL=automem
//
// Version 1.2, 17/11/93, Thorsten Tews
// ===========================================================================

#include "tree_vdm.hpp"
#include "tree_test.def"

// Beschreibung des Testmakros TEST:
//    - Parameter e : Ausdruck, der, falls Anweisungen ok, TRUE ergeben muss
//    - Parameter c1: Variable v1 darf (YES) bzw. darf nicht (NO) durch den
//                    Ausdruck e veraendert werden
//    - Parameter c2: Variable v2 darf (YES) bzw. darf nicht (NO) durch den
//                    Ausdruck e veraendert werden

#define YES FALSE
#define NO  TRUE
#define TEST(e,c1,c2) {\
                   TREE v1=TREE(E11,E21), v2=TREE(E12,E22), vtmp;\
                   unsigned long err_code = 0;\
                   if ((e) != TRUE) err_code |= 1;\
                   if (c1 && !(v1.is_eq(TREE(E11,E21)))) err_code |= 2;\
                   if (c2 && !(v2.is_eq(TREE(E12,E22)))) err_code |= 4;\
                   if (err_code != 0)\
                      cout << " ### TEST FAILED: " #e ", error code = " << err_code << " ###\n";\
                   else\
                      cout << " Test passed : " #e "\n";\
                 }

// ---------------------------------------------------------------------------

void perform_test(void)
{
   cout << "Used objects in tests:\n" ;
   cout << "   v1 ... the tree TREE(cmp1=E11,cmp2=E12)\n";
   cout << "   v2 ... the tree TREE(cmp1=E21,cmp2=E22) ( with v2 > v1 !)\n\n";
   
   TEST(!v1.is_eq(v2),NO,NO)      // Praedikate, nicht zerstoerend
   TEST(v1.is_eq(v1),NO,NO) 
#if !defined(NO_GREATER)   
   TEST(!v1.is_ge(v2),NO,NO)    
   TEST(v2.is_ge(v1),NO,NO)    
   TEST(!v1.is_gr(v2),NO,NO)  
   TEST(v2.is_gr(v1),NO,NO) 
#endif   
   
   TEST((v1.cmp1()).is_eq(E11),NO,NO)  // Selektoren, nicht zerstoerend
   TEST((v1.cmp2()).is_eq(E21),NO,NO) 

   // subst-Operation, nicht zerstoerend
   TEST((v1.subst_cmp1(E12)).is_eq(TREE(E12,E21)),NO,NO)    
   TEST((v1.subst_cmp2(E22)).is_eq(TREE(E11,E22)),NO,NO)
   
   // chg-Operation, zerstoerend
   TEST((v1.chg_cmp1(E12)).is_eq(TREE(E12,E21)),YES,NO)    
   TEST((v2.chg_cmp2(E21)).is_eq(TREE(E12,E21)),NO,YES)   

   TEST(TREE(E11,E21) == vtmp.cvfromtupl(v1.cvtotupl()), NO,NO);

   cout << "\n--- Please check the following test results: ---\n\n";
   TREE v1=TREE(E11,E21),v2=TREE(E12,E22);
   cout << " fpr       :\n   v1 = " << v1 << "   v2 = " << v2; 
   cout << " v1.hash() : " << v1.hash() << "\n";
   cout << " finput    : \n   Please enter a value  v1 = ";
   cin  >> v1;
   cout << "   Value read by finput  v1 = " << v1;
}

// ---------------------------------------------------------------------------
 
main()
{
   cout << "\n\n=== Test for TYPGEN: Tree, IMPL: Automem (17/11/93) ===\n\n";
   
   perform_test();
   profile_tree();

   return 0;
}   
