// Testprogramm TYPGEN=Tuple, IMPL=automem
//
// Version 1.2, 17/11/93, Thorsten Tews
// ===========================================================================

#include "tuple_vdm.hpp"
#include "tuple_test.def"

// Beschreibung des Testmakros TEST:
//    - Parameter e : Ausdruck, der, falls Anweisungen ok, TRUE ergeben muss
//    - Parameter ce: Variable ve darf (YES) bzw. darf nicht (NO) durch den
//                    Ausdruck e veraendert werden
//    - Parameter c1: Variable v1 darf (YES) bzw. darf nicht (NO) durch den
//                    Ausdruck e veraendert werden
//    - Parameter c2: Variable v2 darf (YES) bzw. darf nicht (NO) durch den
//                    Ausdruck e veraendert werden
//    - Parameter c3: Variable v2 darf (YES) bzw. darf nicht (NO) durch den
//                    Ausdruck e veraendert werden

#define YES FALSE
#define NO  TRUE
#define TEST(e,ce,c1,c2,c3) {\
   TUPLE ve = TUPLE(empty);\
   TUPLE v1 = TUPLE(ELEM1), v2 = TUPLE(ELEM1,ELEM1), v3 = TUPLE(ELEM1,ELEM2,ELEM3);\
   unsigned long err_code = 0;\
   if ((e) != TRUE) err_code |= 1;\
   if (ce && !(ve.is_eq(TUPLE(empty)))) err_code |= 2;\
   if (c1 && !(v1.is_eq(TUPLE(ELEM1)))) err_code |= 4;\
   if (c2 && !(v2.is_eq(TUPLE(ELEM1,ELEM1)))) err_code |= 8;\
   if (c3 && !(v3.is_eq(TUPLE(ELEM1,ELEM2,ELEM3)))) err_code |= 16;\
   if (err_code != 0)\
      cout << " ### TEST FAILED: " #e ", error code: " << err_code << " ###\n";\
   else\
      cout << " Test passed : " #e "\n";\
}

// ---------------------------------------------------------------------------

void perform_test(void)
{
   cout << "Used objects in tests:\n" ;
   cout << "   ve ... the empty tuple <>\n";
   cout << "   v1 ... the tuple <ELEM1>\n";
   cout << "   v2 ... the tuple <ELEM1,ELEM1>\n";
   cout << "   v3 ... the tuple <ELEM1,ELEM2,ELEM3>\n\n";
   
   TEST(ve.is_empty(),NO,NO,NO,NO)   // Praedikate, nicht zerstoerend !
   TEST(!v1.is_empty(),NO,NO,NO,NO)
   TEST(v1 == v1,NO,NO,NO,NO)
   TEST(v1 != v2,NO,NO,NO,NO)
#if !defined(NO_GREATER)   
   TEST(v2.is_gr(v1),NO,NO,NO,NO)
   TEST(!v2.is_gr(v3),NO,NO,NO,NO)
   TEST(v2.is_ge(v1),NO,NO,NO,NO)
   TEST(!v2.is_ge(v3),NO,NO,NO,NO)
#endif   
   
   TEST(v3.len() == 3,NO,NO,NO,NO)  // Attribute, nicht zerstoerend
   TEST(ve.len() == 0,NO,NO,NO,NO)
   TEST(v3.index(ELEM2) == 2,NO,NO,NO,NO)
   TEST(ve.index(ELEM2) == 0,NO,NO,NO,NO)
   TEST(v3.rindex(ELEM1) == 1,NO,NO,NO,NO)
   TEST(ve.rindex(ELEM1) == 0,NO,NO,NO,NO)
   
   TEST(v3.hd() == ELEM1,NO,NO,NO,NO)     // Selektoren, nicht zerstoerend
   TEST(v3.sel(2) == ELEM2,NO,NO,NO,NO)
   TEST(v3.last() == ELEM3,NO,NO,NO,NO)
   
   TEST(TUPLE(ELEM1,NAT0(2)) == v2,NO,NO,NO,NO)  // Extra-Konstruktoren, nicht z.
   TEST(TUPLE(v3,2,3) == TUPLE(ELEM2,ELEM3),NO,NO,NO,NO)
   
   TEST(v1.appl(ELEM1) == v2,NO,YES,NO,NO)  // Kombinatoren, z.T. zerstoerend
   TEST((ELEM1+v1) == v2,NO,NO,NO,NO)
   TEST((v1+=ELEM1) == v2,NO,YES,NO,NO)
   TEST((v1+ELEM1) == v2,NO,NO,NO,NO)
   TEST((v1+=v1) == v2,NO,YES,NO,NO)
   TEST((v1 + v1) == v2,NO,NO,NO,NO)
   TEST(v1.insert(1,ELEM1) == v2,NO,YES,NO,NO)
   TEST(v2.rmv1(2) == v1,NO,NO,YES,NO)
   TEST(v3.sub_tup(1,1) == v1,NO,NO,NO,YES)
   TEST(v2.tl() == v1,NO,NO,YES,NO)
   TEST(v1.ovwrt1(1,ELEM2) == TUPLE(ELEM2),NO,YES,NO,NO)

#if defined(FILE_TEST)
   TUPLE vf1 = TUPLE(ELEM1);
   vf1.writef(TESTFILE);
   TUPLE vf2 = TUPLE(READF(TESTFILE));
   if (vf1 != vf2)
      cout << " ### TEST FAILED: fwrite/fread ###\n";
   else
      cout << " Test passed : fwrite/fread\n";
   vf1.fapp(TESTFILE);
   TUPLE vf3 = TUPLE(READF(TESTFILE));
   if (vf3 != TUPLE(ELEM1,ELEM1))
      cout << " ### TEST FAILED: fapp ###\n";
   else
      cout << " Test passed : fapp\n";
#endif

#if defined(STRING_TEST)
   char tst_str1[] = "a test string for tuples";
   char tst_str2[] = "A TEST string FOR tUpLeS";
   char tst_str3[50];
   
   TUPLE vs1 = TUPLE(tst_str1);         /* VDM-Konstruktor: mkc  */
   TUPLE vs2 = TUPLE(mklc(tst_str2));   /* VDM-Konstruktor: mklc */
   vs1.cpc(tst_str3);
   if (strcmp(tst_str1,tst_str3)!=0)
      cout << " ### TEST FAILED: mkc/cpc ###\n";
   else
      cout << " Test passed : mkc/cpc\n";
   vs2.cpc(tst_str3);   
   if (strcmp(tst_str1,tst_str3)!=0)
      cout << " ### TEST FAILED: mklc/cpc ###\n";
   else
      cout << " Test passed : mklc/cpc\n";   
#endif

   cout << "\n--- Please check the following test results: ---\n\n";
   
   TUPLE ve = TUPLE(empty), v1 = TUPLE(ELEM1);
   TUPLE v2 = TUPLE(ELEM1,ELEM1), v3 = TUPLE(ELEM1,ELEM2,ELEM3);
   
   cout << " fprint    : \n   ve = " << ve << "   v1 = " << v1;
   cout << "   v2 = " << v2 << "   v3 = " << v3;
   cout << " v2.hash() : " << v2.hash() << "\n";
   cout << " finput    : \n   Please enter a value  v1 = ";
   cin  >> v1;
   cout << "   Value read by finput  v1 = " << v1;   
}

// ---------------------------------------------------------------------------
 
main()
{
   cout << "\n\n=== Test for TYPGEN: Tuple, IMPL: Automem (17/11/93) ===\n\n";
   
   perform_test();
   profile_tuple();
   profile_nset();

   return 0;
}   
