// $Id: iterex.cc,v 1.3 1995/12/03 17:05:09 vdm Exp $

// example for use of iterators with C++ 

#include "iter.hpp"

#define OUT(x) cout << #x " = " << x << endl

//---------------------------------------------------------------------------------------

void settest() {
  SET s1 = empty;
  s1 |= 20;
  s1 |= 21;
  s1 |= 22;

  OUT(s1);

  ITER_SET is = s1;		// while loop with iterator
  while (! is ) {
    NAT0 e = is.sel();
    cout << "e = " << e << endl;
    ++is;			// post- and pre- increment are equivalent
  }
//  NAT0 e1 = is.sel();		// will produce runtime error
//  is++;			//	dto. 
}

//---------------------------------------------------------------------------------------

void bagtest() {
  BAG b1 = empty;
  b1 |= 20;
  b1 |= 21;
  b1 |= 22; b1 |= 22; b1 |= 22;

  OUT(b1);
				// for loop with iterator
  for (ITER_BAG ib = b1; ! ib; ib++) {
    NAT0 e = ib.sel();
    cout << "e = " << e << endl;
  }
//  NAT0 e1 = ib.sel();		// will produce runtime error
//  ib++;
}

//---------------------------------------------------------------------------------------

void maptest() {
  MAP m1 = empty;
  m1 |= PAIR_MAP("1",2);		// same as: m1.plus1("1",2);
  m1 |= PAIR_MAP("4",5);
  m1 |= PAIR_MAP("3",4);
  m1 |= PAIR_MAP("2",3);

  OUT(m1);

  ITER_MAP im(m1);
  while ( ! im ) {
    NAT1 a;
    STR k = im.sel(a);
    cout << "k = " << k << "a = " << a << endl;
    im++;
  }
//  im++;		// will produce runtime error
}

//---------------------------------------------------------------------------------------

void strtest() {
  STR s1 = empty;
  s1 += 'a';
  s1 += 'b';
  s1 += 'c';

  OUT(s1);

  ITER_STR is(s1);
  while (! is ) {
    CHAR e = is.sel();
    cout << "e = " << e << endl;
    is++;
  }
//  is++;		// will produce runtime error

}

//---------------------------------------------------------------------------------------

void bittest() {
  BITS s1 = empty;
  s1 += true;
  s1 += false;
  s1 += true;

  OUT(s1);

  ITER_BITS is(s1);
  while (! is ) {
    BOOL e = is.sel();
    cout << "e = " << e << endl;
    is++;
  }
//  is++;		// will produce runtime error

}

//---------------------------------------------------------------------------------------

// heapmap implemetation is not supported for automem C++
// therefore conditional compilation
// for target AUTO

void tokentest() {
#ifndef AUTOMEM
  HEAP h1 = empty;
  PTR p1 = new_val;
  PTR p2 = new_val;
  PTR p3 = new_val;

  h1.plus1(p1,VALUE(INTG(3)));
  h1.plus1(p2,VALUE(INTG(5)));
  h1.plus1(p3,VALUE(INTG(4)));

  OUT(h1);

  ITER_HEAP ih(h1);
  while (! ih ) {
    VALUE v;
    PTR p = ih.sel(v);
    cout << "p = " << p << "v = " << v.val() << endl;
    ih++;
  }
#endif  
}

//---------------------------------------------------------------------------------------

main() {
  settest();
  bagtest();
  maptest();
  strtest();
  bittest();
  tokentest();

  return 0;
}

//---------------------------------------------------------------------------------------

