/* ----- demo_map.c --- Jan Peter Mohr --- WI 4487 ----- */

/*        Demo    fuer eine Stuecklistenverwaltung       */
/* Jedem Teil ist die Menge seiner Baugruppen zugeordnet */
/* Ist die Menge leer, ist das Teil ein Bauelement       */
/* Ein Teil ist durch seine Teil-Nummer dargestellt      */
/* Sl   =  Pn -m-> Pset */
/* Pset =  Pn -set      */
/* Pn   = Nat0          */

#include <stdio.h>
#include "jp-map.h"

/*-----------------------------------------------------------------------*/
/* Aufbau einer Demo-Stueckliste */
/*-----------------------------------------------------------------------*/
Sl build_sl()
{
 Pset ps;
 Sl   sl;

 ps = mk_one_pset((Pn)2);      /* Teile von Teil 1 */
 ps = union1_pset(ps,(Pn)4);
 ps = union1_pset(ps,(Pn)6);
 ps = union1_pset(ps,(Pn)8);
 ps = union1_pset(ps,(Pn)10);
 ps = union1_pset(ps,(Pn)12);

 sl = mk_one_sl((Pn)1,ps);

 ps = mk_one_pset((Pn)3);      /* Teile von Teil 2 */
 ps = union1_pset(ps,(Pn)6);
 ps = union1_pset(ps,(Pn)9);
 ps = union1_pset(ps,(Pn)12);

 sl = union1_sl(sl,(Pn)2,ps);

 ps = mk_one_pset((Pn)4);      /* Teile von Teil 3 */
 ps = union1_pset(ps,(Pn)8);
 ps = union1_pset(ps,(Pn)12);

 sl = union1_sl(sl,(Pn)3,ps);

 ps = mk_one_pset((Pn)9);      /* Teile von Teil 4 */
 ps = union1_pset(ps,(Pn)12);

 sl = union1_sl(sl,(Pn)4,ps);

 ps = mk_empty_pset();     /* 6,8,9,10,12 -> Elementarbausteine */
 sl = union1_sl(sl,(Pn)6,ps);

 ps = mk_empty_pset();
 sl = union1_sl(sl,(Pn)8,ps);

 ps = mk_empty_pset();
 sl = union1_sl(sl,(Pn)9,ps);

 ps = mk_empty_pset();
 sl = union1_sl(sl,(Pn)10,ps);

 ps = mk_empty_pset();
 sl = union1_sl(sl,(Pn)12,ps);

 return sl;
}

/*-----------------------------------------------------------------------*/
/* Konvertierungsfunktion fuer 'fold.subparts': pn --> Bestandteile(pn)  */
/*-----------------------------------------------------------------------*/
Pset get_pn (pn,sl)
Pn pn;
Sl sl;
{ return copy1_pset(sel_sl(sl,pn)); }

/*-----------------------------------------------------------------------*/
/* Konvertierungsfunktion fuer 'fold.dom': Pn --> ein-elementiger Set    */
/*-----------------------------------------------------------------------*/
Pset to_pset_pn_pset (pn,ps)
Pn pn;
Pset ps;
{ return mk_one_pset(pn); }

/*-----------------------------------------------------------------------*/
/* Konvertierungsfunktion fuer 'fold.drng': ps --> ps (IdentitM-^Dt)        */
/*-----------------------------------------------------------------------*/
Pset distr_pn_pset (pn,ps)
Pn pn;
Pset ps;
{ return copy1_pset(ps); }

/*-----------------------------------------------------------------------*/
/* Liefert die Bestandteile aller Teile in 'ps' INKLUSIVE 'ps' selbst    */
/*-----------------------------------------------------------------------*/
Pset parts_pset (ps,sl)
Pset ps;
Sl   sl;
{
 Pset ps_sub = copy1_pset(ps);
 Bool ready;

 ps_sub = subparts_pset(ps_sub,sl);
 if (is_subset_pset(ps_sub,ps)) {
    del1_pset(ps_sub);
    return ps;
 }
 else {
    ps = union_pset(ps,ps_sub);
    return parts_pset(ps,sl);
 }
}

/*-----------------------------------------------------------------------*/
/* Liefert die Bestandteile eines Teils 'p' EXKLUSIVE 'p' selbst !!!     */
/*-----------------------------------------------------------------------*/
Pset parts_pn (p,sl)
Pn p;
Sl sl;
{ return parts_pset(copy1_pset(sel_sl(sl, p)), sl); }

/*-----------------------------------------------------------------------*/
/* Stuecklisten-Invariante: Testet die Konsistenz der StM-^Ackliste         */
/*-----------------------------------------------------------------------*/
Bool inv_sl (sl)
Sl sl;
{
 Bool ok = TRUE;
 Sl   sav1;
 Pset dom,rng,parts,ps1;
 Pn   pn;

 sav1 = copy1_sl(sl);
 dom = dom_sl(sav1);

 sav1 = copy1_sl(sl);
 rng = drng_sl(sav1);

 ok = is_subset_pset(rng,dom);  /* rng not e dom ? */
 del1_pset(rng);
 while (ok && !is_empty_pset(dom)) {
    pn = sel_pset(dom);
    dom = sub1_pset(dom,pn);
    parts = parts_pn(pn, sl);
    ok = !is_in_pset(pn,parts);
 }
 if (!is_empty_pset(dom))
    del1_pset(dom);

 return ok;
}

main() /* -------------------- Hauptprogramm --------------------*/
{
 Sl sl;
 Pset parts;

 puts("VDM-Demo Jan Peter Mohr --- MAP --- qualitative Stueckliste\n");
 puts("Test der Invariante & Ermittlung aller Bestandteile eines Teils\n");
 sl = build_sl();
 printf("Inhalt der Stueckliste:\n");
 pr_sl(sl);
 printf("\nDie Stueckliste ist %skonsistent !\n\n", inv_sl(sl) ? "" : "in");
 parts = parts_pn((Pn)1,sl);
 printf("\nBestandteile des Teils 1:\n");
 pr_pset(parts);
 parts = parts_pn((Pn)2,sl);
 printf("\nBestandteile des Teils 2:\n");
 pr_pset(parts);
 parts = parts_pn((Pn)4,sl);
 printf("\nBestandteile des Teils 4:\n");
 pr_pset(parts);
 parts = parts_pn((Pn)10,sl);
 printf("\nBestandteile des Teils 10:\n");
 pr_pset(parts);
 return 0;
}

/* 1 -> {2,4,6,8,10,12,3,9}  2 -> {3,6,9,12,4,8}  4 -> {9,12}  10 -> {} */
