/* demo-set.c --- Jan Peter Mohr --- WI 4487 -- */

/* Loesung des Problems "coarsest partitioning" */
/* Umwandlung einer Menge von Mengen in eine    */
/* Menge von ueberschneidungsfreien Mengen      */

/* Aset = Nat0 -set */
/* Sset = Aset -set */

#include<stdio.h>
#include"jp-set.h"

/*-------------------- Hilfsfunktion 'build_sset' -------------------------*/
/* Aufbau einer Menge von nicht-ueberschneidungsfreien Mengen              */
/*-------------------------------------------------------------------------*/

Sset build_sset()
{
 Sset ss;
 Aset a;

 a = mk_one_aset(1);    /* ungerade Zahlen */
 a = union1_aset(a,3);
 a = union1_aset(a,5);
 a = union1_aset(a,7);
 a = union1_aset(a,9);
 a = union1_aset(a,11);
 a = union1_aset(a,13);
 a = union1_aset(a,15);

 ss = mk_one_sset(a);

 a = mk_one_aset(2);    /* gerade Zahlen */
 a = union1_aset(a,4);
 a = union1_aset(a,6);
 a = union1_aset(a,8);
 a = union1_aset(a,10);
 a = union1_aset(a,12);
 a = union1_aset(a,14);
 a = union1_aset(a,16);
                        
 ss = union1_sset(ss,a);

 a = mk_one_aset(1);     /* natuerliche Zahlen */
 a = union1_aset(a,2);
 a = union1_aset(a,3);
 a = union1_aset(a,4);
 a = union1_aset(a,5);
 a = union1_aset(a,6);
 a = union1_aset(a,7);
 a = union1_aset(a,8);

 ss = union1_sset(ss,a);

 a = mk_one_aset(2);     /* Primzahlen */
 a = union1_aset(a,3);
 a = union1_aset(a,5);
 a = union1_aset(a,7);
 a = union1_aset(a,11);
 a = union1_aset(a,13);
 a = union1_aset(a,17);
 a = union1_aset(a,19);

 ss = union1_sset(ss,a);

 return(ss);             
} /* end of build_sset */

/*-------------------- Hilfsfunktion 'find_partner' -----------------------*/
/* Sucht zu einer Menge as eine nicht-elementfremde Menge in 'ss'          */
/*-------------------------------------------------------------------------*/
/* Die Menge 'ss' wird zerstoert !!                                        */
/* Return: Wenn nur elementfremde Mengen vorhanden: leere Menge            */
/*-------------------------------------------------------------------------*/

Aset find_partner (ss,as)
Sset ss;  /* Menge von Mengen */
Aset as;  /* vorgegeben */
{
 Aset sav1,sav2;
 Aset yaas = mk_empty_aset(); /* yet another Aset */

 if (is_empty_sset(ss))
    return(yaas);
 else {
    yaas = sel_sset(ss);
    ss = sub1_sset(ss,yaas);

    sav1 = copy1_aset(as);
    sav2 = copy1_aset(yaas);

    if (!is_empty_aset(inters_aset(sav1,sav2))) {
       del1_sset(ss);
       return(yaas);
    }
    else
       return(find_partner(ss,as));
 }
} /* end of find_partner */

/*---------------- semantische Funktion 'decompose_sset' ------------------*/
/* Zerlegung einer Menge von Mengen in Menge von ueberschneidungsfreien M. */
/*-------------------------------------------------------------------------*/
/* Die Menge 'ss' wird zerstoert (sie dient als Speicher des Resultats)    */
/*-------------------------------------------------------------------------*/

Sset decompose_sset (ss)
Sset ss; /* zu zerlegende Menge (muss wohlgeformt sein) */
{
 Aset a1,a2,sav1,sav2;
 Sset ss1,ss2;
 Bool found = FALSE;

 ss1 = copy1_sset(ss);
 while (!found && !is_empty_sset(ss1)) {
    a1 = sel_sset(ss1);
    ss1 = sub1_sset(ss1,a1);
    ss2 = copy1_sset(ss1);
    a2 = find_partner(ss2,a1);
    found = !is_empty_aset(a2);
 }
 del1_sset(ss1);

 if (!found)
    return(ss);
 else {
    ss = sub1_sset(ss,a1);  /* ss \ a1 */
    ss = sub1_sset(ss,a2);  /* ss \ a2 */

    sav1 = copy1_aset(a1);
    sav2 = copy1_aset(a2);

    ss = union1_sset(ss,inters_aset(sav1,sav2)); /* ss u (a1 n a2) */

    sav1 = copy1_aset(a1);
    a1 = diff_aset(a1,a2);
    a2 = diff_aset(a2,sav1);
    del1_aset(sav1);

    if (!is_empty_aset(a1))
       ss = union1_sset(ss,a1);  /* ss u (a1 \ a2) */

    if (!is_empty_aset(a2))
       ss = union1_sset(ss,a2);  /* ss u (a2 \ a1) */

    return(decompose_sset(ss));  /* neue Menge ss zerlegen */
 }
} /* end of decompose_sset */

main() /* ---------------------- Hauptprogramm ----------------------------*/
{
 Sset ss;
 printf("\nVDM-Demo Jan Peter Mohr\n");
 printf("Ermittlung der 'coarsest partitioning' einer Menge von Mengen\n\n");
 ss = build_sset();
 printf("Menge VOR der Zerlegung:\n");
 pr_sset(ss);
 ss = decompose_sset(ss);
 printf("Menge NACH der Zerlegung:\n");
 pr_sset(ss);
 return 0;
} /* Ende */

/* CP = {11,13};{17,19};{1};{2};{3,5,7};{9,15};{4,6,8};{10,12,14,16} */
