// $Id: mmm.cc,v 1.1 1994/01/30 20:17:07 vdm Exp $

// performance test for map implementations
// keys and attribut values are of type NAT0
// 1. a  map with keys in the range of 0..2*MAX-1 is built
//    key are generated with a random function
//    attribute values range from 0..MAX-1
// 2. the cardinality function is called
// 3. the arithmetic mean of the key values and the attributes is
//    computed with an iterator function
// 4. 4*MAX random lookups are performed
// about 1/2 of them should be found in the map

#include <stdlib.h>

#include "map.hpp"

#if R == 1
#define OUT(x) cout << #x " = " << x << endl
#else
#define OUT(x)
#endif

void maptest(unsigned MAX) {
  cout << "build a map with " << MAX << " elements" << endl;
  cout << "keys are randomly choosen from the range 0.."      <<2*MAX-1<<endl;
  cout << "attributes are randomly choosen from the range 0.."<<  MAX-1<<endl;

  unsigned LMAX = 5 * MAX; 
  unsigned MAX2 = 2 * MAX;
  MAP m1 = empty;

  for (unsigned i = 0; i < MAX; i++) {
      NAT0 k = NAT0(random() % (MAX2));
      while ( m1.is_in(k) ) {
         k = NAT0(random() % (MAX2));
      }
      m1.plus1(k,NAT0(random() % MAX));
  }

  cout << "compute cardinality  == " << m1.card()
       << " (should be " << MAX << ")"
       << endl;

  cout << "process all elements with an iterator (arithmetic means)"
       << endl;

  {
  ITER_MAP im(m1);
  NAT0     ksum = 0;
  NAT0     asum = 0;

  while (! im) {
    NAT0 a;
    NAT0 k = im.sel(a);

    ksum = ksum + k;
    asum = asum + a;

    im++;
  }
  cout << "arithmetic mean of keys       == " << (ksum + MAX % 2) / MAX
       << endl
       << "arithmetic mean of attributes == " << (asum + MAX % 2) / MAX
       << endl;
  }
  cout << "lookup " << LMAX << " elems randomly"
       << endl;

  unsigned found = 0;
  for (unsigned ii = 0; ii < LMAX; ii++) {
     NAT0 k = NAT0(random() % (MAX2));
     NAT0 a;
     if ( m1.is_in_dom(k,a) )
        found++;
  }
  cout << LMAX << " lookups and " << found << " successes"
       << endl;

}

main(int argc, char * argv[]) {
  maptest((argv[1])?atoi(argv[1]):10);

//  profile_map(); 
//  profile_rep_map();

return 0;
}

