// $Id: main.cc,v 1.2 1995/02/22 10:32:42 vdm Exp $

// a simple main program for demonstrating the implementation of
// user defined operations for signatures of functions
// defined in the specification 
// and their use
//
// in larger projects, the member functions should be separated from
// the application (in this case the main program) by putting them
// into a separate .cc file

#include "sig.hpp"

//------------------------------------------------------------------------------

				// the attribute functions of class TYPE1
				// the bodies just consists of example expressions

NAT0 TYPE1 :: attribute1 () const {
  return  card();
}

INTG TYPE1 :: attribute2 (const NAT0 & x) const {
  return INTG(card()) - INTG(x);
}

TYPE2 TYPE1 :: attribute3 (TYPE3 &) const {
  return TYPE2(STR("abc"));
}

NAT1 NAT1 :: attribute4 () const {
  return (*this) * (*this);	// a square function
}

//------------------------------------------------------------------------------

				// the predicate functions of class TYPE1
BOOL TYPE1 :: predicate1 () const {
  return is_in(0);
}

BOOL TYPE1 :: predicate2 (const TYPE1 & e2) const {
  return *this == e2;
}

BOOL TYPE1 :: predicate3 (const TYPE2 & e2, TYPE3 & e3) const {
  return (card() == e2.len()) == e3.f2();
}

//------------------------------------------------------------------------------

				// the update functions of class TYPE1
TYPE1 & TYPE1 :: modify1 () {
  return (*this |= 42);
}

TYPE1 & TYPE1 :: modify2 (const TYPE2 & e2) {
  return (*this |= e2.len());	// state is modified
}

TYPE2 TYPE1 :: modify3 () {
  *this |= 0;			// state is modified
  return TYPE2(empty);		// and a return value is constructed
}

TYPE2 TYPE1 :: modify4 (TYPE3 & e3) {
  *this = empty;		// state is modified
  e3 = TYPE3(*this,false);	// 2. output parameter is manipulated
  return TYPE2(empty);		// result value is constructed
}

TYPE2 TYPE1 :: modify5 (const NAT0 & e1, TYPE3 & e3) {
  *this |= e1;			// state is modified with 1. parameter
  e3 = TYPE3(*this,false);	// 2. output parameter is manipulated
  return TYPE2(empty);		// result value is constructed
}

//------------------------------------------------------------------------------

				// the following three member functions implement
				// the "predicate" operation for the
				// subtypes of Uniontype, they are called from
				// the dispatch function "predicate" in
				// class UNIONTYPE

BOOL TYPE1 :: predicate (const NAT0 & e1) const {
  return card() == e1;
}

BOOL TYPE2 :: predicate (const NAT0 & e1) const {
  return len() == e1;
}

BOOL TYPE3 :: predicate (const NAT0 & e1) const {
  return f2();
}

				// the following three member functions implement
				// "modifyfct" for the three subtypes of Uniontype
				// they are called from the generated dispatch
				// function "modifyfct" in class UNIONTYPE

TYPE1 & TYPE1 :: modifyfct (const NAT0 & e1) {
  return *this |= e1;
}
				// the update function for TYPE2
TYPE2 & TYPE2 :: modifyfct (const NAT0 & e1) {
  return *this += STR("123");
}
				// the update function for TYPE3
TYPE3 & TYPE3 :: modifyfct (const NAT0 & e1) {
  chg_f2(e1 == 0);
  return *this;
}

//------------------------------------------------------------------------------

int main() {
{
  TYPE1 t1 = TYPE1(NAT0(42));
  TYPE2 t2 = TYPE2(STR("abcde"));
  TYPE3 t3 = TYPE3(t1,true);

  UNIONTYPE u1 = UNIONTYPE(t1);
  UNIONTYPE u2 = UNIONTYPE(t2);
  UNIONTYPE u3 = UNIONTYPE(t3);

  cout << t1 << t2 << t3 << endl;
  cout << u1 << u2 << u3 << endl;
  cout << u1.predicate(1) << u2.predicate(1) << u3.predicate(0) << endl;

  u1.modifyfct(13);
  u2.modifyfct(14);
  u3.modifyfct(15);
 
  cout << t1 << t2 << t3 << endl;
  cout << u1 << u2 << u3 << endl;
  cout << u1.predicate(0) << u2.predicate(0) << u3.predicate(0) << endl;
}

  profile_type1();
  profile_rep_type1();
  profile_type2();
  profile_rep_type2();
  profile_type3();
  profile_str();
  profile_rep_str();

  return 0;
}
