// $Id: stuecklistenops.cc,v 1.1 1996/09/19 15:46:53 vdm Exp $

#include "stueckliste.hpp"

// -----------------------------------------------------------------------------
//
// Operationen auf STUECKLISTEn
//
// -----------------------------------------------------------------------------
//
// Invarinte: alle Teile mssen bekannt sein
// und ...

BOOL
STUECKLISTE::inv() const
{
  TEILEMENGE
    alle_teile			= alleTeile(),
    alle_bekannten_teile	= dom();

  BOOL
    alle_teile_bekannt = alle_teile.is_subset(alle_bekannten_teile);

  if ( ! alle_teile_bekannt )
    {
      cerr
	<< "STUECKLISTE::inv: unbekannte Teile in der Stckliste: "
	<< (alle_teile - alle_bekannten_teile) << endl;
    }

  BOOL
    zyklen_frei = zyklenfrei();

  return
    alle_teile_bekannt
    &
    zyklen_frei;
}

// -----------------------------------------------------------------------------
//
// zyklenfrei Test

BOOL
STUECKLISTE::zyklenfrei() const
{
  ITER_STUECKLISTE i(*this);
  BOOL zyklenfrei(true);

  while ( ! i ) {
    BESTANDTEILE bt;
    TEILENR t = i.sel(bt);
    TEILEMENGE tm = bt.verwendeteTeile(*this,TEILEMENGE(empty));

    if ( tm.is_in(t) ) {
      zyklenfrei = false;
      cerr
	<< "STUECKLISTE::zyklenfrei: Teil enthlt sich selbst: " << t << endl;
    }
    ++i;
  }
  return zyklenfrei;
}

// -----------------------------------------------------------------------------
//
// ein Teil eintragen

STUECKLISTE &
STUECKLISTE::neuesTeil(const TEILENR & t)
{
  ovwrt1(t,BESTANDTEILE(empty));
  return *this;
}

// -----------------------------------------------------------------------------
//
// fr ein Teil ein Bestandteil eintragen

STUECKLISTE &
STUECKLISTE::neuesBestandTeil(const TEILENR & t,
			      const TEILENR & et,
			      const ANZAHL & n)
{
  BESTANDTEILE bt;
  if ( ! is_in_dom(t,bt) ) {
    bt = BESTANDTEILE(et,n);
  }
  else {
    bt.ovwrt1(et,n);
  }
  ovwrt1(t,bt);
  return *this;
}

// -----------------------------------------------------------------------------
//
// ein Teil lschen

STUECKLISTE &
STUECKLISTE::loescheTeil(const TEILENR & t)
{
  *this -= t;
  return *this;
}

// -----------------------------------------------------------------------------
//
// auflsen einer Stckliste

BESTANDTEILE
STUECKLISTE::aufloesen(const TEILENR & t) const
{
  BESTANDTEILE
    bt = (*this)[t],
    res;

  if ( bt.is_empty() ) {
    res = BESTANDTEILE(t,1);
  }
  else {
    res = bt.aufloesen(*this);
  }

  return res;
}

// -----------------------------------------------------------------------------
//
// alle Teilenummern von Produkten (Teilen die nirgends eingebaut werden)

TEILEMENGE
STUECKLISTE::alleProdukte() const
{
  TEILEMENGE tm = dom();
  tm -= alleTeile();
  return tm;
}

// -----------------------------------------------------------------------------
//
// alle Teilenummern von einfachen Teilen berechnen

TEILEMENGE
STUECKLISTE::alleEinfachenTeile() const
{
  ITER_STUECKLISTE i(*this);
  TEILEMENGE tm(empty);

  while ( !i ) {
    BESTANDTEILE bt;
    TEILENR t = i.sel(bt);
    if ( bt.is_empty() )
      tm |= t;
    ++i;
  }

  return tm;
}

// -----------------------------------------------------------------------------
//
// alle TeileNummern berechnen

TEILEMENGE
STUECKLISTE::alleTeile() const
{
  ITER_STUECKLISTE i(*this);
  TEILEMENGE tm(empty);

  while ( !i ) {
    BESTANDTEILE bt;
    TEILENR t = i.sel(bt);
    tm |= bt.dom();
    ++i;
  }

  return tm;
}

// -----------------------------------------------------------------------------
//
// Operationen auf BestandTeile
//
// -----------------------------------------------------------------------------
//
// auflsen von einer Tabelle von Bestandteilen
//

BESTANDTEILE
BESTANDTEILE::aufloesen(const STUECKLISTE & sl) const
{
  ITER_BESTANDTEILE i(*this);
  BESTANDTEILE res(empty);

  while ( ! i ) {	// Schleife ber alle Teile
    ANZAHL		n;
    TEILENR		t = i.sel(n);
    BESTANDTEILE	bt;

    bt = sl.aufloesen(t);	// ein Teil auflsen

    bt = bt.skalieren(n);	// mit Anzahl skalieren

    res = res.mischen(bt);	// mit dem schon aufgelsten mischen

    ++i;
  }
  return res;
}

// -----------------------------------------------------------------------------
//
// eine Bestandteil-Liste mit einer Anzahl skalieren
//

BESTANDTEILE
BESTANDTEILE::skalieren(const ANZAHL & n) const
{
  ITER_BESTANDTEILE i(*this);
  BESTANDTEILE res(empty);

  while ( ! i ) {		// Schleife ber alle Teile
    ANZAHL n1;
    TEILENR t = i.sel(n1);
    res.plus1(t,n1*n);		// Anzahl skalieren
    ++i;
  }
  return res;
}

// -----------------------------------------------------------------------------
//
// zwei Bestandteil-Listen "addieren" mischen
//

BESTANDTEILE
BESTANDTEILE::mischen(const BESTANDTEILE & bt) const
{
  BESTANDTEILE
    res = (*this - bt) | (bt - *this);	// die Teile, die nur in einer Tabelle sind

  TEILEMENGE
    gemeinsameTeile = dom() & bt.dom();	// die Teile, die in beiden Tabellen sind

  ITER_TEILEMENGE i(gemeinsameTeile);

  while ( ! i ) {
    TEILENR t = i.sel();

    res.plus1(t, (*this)[t] + bt[t]);	// die Anzahlen summieren
    					// und das Paar [t :-> #] in res eintragen
    ++i;
  }
  return res;
}

// -----------------------------------------------------------------------------

TEILEMENGE
BESTANDTEILE::verwendeteTeile(const STUECKLISTE & sl, const TEILEMENGE & besucht) const
{
  ITER_BESTANDTEILE i(*this);
  TEILEMENGE res(besucht);

  while ( ! i ) {	// Schleife ber alle Teile
    ANZAHL		n;
    TEILENR		t = i.sel(n);

    // wenn ein Teil t noch nicht in der besucht Menge enthalten ist,
    // wird es hinzugefgt und alle verwendeten Teile dieses Teils berechnet

    if ( ! res.is_in(t) ) {

      // nur wenn das Teil auch bekannt ist, wird es bercksichtigt
      // sonst Fehler bei Stcklisten mit unbekannten Teilen

      if ( sl.is_in(t) ) {
	BESTANDTEILE	bt = sl[t];
	res |= t;
	res = bt.verwendeteTeile(sl,res);
      }
    }
    ++i;
  }
  return res;
}

