/*----------------------------------------------------------------------*/

/* Modul:                  @(#)db.c */
/* Version:                1.1 */
/* Ersteller:              Gerrit Petrich, WI 8215, WS 93 */
/* Datum der letzten */
/* Aenderung:              93/02/07 */

/*----------------------------------------------------------------------*/

/* Aufgabe:   "db.c" ist die Programmschnittstelle fuer die */
/*            Funktionen aus "dbs.c". */

/*----------------------------------------------------------------------*/

/* Aenderungen: */
/* wann:                 warum: */
/*----------------------------------------------------------------------*/
/* 16.02.1993 pre_getsx_SignComponent eingefuegt. */

/* 16.02.1993  include der Systemheader entfernt. Sie werden jetzt in header.h eingebunden */

/*----------------------------------------------------------------------*/

#include "header.h"          /* allgemeines Headerfile */
#include "gen.h"             /* aus gen.vdm generierter Header */
#include "proto.h"           /* allgemeines Headerfile */
#include "dbs.h"             /* Headerfile fuer dbs.c */

void Usage()                 /* Usage gibt die moeglichen Programm- */
                             /* parameter aus, wenn das Programm */
                             /* fehlerhaft aufgerufen wird */
{
  fprintf(stderr,"usage: db [-Bvdmdbname ]* -v|-a|-i|-o|-c|-g params\n");
  fprintf(stderr,"options:\n");
  fprintf(stderr,"  -B : work on this VDM-database\n");
  fprintf(stderr,"       vdmdbname: VDM-databasename\n");
  fprintf(stderr,"       this option can be repeated to work on several databases\n");
  fprintf(stderr,"  -v : check if databases exist\n");
  fprintf(stderr,"       params: none\n");
  fprintf(stderr,"  -a : check if ADT exists in databases\n");
  fprintf(stderr,"       params: <ADTname>\n");
  fprintf(stderr,"  -i : check if Implementation exists in databases\n");
  fprintf(stderr,"       params: <ADTname> <Implname>\n");
  fprintf(stderr,"  -o : check if Operation exists in databases\n");
  fprintf(stderr,"       params: <ADTname> <Implname> <Opname>\n");
  fprintf(stderr,"  -s : check if Signature exists in databases\n");
  fprintf(stderr,"       params: <ADTname> <Opname>\n");
  fprintf(stderr,"  -c : check if Component exists in databases\n");
  fprintf(stderr,"       params: <ADTname> <Implname> <Opname> <Componentname>\n");
  fprintf(stderr,"  -S : check if SignComponent exists in databases\n");
  fprintf(stderr,"       params: <ADTname> <Opname> <SignComponentname>\n");
  fprintf(stderr,"  -g : prints Component on stdout\n");
  fprintf(stderr,"       params: <ADTname> <Implname> <Opname> <Componentname>\n");
  fprintf(stderr,"  -x : prints Component on stdout, includes handled\n");
  fprintf(stderr,"       params: <ADTname> <Implname> <Opname> <Componentname>\n");
  fprintf(stderr,"  -G : prints SignComponent on stdout\n");
  fprintf(stderr,"       params: <ADTname> <Opname> <SignComponentname>\n");
  fprintf(stderr,"  -X : prints SignComponent on stdout, includes handled\n");
  fprintf(stderr,"       params: <ADTname> <Opname> <SignComponentname>\n");
  fprintf(stderr,"  -d : prints docu-information on stdout\n");
  fprintf(stderr,"       params: <ADTname> [<Implname> [<Opname>]]\n");
}

/*----------------------------------------------------------------------*/

void pre_exists_Vdmdbs F4(
        int,i,             /* Zaehler fuer verarbeitete Parameter */
        int,argc,          /* Anzahl der Parameter */
        char **,argv,      /* Tupel der Parameter */
        Dbs,dbs)           /* Tupel der Datenbanknamen */
                        /* pre_exists_Vdmdbs wertet die Programm- */
                        /* parameter aus, und startet die Funktion */
                        /* exists_Vdmdbs */
{
Vdmdbs vdmdbs;     /* Datenbanktupel */

  HALLO("pre Vdmdbs\n");
  while (argc > i)                     /* zu viele Parameter: */
    {
      fprintf(stderr,"superfluos parameter %s ignored\n",argv[i++]);
    }
  vdmdbs=def_Vdmdbs(dbs);              /* Datenbanken oeffnen */
  OUT_bool(exists_Vdmdbs(vdmdbs));
  del_vdmdbs(vdmdbs);                  /* Speicher freigeben */
}

/*----------------------------------------------------------------------*/

void pre_exists_Adt F4(
        int,i,             /* Zaehler fuer verarbeitete Parameter */
        int,argc,          /* Anzahl der Parameter */
        char **,argv,      /* Tupel der Parameter */
        Dbs,dbs)           /* Tupel der Datenbanknamen */
                             /* pre_exists_Adt wertet die Programm- */
                             /* parameter aus, und startet die Funktion */
                             /* exists_Adt */
{
Adtname adt_name;
Vdmdbs vdmdbs;     /* Datenbanktupel */

  HALLO("pre Adt\n");
  if (argc<i+1)                        /* zu wenig Parameter: */
    {
      fprintf(stderr,"missing functionparameter: no adtname given\n");
      Usage();
    }
  else
    {                                  /* Parameter auswerten */
      adt_name=mkc_str(argv[i++]);
      while (argc > i)                 /* zu viele Parameter: */
        {
          fprintf(stderr,"superfluos parameter %s ignored\n",argv[i++]);
        }
      vdmdbs=def_Vdmdbs(dbs);          /* Datenbanken oeffnen */
      OUT_bool(exists_Adt(vdmdbs,adt_name)); /* exists_Adt starten */
      del_str(adt_name);               /* Speicher freigeben */
      del_vdmdbs(vdmdbs);
    }
}

/*----------------------------------------------------------------------*/

void pre_exists_Impl F4(
        int,i,             /* Zaehler fuer verarbeitete Parameter */
        int,argc,          /* Anzahl der Parameter */
        char **,argv,      /* Tupel der Parameter */
        Dbs,dbs)           /* Tupel der Datenbanknamen */
                             /* pre_exists_Impl wertet die Programm- */
                             /* parameter aus, und startet die Funktion */
                             /* exists_Implementation */
{
Adtname adt_name;
Implname impl_name;
Vdmdbs vdmdbs;     /* Datenbanktupel */

  HALLO("pre Impl\n");
  if (argc<i+2)                        /* zu wenig Parameter: */
    {
      fprintf(stderr,"missing functionparameter: no implname given\n");
      if (argc<i+1)
          fprintf(stderr,"missing functionparameter: no adtname given\n");
      Usage();
    }
  else
    {                                  /* Parameter auswerten */
      adt_name=mkc_str(argv[i++]);
      impl_name=mkc_str(argv[i++]);
      while (argc > i)                 /* zu viele Parameter: */
        {
          fprintf(stderr,"superfluos parameter %s ignored\n",argv[i++]);
        }
      vdmdbs=def_Vdmdbs(dbs);          /* Datenbanken oeffnen */
      OUT_bool(exists_Implementation(vdmdbs,adt_name,impl_name));
                                       /* exists_Implementation starten */
      del_str(adt_name);               /* Speicher freigeben */
      del_str(impl_name);
      del_vdmdbs(vdmdbs);
    }
}

/*----------------------------------------------------------------------*/

void pre_exists_Operation F4(
        int,i,             /* Zaehler fuer verarbeitete Parameter */
        int,argc,          /* Anzahl der Parameter */
        char **,argv,      /* Tupel der Parameter */
        Dbs,dbs)           /* Tupel der Datenbanknamen */
                             /* pre_exists_Operation wertet die Programm-*/
                             /* parameter aus, und startet die Funktion */
                             /* exists_Operation */
{
Adtname adt_name;
Implname impl_name;
Opname op_name;
Vdmdbs vdmdbs;            /* Datenbanktupel */

  HALLO("pre Op\n");
  if (argc<i+3)                        /* zu wenig Parameter: */
    {
      fprintf(stderr,"missing functionparameter: no opname given\n");
      if (argc<i+2)
         fprintf(stderr,"missing functionparameter: no implname given\n");
      if (argc<i+1)
          fprintf(stderr,"missing functionparameter: no adtname given\n");
      Usage();
    }
  else
    {                                  /* Parameter auswerten */
      adt_name=mkc_str(argv[i++]);
      impl_name=mkc_str(argv[i++]);
      op_name=mkc_str(argv[i++]);
      while (argc > i)                 /* zu viele Parameter: */
        {
          fprintf(stderr,"superfluos parameter %s ignored\n",argv[i++]);
        }
      vdmdbs=def_Vdmdbs(dbs);          /* Datenbanken oeffnen */
      OUT_bool(exists_Operation(vdmdbs,adt_name,impl_name,op_name));
      del_str(adt_name);               /* Speicher freigeben */
      del_str(impl_name);
      del_str(op_name);
      del_vdmdbs(vdmdbs);
    }
}

/*----------------------------------------------------------------------*/

void pre_exists_Signature F4(
        int,i,             /* Zaehler fuer verarbeitete Parameter */
        int,argc,          /* Anzahl der Parameter */
        char **,argv,      /* Tupel der Parameter */
        Dbs,dbs)           /* Tupel der Datenbanknamen */
                             /* pre_exists_Signature wertet die Programm-*/
                             /* parameter aus, und startet die Funktion */
                             /* exists_Signature */
{
Adtname adt_name,
        dummy_adt_name;   /* found-ADT */
Operation dummy_op;
Opname op_name;
Vdmdbs vdmdbs;           /* Datenbanktupel */

  HALLO("pre ex Sign\n");
  if (argc<i+2)                        /* zu wenig Parameter: */
    {
         fprintf(stderr,"missing functionparameter: no opname given\n");
      if (argc<i+1)
          fprintf(stderr,"missing functionparameter: no adtname given\n");
      Usage();
    }
  else
    {                                  /* Parameter auswerten */
      adt_name=mkc_str(argv[i++]);
      op_name=mkc_str(argv[i++]);
      while (argc > i)                 /* zu viele Parameter: */
        {
          fprintf(stderr,"superfluos parameter %s ignored\n",argv[i++]);
        }
      vdmdbs=def_Vdmdbs(dbs);          /* Datenbanken oeffnen */
      OUT_bool(exists_Signature(vdmdbs,adt_name,op_name,
                                &dummy_adt_name,&dummy_op));
      del_str(adt_name);               /* Speicher freigeben */
      del_str(op_name);
      del_vdmdbs(vdmdbs);
    }
}

/*----------------------------------------------------------------------*/

void pre_exists_Component F4(
        int,i,             /* Zaehler fuer verarbeitete Parameter */
        int,argc,          /* Anzahl der Parameter */
        char **,argv,      /* Tupel der Parameter */
        Dbs,dbs)           /* Tupel der Datenbanknamen */
                             /* pre_exists_Component wertet die Programm-*/
                             /* parameter aus, und startet die Funktion */
                             /* exists_Component */
{
    Adtname adt_name,
	Implname impl_name;
    Opname op_name;
    Compname comp_name;
    Vdmdbs vdmdbs;            /* Datenbanktupel */

    HALLO("pre Comp\n");
    if (argc<i+4)                        /* zu wenig Parameter: */
	{
	    fprintf(stderr,"missing functionparameter: no compname given\n");
	    if (argc<i+3)
		fprintf(stderr,"missing functionparameter: no opname given\n");
	    if (argc<i+2)
		fprintf(stderr,"missing functionparameter: no implname given\n");
	    if (argc<i+1)
		fprintf(stderr,"missing functionparameter: no adtname given\n");
	    Usage();
	}
    else
	{                                  /* Parameter auswerten */
	    adt_name=mkc_str(argv[i++]);
	    impl_name=mkc_str(argv[i++]);
	    op_name=mkc_str(argv[i++]);
	    comp_name=mkc_str(argv[i++]);
	    while (argc > i)                 /* zu viele Parameter: */
		{
		    fprintf(stderr,"superfluos parameter %s ignored\n",argv[i++]);
		}
	    vdmdbs=def_Vdmdbs(dbs);          /* Datenbanken oeffnen */
	    OUT_bool(exists_Component(vdmdbs,adt_name,impl_name,op_name,comp_name));
	    /* exists_Component starten */
	    /* in den Referenzparametern */
	    /* werden found-ADT und found-Impl */
	    /* zurueckgegeben */
	    del_str(adt_name);               /* Speicher freigeben */
	    del_str(impl_name);
	    del_str(op_name);
	    del_str(comp_name);
	    del_vdmdbs(vdmdbs);
	}
}

/*----------------------------------------------------------------------*/

void pre_exists_SignComponent F4(
        int,i,             /* Zaehler fuer verarbeitete Parameter */
        int,argc,          /* Anzahl der Parameter */
        char **,argv,      /* Tupel der Parameter */
        Dbs,dbs)           /* Tupel der Datenbanknamen */
                             /* pre_exists_SignComponent wertet die */
                             /* Programmparameter aus, und startet die */
                             /* Funktion exists_SignComponent */
{
Adtname adt_name;
Opname op_name;
Compname signcomp_name;
Vdmdbs vdmdbs;              /* Datenbanktupel */

  HALLO("pre ex SignComp\n");
  if (argc<i+3)                        /* zu wenig Parameter: */
    {
      fprintf(stderr,"missing functionparameter: no signcompname given\n");
      if (argc<i+2)
        fprintf(stderr,"missing functionparameter: no opname given\n");
      if (argc<i+1)
          fprintf(stderr,"missing functionparameter: no adtname given\n");
      Usage();
    }
  else
    {                                  /* Parameter auswerten */
      adt_name=mkc_str(argv[i++]);
      op_name=mkc_str(argv[i++]);
      signcomp_name=mkc_str(argv[i++]);
      while (argc > i)                 /* zu viele Parameter: */
        {
          fprintf(stderr,"superfluos parameter %s ignored\n",argv[i++]);
        }
      vdmdbs=def_Vdmdbs(dbs);          /* Datenbanken oeffnen */
      OUT_bool(exists_SignComponent(vdmdbs,adt_name,op_name,signcomp_name));
      del_str(adt_name);               /* Speicher freigeben */
      del_str(op_name);
      del_str(signcomp_name);
      del_vdmdbs(vdmdbs);
    }
}

/*----------------------------------------------------------------------*/

void pre_gets_Component F4(
        int,i,             /* Zaehler fuer verarbeitete Parameter */
        int,argc,          /* Anzahl der Parameter */
        char **,argv,      /* Tupel der Parameter */
        Dbs,dbs)           /* Tupel der Datenbanknamen */
                             /* pre_gets_Component wertet die Programm- */
                             /* parameter aus, und startet die Funktion */
                             /* gets_Component */
{
Adtname adt_name;
Implname impl_name;
Opname op_name;
Compname comp_name;
Vdmdbs vdmdbs;             /* Datenbanktupel */
Component comp;

  HALLO("pre gets Comp\n");
  if (argc<i+4)                        /* zu wenig Parameter: */
    {
      fprintf(stderr,"missing functionparameter: no compname given\n");
      if (argc<i+3)
        fprintf(stderr,"missing functionparameter: no opname given\n");
      if (argc<i+2)
         fprintf(stderr,"missing functionparameter: no implname given\n");
      if (argc<i+1)
          fprintf(stderr,"missing functionparameter: no adtname given\n");
      Usage();
    }
  else
    {                                  /* Parameter auswerten */
      adt_name=mkc_str(argv[i++]);
      impl_name=mkc_str(argv[i++]);
      op_name=mkc_str(argv[i++]);
      comp_name=mkc_str(argv[i++]);
      while (argc > i)                 /* zu viele Parameter: */
        {
          fprintf(stderr,"superfluos parameter %s ignored\n",argv[i++]);
        }
      vdmdbs=def_Vdmdbs(dbs);          /* Datenbanken oeffnen */
      comp=gets_Component(vdmdbs,adt_name,impl_name,op_name,comp_name);
      puts_str(comp,stdout);           /* Ergebnis auswerten */
      del_str(adt_name);               /* Speicher freigeben */
      del_str(impl_name);
      del_str(op_name);
      del_str(comp_name);
      del_vdmdbs(vdmdbs);
      del_str(comp);
     }
}

/*----------------------------------------------------------------------*/

void pre_getsx_Component F4(
        int,i,             /* Zaehler fuer verarbeitete Parameter */
        int,argc,          /* Anzahl der Parameter */
        char **,argv,      /* Tupel der Parameter */
        Dbs,dbs)           /* Tupel der Datenbanknamen */
                             /* pre_getsx_Component wertet die Programm- */
                             /* parameter aus, und startet die Funktion */
                             /* getsx_Component */
{
Adtname adt_name;
Implname impl_name;
Opname op_name;
Compname comp_name;
Vdmdbs vdmdbs;            /* Datenbanktupel */
Component comp;

  HALLO("pre gets Comp\n");
  if (argc<i+4)                        /* zu wenig Parameter: */
    {
      fprintf(stderr,"missing functionparameter: no compname given\n");
      if (argc<i+3)
        fprintf(stderr,"missing functionparameter: no opname given\n");
      if (argc<i+2)
         fprintf(stderr,"missing functionparameter: no implname given\n");
      if (argc<i+1)
          fprintf(stderr,"missing functionparameter: no adtname given\n");
      Usage();
    }
  else
    {                                  /* Parameter auswerten */
      adt_name=mkc_str(argv[i++]);
      impl_name=mkc_str(argv[i++]);
      op_name=mkc_str(argv[i++]);
      comp_name=mkc_str(argv[i++]);
      while (argc > i)                 /* zu viele Parameter: */
        {
          fprintf(stderr,"superfluos parameter %s ignored\n",argv[i++]);
        }
      vdmdbs=def_Vdmdbs(dbs);          /* Datenbanken oeffnen */
      comp=getsx_Component(vdmdbs,adt_name,impl_name,op_name,comp_name);
      puts_str(comp,stdout);           /* Ergebnis auswerten */
      del_str(adt_name);               /* Speicher freigeben */
      del_str(impl_name);
      del_str(op_name);
      del_str(comp_name);
      del_str(comp);
      del_vdmdbs(vdmdbs);
     }
}

/*----------------------------------------------------------------------*/

void pre_gets_SignComponent F4(
        int,i,             /* Zaehler fuer verarbeitete Parameter */
        int,argc,          /* Anzahl der Parameter */
        char **,argv,      /* Tupel der Parameter */
        Dbs,dbs)           /* Tupel der Datenbanknamen */
                             /* pre_gets_SignComponent wertet die */
                             /* Programmparameter aus, und startet die */
                             /* Funktion gets_SignComponent */
{
Adtname adt_name;
Opname op_name;
Compname signcomp_name;
Vdmdbs vdmdbs;              /* Datenbanktupel */
Component signcomp;

  HALLO("pre gets SignComp\n");
  if (argc<i+3)                        /* zu wenig Parameter: */
    {
      fprintf(stderr,"missing functionparameter: no signcompname given\n");
      if (argc<i+2)
         fprintf(stderr,"missing functionparameter: no opname given\n");
      if (argc<i+1)
          fprintf(stderr,"missing functionparameter: no adtname given\n");
      Usage();
    }
  else
    {                                  /* Parameter auswerten */
      adt_name=mkc_str(argv[i++]);
      op_name=mkc_str(argv[i++]);
      signcomp_name=mkc_str(argv[i++]);
      while (argc > i)                 /* zu viele Parameter: */
        {
          fprintf(stderr,"superfluos parameter %s ignored\n",argv[i++]);
        }
      vdmdbs=def_Vdmdbs(dbs);          /* Datenbanken oeffnen */
      signcomp=gets_SignComponent(vdmdbs,adt_name,op_name,signcomp_name);
      puts_str(signcomp,stdout);       /* Ergebnis auswerten */
      del_str(adt_name);               /* Speicher freigeben */
      del_str(op_name);
      del_str(signcomp_name);
      del_vdmdbs(vdmdbs);
      del_str(signcomp);
    }
}

/*----------------------------------------------------------------------*/

void pre_getsx_SignComponent F4(
        int,i,             /* Zaehler fuer verarbeitete Parameter */
        int,argc,          /* Anzahl der Parameter */
        char **,argv,      /* Tupel der Parameter */
        Dbs,dbs)           /* Tupel der Datenbanknamen */
                             /* pre_getsx_SignComponent wertet die */
                             /* Programmparameter aus, und startet die */
                             /* Funktion getsx_SignComponent */
{
Adtname adt_name;
Opname op_name;
Compname signcomp_name;
Vdmdbs vdmdbs;              /* Datenbanktupel */
Component signcomp;

  HALLO("pre gets SignComp\n");
  if (argc<i+3)                        /* zu wenig Parameter: */
    {
      fprintf(stderr,"missing functionparameter: no signcompname given\n");
      if (argc<i+2)
         fprintf(stderr,"missing functionparameter: no opname given\n");
      if (argc<i+1)
          fprintf(stderr,"missing functionparameter: no adtname given\n");
      Usage();
    }
  else
    {                                  /* Parameter auswerten */
      adt_name=mkc_str(argv[i++]);
      op_name=mkc_str(argv[i++]);
      signcomp_name=mkc_str(argv[i++]);
      while (argc > i)                 /* zu viele Parameter: */
        {
          fprintf(stderr,"superfluos parameter %s ignored\n",argv[i++]);
        }
      vdmdbs=def_Vdmdbs(dbs);          /* Datenbanken oeffnen */
      signcomp=getsx_SignComponent(vdmdbs,adt_name,op_name,signcomp_name);
      puts_str(signcomp,stdout);       /* Ergebnis auswerten */
      del_str(adt_name);               /* Speicher freigeben */
      del_str(op_name);
      del_str(signcomp_name);
      del_vdmdbs(vdmdbs);
      del_str(signcomp);
    }
}

/*----------------------------------------------------------------------*/

void get_Docu F4(
        int,i,             /* Zaehler fuer verarbeitete Parameter */
        int,argc,          /* Anzahl der Parameter */
        char **,argv,      /* Tupel der Parameter */
        Dbs,dbs)           /* Tupel der Datenbanknamen */
                             /* get_Docu ermittelt die Dokumentations- */
                             /* komponente fuer ADTs, Impls oder Ops. */
                             /* Es wird nur in der ersten Datenbank des */
                             /* Tupels dbs gesucht. */
                             /* Die Zahl der Parameter entscheidet, ob */
                             /* fuer ADT (1 Parameter), Impl (2 Params) */
                             /* oder Operation (3 Parameter). Existiert */
                             /* keine Dokukomponente, wird eine leere */
                             /* Komponente zurueckgegeben */
{
Adtname adt_name;
Implname impl_name;
Opname op_name;
Component docu_comp;
Vdmdb vdmdb;             /* Datenbank */

  HALLO("get_Docu\n");
  if (argc<i+1)                        /* zu wenig Parameter: */
    {
      fprintf(stderr,"missing functionparameter: no adtname given\n");
      Usage();
    }
  else
    {
      vdmdb=hd_vdmdbs(def_Vdmdbs(dbs));/* Datenbank oeffnen */
      adt_name=mkc_str(argv[i++]);     /* ADT-Namen ermitteln */
      if (argc<i+1)                    /* => ADT-DocuComp selektieren */
          docu_comp=get_Component(vdmdb,adt_name,dmCOM,dmBASIC,dmDOCU);
      else
        {
          impl_name=mkc_str(argv[i++]);
          if (argc<i+1)              /* => Impl-DocuComp selektieren */
             docu_comp=get_Component(vdmdb,adt_name,impl_name,dmBASIC,dmDOCU);
          else
            {                          /* => Op-DocuComp selektieren */
              op_name=mkc_str(argv[i++]);
              docu_comp=get_Component(vdmdb,adt_name,impl_name,op_name,dmDOCU);

              while (argc > i )        /* zu viele Parameter: */
                {
                  fprintf(stderr,"superfluos parameter %s ignored\n",
                              argv[i++]);
                }
              del_str(op_name);
            }
          del_str(impl_name);
        }
      puts_str(docu_comp,stdout);      /* DocuComp ausgeben */
      del_str(adt_name);               /* Speicher freigeben */
      del_str(docu_comp);
      del_vdmdb(vdmdb);
    }
}

/*----------------------------------------------------------------------*/

int main F2(
        int, argc,
        char **,argv)
                             /* Das Hauptprogramm liest die Programm- */
                             /* parameter ein und startet die Funktionen */
{
char *p;                /* Zeiger auf zu bearbeiteten Parameter */
int i=1;                /* Zaehler fuer zu bearbeiteten Parameter */
Str tracestring;        /* String fuer Traceparameter */
Vdmdbname vdmdb_name;   /* Datenbankname */
Dbs dbs;                /* Datenbanknamentupel */

  init_vdmdm();

  if (argc==1)                         /* keine Programmparameter */
    Usage();                           /* => usage ausgeben */
  else
    {
      dbs=mk_empty_dbs();
      while (argc>i)                   /* Funktionen starten */
        {
          p=argv[i++];
          if (*p++=='-')
            switch(*p)
              {
                case 'T' : /* Trace-variable setzen */
                           tracestring=mkc_str(argv[i-1]);
                           gptrace=len_str(tracestring)-1;
                           del_str(tracestring);
                           break;
                case 'B' : /* Datenbanknamentupel erstellen */
                           vdmdb_name=mkc_str(argv[i-1]);
                           dbs=appr_dbs(dbs,sub_tup_str(vdmdb_name,
                                            3,len_str(vdmdb_name)));
                           break;
                case 'v' : pre_exists_Vdmdbs(i,argc,argv,dbs);i=argc;break;
                case 'a' : pre_exists_Adt(i,argc,argv,dbs);i=argc;break;
                case 'i' : pre_exists_Impl(i,argc,argv,dbs);i=argc;break;
                case 'o' : pre_exists_Operation(i,argc,argv,dbs);i=argc;break;
                case 's' : pre_exists_Signature(i,argc,argv,dbs);i=argc;break;
                case 'c' : pre_exists_Component(i,argc,argv,dbs);i=argc;break;
                case 'S' : pre_exists_SignComponent(i,argc,argv,dbs);i=argc;break;
                case 'g' : pre_gets_Component(i,argc,argv,dbs);i=argc;break;
                case 'x' : pre_getsx_Component(i,argc,argv,dbs);i=argc;break;
                case 'G' : pre_gets_SignComponent(i,argc,argv,dbs);i=argc;break;
                case 'X' : pre_getsx_SignComponent(i,argc,argv,dbs);i=argc;break;
                case 'd' : get_Docu(i,argc,argv,dbs);i=argc;break;
                default  : fprintf(stderr,"wrong function parameter\n");
              }
          else
            {
              fprintf(stderr,"function parameter has to start with -\n");
              Usage();
            }
        }
      TRACE1({ profile_dbs(); profile_vdmdbs(); profile_str(); })
    }
  finish_vdmdm();

  return 0;
}
