/*----------------------------------------------------------------------

$Id: dbs.c,v 1.5 1996/09/08 11:45:13 vdm Exp $

----------------------------------------------------------------------

   Aufgabe:   "dbs.c" realisiert die Funktionen fuer Selektions-
              operationen und Existenzpruefungen 
              fuer die Datenbankelemente.

----------------------------------------------------------------------

 Changes:
 --------

   $Log: dbs.c,v $
   Revision 1.5  1996/09/08 11:45:13  vdm
   error in param_defined removed:
   Implid was not checked against NIL, before accessing the database
   this case occurs for PREDEF domain ids
   "s_str_implid(...)" was not called in parameter expression for "get_params"
   "Str" and "Implid" are in C the same type, so no compiler error was
   issued ( :-( ).

   Revision 1.4  1996/09/07 14:12:09  vdm
   vdm as client

   Revision 1.3  1996/01/15 10:02:11  vdm
   cvs backup

 * Revision 1.2  1993/09/01  18:11:39  vdm
 * Hans-Martins Version vom 1.9.93 als neue Hauptversion
 *
 * Revision 1.2  1993/07/10  07:29:50  hmh
 * Handling of 'VDMROOT' environment variable corrected:
 *   - removed reading from domaincomp ('vdmmain') completely
 *   - in Dm: if VDMROOT is undefined, set to default path ('/usr').
 *

----------------------------------------------------------------------*/

#include "header.h"          /* allgemeines Headerfile */
#include "dbs.h"             /* eigenes Headerfile */
#include "genext.h"          /* Headerfile fuer fold-Funktionen */
#include "invar.h"           /* exist_wf_Operation, exist_wf_Signature */


Bool run_as_client = FALSE;
char * host = "localhost";
char * port = "4711";

/*----------------------------------------------------------------------*/

Str dmADTTYPE   ;
Str dmAUX       ;
Str dmAUXILIARY ;
Str dmBASIC     ;
Str dmCOM       ;
Str dmDOCU      ;
Str dmINTERNAL  ;
Str dmPOST      ;
Str dmREAL      ;
Str dmSIGN      ;
Str dmSTDLIB    ;
Str dmVDMLIB    ;

void init_vdmdm F0()
{
    dmADTTYPE  =mkc_str("adttype");
    dmAUX      =mkc_str("aux");
    dmAUXILIARY=mkc_str("AUXILIARY");
    dmBASIC    =mkc_str("basic");
    dmCOM      =mkc_str("com");
    dmDOCU     =mkc_str("docu");
    dmINTERNAL =mkc_str("INTERNAL");
    dmPOST     =mkc_str("post");
    dmREAL     =mkc_str("REAL");
    dmSIGN     =mkc_str("sign");
    dmSTDLIB   =mkc_str("stdlib");
    dmVDMLIB   =mkc_str("vdmlib");
}

/*----------------------------------------------------------------------*/

void finish_vdmdm F0()
{
    del_str(dmADTTYPE  );
    del_str(dmAUX      );
    del_str(dmAUXILIARY);
    del_str(dmBASIC    );
    del_str(dmCOM      );
    del_str(dmDOCU     );
    del_str(dmINTERNAL );
    del_str(dmPOST     );
    del_str(dmREAL     );
    del_str(dmSIGN     );
    del_str(dmSTDLIB   );
    del_str(dmVDMLIB   );
}

/*----------------------------------------------------------------------*/

Bool is_include F1(char *,s)
     /* includes must begin at column 1 */
     /* no leading blanks or tabs */
     /* one blank between directive and file name */
{
    return *s++ == '#'
	&& *s++ == 'i'
	&& *s++ == 'n'
	&& *s++ == 'c'
	&& *s++ == 'l'
	&& *s++ == 'u'
	&& *s++ == 'd'
	&& *s++ == 'e'
	&& *s++ == ' '
	&& *s   == '<';
}

/*----------------------------------------------------------------------*/

Str solve2_include F2(
        char *,s,             /* contains include directive */
        Vdmdbs,vdmdbs)
{
    char
	* adtcs,
	* implcs,
	* opcs,
	* compcs;

    Str
	adtstr,
	implstr,
	opstr,
	compstr;

    char * l;
    Str r;

    HALLO("solve2_include\n");
    TRACE1( { fprintf(stderr,"%s\n",s); })

           /* scan include file */
           /* assumed syntax : <adt/impl/op/comp> */
    l = s;
    while (*l++ != '<');       /* search start of adt name */
    adtcs=l;                   /* remember start of adt name */
    while (*l != '/') l++;     /* search end of adt name */
    *l++=0;                    /* split string */
    implcs=l;                  /* remember start of impl name */
    while (*l != '/') l++;     /* search end of impl name */
    *l++=0;                    /* split string */
    opcs=l;                    /* remember start of op name */
    while (*l != '/') l++;     /* search end of op name */
    *l++=0;                    /* split string */
    compcs=l;                  /* remember start of comp name */
    while (*l != '>') l++;     /* search end of comp name */
    *l++=0;                    /* split string */

    adtstr  = mkc_str(adtcs);
    implstr = mkc_str(implcs);
    opstr   = mkc_str(opcs);
    compstr = mkc_str(compcs);

    TRACE1( { OUT4_str(adtstr,implstr,opstr,compstr); })

    r = getsx_Component(vdmdbs,adtstr,implstr,opstr,compstr);

    del_str(adtstr);
    del_str(implstr);
    del_str(opstr);
    del_str(compstr);

    return r;
}

/*----------------------------------------------------------------------*/

Str solve1_include F2(
        char *,s,
        Vdmdbs,vdmdbs)
{
    char * l;

    HALLO("solve1_include\n");
    TRACE1( { fprintf(stderr,"%s\n",s); })

    l = s;
    while (*l)
	{
	    if (is_include(l))
		{
		    char *ia = l;                        /* include line found */
		    char *ie = ia;
		    
		    while (*ie && *ie != '\n') ie++;     /* search end of include line */

		    if (*ie) *ie++ = 0;                    /* split string */
		    *ia++ = 0;
		    TRACE1( { fprintf(stderr,"prefix:\n%s\n",s); })
		    TRACE1( { fprintf(stderr,"include:\n%s\n",ia); })
		    TRACE1( { fprintf(stderr,"rest:\n%s\n",ie); })

		    return
			conc3_str(mkc_str(s),                 /* take prefix */
				  solve2_include(ia,vdmdbs),  /* append include file */
				  solve1_include(ie,vdmdbs)); /* process rest */
		}
	    while (*l && (*l++ != '\n'));             /*skip to next line */
	}

    return mkc_str(s);
}

/*----------------------------------------------------------------------*/

Str solve_include F2(
        Str,s,
        Vdmdbs,vdmdbs)
{
    char *cs;
    Str r;

    HALLO("solve_include\n");
    TRACE1( { OUT_str(s); } )

    cs = malloc(len_str(s)+1);       /* Speicher fuer C-String holen */

    if ( ! cs) {}
    cpc_str(s,cs);
    del_str(s);
    r = solve1_include(cs,vdmdbs);

    HALLO("solve_include result:\n");
    TRACE1( { OUT_str(r); } )

    free(cs);
    return r;
}

/*----------------------------------------------------------------------*/

static
Vdmdbs def_local_Vdmdbs F1(Dbs,dbs)
                             /* oeffnet alle Datenbanken aus dem */
                             /* Datenbanknamentupel 'dbs' */
                             /* Suchstrategie, falls dbs leer: */
                             /*      DEFAULT_VDMDB_NAME aus dbs.h */
                             /* der Wert dbs wird verbraucht */
{
    char * vdmdb_name_as_char;
    Vdmdbname vdmdb_name;
    Vdmdbs vdmdbs;

    if (is_empty_dbs(dbs))
	{                                          /* Falls DB-Namentupel leer */
	    vdmdb_name_as_char=getenv(VDMROOT);    /* VDMROOT lesen */

	    if (vdmdb_name_as_char == NULL)
		{
		    vdmdb_name_as_char = DEFAULT_VDMROOT;
		}
	    vdmdb_name=mkc_str(vdmdb_name_as_char);
	    vdmdb_name=conc_path_str(vdmdb_name,mkc_str(DEFAULT_VDMDB_NAME));
                                                    /* DEFAULT_VDMDB_NAME */
	    dbs=appr_dbs(dbs,vdmdb_name);
	}

    vdmdbs=mk_tuple_dbs(dbs);                    /* Datenbanken oeffnen */
    del_dbs(dbs);                                /* Speicher freigeben */
    return vdmdbs;
}

/*----------------------------------------------------------------------*/

static
Bool exists_local_Vdmdbs F1(Vdmdbs,vdmdbs)
                             /* Ueberprueft die Existenz eines DB-Tupels*/
                             /* Ein Datenbanktupel exisiert, */
                             /* wenn alle Elemente existieren */
{
    HALLO("exists Vdmdbs\n");
    return
	(IF is_empty_vdmdbs(vdmdbs)           /* falls leer: */
	 THEN FALSE                           /* FALSE */
	 ELSE fold_exists_vdmdbs(vdmdbs));    /* sonst: fold_exists_vdmdbs */
}

/*----------------------------------------------------------------------*/

Bool exist_Adt F2(Vdmdb,vdmdb,Adtname,adt_name)
                             /* Ueberprueft die Existenz eines ADT */
                             /* in einer Datenbank */
{
    return
	is_in_vdmdb(adt_name,vdmdb);
}

/*----------------------------------------------------------------------*/

static
Bool exists_local_Adt F2(Vdmdbs,vdmdbs,Adtname,adt_name)
                             /* Ueberpueft die Existenz eines ADTs in */
                             /* den uebergebenen Datenbanken */
                             /* Dazu wird die Funktion exist_Adt fuer */
                             /* jede Datenbank ausgefuehrt. */
{
    Bool bool;
    Nat0 index,length;
    Vdmdb vdmdb;

    HALLO("exists Adt:\n");
    TRACE1( { OUT_str(adt_name); } )

    LET length=len_vdmdbs(vdmdbs) IN

    bool=FALSE;
    index=0;

    while ((index<length) AND (NOT bool))
	{
	    index++;
	    vdmdb=sel_vdmdbs(vdmdbs,index);
	    bool=exist_Adt(vdmdb,adt_name);
	}
    return bool;
}

/*----------------------------------------------------------------------*/

Bool exist_Implementation F3(Vdmdb,vdmdb,Adtname,adt_name,Implname,impl_name)
                             /* Ueberprueft die Existenz einer */
                             /* Implementation in einer Datenbank */
{
    return
	exist_Adt(vdmdb,adt_name)
	AND
	is_in_adt(impl_name,sel_vdmdb(vdmdb,adt_name));
}

/*----------------------------------------------------------------------*/

static
Bool exists_local_Implementation F3(Vdmdbs,vdmdbs,Adtname,adt_name,Implname,impl_name)
                             /* Ueberprueft die Existenz einer Impl. in */
                             /* den uebergebenen Datenbanken */
                             /* Dazu wird die Funktion exist_Impl fuer */
                             /* jede Datenbank ausgefuehrt. */
{
    Bool bool;
    Nat0 index,length;
    Vdmdb vdmdb;

    HALLO("exists Impl:\n");
    TRACE1( { OUT2_str(adt_name,impl_name); } )

    LET length=len_vdmdbs(vdmdbs) IN

    bool=FALSE;
    index=0;

    while ((index<length) AND (NOT bool))
	{
	    index++;
	    vdmdb = sel_vdmdbs(vdmdbs,index);
	    bool = exist_Implementation(vdmdb,adt_name,impl_name);
	}
    return bool;
}

/*----------------------------------------------------------------------*/

Bool exist_Operation F4(
     Vdmdb,vdmdb,
     Adtname,adt_name,
     Implname,impl_name,
     Opname,op_name)
                             /* Ueberprueft die Existenz einer Operation*/
                             /* in einer Datenbank */
{
    Adtname found_in_adt;
    Implname found_in_impl;
    Operation found_op;

    return
	exist_Implementation(vdmdb,adt_name,impl_name)
	AND
	exist_wf_Operation(vdmdb,adt_name,impl_name,op_name,
			   &found_in_adt,&found_in_impl,&found_op);
}

/*----------------------------------------------------------------------*/

Bool exists_local_Operation F4(
     Vdmdbs,vdmdbs,
     Adtname,adt_name,
     Implname,impl_name,
     Opname,op_name)
                             /* Ueberprueft die Existenz einer Operation*/
                             /* in den uebergebenen Datenbanken */
{
    Bool bool;
    Nat0
	index,
	length;
    Vdmdb vdmdb;

    HALLO("exist Op:\n");
    TRACE1( { OUT3_str(adt_name,impl_name,op_name); } )

    LET length=len_vdmdbs(vdmdbs) IN

    bool=FALSE;
    index=0;

    while ((index<length) AND (NOT bool))
	{
	    index++;
	    vdmdb = sel_vdmdbs(vdmdbs,index);
	    bool = exist_Operation(vdmdb,adt_name,impl_name,op_name);
    }

    return bool;
}

/*----------------------------------------------------------------------*/

Bool exist_Component F8(
			Vdmdb,vdmdb,
			Adtname,adt_name,
			Implname,impl_name,
			Opname,op_name,
			Compname,comp_name,
			Adtname *,found_in_adt,
			Implname *,found_in_impl,
			Component *,found_comp)
                             /* Ueberprueft die Existenz eines Component*/
                             /* in einer Datenbank */
                             /* In den Referenzparametern werden */
                             /* found-ADT und found-Implementation */
                             /* und die Komponente zurueckgegeben */
{
    Operation found_operation;

    return
	exist_Implementation(vdmdb,adt_name,impl_name)
	AND
	exist_wf_Operation(vdmdb,adt_name,impl_name,op_name,
			   found_in_adt,found_in_impl,&found_operation)
	AND
	is_in_dom_operation(comp_name,found_operation,found_comp);
}

/*----------------------------------------------------------------------*/

static
Bool exists_local_Component F5(
			 Vdmdbs,vdmdbs,
			 Adtname,adt_name,
			 Implname,impl_name,
			 Opname,op_name,
			 Compname,comp_name)

{
    Bool bool;
    Nat0 index,length;
    Vdmdb vdmdb;

    Adtname found_in_adt;
    Implname found_in_impl;
    Component found_comp;

    HALLO("exist Comp:\n");
    TRACE1( { OUT4_str(adt_name,impl_name,op_name,comp_name); } )

    LET length=len_vdmdbs(vdmdbs) IN

    bool=FALSE;
    index=0;

    while ((index<length) AND (NOT bool))
	{
	    index++;
	    vdmdb=sel_vdmdbs(vdmdbs,index);
	    bool=exist_Component(vdmdb,
				 adt_name,impl_name,op_name,comp_name,
				 &found_in_adt,&found_in_impl,
				 &found_comp);
	}
    return bool;
}

/*----------------------------------------------------------------------*/

Component get_Component F5(
			   Vdmdb,vdmdb,
			   Adtname,adt_name,
			   Implname,impl_name,
			   Opname,op_name,
			   Compname,comp_name)
{
    Adtname found_in_adt;
    Implname found_in_impl;
    Component found_comp;

    return
	IF exist_Component(vdmdb,
			   adt_name,impl_name,op_name,comp_name,
			   &found_in_adt,&found_in_impl,
			   &found_comp)
	THEN copy_str(found_comp)
	ELSE mk_empty_str();
}

/*----------------------------------------------------------------------*/

static
Component gets_local_Component F5(
				  Vdmdbs,vdmdbs,
				  Adtname,adt_name,
				  Implname,impl_name,
				  Opname,op_name,
				  Compname,comp_name)
                             /* Selektiert einen Component aus */
                             /* den uebergebenen Datenbanken */
                             /* Dazu wird der Component in den DBs */
                             /* mit exist_Component gesucht. */
{
    Bool bool;
    Nat0
	index,
	length;
    Vdmdb vdmdb;
    Adtname found_in_adt;
    Implname found_in_impl;
    Component found_comp;

    HALLO("gets Comp:\n");
    TRACE1( { OUT4_str(adt_name,impl_name,op_name,comp_name); } )

    LET length=len_vdmdbs(vdmdbs) IN

    bool=FALSE;
    index=0;

    while ((index<length) AND (NOT bool))
	{
	    index++;
	    vdmdb=sel_vdmdbs(vdmdbs,index);
	    bool=exist_Component(vdmdb,
				 adt_name,impl_name,op_name,comp_name,
				 &found_in_adt,&found_in_impl,
				 &found_comp);
	}
    return
	(IF bool
	 THEN copy_str(found_comp)
	 ELSE mk_empty_str());

}

/*----------------------------------------------------------------------*/

static
Component getsx_local_Component F5(
			     Vdmdbs,vdmdbs,
			     Adtname,adt_name,
			     Implname,impl_name,
			     Opname,op_name,
			     Compname,comp_name)
                             /* Selektiert eine Komponente aus */
                             /* den uebergebenen Datenbanken */
                             /* und loesst die includes auf */
{
    HALLO("getsx Comp:\n");
    TRACE1( { OUT4_str(adt_name,impl_name,op_name,comp_name); } )

    return
	solve_include(
		      gets_local_Component(vdmdbs,adt_name,impl_name,op_name,comp_name),
		      vdmdbs);
}

/*----------------------------------------------------------------------*/

Bool exist_Signature F5(
			Vdmdb,vdmdb,
			Adtname,adt_name,
			Opname,op_name,
			Adtname *,found_in_adt,
			Operation *,found_op)
                             /* Ueberprueft die Existenz einer Signatur */
                             /* in einer Datenbank */
{
    return
	exist_Adt(vdmdb,adt_name)
	AND
	exist_wf_Signature(vdmdb,adt_name,op_name,found_in_adt,found_op);
}

/*----------------------------------------------------------------------*/

Bool exists_Signature F5(
			 Vdmdbs,vdmdbs,
			 Adtname,adt_name,
			 Opname,op_name,
			 Adtname *,found_in_adt,
			 Operation *,found_operation)
                             /* Ueberprueft die Existenz einer Signatur */
                             /* in den uebergebenen Datenbanken */
{
    Bool bool;
    Nat0
	index,
	length;
    Vdmdb vdmdb;

    HALLO("exists Sign:\n");
    TRACE1( { OUT2_str(adt_name,op_name); } )

    LET length=len_vdmdbs(vdmdbs) IN

    bool=FALSE;
    index=0;

    while ((index<length) AND (NOT bool))
	{
	    index++;
	    vdmdb = sel_vdmdbs(vdmdbs,index);
	    bool = exist_Signature(vdmdb,adt_name,op_name,
				   found_in_adt,found_operation);
	}
    return bool;
}



/*----------------------------------------------------------------------*/

Bool exist_SignComponent F6(Vdmdb,vdmdb,
                            Adtname,adt_name,
                            Opname,op_name,
                            Compname,signcomp_name,
                            Adtname *,found_in_adt,
                            Component *,found_comp)
                             /* Ueberprueft die Existenz einer Signatur-*/
                             /* komponente in einer Datenbank */
{
    Operation found_op;

    return
	exist_Adt(vdmdb,adt_name)
	AND
	exist_wf_Signature(vdmdb,adt_name,op_name,found_in_adt,&found_op)
	AND
	is_in_dom_operation(signcomp_name,found_op,found_comp);
}

/*----------------------------------------------------------------------*/

static
Bool exists_local_SignComponent F4(
			     Vdmdbs,vdmdbs,
			     Adtname,adt_name,
			     Opname,op_name,
			     Compname,signcomp_name)
                             /* Ueberprueft die Existenz einer Signatur-*/
                             /* komponente in den uebergebenen DBen */
                             /* Dazu wird die Funktion */
                             /* exist_SignComponent */
                             /* fuer jede Datenbank ausgefuehrt. */
{
    Bool bool;
    Nat0
	index,
	length;
    Vdmdb vdmdb;
    Adtname found_in_adt;
    Component found_comp;

    HALLO("exists SignComp:\n");
    TRACE1( { OUT2_str(adt_name,op_name); } )

    LET length=len_vdmdbs(vdmdbs) IN

    bool=FALSE;
    index=0;

    while ((index<length) AND (NOT bool))
	{
	    index++;
	    vdmdb=sel_vdmdbs(vdmdbs,index);
	    bool=exist_SignComponent(vdmdb,adt_name,op_name,signcomp_name,
				     &found_in_adt,&found_comp);
	}
    return bool;
}

/*----------------------------------------------------------------------*/

static
Component gets_local_SignComponent F4(
				Vdmdbs,vdmdbs,
				Adtname,adt_name,
				Opname,op_name,
				Compname,signcomp_name)
                             /* Selektiert eine SignKomponente aus */
                             /* den uebergebenen Datenbanken */
{
    Bool bool;
    Nat0 index,length;
    Vdmdb vdmdb;
    Adtname found_in_adt;
    Component found_comp;

    HALLO("gets SignComp:\n");
    TRACE1( { OUT3_str(adt_name,op_name,signcomp_name); } )

    LET length=len_vdmdbs(vdmdbs) IN

    bool=FALSE;
    index=0;

    while ((index<length) AND (NOT bool))
	{
	    index++;
	    vdmdb = sel_vdmdbs(vdmdbs,index);
	    bool=exist_SignComponent(vdmdb,adt_name,op_name,signcomp_name,
				     &found_in_adt,&found_comp);
	}
    return
	IF bool
	THEN copy_str(found_comp)
	ELSE mk_empty_str();
}

/*----------------------------------------------------------------------*/

static
Component getsx_local_SignComponent F4(
				 Vdmdbs,vdmdbs,
				 Adtname,adt_name,
				 Opname,op_name,
				 Compname,signcomp_name)
                             /* Selektiert eine SignKomponente aus */
                             /* den uebergebenen Datenbanken */
                             /* und loesst includes auf */
{

    HALLO("gets SignComp:\n");
    TRACE1( { OUT3_str(adt_name,op_name,signcomp_name); } )

    return
	solve_include(
		      gets_SignComponent(vdmdbs,adt_name,op_name,signcomp_name),
		      vdmdbs);
}

/*----------------------------------------------------------------------*/

#if CLIENT

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <string.h>

extern int domtrace;
extern void error H2(Str, Str);
#define Empty   mk_empty_str()

#define MEMDIMSTEP 1024

static char * out_buffer = 0;
static size_t out_buffer_size = 0;

static char * in_buffer = 0;
static size_t in_buffer_size = 0;

static int sockfd, portfd;
static struct sockaddr_in serv_addr;
static struct hostent *hp;
static struct servent *sp;


static int rpc_MakeClient H0();
static int rpc_Exec H0();
static int rpc_Close H0();

static
size_t writen F2(char *, ptr, size_t, nbytes)
{
    size_t
	nleft,
	nwritten;
  
    nleft = nbytes;
    while (nleft > 0) {
  	nwritten = write(sockfd, ptr, nleft);
  	if (nwritten <= 0)
	    return nwritten;
  	
  	nleft -= nwritten;
  	ptr += nwritten;
    }
    return (nbytes - nleft);
}

static
size_t readn F2(char *, ptr, size_t, nbytes)
{
    size_t
	nleft,
	nread;

    nleft = nbytes;
    while (nleft > 0) {
	nread = read(sockfd, ptr, nleft);

	if (nread < 0)
	    return nread;
	else
	    if (nread==0)
		break;
	
	nleft -= nread;
	ptr += nread;
    }

    return (nbytes - nleft);
}

static
char * readcr F0()
{
    size_t n,rc;
    char c;

    for (n=0;;) {
	if (n+1 >= out_buffer_size) {
	    char * p;

	    p = malloc(out_buffer_size += MEMDIMSTEP);
	    if ( ! p ) {
		error(mkc_str("can't allocate heap memory"),Empty);
		exit(1);
	    }
	    if (out_buffer)
		{
		    memcpy(p,out_buffer,n);
		    free(out_buffer);
		}
	    out_buffer = p;	
	}
	if ( (rc = readn(&c, 1) ) == 1) {
	    if (c == '\n') 
		break;
	    out_buffer[n++] = c;
	} else {
	    return NULL;
	}		
    } 
    out_buffer[n]=0; 

    return out_buffer;
}

static
char *readdata F1(size_t, size)
{
    while (size + 1 >= out_buffer_size)
	{
	    char * p;
	    p = malloc(out_buffer_size += MEMDIMSTEP);
	    if ( ! p ) {
		error(mkc_str("can't allocate heap memory"),Empty);
		exit(1);
	    }
	    if (out_buffer)
		free(out_buffer);
	    out_buffer = p;	
	}
    
    out_buffer[0] = 0;
    if (readn(out_buffer, size) != size) {
	return NULL;
    } else {
	out_buffer[size]=0;
	return out_buffer;
    }
}

static
int writedata F1(char *,ptr)
{
    size_t l = strlen(ptr);

    return
	writen(ptr, l) == l;   
}

static
int rpc_MakeClient F0()
{
    char *s;

    if ( (s = readcr() ) == NULL)
	return 0;

    return
	(strstr(s, "accepted") == NULL) ? -1 : 0;
}

static
int rpc_Close F0()
{
    char *s;
  
    if ( ! writedata("release.request\n") )
	return -1;

    if ( (s = readcr()) == NULL)
	return -1;

    return
	(strstr(s,"release.response") == NULL) ? -1 : 0;
}

static
int rpc_Exec F0()
{
    char *s,
	len[12],
	status[10];
    size_t n;

    if ( ! writedata("data.request\n"))
	return -1;

    sprintf(len,"%d\n",strlen(in_buffer));

    if ( ! writedata(len)
	 || !writedata(in_buffer))
	return -1; 

    if ( (s = readcr()) == NULL )
	return -1;

    if ( sscanf(s,"%d %s", &n, status) != 2 )
	    return -1;

    if ( (readdata(n)) == NULL )
	return -1;

    if ( ! strcmp(status,"ok") )
	return 0;
    
    return 1;
}

static
Str communicate F1(Str, in)
{
    int len = len_str(in) + 1;

    while ( len > in_buffer_size )
	{
	    char * p = malloc(in_buffer_size += MEMDIMSTEP);
	    if ( ! p ) {
		error(mkc_str("can't allocate heap memory"),Empty);
		exit(1);
	    }
	    if ( ! in_buffer )
		free(in_buffer);
	    in_buffer = p;
	}

    cpc_str(in,in_buffer);
    del_str(in);

    TRACE1({ fprintf(stderr,"server cmd:\n%s\n",in_buffer); })

    switch (rpc_Exec())
	{
	case -1:
	    error(mkc_str("connection error"),Empty);
	    exit(1);
	case  1:
	    TRACE1({ fprintf(stderr,"server error:\n%s\n",out_buffer); fflush(stderr); })
	    error(mkc_str("server error: "),mkc_str(out_buffer));
	    exit(1);
	case  0:
	    TRACE1({ fprintf(stderr,"server answer:\n%s\n",out_buffer); fflush(stderr); })
	}
    return
	mkc_str(out_buffer);
}

static
void open_client F0()
{
    if ( (hp = gethostbyname(host) ) == NULL)
	{
	    error(mkc_str("unknown host "),mkc_str(host));
	    exit(1);
	}
  
    memset( (char*) &serv_addr, 0, sizeof(serv_addr));

    serv_addr.sin_family = hp->h_addrtype;

    memcpy(hp->h_addr_list[0], (caddr_t)&serv_addr.sin_addr, hp->h_length);

    if ((portfd=atoi(port)) == 0)
	{
	    if ((sp = getservbyname(port,"tcp")) == NULL)
		{
		    char str[255];
		    sprintf(str,"service %s/tcp unknown\n",port);
		    error(mkc_str(str),Empty);
		    exit(1);
		}
	    else
		portfd = sp->s_port;
	}

    serv_addr.sin_port = htons(portfd);
  
    if ( (sockfd = socket(AF_INET,SOCK_STREAM, 0)) < 0)
	{
	    error(mkc_str("cannot open stream socket"),Empty);
	    exit(1);
	}  
    
    if (connect(sockfd, (struct sockaddr *) &serv_addr, sizeof(serv_addr)) < 0)
	{
	    error(mkc_str("cannot connect to server"),Empty);
	    exit(1);
	}

    TRACE1({ fprintf(stderr,"connected to server\n"); })

    if (rpc_MakeClient() < 0)
	{
	    error(mkc_str("connection rejected"),Empty);
	    exit(1);
	}  

    TRACE1({ fprintf(stderr,"connection to server initialized\n"); })
}

void close_client F0()
{
    close(sockfd);
    exit(0);
}
  

#endif

/*----------------------------------------------------------------------*/

#define ini_cmd(var,cmd) var = mkc_str(cmd)
#define app_arg(var,arg) var = conc_str(conc_str(var,mkc_str(" ")),copy_str(arg))
#define to_bool(val) (hd_str(val) == '1')
#define exec_pred(res,cmd) {Str val; val = communicate(cmd); res = to_bool(val); del_str(val);}

Vdmdbs def_Vdmdbs F1(Dbs,dbs)
{
    if (! run_as_client)
	return
	    def_local_Vdmdbs(dbs);

#if CLIENT
    gptrace = domtrace;

    open_client();
    {
	Str res;
	res = communicate(mkc_str("if {\"[currdb]\"==\"\"} {loaddb}"));
	del_str(res);
    }

    /*
    exists_Adt(mk_empty_vdmdbs(),mkc_str("tupl"));
    exists_Implementation(mk_empty_vdmdbs(),mkc_str("tupl"),mkc_str("linklist"));
    exists_Operation(mk_empty_vdmdbs(),mkc_str("tupl"),mkc_str("linklist"),mkc_str("basic"));
    exists_Component(mk_empty_vdmdbs(),mkc_str("enum"),mkc_str("enum"),mkc_str("copy"),mkc_str("ccod"));
    exists_SignComponent(mk_empty_vdmdbs(),mkc_str("enum"),mkc_str("copy"),mkc_str("vsig"));

    getsx_Component(mk_empty_vdmdbs(),mkc_str("enum"),mkc_str("enum"),mkc_str("copy"),mkc_str("ccod"));
    getsx_SignComponent(mk_empty_vdmdbs(),mkc_str("enum"),mkc_str("copy"),mkc_str("vsig"));

    exit(0);

    */

    del_dbs(dbs);
    return
	mk_empty_vdmdbs();
#endif

}

void undef_Vdmdbs F0()
{
#if CLIENT
    close_client();
#endif
}

Bool exists_Vdmdbs F1(Vdmdbs,vdmdbs)
{
    if (! run_as_client)
	return
	    exists_local_Vdmdbs(vdmdbs);

#if CLIENT
    return
	TRUE;
#endif
}

Bool exists_Adt F2(Vdmdbs,vdmdbs,Adtname,adt_name)
{
    if (! run_as_client)
	return
	    exists_local_Adt(vdmdbs,adt_name);

#if CLIENT
    {
	Str cmd;
	Bool res;

	ini_cmd(cmd,"existadt");
	app_arg(cmd,adt_name);

	exec_pred(res,cmd);

	return res;
    }
#endif
}

Bool exists_Implementation F3(Vdmdbs,vdmdbs,Adtname,adt_name,Implname,impl_name)
{
    if (! run_as_client)
	return
	    exists_local_Implementation(vdmdbs,adt_name,impl_name);

#if CLIENT
    {
	Str cmd;
	Bool res;

	ini_cmd(cmd,"existimpl");
	app_arg(cmd,adt_name);
	app_arg(cmd,impl_name);

	exec_pred(res,cmd);

	return res;
    }
#endif
}

Bool exists_Operation F4(
     Vdmdbs,vdmdbs,
     Adtname,adt_name,
     Implname,impl_name,
     Opname,op_name)
{
    if (! run_as_client)
	return
	    exists_local_Operation(vdmdbs,adt_name,impl_name,op_name);

#if CLIENT
    {
	Str cmd;
	Bool res;

	ini_cmd(cmd,"checkoper");
	app_arg(cmd,adt_name);
	app_arg(cmd,impl_name);
	app_arg(cmd,op_name);

	exec_pred(res,cmd);

	return res;
    }
#endif
}

Bool exists_Component F5(
			 Vdmdbs,vdmdbs,
			 Adtname,adt_name,
			 Implname,impl_name,
			 Opname,op_name,
			 Compname,comp_name)
{
    if (! run_as_client)
	return
	    exists_local_Component(vdmdbs,adt_name,impl_name,op_name,comp_name);
#if CLIENT
    {
	Str cmd;
	Bool res;

	ini_cmd(cmd,"checkcomp");
	app_arg(cmd,adt_name);
	app_arg(cmd,impl_name);
	app_arg(cmd,op_name);
	app_arg(cmd,comp_name);

	exec_pred(res,cmd);

	return res;
    }
#endif
}

Bool exists_SignComponent F4(
			     Vdmdbs,vdmdbs,
			     Adtname,adt_name,
			     Opname,op_name,
			     Compname,signcomp_name)
{
    if (! run_as_client)
	return
	    exists_SignComponent(vdmdbs,adt_name,op_name,signcomp_name);

#if CLIENT
    {
	Str cmd;
	Bool res;

	ini_cmd(cmd,"checksign");
	app_arg(cmd,adt_name);
	app_arg(cmd,op_name);
	app_arg(cmd,signcomp_name);

	exec_pred(res,cmd);

	return res;
    }
#endif
}

Component gets_Component F5(
			    Vdmdbs,vdmdbs,
			    Adtname,adt_name,
			    Implname,impl_name,
			    Opname,op_name,
			    Compname,comp_name)
                             /* Selektiert einen Component aus */
                             /* den uebergebenen Datenbanken */
                             /* Dazu wird der Component in den DBs */
                             /* mit exist_Component gesucht. */
{
    if (! run_as_client)
	return
	    gets_local_Component(vdmdbs,adt_name,impl_name,op_name,comp_name);

#if CLIENT
    {
	Str cmd;

	ini_cmd(cmd,"searchcomp");
	app_arg(cmd,adt_name);
	app_arg(cmd,impl_name);
	app_arg(cmd,op_name);
	app_arg(cmd,comp_name);

	return
	    communicate(cmd);
    }
#endif
}

Component getsx_Component F5(
			     Vdmdbs,vdmdbs,
			     Adtname,adt_name,
			     Implname,impl_name,
			     Opname,op_name,
			     Compname,comp_name)
                             /* Selektiert eine Komponente aus */
                             /* den uebergebenen Datenbanken */
                             /* und loesst die includes auf */
{
    if (! run_as_client)
	return
	    getsx_local_Component(vdmdbs,adt_name,impl_name,op_name,comp_name);

#if CLIENT
    /* includes are solved on server side */
    /* even in gets_Component */

    return
	gets_Component(vdmdbs,adt_name,impl_name,op_name,comp_name);
#endif
}


Component gets_SignComponent F4(
				Vdmdbs,vdmdbs,
				Adtname,adt_name,
				Opname,op_name,
				Compname,signcomp_name)
{
    if (! run_as_client)
	return
	    gets_local_SignComponent(vdmdbs,adt_name,op_name,signcomp_name);

#if CLIENT
    {
	Str cmd;

	ini_cmd(cmd,"searchsign");
	app_arg(cmd,adt_name);
	app_arg(cmd,op_name);
	app_arg(cmd,signcomp_name);

	return
	    communicate(cmd);
    }
#endif
}

Component getsx_SignComponent F4(
				 Vdmdbs,vdmdbs,
				 Adtname,adt_name,
				 Opname,op_name,
				 Compname,signcomp_name)
{
    if (! run_as_client)
	return
	    getsx_local_SignComponent(vdmdbs,adt_name,op_name,signcomp_name);

#if CLIENT
    /* includes are solved on server side */
    /* even in gets_SignComponent */

    return
	gets_SignComponent(vdmdbs,adt_name,op_name,signcomp_name);
#endif
}

/*----------------------------------------------------------------------*/
