/*----------------------------------------------------------------------*/

/* Modul:                  @(#)inf.c */
/* Version:                2.1 */
/* Ersteller:              Gerrit Petrich, WI 8215, WS 93 */
/* Datum der letzten */
/* Aenderung:              93/03/29 */

/*----------------------------------------------------------------------*/

/* Aufgabe:   "inf.c" ist die Programmschnittstelle fuer die */
/*            Funktionen fuer das VDM-Informationssystem. */

/*----------------------------------------------------------------------*/

/* Aenderungen: */
/* wann:                 warum: */
/*----------------------------------------------------------------------*/
/* 16.02.1993                 include der Systemheader entfernt */
/*                       Sie werden jetzt in header.h eingebunden */
/*                          geaendert von: G. Petrich */
/*----------------------------------------------------------------------*/


#include "header.h"          /* allgemeiner Header */
#include "gen.h"             /* aus gen.vdm generierter Header */
#include "proto.h"           /* allgemeiner Header */
#include "dbs.h"             /* Header fuer gets_... und exists_... */
#include "info.h"            /* Header fuer info.c */

void Usage()                 /* Usage gibt die moeglichen Programm- */
                             /* parameter aus, wenn das Programm */
                             /* fehlerhaft aufgerufen wird */
{
  fprintf(stderr,"usage: inf [-Bvdmdbname] [-Dadttype]* -v|-a|-i|-e params\n");
  fprintf(stderr,"options:\n");
  fprintf(stderr," -B : work on this VDM-database\n");
  fprintf(stderr,"      vdmdbname: VDM-databasename\n");
  fprintf(stderr," -D : work on ADTs of this type\n");
  fprintf(stderr,"      adttype = { REAL, AUXILIARY, INTERNAL }\n");
  fprintf(stderr,"      this option can be repeated to work on several adttypes\n");
  fprintf(stderr," -v : prints all ADTs included in database on stdout\n");
  fprintf(stderr,"      params: none\n");
  fprintf(stderr," -a : prints all Implementations of an ADT on stdout\n");
  fprintf(stderr,"      params: <ADTname>\n");
  fprintf(stderr," -i : prints all Operations of an Implementation on stdout\n");
  fprintf(stderr,"      params: <ADTname> <Implementationname>\n");
  fprintf(stderr," -e : prints all Operations of an Implementation and\n");
  fprintf(stderr,"             all common Operations of an ADT on stdout\n");
  fprintf(stderr,"      params: <ADTname> <Implementationname>\n");
}


/*----------------------------------------------------------------------*/

void printout F1(Str,str)    /* gibt einen String auf stdout aus */
{
  puts_str(str,stdout);
  fprintf(stdout,"\n");
}

/*----------------------------------------------------------------------*/

void pre_contents_Vdmdb F5(
        int,i,             /* Zaehler fuer verarbeitete Parameter */
        int,argc,          /* Anzahl der Parameter */
        char **,argv,      /* Tupel der Parameter */
        Dbs,dbs,           /* Tupel der Datenbanknamen */
        Strset,typeset)    /* Menge der auszuwaehlenden ADT-Typen */
                             /* pre_contents_Vdmdb wertet die Programm- */
                             /* parameter aus, und startet die Funktion */
                             /* contents_Vdmdb */
{
Vdmdb vdmdb;
Adtname adt_name,
        found_in_adt;        /* found-ADT */
Implname found_in_impl;      /* found-Impl */
Opname basic_op;
Compname adttype_comp,found_comp;
Adtnameset adt_name_set,
           return_set;       /* Rueckgabemenge der ADT-Namen */
Str adttype;
Nat0 idx;

  HALLO("pre Vdmdbs\n");
  while (argc > i)                     /* zu viele Parameter: */
    {
      fprintf(stderr,"superflous parameter %s ignored\n",argv[i++]);
    }
  vdmdb=hd_vdmdbs(def_Vdmdbs(dbs));    /* Datenbank oeffnen */
  basic_op=dmBASIC;           /* wird die Komponente */
  adttype_comp=dmADTTYPE;     /* adtname/com/basic/adttype */
                                       /* gelesen */
  return_set=mk_empty_strset();
  adt_name_set=contents_Vdmdb(vdmdb);  /* contents_Vdmdb aufrufen */
  while (NOT is_empty_strset(adt_name_set))
    {
      adt_name=sel_strset(adt_name_set);
      if (exist_Component(vdmdb,
                          adt_name,dmCOM,basic_op,adttype_comp,
                          &found_in_adt,&found_in_impl,
                          &found_comp)
          AND (is_eq_str(adt_name,found_in_adt))
          AND (is_eq_str(dmCOM,found_in_impl))
         )
         adttype=copy_str(found_comp);
      else
         adttype=mk_empty_str();
      idx=index_str(adttype,'\n');
      while (idx!=0)                   /* EOLs entfernen */
        {
          adttype=sub_tup_str(adttype,1,idx-1);
          idx=index_str(adttype,'\n');
        }
      if (is_in_strset(adttype,typeset))    /* Ergebnismenge filtern: */
                                            /* nur die ADTs mit richtigem */
                                            /* ADT-Typ werden ausgegeben */
        return_set=union1_strset(return_set,copy_str(adt_name));
      adt_name_set=sub1d_strset(adt_name_set,adt_name);
      del_str(adttype);
    }
  forall_strset(return_set,printout);       /* Ergebnismenge ausgeben */
  del_strset(adt_name_set);                 /* Speicher freigeben */
  del_strset(return_set);
  del_str(basic_op);
  del_str(adttype_comp);
  del_vdmdb(vdmdb);
}

/*----------------------------------------------------------------------*/

void pre_contents_Adt F4(
        int,i,             /* Zaehler fuer verarbeitete Parameter */
        int,argc,          /* Anzahl der Parameter */
        char **,argv,      /* Tupel der Parameter */
        Dbs,dbs)           /* Tupel der Datenbanknamen */
                             /* pre_contents_Adt wertet die Programm- */
                             /* parameter aus, und startet die Funktion */
                             /* contents_Adt */
{
Vdmdb vdmdb;
Adtname adt_name;
Implnameset impl_name_set;

  HALLO("pre Vdmdbs\n");
  if (argc<i+1)                        /* zu wenig Parameter: */
    {
      fprintf(stderr,"missing functionparameter: no adtname given\n");
      Usage();
    }
  else
    {                                  /* Parameter auswerten */
      adt_name=mkc_str(argv[i++]);
      while (argc > i)
        {                              /* zu viele Parameter: */
          fprintf(stderr,"superflous parameter %s ignored\n",argv[i++]);
        }
      vdmdb=hd_vdmdbs(def_Vdmdbs(dbs));/* Datenbank oeffnen */
      impl_name_set=contents_Adt(vdmdb,adt_name);
                                       /* contents_Adt aufrufen */
      impl_name_set=sub1d_strset(impl_name_set,dmCOM); /* Aus der */
                                                          /* Ergebnismenge */
                                                          /* werden "com" */
      impl_name_set=sub1d_strset(impl_name_set,dmSIGN);/* und "sign" */
                                                          /* herausgenommen*/
      forall_strset(impl_name_set,printout);
                                       /* Ergebnismenge ausgeben */
      del_strset(impl_name_set);       /* Speicher freigeben */
      del_vdmdb(vdmdb);
      del_str(adt_name);
    }
}

/*----------------------------------------------------------------------*/

void pre_contents_Impl F4(
        int,i,             /* Zaehler fuer verarbeitete Parameter */
        int,argc,          /* Anzahl der Parameter */
        char **,argv,      /* Tupel der Parameter */
        Dbs,dbs)           /* Tupel der Datenbanknamen */
                             /* pre_contents_Impl wertet die Programm- */
                             /* parameter aus, und startet die Funktion */
                             /* contents_Impl */
{
Adtname adt_name;
Implname impl_name;
Vdmdb vdmdb;
Opnameset op_name_set;

  HALLO("pre Impl\n");
  if (argc<i+2)                        /* zu wenig Parameter: */
    {
      fprintf(stderr,"missing functionparameter: no implname given\n");
      if (argc<i+1)
          fprintf(stderr,"missing functionparameter: no adtname given\n");
      Usage();
    }
  else
    {                                  /* Parameter auswerten */
      adt_name=mkc_str(argv[i++]);
      impl_name=mkc_str(argv[i++]);
      while (argc > i)                 /* zu viele Parameter: */
        {
          fprintf(stderr,"superflous parameter %s ignored\n",argv[i++]);
        }
      vdmdb=hd_vdmdbs(def_Vdmdbs(dbs));/* Datenbank oeffnen */
      op_name_set=contents_Impl(vdmdb,adt_name,impl_name);
                                       /* contents_Impl aufrufen */
      forall_strset(op_name_set,printout);  /* Ergebnismenge ausgeben */
      del_strset(op_name_set);              /* Speicher freigeben */
      del_vdmdb(vdmdb);
      del_str(adt_name);
      del_str(impl_name);
    }
}

/*----------------------------------------------------------------------*/

void pre_ext_contents_Impl F4(
        int,i,             /* Zaehler fuer verarbeitete Parameter */
        int,argc,          /* Anzahl der Parameter */
        char **,argv,      /* Tupel der Parameter */
        Dbs,dbs)           /* Tupel der Datenbanknamen */
                             /* pre_ext_contents_Impl wertet die Programm-*/
                             /* parameter aus, und startet die Funktion */
                             /* ext_contents_Impl */
{
Adtname adt_name;
Implname impl_name;
Vdmdb vdmdb;
Opnameset op_name_set;

  HALLO("pre Impl\n");
  if (argc<i+2)                        /* zu wenig Parameter: */
    {
      fprintf(stderr,"missing functionparameter: no implname given\n");
      if (argc<i+1)
          fprintf(stderr,"missing functionparameter: no adtname given\n");
      Usage();
    }
  else
    {                                  /* Parameter auswerten */
      adt_name=mkc_str(argv[i++]);
      impl_name=mkc_str(argv[i++]);
      while (argc > i)                 /* zu viele Parameter: */
        {
          fprintf(stderr,"superflous parameter %s ignored\n",argv[i++]);
        }
      vdmdb=hd_vdmdbs(def_Vdmdbs(dbs));/* Datenbank oeffnen */
      op_name_set=ext_contents_Impl(vdmdb,adt_name,impl_name);
                                       /* ext_contents_Impl aufrufen */
      forall_strset(op_name_set,printout);  /* Ergebnismenge ausgeben */
      del_strset(op_name_set);         /* Speicher freigeben */
      del_vdmdb(vdmdb);
      del_str(adt_name);
      del_str(impl_name);
    }
}

/*----------------------------------------------------------------------*/

int main F2(
        int,argc,
        char **,argv)
                             /* Das Hauptprogramm liest die Programm- */
                             /* parameter ein und startet die Funktionen */
{
char *p;
int i=1;
Vdmdbname vdmdb_name;
Dbs dbs;
Strset adttypeset,typeset;
Str adttype,tracestring;

  init_vdmdm();

  if (argc==1)                         /* keine Programmparameter */
    Usage();                           /* => usage ausgeben */
  else
    {
                                      /* Menge der verfuegbaren */
                                      /* ADT-Typen zusammenstellen */
      typeset=mk_empty_strset();
      adttypeset=mk1_strset(copy_str(dmREAL));
      adttypeset=union1_strset(adttypeset,copy_str(dmAUXILIARY));
      adttypeset=union1_strset(adttypeset,copy_str(dmINTERNAL));
      dbs=mk_empty_dbs();
      while (argc>i)                   /* Funktionen starten */
        {
          p=argv[i++];
          if (*p++=='-')
            switch(*p)
              {
                case 'T' : /* Trace-variable setzen */
                           tracestring=mkc_str(argv[i-1]);
                           gptrace=len_str(tracestring)-1;
                           del_str(tracestring);
                           break;
                case 'B' : /* Datenbanknamen ermitteln */
                           vdmdb_name=mkc_str(argv[i-1]);
                           if (len_dbs(dbs)==0)
                             dbs=appr_dbs(dbs,sub_tup_str(vdmdb_name,
                                          3,len_str(vdmdb_name)));
                           else
                             {
                               fprintf(stderr,"you can only work on one database\n");
                               fprintf(stderr,"parameter ignored :");
                               OUT_str(vdmdb_name);
                               del_str(vdmdb_name);
                             }
                           break;
                case 'D' :             /* ADT-Typ einlesen */
                           adttype=mkc_str(argv[i-1]);
                           adttype=sub_tup_str(adttype,3,len_str(adttype));
                           if (is_in_strset(adttype,adttypeset))
                             typeset=union1_strset(typeset,adttype);
                           else
                             {
                               fprintf(stderr,"adttype unknown: ");
                               OUT_str(adttype);
                               fprintf(stderr,"parameter ignored \n");
                               del_str(adttype);
                             }
                           break;
                case 'v' : pre_contents_Vdmdb(i,argc,argv,dbs,typeset);i=argc;break;
                case 'a' : pre_contents_Adt(i,argc,argv,dbs);i=argc;break;
                case 'i' : pre_contents_Impl(i,argc,argv,dbs);i=argc;break;
                case 'e' : pre_ext_contents_Impl(i,argc,argv,dbs);i=argc;break;
                default  : fprintf(stderr,"wrong function parameter\n");
                           Usage();
              }
            else
              {
                fprintf(stderr,"function parameter has to start with -\n");
                Usage();
              }
        }
      del_strset(adttypeset);
      del_strset(typeset);
      TRACE1( { profile_dbs(); profile_strset(); profile_str(); } )

    }
  finish_vdmdm();

  return 0;
}
