/*----------------------------------------------------------------------*/

/* Modul:                  @(#)invx.c */
/* Version:                2.1 */
/* Ersteller:              Gerrit Petrich, WI 8215, WS 93 */
/* Datum der letzten */
/* Aenderung:              93/03/29 */

/*----------------------------------------------------------------------*/

/* Aufgabe:   "invx.c" enthaelt die Konsistenzpruefungsfunktionen */
/*            fuer die VDM-Klassenbibliothek. */

/*----------------------------------------------------------------------*/

/* Aenderungen: */
/* wann:                 warum: */
/*----------------------------------------------------------------------*/
/* 16.02.1993                 include der Systemheader entfernt */
/*                       Sie werden jetzt in header.h eingebunden */
/*                          geaendert von: G. Petrich */
/*----------------------------------------------------------------------*/

#include "header.h"              /* Standard-header-file */
#include "invx.h"                /* eigenes header-file */
#include "invar.h"               /* Header fuer exist_wf_... */
#include "dbs.h"                 /* Header fuer exis_... */

Bool consistent=TRUE;            /* bleibt TRUE, wenn vdmdb konsistent */

/*----------------------------------------------------------------------*/

Adtnameset special_names()
                             /* special_names erzeugt die Menge von */
                             /* ADT-Namen, die in einer DB enthalten */
                             /* sein muessen: com, aux, stdlib, vdmlib */

{
  Adtnameset special_set;

  HALLO("special_names\n");
  LET special_set=mk1_strset(copy_str(dmCOM)) IN
  LET special_set=union1_strset(special_set,copy_str(dmAUX)) IN
  LET special_set=union1_strset(special_set,copy_str(dmSTDLIB)) IN
  LET special_set=union1_strset(special_set,copy_str(dmVDMLIB)) IN
  return special_set;
}

/*----------------------------------------------------------------------*/

Bool exist_special_names F1(Vdmdb,vdmdb)
                             /* exist_special_names ueberprueft die */
                             /* Existenz der ADTs aus 'special_set' */
{
Bool bool;
Adtname adt_name;
Adtnameset adt_name_set;

  HALLO("exist_special_names\n");
  adt_name_set=special_names();
  while ((NOT is_empty_strset(adt_name_set)))
    {
      adt_name=sel_strset(adt_name_set);     /* ADT auswaehlen */
      bool=is_in_vdmdb(adt_name,vdmdb);      /* ADT in DB ? */
      if (NOT bool)                          /* wenn nicht: Meldung */
        {
          fprintf(stderr,"database inconsistent:\n");
          fprintf(stderr,"the following adt does not exist:\n");
          OUT_str(adt_name);
          consistent=FALSE;
        }
      adt_name_set=sub1d_strset(adt_name_set,adt_name);
    }
  del_strset(adt_name_set);
  return consistent;
}

/*----------------------------------------------------------------------*/

Str exist_includes F7(
        Str,s,               /* Liste an die der VDM-String 'c' anzuhaengen ist */
        Str,c,               /* anzuhaengender String 'c' */
        Vdmdb,vdmdb,         /* VDM-Datenbank */
        Str,constr,          /* ADT-Name (z.B. tree) des ADT */
        Str,implid,          /* Implementationsname (z.B. linklist) des ADT */
        Str,operationid,     /* Operationsname (z.B. union), aus 'c' */
        Nat0, counter)       /* Zaehler fuer include-Aufrufe */

                             /* exist_includes ermittelt die in einer */
                             /* Komponente enthaltenen includes. */
                             /* Diese werden mit der Funktion */
                             /* exist_includes aufgeloest. */
                             /* in 'counter' werden die Anzahl der */
                             /* include-Aufrufe gezaehlt. */
{
 Str exist_include();        /* Funktionsdeklaration fuer exist_include */
                             /* Die Funktionen exist_include und */
                             /* exist_includes rufen sich wechselseitig */
                             /* rekursiv auf */
  char *mcs=malloc(len_str(c)+1);       /* Speicher fuer C-String holen */
  HALLO("exist_includes\n");
  TRACE1( { OUT3_str(constr,implid,operationid);
            fprintf(stderr,"%i \n",counter); } )

 if (mcs!=0)
 {
  Nat0 lno=0;     /* Anzahl der verarbeiteten Zeilen */
  Bool il;        /* '#include <...>' -Zeile gefunden */
  char *cs=mcs;   /* Zeiger zum Durchlaufen des Strings */
  char *line;     /* Zeiger auf Zeilenanfang */

  cpc_str(c,cs);                       /* VDM-String zu C-String machen */
  del_str(c);                          /* Speicher freigeben */
  while (*cs)                          /* Solange C-Strings nicht zuende */
  {
   lno++;                              /* Anzahl Zeilennr erhoehen */
   line=cs;                            /* Zeiger auf Zeilenanfang */
   while ((*cs) && (*cs!='\n')) cs++;  /* Zeilenende suchen */
   if (*cs) *cs++='\0';                /* '\n' durch '\0' ersetzen */
                                       /* (String-Ende Markierung) fuer */
                                       /* spaetere Umwandlung in einen */
                                       /* VDM-String */
   il=FALSE;
   if (*line=='#')                     /* '#' am Zeilenanfang */
   {
    char *incl="#include", *ic, *l, *p[4];
    int i;
    /* Durchsuchen der aktuellen Zeile nach '#include' -Anweisungen */
    l=line;                            /* Zeiger auf Zeilenanfang */
    ic=incl;                           /* Zeiger auf '#include' */
    while ((*ic) && (*l++==*ic++));    /* '#include' -Anweisung suchen */
    if (!*ic)                          /* '#include' -Anweisung gefunden */
    {
     for (i=4;i;p[--i]=0);          /* Leerstellen ueberlesen */
     while (*l==' ') l++;
     /* nur mit '<' beginnende Include-Komponenten werden aufgeloest. */
     if (*l=='<')
     {                                 /* aufzuloesende Zeile gefunden */
      il=TRUE;
      /* Trennen der einzelnen durch '/' getrennten Komponenten */
      do
      {
       /* '<' beim ersten und '/' bei jedem weiteren Durchlauf 'loeschen'. */
       /* C-String Zeilenende-Markierung setzen, die Variablen 'char *p[4] */
       /* werden im Folgenden auf den Anfang der entsprechenden Komponenten */
       /* '<comp1/comp2/comp3/comp4>' gesetzt. Die Zeile wird in */
       /* '\0comp1\0comp2\0comp3\0comp4>' umgesetzt. */
       *l++='\0';                      /* ueberlesen von Leerstellen */
       while (*l==' ') l++;            /* Zeiger auf Anfang von 'comp?' */
       line=l;                         /* Ende der Komponente suchen */
       while ((*l) && (*l!='\n') && (*l!='/') && (*l!='>')) l++;
                                       /* Zeiger auf Anfang der Komponente */
       p[i++]=line;
      } while ((*l=='/') && (i<4));
      /* Korrekte Include-Zeile enthM-^Dlt 4 Komponenten und '>' zum AbschluM-a */
      if ((i==4) && (*l=='>'))
      {
       /* Umwandlung der ersten drei Komponenten der Include-Zeile in */
       /* VDM-Strings */
       Str iconstr=mkc_str(p[0]);
       Str iimplid=mkc_str(p[1]);
       Str ioperationid=mkc_str(p[2]);
       Str icomp;
       /* Setzen der Zeilenende-Markierung fM-^Ar 'comp4' die Include- */
       /* Komponente hat jetzt die Form '\0comp1\0comp2\0comp3\0comp4\0' */
       *l='\0';
       /* Umwandlung der vierten Komponente in einen VDM-String */
       icomp=mkc_str(p[3]);
       /* Aufruf von exist_include zum Aufloesen der gefundenen */
       /* Include-Anweisung */
       s=exist_include(s,vdmdb,iconstr,iimplid,ioperationid,icomp,counter);
       del_str(iconstr);               /* Speicher freigeben */
       del_str(iimplid);
       del_str(ioperationid);
       del_str(icomp);
      }
     }
    }
   }
   /* Keine Include-Zeile --> anhaengen der Zeile an die Liste 's' */
   if (!il) {s=conc_str(s,mkc_str(line));s=appr_str(s,'\n');}
  }
  free(mcs);                           /* Speicher freigeben */
 }
 return s;
}

/*----------------------------------------------------------------------*/

void Error F3(Nat0,errorcode,Str *,c,Bool *,flag)
                             /* Error ist die Fehlerroutine fuer */
                             /* die Funktion exist_include. */
                             /* Error gibt eine Fehlermeldung aus, */
                             /* c wird initialisiert und flag auf FALSE */
                             /* gesetzt. */
{
  switch (errorcode)
    {
      case  1 : fprintf(stderr,"Component does not exist\n");
                break;
      default : fprintf(stderr,"Component not wellformed\n");
                fprintf(stderr,"An include of the %i. Component does not exist\n",errorcode-1);
    }
  *c=mk_empty_str();
  *flag=FALSE;
}

/*----------------------------------------------------------------------*/

Str exist_include F7(
        Str, s,
        Vdmdb, vdmdb,        /* VDM-Datenbank */
        Str, adt_name,       /* ADT-Name (z.B. tree) des ADT */
        Str, impl_name,      /* Implementationsname (z.B. linklist) des ADT */
        Str, op_name,        /* Operationsname (z.B. union), aus 'c' */
        Str, comp_name,      /* Komponentenname (z.B. 'vdm') */
        Nat0, counter)       /* Zaehler fuer include-Aufrufe */

                             /* exist_include ueberprueft die Existenz */
                             /* einer include-Anweisung. Wenn sie */
                             /* existiert, wird sie mit exist_includes */
                             /* weiterbearbeitet, wenn nicht eine */
                             /* Fehlermeldung ausgegeben */
{
 Adt adt;
 Impl impl;
 Operation op;
 Component comp;
 Bool eflag=FALSE;

  HALLO("exist_include\n");
  TRACE1( { OUT4_str(adt_name,impl_name,op_name,comp_name);
          fprintf(stderr,"%i \n",counter); } )

  counter++;                           /* Aufrufzaehler erhoehen */
  if (is_in_vdmdb(adt_name,vdmdb))
    {
      adt=sel_vdmdb(vdmdb,adt_name);    /* wenn ADT, Impl, Operation */
      if (is_in_adt(impl_name,adt))     /* und Komponente vorhanden, */
        {                               /* dann selektieren */
          impl=sel_adt(adt,impl_name);  /* sonst Error aufrufen und */
          if (is_in_impl(op_name,impl)) /* leeren String zurueckgeben */
            {
              op=sel_impl(impl,op_name);
              if (is_in_operation(comp_name,op))
                comp=get_Component(vdmdb,adt_name,impl_name,op_name,comp_name);
              else Error(counter,&comp,&eflag);
            }
          else Error(counter,&comp,&eflag);
        }
      else Error(counter,&comp,&eflag);
    }
  else Error(counter,&comp,&eflag);
  if (eflag)                           /* Falls include nicht gefunden */
    {                                  /* Meldung ausgeben */
      if (counter>1)
        {
          fprintf(stderr,"database inconsistent:\n");
          fprintf(stderr,"component refers to non existing include-statement\n");
          fprintf(stderr,"the following include does not exist:\n");
        }
      OUT_str(comp_name);
      fprintf(stderr,"from Operation:\n");
      OUT_str(op_name);
      fprintf(stderr,"from Implementation:\n");
      OUT_str(impl_name);
      fprintf(stderr,"from ADT:\n");
      OUT_str(adt_name);
      consistent=FALSE;
      eflag=FALSE;
      return s;
   }
  return exist_includes(s,comp,vdmdb,adt_name,impl_name,op_name,counter);
}

/*----------------------------------------------------------------------*/

Bool is_wf_Signature F3(Vdmdb,vdmdb,Adtname,adt_name,Opname,op_name)
                             /* is_wf_Signature ueberprueft die Konsistenz */
                             /* einer Signatur. */
                             /* Eine Signatur ist konsistent, wenn alle */
                             /* Implementationen des angegebenen ADTs */
                             /* eine Operation gleichen Namens haben. */
{
Bool bool;
Adtname found_in_adt;
Implname impl_name,found_in_impl;
Operation found_op;
Implnameset domadt,special_set,impl_name_set;

  HALLO("is_wf_Sign\n");
  TRACE1( { OUT2_str(adt_name,op_name); } )

  special_set=mk1_strset(copy_str(dmCOM));
  special_set=union1_strset(special_set,copy_str(dmSIGN));
  special_set=union1_strset(special_set,copy_str(dmAUX));
  if (exist_Adt(vdmdb,adt_name))       /* wenn ADT und Signatur vorhanden */
    if (exist_Signature(vdmdb,adt_name,op_name,&found_in_adt,&found_op)
        AND (is_eq_str(adt_name,found_in_adt)))
      {
        domadt=dom_adt(sel_vdmdb(vdmdb,adt_name));
        impl_name_set=diffd_strset(domadt,special_set);
        while ((NOT is_empty_strset(impl_name_set))) /* Fuer alle Impls: */
          {
            impl_name=sel_strset(impl_name_set);     /* Operation pruefen */
            bool=exist_wf_Operation(vdmdb,adt_name,impl_name,op_name,
                                    &found_in_adt,&found_in_impl,
                                    &found_op);
            if (NOT bool)
              {
                fprintf(stderr,"database inconsistent:\n");
                fprintf(stderr,"a signature for a non existing operation found:\n");
                fprintf(stderr,"the following operation does not exist:\n");
                OUT_str(adt_name);
                OUT_str(impl_name);
                OUT_str(op_name);
                fprintf(stderr,"the following signature was found:\n");
                OUT_str(adt_name);
                OUT_str(dmSIGN);
                OUT_str(op_name);
                consistent=FALSE;
              }
            impl_name_set=sub1d_strset(impl_name_set,impl_name);
          }
        del_strset(impl_name_set);
      }
    else
      {
        fprintf(stderr,"Signature does not exist:\n");
        OUT_str(op_name);
        fprintf(stderr,"from Implementation:\n");
        fprintf(stderr,"sign\n");
        fprintf(stderr,"from ADT:\n");
        OUT_str(adt_name);
        consistent=FALSE;
      }
  else
    {
      fprintf(stderr,"ADT does not exist:\n");
      OUT_str(adt_name);
      consistent=FALSE;
    }
  del_strset(special_set);
  return consistent;
}

/*----------------------------------------------------------------------*/

Bool is_wf_Component F5(Vdmdb,vdmdb,Adtname,adt_name,Implname,impl_name,
                        Opname,op_name,Compname,comp_name)
                             /* is_wf_Component ueberprueft die */
                             /* Konsistenz einer Komponente. */
                             /* Eine Komponente ist konsistent, wenn ihre*/
                             /* includes aufgeloest werden koennen. */
                             /* Dies wird mit exist_include geprueft. */
{
Adtname found_in_adt;
Implname found_in_impl;
Str s;
Nat0 init=0;

  HALLO("is_wf_Comp\n");
  TRACE1( { OUT4_str(adt_name,impl_name,op_name,comp_name); } )

  if (exist_Adt(vdmdb,adt_name))                         /* Wenn ADT und */
    if (exist_Implementation(vdmdb,adt_name,impl_name))  /* Impl und */
      {
        if ((exist_Operation(vdmdb,adt_name,impl_name,op_name,
                             &found_in_adt,&found_in_impl))
            AND (is_eq_str(adt_name,found_in_adt))
            AND (is_eq_str(impl_name,found_in_impl)))
          {
            s=mk_empty_str();          /* includes ueberpruefen */
            s=exist_include(s,vdmdb,adt_name,impl_name,
                            op_name,comp_name,init);
            del_str(s);
          }
        else
          {
            fprintf(stderr,"Operation does not exist:\n");
            OUT_str(op_name);
            fprintf(stderr,"from Implementation:\n");
            OUT_str(impl_name);
            fprintf(stderr,"from ADT:\n");
            OUT_str(adt_name);
            consistent=FALSE;
          }
        }
    else
      {
        fprintf(stderr,"Implementation does not exist:\n");
        OUT_str(impl_name);
        fprintf(stderr,"from ADT:\n");
        OUT_str(adt_name);
        consistent=FALSE;
      }
  else
    {
      fprintf(stderr,"ADT does not exist:\n");
      OUT_str(adt_name);
      consistent=FALSE;
    }
  return consistent;
}

/*----------------------------------------------------------------------*/

Bool is_wf_Operation F4(Vdmdb,vdmdb,Adtname,adt_name,
                        Implname,impl_name,Opname,op_name)
                             /* is_wf_Operation ueberprueft die Konsistenz */
                             /* einer Operation. */
                             /* Eine Operation ist konsistent, wenn es fuer*/
                             /* sie eine Signatur gibt und alle Komponenten*/
                             /* konsistent sind. */
{
Bool bool;
Adtname found_in_adt;
Implname found_in_impl;
Operation found_op;
Compname comp_name;
Compnameset comp_name_set;

  HALLO("is_wf_Op\n");
  TRACE1( { OUT3_str(adt_name,impl_name,op_name); } )

  if (exist_Adt(vdmdb,adt_name))                          /* Wenn ADT und */
    if (exist_Implementation(vdmdb,adt_name,impl_name))   /* Impl und */
      {
        if ((exist_Operation(vdmdb,adt_name,impl_name,op_name,
                                &found_in_adt,&found_in_impl))
            AND (is_eq_str(adt_name,found_in_adt))
            AND (is_eq_str(impl_name,found_in_impl)))
          {
            if (is_eq_str(op_name,dmBASIC)            /* basic und post */
                OR is_eq_str(op_name,dmPOST))         /* braucht es */
              bool=TRUE;                                 /* keine Signatur */
            else                                         /* geben */
              bool=exist_wf_Signature(vdmdb,adt_name,op_name,
                                      &found_in_adt,&found_op);
            if (NOT bool)
              {
                fprintf(stderr,"database inconsistent:\n");
                fprintf(stderr,"the following operation has no signature:\n");
                OUT_str(adt_name);
                OUT_str(impl_name);
                OUT_str(op_name);
                fprintf(stderr,"the following signature is missing:\n");
                OUT_str(adt_name);
                OUT_str(dmSIGN);
                OUT_str(op_name);
                consistent=FALSE;
              }
            comp_name_set=dom_operation(sel_impl(sel_adt(sel_vdmdb(vdmdb,adt_name),
                                                 impl_name),op_name));
            while ((NOT is_empty_strset(comp_name_set)))
              {
                comp_name=sel_strset(comp_name_set);
                bool=is_wf_Component(vdmdb,adt_name,impl_name,op_name,comp_name);
                comp_name_set=sub1d_strset(comp_name_set,comp_name);
              }
            del_strset(comp_name_set);
          }
        else
          {
            fprintf(stderr,"Operation does not exist:\n");
            OUT_str(op_name);
            fprintf(stderr,"from Implementation:\n");
            OUT_str(impl_name);
            fprintf(stderr,"from ADT:\n");
            OUT_str(adt_name);
            consistent=FALSE;
          }
        }
    else
      {
        fprintf(stderr,"Implementation does not exist:\n");
        OUT_str(impl_name);
        fprintf(stderr,"from ADT:\n");
        OUT_str(adt_name);
        consistent=FALSE;
      }
  else
    {
      fprintf(stderr,"ADT does not exist:\n");
      OUT_str(adt_name);
      consistent=FALSE;
    }
  return consistent;
}

/*----------------------------------------------------------------------*/

Bool is_wf_Implementation F3(Vdmdb,vdmdb,Adtname,adt_name,Implname,impl_name)
                             /* is_wf_Implementation ueberprueft die */
                             /* Konsistenz einer Implementation. */
                             /* Eine Implementation ist konsistent, wenn */
                             /* alle Operationen konsistent sind. */
                             /* Die Operationen 'basic' und 'post' */
                             /* brauchen keine Signatur zu haben. */
{
Bool bool;
Opname op_name;
Compname comp_name;
Opnameset op_name_set;
Compnameset comp_name_set;


  HALLO("is_wf_Impl\n");
  TRACE1( { OUT2_str(adt_name,impl_name); } )

  if (exist_Adt(vdmdb,adt_name))       /* Wenn ADT und Impl vorhanden */
    if (exist_Implementation(vdmdb,adt_name,impl_name))
      {
        if (is_eq_str(impl_name,dmSIGN))      /* Wenn Impl = 'sign', */
          {                                      /* Signatures pruefen */
            op_name_set=dom_impl(sel_adt(sel_vdmdb(vdmdb,adt_name),
                                         impl_name));
            while ((NOT is_empty_strset(op_name_set)))
              {
                op_name=sel_strset(op_name_set);
                bool=is_wf_Signature(vdmdb,adt_name,op_name);
                op_name_set=sub1d_strset(op_name_set,op_name);
              }
            del_strset(op_name_set);
          }
        else                                     /* Wenn Impl = 'aux', */
          if (is_eq_str(impl_name,dmAUX))     /* nix tun */
            bool=TRUE;
          else
            {
              op_name_set=dom_impl(sel_adt(sel_vdmdb(vdmdb,adt_name),impl_name));
              while ((NOT is_empty_strset(op_name_set)))
                {
                  op_name=sel_strset(op_name_set);
                  if (is_eq_str(op_name,dmBASIC) OR is_eq_str(op_name,dmPOST))
                    {                  /* 'basic' oder 'post' => Komponenten
                                                                 pruefen */
                      comp_name_set=dom_operation(sel_impl(sel_adt(sel_vdmdb(vdmdb,adt_name),
                                                      impl_name),op_name));
                      while ((NOT is_empty_strset(comp_name_set)))
                        {
                          comp_name=sel_strset(comp_name_set);
                          bool=is_wf_Component(vdmdb,adt_name,impl_name,op_name,comp_name);
                          comp_name_set=sub1d_strset(comp_name_set,comp_name);
                        }
                      del_strset(comp_name_set);
                    }
                  else                 /* sonst Ops pruefen */
                    bool=is_wf_Operation(vdmdb,adt_name,impl_name,op_name);
                  op_name_set=sub1d_strset(op_name_set,op_name);
                }
              del_strset(op_name_set);
            }
      }
    else
      {
        fprintf(stderr,"Implementation does not exist:\n");
        OUT_str(impl_name);
        fprintf(stderr,"from ADT:\n");
        OUT_str(adt_name);
        consistent=FALSE;
      }
  else
    {
      fprintf(stderr,"ADT does not exist:\n");
      OUT_str(adt_name);
      consistent=FALSE;
    }
  return consistent;
}

/*----------------------------------------------------------------------*/

Bool is_wf_Combase F1(Vdmdb,vdmdb)
                             /* is_wf_Combase ueberprueft die Konsistenz */
                             /* des 'com'-Datenbankteils. */
{
Bool bool;
Adtname adt_name;
Opname op_name;
Adtnameset domvdmdb,adt_name_set,specialset;
Opnameset op_name_set;

  HALLO("is_wf_Combase\n");
  domvdmdb=dom_vdmdb(vdmdb);
  specialset=special_names();
  adt_name_set=diffd_strset(domvdmdb,specialset);
  del_strset(specialset);                      /* 'com' 'sign' pruefen */
  while ((NOT is_empty_strset(adt_name_set)))
    {
      adt_name=sel_strset(adt_name_set);
      op_name_set=dom_impl(sel_adt(sel_vdmdb(vdmdb,adt_name),dmSIGN));
      while ((NOT is_empty_strset(op_name_set)))
        {
          op_name=sel_strset(op_name_set);
          bool=is_wf_Signature(vdmdb,adt_name,op_name);
          op_name_set=sub1d_strset(op_name_set,op_name);
        }
      del_strset(op_name_set);
      adt_name_set=sub1d_strset(adt_name_set,adt_name);
    }
  del_strset(adt_name_set);                    /* 'com' 'com' pruefen */
  op_name_set=dom_impl(sel_adt(sel_vdmdb(vdmdb,dmCOM),dmCOM));
  while ((NOT is_empty_strset(op_name_set)))
    {
      op_name=sel_strset(op_name_set);
      bool=is_wf_Operation(vdmdb,dmCOM,dmCOM,op_name);
      op_name_set=sub1d_strset(op_name_set,op_name);
    }
  del_strset(op_name_set);
  return consistent;
}

/*----------------------------------------------------------------------*/

Bool is_wf_Adt F2(Vdmdb,vdmdb,Adtname,adt_name)
                             /* is_wf_Adt ueberprueft die Konsistenz */
                             /* eines ADTs */
                             /* Ein Adt ist konsistent, wenn seine Impls */
                             /* und Signaturen konsistent sind. */
{
Bool bool;
Adtname stdlib_adt,vdmlib_adt;
Implname impl_name;
Opname op_name;
Implnameset impl_name_set;
Opnameset op_name_set;

  HALLO("is_wf_Adt\n");
  TRACE1( { OUT_str(adt_name); } )

  if (exist_Adt(vdmdb,adt_name))
    {
      stdlib_adt=dmSTDLIB;
      vdmlib_adt=dmVDMLIB;
      if (is_eq_str(adt_name,dmAUX) OR is_eq_str(adt_name,stdlib_adt)
            OR is_eq_str(adt_name,vdmlib_adt))
        bool=TRUE;                     /* aux, stdlib oder vdmlib => TRUE */
      else
        if (is_eq_str(adt_name,dmCOM))
          bool=is_wf_Combase(vdmdb);   /* com => Combase pruefen */
        else
          {
            impl_name_set=dom_adt(sel_vdmdb(vdmdb,adt_name));
            while (NOT is_empty_strset(impl_name_set))
              {                        /* Signatures   pruefen */
                impl_name=sel_strset(impl_name_set);
                if (is_eq_str(impl_name,dmSIGN))
                  {
                    op_name_set=dom_impl(sel_adt(sel_vdmdb(vdmdb,adt_name),
                                         impl_name));
                    while ((NOT is_empty_strset(op_name_set)))
                      {
                        op_name=sel_strset(op_name_set);
                        bool=is_wf_Signature(vdmdb,adt_name,op_name);
                        op_name_set=sub1d_strset(op_name_set,op_name);
                      }
                    del_strset(op_name_set);
                  }
                else
                  if (is_eq_str(impl_name,dmAUX))
                    bool=TRUE;
                  else                  /* sonst: Impls pruefen */
                    bool=is_wf_Implementation(vdmdb,adt_name,impl_name);
                impl_name_set=sub1d_strset(impl_name_set,impl_name);
              }
            del_strset(impl_name_set);
          }
      del_str(stdlib_adt);
      del_str(vdmlib_adt);
    }
  else
    {
      fprintf(stderr,"ADT does not exist:\n");
      OUT_str(adt_name);
      consistent=FALSE;
    }
  return consistent;
}

/*----------------------------------------------------------------------*/

Bool is_wf_Vdmdb F1(Vdmdb,vdmdb)
                             /* is_wf_Vdmdb ueberprueft die Konsistenz */
                             /* einer Datenbank. */
                             /* Eine Datenbank ist konsistent, wenn */
                             /* der 'com'-Teil konsistent ist, */
                             /* und alle ADTs konsistent ist. */
{
Bool bool;
Adtname adt_name,stdlib_adt,vdmlib_adt;
Adtnameset adt_name_set;

  HALLO("is_wf_Vdmdb\n");
  stdlib_adt=dmSTDLIB;
  vdmlib_adt=dmVDMLIB;
  adt_name_set=dom_vdmdb(vdmdb);
  while ((NOT is_empty_strset(adt_name_set)))
    {
      adt_name=sel_strset(adt_name_set);
      bool=
        (IF is_eq_str(adt_name,dmAUX) OR is_eq_str(adt_name,stdlib_adt)
            OR is_eq_str(adt_name,vdmlib_adt)
         THEN TRUE
         ELSE
           (IF is_eq_str(adt_name,dmCOM)
            THEN is_wf_Combase(vdmdb)            /* Combase pruefen */
            ELSE (is_wf_Adt(vdmdb,adt_name))     /* ADTs pruefen */
           )
        );
      adt_name_set=sub1d_strset(adt_name_set,adt_name);
    }
  del_strset(adt_name_set);
  del_str(stdlib_adt);
  del_str(vdmlib_adt);
  return consistent;
}

/*----------------------------------------------------------------------*/
/*----------------------------------------------------------------------*/


