/*----------------------------------------------------------------------*/

/* Modul:                  @(#)kons.c */
/* Version:                2.1 */
/* Ersteller:              Gerrit Petrich, WI 8215, WS 93 */
/* Datum der letzten */
/* Aenderung:              93/03/29 */

/*----------------------------------------------------------------------*/

/* Aufgabe:   "kons.c" ist die Programmschnittstelle fuer die */
/*            Konsistenzpruefungsfunktionen aus "invx.c". */

/*----------------------------------------------------------------------*/

/* Aenderungen: */
/* wann:                 warum: */
/*----------------------------------------------------------------------*/
/* 16.02.1993                 include der Systemheader entfernt */
/*                       Sie werden jetzt in header.h eingebunden */
/*                          geaendert von: G. Petrich */
/*----------------------------------------------------------------------*/

#include "header.h"          /* allgemeiner Header */
#include "gen.h"             /* aus gen.vdm generierte Funktionen */
#include "proto.h"           /* allgemeiner Header */
#include "invx.h"            /* Konsistenzoperationen */
#include "dbs.h"             /* Datenbank oeffnen, ... */

void Usage()                 /* Usage gibt die moeglichen Programm- */
                             /* parameter aus, wenn das Programm */
                             /* fehlerhaft aufgerufen wird */
{
  fprintf(stderr,"usage: kons [-Bvdmdbname] -v|-a|-i|-o|-s|-c params\n");
  fprintf(stderr,"options:\n");
  fprintf(stderr," -B : work on this VDM-database\n");
  fprintf(stderr,"      vdmdbname: VDM-databasename\n");
  fprintf(stderr," -v : check if database is wellformed\n");
  fprintf(stderr,"      params: none\n");
  fprintf(stderr," -C : check if Combase is wellformed\n");
  fprintf(stderr,"      params: none\n");
  fprintf(stderr," -a : check if Adt is wellformed\n");
  fprintf(stderr,"      params: <ADTname>\n");
  fprintf(stderr," -i : check if Implementation is wellformed\n");
  fprintf(stderr,"      params: <ADTname> <Implname>\n");
  fprintf(stderr," -o : check if Operation is wellformed\n");
  fprintf(stderr,"      params: <ADTname> <Implname> <Operationname>\n");
  fprintf(stderr," -s : check if Signature is wellformed\n");
  fprintf(stderr,"      params: <ADTname> <Operationname>\n");
  fprintf(stderr," -c : check if Component is wellformed\n");
  fprintf(stderr,"      params: <ADTname> <Implname> <Operationname> <Componentname>\n");
}

/*----------------------------------------------------------------------*/

void pre_is_wf_Vdmdb F4(
        int,i,             /* Zaehler fuer verarbeitete Parameter */
        int,argc,          /* Anzahl der Parameter */
        char **,argv,      /* Tupel der Parameter */
        Dbs,dbs)           /* Tupel der Datenbanknamen */

                             /* pre_is_wf_Vdmdb wertet die Programm- */
                             /* parameter aus, und startet die Funktion */
                             /* is_wf_Vdmdb */
{
Vdmdb vdmdb;

  HALLO("pre is_wf_Vdmdbs\n");
  while (argc > i)                          /* zu viele Parameter: */
    {
      fprintf(stderr,"superflous parameter %s ignored\n",argv[i++]);
    }
  vdmdb=hd_vdmdbs(def_Vdmdbs(dbs));         /* Datenbank oeffnen */
  OUT_bool(is_wf_Vdmdb(vdmdb));
  del_vdmdb(vdmdb);                         /* Speicher freigeben */
}

/*----------------------------------------------------------------------*/

void pre_is_wf_Combase F4(
        int,i,             /* Zaehler fuer verarbeitete Parameter */
        int,argc,          /* Anzahl der Parameter */
        char **,argv,      /* Tupel der Parameter */
        Dbs,dbs)           /* Tupel der Datenbanknamen */

                             /* pre_is_wf_Combase wertet die Programm- */
                             /* parameter aus, und startet die Funktion */
                             /* is_wf_Combase */
{
Vdmdb vdmdb;

  HALLO("pre is_wf_Combase\n");
  while (argc > i)                          /* zu viele Parameter: */
    {
      fprintf(stderr,"superflous parameter %s ignored\n",argv[i++]);
    }
  vdmdb=hd_vdmdbs(def_Vdmdbs(dbs));         /* Datenbank oeffnen */
  OUT_bool(is_wf_Combase(vdmdb));
  del_vdmdb(vdmdb);                         /* Speicher freigeben */
}

/*----------------------------------------------------------------------*/

void pre_is_wf_Adt F4(
        int,i,             /* Zaehler fuer verarbeitete Parameter */
        int,argc,          /* Anzahl der Parameter */
        char **,argv,      /* Tupel der Parameter */
        Dbs,dbs)           /* Tupel der Datenbanknamen */

                             /* pre_is_wf_Adt wertet die Programm- */
                             /* parameter aus, und startet die Funktion */
                             /* is_wf_Adt */
{
Adtname adt_name;
Vdmdb vdmdb;

  HALLO("pre is_wf_Adt\n");
  if (argc<i+1)                             /* zu wenig Parameter: */
    {
      fprintf(stderr,"missing functionparameter: no adtname given\n");
      Usage();
    }
  else
    {                                       /* Parameter auswerten */
      adt_name=mkc_str(argv[i++]);
      while (argc > i)                      /* zu viele Parameter: */
        {
          fprintf(stderr,"superflous parameter %s ignored\n",argv[i++]);
        }
      vdmdb=hd_vdmdbs(def_Vdmdbs(dbs));     /* Datenbank oeffnen */
      OUT_bool(is_wf_Adt(vdmdb,adt_name));
      del_str(adt_name);                    /* Speicher freigeben */
      del_vdmdb(vdmdb);
    }
}

/*----------------------------------------------------------------------*/

void pre_is_wf_Impl F4(
        int,i,             /* Zaehler fuer verarbeitete Parameter */
        int,argc,          /* Anzahl der Parameter */
        char **,argv,      /* Tupel der Parameter */
        Dbs,dbs)           /* Tupel der Datenbanknamen */

                             /* pre_is_wf_Impl wertet die Programm- */
                             /* parameter aus, und startet die Funktion */
                             /* is_wf_Implementation */
{
Adtname adt_name;
Implname impl_name;
Vdmdb vdmdb;

  HALLO("pre is_wf_Impl\n");
  if (argc<i+2)                             /* zu wenig Parameter: */
    {
      fprintf(stderr,"missing functionparameter: no implname given\n");
      if (argc<i+1)
          fprintf(stderr,"missing functionparameter: no adtname given\n");
      Usage();
    }
  else
    {                                       /* Parameter auswerten */
      adt_name=mkc_str(argv[i++]);
      impl_name=mkc_str(argv[i++]);
      while (argc > i)                      /* zu viele Parameter: */
        {
          fprintf(stderr,"superflous parameter %s ignored\n",argv[i++]);
        }
      vdmdb=hd_vdmdbs(def_Vdmdbs(dbs));    /* Datenbank oeffnen */
      OUT_bool(is_wf_Implementation(vdmdb,adt_name,impl_name));
      del_str(adt_name);                    /* Speicher freigeben */
      del_str(impl_name);
      del_vdmdb(vdmdb);
    }
}

/*----------------------------------------------------------------------*/

void pre_is_wf_Op F4(
        int,i,             /* Zaehler fuer verarbeitete Parameter */
        int,argc,          /* Anzahl der Parameter */
        char **,argv,      /* Tupel der Parameter */
        Dbs,dbs)           /* Tupel der Datenbanknamen */

                             /* pre_is_wf_Op wertet die Programm- */
                             /* parameter aus, und startet die Funktion */
                             /* is_wf_Operation */

{
Adtname adt_name;
Implname impl_name;
Opname op_name;
Vdmdb vdmdb;

  HALLO("pre is_wf_Op\n");
  if (argc<i+3)                             /* zu wenig Parameter: */
    {
      fprintf(stderr,"missing functionparameter: no opname given\n");
      if (argc<i+2)
         fprintf(stderr,"missing functionparameter: no implname given\n");
      if (argc<i+1)
          fprintf(stderr,"missing functionparameter: no adtname given\n");
      Usage();
    }
  else
    {                                       /* Parameter auswerten */
      adt_name=mkc_str(argv[i++]);
      impl_name=mkc_str(argv[i++]);
      op_name=mkc_str(argv[i++]);
      while (argc > i)                      /* zu viele Parameter: */
        {
          fprintf(stderr,"superflous parameter %s ignored\n",argv[i++]);
        }
      vdmdb=hd_vdmdbs(def_Vdmdbs(dbs));     /* Datenbank oeffnen */
      OUT_bool(is_wf_Operation(vdmdb,adt_name,impl_name,op_name));
      del_str(adt_name);                    /* Speicher freigeben */
      del_str(impl_name);
      del_str(op_name);
      del_vdmdb(vdmdb);
    }
}

/*----------------------------------------------------------------------*/

void pre_is_wf_Sign F4(
        int,i,             /* Zaehler fuer verarbeitete Parameter */
        int,argc,          /* Anzahl der Parameter */
        char **,argv,      /* Tupel der Parameter */
        Dbs,dbs)           /* Tupel der Datenbanknamen */

                             /* pre_is_wf_Sign wertet die Programm- */
                             /* parameter aus, und startet die Funktion */
                             /* is_wf_Signature */
{
Adtname adt_name;
Opname op_name;
Vdmdb vdmdb;

  HALLO("pre is_wf_Sign\n");
  if (argc<i+2)                             /* zu wenig Parameter: */
    {
      fprintf(stderr,"missing functionparameter: no opname given\n");
      if (argc<i+1)
          fprintf(stderr,"missing functionparameter: no adtname given\n");
      Usage();
    }
  else
    {                                       /* Parameter auswerten */
      adt_name=mkc_str(argv[i++]);
      op_name=mkc_str(argv[i++]);
      while (argc > i)                      /* zu viele Parameter: */
        {
          fprintf(stderr,"superflous parameter %s ignored\n",argv[i++]);
        }
      vdmdb=hd_vdmdbs(def_Vdmdbs(dbs));    /* Datenbank oeffnen */
      OUT_bool(is_wf_Signature(vdmdb,adt_name,op_name));
      del_str(adt_name);                    /* Speicher freigeben */
      del_str(op_name);
      del_vdmdb(vdmdb);
    }
}

/*----------------------------------------------------------------------*/

void pre_is_wf_Comp F4(
        int,i,             /* Zaehler fuer verarbeitete Parameter */
        int,argc,          /* Anzahl der Parameter */
        char **,argv,      /* Tupel der Parameter */
        Dbs,dbs)           /* Tupel der Datenbanknamen */

                             /* pre_is_wf_Comp wertet die Programm- */
                             /* parameter aus, und startet die Funktion */
                             /* is_wf_Component */

{
Adtname adt_name;
Implname impl_name;
Opname op_name;
Compname comp_name;
Vdmdb vdmdb;

  HALLO("pre is_wf_Comp\n");
  if (argc<i+4)                             /* zu wenig Parameter: */
    {
      fprintf(stderr,"missing functionparameter: no compname given\n");
      if (argc<i+3)
        fprintf(stderr,"missing functionparameter: no opname given\n");
      if (argc<i+2)
         fprintf(stderr,"missing functionparameter: no implname given\n");
      if (argc<i+1)
          fprintf(stderr,"missing functionparameter: no adtname given\n");
      Usage();
    }
  else
    {                                       /* Parameter auswerten */
      adt_name=mkc_str(argv[i++]);
      impl_name=mkc_str(argv[i++]);
      op_name=mkc_str(argv[i++]);
      comp_name=mkc_str(argv[i++]);
      while (argc > i)                      /* zu viele Parameter: */
        {
          fprintf(stderr,"superflous parameter %s ignored\n",argv[i++]);
        }
      vdmdb=hd_vdmdbs(def_Vdmdbs(dbs));     /* Datenbank oeffnen */
      OUT_bool(is_wf_Component(vdmdb,adt_name,impl_name,op_name,comp_name));
      del_str(adt_name);                    /* Speicher freigeben */
      del_str(impl_name);
      del_str(op_name);
      del_str(comp_name);
      del_vdmdb(vdmdb);
    }
}

/*----------------------------------------------------------------------*/

int main F2(
        int,argc,
        char **,argv)
                             /* Das Hauptprogramm liest die Programm- */
                             /* parameter ein und startet die Funktionen */
{
char *p;
int i=1;
Str tracestring;
Vdmdbname vdmdb_name;
Dbs dbs;

  init_vdmdm();

  if (argc==1)                         /* keine Programmparameter */
    Usage();                           /* => usage ausgeben */
  else
    {
      dbs=mk_empty_dbs();
      while (argc>i)
        {
          p=argv[i++];
          if (*p++=='-')
            switch(*p)                 /* Funktionen starten */
              {
                case 'T' : /* Trace-variable setzen */
                           tracestring=mkc_str(argv[i-1]);
                           gptrace=len_str(tracestring)-1;
                           del_str(tracestring);        
                           break;
                case 'B' : /* Datenbanknamen ermitteln */
                           vdmdb_name=mkc_str(argv[i-1]);
                           if (len_dbs(dbs)==0)
                             dbs=appr_dbs(dbs,sub_tup_str(vdmdb_name,
                                          3,len_str(vdmdb_name)));
                           else
                             {
                               fprintf(stderr,"you can only work on one database\n");
                               fprintf(stderr,"parameter ignored :");
                               OUT_str(vdmdb_name);
                               del_str(vdmdb_name);
                             }
                           break;
                case 'v' : pre_is_wf_Vdmdb(i,argc,argv,dbs);i=argc;break;
                case 'C' : pre_is_wf_Combase(i,argc,argv,dbs);i=argc;break;
                case 'a' : pre_is_wf_Adt(i,argc,argv,dbs);i=argc;break;
                case 'i' : pre_is_wf_Impl(i,argc,argv,dbs);i=argc;break;
                case 'o' : pre_is_wf_Op(i,argc,argv,dbs);i=argc;break;
                case 's' : pre_is_wf_Sign(i,argc,argv,dbs);i=argc;break;
                case 'c' : pre_is_wf_Comp(i,argc,argv,dbs);i=argc;break;
                default  : fprintf(stderr,"wrong function parameter\n");
                           Usage();
               }
          else
            {
              fprintf(stderr,"function parameter has to start with -\n");
              Usage();
            }
        }
      TRACE1( { profile_dbs(); profile_vdmdb(); profile_strset(); profile_str(); } )

    }
  finish_vdmdm();

  return 0;
}
