/*----------------------------------------------------------------------*/

/* Modul:                  $Id: tool.c,v 1.2 1994/03/07 13:43:39 vdm Exp $ */
/* Ersteller:              Gerrit Petrich, WI 8215, WS 93 */

/*----------------------------------------------------------------------*/

/* Aufgabe:   "tool.c" realisiert die Einfuege- und Loesch- */
/*            funktionen fuer die VDM-Klassenbibliothek. */

/*----------------------------------------------------------------------*/

#include "header.h"     /* header fuer Booleschen Datentyp, IF THEN ELSE,... */
#include "tool.h"       /* eigener header */
#include "dbs.h"        /* header fuer exists_... */

/*----------------------------------------------------------------------*/

Bool write_Vdmdb F2(Vdmdbname,vdmdb_name,Vdmdb,vdmdb)

/* speichert eine VDM-Datenbank im File-System ab */

{
Bool bool;

  TRACE2( { fprintf(stderr,"writing database: ");
            pr_vdmdb(vdmdb); } )

#ifdef FLAT
  {
    Outbytes ob;
    Vdmdbname vn;
    vn = conc_str(copy_str(vdmdb_name),mkc_str( DEFAULT_FLAT_EXTENSION ));
    ob = cvtotupl_vdmdb(vdmdb);
    bool = fwrite_outbytes(ob,vn);
    del_str(vn);
  }
#else
  bool=fwrite_vdmdb(vdmdb,vdmdb_name);
#endif

  fprintf(stderr,"writing of ");
  puts_str(vdmdb_name,stderr);
  if (bool)
    fprintf(stderr," successful\n");
  else fprintf(stderr," not successful");
  return(bool);
}

/*----------------------------------------------------------------------*/

Vdmdb create_Vdmdb F1(Vdmdbname,vdmdb_name)

/* Erzeugt eine neue VDM-Datenbank */

{
  return(mk1_vdmdb(vdmdb_name,mk_empty_adt()));  /* -- action -- */
}

/*----------------------------------------------------------------------*/

Vdmdb create_Adt F2(Vdmdbname,vdmdb_name,Adtname,adtname)

/* Erzeugt einen neuen ADT */

{
  HALLO("starting create_Adt\n");
  TRACE1( { OUT2_str(vdmdb_name,adtname); } )

  return(mk1_vdmdb(copy_str(adtname),mk_empty_adt()));
}

/*----------------------------------------------------------------------*/

Vdmdb create_Impl F3(Vdmdbname,vdmdb_name,Adtname,adtname,Implname,implname)

/* Erzeugt eine neue Implementation */

{
  HALLO("starting create_Impl\n");
  TRACE1( { OUT3_str(vdmdb_name,adtname,implname); } )

  return(mk1_vdmdb(copy_str(adtname),
                      mk1_adt(copy_str(implname),mk_empty_impl())));
}

/*----------------------------------------------------------------------*/

Vdmdb create_Operation F4(Vdmdbname,vdmdb_name,Adtname,adtname,
                          Implname,implname,Opname,opname)

/* Erzeugt eine neue Operation */

{
  HALLO("starting create_Op\n");
  TRACE1( { OUT4_str(vdmdb_name,adtname,implname,opname); } )

  return(mk1_vdmdb(copy_str(adtname),
                      mk1_adt(copy_str(implname),
                                 mk1_impl(copy_str(opname),
                                             mk_empty_operation()))));
}

/*----------------------------------------------------------------------*/

Vdmdb insert_Component F6(Vdmdbname,vdmdb_name,Adtname,adtname,
                          Implname,implname,Opname,opname,
                          Compname,compname,Component,component)

/* Fuegt eine Komponente in die Datenbank ein */

{
  HALLO("starting insert_Comp\n");
  TRACE1( { OUT5_str(vdmdb_name,adtname,implname,opname,compname); } )

  return(mk1_vdmdb(copy_str(adtname),         /* -- action -- */
                mk1_adt(copy_str(implname),
                       mk1_impl(copy_str(opname),
                              mk1_operation(copy_str(compname),
                                               component)))));
}

/*----------------------------------------------------------------------*/

int dummy_unlink F1(char *,path)

/* loescht eine Datei (=Komponente) in der Datenbank,
   bzw. gibt eine Loeschmeldung bei gptrace=3 aus.
   Bei erfolgreicher Ausfuehrung wird 0 zurueckgegeben */

{
 TRACE2( { fprintf(stderr,"following file would have been deleted: ");
           fprintf(stderr,path);
           fprintf(stderr,"\n");
           return 0; } )
 return unlink(path);
}

/*----------------------------------------------------------------------*/

int dummy_rmdir F1(char *,path)

/* loescht ein Verzeichnis in der Datenbank,
   bzw. gibt eine Loeschmeldung bei gptrace=3 aus.
   Bei erfolgreicher Ausfuehrung wird 0 zurueckgegeben */

{
   TRACE2( { fprintf(stderr,"following path would have been removed: ");
             fprintf(stderr,path);
             fprintf(stderr,"\n");
             return 0; } )

   return rmdir(path);
}

/*----------------------------------------------------------------------*/

Bool rm_Component F5(Vdmdbname,vdmdb_name,Adtname,adtname,Implname,implname,
                     Opname,opname,Compname,compname)

/* Entfernt eine Komponente aus der DB.
   Bei Erfolg wird TRUE zurueckgegeben */

{
Str tmp;
char *path;

  HALLO("starting rm_Comp\n");
                                                 /* Datei- und Pfadname */
                                                 /* zum Loeschen */
                                                 /* zusammenstellen */
  LET tmp=conc_path_str(copy_str(vdmdb_name),copy_str(adtname)) IN
  LET tmp=conc_path_str(tmp,copy_str(implname)) IN
  LET tmp=conc_path_str(tmp,copy_str(opname)) IN
  LET tmp=conc_path_str(tmp,copy_str(compname)) IN
  LET path=malloc(len_str(tmp)+1) IN
  LET cpc_str(tmp,path) IN
  del_str(tmp);
  return                                         /* -- action -- */
    (IF dummy_unlink(path)==0
     THEN free(path),
          TRUE
     ELSE (free(path),
           FALSE));
}

/*----------------------------------------------------------------------*/

Bool rm_Operation F4(Vdmdbname,vdmdb_name,Adtname,adtname,
                     Implname,implname,Opname,opname)

/* Entfernt eine Operation aus der DB.
   Bei Erfolg wird TRUE zurueckgegeben */

{
Str tmp;
char *path;

  HALLO("starting rm_Op\n");
                                                 /* Pfadnamen */
                                                 /* zum Loeschen */
                                                 /* zusammenstellen */
  LET tmp=conc_path_str(copy_str(vdmdb_name),copy_str(adtname)) IN
  LET tmp=conc_path_str(tmp,copy_str(implname)) IN
  LET tmp=conc_path_str(tmp,copy_str(opname)) IN
  LET path=malloc(len_str(tmp)+1) IN
  LET cpc_str(tmp,path) IN
  del_str(tmp);
  return                                         /* -- action -- */
    (IF dummy_rmdir(path)==0
     THEN free(path),
          TRUE
     ELSE (free(path),
           FALSE));
}

/*----------------------------------------------------------------------*/

Bool rm_Impl F3(Vdmdbname,vdmdb_name,Adtname,adtname,Implname,implname)

/* Entfernt eine Implementation aus der DB.
   Bei Erfolg wird TRUE zurueckgegeben */

{
Str tmp;
char *path;

  HALLO("starting rm_Impl\n");
                                                 /* Pfadnamen */
                                                 /* zum Loeschen */
                                                 /* zusammenstellen */
  LET tmp=conc_path_str(copy_str(vdmdb_name),copy_str(adtname)) IN
  LET tmp=conc_path_str(tmp,copy_str(implname)) IN
  LET path=malloc(len_str(tmp)+1) IN
  LET cpc_str(tmp,path) IN
  del_str(tmp);
  return                                         /* -- action -- */
    (IF dummy_rmdir(path)==0
     THEN free(path),
          TRUE
     ELSE (free(path),
           FALSE));
}

/*----------------------------------------------------------------------*/

Bool rm_Adt F2(Vdmdbname,vdmdb_name,Adtname,adtname)

/* Entfernt einen ADT aus der DB.
   Bei Erfolg wird TRUE zurueckgegeben */

{
Str tmp;
char *path;

  HALLO("starting rm_Adt\n");
                                                 /* Pfadnamen */
                                                 /* zum Loeschen */
                                                 /* zusammenstellen */
  LET tmp=conc_path_str(copy_str(vdmdb_name),copy_str(adtname)) IN
  LET path=malloc(len_str(tmp)+1) IN
  LET cpc_str(tmp,path) IN
  del_str(tmp);
  return                                         /* -- action -- */
    (IF dummy_rmdir(path)==0
     THEN free(path),
          TRUE
     ELSE (free(path),
           FALSE));
}

/*----------------------------------------------------------------------*/

Bool rm_Vdmdb F1(Vdmdbname,vdmdb_name)

/* Loescht eine Datenbank.
   Bei Erfolg wird TRUE zurueckgegeben */

{
char *path;

  HALLO("starting rm_Vdmdb\n");
                                                 /* Pfadnamen */
                                                 /* zum Loeschen */
                                                 /* zusammenstellen */
  LET path=malloc(len_str(vdmdb_name)+1) IN
  LET cpc_str(vdmdb_name,path) IN
  return                                         /* -- action -- */
    (IF dummy_rmdir(path)==0
     THEN free(path),
          TRUE
     ELSE (free(path),
           FALSE));
}

/*----------------------------------------------------------------------*/

