/***************************************************************************/
/*                                                                         */
/* IDENTITY:         define                                                */
/*                                                                         */
/* TYPE:             Module (project: cpp, sub-module to: parser)          */
/*                                                                         */
/* HISTORY:          DESIGN           Schlotfeldt  03.04.91                */
/*                   CODE             Schlotfeldt  27.04.91                */
/*                   VERSION(CURRENT) Schlotfeldt  11.06.91  16:55         */
/*                                                                         */
/* FUNCTION:         Parses a stream of 'symbols' generated by 'next_lex_e'*/
/*                   (module 'expandm') and terminated (for this module)   */
/*                   by the ENDOFFILE or NEWLINE 'symbol'.                 */
/*                   'define' assumes the stream to describe a pp-command  */
/*                   line, which was already identified as #define-line by */
/*                   reading the first 'symbol' of this line by the        */
/*                   'parser' module.                                      */
/*                                                                         */
/* DATA/TABLES:      -                                                     */
/*                                                                         */
/* PROCEDURES:       for external use: 'ppDEFINE'                          */
/*                                                                         */
/***************************************************************************/
#include "define.h"

#include "expandm.h"
#include "macro.h"
#include "lexbuf.h"

#include "myalloc.h"
#include "errmsg.h"

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         ppDEFINE                                              */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         #define-line sub-parser (see function of module)      */
/*                                                                         */
/* PARAMETERS:       -                                                     */
/*                                                                         */
/* RETURN VALUE:     -                                                     */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   See inline documentation                              */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  Last 'symbol' read by the 'parser' module before      */
/*                   calling this function is assumed to be the DEFINE     */
/*                   'symbol'                                              */
/*                                                                         */
/* ERROR CONDITIONS: Calls 'errmsg' module/function on illegal syntax of   */
/*                   the pp-command line or illegal re-definition of the   */
/*                   'macro', which was to be defined by this function.    */
/*                                                                         */
/***************************************************************************/
void ppDEFINE S0(void)
{
# define reterr(n,s)                   /* reterr:                          */\
  {                                    /*   generates the error message 'n'*/\
    errmsg(1,n,s,NULL);                /*   with sub-string 's' and jumps  */\
    goto error;                        /*   to the error exit of this      */\
  }                                    /*   function.                      */

  symbol name,                         /* name of 'macro' to be defined    */
	 p,                            /* temporary storage for a 'symbol' */
	 *param=NULL;                  /* list of 'symbols' used as param. */
				       /* for the new 'macro'              */
  size_t cnt;                          /* temporary variable for counting  */

  p=next_lex_e(NO_EXP);                /* skip optional WHITESPace         */
  skipsp(p,NO_EXP);
  if((name=p)->type!=IDENT)            /* check 'macro' name to be an      */
    reterr( 7,NULL)                    /* identifier -> else: error        */
  p=next_lex_e(NO_EXP);
  if(p==L_PARENTH)                     /* 'macro' has parameter ?          */
  {
    cnt=0;                             /* YES: parse parameter list        */
    p=next_lex_e(NO_EXP);
    skipsp(p,NO_EXP);
    if(p!=R_PARENTH)                        /* there is at least 1 param.? */
    while(1)
    {
      size_t i;

      if(p->type!=IDENT)                    /* parameter is a identifier ? */
	reterr(19,NULL)                     /* no: -> error                */
      param=(symbol *)realloc(param,++cnt*sizeof(symbol));
      param[cnt-1]=p;                       /* append parameter name (as   */
      for(i=0;i<cnt-1;i++)                  /* 'symbol') to the initialized*/
	if(param[i]==p)                     /* param.list and check names  */
	  reterr(20,p->text)                /* to be unique!               */
      p=next_lex_e(NO_EXP);
      skipsp(p,NO_EXP);
      if(p==R_PARENTH) break;               /* list is complete if ')' is  */
      if(p!=COMMA)                          /* reached. otherwise there    */
	reterr(21,NULL)                     /* must follow a COMMA with    */
      p=next_lex_e(NO_EXP);                 /* opt.WHITESPace to separate  */
      skipsp(p,NO_EXP);                     /* the next parameter name     */
    }
    p=next_lex_e(NO_EXP);
    param=(symbol *)realloc(param,++cnt*sizeof(symbol));
    param[cnt-1]=NULL;                      /* append NULL termination to  */
					    /* the parameter list          */
  }                                    /* NO : leave param.list unchanged  */
				       /*      (=NULL)                     */
  skipsp(p,NO_EXP);                    /* skip space between 'macro' name  */
  uget1_lex(p);			       /* (+param. list) and replace text  */

  if(!definem(name,param,              /* try to define the new 'macro'    */
	      next_line(NO_EXP),0))    /* with the replace text read from  */
				       /* the remaining line               */
    errmsg(1,22,name->text,NULL);      /* generate error on "illegal re-   */
  return;                              /* defintion"                       */

  error:                               /* error exit:                      */
  skip2eol(p,NO_EXP);                  /*   ignore remaining line and      */
  free(param);                         /*   'free' 'param' (if prevously   */
}                                      /*   allocated)                     */
