/***************************************************************************/
/*                                                                         */
/* IDENTITY:         myalloc                                               */
/*                                                                         */
/* TYPE:             Module (project: cpp)                                 */
/*                                                                         */
/* HISTORY:          DESIGN           Schlotfeldt  12.03.91                */
/*                   CODE             Schlotfeldt  12.03.91                */
/*                   VERSION(CURRENT) Schlotfeldt  18.06.91  15:29         */
/*                                                                         */
/* FUNCTION:         'myalloc.h' should be used instead of 'alloc.h' (or   */
/*		     other include files defining function for memory      */
/*                   allocation). 'myalloc.h' defines the macros 'malloc', */
/*                   'realloc', 'free' and 'strdup'. Calls to this macros  */
/*                   (usually known as library functions) are redirected   */
/*                   to this module.                                       */
/*                   The functions of this module are a little different   */
/*                   and have their own build-in error check for "out of   */
/*                   memory" (uses 'errmsg' module)                        */
/*                   'myalloc' uses type *char for memory block pointers,  */
/*                   cause some compilers don't like the type *void        */
/*                                                                         */
/* DATA/TABLES:      -                                                     */
/*                                                                         */
/* PROCEDURES:       for external use: 'my_malloc', 'my_realloc',          */
/*                       	       'my_free', 'my_strdup'              */
/*                                                                         */
/***************************************************************************/
#include "errmsg.h"

#include "myalloc.h"

#undef malloc                          /* disable macros defined in        */
#undef realloc                         /* 'myalloc.h' to enable use of     */
#undef free                            /* library functions (with the same */
#undef strdup                          /* name)                            */

                                       /* include prototype for malloc     */
#if __STDC__
#include <stdlib.h>
#endif

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         my_malloc                                             */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         works as library function 'malloc'                    */
/*                                                                         */
/* PARAMETERS:       Size of memory block to allocate                      */
/*                                                                         */
/* RETURN VALUE:     Pointer to the allocated block or NULL if size was 0  */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   -                                                     */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  -                                                     */
/*                                                                         */
/* ERROR CONDITIONS: Calls 'errmsg' module for "out of memory" message     */
/*                   and program termination, if size was !=0 and library  */
/*                   function returned NULL with the passed parameter      */
/*                                                                         */
/***************************************************************************/
char *my_malloc S1(size_t,size)
{
  char *p;

  if(!size) return(NULL);
  if(!(p=(char *)malloc(size)))
    errmsg(0,28,NULL,NULL);
  return(p);
}

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         my_realloc                                            */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         works as library function 'realloc'                   */
/*                                                                         */
/* PARAMETERS:       1) Pointer to memory block to re-allocate             */
/*                   2) New size of memory block                           */
/*                                                                         */
/* RETURN VALUE:     Pointer to the re-allocated block or NULL if size was */
/*                   0                                                     */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   -                                                     */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  -                                                     */
/*                                                                         */
/* ERROR CONDITIONS: Calls 'errmsg' module for "out of memory" message     */
/*                   and program termination, if size was !=0 and library  */
/*                   function returned NULL with the passed parameter      */
/*                                                                         */
/***************************************************************************/
char *my_realloc S2(char *,block,size_t,size)
{
  char *p;

  if(!block) return(my_malloc(size));
  if(!(p=(char *)realloc(block,size)) && size)
    errmsg(0,28,NULL,NULL);
  return(p);
}

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         my_free                                               */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         works as library function 'free', calls with          */
/*                   parameter NULL are ignored (will not cause an error)  */
/*                                                                         */
/* PARAMETERS:       Pointer to memory block to set 'free'                 */
/*                                                                         */
/* RETURN VALUE:     -                                                     */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   -                                                     */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  -                                                     */
/*                                                                         */
/* ERROR CONDITIONS: -                                                     */
/*                                                                         */
/***************************************************************************/
void my_free S1(char *,block)
{
  if(block) free(block);
}

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         my_strdup                                             */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         works as library function 'strdup'                    */
/*                                                                         */
/* PARAMETERS:       String to duplicate (allocation and copying)          */
/*                                                                         */
/* RETURN VALUE:     Pointer to the new string or NULL if parameter was    */
/*                   NULL                                                  */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   -                                                     */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  -                                                     */
/*                                                                         */
/* ERROR CONDITIONS: Calls 'my_malloc' instead of 'malloc' to check for    */
/*                   "out of memory"                                       */
/*                                                                         */
/***************************************************************************/
char *my_strdup S1(char *,s1)
{
  char *s2;

  if(!s1) return(NULL);
  s2=my_malloc(strlen(s1)+1);
  return(strcpy(s2,s1));
}
