/***************************************************************************/
/*                                                                         */
/* IDENTITY:         pragma                                                */
/*                                                                         */
/* TYPE:             Module (project: cpp, sub-module to: parser)          */
/*                                                                         */
/* HISTORY:          DESIGN           Schlotfeldt  03.04.91                */
/*                   CODE             Schlotfeldt  27.04.91                */
/*                   VERSION(CURRENT) Schlotfeldt  30.05.91  11:17         */
/*                                                                         */
/* FUNCTION:         Parses a stream of 'symbols' generated by 'next_lex_e'*/
/*                   (module 'expandm') and terminated (for this module)   */
/*                   by the ENDOFFILE or NEWLINE 'symbol'.                 */
/*                   In fact there is nothing to parse then the end of     */
/*                   line, cause there is no #pragma flag implemented yet. */
/*                                                                         */
/* DATA/TABLES:      -                                                     */
/*                                                                         */
/* PROCEDURES:       for external use: 'ppPRAGMA'                          */
/*                                                                         */
/***************************************************************************/
#include "pragma.h"

#include "expandm.h"

#include "errmsg.h"

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         ppPRAGMA                                              */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         #pragma-line "sub-parser" (see function of module)    */
/*                                                                         */
/* PARAMETERS:       -                                                     */
/*                                                                         */
/* RETURN VALUE:     -                                                     */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   See inline documentation                              */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  Last 'symbol' read by the 'parser' module before      */
/*                   calling this function is assumed to be the PRAGMA     */
/*                   'symbol'                                              */
/*                                                                         */
/* ERROR CONDITIONS: Calls 'errmsg' module/function on illegal syntax of   */
/*                   the pp-command line                                   */
/*                                                                         */
/***************************************************************************/
void ppPRAGMA S0(void)
{
  symbol p;                            /* temporary storage for a 'symbol' */

  p=next_lex_e(NO_EXP);                /* get next 'symbol' and check for  */
  if(!endofline(p))                    /* end of line                      */
  {
    errmsg(1,25,NULL,NULL);            /* generate error message if last   */
				       /* 'symbol' doesn't terminate the   */
				       /* actual line                      */
    skip2eol(p,NO_EXP)                 /* skip to end of line any way      */
  }
}
