/***************************************************************************/
/*                                                                         */
/* IDENTITY:         stringu                                               */
/*                                                                         */
/* TYPE:             Module (project: cpp)                                 */
/*                                                                         */
/* HISTORY:          DESIGN           Schlotfeldt  13.03.91                */
/*                   CODE             Schlotfeldt  05.04.91                */
/*                   VERSION(CURRENT) Schlotfeldt  18.06.91  18:43         */
/*                                                                         */
/* FUNCTION:         Defines some utility functions for special string     */
/*                   handling.                                             */
/*                                                                         */
/* ------------------------------------------------------------------------

   %W% %D%

   Change History:
   ===============
   Apr-25-92:	'digt' and 'd' changed from 'char *' to 'char .[]'

   ------------------------------------------------------------------------*/
/*                                                                         */
/* DATA/TABLES:      -                                                     */
/*                                                                         */
/* PROCEDURES:       for external use: 'str2ul', 'strtoc', 'app_string',   */
/*                       	       'conv2name', 'strcomp'              */
/*                                                                         */
/***************************************************************************/
#include "stringu.h"

#include "myalloc.h"
#include "myctype.h"

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         str2ul                                                */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         Converts a string to a unsigned long value using      */
/*                   a optional specified radix                            */
/*                                                                         */
/* PARAMETERS:       1) String to be parsed / converted                    */
/*                   2) Storage for a pointer pointing to the first non    */
/*                      recognized character in the selected string        */
/*                   3) radix to use or 0 for autodetect ("0x..." for 16,  */
/*                      "0..." for 8 and any other for 10)                 */
/*                                                                         */
/* RETURN VALUE:     unsigned long value represented by the string         */
/*                   and the modified 2nd parameter (for error handling)   */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   See inline documentation                              */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  The passed string must not be NULL, it must match one */
/*                   of the following formats:                             */
/*                                                                         */
/*                      0x[hex-digits] or 0X[hex-digits],                  */
/*                      0[octal-digits],                                   */
/*                      or [decimal-digits]                                */
/*                                                                         */
/*                   1st format is unly valid for radix==0 (autodetect)    */
/*                                                                         */
/* ERROR CONDITIONS: returns the position of the 1st not parsed character  */
/*                   in the given string (must be checked for end of string*/
/*                   by the calling function for error detection)          */
/*                                                                         */
/***************************************************************************/
unsigned long str2ul S3(char *,s,char **,endptr,int,radix)
{
  char          *dig="0123456789abcdef",    /* string to look up the value */
					    /* of a hexadecimal digit      */
		*p;                         /* pointer to a position in the*/
					    /* above string (p-dig is the  */
					    /* searched value)             */
  unsigned long d,                          /* value of the actual digit   */
		r=0;                        /* result of conversion        */

  if(!radix)                           /* scan first characters of the     */
  {                                    /* string for 'radix' detect if no  */
    if(*s!='0') radix=10;              /* 'radix' specified                */
	   else if(tolower(*++s)=='x')
		{
		  radix=16;
		  if(!strchr(dig,*++s))     /* "0x"/"0X" must be followed  */
		  {                         /* at least one hex.dec.digit! */
		    *endptr=s-1;
		    return(r);
		  }
		}
		else radix=8;
  }

  if(radix>0 && radix<=16)             /* see if radix is valid            */
  while((p=strchr(dig,tolower(*s)))    /* scan til NUL or illegal digit is */
	&& *s)                         /* reached                          */
  {
    d=(unsigned)(p-dig);               /* calculate value of actual digit  */
    if(d>=radix) break;                /* check if it's less than 'radix'  */
    r*=radix;                          /* and calculate new result value   */
    r+=d;
    s++;
  }
  *endptr=s;                           /* save last position of scanning   */
  return(r);                           /* to 'endptr' and return           */
}

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         strtoc                                                */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         Converts a string to a character (one character may   */
/*                   be expressed by more than one character in c-syntax)  */
/*                                                                         */
/* PARAMETERS:       1) Pointer to string to be parsed / converted         */
/*                      (returns new string excluding the parsed           */
/*                      characters)                                        */
/*                   2) Flag: 1 = character to parse belongs to a string   */
/*                                (no check on termination)                */
/*                            0 = character to parse belongs to a character*/
/*                                constant (terminated by a single quote)  */
/*                   3) Storage for error flag                             */
/*                                                                         */
/* RETURN VALUE:     character value represented by the string             */
/*                   and the modified parameters 3 and 1                   */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   See inline documentation                              */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  The passed string must not be NULL.                   */
/*                   Pay attention to escape sequences \\, \", \? and \' : */
/*                   these sequences may be already converted to single    */
/*                   (special) characters. See 'fetch' module !            */
/*                                                                         */
/* ERROR CONDITIONS: returns the position of the 1st not parsed character  */
/*                   in the given string (must be checked for end of string*/
/*                   by the calling function for error detection)          */
/*                                                                         */
/***************************************************************************/
char strtoc S3(char **,s,char,str,char *,e_flag)
{
  char   res,                          /* result of function               */
	 *escp="abfnrtv'\"\\?",
	 *escc="\007\010\014\012\015\011\013\047\042\134\077";
				       /* 'escp' contains all characters   */
				       /* valid to use in a 2 character    */
				       /* escape sequence (ANSI !!)        */
				       /* 'escc' contains the corresponding*/
				       /* control characters (not written  */
				       /* as normal esc.sequ. cause not    */
				       /* every compiler knows each sequ.!)*/

  char   digt[4],                      /* temporary storage for string of  */
				       /* max 3 characters (analyzeable by */
				       /* 'str2ul')                        */
	 *p;                           /* temporary storage for error pos. */
				       /* returned by 'str2ul' in this con-*/
				       /* text                             */

  int    hex=1;                        /* flag to differ hex./oct. numbers */

  digt[0]=digt[1]=digt[2]=digt[3]=0;
    if((res= *(*s))==ESC)
    switch(*(++(*s)))                  /* handle escape sequences:         */
    {
      case '0' :                            /* esc.sequ. using numbers:    */
      case '1' :
      case '2' :
      case '3' :
      case '4' :
      case '5' :
      case '6' :
      case '7' : hex=0;                     /* set 'hex' to 0 (init.value  */
					    /* was 1) for octal digits     */
      case 'x' :
      case 'X' : (*s)+=hex;                 /* 'x'/'X' for hex.dec.digits  */
		 strncpy(digt,(*s),3);      /* copy following characters   */
					    /* (max.3) to 'digt'           */
		 res                        /* try to convert 'digt' to a  */
		  =str2ul(digt,&p,hex?16:8);/* num.value using radix 8/16  */
		 if(p!=digt)                /* check if there was at least */
					    /* 1 character convertable     */
		   (*s)+=(p-digt);          /* skip these character(s)     */
		 else
		 {                          /* or set error flag and undo  */
		   *e_flag=1;               /* all increments to '*s'      */
		   (*s)-=1+hex;
		 }
		 break;
      default  :                            /* other esc.sequences :       */
		 if(strchr(escp,*(*s))
		    && *(*s))               /* translate sequence to single*/
		   res                      /* character if known sequence */
		    =escc[strchr(escp,*(*s)++)-escp];
		 else
		 {
		   *e_flag=1;               /* or set error flag and undo  */
		   (*s)-=2;                 /* all increments to '*s'      */
		 }
		 break;
    }
  else                                 /* handle normal characters (mostly */
  {                                    /* represented by itself)           */
    if(str || res!='\'')
      (*s)++;                               /* skip character or set error */
    else                                    /* flag and leave '*s' un-     */
      *e_flag=1;                            /* changed if character was    */
					    /* illegal (for use while the  */
					    /* calling functions parses a  */
					    /* character constant!)        */
    switch(res)
    {
      case APOSTR : return('\'');           /* re-translate special charac-*/
      case QUOTES : return('\"');           /* ters (see module 'fetch')   */
      case HASHM  : return('#');
    }
  }
  return(res);
}

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         app_string                                            */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         Appends to a dynamic allocated string (may be NULL)   */
/*                   another string (no restrictions including NULL)       */
/*                                                                         */
/* PARAMETERS:       1) Pointer to string to extend                        */
/*                   2) String to append to the first string               */
/*                                                                         */
/* RETURN VALUE:     modified 1st parameter                                */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   Uses 'realloc' and 'strcpy' to resize and extend the  */
/*                   old string (1st parameter)                            */
/*                   Calls with 2nd parameter == NULL are ignored.         */
/*                   Calls with 1st parameter == NULL will create a new    */
/*                   string (dynamic allocated)                            */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  The first string must be dynamic allocated or NULL.   */
/*                   (re-allocation of static objects will cause chaos!)   */
/*                                                                         */
/* ERROR CONDITIONS: -                                                     */
/*                                                                         */
/***************************************************************************/
void app_string S2(char **,dest,char *,src)
{
  size_t dl;                           /* length of string to be extended  */

  if(src)                              /* leave old string un-changed on   */
  {                                    /* 'src'==NULL                      */
    dl= *dest?strlen(*dest):0;
    *dest=(char *)realloc(*dest,dl+strlen(src)+1);
    strcpy(*dest+dl,src);
  }
}

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         conv2name                                             */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         Converts a STRING 'symbol' to a normal string         */
/*                   (the original string may contain escape sequences     */
/*                   represented by multiple characters!)                  */
/*                                                                         */
/* PARAMETERS:       'symbol' to convert                                   */
/*                                                                         */
/* RETURN VALUE:     name (as character string) or NULL for error          */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   The 'text' part of the 'symbol' is processed by       */
/*                   'strtoc'. Each character returned by 'strtoc' is      */
/*                   appended to the string 's' building the return value. */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  -                                                     */
/*                                                                         */
/* ERROR CONDITIONS: Returns error (NULL) on illegal 'symbols', illegal    */
/*                   strings (must not start with 'L' or be empty) and     */
/*                   illegal characters in the string (remarked by         */
/*                   'strtoc'                                              */
/*                                                                         */
/***************************************************************************/
char *conv2name S1(symbol,sp)
{
  char   *p,                           /* pointer to positions in 'text'   */
				       /* part of 'symbol' 'sp'            */
	 *s=NULL,                      /* return value                     */
	 d[2],                         /* dummy string ('conv2name' will   */
				       /* change its 1st character) to call*/
				       /* 'strtoc'                         */
	 err=0;                        /* error flag                       */
  d[0]=' ';d[1]=0;
  if(!sp)              return(s);      /* return error on parameter NULL,  */
  if(sp->type!=STRING) return(s);      /* 'symbol' of other type than      */
  p=sp->text;                          /* STRING, string of w_char charac- */
  if(p[0]=='L') return(s);             /* ters (starting with 'L') or an   */
  if(!p[2]    ) return(s);             /* empty string.                    */

  p++;
  while(*(p+1) && !err)                /* process string to build 's'      */
  {
    d[0]=strtoc(&p,1,&err);                 /* convert each character (may */
    if(err)				    /* be a escape sequence)       */
    {
      free(s);
      s=NULL;                               /* return error (err!=0 ends   */
    }                                       /* loop)                       */
    else
      app_string(&s,d);                     /* or build 's' by appending   */
					    /* the modified dummy string   */
					    /* 'd' for the just processed  */
  }                                         /* character                   */
  return(s);
}

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         strcomp                                               */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         Compares 2 strings (like 'strcmp').                   */
/*                                                                         */
/* PARAMETERS:       1+2) Strings to compare                               */
/*                                                                         */
/* RETURN VALUE:     Same return value as 'strcmp', NULL strings are       */
/*                   interpreted as empty strings!                         */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   -                                                     */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  -                                                     */
/*                                                                         */
/* ERROR CONDITIONS: -                                                     */
/*                                                                         */
/***************************************************************************/
int strcomp S2(char *,s1,char *,s2)
{
  if(s1==s2)   return( 0);
  if(s1==NULL) return(-1);
  if(s2==NULL) return( 1);
  return(strcmp(s1,s2));
}
