/*----------------------------------------------------------------------

 $Id: domcheck.c,v 1.11 1996/05/01 17:04:13 vdm Exp $ 

 Aufgabe  : Ueberpruefung eines Parse-Baum auf semantische Fehler 

----------------------------------------------------------------------
 Parameter

 Eingabe  :

 Ausgabe  : 

----------------------------------------------------------------------

 Ersteller: Kai Steinau 

----------------------------------------------------------------------

 Aenderungen: 

 $Log: domcheck.c,v $
 Revision 1.11  1996/05/01 17:04:13  vdm
 cyclic dependeny graph routines changed
 (hopefully the last time)

 Revision 1.10  1996/05/01 15:50:09  vdm
 cyclic dependeny graph routines changed

 Revision 1.9  1996/04/30 20:24:26  vdm
 extensions for implementing alias as inheritance in C++
 solution: a new constructor -isa with external syntax
 Subtype is-a Basetype
 Semantics for C like "=", but for C++ class SUBTYPE
 becomes a subclass of class BASETYPE
 Problem: constructors are not inherited, so the
 constructors are passed from Basetype to Subtype
 in the generator, therefore a special operation
 constructor.mk??? introduced, wich is copied in the generator
 to the Subtype domain.
 The generation order computation did not work save,
 therefore an explicit dependency graph is constructed, with this
 graph the generation order is computed, old code still there
 function "old_gen_order"

 Revision 1.8  1996/01/15 10:01:47  vdm
 cvs backup

 * Revision 1.7  1995/11/25  13:50:36  vdm
 * minor warning correction from sun cc and g++ 2.7.0
 *
 * Revision 1.6  1994/05/01  17:08:29  vdm
 * Hans-Martins Aenderungen, make FLAT INSTFLAT neu
 *
 * Revision 1.6  1994/04/08  21:45:49  hmh
 * Merging Uwes and my version
 *
 * Revision 1.5  1993/11/27  23:38:48  hmh
 * gen_order1 and gen_order2 simplified

 * Minor correction in generation order computation
 * - in 'cyclic_dep1_str' the name of the actual domain ('e') is removed
 *   from the list to be checked for dependencies using the new 
 *   macro 'get_depend'.  This otherwise
 *   lead to problems, where 'DOMAIN -> DOMAIN' might result
 *   from some generic domain equation (like 'DOMAIN -> RESULT' for
 *   'fold' operations).
 * - the same was done in 'gen_order[12]_str.

 * Revision 1.4  1993/10/24  21:57:58  hmh
 * Changes built in to handle the '-r' option used to check
 * if a given enrichment file can be used as starting point of
 * enrichment process.
 *
 * - trace messages generated using 'trace*'.

 * Revision 1.3  1993/07/25  08:50:48  hmh
 * Several changes
 *
 * In function 'domain_check1_str_description'
 * 'missing domain equation' is made an error instead of a warning

 * Revision 1.2  1993/05/28  11:03:17  hmh
 * Compiler extended to accept the ISO VDM-SL ASCII syntax. These
 * extensions are activated through the '-N' switch.
 *
 * New functions 'check_domid*'

----------------------------------------------------------------------*/

#include "domadt.h"
#include "domfct.h"
#include "domdb.h"
#include "domcheck.h"

/*----------------------------------------------------------------------*/
/* auxiliary Macro */
#define get_depend(d,e)	sub1d_depend(copy_depend(s_depend(d)),e)

/*----------------------------------------------------------------------*/

/* Variable zur Aufnahme der Generierungsreihenfolge der ADT's */

Elems gen_order;

/*----------------------------------------------------------------------*/


/*----------------------------------------------------------------------

 Modulname: domain_check_dominfo 

 Aufgabe  : Ueberpruefung der ADT's eines Parse-Baum auf fehlende
            Definition (=PREDEF) und Implementations-Beschreibung 


 Eingabe  : Parse-Baum 'dominfo' 

 Ausgabe  : Beschreibungen korrekt Ja/Nein (TRUE/FALSE) 

----------------------------------------------------------------------*/

/* Bool domain_check_dominfo(Dominfo dominfo) */
/* Fold-Funktion, definiert in 'domadt' */


/*----------------------------------------------------------------------

 Modulname: domain_check1_str_description 

 Aufgabe  : Ueberpruefung eines ADT des Parse-Baum auf vollstaendige 
            Definition und Implementations-Beschreibung 

 Eingabe  : Name des ADT 'dn' 
            Beschreibung des ADT 'desc' 

 Ausgabe  : Beschreibung korrekt Ja/Nein (TRUE/FALSE) 

----------------------------------------------------------------------*/

Bool
domain_check1_str_description F2(
				 Str, dn,                           /* ADT-Name */
				 Description, desc)                 /* ADT-Beschreibung */
{
    /* Funktionsresultat */
    Bool res=TRUE;

    /* Der ADT hat keinen Konstruktor */
    if (is_eq_str(s_constr(s_domain(desc)),PREDEF))
	{
	    if ((is_in_crossref(dn,crossref))
		AND (is_in_params(PARAM_['W'],global_params)))
		{
		    error(mkc_str("missing definition for  "),copy_str(dn));
		    message(mkc_str(" Domain occurs first in  "),get_root_name(dn));
		}
	}

    else
	/* Fuer einen ADT mit gegebenem Konstruktor ist keine Implementation */
	/* gegeben */
	if (is_nil_implid(s_implid(s_implem(desc))))
	    {
		error(mkc_str("missing implementation description for  "),copy_str(dn));
		message(mkc_str(" Domain is first defined in  "),get_root_name(dn));
		res=FALSE;
	    }
    
    return res;
}


/*----------------------------------------------------------------------
 Modulname: domain_check 
 Aufgabe  : Steuerung der Ueperpruefung der ADT's eines Parse-Baum auf 
            fehlende Definition (=PREDEF) und Implementations-Beschreibung
----------------------------------------------------------------------
 Parameter 
 Eingabe  : Parse-Baum 'dominfo' 
 Ausgabe  : Definitions-Gleichungen korrekt Ja/Nein (TRUE/FALSE) 
----------------------------------------------------------------------*/

Bool
domain_check F1(
		Dominfo, dominfo)          /* Parse-Baum */
{
    trace1(mkc_str("Checking for complete descriptions..."),Empty);

    /* Aufruf der in 'domadt' definierten Foldfunktion */
    return
	domain_check_dominfo(dominfo);
}


/*----------------------------------------------------------------------*/


/*----------------------------------------------------------------------
 Modulname: is_domsub_depend 
 Aufgabe  : Ueberpruefung, ob alle Elemente einer Menge von ADT's in 
            der Abbildung 'crossref' enthalten sind 
----------------------------------------------------------------------
 Parameter 
 Eingabe  : Menge von ADT-Namen 
 Ausgabe  : Elemente sind in 'crossref' enthalten Ja/Nein (TRUE/FALSE) 
----------------------------------------------------------------------*/

/* Bool is_domsub_depend(Depend de) */
/* Fold-Funktion, definiert in 'domadt' */


/*----------------------------------------------------------------------
 Modulname: is_domsub1_str 
 Aufgabe  : Ueberpruefung, ob ein ADT in der Abbildung 'crossref' 
            enthalten ist 
----------------------------------------------------------------------
 Parameter 
 Eingabe  : Name des ADT 'de' 
 Ausgabe  : ADT ist in 'crossref' enthalten Ja/Nein (TRUE/FALSE) 
----------------------------------------------------------------------*/

Bool
is_domsub1_str F1(
		  Str, de)                            /* Elemente 'Elems' und 'Depend' */
{
    /* Funktionsresultat */
    Bool res;

    /* Das Element ist nicht in der Abbildung (Map) der definierten ADT's */
    /* enthalten. --> Das Element wird von einem ADT verwendet, wird aber */
    /* selbst nicht definiert. */
    if ((res=is_in_crossref(de,crossref))==FALSE)
	error(mkc_str("missing domain equation for  "),copy_str(de));

    return res;
}


/*----------------------------------------------------------------------
 Modulname: get_elems 
 Aufgabe  : Auswahl aller Elemente eines ADT, die selber ADT's sind 
            (keine Selektoren) 
----------------------------------------------------------------------
 Parameter 
 Eingabe  : Elemente des ADT 'elems' 
            Konstruktor des ADT 'constr' 
 Ausgabe  : Menge (Depend) aller Elemente die ADT's sind 
----------------------------------------------------------------------*/

/* Depend get_elems(Elems elems, Str constr) */
/* Fold-Funktion, definiert in 'domadt' */


/*----------------------------------------------------------------------
 Modulname: get_elems1_str 
 Aufgabe  : Ruekgabe eines Elements eines ADT, wenn es selber ein ADT ist 
----------------------------------------------------------------------
 Parameter 
 Eingabe  : Element 'elem' des ADT
            Index 'i' des Elements 
            Konstruktor des ADT 
 Ausgabe  : Menge (Depend) mit dem Element 'elem' oder leere Menge, wenn 
            das Element kein ADT ist 
----------------------------------------------------------------------*/

Depend 
get_elems1_str F3(
		  Str, elem,                             /* Element 'Elem' */
		  Nat1, i,                               /* Index des Elements */
		  Str, constr)                           /* Konstruktor des zugehoerigen ADT */
{

    /* fuer ADT's mit 'Tree' und 'Union' - Konstruktoren sind die */
    /* Elemente mit geradem Index ADT's */

    return
	IF is_eq_str(constr,TREE) OR is_eq_str(constr,UNION)
	THEN
           IF (NOT (i%2))
           THEN
              mk1_depend(copy_str(elem))
           ELSE
              mk_empty_depend()

        ELSE
        /* fuer ADT's mit 'Tokuni' - Konstruktoren sind keine der Elemente */
        /* ADT's */
        IF is_eq_str(constr,TOKUNI)
        THEN
           mk_empty_depend()
        ELSE
        /* fuer alle uebrigen Konstruktoren, sind alle Elemente ADT's */
           mk1_depend(copy_str(elem));
}


/*----------------------------------------------------------------------
 Modulname: complete_depends_dominfo
 Aufgabe  : Ueberpruefung, ob fuer alle Elemente aller ADT's im 
            Parse-Baum, die selber ADT's sind, eine Definitions-Gleichung 
            vorhanden ist 
----------------------------------------------------------------------
 Parameter 
 Eingabe  : Parse-Baum 'dominfo' 
 Ausgabe  : Definitions-Gleichungen komplett Ja/Nein (TRUE/FALSE) 
----------------------------------------------------------------------*/

/* Bool complete_depends_dominfo(Dominfo dominfo) */
/* Fold-Funktion definiert in 'domadt' */


/*----------------------------------------------------------------------
 Modulname: complete_depends1_str_description 
 Aufgabe  : Ueberpruefung, ob die Elemente eines ADT, die selber ADT's 
            im Parse-Baum definiert werden 
----------------------------------------------------------------------
 Parameter 
 Eingabe  : Name des ADT 'dn' 
            Beschreibung des ADT 'desc' 
 Ausgabe  : Elemente des ADT sind definiert   Ja/Nein (TRUE/FALSE) 
----------------------------------------------------------------------*/

Bool
complete_depends1_str_description F2(
				     Str, dn,                         /* ADT-Name */
				     Description, desc)               /* ADT-Beschreibung */
{
    Depend g;                      /* Menge der Elemente eines ADT, die selber
				      ADT's sind und alle Elemente, von denen
				      der ADT abhaengt */
    Bool res=TRUE;                 /* Funktionsresultat */

    /* Die Elemente des ADT, die selber ADT's sind und die, von denen er */
    /* abhaengt (Menge 'Depend'), werden ermittelt */

    g=get_elems(s_elems(s_domain(desc)),
		s_constr(s_domain(desc)));
    g=uniond_depend(g,
		    copy_depend(s_depend(s_domain(desc))));

    /* nicht alle Elemente von denen der ADT abhaengt (seine Elemente, die */
    /* ADT's sind und die Menge 'Depend') sind definiert */
    if (NOT is_domsub_depend(g))
	printstr(Empty,mkc_str(" Domain is used by  "),copy_str(dn)),
	    message(mkc_str("   defined in  "),get_root_name(dn)),
	    res=FALSE;

    del_depend(g);

    return res;
}


/*----------------------------------------------------------------------

 Modulname: complete_domains_check 

 Aufgabe  : Steuerung der Ueberpruefung der Definitionen aller Elemente 
            aller im Parse-Baum enthaltenen ADT's, die selber ADT's sind 

 Eingabe  : Parse-Baum 

 Ausgabe  : Definitionen vollstaendig Ja/Nein (TRUE/FALSE) 

----------------------------------------------------------------------*/

Bool
complete_domains_check F1(
			  Dominfo, dominfo)                     /* Parse-Baum */
{
    Bool res;                          /* Funktionsresultat */

    trace1(mkc_str("Checking for complete domains..."),Empty);

    /* Test der ADT's des Parse Baum's auf Vollstaendigkeit */
    res=complete_depends_dominfo(dominfo);

    return res;
}


/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------

 Modulname: get_unique_elems 

 Aufgabe  : Auswahl aller Elemente eines ADT, die selber keine ADT's sind 
            (Selektoren) 

 Eingabe  : Elemente des ADT 'elems' 
            Konstruktor des ADT 

 Ausgabe  : Menge aller Selektoren des ADT 

----------------------------------------------------------------------*/

/* Depend get_unique_elems(Elems elems, Str constr) */
/* Fold-Funktion, definiert in 'domadt' */

/*----------------------------------------------------------------------

 Modulname: get_unique_elems1_str 

 Aufgabe  : Ruekgabe eines Elements eines ADT, wenn es selber ein 
            Selektor ist 

 Eingabe  : Element 'elem' 
            Index des Elements 'i' 
            Konstruktor 'constr' des Elements 

 Ausgabe  : Menge (Depend) mit dem Element 'elem' oder leere Menge, wenn 
            das Element ein ADT ist 
----------------------------------------------------------------------*/

Depend
get_unique_elems1_str F3(
			 Str, elem,                    /* Element 'Elem' */
			 Nat1, i,                      /* Index des Elements */
			 Str, constr)                  /* Konstruktor des entsprechenden Elements */
{
    
    /* fuer ADT's mit 'Tree' und 'Union' - Konstruktoren sind die */
    /* Elemente mit ungeradem Index Selektoren, deren Namen innerhalb */
    /* des Gesamt-Parse-Baums eindeutig sein muessen */

    return
	IF is_eq_str(constr,TREE)
	   OR is_eq_str(constr,UNION)
	THEN
	    IF (i%2)
	    THEN
	        mk1_depend(copy_str(elem))
            ELSE
                mk_empty_depend()

        ELSE
        /* fuer ADT's mit 'Tokuni' - Konstruktoren muessen alle Elemente */
        /* eindeuitg sien */
        IF is_eq_str(constr,TOKUNI)
        THEN
            mk1_depend(copy_str(elem))

        ELSE
        /* fuer alle uebrigen Konstruktoren ist keines der Elemente ein */
        /* Selektor */
            mk_empty_depend();
}


/*----------------------------------------------------------------------

 Modulname: unique_elems_dominfo 

 Aufgabe  : Ermittlung aller Elemente aller ADT's eines Parse-Baum, die 
            innerhalb des gesammten Parse-Baums eindeutig sein muessen 
            (Selektoren) 

 Eingabe  : Parse-Baum 'dominfo' 

 Ausgabe  : Menge (Depend) aller Elemente des Parse-Baum, die eindeutig 
            sein muessen (Selektoren) 

----------------------------------------------------------------------*/

/* Depend unique_elems_dominfo(Dominfo dominfo) */
/* Fold-Funktion, definiert in 'domadt' */

/*----------------------------------------------------------------------

 Modulname: unique_elems1_str_description 

 Aufgabe  : Ermittlung der Elemente eines ADT, die innerhalb des 
            gesamten Parse-Baum eindeutig sein muessen (Selektoren) 

 Eingabe  : Name des ADT 'dn' 
            Beschreibund des ADT 'desc' 

 Ausgabe  : Menge (Depend) der Elemente des ADT, die eindeutig sein 
            muessen 

----------------------------------------------------------------------*/

Depend
unique_elems1_str_description F2(
				 Str, dn,                          /* ADT-Name */
				 Description, desc)                /* ADT-Beschreibung */
{
    /* Rueckgabe der Elemente des ADT, deren Namen innerhalb des */
    /* Gesamt-Parse-Baums eindeutig sein muessen */
    return
	get_unique_elems(s_elems(s_domain(desc)),
			 s_constr(s_domain(desc)));
}


/*----------------------------------------------------------------------

 Modulname: make_print_depend 

 Aufgabe  : Erzeugen eines VDM-Strings aus den Elementen einer Menge 

 Eingabe  : Menge 'de' 

 Ausgabe  : VDM-String mit den durch ',' getrennten Elementen aus 'de' 

----------------------------------------------------------------------*/

/* Str make_print_depend(Depend de) */
/* Fold-Funktion, definiert in 'domadt' */


/*----------------------------------------------------------------------

 Modulname: make_print_depend1_str 

 Aufgabe  : Erzeugen eines VDM-Strings mit voranstehendem ',' aus einem 
            VDM-String 

 Eingabe  : VDM-String 's' 

 Ausgabe  : VDM-String ',' + 's' 

----------------------------------------------------------------------*/

Str make_print_depend1_str F1(
			      Str, s)                        /* Element aus 'Depend' */
{
    return conc_str(mkc_str(","),copy_str(s));
}


/*----------------------------------------------------------------------

 Modulname: nonunique_elems_dominfo 

 Aufgabe  : Ermitteln der Elemente der Beschreibungen der ADT's eines 
            Parse-Baum, die nicht eindeutig sind 

 Eingabe  : Parse-Baum 'dominfo' 

 Ausgabe  : Elemente eindeutig Ja/Nein (TRUE/FALSE) 

----------------------------------------------------------------------*/

/* Bool nonunique_elems_dominfo(Dominfo dominfo, Dominfo dominfo) */
/* Fold-Funktion, definiert in 'domadt' */

/*----------------------------------------------------------------------

 Modulname: nonunique_elems1_str_description 

 Aufgabe  : Ermitteln der Elemente der Beschreibung eines ADT des 
            Parse-Baum, die nicht eindeutig sind 

 Eingabe  : ADT-Name 'dn' 
            Beschreibung des ADT 'desc' 
            Parse-Baum 'dominfo' 

 Ausgabe  : Elemente des ADT eindeutig Ja/Nein (TRUE/FALSE) 

----------------------------------------------------------------------*/

Bool
nounique_elems1_str_description F3(
				   Str, dn,                /* ADT-Name */
				   Description, desc,      /* ADT-Beschreibung */
				   Dominfo, dominfo)       /* Parse-Baum */
{
    Depend ue;            /* Eindeutige Elemente aus 'Elems' der ADT's 
			     eines Parse-Baums ohne die Elemente des ADT 'dn' */
    Depend de;            /* Eindeutige Elemente aus 'Elems' des ADT 'dn' */
    Bool res=TRUE;        /* Funktionsresultat */
    Dominfo di;           /* Parse-Baum fuer Elemente des Parse-Baums ohne 'dn'*/

    /* Nur ADT's mit 'Tree', 'Union' und 'Tokuni' - Konstruktoren haben */
    /* Selektoren (Elemente, die eindeutige Namen haben muessen) */

    if ( (is_eq_str(s_constr(s_domain(desc)),TREE))
	 OR (is_eq_str(s_constr(s_domain(desc)),UNION))
	 OR (is_eq_str(s_constr(s_domain(desc)),TOKUNI)))
	{
	    /* Eindeutige Elemente der ADT's des Parse-Baums, ohne den ADT 'dn' */
	    ue=unique_elems_dominfo(di=sub1_dominfo(copy1_dominfo(dominfo),dn));
	    del1_dominfo(di);

	    /* Eindeutige Elemente des ADT 'dn' */
	    de=get_unique_elems(s_elems(s_domain(desc)),
				s_constr(s_domain(desc)));

	    /* Schnittmenge der Elemente des ADT's und der Elemente des */
	    /* Rest-Parse-Baums --> ist die Schnittmenge nicht leer, sind Elemente */
	    /* des ADT nicht eindeutig */
	    de=intersd_depend(ue,de);
	    if (NOT is_empty_depend(de))
		{
		    Str ermsg=make_print_depend(de);
		    ermsg=sub_tupd_str(ermsg,2,len_str(ermsg));
		    error(mkc_str("need unique selector names for  "),copy_str(dn));
		    message(mkc_str(" duplicate selectors are  "),ermsg);
		    message(mkc_str(" Domain is first defiened in  "),get_root_name(dn));
		    res=FALSE;
		}
	    del_depend(de);
	}
    return res;
}


/*----------------------------------------------------------------------
 
Modulname: unique_elems_check 

 Aufgabe  : Steuerung der Ueberpruefung, ob alle Elemente aller ADT's des 
            Parse-Baums, die Selektoren sind, innerhalb des gesammten 
            Parse-Baums eindeutig sind 

 Eingabe  : Parse-Baum 'dominfo' 

 Ausgabe  : Selektoren sind eindeutig Ja/Nein (TRUE/FALSE) 

----------------------------------------------------------------------*/

Bool
unique_elems_check F1(
		      Dominfo, dominfo)       /* Parse-Baum */
{
    trace1(mkc_str("Checking for unique selector names..."),Empty);

    return nounique_elems_dominfo(dominfo,dominfo);
}

/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------

 Modulname: make_elems 

 Aufgabe  : Umsetzen der Elemente einer Liste in Elemente einer Menge 

 Eingabe  : Liste der Elemente 'elems' 

 Ausgabe  : Menge (Depend) der Elemente 

----------------------------------------------------------------------*/

/* Depend make_elems(Elems e) */
/* Fold-Funktion, definiert in 'domadt' */


/*----------------------------------------------------------------------

 Modulname: make1_str 

 Aufgabe  : Umsetzen eines Elements einer Liste in eine Menge mit dem 
            diesem Element 

 Eingabe  : Element der Liste 's' 

 Ausgabe  : Menge (Depend) mit dem Element 

----------------------------------------------------------------------*/

Depend make1_str F1(
		    Str, e)                 /* Element aus 'Elems' */
{
    return mk1_depend(copy_str(e));
}

/*----------------------------------------------------------------------

 Modulname: is_indomforward_depend 

 Aufgabe  : Ueberpruefung, ob die Elemente einer Liste bereits generiert 
            sind oder als 'Forward' deklariert werden koennen 

 Eingabe  : Liste der zu ueberpruefenden Elemente 'elems' 
            Menge der bereits generierten Elemente 'gen_list' 
            Parse-Baum 'dominfo' 

 Ausgabe  : Elemente sind bereits generiert oder koennen als 'Forward' 
            deklariert werden Ja/Nein (TRUE/FALSE) 

----------------------------------------------------------------------*/

/* Bool is_indomforward_depend(Elems e, Depend gen_list, Dominfo dominfo) */
/* Foldfunktion, definiert in 'domadt' */


/*----------------------------------------------------------------------

 Modulname: 
 Aufgabe  : Ueberpruefung, ob ein Element (ADT) bereits generiert ist 
            oder als 'Forward' deklariert werden kann 

 Eingabe  : ADT 'adt'
            Menge der bereits generierten Elemente 'gen_list' 
            Parse-Baum 'dominfo' 

 Ausgabe  : Element ist bereits generiert oder kann als 'Forward' 
            deklariert werden Ja/Nein (TRUE/FALSE) 

----------------------------------------------------------------------*/


Bool
is_indomforward1_str F3(
			Str, adt,               /* Element aus 'Elems' (ADT-Name) */
			Depend, gen_list,       /* Menge der ADT's fuer die bereits eine
						   Generierungsreihenfolge gefunden wurde */
			Dominfo, dominfo)       /* Parse-Baum */

{
    return
	is_in_depend(adt,gen_list)
	|| is_forward_allowed(adt,dominfo);
}


/*----------------------------------------------------------------------

 Modulname: gen_order2_depend 

 Aufgabe  : Ermittlung, welche Elemente einer Menge von ADT's generiert 
            werden koennen. 
            (mit Beruecksichtigung von Forward-Deklarationen) 

 Eingabe  : Menge der ADT's, die generiert werden sollen 'depend' 
            Parse-Baum 'dominfo' 

 Ausgabe  : mindestens eines der Elemente aus 'depend' kann generiert 
            werden Ja/Nein (TRUE/FALSE) 

----------------------------------------------------------------------*/

/* Bool gen_order2_depend(Depend depend, Dominfo dominfo) */
/* Foldfunktion, definiert in 'domadt' */

/*----------------------------------------------------------------------

 Modulname: gen_order_str 

 Aufgabe  : Ermittlung, ob ein ADT generiert werden kann 

 Eingabe  : Name des ADT 'e' 
            Parse-Baum  'dominfo' 
            Vorwaerts Referenzen erlaubt ' forward'

 Ausgabe  : ADT kann generiert werden Ja/Nein

----------------------------------------------------------------------*/

Bool
gen_order_str F3(
		 Str, e,                /* Element aus 'depend' */
		 Dominfo, dominfo,      /* Parse-Baum */
		 Bool, forward)
{
    Depend m,g;           /* Hilfsvarablen : Umwandlung einer Liste 'Elems' in
			     eine Menge 'Depend' */
    Description desc;     /* Beschreibung des ADT */
    Str constr;           /* Konstruktor des ADT */
    Elems elems;          /* Elemente des ADT */
    Depend depend;        /* Menge der ADT's von denen der ADT 'e' abhaengt */
    Bool res;             /* Funktionsresultat : ADT 'e' kann generiert werden */

    LET desc	= sel_dominfo(dominfo,e) IN
    LET constr	= s_constr(s_domain(desc)) IN
    LET elems	= s_elems(s_domain(desc)) IN
    LET depend	= copy_depend(s_depend(s_domain(desc))) IN

    LET m		= make_elems(gen_order) IN
    LET g		= uniond_depend(get_elems(elems,constr),depend) IN
    LET res	= IF ( forward )
                  THEN is_indomforward_depend(g,m,dominfo)
                  ELSE is_subset_depend(g,m)
    IN
	if ( res )
	    gen_order = appr_elems(gen_order,copy_str(e));

    del_depend(g);
    del_depend(m);

    return res;
}

/*--------------------------------------------------------------------*/

Bool gen_order2_str F2(
		       Str, e,                /* Element aus 'depend' */
		       Dominfo, dominfo)      /* Parse-Baum */
{
    return gen_order_str(e,dominfo,TRUE);
}

Bool gen_order1_str F2(
		       Str, e,
		       Dominfo, dominfo)
{
    return gen_order_str(e,dominfo,FALSE);
}

/*----------------------------------------------------------------------

 Modulname: gen_order1_depend 

 Aufgabe  : Ermittlung, welche Elemente einer Menge von ADT's generiert 
            werden koennen. 
            (ohne Beruecksichtigung von Forward-Deklarationen) 

 Eingabe  : Menge der ADT's, die generiert werden sollen 'depend' 
            Parse-Baum 'dominfo' 

 Ausgabe  : mindestens eines der Elemente aus 'depend' kann generiert 
            werden Ja/Nein (TRUE/FALSE) 

----------------------------------------------------------------------*/

/* Bool gen_order1_depend(Depend depend, Dominfo dominfo) */
/* Foldfunktion, definiert in 'domadt' */


/*----------------------------------------------------------------------

 Modulname: make_cyc_elems 

 Aufgabe  : Erzeugen eines VDM-Strings zur Ausgabe einer Fehlermeldung 
            bei zyklischen Definitionen 

 Eingabe  : Elemente der zyklischen Definition 'elems' 

 Ausgabe  : VDM-String 

----------------------------------------------------------------------*/

/* Str make_cyc_elems(Elems elems) */
/* Fold-Funktion, definiert in 'domadt' */

/*----------------------------------------------------------------------

 Modulname: make_cyc_str 

 Aufgabe  : Erzeugen eines Elements einer Fehlermeldung fuer zyklische 
            Definitionen 

 Eingabe  : Element 'e' 

 Ausgabe  : VDM-String 

----------------------------------------------------------------------*/

Str make_cyc_str F1(
		    Str, e)                 /* Element aus 'elems' */
{
    return
	conc_str(copy_str(e),
		 mkc_str(" --> "));
}


/*----------------------------------------------------------------------

 Modulname: cyclic_dep_depend 

 Aufgabe  : Ermittlung zyklischer Definitionen 

 Eingabe  : Menge mit ADT-Namen, die auf zyklische Definitionen getestet 
            werden sollen 'depend' 
            Parse-Baum 'dominfo' 

 Ausgabe  : zyklische Definitionen gefunden Ja/Nein (TRUE/FALSE) 

----------------------------------------------------------------------*/

/* Bool cyclic_dep_depend(Depend depend, Dominfo dominfo); */
/* Fold-Funktion, definiert in 'domadt' */


/*----------------------------------------------------------------------

 Modulname: cyclic_dep1_str 

 Aufgabe  : Ueberpruefung eines ADT auf zyklische Definition 

 Eingabe  : Name des ADT 'e' 
            Parse-Baum 'dominfo' 

 Ausgabe  : zyklische Definition gefunden Ja/Nein (TRUE/FALSE) 

----------------------------------------------------------------------*/

Bool
cyclic_dep1_str F2(
		   Str, e,                /* Element aus 'depend' */
		   Dominfo, dominfo)      /* Parse-Baum */
{
    Description desc;    /* Beschreibung des ADT */
    Depend m;            /* Elemente Liste 'gen_order' als Menge */
    Bool cyc=FALSE;      /* Funktionsresultat : TRUE -> Zyklus gefunden */

    LET desc=sel_dominfo(dominfo,e) IN

    /* keine zyklische Abhaengigkeit des ADT bei moeglicher */
    /* 'Forward' - Deklaration der Implementation des ADT */
    if ( NOT is_forward_allowed(e,dominfo) )
	{
	    LET m=make_elems(gen_order) IN
	    LET cyc=is_in_depend(e,m) IN
		del_depend(m);

	    /* kein Zyklus */
	    if (NOT cyc)
		{
		    Depend el; /* Elemente des ADT und ADT's von denen er abhaengt */
		    if ((NOT is_empty_elems(s_elems(s_domain(desc))))
			OR (NOT is_empty_depend(s_depend(s_domain(desc)))))
			{
			    LET el=uniond_depend(get_elems(s_elems(s_domain(desc)),
							   s_constr(s_domain(desc))),
						 get_depend(s_domain(desc),e))
			    IN
			    /* anfuegen des aktuellen ADT an die Liste der ueberprueften ADT's */
			    LET gen_order=appr_elems(gen_order,copy_str(e))
			    IN

			    /* Suche nach Zyklus durch die Elemente des ADT */
			    cyclic_dep_depend(el,dominfo);

			    del_depend(el);

			    /* entfernen des ueberprueften ADT aus der Liste */
			    if (len_elems(gen_order)>1)
				gen_order=sub_tupd_elems(gen_order,1,len_elems(gen_order)-1);
			    else
				del_elems(gen_order),gen_order=mk_empty_elems();
			}
		}
	    /* Zyklus gefunden */
	    else
		if (is_eq_str(e,hd_elems(gen_order)))
		    {
			Str errmsg=make_cyc_elems(gen_order);
			errmsg=conc_str(errmsg,copy_str(e));
			message(mkc_str(" Cyclic dependency detected :  "),errmsg);
		    }
	}
    return cyc;
}


/*----------------------------------------------------------------------

 Modulname: trans_closure

 Aufgabe  : Berechnung der transitiven Huelle

 Eingabe  : ein Abhaengingkeitsgraph

 Ausgabe  : die transitive Huelle

----------------------------------------------------------------------*/

Dependgraph
trans_closure F1(Dependgraph,dg)
{
    Bool   cont   = TRUE;

    Depend domdg0 = dom_dependgraph(dg);
    Depend domdg  = copy_depend(domdg0);
    del1_depend(domdg0);
    /* now domdg and dg don't share any data */
    
    while ( cont )
	{
	    /* loop over all domain names */

	    Depend dom = copy1_depend(domdg);

	    cont = FALSE;
    
	    while ( ! is_empty_depend(dom) )
		{
		    Str    id   = sel_depend(dom);

		    /* loop over all successors of id */

		    /* ids doesn't share any data with dg */
		    Depend ids  = copy_depend(sel_dependgraph(dg,id));

		    while ( ! is_empty_depend(ids) )
			{
			    Str    id1   = sel_depend(ids);
			    Depend succ1 = sel_dependgraph(dg,id1);

			    /* don't move to outer loop */
			    Depend succ  = sel_dependgraph(dg,id);

			    if ( ! is_subset_depend(succ1,succ) )
				{
				    cont = TRUE;
				    dg = ovwrt1d_dependgraph(dg,
							     copy_str(id),
							     uniond_depend(copy_depend(succ),
									   copy_depend(succ1)));
				}
			    ids = sub1d_depend(ids,id1);
			}
		    del_depend(ids);

		    dom = sub1_depend(dom,id);
		}
	    del1_depend(dom);
	}

    del_depend(domdg);

    return dg;
}

Dependgraph
remove_cycles F2(Dependgraph,dg,
		 Depend,fwal)
{
    Dependgraph trcl_dg = trans_closure(copy_dependgraph(dg));

    Depend ids = dom_dependgraph(trcl_dg);

    if (domtrace>1)
	{
	    puts_str(mkc_str("\n\nTransitive Closure of dependency graph :\n"),stdout);
	    pr_dependgraph(trcl_dg);
	}

    while ( ! is_empty_depend(ids) )
	{
	    Str id = sel_depend(ids);

	    if ( is_in_depend(id,fwal)
		 && is_in_depend(id,sel_dependgraph(trcl_dg,id)) )
		{
		    /* domain is in forward allowed set and is cyclic */
		    /* remove all direct successors on cyclic dependencies */ 
		    Depend succ = copy_depend(sel_dependgraph(trcl_dg,id));
		    while ( ! is_empty_depend(succ) )
			{
			    Str id1 = sel_depend(succ);

			    if ( is_in_depend(id,sel_dependgraph(dg,id1)) )
				{
				    /* remove id1 from dependent set for id */
				    Depend succ = copy_depend(sel_dependgraph(dg,id1));

				    dg = ovwrt1d_dependgraph(dg,
							     copy_str(id1),
							     sub1d_depend(succ,id));
				}
			    succ = sub1d_depend(succ,id1);
			}
		    del_depend(succ);
		}
	    ids = sub1_depend(ids,id);
	}

    del1_depend(ids);

    del_dependgraph(trcl_dg);

    if (domtrace>1)
	{
	    puts_str(mkc_str("\n\nCyclic dependencies removed from dependency graph :\n"),stdout);
	    pr_dependgraph(dg);
	}

    return dg;
}

/*----------------------------------------------------------------------

 Modulname: get_gen_order 

 Aufgabe  : Ermittlung einer moeglichen Generierungsreihenfolge fuer die 
            ADT's eines Parse-Baums 

 Strategie:
         1. es wird der vollstndige Abhngingkeitsgraph und die
            Menge der forward allowed domains berechnet
	 2. es wird die transitive Hlle des Abhngigkeitsgraphen
	    berechnet
	 3. aus diesem werde fr die forward allowed domains,
	    die transitiv von sich selbst abhngen, die Abhngigkeiten
	    gelscht, die auf Zyklen liegen, nur hier geht die
	    forward allowed info ein.
	 4. aus diesem, hoffentlich zyklenfreien, Graphen werden
	    successive die Knoten selektiert, deren abhngige domains
	    schon verarbeitet sind
	 5. der resultierende Graph mu eine partielle Ordnung festlegen
	    wenn immer noch Zyklen enthalten sind, gibts eine Fehlermeldung

 Eingabe  : Parse-Baum 'dominfo' 

 Ausgabe  : Generierungsreihenfolge gefunden Ja/Nein (TRUE/FALSE) 
            und globale variable gen_order mit der Reihenfolge

----------------------------------------------------------------------*/

Bool
get_gen_order F1(Dominfo, dominfo)
{
    Depend      dids1 = dom_dominfo(dominfo);
    Depend      dids  = copy1_depend(dids1);
    Dependgraph dg    = mk_empty_dependgraph();
    Depend      fwal  = mk_empty_depend();
    Bool        res   = TRUE;

    trace1(mkc_str("Building dependeny graph ..."),Empty);

    {
	/* compute dependency graph and forward allowed set */
	while ( ! is_empty_depend(dids) )
	    {
		Str did = sel_depend(dids);

		Description descr    = sel_dominfo(dominfo,did);
		Domain      dm       = s_domain(descr);

		Depend      dep_dids = elems_elems(s_elems(dm));

		/* filter selector names */
		dep_dids = intersd_depend(dep_dids,copy_depend(dids1));

		/* add dependant domains */
		dep_dids = uniond_depend(dep_dids,copy_depend(s_depend(dm)));

		dg = union1_dependgraph(dg,
					copy_str(did),
					dep_dids);

		if ( is_forward_allowed(did,dominfo) )
		    fwal = union1d_depend(fwal,
					  copy_str(did));
		
		dids = sub1_depend(dids,did);
	    }

	del1_depend(dids);
    }

    if (domtrace>1)
	{
	    puts_str(mkc_str("\n\nDependency graph :\n"),stdout);
	    pr_dependgraph(dg);
	    puts_str(mkc_str("\n\nForward allowed set :\n"),stdout);
	    pr_depend(fwal);
	}

    dg = remove_cycles(dg,fwal);

    trace1(mkc_str("\n\nCalculating generation order..."),Empty);

    {
	/* compute generation order */
	Bool   cycle     = FALSE;

	Elems  order     = mk_empty_elems();

	Depend dids_left      = copy1_depend(dids1);
        Depend dids_processed = mk_empty_depend();

	while ( ! is_empty_depend(dids_left)
		AND NOT cycle )
	    {
		trace1(mkc_str("step ..."),Empty);

		cycle     = TRUE;
		dids      = copy1_depend(dids_left);

		while ( ! is_empty_depend(dids) )
		    {
			Str did = sel_depend(dids);
		    
			if ( is_subset_depend(sel_dependgraph(dg,did),dids_processed) )
			    {
				cycle          = FALSE;
				order          = appr_elems(order,copy_str(did));
				dids_left      = sub1_depend(dids_left,did);
				dids_processed = union1_depend(dids_processed,did);
			    }
		    
			dids = sub1_depend(dids,did);
		    }
		del1_depend(dids);
	    }

	/* forward allowed types are put at the end */
	gen_order = order;

	res = ! cycle;

	if ( cycle )
	    {
		/* error message cyclic dependencies in domains */

		Str ermsg=make_print_depend(dids_left);
		ermsg=sub_tupd_str(ermsg,2,len_str(ermsg));
		error(mkc_str("No generation order for  "),ermsg);

		del_elems(gen_order);
		gen_order=mk_empty_elems();

		/* Suchen zyklischer Definitionen */
		cyclic_dep_depend(dids_left,dominfo);
		del_elems(gen_order);
	    }

	del1_depend(dids_left);
    }

    if (domtrace>1)
	{
	    puts_str(mkc_str("\n\nGeneration order :\n"),stdout);
	    pr_elems(gen_order);
	}

    del_dependgraph(dg);
    del_depend(fwal);
    del1_depend(dids1);

    /* preliminary */
    return TRUE;
}

/* -------------------------------------------------------------------------- */
/* dies und ein paar aufgerufene Routinen sind alt (hoffentlich) */

Bool
old_get_gen_order F1(
		 Dominfo, dominfo)      /* Parse-Baum */
{
    Depend ngo;           /* ADT's fuer die (noch) keine Generierungsfolge
			     gefunden wurde */
    Bool res=TRUE;        /* Funktionsresultat : FALSE -> nicht fuer alle ADT's 
			     wurde eine Generierungsreihenfolge gefunden */

    trace1(mkc_str("Calculating generation order..."),Empty);

    gen_order = mk_empty_elems();

    /* Initialisierung : noch keine Reihenfolge fuer ein Element ermittelt */
    {
	Depend ngo1 = dom_dominfo(dominfo);

	ngo = copy_depend(ngo1);
	del1_depend(ngo1);
    }

    /* Ermittlung der Reihenfolge ohne Beruecksichtigung der moeglichen */
    /* Forward - Deklaration der Implementation eines ADT --> elementare */
    /* Datentypen werden zuerst generiert */

    while ( gen_order1_depend(ngo,dominfo) )
	ngo = diffd_depend(ngo,
			   make_elems(gen_order));

    /* noch keine Reihenfolge fuer alle Elemente gefunden */
    /* --> Ermittlung der Reihenfolge mit Beruecksichtigung der moeglichen */
    /* Forward - Deklaration der Implementation eines ADT */

    if ( len_elems(gen_order) < card_dominfo(dominfo))
	{
	    while ( gen_order2_depend(ngo,dominfo) )
		ngo = diffd_depend(ngo,
				   make_elems(gen_order));
	}

    /* keine Reihenfolge fuer alle ADT's gefunden */
    if (NOT is_empty_depend(ngo))
	{
	    Str ermsg=make_print_depend(ngo);
	    ermsg=sub_tupd_str(ermsg,2,len_str(ermsg));
	    error(mkc_str("No generation order for  "),ermsg);
	    del_elems(gen_order),gen_order=mk_empty_elems();

	    /* Suchen zyklischer Definitionen */
	    cyclic_dep_depend(ngo,dominfo);
	    del_elems(gen_order);
	    del_depend(ngo);
	    res=FALSE;
	}
    else
	if (domtrace>1)
	    {
		puts_str(mkc_str("\n\nGeneration order :\n"),stdout);
		pr_elems(gen_order);
	    }

    return res;
}

/*----------------------------------------------------------------------

 Modulname: casesen_domid 

 Aufgabe  : Pruefe, ob alle domain namen verschieden sind nach
            fortlassung von Gross/Kleinchreibung 

 Eingabe  : Parse-Baum 'dominfo' 

 Ausgabe  : OK (TRUE/FALSE) 

----------------------------------------------------------------------*/

static Str curr_dnlc;
static Str curr_dn;
static Depend checked_di;

static
Bool 
is_eq_lc F2(Str,dn,
	    Description,d)
{
    Str dnlc = lowercase_str(dn);

    if ( !is_eq_str(dn,curr_dn)
	 && is_eq_str(dnlc,curr_dnlc)
	 )
	{
	    Str em = conc3_str(copy_str(dn),
			       mkc_str(", "),
			       copy_str(curr_dn));
	    error(mkc_str("Domain names differ in cases only: "),em);
	    del_str(em);
	    return TRUE;
	}
    del_str(dnlc);

    return FALSE;
}

Bool
casesen_domid_str_description F3(Str, dn,
				 Description, d,
				 Dominfo, di)
{
    Str dnlc = lowercase_str(dn);
    Bool result;

    curr_dn = dn;
    curr_dnlc = dnlc;

    result =
	(is_in_depend(dnlc,checked_di))
	? TRUE
	: !exist_dominfo(di,is_eq_lc);

    checked_di = union1_depend(checked_di,dnlc);

    return result;
}

Bool
casesen_domid F1(
		 Dominfo, dominfo)            /* Parse-Baum */
{
    Bool result;

    trace1(mkc_str("Checking domain names..."),Empty);

    checked_di = mk_empty_depend();
    result = casesen_domid_dominfo(dominfo,dominfo);

    del_depend(checked_di);

    return result;
}
  

/*----------------------------------------------------------------------

 Modulname: semantic_check 

 Aufgabe  : Ueberpruefung eines Parse-Baums auf semantische Fehler 

 Eingabe  : Parse-Baum 'dominfo' 

 Ausgabe  : Parse-Baum fehlerfrei Ja/Nein (TRUE/FALSE) 

----------------------------------------------------------------------*/

Bool
semantic_check F1(
		  Dominfo, dominfo)            /* Parse-Baum */
{
    Bool res=TRUE;              /* Funktionsresultat */

    /* Test auf ADT-Definition und Implementation */
    res&=domain_check(dominfo);

    /* Test auf fehlende Definitionen fuer die Elemente der ADT's */
    res&=complete_domains_check(dominfo);

    /* Test auf Eindeutigkeit der Selektornamen von ADT's */
    res&=unique_elems_check(dominfo);

    /* Test, ob Konflikte durch Gross/Kleinschreibung auftreten */
    res&=casesen_domid(dominfo);
    if (res)
	{
	    res&=get_gen_order(dominfo);
	    del_elems(gen_order);
	}
    
    return res;
}
