
/* $Id: domcomment.c,v 1.2 1996/01/15 10:01:49 vdm Exp $

  Comments in vdm files

  $Log: domcomment.c,v $
  Revision 1.2  1996/01/15 10:01:49  vdm
  cvs backup

 * Revision 1.1  1994/12/29  16:52:29  vdm
 * new option -X for collecting comments
 * during parsing and writing these comments into the .enr and .pre files
 * for parsing and importing .vdm files into vdmedit
 *

*/

#include "domadt.h"
#include "domfct.h"
#include "domcomment.h"

int collectcomments = 0;
static Comments comments;         /* map for collecting comments */
static Str currcommentkey;      /* current comment key: general comment, */
                                /*   domain specific, implementation spec. */

static Str emptystr;

static char * ini_prefix = "Ini.";
static char * dom_prefix = "Dom.";
static char * imp_prefix = "Imp.";
static char * fct_prefix = "Fct.";

void init_comment F0() {
  collectcomments = 0;
  emptystr = mk_empty_str();
  comments = mk_empty_comments();
  currcommentkey = mkc_str(ini_prefix);
}

void set_collect_comment F0() {
  collectcomments = 1;
}

void reset_collect_comment F0() {
  collectcomments = 0;
}

static void set_comment_key F1(Str, key) {
  if ( collectcomments ) {
    del_str(currcommentkey);
    currcommentkey = copy_str(key);
  }
}

void domain_comment F0() {
  if ( collectcomments ) {
    set_comment_key(mkc_str(dom_prefix));
  }
}

void impl_comment F0() {
  if ( collectcomments ) {
    set_comment_key(mkc_str(imp_prefix));
  }
}

void fct_comment F0() {
  if ( collectcomments ) {
    set_comment_key(mkc_str(fct_prefix));
  }
}

void specdom_comment F1(Str,dom) {
  if ( collectcomments ) {
    set_comment_key(conc_str(mkc_str(dom_prefix),copy_str(dom)));
  }
}

void specimp_comment F1(Str,dom) {
  if ( collectcomments ) {
    set_comment_key(conc_str(mkc_str(imp_prefix),copy_str(dom)));
  }
}

void specfct_comment F2(Str,name,Str,param) {
  if ( collectcomments ) {
    set_comment_key(conc_str(conc_str(mkc_str(fct_prefix),copy_str(name))
                            ,conc_str(mkc_str("."),copy_str(param))));
  }
}

void insert_comment F1(char *,l) {
				/* a line of comment is inserted */
				/* if comment already exist, line is appended */
				/* line is not copied !!! */
  if ( collectcomments ) {
     Str currcom;
     Str line;
     line = conc_str(mkc_str(l),mkc_str("\n"));

     if ( is_in_dom_comments(currcommentkey,comments,&currcom) ) {
       line = conc_str(copy_str(currcom),line);
       comments = sub1d_comments(comments,currcommentkey);
     }
     comments = union1_comments(comments,copy_str(currcommentkey),line);
  }
}

static Str give_comment F1(Str,key) {
  Str comment;
  if ( ! is_in_dom_comments(key,comments,&comment) ) {
    comment = emptystr;
  }
  del_str(key);

  return copy_str(comment);
}

Str give_init_comment   F0() {
  return collectcomments
    ? give_comment(mkc_str(ini_prefix))
    : mk_empty_str();
}

Str give_domain_comment   F0() {
  return collectcomments
    ? give_comment(mkc_str(dom_prefix))
    : mk_empty_str();
}

Str give_impl_comment   F0() {
  return collectcomments
    ? give_comment(mkc_str(imp_prefix))
    : mk_empty_str();
}

Str give_fct_comment   F0() {
  return collectcomments
    ? give_comment(mkc_str(fct_prefix))
    : mk_empty_str();
}

Str give_specdom_comment F1(Str,name) {
  return collectcomments
    ? give_comment(conc_str(mkc_str(dom_prefix),copy_str(name)))
    : mk_empty_str();
}

Str give_specimp_comment F1(Str,name) {
  return collectcomments
    ? give_comment(conc_str(mkc_str(imp_prefix),copy_str(name)))
    : mk_empty_str();
}

Str give_specfct_comment F2(Str,name,Str,param) {
  return collectcomments
    ? give_comment(conc_str(conc_str(mkc_str(fct_prefix),copy_str(name))
                           ,conc_str(mkc_str("."),copy_str(param))))
    : mk_empty_str();
}
