/*----------------------------------------------------------------------

 $Id: domcpp.c,v 1.4 1996/01/15 10:01:52 vdm Exp $

 Aufgabe  : Steuerung des Aufrufs des Praeprozessors 

----------------------------------------------------------------------
 Parameter 

 Eingabe  : zu verarbeitender VDM-String 

 Ausgabe  : VDM-String 

----------------------------------------------------------------------

 Ersteller: Kai Steinau 

----------------------------------------------------------------------

 Aenderungen: 

 $Log: domcpp.c,v $
 Revision 1.4  1996/01/15 10:01:52  vdm
 cvs backup

 * Revision 1.3  1994/01/30  17:16:55  vdm
 * -D defines only passed to 1. call of preprocessor
 * and not to internal calls
 * there syntax of vdm files can be destroied
 *
 * Revision 1.2  1993/11/03  14:16:43  vdm
 * Hans-Martins Aenderungen eingemischt
 *
 * Revision 1.2  1993/10/24  21:58:03  hmh
 * Changes built in to handle the '-r' option used to check
 * if a given enrichment file can be used as starting point of
 * enrichment process.
 *
 * - trace messages generated using 'trace*'.

----------------------------------------------------------------------*/

#include "domadt.h"
#include "domfct.h"
#include "domcpp.h"

/*----------------------------------------------------------------------*/


/*----------------------------------------------------------------------*/

/* Modulname: count_params */

/* Aufgabe  : Ermitteln der Anzahl der zu einem Parameter gegebenen Werte */
/*            (noetig um fuer die Parameter des Praeprozessor-Aufrufs */
/*             eine entsprechende Anzahl von dynamisch belegten Speicher- */
/*             bereichen zu erzeugen) */

/*----------------------------------------------------------------------*/
/* Parameter */

/* Eingabe  : Paramter 'param' */

/* Ausgabe  : Anzahl der zu 'param' gegebenen Werte */

/*----------------------------------------------------------------------*/

Nat0 count_params F1(
     Str, param)
{
 Str pv,v;
 Nat0 p,n=0;
 if (is_in_dom_params(param,global_params,&pv))
 {
    v=copy_str(pv);
    while (NOT is_empty_str(v))  /* Parameter-Wert vorhanden */
    {
       n++;
       p=index_str(v,'\n');
       if (NOT p) p=len_str(v);
       v=sub_tupd_str(v,p+1,len_str(v));
    }
   del_str(v);
 }
 return n;
}


/*----------------------------------------------------------------------*/

/* Modulname: cppvdm */

/* Aufgabe  : Aufruf des Praeprozessors */

/*----------------------------------------------------------------------*/
/* Parameter */

/* Eingabe  : Zu verarbeitender VDM-String 's' */
/*            Name der zu verarbeitenden Komponente 'comp' */
/*            (bei 'comp'=VDM werden '#line'-Informationen erzeugt) */

/* Ausgabe  : durch Praeprozessor verarbeiteter VDM-String */

/*----------------------------------------------------------------------*/

Str cppvdm F3(
Str, s,
Str, comp,
Bool, includedefs)
{
 int cppmain();                 /* Deklartation Praeprozessor-Aufruf */
 int arga;                      /* Anzahl der Argumente abhaengig von */
                                /* Aufruf mit oder ohne '#line'-Information */
 int argc=0;                    /* Gesammt-Anzahl der Argumente bei Aufruf */
 int argcbak;                   /* Kopie von argc */
 char **argv;                   /* Zeiger auf Argumenten-Liste */
 char **argvbak;                /* Kopie von argv */
 Str fi=create_tmpfile();       /* Name fuer Eingabedatei */
 Str fo=create_tmpfile();       /* Name fuer Ausgabedatei */
 Str pv,cpv,ppv,p=mk_empty_str();       /* Hilfsvariablen */
 Nat0 i,n;                              /* Hilfsvariablen */

 trace1(mkc_str("Preprocessing..."),mk_empty_str());
                   /* Anzahl Argumente: Praeprozessor-Name, Eingabe- und Ausgabe-Datei */
               /* Bei nicht VDM zusaetzlich Parameter um '#line'-Info zu unterdruecken */
 arga=IF (is_eq_str(comp,VDM)) THEN 3 ELSE 4;
                               /* Anzahl um Parameter-Angaben fuer -I und -D erweitern */
 n=arga+count_params(PARAM_['I'])
   + (IF includedefs THEN count_params(PARAM_['D']) ELSE 0);
                                                     /* Speicher fuer Argumenten-Liste */
 argv=(char **)tmpmalloc(n*sizeof(char *));
 argvbak=(char **)tmpmalloc(n*sizeof(char *));
 if ((argv!=NULL) AND (argvbak!=NULL))
 if ((NOT is_empty_str(fi)) AND (NOT is_empty_str(fo)))
 if (fwrite_str(s,fi)==FALSE)
    error(mkc_str("Can't write source for cpp into tmpfile "),copy_str(fi));
 else
 {
                                                                 /* Praeprozessor-Name */
    if ((argv[argc]=(char *)tmpmalloc(len_str(CPP_NAME)+1))!=NULL)
       cpc_str(CPP_NAME,argv[argc++]);
                                            /* bei nicht VDM keine '#line'-Information */
    if ((NOT is_eq_str(comp,VDM)) AND (domtrace<3))
    if ((argv[argc]=(char *)tmpmalloc(len_str(CPP_NODEBUG)+1))!=NULL)
       cpc_str(CPP_NODEBUG,argv[argc++]);
                                                           /* global definierte Makros */
    if ( includedefs && is_in_dom_params(PARAM_['D'],global_params,&pv))
    {
       cpv=copy_str(pv);
       while (NOT is_empty_str(cpv))
       {
          i=index_str(cpv,'\n');
          if (NOT i) i=len_str(cpv)+1;
          ppv=mk_sub_tup_str(cpv,1,i-1);
          p=conc_str(copy_str(PARAM_['D']),ppv);
          cpv=sub_tupd_str(cpv,i+1,len_str(cpv));
          if ((argv[argc]=(char *)tmpmalloc(len_str(p)+1))!=NULL)
             cpc_str(p,argv[argc++]);
          del_str(p);
       }
       del_str(cpv);
    }
                                                              /* Include-Verzeichnisse */
    if (is_in_dom_params(PARAM_['I'],global_params,&pv))
    {
       cpv=copy_str(pv);
       while (NOT is_empty_str(cpv))
       {
          i=index_str(cpv,'\n');
          if (NOT i) i=len_str(cpv)+1;
          ppv=mk_sub_tup_str(cpv,1,i-1);
          p=conc_str(copy_str(PARAM_['I']),ppv);
          cpv=sub_tupd_str(cpv,i+1,len_str(cpv));
          if ((argv[argc]=(char *)tmpmalloc(len_str(p)+1))!=NULL)
             cpc_str(p,argv[argc++]);
          del_str(p);
       }
       del_str(cpv);
    }
                                                             /* Name der Ausgabe-Datei */
    p=conc_str(copy_str(PARAM_['o']),copy_str(fo));
    if ((argv[argc]=(char *)tmpmalloc(len_str(p)+1))!=NULL)
       cpc_str(p,argv[argc++]);
    del_str(p);
                                                             /* Name der Eingabe-Datei */
    if ((argv[argc]=(char *)tmpmalloc(len_str(fi)+1))!=NULL)
       cpc_str(fi,argv[argc++]);
                        /* Sichern der Laenge und Adressen der Parameter-Liste --> der */
                                               /* Praeprozessor veraendert diese Werte */
    argcbak=argc;
    for (;argcbak--;) {
       argvbak[argcbak]=argv[argcbak];
    }
    argcbak=argc;
    del_str(s);
                                                               /* Praeprozessor-Aufruf */
    cppmain(argc,argv);
    s=fread_str(fo);
                                                   /* freigeben des belegten Speichers */
    for (;argcbak--;)
      if (argvbak[argcbak]!=NULL) free(argvbak[argcbak]);
 }
 if ((NOT is_empty_str(fi)) AND (domtrace<3)) remove_file(fi);
 del_str(fi);
 if ((NOT is_empty_str(fo)) AND (domtrace<3)) remove_file(fo);
 del_str(fo);
 return s;
}
