/*----------------------------------------------------------------------*/

/* $Id: domdb.c,v 1.11 1996/09/08 11:45:11 vdm Exp $ */

/* Aufgabe  : Durchfuehrung saemtlicher Datenbankzugriffe */

/*----------------------------------------------------------------------*/
/* Parameter */

/* Eingabe  : */

/* Ausgabe  : */

/*----------------------------------------------------------------------*/

/* Ersteller: Kai Steinau */

/*----------------------------------------------------------------------

 Aenderungen: 
 
 $Log: domdb.c,v $
 Revision 1.11  1996/09/08 11:45:11  vdm
 error in param_defined removed:
 Implid was not checked against NIL, before accessing the database
 this case occurs for PREDEF domain ids
 "s_str_implid(...)" was not called in parameter expression for "get_params"
 "Str" and "Implid" are in C the same type, so no compiler error was
 issued ( :-( ).

 Revision 1.10  1996/09/07 14:12:04  vdm
 vdm as client

 Revision 1.9  1996/04/30 20:24:26  vdm
 extensions for implementing alias as inheritance in C++
 solution: a new constructor -isa with external syntax
 Subtype is-a Basetype
 Semantics for C like "=", but for C++ class SUBTYPE
 becomes a subclass of class BASETYPE
 Problem: constructors are not inherited, so the
 constructors are passed from Basetype to Subtype
 in the generator, therefore a special operation
 constructor.mk??? introduced, wich is copied in the generator
 to the Subtype domain.
 The generation order computation did not work save,
 therefore an explicit dependency graph is constructed, with this
 graph the generation order is computed, old code still there
 function "old_gen_order"

 Revision 1.8  1996/01/15 10:01:53  vdm
 cvs backup

 * Revision 1.7  1994/12/29  16:52:30  vdm
 * new option -X for collecting comments
 * during parsing and writing these comments into the .enr and .pre files
 * for parsing and importing .vdm files into vdmedit
 *
 * Revision 1.6  1993/11/28  09:54:30  vdm
 * Options sortiert
 *
 * Revision 1.5  1993/11/21  20:03:48  vdm
 * Torsten Tews 1.Version eingemischt
 * Dabei interne Domain Gleichungen in C++ ausgeblendet
 * mit einem internen Implementierungsparameter INTERALDOMAIN
 *
 * Revision 1.4  1993/11/03  14:16:49  vdm
 * Hans-Martins Aenderungen eingemischt
 *
 * Revision 1.3  1993/07/05  21:33:57  hmh
 * Merging Uwes changes into mine.
 *
 * Revision 1.2  1993/05/28  11:03:20  hmh
 * Compiler extended to accept the ISO VDM-SL ASCII syntax. These
 * extensions are activated through the '-N' switch.
 *
 * In 'tolowerunder' the whole string is converted to lower case
 * instead of the first character only. 'lowercase_str' is used for
 * this (needed since domain names may contain upper case characters
 * now).

   93/02/11: System-Includes (#include <...>) nach domfct.h 
             Include-Direktive fuer dbs.h ohne Pfadname 

----------------------------------------------------------------------*/

#include "domadt.h"
#include "domfct.h"
#include "dbs.h"
#include "domdb.h"

/*----------------------------------------------------------------------*/


/* Variable zur Definition der verwendeten Datenbanken */
/* (Vdmdbs definiert in 'dbs.h') */

Vdmdbs vdmdbs;

/*----------------------------------------------------------------------*/

/* Modulname: tolowerunder */

/* Aufgabe  : Umwandeln des ersten Buchstabens eines VDM-Strings in einen */
/*            Kleinbuchstaben und voranstellen eines Unterstrichs */

/*----------------------------------------------------------------------*/
/* Parameter */

/* Eingabe  : VDM-String 'v' */

/* Ausgabe  : VDM-String mit Unterstrich in Kleinschreibweise */

/*----------------------------------------------------------------------*/

Str tolowerunder F1(
		    Str, v)
{
    return prefix_underline_str(lowercase_str(v));
}


/*----------------------------------------------------------------------*/

/* Modulname: define2 */

/* Aufgabe  : Erzeugen von zwei '#define'-Anweisungen */

/*----------------------------------------------------------------------*/
/* Parameter */

/* Eingabe  : VDM-String, in dem die '#define'-Anweisungen erzeugt werden */
/*            sollen 's' */
/*            Name der zu definierenden 'Variablen' 'p' */
/*            Wert der 'Variablen' 'v' */

/* Ausgabe  : Um '#define'-Anweisungen ergaenzter VDM-String */

/*----------------------------------------------------------------------*/

Str define2 F3(
	       Str, s,
	       Str, p,
	       Str, v)
{
    Str cpp;
    app_str(s,define_str(copy_str(p),copy_str(v)));
    cpp=mk_sub_tup_str(p,1,len_str(CPP_PARAM));
    if (NOT is_eq_str(cpp,CPP_PARAM))
	app_str(s,define_str(prefix_underline_str(copy_str(p)),tolowerunder(v)));
    del_str(cpp);
    return s;
}


/*----------------------------------------------------------------------*/

/* Modulname: define */

/* Aufgabe  : Erzeugen einer Menge von '#define'-Anweisungen */

/*----------------------------------------------------------------------*/
/* Parameter */

/* Eingabe  : VDM-String, in dem die '#define'-Anweisungen erstellt werden */
/*            sollen 's' */
/*            Abbildung mit Paramter-Name (=zu definierende Variable) auf */
/*            Parameter-Wert (=Wert der Variablen) 'params' */

/* Ausgabe  : Um '#define'-Anweisungen ergaenzter VDM-String */

/*----------------------------------------------------------------------*/

Str define F2(
     Str, s,                  /* VDM-String, in dem die Parameter 'params' */
                              /* mit '#define' definiert werden */
     Params, params)          /* Parameter */
{
    /* Die Abbildung (Map) der Parameter ist nicht leer */
    if (NOT(is_empty_params(params)))
	{
	    Str p;  /* Parameter */
	    Str v;  /* Wert des Parameters */

                              /* Auswahl eines beliebigen Parameters */
	    LET p=sel_dom_params(params) IN
                                /* Wert des ausgewaehlten Parameters */
	    LET v=sel_params(params,p) IN

                           /* Der Wert des Parameters ist nicht leer */
	    s = IF (NOT(is_empty_str(v)))
		THEN
		/* Der Parameter wird definiert */
		    define2(s,p,v)
                                 /* Der Wert des Parameters ist leer */
		ELSE
              /* Eine Zeile '#undef' fuer den Parameter wird erzeugt */
		    undefine(s,mk1_params(copy_str(p),copy_str(v)));
                      /* Die restlichen Parameter werden verarbeitet */
	    s=define(s,sub1d_params(params,p));
	}
    return s;
}


/*----------------------------------------------------------------------*/

/* Modulname: undefine2 */

/* Aufgabe  : Erzeugen von zwei '#undef'-Anweisungen */

/*----------------------------------------------------------------------*/
/* Parameter */

/* Eingabe  : VDM-String, in dem die '#undef'-Anweisungen erzeugt werden */
/*            sollen 's' */
/*            Name der 'Variablen' 'p' */

/* Ausgabe  : Um '#undef'-Anweisungen ergaenzter VDM-String */

/*----------------------------------------------------------------------*/

Str undefine2 F2(
		 Str, s,
		 Str, p)
{
    Str cpp;
    app_str(s,undef_str(copy_str(p)));
    cpp=mk_sub_tup_str(p,1,len_str(CPP_PARAM));
    if (NOT is_eq_str(cpp,CPP_PARAM))
	app_str(s,undef_str(prefix_underline_str(copy_str(p))));
    del_str(cpp);
    return s;
}


/*----------------------------------------------------------------------*/

/* Modulname: undefine */

/* Aufgabe  : Erzeugen einer Menge von '#undef'-Anweisungen */

/*----------------------------------------------------------------------*/
/* Parameter */

/* Eingabe  : VDM-String, in dem die '#undef'-Anweisungen erstellt werden */
/*            sollen 's' */
/*            Abbildung mit Paramter-Name (=zu definierende Variable) auf */
/*            Parameter-Wert (=Wert der Variablen) 'params' */

/* Ausgabe  : Um '#undef'-Anweisungen ergaenzter VDM-String */

/*----------------------------------------------------------------------*/

Str undefine F2(
     Str, s,                  /* VDM-String, in dem die Parameter 'params' */
                              /* mit '#undef' als nicht definiert gekennzeichnet */
                              /* werden */
     Params, params)          /* Parameter */

{
                 /* Die Abbildung (Map) der Parameter ist nicht leer */
    if (NOT(is_empty_params(params)))
	{
                              /* Auswahl eines beliebigen Parameters */
	    Str p = sel_dom_params(params);
		/* Eine Zeile '#undef' fuer den Parameter wird erzeugt */
	    s=undefine2(s,p);
                      /* Die restlichen Parameter werden verarbeitet */
	    s=undefine(s,sub1d_params(params,p));
	}
    return s;
}


/*----------------------------------------------------------------------*/

/* Modulname: include */

/* Aufgabe  : Einlesen einer Datenbank-Komponente oder einer DSL- */
/*            Spezifikation des Anwenders */

/*----------------------------------------------------------------------*/
/* Parameter */

/* Eingabe  : VDM-String, an den die gelesene Komponente angehaengt werden */
/*            soll 's' */
/*            Name des ADT, der das Anhaengen der Komponente "verursacht" */
/*            'domainid' */
/*            Implementation des ADT 'implid' */
/*            Operation 'operationid' */
/*            Komponente 'comp' */

/* Ausgabe  : Um gelesene Komponente erweiterter String 's' */

/*----------------------------------------------------------------------*/

Str include F6(
     Str, s,              /* Liste an die der der Komponente 'comp' entsprechende*/
                          /* VDM-String anzuhaengen ist */
     Str, domainid,       /* Name des ADT, der das Anhaengen der Komponente */
                          /* 'verursacht' */
     Str, constr,         /* Konstruktor (z.B. tree) des ADT */
     Str, implid,         /* Implementation (z.B. linklist) des ADT */
     Str, operationid,    /* Operation (z.B. union') */
     Str, comp)           /* Komponente (z.B. 'vdm') */
{
    Str c;

    LET c=IF (is_empty_str(constr))
	THEN
	/* Bei leerem Konstruktor ist in 'domainid' der Name einer vom */
	/* Anwender definierten DSL-Spezifikation enthalten */
	   fread_str(domainid)
	ELSE
	/* Die VDM-Datenbank Komponente wird gelesen */
	IF (NOT is_in_components(comp,Sign_components))
	THEN
	    get_component(constr,implid,operationid,comp)
	ELSE
	    get_signature(constr,operationid,comp)
	IN return conc_str(s,c);
}

/*----------------------------------------------------------------------*/

/* Modulname: get_component */

/* Aufgabe  : Lesen einer Datenbank-Komponente mit der vom VDM-DB Manager */
/*            bereitgestellten Funktion */

/*----------------------------------------------------------------------*/
/* Parameter */

/* Eingabe  : Konstruktor 'constr', Implementation 'implid', Operation */
/*            'operationid' und Komponente 'comp' */

/* Ausgabe  : In der Datenbank gefundene Komponente oder leerer String */

/*----------------------------------------------------------------------*/

Str get_component F4(
		     Str, constr,
		     Str, implid,
		     Str, operationid,
		     Str, comp)
{
    return getsx_Component(vdmdbs,constr,implid,operationid,comp);
}

/*----------------------------------------------------------------------*/

Str
get_adttype F1(
	       Str, constr)
{
    return getsx_Component(vdmdbs,constr,COM,BASIC,ADTTYPE);
}

/*----------------------------------------------------------------------*/

/* Modulname: get_signature */

/* Aufgabe  : Lesen einer Signatur-Komponente aus der Datenbank mit der vom */
/*            VDM-DB Manager bereitgestellten Funktion */

/*----------------------------------------------------------------------*/
/* Parameter */

/* Eingabe  : Konstruktor 'constr', Operation 'operationid' und Komponente */
/*            'comp' */

/* Ausgabe  : In der Datenbank gefundene Komponente oder leerer String */

/*----------------------------------------------------------------------*/

Str get_signature F3(
		     Str, constr,
		     Str, operationid,
		     Str, comp)
{
    return getsx_SignComponent(vdmdbs,constr,operationid,comp);
}


/*----------------------------------------------------------------------*/

/* Modulname: exists_component */

/* Aufgabe  : Ueberpruefen, ob eine bestimmte Komponente in der Datenbank */
/*            vorhanden ist */

/*----------------------------------------------------------------------*/
/* Parameter */

/* Eingabe  : Konstruktor 'constr', Implementation 'implid', Operation */
/*            'operationid' und Komponente 'comp' */

/* Ausgabe  : Komponente ist vorhanden (Ja/Nein) (TRUE oder FALSE) */

/*----------------------------------------------------------------------*/

Bool exists_component F4(
			 Str, constr,
			 Str, implid,
			 Str, operationid,
			 Str, comp)
{
    return
	exists_Component(vdmdbs,constr,implid,operationid,comp);
}


/*----------------------------------------------------------------------*/

/* Modulname: exists_operation */

/* Aufgabe  : Ueberpruefen, ob eine bestimmte Operation in der Datenbank */
/*            vorhanden ist */

/*----------------------------------------------------------------------*/
/* Parameter */

/* Eingabe  : Konstruktor 'constr', Implementation 'implid', und Operation */
/*            'operationid' */

/* Ausgabe  : Operation ist vorhanden (Ja/Nein) (TRUE oder FALSE) */

/*----------------------------------------------------------------------*/

Bool exists_operation F3(
			 Str, constr,
			 Str, implid,
			 Str, operationid)
{
    Str econstr;
    Str eimplid;
    return exists_Operation(vdmdbs,constr,implid,operationid,
			    &econstr,&eimplid);
}


/*----------------------------------------------------------------------*/

/* Modulname: exists_implementation */

/* Aufgabe  : Ueberpruefen, ob eine bestimmte Implementation in der Daten- */
/*            bank vorhanden ist */

/*----------------------------------------------------------------------*/
/* Parameter */

/* Eingabe  : Konstruktor 'constr' und  Implementation 'implid' */

/* Ausgabe  : Implementation ist vorhanden (Ja/Nein) (TRUE oder FALSE) */

/*----------------------------------------------------------------------*/

Bool exists_implementation F2(
			      Str, constr,
			      Str, implid)
{
    return exists_Implementation(vdmdbs,constr,implid);
}


/*----------------------------------------------------------------------*/

/* Modulname: exists_adt */

/* Aufgabe  : Ueberpruefen, ob ein bestimmter Konstruktor in der Datenbank */
/*            vorhanden ist */

/*----------------------------------------------------------------------*/
/* Parameter */

/* Eingabe  : Konstruktor 'constr' */

/* Ausgabe  : Konstruktor ist vorhanden (Ja/Nein) (TRUE oder FALSE) */

/*----------------------------------------------------------------------*/

Bool exists_adt F1(
     Str, constr)
{
    return exists_Adt(vdmdbs,constr);
}


/*----------------------------------------------------------------------*/

/* Modulname: is_in_components */

/* Aufgabe  : Ueberpruefen, ob eine bestimmte Komponente in einer Liste von */
/*            Komponenten vorhanden ist */

/*----------------------------------------------------------------------*/
/* Parameter */

/* Eingabe  : Komponente 'comp' */
/*            Liste der Komponenten 's' */

/* Ausgabe  : Komponente ist in der Liste vorhanden (Ja/Nein) (TRUE o.FALSE)*/

/*----------------------------------------------------------------------*/

Bool is_in_components F2(
			 Str, comp,
			 Str, s)
{
    Bool res=FALSE;
    if (NOT is_empty_str(s))
	{
	    Nat0 i;
	    Str c;
	    if ((i=index_str(s,','))==0) i=len_str(s)+1;
	    LET c=mk_sub_tup_str(s,1,i-1) IN
	    if (is_eq_str(comp,c))
		res=TRUE;
	    else
		res=is_in_components(comp,sub_tupd_str(s,i+1,len_str(s)));
	    del_str(c);
	}
    else del_str(s);
    return res;
}


/*----------------------------------------------------------------------*/

/* Modulname: get1_params */

/* Aufgabe  : Umsetzen der in einer Liste gegebenen Parameter-Namen in eine */
/*            Abbildung vom Typ 'Params' */

/*----------------------------------------------------------------------*/
/* Parameter */

/* Eingabe  : Liste der Parameter-Namen 's' */
/*            Abbildung der Parameter 'params' */

/* Ausgabe  : Abbildung der Parameter aus der Liste */

/*----------------------------------------------------------------------*/

Params get1_params F2(
		      Str, s,
		      Params, params)
{
    Nat0 i;
    if (NOT is_empty_str(s))
	{
	    if ((i=index_str(s,'\n'))==0) i=len_str(s)+1;
	    params=union1_params(params,mk_sub_tup_str(s,1,i-1),mk_empty_str());
	    params=get1_params(sub_tupd_str(s,i+1,len_str(s)),params);
	}
    else del_str(s);
    return params;
}


/*----------------------------------------------------------------------*/

/* Modulname:  get_params */

/* Aufgabe  :  Lesen der Parameter einer Komponente aus der Datenbank */

/*----------------------------------------------------------------------*/
/* Parameter */

/* Eingabe  : Konstruktor 'constr', Implementation 'implid', und Operation */
/*            'operationid' */

/* Ausgabe  : Parameter der Komponente */

/*----------------------------------------------------------------------*/

Params
get_params F3(
	      Str, constr,
	      Str, implid,
	      Str, operationid)
{
    Str s;
    Params params;
    LET s=get_component(constr,implid,operationid,PARAM) IN
    LET params=get1_params(s,mk_empty_params()) IN

	return params;
}


Params
get_all_params F3(
		  Str, constr,
		  Str, implid,
		  Str, operationid)
{
    Params params,
	globalparams;
    
    LET params=get_params(constr,implid,operationid) IN
    LET globalparams=get_globalimpl_params() IN

	return ovwrtd_params(params,globalparams);
}

/*--------------------------------------------------------------------*/

Bool
param_defined F3(Str,    param,
		 Str,	 adt,
		 Dominfo,dominfo)

{
    Bool        res;
    Description	desc = sel_dominfo(dominfo,adt);
    Str		cons = s_constr(s_domain(desc));
    Implid	iid  = s_implid(s_implem(desc));

    if ( is_nil_implid(iid) )
	{
	    /* predefined types don't have implementations */
	    return FALSE;
	}
    else
	{
	    Params	vdmparams = get_params(cons,s_str_implid(iid),BASIC);
	    Bool	res       = is_in_params(param,vdmparams);

	    del_params(vdmparams);
	    return res;
	}
}

/*--------------------------------------------------------------------*/

Bool
is_forward_allowed F2(Str,adt,Dominfo,dominfo)
{
    return
	param_defined(FORWARD_ALLOWED,adt,dominfo);
}

/*----------------------------------------------------------------------*/

/* Modulname: open_vdmdb */

/* Aufgabe  : Oeffnen einer Datenbank mit der vom VDM-DB Manager */
/*            bereitgestellten Funktion */

/*----------------------------------------------------------------------*/
/* Parameter */

/* Eingabe  : Name der Datenbank 'dbname' */

/* Ausgabe  : Datenbank konnte geoeffnet werden (Ja/Nein) (TRUE/FALSE) */

/*----------------------------------------------------------------------*/

Bool open_vdmdb F1(
		   Str, dbname)
{
    Vdmdbs vdbs;
    Dbs ovdb;
    Bool res=TRUE;

    init_vdmdm();

    ovdb = IF (NOT is_empty_str(dbname))
        THEN mk1_dbs(copy_str(dbname))
        ELSE mk_empty_dbs();

    vdbs=def_Vdmdbs(ovdb);           /* ovdb is consumed by def_Vdmdbs */
    if (exists_Vdmdbs(vdbs))
	vdmdbs=conc_vdmdbs(vdmdbs,vdbs);
    else
	error(mkc_str("can't open vdmdb  "),copy_str(dbname)),
	    res=FALSE;
    return res;
}


/*----------------------------------------------------------------------*/

/* Modulname: delete_vdmdb */

/* Aufgabe  : Schliessen der verwendeten Datenbanken und loeschen der */
/*            entsprechenden Variablen durch die vom DB Manager */
/*            bereitgestellte Funktion */

/*----------------------------------------------------------------------*/
/* Parameter */

/* Eingabe  : (globale Variable 'vdmdbs') */

/* Ausgabe  : */

/*----------------------------------------------------------------------*/

void 
delete_vdmdb F0()
{
    del_vdmdbs(vdmdbs);
    undef_Vdmdbs();
    finish_vdmdm();
}

