/*----------------------------------------------------------------------

 $Id: domfile.c,v 1.12 1996/04/30 20:24:27 vdm Exp $ 

 Aufgabe  : erstellen des 'Erichments' (bestehend aus .pre und .enr
            Teil) aus den im Parse-Baum enthaltenen Informationen 

 Eingabe  : Parse-Baum 

 Ausgabe  : Informationen der .pre bzw. .enr -Datei in Form eines 
            VDM-Strings 

----------------------------------------------------------------------

 Ersteller: Kai Steinau 

----------------------------------------------------------------------

 Flags for conditional compilation:

 SHORTFILENAMES:	IF defined, the files generated by '-s' option
			will be called 'dom_<nnn>' instead of 
			domainname.

----------------------------------------------------------------------

Aenderungen: 

$Log: domfile.c,v $
Revision 1.12  1996/04/30 20:24:27  vdm
extensions for implementing alias as inheritance in C++
solution: a new constructor -isa with external syntax
Subtype is-a Basetype
Semantics for C like "=", but for C++ class SUBTYPE
becomes a subclass of class BASETYPE
Problem: constructors are not inherited, so the
constructors are passed from Basetype to Subtype
in the generator, therefore a special operation
constructor.mk??? introduced, wich is copied in the generator
to the Subtype domain.
The generation order computation did not work save,
therefore an explicit dependency graph is constructed, with this
graph the generation order is computed, old code still there
function "old_gen_order"

Revision 1.11  1996/01/15 10:01:59  vdm
cvs backup

 * Revision 1.10  1995/01/31  18:13:48  vdm
 * SIGNATURE --> SIGNATURESTRING
 * for -X all domains in .enr file written, not only defined domains'
 *
 * Revision 1.9  1994/12/29  16:52:35  vdm
 * new option -X for collecting comments
 * during parsing and writing these comments into the .enr and .pre files
 * for parsing and importing .vdm files into vdmedit
 *
 * Revision 1.8  1994/08/24  19:32:17  vdm
 * warning eliminated
 *
 * Revision 1.7  1993/11/21  20:03:52  vdm
 * Torsten Tews 1.Version eingemischt
 * Dabei interne Domain Gleichungen in C++ ausgeblendet
 * mit einem internen Implementierungsparameter INTERALDOMAIN
 *
 * Revision 1.6  1993/11/03  14:16:54  vdm
 * Hans-Martins Aenderungen eingemischt
 *
 * Revision 1.10  1993/10/29  21:21:48  hmh
 * Several corrections dealing with storage allocation:
 *
 * - In 'gen_deps' call to 'copy_depend' removed within updating
 *   of 'diff'.
 * - In function 'write_enr_file' changes against old '.vdm' file
 *   recognized by reading existing file and using 'is_eq_str'
 *   instead of parsing both files and comparing Dominfos.

 * Revision 1.9  1993/10/24  21:58:21  hmh
 * Changes built in to handle the '-r' option used to check
 * if a given enrichment file can be used as starting point of
 * enrichment process.
 *
 * - trace messages generated using 'trace*'.
 * - new operation 'is_enriched_by' used to check if the enirchment
 *   process can be accellerated

 * Revision 1.8  1993/08/07  05:59:54  hmh
 * Further changes neede for 'VDMGEN':
 *
 * - Generate files 'dom_<nnn>' instead of domain name (bloody 14
 *   character filenames). This may cause too much files to be 
 *   regenerated if the set of domains changes (i.e. some domains may
 *   get other filenames).
 * - conditional compilation 'SHORTFILENAMES' to enable this feature.
 * - Generation order used in 'split_enr_file' to get 'DOMAINLIST' right

 * Revision 1.7  1993/08/06  07:15:19  hmh
 * Changes needed for VDMGEN.
 * - New funciotn 'gen_deps' to generate list of domains a given
 *   domain is depending on (needed for the 'VDMGEN' command to
 *   control recompilation of domain libraries).

 * Revision 1.6  1993/07/25  12:41:21  hmh
 * Initialized decls removed
 *
 * Revision 1.5  1993/07/25  08:50:58  hmh
 * Several changes
 * - new function 'write_enr_file' derived from file 'domgen.c'
 * - in 'gen_params_str_str' '"' replaced by '\"' in parameter value.

 * Revision 1.4  1993/07/05  21:34:06  hmh
 * Merging Uwes changes into mine.
 *
 * Revision 1.4  1993/07/01  14:10:23  vdm
 * Uwes changes for C++ inserted
 * Naming conventions for selector names
 * "for" routine interfaces removed
 *
 * Revision 1.3  1993/05/28  13:27:44  hmh
 * Minor correction concerning standard syntax
 * In 'gen_tree_str' no 's_' prefix added if '-N' is active.

 * Revision 1.2  1993/05/28  11:03:28  hmh
 * Compiler extended to accept the ISO VDM-SL ASCII syntax. These
 * extensions are activated through the '-N' switch.
 *

  * In function 'gen_domain_str_description' special case for 'TOKEN'
  * removed to enforce default output as '() -token'.
 
----------------------------------------------------------------------*/

#include "domadt.h"
#include "domfct.h"
#include "domdsl.h"
#include "domcpp.h"
#include "domfile.h"
#include "domcheck.h"
#include "domcomment.h"


/*----------------------------------------------------------------------

 Modulname: gen_fct_elems 
 Aufgabe  : erzeugen eines VDM-Strings fuer die Elemente eines ADT vom 
            Typ FCT  (Elemente der Domain-Gleichung des ADT) 

 Eingabe  : Elemente des ADT 'elems' 
            Anzahl der Elemente 'n' 
 Ausgabe  : Elemente der Domain-Gleichung des ADT als VDM-String 
----------------------------------------------------------------------*/


/* Str gen_fct_elems(Elems elems, Nat0 n) 	*/
/* Fold-Funktion, definiert 'domadt' 		*/


/*---------------------------------------------------------------------

 Modulname: gen_fct_str 
 Aufgabe  : erzeugen des Strings fuer ein Element eines ADT 

 Eingabe  : Element 'e' 
            Index des Elements 'i' 
            Anzahl der Elemente 'n' der Beschreibung des ADT 
 Ausgabe  : String des Elements fuer Domain-Gleichung 

----------------------------------------------------------------------*/

Str gen_fct_str F3(
		   Str, e,                  /* Element aus 'Elems' */
		   Nat1, i,                 /* Index des Elements */
		   Nat0, n)                 /* Anzahl der Elemente */
{
    return
	IF i<=n
	THEN
	    IF i<n
	    THEN conc_str(copy_str(e),mkc_str(" "))
	    ELSE copy_str(e)
        ELSE
            mk_empty_str();
}


/*----------------------------------------------------------------------*/

/* Modulname: gen_tree_elems */

/* Aufgabe  : erzeugen eines VDM-Strings fuer die Elemente eines ADT vom */
/*            Typ TREE (Elemente der Domain-Gleichung des ADT) */

/* Eingabe  : Elemente des ADT 'elems' */

/* Ausgabe  : Elemente der Domain-Gleichung des ADT als VDM-String */

/*----------------------------------------------------------------------*/

/* Str gen_tree_elems(Elems elems, Str dn) */
/* Fold-Funktion, definiert im 'domadt' */


/*----------------------------------------------------------------------*/

/* Modulname: gen_tree_str */

/* Aufgabe  : erzeugen des Strings fuer ein Element eines ADT */

/* Eingabe  : Element 'e' */
/*            Index des Elements 'i' */

/* Ausgabe  : String des Elements fuer Domain-Gleichung */

/*----------------------------------------------------------------------*/

Str gen_tree_str F3(
		    Str, e,                 /* Element aus 'Elems' */
		    Nat1, i,                /* Index des Elements */
		    Str, dn)                /* Domain name */
{
    return
	IF i%2
	THEN
	    conc3_str(IF (is_in_params(PARAM_['N'], global_params))
		      THEN mkc_str("")
		      ELSE mkc_str("s_"),
		      IF (is_in_params(PARAM_['a'],global_params))
                            /* remove implicit selector extension */
                      THEN mk_sub_tup_str(e,1,len_str(e)-len_str(dn)-1)
                      ELSE copy_str(e),
                      mkc_str(" : "))
        ELSE
            conc_str(copy_str(e),mkc_str("\n\t\t"));
}


/*----------------------------------------------------------------------*/

/* Modulname: gen_union_elems */

/* Aufgabe  : erzeugen eines VDM-Strings fuer die Elemente eines ADT vom */
/*            Typ UNION (Elemente der Domain-Gleichung des ADT) */

/* Eingabe  : Elemente des ADT 'elems' */

/* Ausgabe  : Elemente der Domain-Gleichung des ADT als VDM-String */

/*----------------------------------------------------------------------*/

/* Str gen_union_elems(Elems elems, Str dn) */
/* Fold-Funktion, definiert in 'domadt' */


/*----------------------------------------------------------------------*/

/* Modulname: gen_union_str */

/* Aufgabe  : erzeugen des Strings fuer ein Element eines ADT */

/* Eingabe  : Element 'e' */
/*            Index des Elements 'i' */

/* Ausgabe  : String des Elements fuer Domain-Gleichung */

/*----------------------------------------------------------------------*/

Str gen_union_str F3(
		     Str, e,                 /* Element aus 'Elems' */
		     Nat1, i,                /* Index des Elements */
		     Str, dn)                /* Domain name */
{
    return
	IF ( i>2 ) AND ( i%2 )
	THEN
	    conc_str(mkc_str(" |"),gen_tree_str(e,i,dn))
        ELSE
            gen_tree_str(e,i,dn);
}

/*----------------------------------------------------------------------*/

/* Modulname: gen_tokuni_elems */

/* Aufgabe  : erzeugen eines VDM-Strings fuer die Elemente eines ADT vom */
/*            Typ TOKUNI (Elemente der Domain-Gleichung des ADT) */


/* Eingabe  : Elemente des ADT 'elems' */
/*            Anzahl der Elemente 'n' */

/* Ausgabe  : Elemente der Domain-Gleichung des ADT als VDM-String */

/*----------------------------------------------------------------------*/

/* gen_tokuni_elems(Elems elems, Nat0 n) */
/* Fold-Funktion, definiert in 'domadt' */


/*----------------------------------------------------------------------*/

/* Modulname: gen_tokuni_str */

/* Aufgabe  : erzeugen des Strings fuer ein Element eines ADT */

/* Eingabe  : Element 'e' */
/*            Index des Elements 'i' */
/*            Anzahl der Elemente 'n' */

/* Ausgabe  : String des Elements fuer Domain-Gleichung */

/*----------------------------------------------------------------------*/

Str gen_tokuni_str F3(
		      Str, e,                    /* Element aus 'Elems' */
		      Nat1, i,                   /* Index des Elements */
		      Nat0, n)                   /* Anzahl der Elemente */
{
    return
	IF i < n
	THEN conc_str(copy_str(e), mkc_str("\n\t |"))
	ELSE copy_str(e);
}


/*----------------------------------------------------------------------*/

/* Modulname: gen_depend */

/* Aufgabe  : erzeugen eines VDM-Strings fuer die Beschreibung der */
/*            Abhaengigeiten eines ADT von anderen ADT's */

/* Eingabe  : ADT's von denen der ADT 'dn' abhaengt 'de' */
/*            Name des ADT 'dn' */

/* Ausgabe  : Domain-Gleichung der Abhaengigkeiten des ADT als VDM-String */

/*----------------------------------------------------------------------*/

/* Str gen_depend(Depend de, Str dn) */
/* Fold-Funktion, definiert in 'domadt' */


/*----------------------------------------------------------------------*/

/* Modulname: gen_depend_str */

/* Aufgabe  : erzeugen eines Strings zur Darstellung der Abhaengigkeit */
/*            eines ADT von einem anderen ADT */

/* Eingabe  : Name des ADT von dem der ADT 'dn' abhaengig ist 'd' */
/*            Name des abhaengigen ADT 'dn' */

/* Ausgabe  : String zur Beschreibung der Abhaengigkeit fuer die */
/*            Domain-Gleichung */

/*----------------------------------------------------------------------*/

Str gen_depend_str F2(
		      Str, d,                   /* Element aus 'Depend' */
		      Str, dn)                  /* ADT-Name */
{
    return conc_str(mkc_str("\n"),
		    conc3_str(copy_str(dn),
			      mkc_str(" -> "),
			      copy_str(d)));
}


/*----------------------------------------------------------------------*/

/* Modulname: gen_domain_dominfo */

/* Aufgabe  : erzeugen von Domain-Gleichungen aus den Informationen des */
/*            Parse-Baums */

/* Eingabe  : Parse-Baum 'dominfo' */
/*            Angabe ob vordefinierte (PREDEF) oder benutzer-definierte */
/*            (USERDEF) Domain-Gleichungen erzeugt werden 'def' */

/* Ausgabe  : Domain-Gleichungen der entsprechenden ADT's als VDM-String */

/*----------------------------------------------------------------------*/

/* Str gen_domain_dominfo(Dominfo dominfo, Str def) */
/* Fold-Funktion, definiert in 'domadt' */


/*----------------------------------------------------------------------*/

/* Modulname: gen_domain_str_description */

/* Aufgabe  : erzeugen der Domain-Gleichung fuer einen ADT */

/* Eingabe  : Name des ADT 'dn' */
/*            Beschreibung des ADT 'desc' */
/*            Angabe, ob vordefinierte oder benutzer-definierte Gleichungen */
/*            erzeugt werden */

/* Ausgabe  : Domain-Gleichung des ADT als VDM-String */

/*----------------------------------------------------------------------*/

Str gen_domain_str_description F3(
      Str, dn,                  /* ADT-Name */
      Description, desc,        /* ADT-Beschreibung */
      Str, def)                 /* 'USERDEF' oder 'PREDEF' oder ALLDEF */
{
 Str s;                  /* Liste der Domain-Definitionen */
 Str constr;             /* Konstruktor des ADT */
 Elems elems;            /* Elemente des ADT */
 Depend depend;          /* ADT's von denen der ADT abhaengig ist */
 s=mk_empty_str();
 constr=s_constr(s_domain(desc));
 elems=s_elems(s_domain(desc));
 depend=s_depend(s_domain(desc));
 /* Aufnahme aller ADT's, die als 'PREDEF' definiert sind bzw. aller ADT's */
 /* ('PREDEF' und nicht 'PREDEF'), wenn 'USERDEF' definiert ist */
 if (   is_eq_str(def,ALLDEF)
    OR  is_eq_str(constr,PREDEF)
    OR (NOT is_eq_str(def,PREDEF))
    )
 {
  /* erzeugen der ADT-Definition */
  s=conc_str(s,copy_str(dn));
  s=conc_str(s,IF is_eq_str(constr,TREE) THEN
                  mkc_str("\t::")
               ELSE
                  mkc_str("\t= "));
  if (is_eq_str(constr,TUPLE))
     s=conc3_str(s,copy_str(hd_elems(elems)),mkc_str("*"));
  else
  if (is_eq_str(constr,OPTI))
     s=conc_str(s,conc3_str(mkc_str("["),copy_str(hd_elems(elems)),mkc_str("]")));
  else
  if (is_eq_str(constr,MAP))
     s=conc_str(s,conc3_str(copy_str(hd_elems(elems)),mkc_str(" -m-> "),copy_str(sel_elems(elems,2))));
  else
  if (is_eq_str(constr,FCT))
     s=conc5_str(s,mkc_str("("),gen_fct_elems(elems,len_elems(elems)-1),
                   mkc_str(") --> "),copy_str(last_elems(elems)));
  else
  if (is_eq_str(constr,TREE))
     s=conc_str(s,gen_tree_elems(elems,dn));
  else
  if (is_eq_str(constr,UNION))
     s=conc_str(s,gen_union_elems(elems,dn));
  else
  if (is_eq_str(constr,TOKUNI))
     s=conc_str(s,gen_tokuni_elems(elems,len_elems(elems)));
  else
  if (is_eq_str(constr,PREDEF))
     s=conc_str(s,copy_str(PREDEF));
  else
  if (is_eq_str(constr,ALIAS))
     s=conc_str(s,copy_str(hd_elems(elems)));
  else
     s=conc5_str(s,mkc_str("("),gen_fct_elems(elems,len_elems(elems)),
                   mkc_str(") -"),copy_str(constr));
  /* erzeugen der Definitionen fuer Abhaengigkeiten */
  s=conc3_str(s,gen_depend(depend,dn),mkc_str("\n"));
 }
 s=conc_str(s,give_specdom_comment(dn));
 return s;
}


/*----------------------------------------------------------------------

 Modulname: gen_params 
 Aufgabe  : erzeugen der Parameter-Definitionen fuer Operationen und 
            Implementationen aus den Informationen des Parse-Baums 

 Eingabe  : Abbildung (Map) der Parameter 'params' 
 Ausgabe  : durch ',' getrennte Aufzaehlung der Parameter-Definitionen 
            als VDM-String 

----------------------------------------------------------------------*/

/* gen_params(Params p)				 */
/* Fold-Funktion, definiert in 'domadt'		 */


/*----------------------------------------------------------------------

 Modulname: gen_params_str_str 
 Aufgabe  : erzeugen eines VDM-Strings fuer eine Parameter-Definition 
	    ersetze '"' durch '\"' in parameter value

 Eingabe  : Parameter-Name 'pn'
            Parameter-Wert 'pv' 
 Ausgabe  : Parameter-Definition 'pn=pv' als VDM-String 

----------------------------------------------------------------------*/

Str gen_params_str_str F2(
      Str, pn,                       /* Parameter-Name */
      Str, pv)                       /* Parameter-Wert */
{
 return conc4_str(gen_quote_str(copy_str(pn)),
                  mkc_str("="),
                  gen_quote2_str(copy_str(pv)),
                  mkc_str(","));
}


/*----------------------------------------------------------------------

 Modulname: gen_operations 
 Aufgabe  : erzeugen der Operations-Definitionen fuer die 
            Implementations-Beschreibung eines ADT aus den Informationen 
            des Parse-Baums 

 Eingabe  : Abbildung (Map) der Operationen 'o' 
 Ausgabe  : Operations-Definitionen als VDM-String 

----------------------------------------------------------------------*/

/* Str gen_operations(Operations o) 		*/
/* Fold-Funktion, definiert in 'domadt' 	*/


/*----------------------------------------------------------------------*/

/* Modulname: gen_operations_str_params */

/* Aufgabe  : erzeugen der Operations-Definition fuer eine Operation */

/* Eingabe  : Operationsname 'on' */
/*            Parameter der Operation 'ov' */

/* Ausgabe  : VDM-String mit der Operationsdefinition '+ on(ov)' */

/*----------------------------------------------------------------------*/

static Str fctstr;

Str gen_operations_str_params F2(
      Str, on,                       /* Operation-Name */
      Params,ov)                     /* Operation-Wert (=Parameter) */
{
  Str s;
  if ( collectcomments && is_prefix_str(SIGNATUREOP,on)) {
    Str name;
    Str dom;
    Str sig;

    name = sel_params(ov,FCTNAME);
    dom  = sel_params(ov,mkc_str("X"));
    sig  = sel_params(ov,SIGNATURESTRING);

    fctstr = conc_str(fctstr,copy_str(sig));
    fctstr = conc_str(fctstr,mkc_str("\n"));
    fctstr = conc_str(fctstr,give_specfct_comment(name,dom));
    
    s = mk_empty_str();
  }
  else {
    s=mkc_str("\n\t + ");
    s=conc_str(s,copy_str(on));
    if (NOT is_empty_params(ov))
    {
      s=conc3_str(s,mkc_str("("),gen_params(ov));
      s=ovwrt1_str(s,len_str(s),')');
    }
  }
  return s;
}


/*----------------------------------------------------------------------*/

/* Modulname: gen_impl_str_params */

/* Aufgabe  : erzeugen der Implementations-Definition eines ADT fuer seine */
/*            Implementations-Beschreibung */

/* Eingabe  : Implementation 'implid' */
/*            Parameter der Implementation 'params' */

/* Ausgabe  : Implementations-Definition als VDM-String 'implid(params)' */

/*----------------------------------------------------------------------*/

Str gen_impl_str_params F2(
      Implid, implid,                /* Implementation-Name */
      Params, params)                /* Implementation-Paramter */
{
 Str s=copy_str(s_str_implid(implid));
 if (NOT is_empty_params(params))
 {
    s=conc3_str(s,mkc_str("("),gen_params(params));
    s=ovwrt1_str(s,len_str(s),')');
 }
 return s;
}


/*----------------------------------------------------------------------*/

/* Modulname: gen_implementation_dominfo */

/* Aufgabe  : erzeugen der Implementations-Beschreibungen aus den */
/*            Informationen des Parse-Baums */

/* Eingabe  : Parse-Baum 'dominfo' */
/*            Angabe ob vordefinierte (PREDEF) oder benutzer-definierte */
/*            (USERDEF) Domain-Gleichungen erzeugt werden 'def' */

/* Ausgabe  : Implementations-Beschreibungen der ADT's als VDM-String */

/*----------------------------------------------------------------------*/

/* Str gen_implementation_dominfo(Dominfo dominfo, Str def) */
/* Fold-Funktion, definiert in 'domadt' */


/*----------------------------------------------------------------------*/

/* Modulname: gen_implementation_str_description */

/* Aufgabe  : erzeugen der Implementations-Beschreibung fuer einen ADT */

/* Eingabe  : Name des ADT 'dn' */
/*            Beschreibung des ADT 'desc' */

/* Ausgabe  : Implementations-Beschreibung des ADT als VDM-String */

/*----------------------------------------------------------------------*/

Str gen_implementation_str_description F3(
      Str, dn,                   /* ADT-Name */
      Description, desc,         /* ADT-Beschreibung */
      Str, def)                  /* 'USERDEF' oder 'PREDEF' */
{
 Str s;                   /* Liste der Implementations-Definitionen */
 Str constr;              /* Konstruktor des ADT */
 Implid implid;           /* Implementations-Name des ADT */
 LET s=mk_empty_str() IN
 LET constr=s_constr(s_domain(desc)) IN
 LET implid=s_implid(s_implem(desc)) IN
 /* Aufnahme nur der nicht-'PREDEF' Definitionen bei 'def'='USERDEF' oder */
 /* aller 'PREDEF'-Definitionen bei 'def'='PREDEF' */
 if ( is_eq_str(def,ALLDEF)
    OR ( is_eq_str(def,PREDEF) == is_eq_str(constr,PREDEF) )
    )
  /* Implementation oder Operationen gegeben */
  s=IF (NOT is_empty_operations(s_operations(s_implem(desc)))) OR
       (NOT is_nil_implid(implid)) THEN
       conc3_str(copy_str(dn),mkc_str("\t= "),
       conc3_str(IF NOT is_nil_implid(implid) THEN
                 gen_impl_str_params(implid,s_iparams(s_implem(desc)))
              ELSE
                 mk_empty_str(),
              gen_operations(s_operations(s_implem(desc))),
              mkc_str("\n")))
    ELSE
       mk_empty_str();
 s=conc_str(s,give_specimp_comment(dn));
 return s;
}


/*----------------------------------------------------------------------

 Modulname: gen_enr_str 
 Aufgabe  : erzeugen einer DSL-Speyifikation aus den Informationen des 
            Parse-Baums 

 Eingabe  : Parse-Baum 'dominfo' 
            Angabe ob vordefinierte (PREDEF) oder benutzer-definierte 
            (USERDEF) oder alle (ALLDEF) Domain-Gleichungen erzeugt werden 'def' 
 Ausgabe  : DSL-Spezifikation als VDM-String 

----------------------------------------------------------------------*/

Str gen_enr_str F2(
      Dominfo, dominfo,          /* Parse Baum */
      Str, def)                  /* 'USERDEF' oder 'PREDEF' */
{
 Str s;

 fctstr = mk_empty_str();

 s=give_init_comment();
 s=conc_str(s,mkc_str("\nDOMAINS\n"));
 s=conc_str(s,give_domain_comment());
 s=conc_str(s,gen_domain_dominfo(dominfo,def));
 s=conc_str(s,mkc_str("\n\nIMPLEMENTATIONS\n"));
 s=conc_str(s,give_impl_comment());
 s=conc_str(s,gen_implementation_dominfo(dominfo,def));
 s=conc_str(s,mkc_str("\nFUNCTIONS\n"));
 s=conc_str(s,give_fct_comment());
 s=conc_str(s,fctstr);

 return s;
}



/*---------------------------------------------------------------------

 Modulname: write_enr_file 
 Aufgabe  : Output of VDM file includeing eventually only, if changed.

 Eingabe  : VDM specification as string (s)
	    Output file name (fn)
	    Option '-E'
 Ausgabe  :

----------------------------------------------------------------------*/

void write_enr_file F2(Str,s,Str,fn)
{ Bool writeflag;

 /* compasison of old and new .enr and .pre files 
    and conditional writing confuses make 
    if .vdm file is touched but not realy changed 
    the computing of .enr is done in every make call, but has no effect 
    So make conditional output dependent of '-E' option */

 if (is_in_params(PARAM_['E'],global_params))
    { /* read old '.vdm' file. If it does not exist, an empty
	 string will be returned which will cause the comparison to 
	 fail. */
	
      Str ovdm = fread_str(fn); 
      writeflag = (is_eq_str(s,ovdm)) ? FALSE : TRUE;
      del_str(ovdm);
    }
 else
    { writeflag = TRUE; }

 if (writeflag)
    { trace1(mkc_str("Writing file "),copy_str(fn));
      if (fwrite_str(s,fn)==FALSE)
         { warning(mkc_str("Can't write file "),copy_str(fn)); }
    }
  else
    { trace1(mkc_str("No changes in file "),copy_str(fn)); }

 del_str(s);
 del_str(fn);

}

/*----------------------------------------------------------------------

 Modulname: gen_deps 
 Aufgabe  : Generate list of dependent domains 

 Eingabe  : domain description
 Ausgabe  : String, contining list of depending domains

----------------------------------------------------------------------*/

Depend gen_deps_str F2(
      Str,dn,
      Dominfo,di)
  { Domain dd;
    dd = s_domain(sel_dominfo(di,dn));
    return union_depend(copy_depend(s_depend(dd))
		       ,get_elems(s_elems(dd),s_constr(dd)));
  }

Str tofile_str F1(Str,dn) 
 { return conc_str(copy_str(dn),mkc_str("\n")); }

static Str gen_deps F2(
      Domain,dd,
      Dominfo,di)
  { Depend dep;
    Depend diff;
    Str    r;

    dep = union_depend(copy_depend(s_depend(dd))
		      ,get_elems(s_elems(dd),s_constr(dd)));
    diff = copy_depend(dep);
    do
     { Depend d;
       d = gen_deps_depend(diff,di);
       del_depend(diff);
       diff = diffd_depend(copy_depend(d),dep);
       dep = union_depend(dep,d);
     }
    while (! is_empty_depend(diff));

    r = tofile_depend(dep);
    del_depend(diff); del_depend(dep);
    return r;
  }

/*----------------------------------------------------------------------

 Modulname: gen_domain 
 Aufgabe  : Generate .vdm file for one domain

 Eingabe  : Domain name
	    domain description

----------------------------------------------------------------------*/

static Dominfo gdi;
static Str filelist;
static Nat0 filecnt;

static Str gen_predefs_str F1(Str, dn)
  { return conc_str(copy_str(dn),mkc_str(" = PREDEF\n")); }

static void gen_domain F1(Str, dn)
  { Str fn,fnd;
    Str s;
    Str depdoms;
    Description dd;
    Domain d;
    Depend dp;
    char n[10];

    dd = sel_dominfo(gdi,dn);
    d = s_domain(dd);
    dp = union_depend(copy_depend(s_depend(d))
		     ,get_elems(s_elems(d),s_constr(d)));
    sprintf(n,"%d",(int)(filecnt++));
#ifdef SHORTFILENAMES
    fn = conc_str(mkc_str("dom_"),mkc_str(n));
#else
    fn = copy_str(dn);
#endif

    s = mkc_str("DOMAINS\n");
    s = conc_str(s,coll_depends_depend(dp,gen_predefs_str));
    s = conc_str(s,gen_domain_str_description(dn,dd,USERDEF));
    s = conc_str(s,mkc_str("IMPLEMENTATIONS\n"));
    s = conc_str(s,gen_implementation_str_description(dn,dd,USERDEF));

    depdoms = gen_deps(d,gdi);

    filelist = conc_str(filelist
		       ,conc4_str(copy_str(fn)
				 ,mkc_str("\t")
				 ,copy_str(dn)
				 ,mkc_str("\n")));
    fnd = conc_str(copy_str(fn),mkc_str(".deps"));
    fwrite_str(depdoms,fnd);
    write_enr_file(s,conc_str(fn,mkc_str(".vdm")));
    del_depend(dp); del_str(depdoms); del_str(fnd);

    profile_all();
  }


/*----------------------------------------------------------------------

 Modulname: split_enr_file 
 Aufgabe  : Split dominfo into a collection of VDM file, one for each 
	    domain

 Eingabe  : Dominfo to be split
	    output file name (dummy argument, ignored)

----------------------------------------------------------------------*/

void split_enr_file F2(Dominfo, di, Str, fn)
  { Str flf = mkc_str("DOMAINLIST");
    
    if (! get_gen_order(di)) { return; }

    trace1(mkc_str("\nSplitting domain equations..."),Empty);
    filelist = mkc_str("");
    filecnt = 100;
    gdi = di;
    forall_elems(gen_order,gen_domain);

    trace1(mkc_str("Writing file "),copy_str(flf));
    if (fwrite_str(filelist,flf)==FALSE)
       { warning(mkc_str("Can't write file "),copy_str(flf)); }

    del_str(filelist); del_str(flf);
  }


/*----------------------------------------------------------------------

 Modulname: is_enriched_by
 Aufgabe  : check if the second domain description is a potential
	    potential enrichment of the first one.

 Eingabe  : domain description
	    potential enrichment domain description
 Ausgabe  : TRUE, if it is an enrichment

----------------------------------------------------------------------*/

/* Bool is_enriched_by(Dominfo,Dominfo): Fold functio over Dominfo */


Bool is_enriched_by_str_description F3(
      Str, dn,
      Description, dd,
      Dominfo, di)
  { Description		dd1;

    Str			cn,cn1;
    Elems		el,el1;
    Depend		dep,dep1;

    Implid		impl,impl1;
    Params		ipars,ipars1;
    Operations		ops,ops1;

    trace2(mkc_str("enrichment check for "),dn);

    if (! is_in_dom_dominfo(dn,di,&dd1)) 
       { trace1(mkc_str("enrichment check failed for new domain: "),dn);
	 return FALSE;
       }

    smk_domain(s_domain(dd),&cn,&el,&dep);
    smk_domain(s_domain(dd1),&cn1,&el1,&dep1); 
    smk_implementation(s_implem(dd),&impl,&ipars,&ops);
    smk_implementation(s_implem(dd1),&impl1,&ipars1,&ops1);
    
    if (! (  ((is_eq_str(cn,cn1) && is_eq_elems(el,el1)) || is_eq_str(cn,PREDEF))
          && is_subset_depend(dep,dep1)
       )  )
       { trace1(mkc_str("domain equation changed:" ), dn);
	 return FALSE;
       }
    
    if (! (is_eq_str(cn,PREDEF)
	  || (is_eq_implid(impl,impl1) && is_eq_params(ipars,ipars1))
       )  )
       { trace1(mkc_str("implementation changed: "),dn);
	 return FALSE;
       }
    
    if (! is_submap_operations(ops,ops1))
       { trace1(mkc_str("operation set changed: "),dn);
	 return FALSE;
       }
    
    return TRUE;

  }
