/*----------------------------------------------------------------------

 Modulname: $Id: domgen.c,v 1.16 1996/04/30 20:24:27 vdm Exp $ 

 Aufgabe  : Steuerung der Generierung einzelner Komponenten von Modulen 

----------------------------------------------------------------------
 Parameter

 Eingabe  :

 Ausgabe  : 

----------------------------------------------------------------------

 Ersteller: Kai Steinau 

----------------------------------------------------------------------

Aenderungen:

   $Log: domgen.c,v $
   Revision 1.16  1996/04/30 20:24:27  vdm
   extensions for implementing alias as inheritance in C++
   solution: a new constructor -isa with external syntax
   Subtype is-a Basetype
   Semantics for C like "=", but for C++ class SUBTYPE
   becomes a subclass of class BASETYPE
   Problem: constructors are not inherited, so the
   constructors are passed from Basetype to Subtype
   in the generator, therefore a special operation
   constructor.mk??? introduced, wich is copied in the generator
   to the Subtype domain.
   The generation order computation did not work save,
   therefore an explicit dependency graph is constructed, with this
   graph the generation order is computed, old code still there
   function "old_gen_order"

   Revision 1.15  1996/01/15 10:02:01  vdm
   cvs backup

 * Revision 1.14  1995/01/31  18:13:49  vdm
 * SIGNATURE --> SIGNATURESTRING
 * for -X all domains in .enr file written, not only defined domains'
 *
 * Revision 1.13  1994/08/24  14:21:24  vdm
 * new TOKEN syntax <ABC>
 *
 * Revision 1.12  1994/02/10  10:36:39  vdm
 * option -J for adding C++ default operations
 *
 * Revision 1.11  1994/02/08  17:55:58  vdm
 * warning message changed
 *
 * Revision 1.10  1994/01/30  17:17:00  vdm
 * -D defines only passed to 1. call of preprocessor
 * and not to internal calls
 * there syntax of vdm files can be destroied
 *
 * Revision 1.9  1994/01/12  21:59:05  vdm
 * Torsten Tews Aenderungen vom Dezember eingebaut
 *
 * Revision 1.4.2.3  1993/12/08  21:17:03  uvdm
 * Neue Version des Compilers von Prof. Schmidt, basiert auf meinen
 * Erweiterungen vom 17.11
 * Anpassungen entsprechend der Version vom 20.11 durchgefuehrt
 *
 * Revision 1.8  1993/11/28  14:56:58  vdm
 * trace output fuer automem optional
 *
 * Revision 1.7  1993/11/28  09:54:32  vdm
 * Options sortiert
 *
 * Revision 1.6  1993/11/21  20:03:53  vdm
 * Torsten Tews 1.Version eingemischt
 * Dabei interne Domain Gleichungen in C++ ausgeblendet
 * mit einem internen Implementierungsparameter INTERALDOMAIN
 *
 * Revision 1.5  1993/11/03  14:16:56  vdm
 * Hans-Martins Aenderungen eingemischt
 *
 * Revision 1.4.2.1  1993/11/20  17:20:30  vdm
 * 1.Version von Torsten Tews
 *
 * Revision 1.7  1993/10/29  21:21:51  hmh
 * Several corrections dealing with storage allocation:
 *
 * - In function 'exten_vdm' call of 'del_str(s)' removed since 
 *   the parser is corrected such that the input string will be
 *   destroyed.
 *
 * Revision 1.6  1993/10/28  20:50:28  hmh
 * Error detection improved.
 *
 * In function 'extend_domainid' local flag 'OK' introduced to 
 * control occurence of errors detected for the current domain.
 * Further processing within this function made dependent of this
 * flag instead of 'NOR yyexitcode' which produced the effect that
 * "missing operation" errors, etc. were ommitted for further
 * domains.

 * Revision 1.5  1993/10/24  21:58:26  hmh
 * Changes built in to handle the '-r' option used to check
 * if a given enrichment file can be used as starting point of
 * enrichment process.
 *
 * Revision 1.4.2.1  1993/11/17  21:03:44  uvdm
 * Aenderungen fuer automatische Speicherverwaltung eingebaut; der
 * eingelesene Parsebaum wird in domgen.c transformiert. Aenderungen sind mir
 * / * TT * / markiert. Auswahl der Transformation ueber neuen Kommandozeilen-
 * parameter -M
 *
 * Revision 1.4  1993/09/01  18:11:16  vdm
 * Hans-Martins Version vom 1.9.93 als neue Hauptversion
 *
 * Revision 1.3  1993/07/01  14:10:24  vdm
 * Uwes changes for C++ inserted
 * Naming conventions for selector names
 * "for" routine interfaces removed
 *
 * - trace messages generated using 'trace*'.
 * - Handling of '-r' option introduced. For this, in 'gen_dominfo'
 *   a test was introduced after parsing all '.vdm' files
 *   to see if the enrichment file supplied with the '-r' option
 *   can be taken as starting point for enrichment computation.

 * Revision 1.4  1993/07/25  08:51:03  hmh
 * Several changes
 *
 * 's_param' -> 's_iparam'

 * Revision 1.3  1993/07/05  21:34:08  hmh
 * Merging Uwes changes into mine.
 *
 * Revision 1.3  1993/07/01  14:10:24  vdm
 * Uwes changes for C++ inserted
 * Naming conventions for selector names
 * "for" routine interfaces removed
 *
 * Revision 1.2  1993/05/22  15:47:24  hmh
 * Condtional writing of '.enr' and '.pre' files caused problems
 * together with traditional 'make' dependencies.
 * Command line parameter '-E' introduced to enable this feature
 * if needed (i.e. by 'VDM'), disabled by default.
 *
   93/02/11: Definition der globalen Parameter (-S) in 
             'extend_dominfo'

----------------------------------------------------------------------*/

#include "domadt.h"
#include "domfct.h"
#include "domdsl.h"
#include "domcpp.h"
#include "domdb.h"
#include "domcheck.h"
#include "domfile.h"
#include "domgen.h"
#include "domcomment.h"

/*----------------------------------------------------------------------*/

Params def_params;  /* defined params for a component */


/*----------------------------------------------------------------------*/

/* Modulname: mk_name */

/* Aufgabe  : Erzeugen von Parametern in Gross- oder Kleinschreibweise */
/*            entsprechend dem Parameter '-U' */

/*----------------------------------------------------------------------*/
/* Parameter */

/* Eingabe  : Paramter 's' */

/* Ausgabe  : Parameter in Gross- bzw. Kleinschreibweise */

/*----------------------------------------------------------------------*/

Str
mk_name F1(
	   Str, s)
{
    register Nat1 i;
    Str r;

    r=copy_str(s);
    for (i=1;
	 i<=len_str(s);
	 i++)
	ovwrt1_str(r,
		   i,
		   IF (is_in_params(PARAM_['U'],global_params))
		   THEN toupper(sel_str(s,i))
		   ELSE tolower(sel_str(s,i)));
    return r;
}


/*----------------------------------------------------------------------*/

/* Modulname: mk_two_params */

/* Aufgabe  : Erzeugen von zwei Parametern (ein Parameter in gegebener */
/*            Schreibweise + ein Parameter in Gross- bzw. Kleinschreibweise)*/

/*----------------------------------------------------------------------*/
/* Parameter */

/* Eingabe  : Parameter-Name 'pn' */
/*            Parameter-Wert 'pv' */

/* Ausgabe  : Erzeugte Parameter */

/*----------------------------------------------------------------------*/

Params
mk_two_params F2(
		 Str, pn,
		 Str, pv)
{
    return
	union1_params(mk1_params(conc_str(copy_str(CPP_PARAM),
					  copy_str(pn)),
				 mk_name(pv)),
		      pn,
		      pv);
}


/*----------------------------------------------------------------------*/

/* Modulname: two_params */

/* Aufgabe  : Erzeugen von jeweils zwei Parametern fuer jeden angegebenen */
/*            Parameter (in Gross- bzw. Kleinschreibweise und in gegebener */
/*            Schreibweise) */

/*----------------------------------------------------------------------*/
/* Parameter */

/* Eingabe  : Abbildung von Parametern 'p' */

/* Ausgabe  : Erzeugte Parameter */

/*----------------------------------------------------------------------*/

Params 
two_params F1(
	      Params, p)
{
    Str pn,pv;
    Params r,c;

    r=mk_empty_params();
    c=copy_params(p);

    while (NOT is_empty_params(c))
	{
	    pn=sel_dom_params(c);
	    pv=sel_params(c,pn);
	    r=ovwrtd_params(r,
			    mk_two_params(copy_str(pn),
					  copy_str(pv)));
	    c=sub1d_params(c,pn);
	}

    del_params(c);
    
    return r;
}
/*----------------------------------------------------------------------*/


/*----------------------------------------------------------------------*/

/* Modulname: get_fct_str */

/* Aufgabe  : Erzeugen eines Parameters fuer ein Element eines ADT vom Typ */
/*            'FCT' */

/*----------------------------------------------------------------------*/
/* Parameter */

/* Eingabe  : Element des ADT 'elem' */
/*            Index des Elements 'i' */
/*            Anzahl der Elemente des ADT 'n' */

/* Ausgabe  : Erzeugter Parameter */

/*----------------------------------------------------------------------*/

Params
get_fct_str F3(
	       Str, elem,
	       Nat1, i,
	       Nat1, n)
{
    return
	IF (i<=n)
        THEN mk_two_params(gen_str(ELEM,i),
			   copy_str(elem))
        ELSE mk_empty_params();
}

/*----------------------------------------------------------------------*/

/* Modulname: get_tree_str */

/* Aufgabe  : Erzeugen eines Parameters fuer ein Element eines ADT vom Typ */
/*            'TREE' */

/*----------------------------------------------------------------------*/
/* Parameter */

/* Eingabe  : Element des ADT 'elem' */
/*            Index des Elements 'i' */

/* Ausgabe  : Erzeugter Parameter */

/*----------------------------------------------------------------------*/

Params
get_tree_str F3(
		Str, elem,
		Nat1, i,
		Str, domainid)
{
    Params p;
    if ( i%2 )
	{
	    Str s,s2;
	    p = mk1_params(gen_str(SEL,(i+1)/2),
			   copy_str(elem));
	    s = conc_str(copy_str(CPP_PARAM),
			 copy_str(SEL));
	    s2 = copy_str(elem);

	    if ( extent_selector_names )
		{
		    /* remove _domain suffix for .cpp component selection */
		    s2 = sub_tup_str(s2,
				     1,
				     len_str(s2) - len_str(domainid) - 1);
		}
	    p = union1_params(p,
			      gen_str(s,(i+1)/2),
			      s2);
	    del_str(s); 
	}
    else 
	{
	    p = mk_two_params(gen_str(COMP,i/2),
			      copy_str(elem));
	}

    return p;
}




/*----------------------------------------------------------------------*/

/* Modulname: get_union_str */

/* Aufgabe  : Erzeugen eines Parameters fuer ein Element eines ADT vom Typ */
/*            'UNION' */

/*----------------------------------------------------------------------*/
/* Parameter */

/* Eingabe  : Element des ADT 'elem' */
/*            Index des Elements 'i' */

/* Ausgabe  : Erzeugter Parameter */

/*----------------------------------------------------------------------*/

Params 
get_union_str F2(
		 Str, elem,
		 Nat1, i)
{
    return
	IF (NOT(i%2))
	THEN
           mk_two_params(gen_str(VAR,i/2),
			 copy_str(elem))
        ELSE
           mk_empty_params();
}


/*----------------------------------------------------------------------*/

/* Modulname: get_tokuni_str */

/* Aufgabe  : Erzeugen eines Parameters fuer ein Element eines ADT vom Typ */
/*            'TOKUNI' */

/*----------------------------------------------------------------------*/
/* Parameter */

/* Eingabe  : Element des ADT 'elem' */
/*            Index des Elements 'i' */
/*            Anzahl der Elemente des ADT 'n' */

/* Ausgabe  : Erzeugter Parameter */

/*----------------------------------------------------------------------*/

Str get_tokuni_str F3(
      Str, elem,
      Nat1, i,
      Nat1, n)
{
 return  IF i<n THEN appr_str(copy_str(elem),',') ELSE copy_str(elem);
}


/*----------------------------------------------------------------------*/

/* Modulname: make_params */

/* Aufgabe  : Erzeugen der Parameter fuer einen ADT */

/*----------------------------------------------------------------------*/
/* Parameter */

/* Eingabe  : Name des ADT 'domainid' */
/*            Konstruktor des ADT 'constr' */
/*            Elemente des ADT 'elems' */

/* Ausgabe  : Paramter fuer den ADT */

/*----------------------------------------------------------------------*/

Params make_params F3(
      Str, domainid,
      Str, constr,
      Elems, elems)
{
 Params params;
#ifdef OLDVDM
 if ((is_eq_str(constr,TUPLE)) OR (is_eq_str(constr,OPTI)))
 {
    params=union_params(mk_two_params(copy_str(ELEM),copy_str(sel_elems(elems,1))),
                        mk_two_params(copy_str(ELEM1),copy_str(sel_elems(elems,1))));
 }
 else
#endif
 if (is_eq_str(constr,TREE))
    params=get_tree_elems(elems,domainid);
 else
 if (is_eq_str(constr,UNION))
    params=union_params(get_tree_elems(elems,domainid)
                       ,get_union_elems(elems));
 else
 if (is_eq_str(constr,TOKUNI))
    params=union1_params(get_fct_elems(elems,len_elems(elems)),
    copy_str(ENUM),get_tokuni_elems(elems,len_elems(elems)));
 else
 if (is_eq_str(constr,FCT))
    params=union_params(get_fct_elems(elems,len_elems(elems)-1),
    mk_two_params(copy_str(ELEM0),copy_str(last_elems(elems))));
 else
    params=get_fct_elems(elems,len_elems(elems));
 return union_params(mk_two_params(copy_str(DOMAINNAME),
                     copy_str(domainid)),params);
}


/*----------------------------------------------------------------------*/

/* Modulname: is_domsubmap1_str_str */

/* Aufgabe  : Feststellen ob der Name eines Parameters im Vorbereich */
/*            einer Abbildung von Parametern vorhanden ist */

/*----------------------------------------------------------------------*/
/* Parameter */

/* Eingabe  : Parameter-Name 'pn1' */
/*            Parameter-Wert 'pv1' */
/*            Abbildung der Parameter 'p2' */

/* Ausgabe  : Parameter-Name 'pn1' ist in Parametern 'p2' enthalten */
/*            (Ja/Nein) (TRUE oder FALSE) */

/*----------------------------------------------------------------------*/

Bool is_domsubmap1_str_str F3(
      Str, pn1,
      Str, pv1,
      Params, p2)
{
 return is_in_params(pn1,p2);
}

/*----------------------------------------------------------------------*/


/*----------------------------------------------------------------------*/

/* Modulname: get_gen_component */

/* Aufgabe  : Festlegen der fuer eine zu erzeugende Komponente zu lesenden */
/*            Teile aus der Datenbank */

/*----------------------------------------------------------------------*/
/* Parameter */

/* Eingabe  : Konstruktor 'constr', Implementation 'implid', Operation */
/*            'operationid', und Komponente 'comp' der aus der Datenbank */
/*            zu lesenden Datei, fuer die Implementation gegebenen */
/*            Parameter 'params' */

/* Ausgabe  : Innhalt der aus der Datenbank gelesenen Datei */

/*----------------------------------------------------------------------*/

Str get_gen_component F5(
      Str, constr,
      Str, implid,
      Str, operationid,
      Str, comp,
      Params, params)
{
 Bool suspress=FALSE;
 Str pv,retcomp=mk_empty_str();
 /* Operationen der Implementation sollen mit Makros implementiert werden */
 if ((is_in_dom_params(MACROS,params,&pv)) AND (NOT is_empty_str(pv)))
 {
    /* Die Erzeugung von Makros beeinflusst die Signatur- und Code- */
    /* Komponenten */
    if (((is_eq_str(comp,CCOD)) OR (is_eq_str(comp,GSIG)))
     AND ((exists_component(constr,implid,operationid,CMCR))
     OR (exists_component(constr,implid,operationid,GCOD))))
       suspress=TRUE;
 }
 else
 /* Operationen der Implementation sollen ohne Makros implementiert werden */
 if (is_eq_str(comp,CMCR))
    suspress=TRUE;
 if (NOT suspress)
 {
    if (is_in_components(comp,Sign_components))
        retcomp=get_signature(constr,operationid,comp);
    else
    {
        retcomp=get_component(constr,implid,operationid,comp);
        if (is_empty_str(retcomp))
           /* Die Komponente GCOD ersetzt die CCOD und CMCR Komponente */
           if ((is_eq_str(comp,CCOD)) OR (is_eq_str(comp,CMCR)))
              retcomp=get_component(constr,implid,operationid,GCOD);
    }
 }
 return (retcomp);
}


/*----------------------------------------------------------------------*/

/* Modulname: extend_enr */

/* Aufgabe  : Erzeugen der Parameter und lesen der Komponente fuer eine */
/*            Operation */

/*----------------------------------------------------------------------*/
/* Parameter */

/* Eingabe  : String an den die gelesene Komponente angehaengt wird 's' */
/*            Name des ADT 'domainid' */
/*            Konstruktor des ADT 'constr' */
/*            Implementation des ADT 'implid' */
/*            zu lesende Komponente 'comp' */
/*            Parameter der Implementation 'implparams' */

/* Ausgabe  : Um gelesene Komponente erweiterter String */

/*----------------------------------------------------------------------*/

Str extend_enr F7(
		  Str, s,
		  Str, domainid,
		  Str, constr,
		  Implid, implid,
		  Operations, operations,
		  Str, comp,
		  Params, implparams)
{
    if (NOT is_empty_operations(operations))
	{
	    Str operationid,ioperationid;
	    Nat0 i;
	    Params params;

	    /* Auswahl einer Operation */
	    LET operationid=sel_dom_operations(operations) IN
	    LET params=sel_operations(operations,operationid) IN

	    /* Umsetzung von 'fold', 'binary' und 'rename' Opearationen */
	    LET ioperationid=copy_str(operationid) IN
		if ((i=index_str(operationid,'.'))!=0)
		    ioperationid=sub_tupd_str(ioperationid,1,i-1);

	    /* Verarbeitung von VDM Komponenten */
	    if (is_eq_str(comp,VDM))
		{
		    /* Die Operation wird nicht gefunden */
		    if (NOT exists_operation(constr,s_str_implid(implid),ioperationid))
			{
			    errorfile(domainid);
			    outstr(conc_str(
					    conc4_str(mkc_str(" Operation "),
						      mk_quot(constr),
						      mkc_str(" "),
						      mk_quot(implid)),
					    conc4_str(mkc_str(" "),
						      mk_quot(operationid),
						      mkc_str(" not found for domain "),
						      mk_quot(domainid))));
			}
		    /* Die Operation wurde gefunden */
		    else
			{
			    Params vdmparams=get_all_params(constr,
							    s_str_implid(implid),
							    ioperationid);
			    /* Die Parameter werden ueberprueft */
			    if (NOT is_domsubmap_params(params,vdmparams))
				{
				    errorfile(domainid);
				    outstr(conc4_str(mkc_str("Invalid parameter for operation "),
						     mk_quot(operationid),
						     mkc_str(" for domain "),
						     mk_quot(domainid)));
				}
			    del_params(vdmparams);
			}

		    if (NOT yyexitcode)
			{
			    /* definieren der Parameter, lesen der Komponente */
			    s=define(s,copy_params(params));
			    def_params=copy_params(params);
			    s=include(s,domainid,constr,s_str_implid(implid),ioperationid,comp);
			    s=undefine(s,def_params);
			}
		}
	    /* Verarbeitung von nicht VDM-Komponenten */
	    else
		{
		    Params implopparams=ovwrtd_params(copy_params(implparams),
						      copy_params(params));

		    /* definieren der Parameter, lesen der Komponente */
		    s=define(s,two_params(params));
		    s=conc_str(s,
			       get_gen_component(constr,implid,ioperationid,comp,implopparams));
		    del_params(implopparams);
		    s=undefine(s,two_params(params));
		}
	    del_str(ioperationid);
	    s=extend_enr(s,
			 domainid,
			 constr,
			 implid,
			 sub1_operations(operations,operationid),
			 comp,
			 implparams);
	}
    return s;
}


/*----------------------------------------------------------------------
 Modulname: extend_domainid
 Aufgabe  : Erzeugen der Parameter und lesen der Komponenten fuer die
            Implementation eines ADT
----------------------------------------------------------------------
 Parameter 
 Eingabe  : String an den die gelesenen Komponenten angehaengt werden 's' 
            Name des ADT 'domainid' 
            Parse-Baum 'dominfo' 
            Komponente 'comp' 
 Ausgabe  : Um gelesene Komponenten erweiterter String 
----------------------------------------------------------------------*/

Str extend_domainid F4(
      Str, s,
      Str, domainid,
      Dominfo, dominfo,
      Str, comp)
{
 Description desc;              /* Beschreibung des ADT */
 Str constr;                    /* Konstruktor des ADT */
 Implid implid;                 /* Implementation des ADT */
 Params params;                 /* Implementationsparameter */
 Bool OK = TRUE;		/* Error found for this domain */

 LET desc=sel_dominfo(dominfo,domainid) IN
 LET constr=s_constr(s_domain(desc)) IN
 LET implid=s_implid(s_implem(desc)) IN
 trace1(mkc_str("parsing VDM files for  "),copy_str(domainid));
 if ((NOT is_eq_str(constr,PREDEF)) AND (NOT is_nil_implid(implid)))
 {
    Params vdmparams;

    params=s_iparams(s_implem(desc));
    vdmparams=get_all_params(constr,s_str_implid(implid),BASIC);
    /* Verarbeitung von VDM Komponenten */
    if (is_eq_str(comp,VDM))
    {
       /* Der Konstruktor ist nicht vorhanden */
       if (NOT exists_adt(constr))
       {
	  OK = FALSE;
          errorfile(domainid);
          outstr(conc4_str(mkc_str(" Constructor "),mk_quot(constr),
                 mkc_str(" not found for domain "),mk_quot(domainid)));
       }
       else
       /* Die Implementation wurde nicht gefunden */
       if (NOT exists_implementation(constr,s_str_implid(implid)))
       {
	  OK = FALSE;
          errorfile(domainid);
          outstr(conc6_str(mkc_str(" Implementation "),mk_quot(constr),mkc_str(" "),
                 mk_quot(implid),mkc_str(" not found for domain "),mk_quot(domainid)));
       }
       else
          /* Ueberpruefung der Parameter */
          if (NOT is_domsubmap_params(params,vdmparams))
          {
	     OK = FALSE;
             errorfile(domainid);
             outstr(conc4_str(mkc_str(" Invalid parameter for implementation "),
                    mk_quot(implid),mkc_str(" for domain "),mk_quot(domainid)));
          }
    }
    if ( OK 
         &&  ! ( ( is_eq_str(comp,CPP) || is_eq_str(comp,CPPCLASS) )
                 && ( is_in_params(INTERNALDOMAIN,params)
                    )
               )
       )
    {
       /* Parameter der Implementation */
       Params adtparams,defparams,glbparams;

       /* Parameter zum ADT und seinen Elementen */
       adtparams=ovwrtd_params(get_globalimpl_params(),
                               make_params(domainid,
                                           constr,
                                           s_elems(s_domain(desc))));
       adtparams=ovwrtd_params(adtparams,copy_params(params));

       /* Globale Implementationsparameter '-S' */
       glbparams=ovwrtd_params(get_globalimpl_params(),
                               copy_params(params));

       s=define(s,copy_params(adtparams));
       def_params=adtparams;
       s=include(s,domainid,constr,s_str_implid(implid),BASIC,comp);
       LET defparams=copy_params(def_params) IN
       del_params(def_params);
       /* lesen der Komponenten der Operationen des ADT */
       s=extend_enr(s,domainid,constr,implid,
            copy1_operations(s_operations(s_implem(desc))),comp,glbparams);
       def_params=defparams;
       s=include(s,domainid,constr,s_str_implid(implid),POST,comp);
       del_params(glbparams);
       s=undefine(s,defparams);
       s=undefine(s,copy_params(vdmparams));
        /* for undefining default definitions of impl params */
    }
    del_params(vdmparams);
 }
 return s;
}


/*----------------------------------------------------------------------*/

/* Modulname: extend_dominf */

/* Aufgabe  : Auswahl der ADT's des Parse-Baums in der benoetigten */
/*            Reihenfolge und Aufruf der Funktion zur Erzeugung der */
/*            Komponenten des ADT */

/*----------------------------------------------------------------------*/
/* Parameter */

/* Eingabe  : String, an die die zu erzeugenden Komponenten angefuegt */
/*            werden sollen 's' */
/*            Parse-Baum 'dominfo' */
/*            Zu erzeugende Komponente 'comp' */

/* Ausgabe  : Um erzeugte Komponenten erweiterter String 's' */

/*----------------------------------------------------------------------*/

Str extend_dominf F3(
      Str, s,
      Dominfo, dominfo,
      Str, comp)
{
 Str domainid;
 if (NOT is_empty_dominfo(dominfo))
 {
    /* die Komponente benoetigt keine Generierungreihenfolge fuer die ADT's */
    if (NOT is_in_components(comp,Order_components))
       domainid=copy_str(sel_dom_dominfo(dominfo));
    /* die Komponente benoetigt eine Generierungreihenfolge fuer die ADT's */
    else
    {
       domainid=copy_str(hd_elems(gen_order));
       gen_order=tld_elems(gen_order);
    }
    s=extend_domainid(s,domainid,dominfo,comp);
    s=extend_dominf(s,sub1_dominfo(dominfo,domainid),comp);
    del_str(domainid);
 }
 return s;
}


/*----------------------------------------------------------------------*/

/* Modulname: extend_dominfo */

/* Aufgabe  : Erstellen einer Liste aus den, aus einem Parse-Baum */
/*            abzuleitenden Komponenten aus der VDM-Datenbank */

/*----------------------------------------------------------------------*/
/* Parameter */

/* Eingabe  : Parse-Baum 'diff' */
/*            Abzuleitende Komponente 'comp' */
/*            Parameter der Komponente 'def' */

/* Ausgabe  : VDM-String der abgeleiteten Komponenten */

/*----------------------------------------------------------------------*/

Str extend_dominfo F3(
      Dominfo, diff,
      Str, comp,
      Params, def)
{
 Str s=mk_empty_str();
 trace1(mkc_str("Generating component "),copy_str(comp));
 if (is_in_components(comp,Order_components))
    get_gen_order(diff);
 def_params=mk_empty_params();
 rootvalues=mk_empty_rootvals();
 s=define(s,def);
 s=define(s,get_globalimpl_params());
 s=include(s,STDADT,STDLIB,COM,SETUP,comp);
 s=undefine(s,get_globalimpl_params());
 del_params(def_params);
 s=extend_dominf(s,copy1_dominfo(diff),comp);
 if (NOT yyexitcode)
    s=cppvdm(s,comp,FALSE);
 else
    {del_str(s);s=mk_empty_str();}
 if (NOT is_eq_str(comp,VDM))
    del_rootvals(rootvalues);
 return s;
}

/*----------------------------------------------------------------------*/


/*----------------------------------------------------------------------*/

/* Modulname: extend_vdm */

/* Aufgabe  : Erweiterung des Parse-Baums 'dominfo' um die aus 'diff' */
/*            abzuleitenden VDM-Definitionen */

/*----------------------------------------------------------------------*/
/* Parameter */

/* Eingabe  : Parse-Baum 'dominfo' */
/*            Parse-Baum 'diff' */

/* Ausgabe  : Erweiterter Parse-Baum 'dominfo' (Vollstaendiger Operations- */
/*            Satz) */

/*----------------------------------------------------------------------*/

Dominfo
extend_vdm F2(
              Dominfo, dominfo,
              Dominfo, diff)
{
    if (domtrace>2)
        {
            puts_str(mkc_str("\n\ndominfo :\n"),stdout);
            pr_dominfo(dominfo);
            puts_str(mkc_str("\n\ndiff :\n"),stdout);
            pr_dominfo(diff);fflush(stdout);
        }

    /* fuer ADT's ohne Konstruktor oder Implementation koennen keine Komponenten */
    /* abgeleitet werden */
    diff=del_predef_dominfo(copy_dominfo(diff),diff);
    if ( (NOT is_empty_dominfo(diff))
         AND (NOT yyexitcode) )
        {
            Dominfo newdominfo;
            Params params;
            Str s;
            LET params=mk1_params(copy_str(DUMMY),copy_str(DEF)) IN
            LET params=union1_params(params,copy_str(VDMDOMAIN),copy_str(DEF)) IN

            /* Ableitung der Komponenten aus den Informationen des Parse-Baums */
            LET s=extend_dominfo(diff,VDM,params) IN
            if (NOT yyexitcode)
                {
                    /* Der 'neue' Parse-Baum ergibt sich aus den Informationen des 'alten' */
                    /* und allen neu gefundenen Komponenten */
                    /* string 's' will be deleted by 'parsevdm' */
                    newdominfo=merge_dominfo(copy_dominfo(dominfo),parsevdm(s));
                    newdominfo=update_alias_dominfo(copy_dominfo(newdominfo),newdominfo);
                    /* aus dem 'neuen' Parse-Baum werden weitere Komponenten abgeleitet */
                    dominfo=extend_vdm(newdominfo,
                                       diff_dominfo(copy_dominfo(newdominfo),
                                                    dominfo));
                }
        }

    del_dominfo(diff);

    return dominfo;
}

/*
 *----------------------------------------------------------------------
 * Modulname : automem_impl (Version 1.1, 16/11/93)
 * Aufgabe   : Transformation einer Implementationsbeschreibung, so
 *             da angegebene Implementation durch eine automem-Implemen-
 *             tion ersetzt wird !
 *
 * Parameter : old_impl  - zu transformierende Impl.-Beschreibung 
 *             Fkt.-wert - transformierte Impl.-Beschreibung
 *----------------------------------------------------------------------
*/ /* TT */

Implementation
automem_impl F1(Implementation,old_impl)
{
    Implementation new_impl;
    Params         old_params = s_iparams(old_impl);
    Str            baseimpl_val = conc_str(gen_quote_str(copy_str(INTERNALDOMAIN))
                                           ,mkc_str("=\"1\""));
    Params         new_params;
   
    if ( automem_impl_with_macros ) {
        baseimpl_val = conc4_str(baseimpl_val
                                 ,mkc_str(",")
                                 ,gen_quote_str(copy_str(MACROS))
                                 ,mkc_str("=\"1\""));
    } 

    while (!is_empty_params(old_params))
        {
            Str pname = sel_dom_params(old_params);
            Str pval  = sel_params(old_params,pname);
            
            old_params   = sub1_params(old_params,pname);
            baseimpl_val = conc5_str(baseimpl_val,
                                     mkc_str(","),
                                     gen_quote_str(pname),
                                     mkc_str("="),
                                     gen_quote2_str(pval));
        }
    del1_params(old_params);
   
    baseimpl_val = conc4_str(s_str_implid(s_implid(old_impl)),
                             mkc_str("("),baseimpl_val,mkc_str(")"));

    new_params = mk1_params(copy_str(BASEIMPL),baseimpl_val);

    new_impl = mk_implementation(mk_opt_implid(copy_str(AUTOMEM)),
                                 new_params,
                                 s_operations(old_impl));
                                 
    return new_impl;   
}

/* --------------------------------------------------------------------*/

Implementation 
add_default_ops F1(Implementation,impl)
{
    if ( automem_impl_with_defaults )
        {
            Operations     ops     = s_operations(impl);
            Implementation new_impl;

            ops = merge1_operations(ops,
                                    copy_str(DEFAULT),
                                    mk_empty_params());

            new_impl = mk_implementation(s_implid(impl),
                                         s_iparams(impl),
                                         ops);
            return new_impl;
        }

    return impl;
}

/*
 *----------------------------------------------------------------------
 * Modulname : automem_transform (Version 1.1, 16/11/93)
 * Aufgabe   : Tranformation des Ableitungsbaumes, so da alle Implemen-
 *             tationen, soweit erlaubt und notwendig, durch eine 
 *             entsprechende automem-Implementation ersetzt werden.
 *
 * Parameter : dominfo  - zu transformierender Ableitungsbaum
 *             Fkt.erg. - transformierter Ableitungsbaum
 *----------------------------------------------------------------------
*/ /* TT */

Dominfo
automem_transform F1(Dominfo,old_dominfo)
{
    Dominfo new_dominfo = mk_empty_dominfo();
   
    while ( !is_empty_dominfo(old_dominfo) )
        {
            Str domain_name;
            Description domain_descr;
            Domain domain;
            Implementation impl;
            Implementation new_impl;
      
            domain_name  = sel_dom_dominfo(old_dominfo);
            domain_descr = sel_dominfo(old_dominfo, domain_name);
            domain       = s_domain(domain_descr);
            impl         = s_implem(domain_descr);
            new_impl     = add_default_ops(impl);
      
            if ( !is_nil_implid(s_implid(impl)) )
                {
                    Str impl_name = s_str_implid(s_implid(impl));
                    Params impl_params = get_params(s_constr(domain),
                                                    impl_name,
                                                    BASIC);
                    if (is_in_params(AUTOMEM_OK,impl_params))
                        new_impl = automem_impl(impl);
                    else   
                        if (is_in_params(AUTOMEM_NOT_ALLOWED,impl_params))
                            {
                                Str w = conc3_str(mkc_str("automatic memory management not supported for implementation ")
                                                  ,mk_quot(impl_name)
                                                  ,mkc_str(" of domain ")
                                                  );
                                warning(w,copy_str(domain_name));
                            }
                    del_params(impl_params);   
                }
      
            new_dominfo = union1_dominfo(new_dominfo,
                                         domain_name,
                                         mk_description(domain,new_impl));
            old_dominfo = sub1_dominfo(old_dominfo,
                                       domain_name);
        }
   
    del1_dominfo(old_dominfo);
    return new_dominfo;
}

Dominfo 
add_constructors_for_isa F1(Dominfo,dominfo)
{
    trace1(mkc_str("Adding constructors for isa domains"),Empty);

    dominfo = add_constructors_for_isa_dominfo(copy_dominfo(dominfo),dominfo);

    if (domtrace>2)
	{
	    puts_str(mkc_str("\n\ndominfo :\n"),stdout);
	    pr_dominfo(dominfo);
	    fflush(stdout);
	}

    return dominfo;
}

/*----------------------------------------------------------------------*/

/* Modulname: gen_dominfo */

/* Aufgabe  : Erstellen eines Gesammt-Parse-Baum aus einer Liste von */
/*            Dateien und ableiten der  VDM-Datenbank-Komponenten aus */
/*            dem Parse-Baum */

/*----------------------------------------------------------------------*/
/* Parameter */

/* Eingabe  : Anfangsindex der Dateinamen in einer Liste 'i' */
/*            Anzahl der Elemente in der Liste 'n' */
/*            Liste der Elemente 's' */

/* Ausgabe  : Abgeleiteter Parse-Baum */

/*----------------------------------------------------------------------*/

Dominfo
gen_dominfo F3(
	       int, i,
	       int, n,
	       char **, s)
{
    extern char *infile;
    extern char *rootfile;

    Dominfo dominfo,
	diff;
    Str file,
	vdm,
	enr,
	pre,
	ps;
    Bool gen;
    Bool compiling_vdm_file;

    gen=FALSE;
    yyexitcode=FALSE;
    crossref=mk_empty_crossref();
    dominfo=mk_empty_dominfo();

    while ((i<n) && (NOT yyexitcode))
	{
	    parse_count=0;
	    file=mkc_str(s[i]);
	    vdm=gen_file_name(file,VDM);
	    enr=gen_file_name(file,ENR);
	    pre=gen_file_name(file,PRE);

	    /* Test auf korrekten Datei-Namen */
	    if ((is_eq_str(file,vdm))
		OR (is_eq_str(file,enr))
		OR (is_eq_str(file,pre)))
		{
		    if (NOT exists_file(file))
			error(mkc_str("file not found :  "),copy_str(file));
		    else
			{
			    trace1(mkc_str("Reading...  "),copy_str(file));
			    if (NOT is_eq_str(file,enr))
				gen=TRUE;

			    compiling_vdm_file = is_eq_str(file,vdm); /* TT */
			    rootvalues=mk_empty_rootvals();
			    def_params=mk_empty_params();
			    ps=include(line_str(conc3_str(mkc_str("#line 1 \""),
							  copy_str(file),
							  mkc_str("\""))),
				       file,
				       Empty,
				       Empty,
				       Empty,
				       Empty);
			    del_params(def_params);
			    diff=parsevdm(cppvdm(ps,VDM,is_eq_str(file,vdm))); /* ps wird verbraucht */
			    del_rootvals(rootvalues);
			    if (domtrace>2)
				{
				    puts_str(mkc_str("\n\n"),stdout);
				    puts_str(file,stdout);
				    puts_str(mkc_str(":\n"),stdout);
				    pr_dominfo(diff);fflush(stdout);
				}
          
			    /* Transformation of read vdm-file for automatic memory management:
			       Replace each implementation by an automem implementation */
			    if (is_eq_str(file,vdm)
				&& is_in_params(PARAM_['M'],global_params))
				{
				    trace1(mkc_str("Transforming dominfo for automem implementation"),Empty);
				    if (domtrace>2)
					{
					    puts_str(mkc_str("\n\ndominfo :\n"),stdout);
					    pr_dominfo(diff); fflush(stdout);
					}
				    diff = automem_transform(diff);
				    trace1(mkc_str("dominfo for automem implementation transformed"),Empty);
				    if (domtrace>2)
					{
					    puts_str(mkc_str("\n\ndominfo :\n"),stdout);
					    pr_dominfo(diff);
					    fflush(stdout);
					}
				}
          
			    /* Zusammensetzen der einzelnen Parse-Baeume zu einem Gesammt- */
			    /* Parse-Baum */
			    if (NOT yyexitcode)
				{
				    rootfile=malloc(len_str(file)+1);
				    infile=malloc(len_str(file)+1);
				    cpc_str(file,rootfile);
				    cpc_str(file,infile);
				    rootvalues=mk1_rootvals(mk_fileval(copy_str(file),
								       Empty,
								       Empty,
								       Empty,
								       1,
								       1));
				    dominfo=merge_dominfo(dominfo,diff);
				    del_rootvals(rootvalues);
				    free(rootfile);
				    free(infile);
				}
			}
		}
	    else
		error(mkc_str("wrong file name (.vdm or .enr or .pre expected)  "),
                      copy_str(file));

	    del_str(file);
	    del_str(vdm);
	    del_str(enr);
	    del_str(pre);
	    i++;
	}

    /* Check if potential enrichment file has been supplied through
       a '-r' option. If so, merge it into the files parsed 
       up to here.
       */
    {
	Str	enrfn;
	Dominfo	enr;

	if (is_in_dom_params(PARAM_['r'],global_params,&enrfn))
	    {
		trace1(mkc_str("Checking possible enrichment file..."),Empty);
		if (! exists_file(enrfn))
		    {
			message(mkc_str("Cannot open enrichment file: "),
				copy_str(enrfn));
		    }
		else
		    {
			rootvalues=mk_empty_rootvals();
			enr = parsevdm(fread_str(enrfn));
			if (is_enriched_by_dominfo(dominfo,enr))
			    {
				trace1(mkc_str("merging enrichment..."),Empty);
				dominfo = merge_dominfo(dominfo,enr);
			    }
		    }
	    }	
    }
    

    parse_count=0;
    if ( (gen)
         AND (!yyexitcode) )
	{
	    Params p;
	    trace1(mkc_str("\nGenerating enrichment file"),Empty);

	    /* ergaenzen des Parse-Baums um Standart-Datentypen, wenn Parameter '-k' */
	    /* nicht angegeben mit Definition sonst als PREDEF (NOSTD=1) */
	    LET p=IF include_standard_domains
		  THEN mk_empty_params()
		  ELSE mk1_params(copy_str(NOSTD),copy_str(DEF))
	    IN
		/* unterdruecken der Erzeugung des vollstaendigen Operationessatzes */
		if ( incomplete_set_of_operations )
		    {
			ps=define(mk_empty_str(),
				  copy_params(p));
			rootvalues=mk1_rootvals(mk_fileval(copy_str(STDADT),
							   copy_str(STDLIB),
							   copy_str(COM),
							   copy_str(BASIC),
							   1,
							   1));
			ps=conc_str(ps,get_component(STDLIB,COM,BASIC,VDM));
			diff=parsevdm(cppvdm(ps,VDM,FALSE)); /* ps wird verbraucht */
			dominfo=merge_dominfo(dominfo,diff);
			del_rootvals(rootvalues);
		    }
	    /* Standart-Datentypen einbinden (als PRDEEF oder mit Definition) */
	    /* abhaengig vom Parameter -k */
	    diff=mk1_dominfo(copy_str(STDADT),
			     mk_description(mk_domain(copy_str(STDLIB),
						      mk_empty_elems(),
						      mk_empty_depend()),
					    mk_implementation(copy_str(COM),
							      p,
							      mk_empty_operations())));
	    dominfo=merge_dominfo(dominfo,diff);

	    /* ergaenzen fehlender Definitionen fuer Alias-Konstruktoren */
	    dominfo=update_alias_dominfo(copy_dominfo(dominfo),dominfo);

	    /* unterdruecken der Erzeugung des vollstaendigen Operationessatzes */
	    if ( ( NOT yyexitcode )
		AND ( NOT incomplete_set_of_operations ) )
		dominfo=extend_vdm(dominfo,copy_dominfo(dominfo));

            /* C++ Konstruktoren zu den isa domains dazunehmen */
	    if ( ( NOT yyexitcode )
		AND ( NOT incomplete_set_of_operations ) )
                dominfo = add_constructors_for_isa(copy_dominfo(dominfo),dominfo);

	    /* semantische Ueberpruefung */
	    if (NOT yyexitcode)
		semantic_check(dominfo);
	}

    if ((yyexitcode)
	AND (NOT is_empty_dominfo(dominfo)))
	{
	    del_dominfo(dominfo);
	    dominfo=mk_empty_dominfo();
	}

    return dominfo;
}

/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/

/* Modulname: gen_enr_comp */

/* Aufgabe  : Steuerung der Ausgabe des Inhalts eines Parse-Baums in eine */
/*            Datei */

/*----------------------------------------------------------------------*/
/* Parameter */

/* Eingabe  : Parse-Baum 'dominfo' */
/*            Name der zu erstellenden Datei 'vdm_file_name' */
/*            Angabe ob '.enr'-Datei (def=USERDEF) oder '.pre'-Datei */
/*            (def=PREDEF) erstellt werden soll  'def' */
/*            Name der Datei-Endung '.enr' oder '.pre'  'comp' */

/* Ausgabe  : */

/*----------------------------------------------------------------------*/

void gen_enr_comp F4(
      Dominfo, dominfo,
      Str, vdm_file_name,
      Str, def,
      Str, comp)
{
 Str fn,s;         /* Hilfsvariablen */
 LET fn=gen_file_name(vdm_file_name,comp) IN
 LET s=gen_enr_str(dominfo,def) IN

 write_enr_file(s,fn);
}


/*----------------------------------------------------------------------*/

/* Modulname: gen_enr_file */

/* Aufgabe  : Erzeugung der '.enr' und '.pre' Dateien aus einem Parse-Baum */

/*----------------------------------------------------------------------*/
/* Parameter */

/* Eingabe  : Parse-Baum 'dominfo' */
/*            Datei-Name 'vdm_file_name' */

/* Ausgabe  : */

/*----------------------------------------------------------------------*/

void gen_enr_file F2(
      Dominfo, dominfo,
      Str, vdm_file_name)
{
 parse_count=0;
 trace1(mkc_str("Generating output files for "),copy_str(vdm_file_name));
 gen_enr_comp(dominfo,vdm_file_name,(collectcomments ? ALLDEF : USERDEF),ENR);
 gen_enr_comp(dominfo,vdm_file_name,PREDEF,PRE);
}

/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/

/* Modulname: open_vdmdbs */

/* Aufgabe  : Oeffnen der in der Abbildung der globalen Parameter */
/*            'global_params' gegeben Datenbanken */

/*----------------------------------------------------------------------*/
/* Parameter */

/* Eingabe  : */

/* Ausgabe  : Datenbanken konnten geoeffnet werden (Ja/Nein) (TRUE o. FALSE)*/

/*----------------------------------------------------------------------*/

Bool open_vdmdbs()
{
 Bool res=TRUE;
 Str dbname;
 vdmdbs=mk_empty_vdmdbs();
 trace1(mkc_str("Opening VDM database..."),Empty);
 if (NOT is_in_params(PARAM_['B'],global_params))
 {
  res&=open_vdmdb(dbname=mk_empty_str());
  del_str(dbname);
 }
 else
 {
  Nat0 p;
  Str v=copy_str(sel_params(global_params,PARAM_['B']));
  while (NOT is_empty_str(v))
  {
    p=index_str(v,'\n');
    if (NOT p) p=len_str(v)+1;
    dbname=mk_sub_tup_str(v,1,p-1);
    res&=open_vdmdb(dbname);
    del_str(dbname);
    v=sub_tupd_str(v,p+1,len_str(v));
  }
  del_str(v);
 }
 return res;
}


/*----------------------------------------------------------------------*/

/* Modulname: delete_vdmdbs */

/* Aufgabe  : Schliessen der geoeffneten Datenbanken */

/*----------------------------------------------------------------------*/
/* Parameter */

/* Eingabe  : */

/* Ausgabe  : */

/*----------------------------------------------------------------------*/

void delete_vdmdbs()
{
 trace1(mkc_str("Closing VDM database..."),Empty);
 delete_vdmdb();
}
