%{
/*----------------------------------------------------------------------

  $Id: domparse.y,v 1.16 1996/04/30 20:24:28 vdm Exp $

  Aenderungen:

  $Log: domparse.y,v $
  Revision 1.16  1996/04/30 20:24:28  vdm
  extensions for implementing alias as inheritance in C++
  solution: a new constructor -isa with external syntax
  Subtype is-a Basetype
  Semantics for C like "=", but for C++ class SUBTYPE
  becomes a subclass of class BASETYPE
  Problem: constructors are not inherited, so the
  constructors are passed from Basetype to Subtype
  in the generator, therefore a special operation
  constructor.mk??? introduced, wich is copied in the generator
  to the Subtype domain.
  The generation order computation did not work save,
  therefore an explicit dependency graph is constructed, with this
  graph the generation order is computed, old code still there
  function "old_gen_order"

  Revision 1.15  1996/01/15 10:02:03  vdm
  cvs backup

 * Revision 1.14  1994/12/29  16:52:36  vdm
 * new option -X for collecting comments
 * during parsing and writing these comments into the .enr and .pre files
 * for parsing and importing .vdm files into vdmedit
 *
 * Revision 1.13  1994/12/27  20:13:17  vdm
 * union domains extended:
 * a union domain can now be defined step by step with the following syntax
 * Uni = D1 | D2
 * ...
 * Uni = ... | D3 | D4
 * ...
 * Uni = ... | D5
 *
 * Revision 1.12  1994/11/27  19:05:12  vdm
 * new keywords set bag nat nat0 int char bool real rat
 * for ISO syntax
 *
 * Revision 1.11  1994/09/05  09:08:42  vdm
 * errors found by test progs corrected
 *
 * Revision 1.10  1994/08/13  19:44:05  vdm
 * TDEPEND --> TFCT,
 * new function arrow => for transformation functions
 *
 * Revision 1.9  1994/08/13  11:48:42  vdm
 * syntax of VDM-DSL extended:
 * a FUNCTIONS part added to specify the signature of user defined functions
 * DOMAINS, IMPLEMENTATIONS and FUNCTIONS parts may be mixed arbitrarily
 * the keyword FUNCTIONS may also be writtes as Functions
 * the syntax of a signature is
 *   fct1,...,fctn : D1 x ... x Dm -> R1 x ... x Ro
 * these signatures are implemeted by the com.com.signature operation
 * x is a new keyword
 * lexical syntax of selectors changed: numbers are not longer allowed as the first char
 * in identifiers
 * the --> symbol (TFCT) did not longer work because of the -- comment
 * therfore the symbol substituted by -> (TDEPEND)
 *
 * Revision 1.8  1993/11/03  14:16:57  vdm
 * Hans-Martins Aenderungen eingemischt
 *
 * Revision 1.6  1993/07/05  21:34:11  hmh
 * Merging Uwes changes into mine.
 *
 * Revision 1.6  1993/07/01  14:10:27  vdm
 * Uwes changes for C++ inserted
 * Naming conventions for selector names
 * "for" routine interfaces removed
 *
 * Revision 1.5  1993/05/28  17:47:30  hmh
 * Tree selector handling corrected.
 *
 * Revision 1.3  1993/05/28  11:03:30  hmh
 * Compiler extended to accept the ISO VDM-SL ASCII syntax. These
 * extensions are activated through the '-N' switch.
 *
 * The new ISO VDM-SL ASCII syntax extensions contain:
 *   - Domain names may consist of arbitrary combinations of
 *     upper and lower case character containing at least one
 *     lower case char (to distinguish from TOKEN element).
 *   - The ASCII form of the type constructors are available:
 *       'seq of', 'map _ to _', 'compose _ of _ end'
 *   - The keywords 'Types' and 'Implementations' are allowed for
 *     'DOMAINS' and 'IMPLEMENTATIONS'.
 *   - Selector names no longer have to start with 's_'.
 *     In this case, the selector functions are taken unchanged
 *     while (for union types) the discriminants will be built
 *     by adding the 'd_' prefix (which might lead to 'd_s_').
 *   - Type definitions may be separated by ';'.
 *   - Since domain idfs and implementation names are the same 
 *     now, the domain equation 
 *        Domain = <adtname>
 *     can no longer be supoerted and must be written as
 *        Domain = () -<adtname>
 *     (This applies even without '-N' option!)
 *
 * These extensions are activated through the '-N' option which
 * also defines the flag 'GEN_ISO' for the database.

 * Revision 1.2  1993/05/22  13:24:13  hmh
 * Yacc stacksize set to 1000.
 *

----------------------------------------------------------------------*/

#include "domadt.h"
#include "domfct.h"
#include "domcomment.h"


#ifdef YDEBUG
#define YYRETURN(i) {printf("\n*** lex: %3d, '%s' ***",i,yytext); return(i);}
#else
#define YYRETURN(i) return(i);
#endif

#define YYMAXDEPTH 1000	/* Yacc state stack */
#define yywrap() 1

Dominfo         yyparse_result;
char           *rootfile;
char           *infile;
Bool            yyexitcode;
Str             curr_domid;
static unsigned lineno=1;
static unsigned posline=0;
static int      count=0;

static void errormsg H1(char *);

static Str tolow_str(s)
Str s;
{ Str r = mk_empty_str();
  Nat0 i;

  for (i = 1; i <= len_str(s); i++)
    { r = appr_str(r,tolower(sel_str(s,i))); }
  return r;
}

#define CHECK_ISO	{ if (! ISO_standard)\
{ yyerror("notation allowed in ISO mode only (use '-N' to activate)"); }}

%}

%union {
        int            vnull;
        Str            vstr;
        Dominfo        vdominfo;
        Domain         vdomain;
        Elems          velems;
        Implementation vimpl;
        Operations     voperations;
        Params         vparams;
       }

%token <vnull>       TBAR TTREE TEQUAL TCOLON TLEFTPAR TRIGHTPAR TLEFTBR
%token <vnull>       TRIGHTBR TPLUS TMULT TCOMMA TDOMAINS TIMPL TILLEGALSY
%token <vnull>       T3DOTS TISA
%token <vstr>        TFCT
%token <vstr>        TMAP TSET TBAG TPREDEF TTOKEN TELEMENTARY
%token <vstr>        TMAP1 TMAP2 TTUPL TCOMPOSE
%token <vnull>       TOF TEND
%token <vstr>        TSELECTOR TUSERDEF
%token <vstr>        TTOKID TIDENT TFCTIDENT TSTRCONST TINTCONST
%token <vnull>       TFCTS TCROSS
%type  <vstr>        isoop impval ival domid gdomid tselector
%type  <vdominfo>    start domspecs domspec domeql domeq impleql impleq
%type  <vdomain>     domconstr
%type  <velems>      domidl unioncomp ucomp tokuni treecomp trcomp
%type  <vimpl>       implml
%type  <voperations> implenrs implm
%type  <vparams>     implpars implparams implpar
%type  <vdominfo>    signaturel signature
%type  <velems>      paramtypes fctnames

%%

start       : domspecs
              {
               $$=$1;
               yyparse_result=$1;
              }
            ;

domspecs    : /* empty */
              {$$=mk_empty_dominfo();}
            | domspecs domspec
              {$$=merge_dominfo($1,$2);}
            ;

domspec     : TDOMAINS {domain_comment();} domeql
              {$$=$3;}
            | TIMPL {impl_comment();} impleql
              {$$=$3;}
            | TFCTS {fct_comment();} signaturel
              {$$=$3;}
            | error
              {$$=mk_empty_dominfo();errormsg("Error in DSL-specifikation");}
            ;

domeql      : /* empty */
              {$$=mk_empty_dominfo();}
            | domeql domeq
              {$$=merge_dominfo($1,$2);}
            | domeql error
              {$$=$1;errormsg("Error in domain equation");}
            ;

domeq       : gdomid {specdom_comment($1);} TEQUAL domconstr
              {$$=mk1_dominfo($1,mk_description($4,mk_empty_implementation()));
               update_crossref(curr_domid,infile,rootfile,lineno);}
            | gdomid {specdom_comment($1);} TISA domid
              {$$=mk1_dominfo($1,mk_description(mk_domain
                  (copy_str(ISA),mk1_elems($4),mk_empty_depend()),mk_empty_implementation()));
               update_crossref(curr_domid,infile,rootfile,lineno);}
            | gdomid {specdom_comment($1);} TTREE treecomp
              {$$=mk1_dominfo($1,mk_description(mk_domain
                  (copy_str(TREE),$4,mk_empty_depend()),mk_empty_implementation()));
               update_crossref(curr_domid,infile,rootfile,lineno);}
            | gdomid {specdom_comment($1);} TFCT domid
              {$$=mk1_dominfo($1,mk_description(mk_domain
                  (copy_str(PREDEF),mk_empty_elems(),mk1_depend($4)),
                   mk_empty_implementation()));
               del_str($3);
              }
            ;

domid       : TIDENT
              { if ( islower(hd_str($1)) && ! ISO_standard )
		     { yyerror("lower case domain identifyers allowed in ISO mode only"); }
		$$=$1;
	      }
            ;

gdomid      : domid
              {$$=$1;curr_domid=$1;}
            ;

domconstr   : domid TMULT
              {$$=mk_domain(copy_str(TUPLE),mk1_elems($1),mk_empty_depend());}
            | domid TPLUS
              {$$=mk_domain(copy_str(TUPLE),mk1_elems($1),mk_empty_depend());}
            | TLEFTBR domid TRIGHTBR
              {$$=mk_domain(copy_str(OPTI),mk1_elems($2),mk_empty_depend());}
            | domid TMAP domid
              {$$=mk_domain($2,mk2_elems($1,$3),mk_empty_depend());}
            | unioncomp
              {$$=mk_domain(copy_str(UNION),$1,mk_empty_depend());}
            | tokuni
              {$$=mk_domain(copy_str(TOKUNI),$1,mk_empty_depend());}
            | domid
              {$$=mk_domain(copy_str(ALIAS),mk1_elems($1),mk_empty_depend());}
            | TPREDEF
              {$$=mk_domain($1,mk_empty_elems(),mk_empty_depend());}
            | TTOKEN
              {$$=mk_domain($1,mk_empty_elems(),mk_empty_depend());}
            | TELEMENTARY
              {$$=mk_domain($1,mk_empty_elems(),mk_empty_depend());}
            | domid TUSERDEF
              {$$=mk_domain($2,mk1_elems($1),mk_empty_depend());}
            | TLEFTPAR domidl TRIGHTPAR TUSERDEF
              {$$=mk_domain($4,$2,mk_empty_depend());}
            | domid TFCT domid
              {$$=mk_domain($2,mk2_elems($1,$3),mk_empty_depend());}
            | TLEFTPAR domidl TRIGHTPAR TFCT domid
              {$$=mk_domain($4,appr_elems($2,$5),mk_empty_depend());}

	      /* ISO Standard Extensions */

            | isoop TOF domid
              {CHECK_ISO; $$=mk_domain($1,mk1_elems($3),mk_empty_depend());}

            | TMAP1 domid TMAP2 domid
              {CHECK_ISO; $$=mk_domain($1,mk2_elems($2,$4),mk_empty_depend());}
            | TCOMPOSE domid TOF treecomp TEND
              { if (! is_eq_str(curr_domid,$2))
		   { yyerror("domain name in composite type must be the same as type name"); }
		CHECK_ISO; del_str($2);
		$$=mk_domain($1,$4,mk_empty_depend());
	      }
            ;

isoop       : TBAG
            | TSET
            | TTUPL
            ;

domidl      : /* empty */
              {$$=mk_empty_elems();}
            | domidl domid
              {$$=appr_elems($1,$2);}
            ;

unioncomp   : ucomp
            | unioncomp TBAR ucomp
              {$$=conc_elems($1,$3);}
            ;

ucomp       : trcomp
            | T3DOTS
              {$$=mk_empty_elems();}
            | domid
              {$$=appr_elems(mk1_elems(conc_str(appr_str(
                  tolow_str($1),'_'),tolow_str(curr_domid))),$1);}
            ;

tokuni      : TTOKID
              {$$=mk1_elems($1);}
            | tokuni TBAR TTOKID
              {$$=appr_elems($1,$3);}
            ;

treecomp    : /* empty */
              {$$=mk_empty_elems();}
            | treecomp trcomp
              {$$=conc_elems($1,$2);}
            ;

tselector   : TSELECTOR
              {if (ISO_standard)
                 { /* ISO Syntax: take selector as it is */
                   $$=$1;
                 }
               else
                 /* old Syntax: remove "s_" prefix */
               if (len_str($1)>=2 && hd_str($1)=='s' && sel_str($1,2)=='_')
                 { $$=sub_tupd_str($1,3,len_str($1));}
               else
                 { yyerror("selector names must start with 's_' in non ISO mode");
                   $$=$1;
                 }
              }
            ;

trcomp      : tselector /* TCOLON */ domid
              {
               if (extent_selector_names)
                  $1=conc_str(appr_str($1,'_'),tolow_str(curr_domid));
               $$=mk2_elems($1,$2);
	      }
	      ;

impleql     : /* empty */
              {$$=mk_empty_dominfo();}
            | impleql impleq
              {$$=merge_dominfo($1,$2);}
            | impleql error
              {$$=$1;errormsg("Error in implementation equation");}
            ;

impleq      : gdomid {specimp_comment($1);} TEQUAL implml
              {$$=mk1_dominfo($1,mk_description(mk_empty_domain(),$4));
               update_crossref(curr_domid,infile,rootfile,lineno);}
            ;

implml      : TIDENT implpars implenrs
              {Str par1;
               par1 = tolow_str($1);
               del_str($1);
               $$=mk_implementation(mk_opt_implid(par1),$2,$3);
              }
            | implenrs
              {$$=mk_implementation(mk_nil_implid(),mk_empty_params(),$1);}
            | error
              {$$=mk_empty_implementation();
               errormsg("Implementation or operation expected");}
            ;

implenrs    : /* empty */
              {$$=mk_empty_operations();}
            | TPLUS implm implenrs
              {$$=merge_operations($2,$3);}
            ;

implm       : TIDENT implpars
              {
		  $$=mk1_operations($1,$2);
	      }
            | TIDENT TFCTIDENT implpars
              {
		  Str s=copy_str($2);
		  $$=mk1_operations(conc_str($1,s),
				    merge_params($3,
						 mk1_params(copy_str(FCTNAME),
							    sub_tupd_str($2,2,len_str($2)))));
	      }
            ;

implpars    : /* empty */
              {$$=mk_empty_params();}
            | TLEFTPAR TRIGHTPAR
              {$$=mk_empty_params();}
            | TLEFTPAR implparams TRIGHTPAR
              {$$=$2;}
            ;

implparams  : implpar
            | implparams TCOMMA implpar
              {$$=merge_params($1,$3);}
            | error
              {$$=mk_empty_params();errormsg("Error in parameter");}
            ;

implpar     : TTOKID impval
              {$$=mk1_params($1,$2);}
            | TSTRCONST impval
              {$$=mk1_params($1,$2);}
            ;

impval      : /* empty */
              {$$=copy_str(DEF);}
            | TEQUAL ival
              {$$=$2;}
            ;

ival        : TINTCONST
            | TIDENT
            | TTOKID
            | TSTRCONST
            ;

signaturel  : /* empty */
              {$$=mk_empty_dominfo();}
            | signaturel signature
              {$$=merge_dominfo($1,$2);}
            | signaturel error
              {$$=$1;errormsg("Error in function signature");}
            ;

signature   : fctnames /* TCOLON */ paramtypes {
                specfct_comment(hd_elems($1),hd_elems($2));
              } TFCT paramtypes
              {
               $$=mk_sig_dominfo_elems($1,$2,$5,$4);
               del_elems($1);
               del_elems($2);
               del_elems($5);
               del_str($4);
              }
            ;

paramtypes  : domid
              {$$=mk1_elems($1);}
            | paramtypes TCROSS domid
              {$$=appr_elems($1,$3);}
            ;

fctnames    : TSELECTOR
              {$$=mk1_elems($1);}
            | TIDENT TCOLON
              {$$=mk1_elems($1);}
            | TIDENT TCOMMA fctnames
              {$$=appl_elems($1,$3);}
            ;
%%

#include "domscan.c"

void yyerror F1(char *,s)
{
#ifdef NOPOS
 fprintf(stderr,"%s:%i: %s\n",infile,lineno,s);
#else
 fprintf(stderr,"%s:%i:%i: %s\n",infile,lineno,posline,s);
#endif
 count=0;
 yyexitcode=TRUE;
}

static void errormsg F1(char *,s)
{
 if (NOT count)
  fprintf(stderr," %s\n",s);
 count++;
}
