%a 7000
%o 6000
UC        [A-Z]
LC        [a-z]
LCN       [a-z0-9_]
LCS       {LC}{LCN}*
LCNS      {LCN}+
DIDS      ("_!"{UC}{LCN}*)*
UCN       [A-Z0-9_]
DIG       [0-9]
DS        {DIG}+
C	  [A-Za-z]
CN	  [A-Za-z0-9_]
CNS       {C}{CN}*
DIDS1     ("_!"{UC}{CN}*)*
SPACES    [ \t]*
COLON	  {SPACES}":"
STRC      [^\"\n]*
STRS      {STRC}(\\\"?{STRC})*

%{
/*----------------------------------------------------------------------

  $Id: domscan.y,v 1.18 1996/06/18 10:10:46 vdm Exp $

  Aenderungen:

  $Log: domscan.y,v $
  Revision 1.18  1996/06/18 10:10:46  vdm
  comments moved into actions

  Revision 1.17  1996/06/13 08:06:46  vdm
  VDM-SL keywords in lowercase added:
  "types", "functions", "token" and "implementations"
  old keywords still work

  Revision 1.16  1996/04/30 20:24:29  vdm
  extensions for implementing alias as inheritance in C++
  solution: a new constructor -isa with external syntax
  Subtype is-a Basetype
  Semantics for C like "=", but for C++ class SUBTYPE
  becomes a subclass of class BASETYPE
  Problem: constructors are not inherited, so the
  constructors are passed from Basetype to Subtype
  in the generator, therefore a special operation
  constructor.mk??? introduced, wich is copied in the generator
  to the Subtype domain.
  The generation order computation did not work save,
  therefore an explicit dependency graph is constructed, with this
  graph the generation order is computed, old code still there
  function "old_gen_order"

  Revision 1.15  1996/01/15 10:02:04  vdm
  cvs backup

 * Revision 1.14  1994/12/29  16:52:37  vdm
 * new option -X for collecting comments
 * during parsing and writing these comments into the .enr and .pre files
 * for parsing and importing .vdm files into vdmedit
 *
 * Revision 1.13  1994/12/27  20:13:18  vdm
 * union domains extended:
 * a union domain can now be defined step by step with the following syntax
 * Uni = D1 | D2
 * ...
 * Uni = ... | D3 | D4
 * ...
 * Uni = ... | D5
 *
 * Revision 1.12  1994/11/27  19:05:13  vdm
 * new keywords set bag nat nat0 int char bool real rat
 * for ISO syntax
 *
 * Revision 1.11  1994/08/24  14:21:26  vdm
 * new TOKEN syntax <ABC>
 *
 * Revision 1.10  1994/08/13  19:44:06  vdm
 * TDEPEND --> TFCT,
 * new function arrow => for transformation functions
 *
 * Revision 1.9  1994/08/13  11:48:43  vdm
 * syntax of VDM-DSL extended:
 * a FUNCTIONS part added to specify the signature of user defined functions
 * DOMAINS, IMPLEMENTATIONS and FUNCTIONS parts may be mixed arbitrarily
 * the keyword FUNCTIONS may also be writtes as Functions
 * the syntax of a signature is
 *   fct1,...,fctn : D1 x ... x Dm -> R1 x ... x Ro
 * these signatures are implemeted by the com.com.signature operation
 * x is a new keyword
 * lexical syntax of selectors changed: numbers are not longer allowed as the first char
 * in identifiers
 * the --> symbol (TFCT) did not longer work because of the -- comment
 * therfore the symbol substituted by -> (TDEPEND)
 *
 * Revision 1.8  1994/05/01  20:24:29  vdm
 * some warnings eliminated
 *
 * Revision 1.7  1993/11/03  14:16:58  vdm
 * Hans-Martins Aenderungen eingemischt
 *
 * Revision 1.5  1993/07/05  21:34:13  hmh
 * Merging Uwes changes into mine.
 *
 * Revision 1.4  1993/05/28  17:47:32  hmh
 * Tree selector handling corrected.
 *
 * Revision 1.3  1993/05/28  13:26:59  hmh
 * Tree selector names checked to be lower case (no longer
 * guaranteed by scanner).
 *
 * Revision 1.2  1993/05/28  11:03:31  hmh
 * Compiler extended to accept the ISO VDM-SL ASCII syntax. These
 * extensions are activated through the '-N' switch.
 *

----------------------------------------------------------------------*/

#ifndef YY_NEW_FILE
 /* new bison requires a yyrestart function for multiple calls of yyparse */
 /* this version is detected by asking for YY_NEW_FILE macro definition   */
void yyrestart F1(FILE *, input_file)
   {}
#endif

%}

%%

"*"       {yylval.vnull=0; posline+=yyleng; YYRETURN(TMULT);}
"+"       {yylval.vnull=0; posline+=yyleng; YYRETURN(TPLUS);}
"|"       {yylval.vnull=0; posline+=yyleng; YYRETURN(TBAR);}
"::"      {yylval.vnull=0; posline+=yyleng; YYRETURN(TTREE);}
"="       {yylval.vnull=0; posline+=yyleng; YYRETURN(TEQUAL);}
":"       {yylval.vnull=0; posline+=yyleng; YYRETURN(TCOLON);}
"("       {yylval.vnull=0; posline+=yyleng; YYRETURN(TLEFTPAR);}
")"       {yylval.vnull=0; posline+=yyleng; YYRETURN(TRIGHTPAR);}
"["       {yylval.vnull=0; posline+=yyleng; YYRETURN(TLEFTBR);}
"]"       {yylval.vnull=0; posline+=yyleng; YYRETURN(TRIGHTBR);}
","       {yylval.vnull=0; posline+=yyleng; YYRETURN(TCOMMA);}
"..."     {yylval.vnull=0; posline+=yyleng; YYRETURN(T3DOTS);}


"-m->"    {yylval.vstr=copy_str(MAP); posline+=yyleng; YYRETURN(TMAP);}
"->"      {yylval.vstr=mkc_str(yytext); posline+=yyleng; YYRETURN(TFCT);}
"=>"      {yylval.vstr=mkc_str(yytext); posline+=yyleng; YYRETURN(TFCT);}


"bool"    {/* VDM-SL basic types */
           yylval.vstr=mkc_str("Bool"); posline+=yyleng; YYRETURN(TIDENT);}
"nat"     {yylval.vstr=mkc_str("Nat0"); posline+=yyleng; YYRETURN(TIDENT);}
"nat1"    {yylval.vstr=mkc_str("Nat1"); posline+=yyleng; YYRETURN(TIDENT);}
"int"     {yylval.vstr=mkc_str("Intg"); posline+=yyleng; YYRETURN(TIDENT);}
"rat"     {yylval.vstr=mkc_str("Rat");  posline+=yyleng; YYRETURN(TIDENT);}
"real"    {yylval.vstr=mkc_str("Real"); posline+=yyleng; YYRETURN(TIDENT);}
"char"    {yylval.vstr=mkc_str("Char"); posline+=yyleng; YYRETURN(TIDENT);}
"token"   {yylval.vstr=copy_str(TOKEN); posline+=yyleng; YYRETURN(TTOKEN);}


"map"	  {/* VDM-SL keywords for type constructors */
           yylval.vstr=copy_str(MAP);   posline+=yyleng; YYRETURN(TMAP1);}
"to"	  {yylval.vstr=copy_str(MAP);   posline+=yyleng; YYRETURN(TMAP2);}
"set"     {yylval.vstr=copy_str(SET);   posline+=yyleng; YYRETURN(TSET);}
"bag"     {yylval.vstr=copy_str(BAG);   posline+=yyleng; YYRETURN(TBAG);}
"seq"     {yylval.vstr=copy_str(TUPLE); posline+=yyleng; YYRETURN(TTUPL);}
"seq1"    {yylval.vstr=copy_str(TUPLE); posline+=yyleng; YYRETURN(TTUPL);}
"compose" {yylval.vstr=copy_str(TREE);  posline+=yyleng; YYRETURN(TCOMPOSE);}

"of"      {yylval.vnull=0; posline+=yyleng; YYRETURN(TOF);}
"end"     {yylval.vnull=0; posline+=yyleng; YYRETURN(TEND);}
"x"       {yylval.vnull=0; posline+=yyleng; YYRETURN(TCROSS);}


"is-a"    {/* type constructor for C++ is-a: subclasses */
           yylval.vnull=0; posline+=yyleng; YYRETURN(TISA);}


"DOMAINS"         {/* section delimiter */
                   yylval.vnull=0; posline+=yyleng; YYRETURN(TDOMAINS);}
"Types"	          {yylval.vnull=0; posline+=yyleng; YYRETURN(TDOMAINS);}
"types"	          {yylval.vnull=0; posline+=yyleng; YYRETURN(TDOMAINS);}

"IMPLEMENTATIONS" {yylval.vnull=0; posline+=yyleng; YYRETURN(TIMPL);}
"Implementations" {yylval.vnull=0; posline+=yyleng; YYRETURN(TIMPL);}
"implementations" {yylval.vnull=0; posline+=yyleng; YYRETURN(TIMPL);}

"FUNCTIONS"       {yylval.vnull=0; posline+=yyleng; YYRETURN(TFCTS);}
"Functions"       {yylval.vnull=0; posline+=yyleng; YYRETURN(TFCTS);}
"functions"       {yylval.vnull=0; posline+=yyleng; YYRETURN(TFCTS);}


"PREDEF"     {/* old or internal keywords */
              yylval.vstr=copy_str(PREDEF);  posline+=yyleng; YYRETURN(TPREDEF);    }
"TOKEN"      {yylval.vstr=copy_str(TOKEN);   posline+=yyleng; YYRETURN(TTOKEN);     }
"ELEMENTARY" {yylval.vstr=copy_str(NUMERIC); posline+=yyleng; YYRETURN(TELEMENTARY);}


"-"{LCS}  {
           unsigned i;
           posline+=yyleng;
           for (i=1;i<=yyleng;i++)
            yytext[i-1]=yytext[i];
           yyleng--;
           yylval.vstr=mkc_str(yytext);
           YYRETURN(TUSERDEF);
          }


{LCS}{COLON} {
           posline+=yyleng;
	   yyleng--;
	   while (yytext[yyleng - 1] == ' ' || yytext[yyleng - 1] == '\t')
		 { yyleng--; }
	   yytext[yyleng] = 0;
           yylval.vstr=mkc_str(yytext);
	   YYRETURN(TSELECTOR);
	  }

{CNS}     {
           yylval.vstr=mkc_str(yytext);
           posline+=yyleng;
	   { int i;
	     for (i = 0; i <= yyleng; i++)
		 { if (isalpha(yytext[i]) && islower(yytext[i]))
		      { YYRETURN(TIDENT); }
		 }
	   }
           YYRETURN(TTOKID);
          }

"<"{UC}{UC}*">" {
           unsigned i;
           posline+=yyleng;

           for(i=1;i<=yyleng;i++) yytext[i-1]=yytext[i];

           yyleng-=2;
           yytext[yyleng] = 0;

           yylval.vstr=mkc_str(yytext);
           if (! ISO_standard) { yyerror("ISO token id not allowed here"); }

           YYRETURN(TTOKID);
          }

"."{CNS}  {
           yylval.vstr=mkc_str(yytext);
           posline+=yyleng;
           YYRETURN(TFCTIDENT);
          }

"."{DS}   {posline+=yyleng;}

"!"[^!\n]*"!" { /* old comment */ posline+=yyleng;}
"//".*$    { /* c++ comment */ }
"--".*$	   { /* VDM-SL ISO Standard comment */
            insert_comment(yytext);
           }
";"	   { /* skip ';' used by ISO standard */ }

"#line"[ \t]*{DIG}+\n {
           posline=0;
           if (sscanf(yytext,"#line %d",&lineno) != 1)
            yyerror("unknown CPP line (1):");
          }

"#"[ \t]*{DIG}+\n {
           posline=0;
           if (sscanf(yytext,"# %d",&lineno) != 1)
            yyerror("unknown CPP line (1):");
          }

"#line"[ \t]*{DIG}+.+\n {
           posline=0;
           if (sscanf(yytext,"#line %d %s",&lineno,infile) !=2)
            yyerror("unknown CPP line (2):");
           {
            unsigned i=0;
            while(infile[++i])
             infile[i-1]=infile[i];
            infile[i-2]='\0';
           }
#ifdef YDEBUG
            printf("new values %d %s\n",lineno,infile);
#endif
          }

"#"[ \t]*{DIG}+.+\n {
           posline=0;
           if (sscanf(yytext,"# %d %s",&lineno,infile) != 2)
            yyerror("unknown CPP line (2):");
           {
            unsigned i=0;
            while(infile[++i])
             infile[i-1]=infile[i];
            infile[i-2]='\0';
           }
#ifdef YDEBUG
            printf("new values %d %s\n",lineno,infile);
#endif
          }

"!"[^!\n]*\n {
           posline=0;
           lineno++;
           yyerror("new line in comment not allowed");
           YYRETURN(TILLEGALSY);
          }

\"{STRS}\" {int i,j;                         
            posline+=yyleng;
            for(i=1,j=0;i<yyleng;i++,j++) {
               if (yytext[i] == '\\' && yytext[i+1] == '"')
                  j--;
               else
                  yytext[j]=yytext[i];
            }
            yytext[j-1]=0;
            yyleng =j-1;
            yylval.vstr = mkc_str(yytext);
            YYRETURN(TSTRCONST);
           }

\"{STRS}\n {
            posline+=yyleng;
            yyerror("new line in string constant not allowed");
            YYRETURN(TILLEGALSY);
           }

{DS}      {
           posline+=yyleng;
           yylval.vstr=mkc_str(yytext);
           YYRETURN(TINTCONST);
          }

{SPACES}  {posline+=yyleng;}

\14       {posline+=yyleng;}
\15       {lineno++; posline=0;}
\n        {lineno++; posline=0;}
.         {yyerror("illegal symbol"); YYRETURN(TILLEGALSY);}

%%
