/*----------------------------------------------------------------------

 Modulname: $Id: domsrc.c,v 1.15 1996/09/07 15:12:53 vdm Exp $ 

 Aufgabe  : Steuerung der Modul-Erzeugung 

----------------------------------------------------------------------
 Parameter 

 Eingabe  : Aufrufparameter des Generators 

 Ausgabe  : Returncode 

----------------------------------------------------------------------

 Ersteller: Kai Steinau 

----------------------------------------------------------------------

Aenderungen:

  $Log: domsrc.c,v $
  Revision 1.15  1996/09/07 15:12:53  vdm
  test output removed

  Revision 1.14  1996/09/07 14:12:05  vdm
  vdm as client

  Revision 1.13  1996/01/15 10:02:05  vdm
  cvs backup

 * Revision 1.12  1994/12/29  16:52:38  vdm
 * new option -X for collecting comments
 * during parsing and writing these comments into the .enr and .pre files
 * for parsing and importing .vdm files into vdmedit
 *
 * Revision 1.11  1994/09/25  18:00:38  vdm
 * conditionally file.h or file.hpp included in file.c
 *
 * Revision 1.10  1994/02/28  15:42:58  vdm
 * cvs backup
 *
 * Revision 1.9  1994/02/10  10:36:41  vdm
 * option -J for adding C++ default operations
 *
 * Revision 1.8  1993/11/28  09:54:34  vdm
 * Options sortiert
 *
 * Revision 1.7  1993/11/21  20:03:55  vdm
 * Torsten Tews 1.Version eingemischt
 * Dabei interne Domain Gleichungen in C++ ausgeblendet
 * mit einem internen Implementierungsparameter INTERALDOMAIN
 *
 * Revision 1.6  1993/11/03  14:16:59  vdm
 * Hans-Martins Aenderungen eingemischt

 * Revision 1.5.2.1  1993/11/20  17:20:32  vdm
 * 1.Version von Torsten Tews
 *
 * Revision 1.13  1993/10/29  21:21:55  hmh
 * Several corrections dealing with storage allocation:
 *
 * Revision 1.12  1993/10/24  21:58:31  hmh
 * Changes built in to handle the '-r' option used to check
 * if a given enrichment file can be used as starting point of
 * enrichment process.

 * Revision 1.5.2.1  1993/11/17  21:03:47  uvdm
 * Aenderungen fuer automatische Speicherverwaltung eingebaut; der
 * eingelesene Parsebaum wird in domgen.c transformiert. Aenderungen sind mir
 * / * TT * / markiert. Auswahl der Transformation ueber neuen Kommandozeilen-
 * parameter -M
 *
 * Revision 1.5  1993/09/01  18:11:22  vdm
 * Hans-Martins Version vom 1.9.93 als neue Hauptversion
 *
 * - trace messages generated using 'trace*'.
 * - new option '-r' introduced

 * Revision 1.4  1993/07/01  14:10:28  vdm
 * Uwes changes for C++ inserted
 * Naming conventions for selector names
 * "for" routine interfaces removed
 *
 * Revision 1.11  1993/08/10  15:50:05  hmh
 * Further changes resulting from VDMGEN
 *
 * - All components gnerated with comments 'PHASE<n>' to allow 
 *   rearrangement using awk.

 * Revision 1.10  1993/08/07  05:59:56  hmh
 * Further changes neede for 'VDMGEN':
 * - DOMAINLIST evaluated to get include filenames in 'gen_c_header_files'

 * Revision 1.9  1993/08/06  07:15:26  hmh
 * Changes needed for VDMGEN.
 * - New function 'gen_c_header_files' to generate C headerfile
 *   seperately, activated through '-H' option.
 *   All includes put into appropriate conditional compilation to
 *   avoid unneccessary includes.
 * - '-X' option removed (no longer needed)

 * Revision 1.8  1993/07/25  17:39:38  hmh
 * typing error corrected
 *
 * Revision 1.7  1993/07/25  12:41:24  hmh
 * Initialized decls removed
 *
 * Revision 1.6  1993/07/25  08:51:05  hmh
 * Several changes
 * - New option '-X' generating includes for all component domains 
 *   upon generation of C header file.
 * - in 'gen_c_header' handling of '-X' option built in.
 * - New option '-s', function 'split_enr_file' called depending on it.

 * Revision 1.5  1993/07/10  07:29:31  hmh
 * Handling of 'VDMROOT' environment variable corrected:
 *   - removed reading from domaincomp ('vdmmain') completely
 *   - in Dm: if VDMROOT is undefined, set to default path ('/usr').
 *
 * Revision 1.4  1993/07/05  21:34:15  hmh
 * Merging Uwes changes into mine.
 *
 * Revision 1.4  1993/07/01  14:10:28  vdm
 * Uwes changes for C++ inserted
 * Naming conventions for selector names
 * "for" routine interfaces removed
 *
 * Revision 1.3  1993/05/28  11:03:33  hmh
 * Compiler extended to accept the ISO VDM-SL ASCII syntax. These
 * extensions are activated through the '-N' switch.
 *
 * Revision 1.2  1993/05/22  15:47:27  hmh
 * Condtional writing of '.enr' and '.pre' files caused problems
 * together with traditional 'make' dependencies.
 * Command line parameter '-E' introduced to enable this feature
 * if needed (i.e. by 'VDM'), disabled by default.
 *
  93/02/11: System-Includes (#include <...>) nach domfct.h 

----------------------------------------------------------------------*/

#include "domadt.h"
#include "domfct.h"
#include "domgen.h"
#include "domfile.h"
#include "domsrc.h"
#include "domcomment.h"

/*----------------------------------------------------------------------*/


/*----------------------------------------------------------------------
 Modulname: set_global_param 
 Aufgabe  : Aufnahme einer Parameter-Definition in die Abbildung der 
            globalen Parameter 'global_params' 
----------------------------------------------------------------------
 Parameter 
 Eingabe  : Parameter-Name 'p' 
            Parameter-Wert 'v'
 Ausgabe  : 
----------------------------------------------------------------------*/

void set_global_param F2(
      Str, p,
      char *,v)
{
    if (NOT is_in_params(p,global_params))
	global_params=union_params(global_params,mk1_params(copy_str(p),mkc_str(v)));
    else
	{
	    Str s=copy_str(sel_params(global_params,p));
	    s=conc3_str(s,mkc_str("\n"),mkc_str(v));
	    global_params=ovwrt1d_params(global_params,copy_str(p),s);
	}
}


/*----------------------------------------------------------------------
 Modulname: include_name 
 Aufgabe  : Umwandeln eines VDM-Strings in eine fuer eine Include- 
            Direktive zulaessige Form 
----------------------------------------------------------------------
 Parameter 
 Eingabe  : VDM-String (Dateiname) 's' 
 Ausgabe  : VDM-String (Dateiname in fuer Include zulaessige Form) 
----------------------------------------------------------------------*/

Str include_name F1(Str, s)
{
    Str o=copy_str(s);
    Nat0 i;
    for(i=1;i<len_str(o);i++)
	if (NOT isalnum(sel_str(o,i)))
	    ovwrt1_str(o,i,'_');
    return o;
}


/*----------------------------------------------------------------------
 Modulname: extend_modul 
 Aufgabe  : Erzeugung einer Komponente eines Quellcode-Moduls 
----------------------------------------------------------------------
 Parameter 
 Eingabe  : Parse-Baum 'dominfo' 
            Parameter zur Erzeugung der Komponente des Moduls 'p' 
            Name des Parameters des Typs des Moduls 'def' 
            Typ des Moduls 'type' 
 Ausgabe  : VDM-String der Komponente des Moduls 
----------------------------------------------------------------------*/

Str extend_modul F4(
		    Dominfo, dominfo,
		    Params, p,
		    Str, def,
		    Str, type)
{
    Params params=union1_params(copy_params(p),copy_str(def),copy_str(DEF));
    return
	extend_dominfo(dominfo,type,params);
}


/*----------------------------------------------------------------------
 Modulname: gen_modul 
 Aufgabe  : Steuerung des Aufrufs der Prozeduren zur Modulgenerierung 
----------------------------------------------------------------------
 Parameter 
 Eingabe  : Parse-Baum 'dominfo' 
            Name der Ausgabe-Datei 'ofile' 
            Name des Parameters fuer das entsprechende Modul 'pn' 
            Fehlercode 'errcode' 
            Zeiger auf aufzurufende Prozedur 'fct'
 Ausgabe  : Fehlercode oder 0 fuer kein Fehler 
----------------------------------------------------------------------*/

int gen_modul F5(
		 Dominfo, dominfo,
		 Str, ofile,
		 Str, pn,
		 int, errcode,
		 fctptr, fct)
{
    if ((NOT yyexitcode) AND (is_in_params(pn,global_params)))
	fct(dominfo,ofile);
    return
	IF (yyexitcode) THEN errcode ELSE 0;
}


/*----------------------------------------------------------------------
 Modulname: gen_c_impl_file 
 Aufgabe  : Erzeugen des C-Implementations-Moduls 
----------------------------------------------------------------------
 Parameter 
 Eingabe  : Parse-Baum 'dominfo' 
            Name der Ausgabe-Datei 'file_name' 
 Ausgabe  : 
----------------------------------------------------------------------*/

void gen_c_impl_file F2(
			Dominfo, dominfo,
			Str, file_name)
{
    Params params;       /* Parameter zur Erzeugung der Komponenten des Moduls */
    Str fn=gen_file_name(file_name,mkc_str("c"));     /* Name der Ausgabedatei */
    Str s=mk_empty_str();               /* String zur Aufnahme der Komponenten */
    trace1(mkc_str("\nGenerating C implementation file"),Empty);

    s=conc_str(s,mkc_str("#if __cplusplus\n"));
    s=conc4_str(s,mkc_str("#include \""),copy_str(file_name),mkc_str(".hpp\"\n"));
    s=conc_str(s,mkc_str("#else\n"));
    s=conc4_str(s,mkc_str("#include \""),copy_str(file_name),mkc_str(".h\"\n"));
    s=conc_str(s,mkc_str("#endif\n\n"));

    s=conc_str(s,mkc_str("/*LINTLIBRARY*/\n\n"));
    s=conc_str(s,mkc_str("#define C_implementation_file\n\n"));

    params=mk1_params(copy_str(GEN_C),copy_str(DEF));
    s=conc_str(s,extend_modul(dominfo,params,CROUTINES,CCOD));
    del_params(params);
    if (NOT fwrite_str(s,fn))
	error(mkc_str("Can't write C implementation file  "),copy_str(fn));
    del_str(fn);
    del_str(s);
}


/*----------------------------------------------------------------------
 Modulname: gen_c_header_files
 Aufgabe  : Erzeugen von drei C-Deklarations-Modulen '.h1', '.h2'
	    und '.h', die die 'gptr', 'gtyp' bzw. 'gsig' und 'cmcr'
	    Komponenten enthalten.
	    Es werden #include statements fuer alle Teildomains erzeugt.
----------------------------------------------------------------------
 Parameter 
 Eingabe  : Parse-Baum 'dominfo' 
            Name der Ausgabe-Dateien 'file_name' 
 Ausgabe  : 
----------------------------------------------------------------------*/

static Params domainmap;

static Bool read_domainlist F0()
{
    FILE *fd;
    char fn[200];
    char dn[200];
    trace1(mkc_str("\nReading DOMAINLIST "),Empty);

    domainmap = mk_empty_params();
    if ( (fd = fopen("DOMAINLIST","r")) == NULL)
	{
	    error(mkc_str("cannot open "), mkc_str("DOMAINLIST"));
	    return
		FALSE;
	}

    while(fscanf(fd,"%s %s",fn,dn) == 2)
      {
	  if (domtrace > 1)
	      {
		  printf("--- %s -> %s\n",dn,fn);
	      }
	  domainmap = union1_params(domainmap,mkc_str(dn),mkc_str(fn));
      }

    fclose(fd);
    return
	TRUE;
  }

/*------------------------------------------------------------------*/

Str coll_includes_str F2(Str,dn,Str,ext)
{
    Str r;
    r = conc5_str(mkc_str("#ifndef VDM_header")
		  ,copy_str(ext)
		  ,mkc_str("_")
		  ,copy_str(sel_params(domainmap,dn))
		  ,mkc_str("\n"));
    r =  conc6_str(r
		   ,mkc_str("#include \"")
		   ,copy_str(sel_params(domainmap,dn))
		   ,mkc_str(".")
		   ,copy_str(ext)
		   ,mkc_str("\"\n#endif\n")); 
    return r;
  }

Str gen_includes_str_description F3(Str, dn, Description, dd, Str, ext)
{
    if (is_eq_str(s_constr(s_domain(dd)), PREDEF))
	{
	    return mkc_str("");
	}
    {
	Domain d;
	Depend dp;
	Str res;
	d = s_domain(dd);
	dp = union_depend(copy_depend(s_depend(d))
			  ,get_elems(s_elems(d),s_constr(d)));
	res = coll_includes_depend(dp,ext);
	del_depend(dp);
	return
	    res;
    }
}
   

/*--------------------------------------------------------------------*/

void gen_c_header_files F2(
			   Dominfo, dominfo,
			   Str, file_name)
{
    Params params;
    Str fn1,fn2,fn;
    Str s;
    trace1(mkc_str("\nGenerating C header fileset "),Empty);

    if (read_domainlist() == FALSE)
	{
	    return;
	}

    params=mk1_params(copy_str(GEN_C),copy_str(DEF));
    fn1=gen_file_name(file_name,mkc_str("h1"));
    fn2=gen_file_name(file_name,mkc_str("h2"));
    fn=gen_file_name(file_name,mkc_str("h"));

    /*- .h1 -*/
    s=mk_empty_str();
    s=conc3_str(s,mkc_str("#ifndef VDM_headerh1_"),include_name(file_name));
    s=conc4_str(s
		,mkc_str("\n#define VDM_headerh1_")
		,include_name(file_name)
		,mkc_str("\n\n"));

    s = conc_str(s,gen_includes_dominfo(dominfo,mkc_str("h1"))); 

    s=conc_str(s,extend_modul(dominfo,params,PTRTYPES,GPTR));
    s=conc_str(s,mkc_str("\n#endif\n"));
    if (NOT fwrite_str(s,fn1))
	error(mkc_str("Can't write C header file  "),copy_str(fn1));
    del_str(s);

    /*- .h2 -*/
    s=mk_empty_str();
    s=conc3_str(s,mkc_str("#ifndef VDM_headerh2_"),include_name(file_name));
    s=conc4_str(s
		,mkc_str("\n#define VDM_headerh2_")
		,include_name(file_name)
		,mkc_str("\n\n"));

    s = conc_str(s,gen_includes_dominfo(dominfo,mkc_str("h2"))); 
    s = conc6_str(s
		  ,mkc_str("\n#ifndef VDM_headerh1_")
		  ,include_name(file_name)
		  ,mkc_str("\n#include \"")
		  ,copy_str(fn1)
		  ,mkc_str("\"\n#endif\n"));
    s=conc_str(s,extend_modul(dominfo,params,TYPES,GTYP));
    s=conc_str(s,mkc_str("\n#endif\n"));
    if (NOT fwrite_str(s,fn2))
	error(mkc_str("Can't write C header file  "),copy_str(fn2));
    del_str(s);

    /*- .h -*/
    s=mk_empty_str();
    s=conc3_str(s,mkc_str("#ifndef VDM_headerh_"),include_name(file_name));
    s=conc4_str(s,mkc_str("\n#define VDM_headerh_"),include_name(file_name),
		mkc_str("\n\n"));

    s = conc_str(s,gen_includes_dominfo(dominfo,mkc_str("h"))); 
    s = conc6_str(s
		  ,mkc_str("\n#ifndef VDM_headerh2_")
		  ,include_name(file_name)
		  ,mkc_str("\n#include \"")
		  ,copy_str(fn2)
		  ,mkc_str("\"\n#endif\n"));
    s=conc_str(s,extend_modul(dominfo,params,SIGNATURE,GSIG));
    s=conc_str(s,extend_modul(dominfo,params,CMACROS,CMCR));

    s=conc_str(s,mkc_str("\n#endif\n"));
    if (NOT fwrite_str(s,fn))
	error(mkc_str("Can't write C header file  "),copy_str(fn));
    del_str(s);


    del_str(fn1); del_str(fn2); del_str(fn);
    del_params(params);
    del_params(domainmap);
}

/*----------------------------------------------------------------------
 Modulname: gen_c_header_file 
 Aufgabe  : Erzeugen des C-Deklarations-Moduls 
----------------------------------------------------------------------
 Parameter 
 Eingabe  : Parse-Baum 'dominfo' 
            Name der Ausgabe-Datei 'file_name' 
 Ausgabe  : 
----------------------------------------------------------------------*/

void gen_c_header_file F2(
			  Dominfo, dominfo,
			  Str, file_name)
{
    Params params;       /* Parameter zur Erzeugung der Komponenten des Moduls */
    Str fn=gen_file_name(file_name,mkc_str("h"));     /* Name der Ausgabedatei */
    Str s=mk_empty_str();               /* String zur Aufnahme der Komponenten */
    trace1(mkc_str("\nGenerating C header file "),Empty);
    s=conc3_str(s,mkc_str("#ifndef VDM_header_"),include_name(file_name));
    s=conc4_str(s,mkc_str("\n#define VDM_header_"),include_name(file_name),
		mkc_str("\n\n"));
    params=mk1_params(copy_str(GEN_C),copy_str(DEF));
    s=conc_str(s,mkc_str("\n/* ------- PHASE1 ------------------ */\n"));
    s=conc_str(s,extend_modul(dominfo,params,PTRTYPES,GPTR));
    s=conc_str(s,mkc_str("\n/* ------- PHASE2 ------------------ */\n"));
    s=conc_str(s,extend_modul(dominfo,params,TYPES,GTYP));
    s=conc_str(s,mkc_str("\n/* ------- PHASE3 ------------------ */\n"));
    s=conc_str(s,extend_modul(dominfo,params,SIGNATURE,GSIG));
    s=conc_str(s,mkc_str("\n/* ------- PHASE4 ------------------ */\n"));
    s=conc_str(s,extend_modul(dominfo,params,CMACROS,CMCR));
    s=conc_str(s,mkc_str("\n#endif\n"));
    del_params(params);
    if (NOT fwrite_str(s,fn))
	error(mkc_str("Can't write C header file  "),copy_str(fn));
    del_str(fn);
    del_str(s);
}


/*----------------------------------------------------------------------
 Modulname: gen_p_header_file 
 Aufgabe  : Erzeugen des Pascal-Deklarations-Moduls 
----------------------------------------------------------------------
 Parameter 
 Eingabe  : Parse-Baum 'dominfo' 
            Name der Ausgabe-Datei 'file_name' 
 Ausgabe  : 
----------------------------------------------------------------------*/

void gen_p_header_file F2(
			  Dominfo, dominfo,
			  Str, file_name)
{
    Params params;       /* Parameter zur Erzeugung der Komponenten des Moduls */
    Str fn=gen_file_name(file_name,mkc_str("defs"));  /* Name der Ausgabedatei */
    Str s=mk_empty_str();               /* String zur Aufnahme der Komponenten */
    trace1(mkc_str("\nGenerating Pascal prolog file "),Empty);
    s=conc4_str(s,mkc_str("CONST\n   PROLOG_"),include_name(file_name),
		mkc_str(" = 0;\n"));
    params=mk1_params(copy_str(GEN_PASCAL),copy_str(DEF));
    s=conc_str(s,mkc_str("\n(* ------- PHASE1 ------------------ *)\n"));
    s=conc4_str(s,extend_modul(dominfo,params,CONSTS,PCST),
		mkc_str("TYPE\n   dummy_type_"),include_name(file_name));
    s=conc_str(s,mkc_str(" = RECORD END;\n"));
    s=conc_str(s,mkc_str("\n(* ------- PHASE2 ------------------ *)\n"));
    s=conc_str(s,extend_modul(dominfo,params,PTRTYPES,GPTR));
    s=conc_str(s,mkc_str("\n(* ------- PHASE3 ------------------ *)\n"));
    s=conc_str(s,extend_modul(dominfo,params,TYPES,GTYP));
    s=conc_str(s,mkc_str("\n(* ------- PHASE4 ------------------ *)\n"));
    s=conc_str(s,extend_modul(dominfo,params,SIGNATURE,GSIG));
    s=conc_str(s,mkc_str("\n/* ------- PHASEEND ------------------ */\n"));
    del_params(params);
    if (NOT fwrite_str(s,fn))
	error(mkc_str("Can't write Pascal prolog file  "),copy_str(fn));
    del_str(fn);
    del_str(s);
}


/*----------------------------------------------------------------------
 Modulname: gen_m_header_file 
 Aufgabe  : Erzeugen des Modula-2 Deklarations-Moduls 
----------------------------------------------------------------------
 Parameter 
 Eingabe  : Parse-Baum 'dominfo' 
            Name der Ausgabe-Datei 'file_name' 
 Ausgabe  : 
----------------------------------------------------------------------*/

void gen_m_header_file F2(
			  Dominfo, dominfo,
			  Str, file_name)
{
    Params params;       /* Parameter zur Erzeugung der Komponenten des Moduls */
    Str fn=gen_file_name(file_name,mkc_str("def"));   /* Name der Ausgabedatei */
    Str s=mk_empty_str();               /* String zur Aufnahme der Komponenten */
    trace1(mkc_str("\nGenerating Modula definition part "),Empty);
    s=conc4_str(s,mkc_str("INTERFACE MODULE "),copy_str(file_name),
		mkc_str(";\n   CONST DummyConst = 0;\n"));
    params=mk1_params(copy_str(GEN_MODULA),copy_str(DEF));
    s=conc_str(s,mkc_str("\n(* ------- PHASE1 ------------------ *)\n"));
    s=conc_str(s,extend_modul(dominfo,params,CONSTS,GCST));
    s=conc_str(s,mkc_str("\nTYPE\n"));
    s=conc_str(s,mkc_str("\n(* ------- PHASE2 ------------------ *)\n"));
    s=conc_str(s,extend_modul(dominfo,params,PTRTYPES,GPTR));
    s=conc_str(s,mkc_str("\n(* ------- PHASE3 ------------------ *)\n"));
    s=conc_str(s,extend_modul(dominfo,params,TYPES,GTYP));
    s=conc_str(s,mkc_str("\n(* ------- PHASE4 ------------------ *)\n"));
    s=conc_str(s,extend_modul(dominfo,params,SIGNATURE,GSIG));
    s=conc4_str(s,mkc_str("\nEND "),copy_str(file_name),mkc_str("."));
    del_params(params);
    if (NOT fwrite_str(s,fn))
	error(mkc_str("Can't write Modula definition part  "),copy_str(fn));
    del_str(fn);
    del_str(s);
}


/*----------------------------------------------------------------------
 Modulname: gen_m_impl_file 
 Aufgabe  : Erzeugen der Modula-2 Import-Statements 
----------------------------------------------------------------------
 Parameter 
 Eingabe  : Parse-Baum 'dominfo' 
            Name der Ausgabe-Datei 'file_name' 
 Ausgabe  : 
----------------------------------------------------------------------*/

void gen_m_impl_file F2(
			Dominfo, dominfo,
			Str, file_name)
{
    Params params;       /* Parameter zur Erzeugung der Komponenten des Moduls */
    Str fn=gen_file_name(file_name,mkc_str("mimp"));  /* Name der Ausgabedatei */
    Str s=mk_empty_str();               /* String zur Aufnahme der Komponenten */
    trace1(mkc_str("\nGenerating Modula import definitions"),Empty);
    s=conc4_str(s,mkc_str("FROM "),copy_str(file_name),mkc_str(" IMPORT\n"));
    params=mk1_params(copy_str(GEN_IMPORT),copy_str(DEF));
    s=conc_str(s,mkc_str("\n(* ------- PHASE1 ------------------ *)\n"));
    s=conc_str(s,extend_modul(dominfo,params,IMPORTS,GIMP));
    s=conc_str(s,mkc_str("\n(* ------- PHASE2 ------------------ *)\n"));
    s=conc_str(s,extend_modul(dominfo,params,SIGNATURE,GSIG));
    s=conc_str(s,mkc_str("   DummyConst;"));
    del_params(params);
    if (NOT fwrite_str(s,fn))
	error(mkc_str("Can't write Modula import definitions  "),copy_str(fn));
    del_str(fn);
    del_str(s);
}


/*----------------------------------------------------------------------
 Modulname:  gen_cpp_header_file 
 Aufgabe  :  Erzeugen der C++-Deklarationsdatei 
----------------------------------------------------------------------
 Parameter 
 Eingabe  : Parse-Baum 'dominfo'
            Name der Ausgabe-Datei 'file_name' 
 Ausgabe  : 
----------------------------------------------------------------------*/

void gen_cpp_header_file F2(
			    Dominfo, dominfo,
			    Str, file_name)
{
    Params params;       /* Parameter zur Erzeugung der Komponenten des Moduls */
    Str fn=gen_file_name(file_name,mkc_str("hpp"));   /* Name der Ausgabedatei */
    Str s=mk_empty_str();               /* String zur Aufnahme der Komponenten */
    trace1(mkc_str("\nGenerating C++ header file "),Empty);
    s=conc3_str(s,mkc_str("#ifndef CPP_header_"),include_name(file_name));
    s=conc3_str(s,mkc_str("\n#define CPP_header_"),copy_str(file_name));
    params=mk1_params(copy_str(GEN_C),copy_str(DEF));
    params=union1_params(params,copy_str(GEN_CPLUS),copy_str(DEF));
    params=union1_params(params,copy_str(CPLUSPLUS),copy_str(DEF));
    if ( NOT friend_functions )
	params=union1_params(params,copy_str(MEMBERS),copy_str(DEF));
    if ( operator_functions )
	params=union1_params(params,copy_str(OPERATOR),copy_str(DEF));
    if ( upper_case_class_names )
	params=union1_params(params,copy_str(UPPER),copy_str(DEF));
    s=conc_str(s,mkc_str("\n/* ------- PHASE1 ------------------ */\n"));
    s=conc4_str(s
		,mkc_str("\n#include \"")
		,include_name(file_name)
		,mkc_str(".h\"\n\n"));
    s=conc_str(s,extend_modul(dominfo,params,DECL,CPPCLASS));
    s=conc_str(s,mkc_str("\n/* ------- PHASE2 ------------------ */\n"));
    s=conc_str(s,extend_modul(dominfo,params,DECL,CPP));
    params=ovwrt1_params(params,copy_str(DECL),Empty);
    s=conc_str(s,mkc_str("\n/* ------- PHASE3 ------------------ */\n"));
    s=conc_str(s,extend_dominfo(dominfo,CPP,params));
    s=conc_str(s,mkc_str("\n/* ------- PHASEEND ------------------ */\n"));
    s=conc_str(s,mkc_str("\n#endif\n"));
    if (NOT fwrite_str(s,fn))
	error(mkc_str("Can't write C++ header file  "),copy_str(fn));
    del_str(fn);
    del_str(s);
}


/*----------------------------------------------------------------------
 Modulname:  vdmmain 
 Aufgabe  :  Schnittstelle zur Steuerung der Modul-Erzeugung 
----------------------------------------------------------------------
 Parameter 
 Eingabe  :  Anzahl der uebergebenen Argumente 'argc' 
             Liste der uebergebenen Argumente 'argv' 
 Ausgabe  :  Returncode 
----------------------------------------------------------------------*/

int vdmmain F2(
	       int, argc,
	       char **,argv)
{
    int exitcode=0;
    char *p;
    int i=1;
    domtrace=0;
    global_params=mk_empty_params();
    mk_init_str();

    init_comment();	/* initialize comment collect module */

    /* all -S parameters are allowed as Implementation parameters */
    /* without giving an error message */
    /* to distiguish between internal domains and user defined domains */
    /* parameter INTERNALDOMAIN is always allowed */

    set_global_param(PARAM_['S'],INTERNALDOMAINstr);

    /* Verarbeitung der Parameterliste */
    while ((i<argc) && (*(p=argv[i])=='-'))
	{
	    switch (*(++p))
		{
		case 'v' :
		    domtrace=1;
		    break;
		case 't' :
		    domtrace=2;
		    break;
		case 'T' :
		    domtrace=3;
		    break;
		case 'r' :
		    if (++i>=argc)
			error(mkc_str("missing enrichment file name for '-r' option"),Empty);
                    else
			global_params=ovwrt1_params(global_params,copy_str(PARAM_['r']),mkc_str(argv[i]));
                    break;
		case 'o' :
		    if (++i>=argc)
			error(mkc_str("missing output name"),Empty);
                    else
			global_params=ovwrt1_params(global_params,
						    copy_str(PARAM_['o']),mkc_str(argv[i]));
                    break;
		case 'B' :
		    set_global_param(PARAM_['B'],p+1);
		case 'I' :
		    set_global_param(PARAM_['I'],p+1);
                    break;
		case 'D' :
		    set_global_param(PARAM_['D'],p+1);
                    break;
		case 'S' :
		    set_global_param(PARAM_['S'],p+1);
                    break;
		case 'N' :
		    global_params=ovwrt1_params(global_params,
						copy_str(PARAM_[(int)*p]),copy_str(DEF));
                    set_global_param(PARAM_['S'],"GEN_ISO");
                    break;

		case 'M' :
		    /* Automatic memory management selected : */ /* TT */
		    set_global_param(PARAM_['M'],"defined");
                    set_global_param(PARAM_['S'],"AUTOMEM_SEL");
                    break;                    
		case 'a' :
		    global_params=ovwrt1_params(global_params,
						copy_str(PARAM_[(int)*p]),copy_str(DEF));
                    set_global_param(PARAM_['S'],"SELSUFFIX");
                    break;
		case 'c' :
		    if ( strcmp(p,"client") == 0 )
			{
			    extern Bool run_as_client;
			    extern char * host;
			    extern char * port;

			    run_as_client = TRUE;

			    if ( i<argc-1
				 && argv[i+1][0] != '-'
				 && strstr(argv[i+1],".vdm")==NULL)
				{
				    /* set host name */
				    ++i;
				    host = argv[i];

				    if ( i<argc-1
					 && argv[i+1][0] != '-'
					 && strstr(argv[i+1],".vdm")==NULL)
					{
					    /*set port name */
					    ++i;
					    port = argv[i];
					}
				}
			    break;
			}
		case 'd' :
		case 'e' :
		case 'h' :
		case 'i' :
		case 'k' :
		case 'm' :
		case 's' :
		case 'u' :
		case 'C' :
		case 'E' :
		case 'F' :
		case 'H' :
		case 'J' :
		case 'K' :
		case 'O' :
		case 'U' :
		case 'W' :
		    global_params=ovwrt1_params(global_params,
						copy_str(PARAM_[(int)*p]),copy_str(DEF));
		    break;
		case 'X' :
		    set_collect_comment();
                    break;
		default  :
		    fprintf(stderr,"%s: unknown option %s ignored\n",Progname,argv[i]);
		}
	    i++;
	}
    if (NOT open_vdmdbs())
	exitcode=1;   /* Fehler beim Datenbank-ffnen */
    else
	{
	    Dominfo dominfo;
	    Str o;
	    if (is_in_dom_params(PARAM_['o'],global_params,&o))
		o=copy_str(o);
	    else
		o=Basename;
	    if (i<argc)
		{
		    dominfo=gen_dominfo(i,argc,argv);
		    if (yyexitcode) exitcode=3; /* Fehler beim Erstellen des Parse_Baums */
		}
	    else
		{
		    dominfo=mk_empty_dominfo();
		    error(mkc_str("missing file name"),Empty);
		    exitcode=2;                 /* Kein Dateiname gegeben */
		}

	    /* Erzeugen der einzelen Module */
	    exitcode=gen_modul(dominfo,o,PARAM_['s'],4,split_enr_file);
	    exitcode=gen_modul(dominfo,o,PARAM_['e'],4,gen_enr_file);
	    exitcode=gen_modul(dominfo,o,PARAM_['c'],5,gen_c_impl_file);
	    exitcode=gen_modul(dominfo,o,PARAM_['H'],6,gen_c_header_files);
	    exitcode=gen_modul(dominfo,o,PARAM_['h'],6,gen_c_header_file);
	    exitcode=gen_modul(dominfo,o,PARAM_['d'],7,gen_p_header_file);
	    exitcode=gen_modul(dominfo,o,PARAM_['m'],8,gen_m_header_file);
	    exitcode=gen_modul(dominfo,o,PARAM_['i'],9,gen_m_impl_file);
	    exitcode=gen_modul(dominfo,o,PARAM_['C'],10,gen_cpp_header_file);
	    del_dominfo(dominfo);
	    del_str(o);
	}
    if (exitcode != 1)
	delete_vdmdbs();
    del_params(global_params);
    del_init_str();

    profile_all();

    return
	exitcode;
}
