/*----------------------------------------------------------------------

 $Id: domuse.c,v 1.12 1996/09/07 14:12:05 vdm Exp $ 

 Aufgabe  : Schnittstelle des VDM-Generators zum Benutzer 

----------------------------------------------------------------------
 Parameter

 Eingabe  :

 Ausgabe  : 

----------------------------------------------------------------------

 Ersteller: Kai Steinau 

----------------------------------------------------------------------

 Aenderungen:
 
   $Log: domuse.c,v $
   Revision 1.12  1996/09/07 14:12:05  vdm
   vdm as client

   Revision 1.11  1996/01/15 10:02:06  vdm
   cvs backup

 * Revision 1.10  1994/12/29  16:52:41  vdm
 * new option -X for collecting comments
 * during parsing and writing these comments into the .enr and .pre files
 * for parsing and importing .vdm files into vdmedit
 *
 * Revision 1.9  1994/02/10  10:36:43  vdm
 * option -J for adding C++ default operations
 *
 * Revision 1.8  1994/01/12  21:59:07  vdm
 * Torsten Tews Aenderungen vom Dezember eingebaut
 *
 * Revision 1.4.2.2  1993/12/08  21:17:14  uvdm
 * Neue Version des Compilers von Prof. Schmidt, basiert auf meinen
 * Erweiterungen vom 17.11
 * Anpassungen entsprechend der Version vom 20.11 durchgefuehrt
 *
 * Revision 1.7  1993/11/28  09:54:35  vdm
 * Options sortiert
 *
 * Revision 1.6  1993/11/21  20:03:56  vdm
 * Torsten Tews 1.Version eingemischt
 * Dabei interne Domain Gleichungen in C++ ausgeblendet
 * mit einem internen Implementierungsparameter INTERALDOMAIN
 *
 * Revision 1.5  1993/11/03  14:17:00  vdm
 * Hans-Martins Aenderungen eingemischt
 * Revision 1.4.2.1  1993/11/20  17:20:33  vdm
 * 1.Version von Torsten Tews
 *
 * Revision 1.8  1993/10/24  21:58:34  hmh
 * Changes built in to handle the '-r' option used to check
 * if a given enrichment file can be used as starting point of
 * enrichment process.
 * Revision 1.4.2.1  1993/11/17  21:03:50  uvdm
 * Aenderungen fuer automatische Speicherverwaltung eingebaut; der
 * eingelesene Parsebaum wird in domgen.c transformiert. Aenderungen sind mir
 * / * TT * / markiert. Auswahl der Transformation ueber neuen Kommandozeilen-
 * parameter -M
 *
 * new option '-r'

 * Revision 1.4  1993/09/01  18:11:24  vdm
 * Hans-Martins Version vom 1.9.93 als neue Hauptversion
 *
 * Revision 1.3  1993/05/28  11:03:35  hmh
 * Compiler extended to accept the ISO VDM-SL ASCII syntax. These
 * extensions are activated through the '-N' switch.
 *
 * Revision 1.7  1993/08/07  05:59:58  hmh
 * Further changes neede for 'VDMGEN':
 *
 * - Explanation of option '-u' changed.

 * Revision 1.6  1993/08/06  07:15:29  hmh
 * Changes needed for VDMGEN.
 * - New option '-H'
 * - Option '-X' removed

 * Revision 1.5  1993/07/25  08:51:07  hmh
 * Several changes
 * - New option '-s'

 * Revision 1.4  1993/07/05  21:34:16  hmh
 * Merging Uwes changes into mine.
 *
 * Revision 1.3  1993/05/28  11:03:35  hmh
 * Compiler extended to accept the ISO VDM-SL ASCII syntax. These
 * extensions are activated through the '-N' switch.
 *
 * Revision 1.2  1993/05/22  15:47:29  hmh
 * Condtional writing of '.enr' and '.pre' files caused problems
 * together with traditional 'make' dependencies.
 * Command line parameter '-E' introduced to enable this feature
 * if needed (i.e. by 'VDM'), disabled by default.
 *
   93/02/11: System-Includes (#include <...>) nach domfct.h
   93/02/23: include "domdekl.h" geloescht 

----------------------------------------------------------------------*/

#include "domfct.h"
#include "domsrc.h"

/*----------------------------------------------------------------------
 Modulname: main 
 Aufgabe  : Hauptprogramm zum Aufruf des VDM-Generators 
----------------------------------------------------------------------
 Parameter 
 Eingabe  : Anzahl der uebergebenen Argumente 'argc' 
            Liste der Uebergebenen Argumente 'argv' 
 Ausgabe  : Returncode: Verarbeitung fehlerfrei (=0), sonnst (<>0) 
----------------------------------------------------------------------*/

int main F2(
	    int, argc,
	    char **,argv)
{
    int exitcode=0;
    if (argc>1)
	exitcode = vdmmain(argc,argv);
    else
	{
	    fprintf(stderr,"usage: vdm [<options>] file ...\n");
	    fprintf(stderr,"options:\n");
	    fprintf(stderr,"   -v        : print trace output\n");
	    fprintf(stderr,"   -W        : produce warning messages\n");
	    fprintf(stderr,"   -B<vdmdb> : give alternative VDM database\n");
	    fprintf(stderr,"   -Sname=def\n");
	    fprintf(stderr,"   -Sname    : set global implementation parameter\n");
	    fprintf(stderr,"   -Dname=def\n");
	    fprintf(stderr,"   -Dname    : define symbol for specification preprocessing\n");
	    fprintf(stderr,"   -N        : accept ISO VDM-SL standard syntax extensions\n");
  
	    /* TT */
	    fprintf(stderr,"   -M        : select automatic memory management (only C++!)\n");  
	    fprintf(stderr,"   -K        : automatic memory management with macros for C functions (-M required)\n");
	    fprintf(stderr,"   -J        : include default set of operations for C++ ( requires -M )\n");
  
	    fprintf(stderr,"   -I<dir>   : give additional directory for specification preprocessing\n\n");
	    fprintf(stderr,"   -s        : generate single VDM file for each domain (internal)\n");
	    fprintf(stderr,"   -H        : generate C header files '.h1', '.h2' and '.h' (internal)\n");
	    fprintf(stderr,"   -E        : write '.enr' file only if changed (internal)\n\n");
	    fprintf(stderr,"   -o <name> : define basename of output files\n");
	    fprintf(stderr,"   -u        : suppress computation of operator enrichments\n");
	    fprintf(stderr,"   -k        : don't add standart domains into generated files\n");
	    fprintf(stderr,"   -r <name> : try to use <name> as enrichment file\n\n");
	    fprintf(stderr,"   -e        : generate enrichment VDM file\n");
	    fprintf(stderr,"   -h        : generate C header file\n");
	    fprintf(stderr,"   -c        : generate C implementation file\n");
	    fprintf(stderr,"   -d        : generate Pascal header file\n");
	    fprintf(stderr,"   -m        : generate Modula-2 definition part\n");
	    fprintf(stderr,"   -i        : generate Modula-2 import statements\n");
	    fprintf(stderr,"   -C        : generate C++ header file\n");
	    fprintf(stderr,"   -U        : generate C++ class names with upper case\n");
	    fprintf(stderr,"   -F        : generate C++ friend functions\n");
	    fprintf(stderr,"   -O        : generate C++ operator functions\n");
	    fprintf(stderr,"   -a        : additional suffix _domain for selectors of trees and unions\n");
	    fprintf(stderr,"   -X        : collect comments in .enr file\n");
	    fprintf(stderr,"   -client [host [port]]\n");
	    fprintf(stderr,"             : run as client for the \"vdmd\" on machine \"host\" at port \"port\"\n");
	    fprintf(stderr,"               default host is \"localhost\"\n");
	    fprintf(stderr,"               default port is \"4711\"\n");
	    fprintf(stderr,"               -B options will be ignored\n");
	}
    exit(exitcode);
    return
	exitcode;
}
