#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <tcl.h>
#include "proto.h"

static Tcl_Trace Tcl_DebuggerToken;
static FILE *Tcl_DebuggerFile=NULL;
static Tcl_Debugging=0;

static void Tcl_Debugger(
			 ClientData clientData,
			 Tcl_Interp *interp,
			 int level,
			 char *command,
			 Tcl_CmdProc *cmdProc,
			 ClientData cmdClientData,
			 int argc,
			 char *argv[])
{
    int i;
    if (Tcl_Debugging) {
	fprintf(Tcl_DebuggerFile,"[%d] %s\n",level,command);
	fprintf(Tcl_DebuggerFile,"===>");
	for (i=0;i < argc;i++) 
	    fprintf(Tcl_DebuggerFile," %s",argv[i]);
	fprintf(Tcl_DebuggerFile,"\n");
	fflush(Tcl_DebuggerFile);
    }
}

int Tcl_debugCmd(ClientData clientData,Tcl_Interp *interp,int argc,char *argv[])
{
    if (argc == 1) {
	sprintf(interp->result,"%s",Tcl_DebuggerFile ? "1" : "0");
	return TCL_OK;
    }
	
    if (argc == 2) {
	if (!strcmp(argv[1],"on"))
	    Tcl_Debugging = 1;
	else if (!strcmp(argv[1],"off"))
	    Tcl_Debugging = 0;
	else if (!strcmp(argv[1],"stop")) {
	    Tcl_DeleteTrace(interp,Tcl_DebuggerToken);
	    if (Tcl_DebuggerFile)
		fclose(Tcl_DebuggerFile);
	    Tcl_DebuggerFile = NULL;
	    Tcl_Debugging = 0;
	}
	else {
	    sprintf(interp->result,"debug [on|off|stop]");
	    return TCL_ERROR;
	}
	return TCL_OK;
    }
    
    if (argc != 3) {
	interp->result = "debug level file";
	return TCL_ERROR;
    }
	
    if (Tcl_DebuggerFile)
	fclose(Tcl_DebuggerFile);
    if ( (Tcl_DebuggerFile = fopen(argv[2],"w+")) == NULL) {
	sprintf(interp->result,"unable to open debugfile");
	return TCL_ERROR;
    }
	
    Tcl_DebuggerToken=Tcl_CreateTrace(interp,atoi(argv[1]),Tcl_Debugger,
				      (ClientData)NULL);
    
    return TCL_OK;
}

