#include <stdlib.h>
#include <unistd.h>
#include "proto.h"
#include "support.h"

#ifdef TCL_CHANNEL_SUPPORT

void
Tcl_dupCmdChannelOption(Tcl_Interp *interp,Tcl_Channel orgchannel,Tcl_Channel newchannel) 
{
    Tcl_DString opt;
    int argc,i;
    char **argv;
    
    Tcl_DStringInit(&opt);
    Tcl_GetChannelOption(orgchannel,NULL,&opt);
    Tcl_SplitList(interp,Tcl_DStringValue(&opt),&argc,&argv);
    for (i=1;i < argc;i+=2) 
	Tcl_SetChannelOption(interp,newchannel,argv[i-1],argv[i]);   
    free((char*)argv);
    Tcl_DStringFree(&opt);
}

int Tcl_dupCmd(ClientData clientData,Tcl_Interp * interp,int argc,char *argv[])
{
    ClientData oldfile,newfile;
    int rc,mode;
    Tcl_Channel channel,orgchannel;
	
    if (argc != 2 && argc != 3) {
	interp->result = "wrong # args: dup fileId ?targetFileId?";
	return TCL_ERROR;
    }
	
    orgchannel=Tcl_GetChannel(interp,argv[1],&mode); 
    mode=(mode==TCL_WRITABLE);
    if (Tcl_GetOpenFile(interp,argv[1],mode,0,&oldfile) != TCL_OK) {
	sprintf(interp->result,"bad file identifier \"%s\"",
		argv[1]);
	return TCL_ERROR;
    }
    
    if (argc == 3) {	
	channel=Tcl_GetChannel(interp,argv[2],&mode); 
	mode=(mode==TCL_WRITABLE);
	if (Tcl_GetOpenFile(interp,argv[2],mode,0,&newfile) != TCL_OK) {
	    sprintf(interp->result,"bad file handle \"%s\"",
		    argv[2]);
	    return TCL_ERROR;
	}

	rc=dup2(fileno((FILE*)oldfile),fileno((FILE*)newfile));
	Tcl_dupCmdChannelOption(interp,orgchannel,channel);
	sprintf(interp->result,"%s",Tcl_GetChannelName(channel));
	return TCL_OK;		

    } else {
	rc = dup(fileno((FILE*)oldfile));
	orgchannel=Tcl_GetChannel(interp,argv[1],&mode);
	if (rc >= 0 && orgchannel) {
	    channel=Tcl_MakeFileChannel((ClientData)rc,(ClientData)rc,mode);
	    Tcl_dupCmdChannelOption(interp,orgchannel,channel);
	    Tcl_RegisterChannel(interp, channel);
	    sprintf(interp->result,"%s",Tcl_GetChannelName(channel));
	    return TCL_OK;
	}
    }

    Tcl_PosixError(interp);
    return TCL_ERROR;
}


#endif /* TCL_CHANNEL_SUPPORT */
