#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "proto.h"

struct sigTableStruct {
    char *signame;
    int signum;
    int sigmark;
    Tcl_Interp *interp;
    Tcl_AsyncHandler asyncHandler;
    char *sigaction;
} sigTable[] = { 

#ifdef SIGALRM
    { "SIGALRM", SIGALRM, 0, NULL },
#endif
#ifdef SIGBUS
    { "SIGBUS", SIGBUS, 0, NULL },
#endif
#ifdef SIGCLD
    { "SIGCLD", SIGCLD, 0, NULL },
#endif
#ifdef SIGEMT
    { "SIGEMT", SIGEMT, 0, NULL },
#endif
#ifdef SIGFPE
    { "SIGFPE", SIGFPE, 0, NULL },
#endif
#ifdef SIGHUP
    { "SIGHUP", SIGHUP, 0, NULL },
#endif
#ifdef SIGILL
    { "SIGILL", SIGILL, 0, NULL },
#endif
#ifdef SIGINT 
    { "SIGINT", SIGINT, 0, NULL },
#endif
#ifdef SIGIO
    { "SIGIO", SIGIO, 0, NULL },
#endif
#ifdef SIGIOT
    { "SIGIOT", SIGIOT, 0, NULL },
#endif
/* SIGKILL normally cames here but can not be changed */
#ifdef SIGKILL
    { "SIGKILL", SIGKILL, 0, NULL },
#endif
#ifdef SIGPIPE
    { "SIGPIPE", SIGPIPE, 0, NULL },
#endif
#ifdef SIGPOLL
    { "SIGPOLL", SIGPOLL, 0, NULL },
#endif
#ifdef SIGPROF
    { "SIGPROF", SIGPROF, 0, NULL },
#endif
#ifdef SIGPWR
    { "SIGPWR", SIGPWR, 0, NULL },
#endif
#ifdef SIGQUIT
    { "SIGQUIT", SIGQUIT, 0, NULL },
#endif
#ifdef SIGSEGV
    { "SIGSEGV", SIGSEGV, 0, NULL },
#endif
/* SIGSTOP normally cames here but can not be changed */
#ifdef SIGSYS
    { "SIGSYS", SIGSYS, 0, NULL },
#endif
#ifdef SIGTERM
    { "SIGTERM", SIGTERM, 0, NULL },
#endif
#ifdef SIGTRAP
    { "SIGTRAP", SIGTRAP, 0, NULL },
#endif
#ifdef SIGTSTP
    { "SIGTSTP", SIGTSTP, 0, NULL },
#endif
#ifdef SIGTTIN
    { "SIGTTIN", SIGTTIN, 0, NULL },
#endif
#ifdef SIGTTOU
    { "SIGTTOU", SIGTTOU, 0, NULL },
#endif
#ifdef SIGURG
    { "SIGURG", SIGURG, 0, NULL }, 
#endif
#ifdef SIGUSR1
    { "SIGUSR1", SIGUSR1, 0, NULL },
#endif
#ifdef SIGUSR2
    { "SIGUSR2", SIGUSR2, 0, NULL },
#endif
#ifdef SIGVTALRM
    { "SIGVTALRM", SIGVTALRM, 0, NULL },
#endif
#ifdef SIGWINCH
    { "SIGWINCH", SIGWINCH, 0, NULL },
#endif
#ifdef SIGXCPU
    { "SIGXCPU", SIGXCPU, 0, NULL },
#endif
#ifdef SIGXFSZ
    { "SIGXFSZ", SIGXFSZ, 0, NULL },
#endif
    { NULL }
};

char *sigName(int signum)
{
    register struct sigTableStruct *st=sigTable;
  
    while (st->signame && (st->signum != signum)) st++;
    return st->signame;
}

int sigNum(char *signame)
{
    register struct sigTableStruct *st=sigTable;
  
    while (st->signame && strcmp(st->signame,signame)) st++;
    return st->signum;
}

int sigIdx(char *signame)
{
    register struct sigTableStruct *st=sigTable; 
    register int i;

    for (i=0;st->signame && strcmp(st->signame,signame) && 
	     st->signum != atoi(signame);st++,i++);
    return st->signame?i:-1;
}

void sighandler(int signum)
{
    register struct sigTableStruct *st=sigTable;
  
    while (st->signame && (st->signum != signum)) st++;
    assert(st->signame); /* unknown signal called by kernel */
    /*  printf("signal %s marked\n",st->signame);  */

    st->sigmark++;
    Tcl_AsyncMark(st->asyncHandler); 
    signal(st->signum,sighandler); 
}

int asyncProc(ClientData clientData, Tcl_Interp *interp, int code)
{
    struct sigTableStruct *st=(struct sigTableStruct*)clientData;

    /*  printf("signal %s called\n",st->signame);  */
    if (--st->sigmark)
	Tcl_AsyncMark(st->asyncHandler);
  
    if (st->sigaction)
	Tcl_GlobalEval(st->interp,st->sigaction); 
  
    return code;
}

int Tcl_signalCmd(ClientData clientData,Tcl_Interp *interp,int argc,char *argv[])
{
    int sigidx;

    if (argc == 1) {
	Tcl_AppendResult(interp, "signal [ signum [\"sigaction\"|SIG_IGN|SIG_DFL] ]\n", (char *)0);

	for (sigidx=0;
	     (sigTable+sigidx)->signame;
	     sigidx++)
	    {
		char str[255];
		sprintf(str,"%2d) %s%c",sigTable[sigidx].signum,
			sigTable[sigidx].signame,((sigidx+1)%4)?'\t':'\n');
		Tcl_AppendResult(interp,str,(char *)0);
	    }
	return TCL_OK;
    }
	
    if (argc > 3) {
	interp->result = "wrong # args";
	return TCL_ERROR;
    }
	
    sigidx = sigIdx(argv[1]);
	
    if (argc == 2) {
	if (sigidx < 0) interp->result="0";
	else interp->result="1";
	return TCL_OK;
    }

    if (sigidx < 0) {
	interp->result = "wrong signal specified";
	return TCL_ERROR;
    }

    sigTable[sigidx].asyncHandler = 
	Tcl_AsyncCreate(asyncProc,(void*)&sigTable[sigidx]);

    if (sigTable[sigidx].sigaction)
	free(sigTable[sigidx].sigaction);

    interp->result = sigTable[sigidx].sigaction = strdup(argv[2]);
    assert(sigTable[sigidx].sigaction);
    sigTable[sigidx].interp = interp;
	
    if (signal(sigTable[sigidx].signum,
	       !strcmp(argv[2],"SIG_IGN")?SIG_IGN:
	       !strcmp(argv[2],"SIG_DFL")?SIG_DFL:sighandler) == SIG_ERR) {
	Tcl_PosixError(interp);
	return TCL_ERROR;
    }
    return TCL_OK;
}

int sigTableSortFnc(const void *p1,const void *p2)
{
    return (((struct sigTableStruct*)p1)->signum - 
	    ((struct sigTableStruct*)p2)->signum);
}

void Tcl_signalCmdInit()
{
    static int initialized=0;
    int nmemb;
    struct sigTableStruct *st=sigTable;

    if (initialized)
	return;

    for (nmemb=0;(st+nmemb)->signame;nmemb++);
    qsort(sigTable,nmemb,sizeof(struct sigTableStruct),sigTableSortFnc);
    initialized++;
}

void Tcl_killCmdInit() 
{
    Tcl_signalCmdInit();
}

void Tcl_signalCmdDone()
{
    struct sigTableStruct *st=sigTable;

    for (;st->signame;st++)
	if (st->sigaction) { 
	    free(st->sigaction); 
	    st->sigaction = NULL;
	}  
}

int Tcl_killCmd(ClientData clientData,Tcl_Interp *interp,int argc,char *argv[])
{
    int rc,sigidx;
	
    if (argc == 1) {
	Tcl_AppendResult(interp, "kill signum pid\n", (char *)0);

	for (sigidx=0;
	     (sigTable+sigidx)->signame;
	     sigidx++)
	    {
		char str[255];
		sprintf(str,"%2d) %s%c",sigTable[sigidx].signum,
			sigTable[sigidx].signame,((sigidx+1)%4)?'\t':'\n');
	    }
	return TCL_OK;
    }
	
    if (argc != 3) {
	interp->result = "kill signal pid";
	return TCL_ERROR;
    }
    
    sigidx = sigIdx(argv[1]);
	
    if (sigidx < 0) {
	interp->result = "wrong signal specified";
	return TCL_ERROR;
    }
	
    if ( (rc=kill(atoi(argv[2]),sigTable[sigidx].signum)) < 0) {
	Tcl_PosixError(interp);
	return TCL_ERROR;
    }
    return TCL_OK;
}

