
/* dbmtcl.cc,v 1.13 1995/03/08 12:08:24 vdm Exp */
			/* include program kernel */
#include "vdmdbm.cc"
			/* include macros for simple tcl interface */
#include "vdmtcl.cc"

#include "./cmds/cmds.h"

extern char configcmd[];        /* find this in tclconfig.cc, produced by
				   sed, source is config.tcl */

/*--- own tcl commands -------------------------------------------------------*/

#ifdef flat
#define dbimpl " (database implemented as flat file)"
#define dbimplementation "flat"
#else
#define dbimpl " (database implemented as 4 level directory structure)"
#define dbimplementation "directory"
#endif

#define dbversion "$Id: dbmtcl.cc,v 1.16 1996/09/07 14:13:35 vdm Exp $" dbimpl

CmdProc(databaseimplementation) {
  ArgEq(1);
  Return(dbimplementation);
}

CmdProc(version) {
  ArgEq(1);
  Return(dbversion);
}

/*----------------------------------------------------------------------------*/

CmdProc(profile) {
  ArgEq(1);

  profile_adt();
  profile_impl();
  profile_operation();
  profile_str();
  profile_stringl();
  profile_strings();
  profile_vdmdb();

  Return("");
}

/*----------------------------------------------------------------------------*/

CmdProc(loaddatabase) {
  ArgEq(2);
  dbm.load_db(argv[1]);
  Return("");
}

CmdProc(storedatabase) {
  ArgEq(2);
  dbm.store_db(argv[1]);
  Return("");
}

CmdProc(databasechanged) {
  ArgEq(1);
  ReturnBOOL(dbm.db_changed());
}

/*--- session functions ------------------------------------------------------*/

CmdProc(session) {
  static char hlptxt[] = 
    "session usage:\n"
    "\tsession create <id>\t\tcreate session and return id or zero\n"
    "\tsession exists <id>\ttest if vaild session\n"
    "\tsession release <id>\trelease session and return boolean\n"
    "\tsession <id>\t\tselect session context return boolean\n"
    "\tsession list\t\treturn all known sessions\n"
    "\tsession\t\t\treturn current session id\n"
    "\tsession help\t\tshow this help\n";

  static char errtxt[] = "unkown session subcommand: type \"session help\"";

  if (argc < 2) { ReturnSTR(dbm.session_current()); }

  if (argc == 2) {
    if (strcmp(argv[1],"create"   )==0) { ReturnSTR(dbm.session_create()); }
    if (strcmp(argv[1],"list"     )==0) { ReturnSTR(dbm.session_list()); } 
    if (strcmp(argv[1],"help"     )==0) { Return(hlptxt); }

    char *err;
    long int sid=strtol(argv[1],&err,10);

    if (*err) { Error(errtxt); }

    ReturnBOOL(dbm.session_switch(INTG(sid)));
  } 

  ArgEq(3);

  if (strcmp(argv[1],"release")==0) {
    ReturnBOOL(dbm.session_release(INTG(atol(argv[2]))));
  }
  else if (strcmp(argv[1],"exists")==0) {
    ReturnBOOL(dbm.session_exists(INTG(atol(argv[2]))));
  }
  else if (strcmp(argv[1],"create")==0) {
    long l=atol(argv[2]);
    ReturnSTR(dbm.session_create(NAT0(l>0?l:0)));
  }

  Error(errtxt);
}

/*--- test functions ---------------------------------------------------------*/

CmdProc(existadt) {
  ArgEq(2);
  ReturnBOOL(dbm.db.exist_adt(argv[1]));
}

CmdProc(existimpl) {
  ArgEq(3);
  ReturnBOOL(dbm.db.exist_impl(argv[1],argv[2]));
}

CmdProc(existoper) {
  ArgEq(4);
  ReturnBOOL(dbm.db.exist_operation(argv[1],argv[2],argv[3]));
}

CmdProc(existcomp) {
  ArgEq(5);
  ReturnBOOL(dbm.db.exist_component(argv[1],argv[2],argv[3],argv[4]));
}

CmdProc(checkoper) {
  ArgEq(4);
  ReturnBOOL(dbm.db.search_operation(argv[1],argv[2],argv[3]));
}

CmdProc(checkcomp) {
  ArgEq(5);
  ReturnBOOL(!dbm.db.search_component(argv[1],argv[2],argv[3],argv[4]).is_empty());
}

CmdProc(checksign) {
  ArgEq(4);
  ReturnBOOL(!dbm.db.search_sign_component(argv[1],argv[2],argv[3]).is_empty());
}

/*--- list database content --------------------------------------------------*/

CmdProc(adts) {
  ArgEq(1);
  ReturnSTR(dbm.sel_adts().concat(" "));
}

CmdProc(implementations) {
  ArgEq(2);
  ReturnSTR(dbm.sel_impl(argv[1]).concat(" "));
}

CmdProc(operations) {
  ArgEq(3);
  ReturnSTR(dbm.sel_oper(argv[1],argv[2]).sort().concat(" "));
}

CmdProc(specialops) {
  ArgEq(3);
  ReturnSTR(dbm.db.sel_op(argv[1],argv[2]).sort().concat(" "));
}

CmdProc(components) {
  ArgEq(4);
  ReturnSTR(dbm.db.sel_comp(argv[1],argv[2],argv[3]).sort().concat(" "));
}

CmdProc(implparams) {
  ArgEq(3);
  STRINGS res  = dbm.db.sel_impl_params(argv[1],argv[2]);
          res -= dbm.filter_params;
  ReturnSTR(res.sort().concat(" "));
}

CmdProc(operparams) {
  ArgEq(4);
  STRINGS res  = dbm.db.sel_oper_params(argv[1],argv[2],argv[3]);
          res -= dbm.filter_params;
  ReturnSTR(res.sort().concat(" "));
}

/*-- selector function -------------------------------------------------------*/

CmdProc(selcomp) {
  ArgEq(5);
  ReturnSTR(dbm.db.sel_component(argv[1],argv[2],argv[3],argv[4]));
}

CmdProc(searchcomp) {
  ArgEq(5);
  ReturnSTR(dbm.db.compute_component(argv[1],argv[2],argv[3],argv[4]));
}

CmdProc(searchsign) {
  ArgEq(4);
  ReturnSTR(dbm.db.compute_sign_component(argv[1],argv[2],argv[3]));
}

/*-- documentation -----------------------------------------------------------*/

CmdProc(predefined) {
  ArgEq(1);
  ReturnSTR(dbm.db.compute_component(stdlib_adt,common_impl,basic_op,vdm_comp));
}

CmdProc(adtdocu) {
  ArgEq(1);
  ReturnSTR(dbm.adtdocu());
}

CmdProc(docu) {
  ArgEq(4);
  STR r = docu_header;
  r += dbm.db.setup_docu();
  r += dbm.db.search_docu_components(argv[1],argv[2],argv[3],
         dbm.sign_comps);
  ReturnSTR(r);
}

CmdProc(op1docu) {
  ArgEq(4);
  STR r = dbm.db.setup_docu();
  r += dbm.op1docu(argv[1],argv[2],argv[3]);
  ReturnSTR(r);
}

CmdProc(oprdocu) {
  ArgEq(3);
  STR r = dbm.db.setup_docu();
  r += dbm.oprdocu(argv[1],argv[2]);
  ReturnSTR(r);
}

CmdProc(impdocu) {
  ArgEq(2);
  ReturnSTR(dbm.impldocu(argv[1]));
}

/*--- database modification ---------------------------------------------------*/

CmdProc(export) {
  ArgEq(1);
  ReturnSTR(dbm.export());
}

CmdProc(partialexport) {
  ArgEq(1);
  ReturnSTR(dbm.export_selected());
}

// Sun Jun 23 16:11:48 MET DST 1996 mt added exportformat to create vdm export 
// format

CmdProc(exportformat)  {
  ArgEq(6);
  ReturnSTR(dbm.exportformat(argv[1],argv[2],argv[3],argv[4],argv[5]));
}

CmdProc(import) {
  ArgEq(2);
  STR co;
  co.fread(argv[1]);
  ReturnBOOL(dbm.import(co));
}

// Sun Jun 23 16:11:48 MET DST 1996 mt added directimport for server feature

CmdProc(directimport) {
  ArgEq(2);
  STR co(argv[1]);
  ReturnBOOL(dbm.import(co));
}

CmdProc(newcomponent) {
  ArgEq(6);
  COMPONENT co;
  co.fread(argv[5]);
  if ( ! co )
    dbm.new_component(argv[1],argv[2],argv[3],argv[4],co);
  Return("");
}

/*--- remove parts from database ---------------------------------------------*/

CmdProc(removedb) {
  ArgEq(1);
  dbm.clear_db();
  Return("");
}

CmdProc(removeadt) {
  ArgEq(2);
  dbm.rm_adt(argv[1]);
  Return("");
}

CmdProc(removeimpl) {
  ArgEq(3);
  dbm.rm_impl(argv[1],argv[2]);
  Return("");
}

CmdProc(removeoper) {
  ArgEq(4);
  dbm.rm_operation(argv[1],argv[2],argv[3]);
  Return("");
}

CmdProc(removecomp) {
  ArgEq(5);
  dbm.rm_component(argv[1],argv[2],argv[3],argv[4]);
  Return("");
}

/* --- new components --------------------------------------------------------*/

CmdProc(newadt) {
  ArgEq(2);
  dbm.new_adt(argv[1]);
  Return("");
}

CmdProc(newimpl) {
  ArgEq(3);
  dbm.new_impl(argv[1],argv[2]);
  Return("");
}

CmdProc(newoper) {
  ArgEq(4);
  dbm.new_operation(argv[1],argv[2],argv[3]);
  Return("");
}

CmdProc(newcomp) {
  ArgEq(5);
  dbm.new_component(argv[1],argv[2],argv[3],argv[4]);
  Return("");
}

/*--- select special components ----------------------------------------------*/

CmdProc(select) {
  ArgGe(2);
  STRINGS dummy  = empty;
  STRINGS & sset =
    (strcmp(argv[1],"adttypes"   )==0)	? dbm.adttypes		:
    (strcmp(argv[1],"specialimpl")==0)	? dbm.special_impl	:
    (strcmp(argv[1],"specialops" )==0)	? dbm.special_ops	:
    (strcmp(argv[1],"signcomps"  )==0)  ? dbm.sign_comps	:
    (strcmp(argv[1],"filteradts" )==0)  ? dbm.filter_adts	:
    (strcmp(argv[1],"filterimpl" )==0)  ? dbm.filter_impl	:
    (strcmp(argv[1],"filteroper" )==0)  ? dbm.filter_oper	:
    (strcmp(argv[1],"filtercomp" )==0)  ? dbm.filter_comp	:
    (strcmp(argv[1],"filterparam")==0)  ? dbm.filter_params	:
    dummy;

  if (& sset == & dummy) {Error("unknown option");}

  if (argc == 2) {
    sset = empty;
  } else {
    for (int i = 2; i < argc; ++i )
      sset |= STR(argv[i]);
  }
  Return("");
}

CmdProc(operationtype) {
  ArgEq(2);

  // marko temme memorial comments. He has wasted hours with coredumps here !
  //  cout << "before" << dbm.operationtype << "end";
  //  dbm.change_operationtype(STR(argv[1]));

  dbm.operationtype = argv[1];

  //  cout << "later" << dbm.operationtype << "end";
  Return("");
}

CmdProc(opertype) {
  ArgEq(4);
  ReturnSTR(dbm.db.classify_op(argv[1],argv[2],argv[3]));
}

CmdProc(solveincludesinfile) {
  ArgEq(2);
  STR ci;
  ci.fread(argv[1]);
  ReturnSTR(dbm.db.solve_include(ci));
}

/*--- the tcl "boot" command was here ---------------------------------------------*/

/* 
   the original bootcmd commands are moved to the proc 
   config_conventional in config.tcl for backward support
   
   The old commands are only executed if the name is "dbm" 
   supplying backward support, if the shell is named "vdmd"
   the config-commands in config.tcl now read the 
   configuration file vdmd.conf

   */

/*-- the "main" programm --------------------------------------------------------*/

int Tcl_AppInit(Tcl_Interp *interp) {

    initdbm();	// global variables must be explicitly initialized
			// error in g++
		// no no, g++ is right, but classes were initialized
		// first. comment mt.

    if (Tcl_Init(interp) == TCL_ERROR) {
	return TCL_ERROR;
    }
 
    CreateCmd(databaseimplementation);
    CreateCmd(version);
    CreateCmd(profile);

    CreateCmd(loaddatabase);
    CreateCmd(storedatabase);
    CreateCmd(databasechanged);

    // Sun Apr 14 23:35:45 MET DST 1996 session cmds by mt for multisession feature
    CreateCmd(session);

    CreateCmd(existadt);
    CreateCmd(existimpl);
    CreateCmd(existoper);
    CreateCmd(existcomp);

    CreateCmd(checkoper);
    CreateCmd(checkcomp);
    CreateCmd(checksign);

    CreateCmd(adts);
    CreateCmd(implementations);
    CreateCmd(operations);
    CreateCmd(specialops);
    CreateCmd(components);
    CreateCmd(implparams);
    CreateCmd(operparams);

    CreateCmd(selcomp);
    CreateCmd(searchcomp);
    CreateCmd(searchsign);

    CreateCmd(predefined);
    CreateCmd(docu);
    CreateCmd(adtdocu);
    CreateCmd(op1docu);
    CreateCmd(oprdocu);
    CreateCmd(impdocu);

    CreateCmd(import);

    // mt
    CreateCmd(directimport);
    CreateCmd(export);
    CreateCmd(partialexport);

    // 06/23/1996 mt needed this to get the vdm format
    CreateCmd(exportformat);
    CreateCmd(newcomponent);

    CreateCmd(removedb);
    CreateCmd(removeadt);
    CreateCmd(removeimpl);
    CreateCmd(removeoper);
    CreateCmd(removecomp);

    CreateCmd(newadt);
    CreateCmd(newimpl);
    CreateCmd(newoper);
    CreateCmd(newcomp);

    CreateCmd(select);
    CreateCmd(operationtype);
    CreateCmd(opertype);
    CreateCmd(solveincludesinfile); 

    // The new c-commands in subdir ./cmd/ are installed
    cmds_Register(interp);
    
    if ( Eval(configcmd) != TCL_OK) { 
       fprintf(stderr,"error evaluating configuration script: %s\n",
       	      interp->result);
       exit(1); 
    }
    Return("");
}

/*-------------------------------------------------------------------------------*/

#if TCLMAJOR >= 7 

#if TCLMINOR >= 4
int main(int argc, char * argv[]) {
    Tcl_Main(argc,argv,Tcl_AppInit);
    return 0;
}
#elif TCLMINOR > 3
int main(int argc, char * argv[]) {
    Tcl_Main(argc,argv);
    return 0;
}
#endif

#endif

/*-------------------------------------------------------------------------------*/
