# $Id: primitives.tcl,v 1.1 1996/09/05 07:03:29 vdm Exp $
#
#  primitives debug options
#
##########################################################################

set primitives_DEBUGFILE ""

proc primitives_debugon file {
    global primitives_DEBUGFILE
    switch $file { 
	stdout { set primitives_DEBUGFILE "stdout" }
	stderr { set primitives_DEBUGFILE "stderr" }
	default { set primitives_DEBUGFILE $file}
    }
}

proc primitives_debugoff {} {
    uplevel #0 set primitives_DEBUGFILE ""
}

proc primitives_debug {peer msg} {
    global primitives_DEBUGFILE
    if {[string length $primitives_DEBUGFILE]} {
	if {[string length $peer]} { 
	    puts -nonewline $primitives_DEBUGFILE "$peer: primitive: " 
	}
	puts -nonewline $primitives_DEBUGFILE "$msg"
	flush $primitives_DEBUGFILE
    }
}

#
#  primitives session service access point 
#
##########################################################################

proc primitives_ssapdefine { login logoff dataindication exception} {
    return [list $login $logoff $dataindication $exception]
}

proc primitives_ssap { ssap no args } {
    if {[string length [lindex $ssap $no]]} {
	return [eval "[lindex $ssap $no] $args"]
    } else {
	return ""
    }
}

#
#  primitives itself
#
##########################################################################

proc S-CONNECT.request { ssap host port } {
    set peer [socket $host $port]
    primitives_debug $peer "S-CONNECT.request $ssap $host $port\n"
    fconfigure $peer -buffering none -translation lf
    return [S-CONNECT.confirm $ssap $peer]
}

proc S-CONNECT.indication {ssap peer cliaddr port} {
    if [catch {
	primitives_debug $peer "S-CONNECT.indication $ssap $cliaddr $port\n"
	fconfigure $peer -buffering none -translation lf
	return [S-CONNECT.response $ssap $peer $cliaddr $port]
    } err]==1 {
	S-P-EXCEPTION-REPORT.request $err $ssap $peer
    }
}

proc S-CONNECT.response {ssap peer cliaddr port} {
    primitives_debug $peer "S-CONNECT.response $ssap $cliaddr $port"

    if [catch {primitives_ssap $ssap 0 $peer $cliaddr $port} msg]==1 {
	primitives_debug "" " -> refused\n" 
	puts $peer "Connection refused: $msg"
	close $peer
	return
    }

    primitives_debug "" " -> accepted\n" 
    puts $peer "Connection accepted"
    atclose $peer appendUnique "S-P-ABORT.indication \{$ssap\} $peer"
    atexit appendUnique "close $peer"
    fileevent $peer readable "S-DATA.indication \{$ssap\} $peer"
}

proc S-CONNECT.confirm {ssap peer} {
    primitives_debug $peer "S-CONNECT.confirm $ssap"
    set return [gets $peer]  
    if {[lindex $return 1] == "refused:"} {
	primitives_debug "" " -> refused\n" 
	close $peer
	error $return
    }

    primitives_debug "" " -> accepted\n" 
    atclose $peer appendUnique "S-RELEASE.request \{$ssap\} $peer"
    atexit appendUnique "S-U-ABORT.request $peer"
    fileevent $peer readable "S-DATA.indication \{$ssap\} $peer"
    return $peer
}

proc S-RELEASE.request {ssap peer} {
    primitives_debug $peer "S-RELEASE.request $ssap\n"
    puts $peer "release.request"
    return [S-RELEASE.confirm $ssap $peer]
}

proc S-RELEASE.indication {ssap peer} {
    primitives_debug $peer "S-RELEASE.indication $ssap\n"
    fileevent $peer readable {}
    return [S-RELEASE.response $ssap $peer]
}

proc S-RELEASE.response {ssap peer} {
    primitives_debug $peer "S-RELEASE.response $ssap\n"
    puts $peer "release.response"
    fileevent $peer writable {}
    atexit delete "close $peer"
    atexit delete "S-U-ABORT.request $peer"
    atclose $peer delete "S-RELEASE.request \{$ssap\} $peer"
    atclose $peer delete "S-P-ABORT.indication \{$ssap\} $peer"
    primitives_ssap $ssap 1 release $peer 
    close $peer
}

proc S-RELEASE.confirm {ssap peer} {
    primitives_debug $peer "S-RELEASE.confirm $ssap\n"
    gets $peer SPDU  
    if [eof $peer] {
	S-U-ABORT.indication "$ssap" $peer
	return
    }
    if {$SPDU != "release.response"} {
	error "malformed SPDU"
    }
    fileevent $peer readable {}
    fileevent $peer writable {}
    atexit delete "close $peer"    
    atexit delete "S-U-ABORT.request"
    atclose $peer delete "S-P-ABORT.indication \{$ssap\} $peer"
    primitives_ssap $ssap 1 release $peer 
    #  close $peer is done by atclose
}

proc S-U-ABORT.indication {ssap peer} {
    primitives_debug $peer "S-U-ABORT.indication $ssap\n"
    atclose $peer delete "S-P-ABORT.indication \{$ssap\} $peer"
    atclose $peer delete "S-RELEASE.request \{$ssap\} $peer"
    primitives_ssap $ssap 1 abort $peer 
    close $peer
}

proc S-P-ABORT.request { ssap peer } {
    primitives_debug $peer "S-P-ABORT.request $ssap\n"
}

proc S-P-ABORT.indication {ssap peer} {
    primitives_debug $peer "S-P-ABORT.indication $ssap\n"
    primitives_ssap $ssap 1 abort $peer 
}

proc S-DATA.request {peer} {
    primitives_debug $peer "S-DATA.request\n"
    puts $peer "data.request"
}

proc S-DATA.indication {ssap peer} {
    if [catch {
	gets $peer SPDU
	if [eof $peer] { S-U-ABORT.indication $ssap $peer; return }
	switch "$SPDU" {
	    release.request { S-RELEASE.indication $ssap $peer}
	    data.request { 
		primitives_debug $peer "S-DATA.indication $ssap\n" 
		primitives_ssap $ssap 2 $peer
	    }
	    default { error "malformed SPDU" }
	}
    } err]==1 {
	S-P-EXCEPTION-REPORT.request $err $ssap $peer
    }
}

proc S-P-EXCEPTION-REPORT.request { err ssap peer } {
    primitives_debug $peer "S-P-EXCEPTION-REPORT.request $ssap $err" 
    primitives_ssap $ssap 3 $err $peer
    catch { close $peer }
}

