# $Id: rpc.tcl,v 1.1 1996/09/05 07:03:29 vdm Exp $

global rpcFile

#
#  rpc debug options
#
##########################################################################

set rpc_DEBUGFILE ""

proc rpc_debugon file {
    global rpc_DEBUGFILE
    switch $file {
	stdout { set rpc_DEBUGFILE "stdout" }
	stderr { set rpc_DEBUGFILE "stderr" }
	default { set rpc_DEBUGFILE "$file" }
    }
}

proc rpc_debugoff {} {
    uplevel #0 set rpc_DEBUGFILE ""
}

proc rpc_debug {peer msg} {
    global rpc_DEBUGFILE
    if {[string length $rpc_DEBUGFILE]} {
	if {[string length $peer]} { puts -nonewline $rpc_DEBUGFILE "$peer: rpc: " }
	puts -nonewline $rpc_DEBUGFILE "$msg"
	flush $rpc_DEBUGFILE
    }
}

#
#  rpc callback
#
##########################################################################

proc rpc_CheckCommand { client checkFunc cmd } {
    rpc_debug "$client" "rpc_CheckCommand $checkFunc $cmd\n"
    for {set i 0} {1} {incr i} {
	switch [catch {
	    $checkFunc [lindex $cmd $i] [lrange $cmd [expr $i +1] end]
	} err] {
	    4 { continue }
	    0 { return }
	    default { break }
	}
    }
    if {$err == {}} { set err "permission denied" }
    error "$err"
}

proc rpc_ProcessCommand {checkFunc client} {
    rpc_debug $client "rpc_ProcessCommand $checkFunc\n"
    global rpcFile
    set rpcFile "$client"
    set cmd [protocol_recvmsg $client]
  
    if {$checkFunc != {}} { 
	if [catch {rpc_CheckCommand $client "$checkFunc" $cmd} err]==1 {
	    protocol_sendmsg $client "$err" "error";return
	}
    }
    if [catch { set res [uplevel \#0 $cmd] } err]==1 {
	protocol_sendmsg $client "$err" "error"
    } else {
	protocol_sendmsg $client "$res" "ok"
    }
    flush stdout
    flush stderr
    return
}

proc rpc_Exception { userFunc err client } {
    rpc_debug $client "rpc_Exception $err $client\n"  
    if {$userFunc != {}} {
	$userFunc "$err" $client
    }
}

#
#  rpc user functions
#
##########################################################################

proc rpc_Exec {server args} {
    rpc_debug $server "rpc_Exec $args\n"
    S-DATA.request $server
    protocol_sendmsg $server "$args"
    set res [protocol_recvtypemsg $server type]
    rpc_debug "" " restype: $type result: $res\n"
    if [string compare $type "ok"]==0 { return $res }
    if [string compare $type "error"]==0 { error $res }
    error "unknown type received"
}

proc rpc_MakeClient {host port {checkFunc ""} {errFunc ""}} {
    rpc_debug "unknown" "rpc_MakeClient $host $port $checkFunc" 
    set ssap [primitives_ssapdefine "" "" \
		  "rpc_ProcessCommand \"$checkFunc\"" \
		  "rpc_Exception \"$errFunc\""]
    return [S-CONNECT.request $ssap $host $port]
}

proc rpc_MakeServer {port {loginFunc ""} {logoffFunc ""} {checkFunc ""} \
			 {errFunc ""}} {
    rpc_debug "unknown" "rpc_MakeServer $port"
    set ssap [primitives_ssapdefine $loginFunc $logoffFunc \
		  "rpc_ProcessCommand \"$checkFunc\"" \
		  "rpc_Exception \"$errFunc\""]
    set listen [socket -server "S-CONNECT.indication \{$ssap\}" $port]
    fconfigure $listen -buffering none
    atexit appendUnique "close $listen"
    atclose $listen append "atexit delete \{close $listen\}"
    return "$listen $port"
}


