#!/bin/sh
# the next line restarts using tclsh \
exec tclsh "$0" "$@"

# $Id: mime.tcl,v 1.1 1996/09/05 07:03:36 vdm Exp $
#
# MIME - Multipurpose Internet Mail Extensions

set mime_ENCODER ""
set mime_DECODER ""
set mime_FILE ""

proc mime_init { encoder decoder {mimefile ""}} {
    global mime_ENCODER mime_DECODER mime_FILE
    set mime_ENCODER "$encoder"
    set mime_DECODER "$decoder"
    set mime_FILE "$mimefile"
}

proc mime_encode_supported {} {
    global mime_ENCODER
    return [string length "$mime_ENCODER"]
}

proc mime_decode_supported {} {
    global mime_DECODER
    return [string length "$mime_DECODER"]
}

proc mime_encode name {
    global mime_ENCODER
    set f [open "|$mime_ENCODER $name" r]
    gets $f x
    while { [gets $f line] > 0 } { append x "\n$line" }
    catch {close $f}
    return $x
}

proc mime_decode {m {name ""}} {
    global mime_DECODER platform_type
    if {[info exists platform_type]} {
	if {$platform_type == "windows" && $name !={}} {
	    set f [open $name.tmp w+]
	    puts $f $m
	    close $f
	    eval exec $mime_DECODER -o $name $name.tmp
	    unlink $name.tmp
	    return
	}
    }
    if {[string length $name]} {
	eval exec $mime_DECODER -o $name {<<$m} 
    } else {
	eval exec $mime_DECODER {<<$m} {>@stdout} 
    }
}

proc mime_type m {
    return [lindex "$m" 0] 
}

proc mime_content m {
    regsub "\[^\n\]*\n" "$m" "" m
    return "$m"
}

proc mime_findtype fname {
    global mime_FILE

    set type "text/plain"

    if {$mime_FILE == {}} {
	return $type 
    }
    catch {
	if {[llength [split $fname .]] > 1} {
	    set ext [lrange [split $fname .] end end]
	    set f [open $mime_FILE r]
	    while {[gets $f line] >=0} {
		if {[lsearch -exact [lrange [split $line] 1 end] $ext] >= 0} {
		    set type [lindex [split $line] 0]
		}
	    }
	    close $f
	}
    }
    return $type
}

proc mime_findext {type {fname ""}} {
    global mime_FILE

    if {$mime_FILE == {}} {
	return $fname
    }
    if [catch \
	    {
		set f [open $mime_FILE r]

		while {[gets $f line] >=0} {
		    if {$type == [lindex [split $line] 0]} {
			set line [string trim [string range $line [string length $type] end]]
			set ext [lindex [split $line] 0]
			if {$ext != {}} { 
			    set fname "[lrange [split $fname .] end end].$ext"
			}
		    }
		}
		
		close $f
	    }] {
	return ""
    }
    return $fname
}
