# $Id: agent.tcl,v 1.1 1996/09/05 07:03:38 vdm Exp $

# global agent agentopts
# global agent_FD
# global agent_Logfile

set agent_Logfile ""

# start log

proc agent_LogOpen logfile {
    global agent_Logfile

    if {[string length [string trim $logfile]]} {
	set agent_Logfile [open $logfile w+]
	atexit appendUnique "close $agent_Logfile"
    }
}

# log a message

proc agent_Log msg {
    global agent_Logfile

    if {$agent_Logfile != {}} { 
	puts $agent_Logfile "$msg"
	flush $agent_Logfile
    }  
}

# write an agent answer

proc agent_sendMessage { fd type msg } { 
    puts $fd $type
    puts $fd [string length $msg]
    puts $fd "$msg"
    flush $fd

    agent_Log "$type"
    agent_Log "$msg"
}

# read an agent message

proc agent_receiveMessage fd {
    gets $fd type
    gets $fd len
    set msg [read $fd $len]
    gets $fd
    if [string compare $type "error"]==0 {
	error "$msg"
    }
    agent_Log "|$type"
    agent_Log "|$msg"
    return "$msg"
}

# serving loop of agent

proc agent_Serve {rfd wfd} {
    while {1} {
	set msg [agent_receiveMessage $rfd]
	if [catch {
	    set msg "[uplevel #0 $msg]"
	} err]==1 {
	    agent_sendMessage $wfd "error" "$err"
	} else {
	    agent_sendMessage $wfd "ok" "$msg"
	}
    }
}

# create local agent or a new process

proc agent_Create { {local "0" } {log ""} {outlog ""} {errlog ""}} {
    global agent_Local

    set agent_Local $local
    agent_LogOpen "$log"
    if $local {
	return "local"
    }
    set wfd [pipe]
    set rfd [pipe]

    if {[set pid [fork]]==0} {
	close [lindex $wfd 1]
	close [lindex $rfd 0]
	if {$outlog != {}} { reopen "$outlog-[pid]" w stdout w }
	if {$errlog != {}} { reopen "$errlog-[pid]" w stderr w }
	if {[config_getbool ServerLogDebug]} {
	    debug "[config_getstr AgentDebugDepth]" \
		"[config_getstr AgentLogRoot]/debug.log-[pid]"
	    debug on
	}
	agent_Serve [lindex $wfd 0] [lindex $rfd 1]
	close [lindex $wfd 0]
	close [lindex $rfd 1]
    }
    close [lindex $wfd 0]
    close [lindex $rfd 1]
    return "[lindex $rfd 0] [lindex $wfd 1] $pid"
}

# quit from agent

proc agent_Release agent {
    agent_sendMessage $agent exit

    catch { 
	close [lindex $agent 0]
	close [lindex $agent 1]
	kill SIGTERM [lindex $agent 2]
    }
}

# communication with agent

proc agent_Request {msg} {
    global agent_FD

    if [string compare $agent_FD "local"]==0 {
	if [catch { set x "[uplevel #0 $msg]" } err]==1 {
	    error [errorInfo $err]
	}
    } else {
	agent_sendMessage [lindex $agent_FD 1] "ok" "$msg"
	set x [agent_receiveMessage [lindex $agent_FD 0]]
    }
    agent_Log "return: $x"
    return $x
}

# id from socket descriptor as is used as session id

proc agent_ConnectionID { rpcfd } {
    if ![regexp \[a-z\]*(\[0-9\]+) "$rpcfd" tmp sid] { 
	error "sorry, tcl file format seems to be bad"
    }
    return $sid
}

# execute command on agent

proc agent_Call msg {
    global rpcFile

    set sid [agent_ConnectionID $rpcFile]
    if {[catch \
	     {
		 set rc [agent_Request "session $sid"]
	     } err]==1} {
	error "sorry, unable to aquire agent session $sid: $err"
    }
    if {!$rc} {
	error "sorry, your session $sid is invalid"
    }
    if {[catch \
	     {
		 set msg [agent_Request "$msg"]
	     } err]==1} {
	error "sorry, your agent failed: $err"
    }
    return $msg
}

# connect to agent

proc agent_Connect {} {
    global rpcFile

    set sid [agent_ConnectionID $rpcFile]
    agent_Log "**** SESSION $sid connected ****"
    if {[catch \
	    {
		if [agent_Request "session exists $sid"] { 
		    agent_Request "session release $sid"
		}
		set sid [agent_Request "session create $sid"]
	    } err]==1} {
	error "sorry, unable to create agent session: $err"
    }

    global rpcFile
    atclose $rpcFile appendUnique "agent_DisconnectInternal $sid"
    return $sid
}

# disconnect session

proc agent_DisconnectInternal sid {
    agent_Log "**** SESSION $sid disconnected ****"
    if {[catch {
	set rc [agent_Request "session release $sid"]
    } err]==1} { 
	error "sorry, unable to release agent session $sid: $err"
    }
    if {!$rc} {
	error "sorry, your session $sid to disconnect is invalid"
    }
}

# disconnet from agent

proc agent_Disconnect {} {
    global rpcFile

    set sid [agent_ConnectionID $rpcFile]
    agent_DisconnectInternal "$sid"
}
