# $Id: config.tcl,v 1.1 1996/09/05 07:03:38 vdm Exp $
# Values initalized with "!" must be specified in logfile and could not 
# be ommitted. Values started with an id followed by # are combined 
# path values and must not have a trailing slash.

# globals
#
# config_enabled		tag for enabled in vdmd.config
# config_disabled		tag for disbaled in vdmd.config
# config_conf			array of configuration values
# VdmSystemRoot 		root dir of vdm package
# VdmSystemLibrary 		root dir of vdm library
# VdmApplicationName		name without path of database manager
# VdmSystemInitFile 		initalization file
# VdmTclDir			directory of conventional tcl files

proc config_defaultvalues {} {
    global config_enabled config_disabled config_conf

    upvar #0 config_conf c
    set c(Version) "!"
    set c(VdmRoot) "/usr/local"
    set c(VdmLibrary) "VdmRoot#./lib"
    set c(ServerRoot) "VdmLibrary#./"
    set c(ServerTCLDir) "ServerRoot#./vdm/server"
    set c(MimeTCLDir)   "ServerRoot#./vdm"
    set c(DPTCLDir)     "ServerRoot#./vdm"	
    set c(ConsoleLogRoot) "VdmRoot#./log"
    set c(ServerLogRoot) "VdmRoot#./log"
    set c(AgentLogRoot) "VdmRoot#./log"
    set c(ExamplesDir) "VdmRoot#./Test"
    set c(ConsoleLogStderr) "$config_disabled"
    set c(ConsoleLogStdout) "$config_disabled"
    set c(ServerLogDebug) "$config_disabled"
    set c(ServerDebugDepth) "100"
    set c(ServerLogOSI5) "$config_disabled"
    set c(ServerLogRPC) "$config_disabled"
    set c(AgentLogSessions) "$config_disabled"
    set c(AgentLogStderr) "$config_enabled"
    set c(AgentLogStdout) "$config_enabled"
    set c(AgentLogDebug) "$config_disabled"
    set c(AgentDebugDepth) "100"
    set c(ServerRPC) "$config_enabled"
    set c(ServerTELNET) "$config_enabled"
    set c(PortRPC) "!"
    set c(PortTELNET) "!"
    set c(ShowErrorDetails) "$config_disabled"
    set c(ManagerPassword) "vdm"
    set c(AgentProcess) "$config_disabled"
    set c(VdmPreprocessor) "VdmLibrary#./vdmpp"
    set c(VdmDefaultDatabase) "VdmLibrary#./vdmdb-c"
    set c(ServerRCScript) "ServerTCLDir#"
    set c(DaemonOperation) "$config_enabled"
    set c(Prompt1) "vdmd > "
    set c(Prompt2) "> "
    set c(TableWidth) "4"
    set c(MimeEncoder) "VdmLibrary#./vdm/mmencode"
    set c(MimeDecoder) "VdmLibrary#./vdm/mmencode -u"
    set c(KnownTags)   ""
    set c(ExampleBrowser) "VdmRoot#./lib/vdm/vdmtest"
    set c(ExampleVdmRoot) "VdmRoot#./"
    set c(ExampleVdmBin) "VdmRoot#./bin"
}

# print information about configuration

proc config_info {} {
    global config_conf
    if ![info exists config_conf] {
	error "Configuration not read yet"
    }
    foreach x [lsort [array names config_conf]] {
	puts "$x=$config_conf($x)"
    }
}

# substitute environment var in path

proc config_substvar ref {
    global config_conf env
    if {![info exists config_conf($ref)]} {
	if ![info exists env($ref)] {
	    error "undefined path substitution \"$ref\" in initialization file or environment variable \"$ref\" not set"
	} else {
	    return "$env($ref)"
	}
    } else {
	return "$config_conf($ref)"
    }
}

# substitute in path

proc config_substpath {conf x} {
    global config_conf
    upvar $conf c

    if ![info exists config_conf($x)] {
	error "undefined path substitution \"$x\" in initalization map config_conf"
    }

    if [info exists c($x)] { set path $c($x) } else { set path "" }

    if [regexp "^(.*)#(.*)" $config_conf($x) s1 s2 s3] {
	config_substpath c $s2
	if [string length $path]==0 { set path "$s3" }
	if [string length $path]==0 { set config_conf($x) ""; return }
        if [regexp "^/.*" $path] { set config_conf($x) $path; return }
	if [regexp "^(./)(.*)" $path a1 a2 a3] { set path "$a3" }
	set config_conf($x) "$config_conf($s2)/$path"
	set c($x) ""
    } else {
    	if [string length $path]!=0 { set config_conf($x) $path }
    }

    set exp "\[$\](\[^\./\]+)(\[/\.\]|$)"
    if [regsub -all $exp $config_conf($x) "\[config_substvar \"\\1\"\]\\2" v] {
	regsub -all "//" "[subst "$v"]" "/" config_conf($x)
    }
 
    if {[string length $config_conf($x)]>1} {
	regsub "/$" $config_conf($x) "" config_conf($x)
    }
}

# substitute configuration

proc config_subst conf {
    global config_conf
    upvar $conf c
    foreach x [array names config_conf] {
	if [string compare $config_conf($x) "!"]==0 {
	    if {![info exists c($x)]} {
		error "Value for \"$x\" must be specified in configuration file"
	    }
	}
    }
    foreach x [array names c] {
	if ![info exists config_conf($x)] {
	    error "Illegal entry \"$x\" found in configuration file"
	}
    }
    foreach x [array names config_conf] {
	config_substpath c $x
	config_substpath c $x
    }
}

# read configuration file

proc config_read { configfile } {
    config_defaultvalues 
    set f [open $configfile r]  
    set line 0
    while {[gets $f s]!=-1} {
        incr line
	regsub #.* "$s" "" s1
	set s [string trim $s1]
        while {![info complete $s]} {
	    if {[gets $f s1]==-1} { 
		error "unclosed command in $configfile, line $line"
	    } else {
		incr line
		set s [subst -nocommands -novariables "$s\n$s1"]
	    }
        }
	if [string length $s] {
	    if {[regexp "^(\[^\t \]+)( |\t)(.*)" $s s1 s2 s3 s4]} {
		set s4 "[string trim "$s4"]"
		if [regexp "^\"(.*)\"$" $s4 s s1]==1 { set s4 "$s1" } 
		set conf([string trim $s2]) "$s4"
	    } else {
		error "illegals in $configfile, line $line"
	    }
	}
    }
    if [info exists conf] {
	config_subst conf
    }
    close $f
}

# get a bool configuration value

proc config_getbool var {
    global config_conf config_enabled config_disabled
    
    if [info exists config_conf($var)] { 
	if [string compare $config_conf($var) $config_enabled]==0 { return "1" }
	if [string compare $config_conf($var) $config_disabled]==0 { return "0" }
	error "Illegal configuration value for boolean \"$var\" is specified, either $config_enabled or $config_disabled is allowed" 
    }
    error "Undefined configuration option \"$var\" referred"
}

# get a string configuration value

proc config_getstr var { 
    global config_conf
  
    if [info exists config_conf($var)] { return $config_conf($var) }
    error "Undefined configuration option \"$var\" referred"
}

# auto update tclindex

proc config_Library { libdir } {
    global auto_path

    if ![file exists $libdir/tclIndex] {
	set doit 1
    } else {
	set age [file mtime $libdir/tclIndex]
	set doit 0
	if {[file mtime $libdir] > $age} {
	    set doit 1
	} else {
	    foreach file [glob $libdir/*.tcl] {
		if {[file mtime $file] > $age} {
		    set doit 1
		    break
		}
	    }
	}
    }
    if {$doit} {
	auto_mkindex $libdir *.tcl
    }
    if {[lsearch -exact $auto_path $libdir]==-1} {
	lappend auto_path $libdir
    }
}

# config fundamentals

proc config_base {} {
    global argv0 auto_path tcl_prompt1 tcl_prompt2
    global databaseloaded defaultdatabase
    global Vdmpp VdmSystemLibrary

    config_Library [config_getstr ServerTCLDir]
    config_Library [config_getstr MimeTCLDir]
    config_Library [config_getstr DPTCLDir]
    init_Source 

    set Vdmpp [config_getstr VdmPreprocessor]
    set VdmSystemLibrary [config_getstr VdmLibrary]

    set tcl_prompt1 "puts -nonewline \"[config_getstr Prompt1]\""
    set tcl_prompt2 "puts -nonewline \"[config_getstr Prompt2]\""

    set databaseloaded ""
    set defaultdatabase [config_getstr VdmDefaultDatabase]

    set rc [config_getstr ServerRCScript]
    if {[string length $rc]} {
	if {[file readable $rc]} {
	    if {[catch { uplevel \#0 source $rc } err]==1} {
		error "error executing vdm system rc file \"$rc\": $err"
	    }
	} else {
	    puts "$argv0: vdm system rc script file \"$rc\" not found or unreadable"
	    exit 1
	}
    }
}

proc config_special {} {
    mime_init "[config_getstr MimeEncoder]" "[config_getstr MimeDecoder]"  
}

# agent configuration

proc config_agent {} {
    global agent_FD

    set agentoutlog ""
    if {[config_getbool AgentLogStdout]} {
	set agentoutlog "[config_getstr AgentLogRoot]/stdout.log"
    } 
    set agenterrlog ""
    if {[config_getbool AgentLogStderr]} {
	set agenterrlog "[config_getstr AgentLogRoot]/stderr.log"
    }
    set agentlog ""
    if {[config_getbool AgentLogSessions]} {
	set agentlog "[config_getstr AgentLogRoot]/agent.log"
    }
    set agentprocess [expr ![config_getbool AgentProcess]]
  
    set agent_FD [agent_Create $agentprocess $agentlog $agentoutlog $agenterrlog]
}

proc config_server {} {
    if {[config_getbool ServerLogOSI5]} {
	set f [open [config_getstr ServerLogRoot]/osi5.log w+]
	primitives_debugon "$f"
    } 
    if {[config_getbool ServerLogRPC]} {
	set f [open [config_getstr ServerLogRoot]/rpc.log w+]
	rpc_debugon "$f"
    } 
}

# main configuration

proc config { configfile } {
    config_read "$configfile"
    config_base
    config_special
    if [config_getbool DaemonOperation] { 
	init_Daemon 
    }

    init_RootDir
    
    init_Server
    config_server
    config_agent
  
    database_init
    language_init 
    preprocess_init
  
    if {[config_getbool ServerLogDebug]} {
	debug "[config_getstr ServerDebugDepth]" "[config_getstr ServerLogRoot]/debug.log"
	debug on
    }
    console
# is never back
}

# The standard configuration for dbm

proc config_conventional {} {
    global VdmSystemRoot VdmSystemLibrary VdmApplicationName
    global VdmSystemInitFile VdmTclDir
    global env auto_path argv0
    if [info exists env(VDMROOT)] {					
	set VdmSystemRoot $env(VDMROOT)
    } else {									
	set VdmSystemRoot /usr/local						
    }										
    set VdmSystemLibrary $VdmSystemRoot/lib				
    set VdmTclDir $VdmSystemLibrary/vdm						
    set VdmSystemInitFile $VdmTclDir/${VdmApplicationName}rc			
    if [file readable $VdmSystemInitFile] {
	uplevel \#0 source $VdmSystemInitFile							
    } else {									
	puts "$argv0: system initialization file \"$VdmSystemInitFile\"	not found"	
	exit 1									
    }										
}

global VdmApplicationName
set VdmApplicationName [file tail $argv0]

set config_FILE "./vdmd.conf"

if {[string compare $VdmApplicationName "vdmd"]!=0} {
    config_conventional
} else {
    if {[info tclversion] < 7.5} {
	puts stderr "stop: tcl version is less than 7.5, no socket support available,"
	puts stderr "the server can not run."
	exit 2
    }
    if {[regexp -- "-c\[\t \]*(\[^ \t\]*)" $argv s1 s2]} {
	if [string length "$s2"]==0 { error "usage is: vdmd -c configfile" }
	set config_FILE $s2
    }
    if {![info exists config_conf]} {
	global config_enabled config_disabled
	set config_enabled "enabled"
	set config_disabled "disabled"
	if [catch {config "$config_FILE"} err] {
	    puts stderr "$err\n"
	    # puts stderr "error details:\n$errorInfo"
	    exit 2
	}
    }
}

