# $Id: html.tcl,v 1.5 1996/09/18 12:23:55 vdm Exp $

# global html_SCRIPT
# global html_VARS
# global html_REQUEST
# global html_QUERY

set html_SCRIPT "dummy.scr"

# set html_GLOBAL_COLS "4"
set html_GLOBAL_COLS [config_getstr TableWidth]

set html_GLOBAL_NAME ""

set html_GLOBAL_CENTERNAME "CenterPoint"

set html_GLOBAL_TAGS "[string toupper [config_getstr KnownTags]]"

# proc html_cgiargs_parse_split
#
# a http protocol parameter in the form a=b is put in the given tcl array 
# parmarr where a is the index b the value. Values in the form %xx represents
# system independent ascii characters and where translated here. Existing
# values are overwritten.

proc html_cgiargs_parse_split {s parmarr} {
    upvar $parmarr arr

    if [regexp (.*)=(.*) "$s" s1 idx val] {
	regsub -all % "$val" "\\x" val
	#    eval set val "$val"
	set val [subst -nocommands -novariables $val]
	set arr($idx) "$val"
	return 1
    } 
    return 0
}

# proc html_cgiargs_parse
#
# a http argument line like name=donald&nickname=donut are put in the given
# tcl array parmarr. Hint: tcl-arrays are maps. The argument parm contains
# the original CGI-GATEWAY args.

proc html_cgiargs_parse {parm {parmarr ""}} {
    if {$parmarr == {}} {
	if [uplevel \#0 info exists html_VARS] { uplevel \#0 unset html_VARS }
	upvar \#0 html_VARS arr
    } else {
	upvar $parmarr arr
    }
    while {[regexp (.*)&(.*) "$parm" s1 parm s2]} {
	if ![html_cgiargs_parse_split "$s2" arr] { return 0 }
    } 
    if ![html_cgiargs_parse_split "$parm" arr] { return 0 }
    return "1"
}

#
# serveral cgi args with the same name are given as a list
#

proc html_cgilist_parse {arg parm} {
    set res ""
    while {[regexp (.*)&(.*) "$parm" s1 parm s2]} {
	if [regexp ($arg)=(.*) "$s2" s1 idx val] {
	    regsub -all % "$val" "\\x" val
	    lappend res [subst -nocommands -novariables $val]
	}
    } 
    if [regexp ($arg)=(.*) "$parm" s1 idx val] {
	regsub -all % "$val" "\\x" val
	lappend res [subst -nocommands -novariables $val]
    } 
    return $res 
}

# html_cgiargs_make
#
# A tcl array (is a map) is encoded to a valid cgi-argument string.

proc html_cgiargs_make { {parmarr ""} } {
    if {$parmarr == {}} {
	upvar \#0 html_VARS arr
    } else {
  	upvar $parmarr arr
    }
    if ![info exists arr] { return "" }
    set s ""
    foreach x [array names arr] {
	if {$s=={}} { 
	    set s "$x=$arr($x)" 
	} else { 
	    set s "$s&$x=$arr($x)"
	} 
    }
    return $s
}

# html_cgiargs_makeexp
#
# Multiple arguments of the transformed cgi-args are expanded

proc html_cgiargs_makeexp parm {
    upvar \#0 html_VARS srcarr

    html_cgiargs_parse $parm arr
    if [info exists srcarr] { 
	foreach x [array names srcarr] {
	    if ![info exists arr($x)] { 
		set arr($x) $srcarr($x)
	    }
	} 
    }
    return [html_cgiargs_make arr]
}

# html_argset
#
# A single argument of the transformed cgi-args is changed or created

proc html_argset {var val} {
    upvar \#0 html_VARS arr 
    set arr($var) $val  
}

# html_arg
#
# A single argument of the transformed cgi-args is retrieved. 
# If the argument is not found, the default value is returned

proc html_arg { name {defval ""}} {
    upvar \#0 html_VARS arr

    if [catch {set res $arr($name)}] { return $defval }

    return $res
}

proc html_oarg { name {defval ""}} {
    upvar \#0 html_OLDVARS arr

    set res [html_arg $name $defval]

    if {"$res"=="$defval"} {
	if [catch {set res $arr($name)}] { set res $defval }
    }

    return $res
}

# html_arg_filter
#
# filter specified vars in arglist from cgi-args

proc html_arg_filter arglist {
    upvar \#0 html_OLDVARS oarr
    upvar \#0 html_VARS arr

    if ![info exists arr] return

    foreach x [array names arr] {
	if [lsearch -exact $arglist $x]==-1 {
	    set oarr($x) $arr($x)
	    unset arr($x)
	}
    }  
}

# html_arg_del
#
# remove an argument

proc html_arg_del arg {
    catch { unset arr($arg) }
}

# html_vdmicon
#
# The standard VDM Class Library icon

proc html_vdmicon {} {
    global html_SCRIPT

    return "<IMG\
 SRC=\"$html_SCRIPT?conttype=mime&adt=mimelib&impl=aux&oper=pics&comp=vdmicon\"\
 ALT=\"\"\
 ALIGN=RIGHT\
>"
}

# html_header
# 
# The standard header of a html-page is produced

proc html_header title {
  return "\
<HTML>
<HEAD><TITLE>$title</TITLE></HEAD>
<BODY>
<H1>[html_vdmicon]$title</H1>
<P>
<HR>
"
}

proc html_trailer {} {
    return "\
<P>
<HR>
<P>
</BODY></HTML>
"
}

# html_comptag
#
# produce an tag in dependance from ntag

proc html_comptag {s1 s2 tag {ntag ""}} {
    if [string compare $s1 $s2]==0 {
	return "$tag"
    } else {
	return "$ntag"
    }
}

# html_name
#
# produce an name reference

proc html_name { text {nameref ""} } {
    global html_GLOBAL_NAME

    if [string length $nameref]==0 {
	set nameref "$html_GLOBAL_NAME"
    }
    if [string length $nameref]==0 {
	return "$text"
    }
    return "<A NAME=\"$nameref\">$text</A>"
} 

# html_actionref
#
# produce an action text

proc html_actionref { {nameref ""} } {
    global html_GLOBAL_NAME

    if [string length $nameref]==0 {
	set nameref "$html_GLOBAL_NAME"
    }
    if [string length $nameref]==0 {
	return "" 
    }
    return [expr [string length "$nameref"]?"\#$nameref":""]
}

# html_actionref
#
# produce an hyperlink

proc html_action {parm text {nameref ""}} {
    global html_SCRIPT

    if {![string length $text]} return ""

    set msg "<A HREF=\"$html_SCRIPT?[html_cgiargs_makeexp $parm]"

    if {$nameref != "none"} {
	append msg "[html_actionref $nameref]"
    }
    return "$msg\">$text</A>"
}

#
# substitute an expression like [contype:][opertype.][adt.][impl.]oper
# as a reference in the vdm database. At the moment conttype is 
# either mime: or text:. If conttype is omitted, text is assumed.
# 

proc html_get_adt {adt} {
    set res $adt
    if {"$adt" == "com"} { set res [html_oarg adt com] }
    return "adt=$res"
}
proc html_get_impl {imp} {
    set res $imp
    if {"$imp" == "com"} { set res [html_oarg impl com] }
    return "impl=$res"
}
proc html_get_oper {opr} {
    if {"$opr" == "basic"} {
	return "feature=yes"
    } else {
	return "oper=$opr"
    }
}

proc html_parse_ref { ref { fulldb "0" } {assume "text"} } {
  global html_SCRIPT

    set w " \t\r\n"
    set p "\[+\]"
    set id "\[a-zA-Z0-9_\]+"     
    set op "c|c$p$p|a$p$p"

    set exp1 "^($id):($id)\\\.($id)\\\.($id)\\\.($id)$"
    set exp4 "^($id)\\\.($id)\\\.($id)\\\.($id)$"
    set exp2 "^($op)\\\.($id)\\\.($id)\\\.($id)$"
    set exp3 "^($id)\\\.($id)\\\.($id)$"
    set exp5 "^($id)\\\.($id)$"
    set exp6 "^($id)$"

    if {[regexp $exp1 $ref dummy conttype adt impl oper comp]} {
	puts  "full tag"
	set    par "conttype=$conttype&[html_get_adt $adt]"
	append par "&[html_get_impl $impl]&[html_get_oper $oper]&comp=$comp"
	return "$html_SCRIPT?[html_cgiargs_makeexp $par]"
    }

    if {$fulldb} {
	puts "mime tag"
	regexp $exp4 $ref dummy adt impl oper comp
	set    par "conttype=$assume&[html_get_adt $adt]"
	append par "&[html_get_impl $impl]&[html_get_oper $oper]&comp=$comp"
	return "$html_SCRIPT?$par"

    }

    if {[regexp $exp2 $ref dummy optype adt impl oper]} {
	puts "optype tag"
	set    par "optype=$optype&[html_get_adt $adt]"
	append par "&[html_get_impl $impl]&[html_get_oper $oper]"
    } elseif {[regexp $exp3 $ref dummy adt impl oper]} {
	puts "adt.impl.oper"
	set    par "[html_get_adt $adt]"
	append par "&[html_get_impl $impl]&[html_get_oper $oper]"
    } elseif {[regexp $exp5 $ref dummy impl oper]} {
	puts "impl oper"
	set    par "[html_get_adt com]"
	append par "&[html_get_impl $impl]&[html_get_oper $oper]"
    } elseif {[regexp $exp6 $ref dummy oper]} {
	puts "oper"
	set    par "[html_get_adt com]"
	append par "&[html_get_impl com]&[html_get_oper $oper]"
    } else {
	set par ""
    }

    return "$html_SCRIPT?[html_cgiargs_makeexp $par]"
}

# html_parse_vdmref
#
# parse an vdm reference

proc html_parse_vdmref { ref } {
    set vdmref "VDMREF"
    set href "HREF"
    set w " \t\r\n"
    set id "\[^$w=\]+"

    # tag = "ref"
    set exp \[$w\]*=\[$w\]*\"\(\[^$w\"\]*)\"

    # TAG = "ref"
    regsub "($id)$exp" "$ref" "\[string toupper \"\\1\"\]=\"\\2\"" ref
    set ref [subst $ref]

    # VDMREF = x.y.z => A HREF = cgi-ref
    set sub "$href=\"\[html_parse_ref \{\\2\}\]\"" 
    regsub "($vdmref)$exp" "$ref" "$sub" ref
    return "[subst $ref]"
}

# html_parse_image
#
# parse an image link 

proc html_parse_image { ref } {
    set w " \t\r\n"
    set ws "\[$w\]"
    set id "\[^$w=\]+" 
    set exp "(.*)(src|SRC)$ws*=$ws*\"($id)\"(.*)"
    regsub "$exp" "$ref" "\\1\\2=\"\[html_parse_ref \{\\3\}\ 1 mime]\"\\4" ref
    return "[subst $ref]"
}

# html_parse_example

proc html_parse_exampleref {runtype example} {
    global html_SCRIPT

    set par "hot=help_examples&runtype=$runtype&example=$example"
    return "$html_SCRIPT?[html_cgiargs_makeexp $par]"  
}

proc html_parse_example { ref } {
    set w " \t\r\n"
    set ws "\[$w\]"
    set id "\[^$w=\]+" 
    set txt "\[^\"\]*"
    set exp "(.*)(example|EXAMPLE)$ws+($id)$ws*=$ws*\"($txt)\"(.*)"
    regsub "$exp" "$ref" "\\1 HREF=\"\[html_parse_exampleref \{\\3\}\ \{\\4\} \]\"\\5" ref
    return [subst $ref]
}

# html_parse_tag
#
# parse an vdm tag

proc html_parse_tag {tag ref expr} {
    global html_GLOBAL_TAGS

    set utag [string toupper $tag]
    if ![string compare $utag "A"] {
	if {[regexp "^EXAMPLE" $ref]} {
	    set ref [html_parse_example "$ref"]
	} else {
	    set ref [html_parse_vdmref "$ref"]
	}
    } elseif ![string compare $utag "IMG"] {
	set ref [html_parse_image "$ref"] 
    } else {
	if {[lsearch -exact "/A $html_GLOBAL_TAGS" $utag] < 0} {
	    return "[latin1_ascii2html "$expr"]"
	}
    }
    return "<[string trim "$tag $ref"]>"
}

# html_parse
#
# parse an vdm text and set latin1 for real text

proc html_parse {html} {

    # all special tcl chars are translated to html escape sequences
    regsub -all \& $html {\&amp} html
    regsub -all \{ $html {\&ob} html
    regsub -all \} $html {\&cb} html
    regsub -all \[\[\] $html \{ html
    regsub -all \[\]\] $html \} html
    set w " \t\r\n"

    #  set exp <(\[^$w/>\]+)\[$w\]*(\[^>\]*)>
    set exp <(\[^$w<>\]+)\[$w\]*(\[^<>\]*)>
    set sub "\[html_parse_tag \{\\1\} \{\\2\} \{\\0\} \]"
    regsub -all $exp $html $sub html
  
    set ob "\[\[\]"; set nob "\[^\[\]"
    set cb "\[\]\]"; set ncb "\[^\]\]"
    set nb "\[^\]\[\]"

    regsub "^($nob+)($ob|$)" $html "\[latin1_ascii2html \{\\1\}\]\\2" html
    regsub "($cb)($ncb+)$" $html "\\1\[latin1_ascii2html \{\\2\}\]" html
    regsub -all "($cb)($nb+)($ob)" $html "\\1\[latin1_ascii2html \{\\2\}\]\\3" html

    # puts $html;puts "----";flush stdout
    set html [subst $html]
    regsub -all \{ $html {[} html
    regsub -all \} $html {]} html

    # braces are translated back to normal ascii
    # else they disapear in <pre> paragraphs
    regsub -all {\&ob} $html \{ html
    regsub -all {\&cb} $html \} html

    # puts $html;flush stdout
    return $html
}
proc html_parse {html} {

    regsub -all \{ $html "\002" html
    regsub -all \} $html "\003" html
    regsub -all \[\[\] $html "\004" html
    regsub -all \[\]\] $html "\005" html
    set w " \t\r\n"

    set exp <(\[^$w<>\]+)\[$w\]*(\[^<>\]*)>
    set sub "\[html_parse_tag \{\\1\} \{\\2\} \{\\0\} \]"
    regsub -all $exp $html $sub html
  
    set ob "\[\[\]"; set nob "\[^\[\]"
    set cb "\[\]\]"; set ncb "\[^\]\]"
    set nb "\[^\]\[\]"

    regsub "^($nob+)($ob|$)" $html "\[latin1_ascii2html \{\\1\}\]\\2" html
    regsub "($cb)($ncb+)$" $html "\\1\[latin1_ascii2html \{\\2\}\]" html
    regsub -all "($cb)($nb+)($ob)" $html "\\1\[latin1_ascii2html \{\\2\}\]\\3" html

    set html [subst -nobackslashes -novariables $html]
    regsub -all "\[\002\]" $html "{" html
    regsub -all "\[\003\]" $html "}" html
    regsub -all "\[\004\]" $html {[} html
    regsub -all "\[\005\]" $html {]} html
 
    return $html
}

proc html_hypertext { text } {
    set msg [html_parse "$text"]
    #  set msg [latin1_ascii2html "$msg"]
    set msg "\n<PRE>$msg</PRE>\n"
}

# html_table functions
#
# calculate the fitting of the table 

proc html_tablefit { list n } { 
    return [expr ([llength $list]+$n-1)/$n]
}

proc html_tablerow { list rows cols line } {
    for {set i "0"; set row ""} { $i < $cols } { incr i } {
	append row "<TD>[lindex "$list" [expr $rows*$i+$line]] "
    }
    return "$row\n"
}

# html_selectortable 
#
# table for database access

proc html_selectortable { seladt selimpl seloper seloptype \
			      adtlist impllist operlist feature general} {
    global html_GLOBAL_COLS html_GLOBAL_CENTERNAME

    set ref "adt=$seladt&impl=$selimpl&oper=$seloper"
    set specialref "feature=$feature&general=$general"
    set cols [html_arg cols]

    if [catch { set cols [expr $cols] }] { set cols "$html_GLOBAL_COLS" }

    html_argset cols "$cols"
    set n1 [html_tablefit $adtlist $cols]
    set n2 [html_tablefit $impllist $cols]
    set n3 [html_tablefit $operlist $cols]
    set nmax [misc_max [misc_max $n1 $n2] $n3]
    set ncols1 [misc_max [html_tablefit $adtlist $nmax] 1]
    set ncols2 [misc_max [html_tablefit $impllist $nmax] 1]
    set ncols3 [misc_max [html_tablefit $operlist $nmax] 1]

    set msg "<TABLE WIDTH=\"100%\" BORDER>"
    append msg "<TR><TH COLSPAN=[expr $ncols1+$ncols2+$ncols3]>
	          [html_comptag "C" "$seloptype" "ANSI C "]
		  [html_comptag "C++" "$seloptype" "CPP "]
		  [html_comptag "A++" "$seloptype" "AUTOMEM "]"  
    append msg "[html_name "Database Contents"]"
    append msg "<TR>"
    append msg "<TH COLSPAN=$ncols1><H3>ADTs</H3>"
    append msg "<TH COLSPAN=$ncols2><H3>Implementations</H3>"
    append msg "<TH COLSPAN=$ncols3><H3>Operations</H3>"
    append msg "<TR><TH COLSPAN=$ncols1>"

    if {$seladt == "Predefined"} { 
	append msg "$seladt"
    } else { 
	append msg "[html_action "$ref&general=yes" "$seladt"]"
    }

    append msg "    <TH COLSPAN=$ncols2>[html_action "$ref&feature=yes" "$selimpl"]
                  <TH COLSPAN=$ncols3>$seloper"

    lappend adtrefs "[html_action "adt=Predefined" Predefined]"

    for {set i "0"} {$i < [llength $adtlist]} { incr i } {
	set adt [lindex $adtlist $i]
	lappend adtrefs "[html_action "adt=$adt" $adt]"
    }
    for {set i "0"; set implrefs ""} { $i < [llength $impllist] } {incr i} {
	set impl [lindex $impllist $i]
	lappend implrefs "[html_action "adt=$seladt&impl=$impl" $impl]" 
    }
    for {set i "0"; set operrefs ""} { $i < [llength $operlist] } { incr i} {
	set oper [lindex $operlist $i]
	lappend operrefs "[html_action "adt=$seladt&impl=$selimpl&oper=$oper" $oper]"
    }
    for {set i "0"} {$i < $nmax } { incr i } {
	append msg "<TR>"
	append msg "[html_tablerow $adtrefs $nmax $ncols1 $i]"
	append msg "[html_tablerow $implrefs $nmax $ncols2 $i]"
	append msg "[html_tablerow $operrefs $nmax $ncols3 $i]"    
	append msg "\n"
    } 
    append msg "</TABLE>" 
    append msg "<HR>"
    append msg "<TABLE WIDTH=\"100%\" BORDER>"
    append msg "<TR><TH ROWSPAN=2>Options"
    append msg "<TH>Mode:
		  <TH COLSPAN=2>[html_action "optype=C&$ref" "ANSI C Prototypes"]
                  <TH COLSPAN=2>[html_action "optype=C++&$ref" "C++ member functions"] 
                  <TH COLSPAN=2>[html_action "optype=A++&$ref" "automem member functions"]"
    append msg "<TR><TH>Display:
                  <TH>[html_action "hot=enlarge&$ref" "Enlarge"]
		  <TH>[html_action "hot=reduce&$ref" "Reduce"]"

    if [string compare "[html_arg text]" "bottom"]==0 {
	append msg "  <TH>[html_action "text=top&$ref&$specialref" "Top-Text" none]"
    } elseif [string compare "[html_arg text]" "top"]==0 {
	append msg "  <TH>[html_action "text=topcenter&$ref&$specialref" "Top-Center" \
	$html_GLOBAL_CENTERNAME]" 
    } else {
	append msg "  <TH>[html_action "text=bottom&$ref&$specialref" "Bottom-Text" none]"
    }

    append msg "    <TH>Other:
                  <TH>[html_action "hot=setup&$ref" "Setup"]
		  <TH>[html_action "hot=help&$ref" "Help"]"
    append msg "</TABLE>"
}

# html_vdmentry 
#
# get component text from the database

proc html_vdmentry {adt impl oper comp} {
    select filteradts
    select filterimpl 
    select filteroper
    select filtercomp

#  select filteradts $adt
#  select filterimpl $impl
#  select filteroper $oper
#  select filtercomp $comp
    return [selcomp $adt $impl $oper $comp]
}

# html_hot_mime 
#
# get mime contents

proc html_hot_mime {} {
    #  puts "mime here"
    html_arg_filter "optype adt impl oper comp"
    return "[html_vdmentry "[html_arg adt]" "[html_arg impl]" "[html_arg oper]" "[html_arg comp]"]"
}

# html_hotquery
#
# query the database

proc html_hot_query { {conttype ""} } {
    html_arg_filter "text optype adt impl oper cols feature general"
    html_argset optype "[string toupper [html_arg optype]]"
    set seloptype [html_arg optype]

    if [string length $seloptype]==0 { 
	set seloptype "A++"
    } 

    set seladt [html_arg adt]
    set selimpl [html_arg impl]
    set seloper [html_arg oper]
    set feature [html_arg feature]
    set general [html_arg general]
    html_arg_filter "optype cols text"
    set impllist ""
    set operlist ""
    setlanguage "$seloptype"
    set adtlist [adts]  
 
    if [string length "$seladt"] { 
	set impllist [implementations $seladt]
	if [string length "$selimpl"] {
	    set operlist [operations $seladt $selimpl]
	}
    }
 
    if {$seladt == "Predefined"} {
	set comptext [docu_predefined]
    } elseif {$general=="yes"} {
	set comptext [docu_general "$seladt"]
    } elseif {$feature=="yes"} {
	set comptext [docu_feature "$seladt" "$selimpl"]
    } else { 
	if [string length "$seloper"] {
	    set comptext [docu_collect $seladt $selimpl $seloper]
	} else {
	    set comptext ""
	}
    }
    upvar html_REQUEST request
    upvar html_QUERY query
  
    set msg "[html_header {VDM Database Info}]"

    #  <H1>VDM Help</H1>
    #  append msg "ptype: $request, query: $query, opt: $seloptype"
    #  filtered: [html_cgiargs_make]

    set selector "[html_selectortable $seladt $selimpl $seloper $seloptype \
	$adtlist $impllist $operlist $feature $general]"

    if [string length $comptext] { 
	set hypertxt "[html_hypertext $comptext]"
    } else {
	set hypertxt ""
    }
    set delim [expr [string length $seloper]?"<HR>":""]
    
    if [string compare "[html_arg text]" "bottom"]==0 {
	append msg "$selector$delim$hypertxt"	
    } else {
	append msg "$hypertxt$delim$selector"
    } 

    append msg "[html_trailer]"
    return "$msg"
}

proc html_hot_colcalc {cols how} {
    global html_GLOBAL_COLS
    if [catch { set cols [expr $cols+($how)] }] { set cols $html_GLOBAL_COLS }
    if {$cols < 1} { set cols 1 }
    return $cols
}

proc html_hot_colmodify how {
    set cols "[html_hot_colcalc "[html_arg cols]" "$how"]"
    html_argset cols "$cols"
    return "$cols"
}

proc html_form_text { size var {val ""}} { 
    return "<INPUT TYPE=\"text\" NAME=\"$var\" VALUE=\"$val\" SIZE=\"$size\">"
}

proc html_form_hidden { var val } {
    return "<INPUT TYPE=\"hidden\" NAME=\"$var\" VALUE=\"$val\">"
}

proc html_form_options {options {selopt ""}} {
    for {set i 0;set msg ""} {$i < [llength "$options"]} {incr i} {
	set s "[lindex $options $i]"
	if [lsearch -exact "$selopt" "$s"]!=-1 { 
	    set opt "<OPTION SELECTED VALUE=\"$s\">$s"
	} else {
	    set opt "<OPTION VALUE=\"$s\">$s"
	}
	append msg "$opt\n"
    }
    return $msg
}

proc html_form_list {name entries {selopt ""} {height "0"}} {
    if $height { 
	set size "SIZE=$height" 
    } else {
	if {$entries=={}} return "" 
	set size "" 
    }
    set msg "<SELECT NAME=\"$name\" MULTIPLE $size>
  [html_form_options "$entries" "$selopt"]
  </SELECT>"

    return $msg
}

proc html_help_search_full {} {
    global html_QUERY

    html_arg_filter "adts impls opers search"
    set s_adts [html_cgilist_parse adts $html_QUERY]
    set s_impls [html_cgilist_parse impls $html_QUERY]
    set s_opers [html_cgilist_parse opers $html_QUERY]
    set msg "[html_header {VDM Fulltext Search}]"

    set l [search_full [html_arg search] $s_adts $s_impls $s_opers]
    append msg "<H2>[llength $l] Targets(s) found for \"[html_arg search]\"</H2><HR>"
    foreach ref $l {
	append msg "<A HREF=\"[html_parse_ref $ref]\"> 
    [lrange [split $ref .] 0 end] </A><BR>" 
    }
    append msg "[html_trailer]"
    return $msg
}

proc html_help_search_commit {} {
    global html_QUERY

    html_arg_filter "adts impls opers"
    set s_adts [html_cgilist_parse adts $html_QUERY]
    set s_impls [html_cgilist_parse impls $html_QUERY]
    set s_opers [html_cgilist_parse opers $html_QUERY]
    set msg "[html_header {VDM Topic Search Results}]"
    set l [search_topic $s_adts $s_impls $s_opers]
    append msg "<H2>[llength $l] Topic(s) found</H2>
  <HR>
  <FORM METHOD=\"GET\">
  Your search: [html_form_text 40 search]
  [html_form_hidden hot help_search_full]
  <INPUT TYPE = \"submit\" VALUE=\"Start Fulltext Search\">"

    foreach a $s_adts { append msg [html_form_hidden adts $a] }
    foreach i $s_impls { append msg [html_form_hidden impls $i] }
    foreach o $s_opers { append msg [html_form_hidden opers $o] }
    append msg " </FORM><HR>"
    foreach ref $l {
	append msg "<A HREF=\"[html_parse_ref $ref]\"> 
    [lrange [split $ref .] 0 end]</A><BR>" 
    }
    append msg "[html_trailer]"
    return $msg
}

proc html_help_search {} {
    global html_QUERY

    html_arg_filter "adts impls opers"

    set adts [set impls [set opers ""]]
    search_alltopics adts impls opers
    set s_adts [html_cgilist_parse adts $html_QUERY]
    set s_impls [html_cgilist_parse impls $html_QUERY]
    set s_opers [html_cgilist_parse opers $html_QUERY]
    set msg "\
[html_header {VDM Search}]
<FORM METHOD=\"GET\">
Please mark the topics you want to search for.
After topic search, a fulltext search can be done.
Be patient, fulltext searching for many topics 
can take a long while. 
Search values can be entered
as regular expressions.
<P>
[html_form_list adts $adts $s_adts 15]
[html_form_list impls $impls $s_impls 15]
[html_form_list opers $opers $s_opers 15]
<HR>
[html_form_hidden hot help_search_commit]
<INPUT TYPE = \"submit\" VALUE=\"Start Topic Search\">
<INPUT TYPE = \"reset\" VALUE=\"Reset Choice\">
</FORM>"

 append msg [html_trailer]

 return $msg
}

proc html_help_usage {} {
    return "\
[html_header {VDM Help Usage}]
<P>
<UL>
<LI>
The most left column titled \"ADTs\" shows the names of the abstract datatypes
available in the database. Select one of the entries by a left button
double click of the mouse. The currently selected ADT is shown at the top
of the column.
</LI>
<P>
<LI>
The list in the middle titled \"Implementations\" will then show the
implementations available for the selected ADT. Select one of these
implementations. The currently selected implementation is shown at the 
top of the list.
</LI>
<P>
<LI>
The right list titled \"Operations\" shows the operations available for the
selected implementation. When selecting one of them by a left button double
click, a text window will appear with information about the operation.
</LI>
<P>
<LI>
As text all detailed information about ADTs, implementations 
and operation is collected. 
</LI>
<P>
<LI>
The \"ADTs\" column has the entry \"Predefined\" for showing the predefined
datatypes. 
</LI>
<P>
<LI>
A click on the highlighted currently selected ADT shows general documentation 
about this ADT.
</LI>
<P>
<LI>
A click on the highlihted currently selected Implementation shows
documentation about the implementation strategie, time and space constraints
and implementation parameters.
</LI>
<P>
<LI>
Selecting the \"ANSI C Prototypes\" button 
shows all operations available for the C interface of the library.
Selectin the \"C++ member functions\" shows all operations for C++ and their
calling conventions. Selecting the \"C++ automem member functions\" shows all operations for C++
with automatic memory management for the generated classes. This interface
is simpler and more comfortable and reliable than the pure C++ interface.
It is highly recomended to use this option with C++.
</LI>
<P>
<LI>
The text marked \"enlarged\" and \"reduced\" configures the width of 
the table.
</LI>
<P>
<LI>
The options \"Top-Text\", \"Button-Text\" and \"Top-Center\" are 
responsible for the position of the text. \"Help\" entries the 
Help-System and \"Setup\" allows to configure the the table width
more comfortable.
</LI>
</UL>
[html_trailer]
"
}

proc html_help {} {
  return "\
[html_header {VDM Help Usage}]
  <H3>
  Contents
  </H3>
<P>
  <UL>
  <LI><B>
  [html_action "hot=help_usage" "Usage Help"]
  </B>
<P>
  <LI><B>
  [html_action "conttype=raw&adt=mimelib&impl=aux&oper=help&comp=common" "Common Help"]
  </B>
<P>
  <LI><B>
  [html_action "hot=help_examples" "VDM Examples" "none"]
  </B>
<P>
  <LI><B>
  [html_action "hot=help_search" "Searching the Library" "none"]
<P>
  </B>
  </UL>
  [html_trailer]
"
}

proc html_setup {} {
    set cols [html_arg cols]
    return "\
[html_header {VDM Setup}]
<H3> Justify COLUMNS </H3>
<H4>Current maximum columns in table: [html_arg cols]</H4>
<BR>
<H4>
[html_action "hot=setup&cols=[html_hot_colcalc $cols +1]" "+1"] 
[html_action "hot=setup&cols=[html_hot_colcalc $cols +5]" "+5"]
[html_action "hot=setup&cols=[html_hot_colcalc $cols +10]" "+10"] 
columns more</H4>
<BR>
<H4>
[html_action "hot=setup&cols=[html_hot_colcalc $cols -1]" "-1"]
[html_action "hot=setup&cols=[html_hot_colcalc $cols -5]" "-5"]
[html_action "hot=setup&cols=[html_hot_colcalc $cols -10]" "-10"] 
columns less</H4>
<HR>
<H3>[html_action "hot=" "Accept Options"]</H3>
[html_trailer]
"
}

proc html_runexamples {} {
    global env html_SCRIPT

    html_arg_filter "hot runtype example"
    
    set browser [config_getstr ExampleBrowser]
    set env(REQUEST_METHOD) "GET"
    set env(QUERY_STRING) "[html_cgiargs_make]"
    set env(SCRIPT_NAME) "$html_SCRIPT"
    set parms "-TESTHOME [config_getstr ExamplesDir]"
    append parms " -VDMROOT [config_getstr ExampleVdmRoot]"
    append parms " -VDMBIN [config_getstr ExampleVdmBin]"

    if [catch {
	set msg "[exec $browser $parms]"
    } err]==1 {
	error "ERROR EXECUTING Example Browser $browser:\n \
           [errorInfo $err]"
    } else { return $msg }
}


# html_main
#
# main entry for html processing

proc html_main {request query script} {
    global html_SCRIPT html_VARS html_REQUEST html_QUERY 
    global html_GLOBAL_NAME html_GLOBAL_CENTERNAME

    set html_SCRIPT $script
    set html_REQUEST $request
    set html_QUERY $query
    set html_GLOBAL_NAME ""

    html_cgiargs_parse "$query"

    loaddb

    set conttype [html_arg conttype]
    set hot [html_arg hot]
    set text [html_arg text]

    if {$text != "top" && $text != "topcenter"} {
	html_argset text "bottom" 
    }

    if {$text == "topcenter"} {
	set html_GLOBAL_NAME html_GLOBAL_CENTERNAME
    }

    switch "$hot" {
	enlarge { html_hot_colmodify "+1" }  
	reduce { html_hot_colmodify "-1" }
    }

    if [string compare "$conttype" "mime"]==0 {
	set msg "mime\n[html_hot_mime]"
    } elseif [string compare "$conttype" "raw"]==0 {
	set msg "html\ntext/html; version=3.0\n"
	set cont "\
[html_vdmentry stdlib aux setup com]
[html_vdmentry stdlib sign setup hyper]
[html_hot_mime]
"

 	append msg "\
[html_header {VDM Documentation Rules}]
[html_hypertext [preprocessDocu 1 $cont]]
[html_trailer]
"  
    } else {
	set msg "html\ntext/html; version=3.0\n"
	switch "$hot" { 
	    setup   { append msg [html_setup] }
	    help    { append msg [html_help] }
	    help_usage { append msg [html_help_usage] }
	    help_examples { set msg [html_runexamples] }
	    help_search { append msg [html_help_search] }
	    help_search_full { append msg [html_help_search_full] }
	    help_search_commit { append msg [html_help_search_commit] }
	    default { append msg "[html_hot_query]" }
	}
    }
    return "$msg"
}

