# $Id: server.tcl,v 1.1 1996/09/05 07:03:40 vdm Exp $
#
# rpc and telnet server main
# 

# global server_RPCfd
# global server_SOCKfd
# global server_TELNETfd
# global server_SOCKclpids

# global server_TELNETmanager

# global status_serverRPCup
# global status_serverSOCKup
# global status_serverTELNETup

set server_TELNETmanager ""

# alert

proc server_Alert { who msg } {
    if [catch {set date [exec date]}]==1 {set date "date not found"}
    puts stderr "\n$who: $date"
    puts stderr "\t$msg"
    flush stderr
}

# error detection

proc server_Error { who err } {
    puts stderr "\n$who: [exec date]"
    puts stderr "\tTCL FAILURE CATCH : $err"
    if [catch {set loc [info level -1]}] { set loc "console"}
    puts stderr "\tLOCATION: $loc"
    flush stderr
}

# check rpc commands for security

proc server_RPCCmdCheck { cmd args } {
    if {[string compare $cmd server_GetVar] == 0} {
	return
    }
    if {[string compare $cmd eval]==0} { return -code continue }
    if {[security_Allowed $cmd]} {
	return
    }
    #	we return here and make no check
    #	this is an example for intruder detection
    return
    #	set f [open "/tmp/intruder.log" a+]
    #	puts $f "$cmd $args"
    #	close $f
	
    return -code break
}

# login for rpc

proc server_RPCLogin { channel host port } {
    server_Alert "rpcserver" "Connection for $host:$port $channel granted"
}

# logout from rpc

proc server_RPCLogoff { logofftype channel } {
    if {$logofftype == "abort"} {
	server_Alert "rpcserver" "Connection for $channel aborted"
    } else {
	server_Alert "rpcserver" "Connection for $channel cleared"
    }
}

# execption, bad abort of client

proc server_RPCException { err channel } {
    server_Alert "rpcserver" "Exception on channel $channel"
}

# shut down rpc-server

proc server_RPCDown {} {
    global server_RPCfd
    global status_serverRPCup
    if !$status_serverRPCup {
	server_Alert "rpcserver" "server already down"	
	return
    }
    server_Alert "rpcserver" "server shutdown"	
    set status_serverRPCup "0"
    close [lindex $server_RPCfd 0]
    atexit delete server_RPCDown
    return
}

# startup rpc-server

proc server_RPCInit port {
    global server_RPCfd
    global status_serverRPCup
    set server_RPCfd \
	[rpc_MakeServer $port server_RPCLogin server_RPCLogoff\
	     server_RPCCmdCheck server_RPCException]
    atexit appendUnique server_RPCDown
    server_Alert "rpcserver" "server is up (port $port)"
    set status_serverRPCup "1"
    return
}

# shutdown telnet-connection

proc server_TELNETConnectionDown  { peer orgstdin orgstdout orgstderr } {
    dup $orgstdin stdin
    dup $orgstdout stdout
    dup $orgstderr stderr
    server_Alert "telnetserver" "Connection for $peer closed"
}

# for telnet manager only

proc quit {} {
    global server_TELNETmanager
    catch { 
	close [lindex $server_TELNETmanager 0]
    }
    set server_TELNETmanager [lrange $server_TELNETmanager 1 end]
    puts $server_TELNETmanager
    return ""
}


# kill telnet negotiation, see RFC0854 for information

proc server_TELNETDontWontNegot {peer msg} {
    regsub -all "\[\\\]" "$msg" "\\" msg
    regsub -all "(\[\[\])" "$msg" "\\\[" msg
    regsub -all "(\[\]\])" "$msg" "\\\]" msg
    regsub -all "(\xff\xfb)(.)" "$msg" "\[puts $peer \"\xff\xfc\\2\"\]" msg
    regsub -all "(\xff\xfd)(.)" "$msg" "\[puts $peer \"\xff\xfe\\2\"\]" msg
    subst -novariables "$msg"
}

proc server_TELNETKillNegot {msg} {
    regsub -all "(\xff\xfb.)|(\xff\xfc.)|(\xff\xfd.)|(\xff\xfe.)" "$msg" "" msg 
    regsub -all "\xff." "$msg" "" msg
    return $msg
}

# telnet connection comes in

proc server_TELNETConnection { peer iaddr port } {
    global server_TELNETmanager
    server_Alert "telnetserver" "Connection for $iaddr:$port granted"

    fconfigure $peer -buffering none 
    puts $peer "Please authentify as manager for the vdm database"
    puts -nonewline $peer "password:" 
    set answ "[gets $peer]"
    server_TELNETDontWontNegot $peer "$answ"
    if {"[server_TELNETKillNegot "$answ"]" != [config_getstr ManagerPassword]} {
	server_Alert "telnetserver" "Connection for $peer denied"
	puts $peer "Access denied, connection refused"
	close $peer
	return
    }
    set savstdin [dup stdin]
    set savstdout [dup stdout]
    set savstderr [dup stderr]
    dup $peer stdin
    dup $peer stdout
    dup $peer stderr
    puts stderr "\nWelcome VDM DATABASE MANAGER (type quit to exit)\n"
    puts -nonewline stderr "Console:"
    set closecmd "server_TELNETConnectionDown $peer $savstdin $savstdout $savstderr"
    atclose $peer appendUnique "$closecmd"
    set server_TELNETmanager [linsert "$server_TELNETmanager" 0 $peer]
}

# shutdown telnet-server

proc server_TELNETDown {} {
    global server_TELNETfd
    global status_serverTELNETup
    if !$status_serverTELNETup {
	server_Alert "telnetservr" "server already down"	
	return
    }
    set status_serverTELNETup "0"
    server_Alert "telnetserver" "server shutdown"
    close [lindex $server_TELNETfd 0]
}

# init telnet server

proc server_TELNETInit port {
    global server_TELNETfd
    global status_serverTELNETup
    set server_TELNETfd [socket -server server_TELNETConnection $port]
    set status_serverTELNETup "1"
    atexit append server_TELNETDown
    server_Alert "telnetserver" "server is up (port $port)"
}

# get a var from server

proc server_GetVar name {
    global $name
    return [eval return \$$name]
}


