# $Id: entryselectbox.tcl,v 1.5 1996/09/05 07:03:43 vdm Exp $

# create a select box for an entry
# parameters can be given as option arguments
# -title	: window title
# -text		: operation description
# -entryname	: entry name
# -entrytext	: initial entry text
# -entrywidth   : entry width
# -borderwidth	: borderwidth
# -answer	: the name of the global variable that returns the result
#		  and for that an tk_wait is executed

# the count EntrySelectBoxCnt makes window names unique to allow
# more than one entry box at a time
# result is the selected text, in case of an abort the empty string

#-------------------------------------------------------------------------------

set EntrySelectBoxCnt 0

proc EntrySelectBox args {
    global EntrySelectBoxCnt

    setopt title "Select Text"

    incr EntrySelectBoxCnt 1

    set answer "EntrySelection$EntrySelectBoxCnt"
    global $answer

    set window ".entryselectbox$EntrySelectBoxCnt"

    set res [ eval topLevelWindow				\
			-window		$window			\
			-title		[list $title]		\
			-proc		createEntrySelectBox	\
			-answer		$answer		\
			$args		]
    unset $answer
    return $res
}

#-------------------------------------------------------------------------------

proc createEntrySelectBox {window args} {
    setopt answer		""
    setopt borderwidth		1
    setopt text			"select text"
    setopt entrytext		""
    setopt entryname		"text"
    setopt entrywidth		30
    setopt state		normal

    global $answer

    frame  $window.headline
    button $window.headline.ok	-text		ok		\
				-borderwidth	$borderwidth	\
				-width		5		\
				-command	"setResultEntrySelect $window $answer"

    button $window.headline.abort				\
				-text		abort		\
				-borderwidth	$borderwidth	\
				-width		5		\
				-command	"set $answer \"\""

    label  $window.headline.tit	-text		$text		\
				-anchor		center		\
				-borderwidth	$borderwidth	\
				-relief		raised

    frame $window.op		-borderwidth	$borderwidth	\
				-relief		raised

    frame $window.op.text
    label $window.op.text.l	-text		"$entryname : "	\
				-anchor		e
    entry $window.op.text.r	-borderwidth	$borderwidth	\
				-relief		sunken		\
				-width		$entrywidth

    pack $window.headline 	-side top	-fill both
    pack $window.headline.ok	-side left
    pack $window.headline.abort	-side right
    pack $window.headline.tit	-side left	-fill both -expand yes
    pack $window.op		-side top	-fill both

    pack $window.op.text	-side top	-fill x    -expand yes
    pack $window.op.text.l	-side left	-fill both
    pack $window.op.text.r	-side right	-fill x    -expand yes

    fillTextEntry $window $entrytext

    $window.op.text.r configure -state $state 

    bind $window.op.text.r <Return> "setResultEntrySelect $window $answer"
    bind $window.op.text.r <Delete> "$window.op.text.r delete 0 end"

    bind $window.op.text.r <ButtonRelease-2> "fillTextEntryWithSelection $window"
}

#-------------------------------------------------------------------------------

proc setResultEntrySelect {window answer} {
    global $answer
    set lanswer [$window.op.text.r get]

    if {"$lanswer" == ""} {
        ErrorMessage "Please, select a text or abort"
        return
    }
    set $answer $lanswer
}

#-------------------------------------------------------------------------------

proc fillTextEntry {window text} {
    $window.op.text.r delete 0 end
    $window.op.text.r insert end $text
}

proc fillTextEntryWithSelection window {
    set str ""
    catch {set str [selection get]}
    set str [string trim $str]
    if {"$str" == "" } return
    fillTextEntry $window $str
}

#-------------------------------------------------------------------------------
