# fileselectbox.tcl,v 1.9 1995/03/08 12:08:28 vdm Exp

# create a file select box
# parameters can be given as option argument
# -text		: operation description
# -path		: directory (default : . )
# -file		: initial file entry
# -geometry	: file list geometry
# -entrywidth   : entry width
# -borderwidth	: borderwidth

# result is the selected file of in case of an abort the empty string

#-------------------------------------------------------------------------------

proc FileSelectBox args {
    setopt title "Select File"

    eval topLevelWindow	-window		.fileselectbox		\
			-title		[list $title]		\
			-proc		createFileSelectBox	\
			-answer		FileSelection		\
			$args
}

#-------------------------------------------------------------------------------

proc createFileSelectBox {window args} {
    setopt answer		""
    setopt borderwidth		1
    setopt text			"select file"
    setopt path			[pwd]
    setopt file			""
    setopt geometry		"20x30"
    setopt entrywidth		30

    global $answer

    frame  $window.headline
    button $window.headline.ok	-text		ok		\
				-width		5		\
				-borderwidth	$borderwidth	\
				-command	"setResultFileSelect $window $answer"

    button $window.headline.abort				\
				-text		abort		\
				-width		5		\
				-borderwidth	$borderwidth	\
				-command	"set $answer \"\""

    label  $window.headline.tit	-text		$text		\
				-anchor		center		\
				-borderwidth	$borderwidth	\
				-relief		raised

    frame $window.op		-borderwidth	$borderwidth	\
				-relief		raised

    frame $window.op.home
    label $window.op.home.l	-text		"home : "	\
				-width		10		\
				-anchor		e

    entry $window.op.home.r	-borderwidth	$borderwidth	\
				-relief		sunken		\
				-width		$entrywidth	\
				-state		disabled

    frame $window.op.cwd
    label $window.op.cwd.l	-text		"cwd : "	\
				-width		10		\
				-anchor		e

    entry $window.op.cwd.r	-borderwidth	$borderwidth	\
				-relief		sunken		\
				-width		$entrywidth	\
				-state		disabled

    frame $window.op.dir
    label $window.op.dir.l	-text		"directory : "	\
				-width		10		\
				-anchor		e
    entry $window.op.dir.r	-borderwidth	$borderwidth	\
				-relief		sunken		\
				-width		$entrywidth	\
				-state		disabled

    frame $window.op.file
    label $window.op.file.l	-text		"file : "	\
				-width		10		\
				-anchor		e

    entry $window.op.file.r	-borderwidth	$borderwidth	\
				-relief		sunken		\
				-width		$entrywidth

    frame $window.fl
    listbox $window.fl.files	-borderwidth	$borderwidth	\
				-relief		raised		\
				-setgrid	yes		\
				-yscroll	"$window.fl.scroll set"

    global tk_version
    if {$tk_version >= 4.0} {
        set wh          [split $geometry x]
        set width       [lindex $wh 0]
        set height      [lindex $wh 1]
        $window.fl.files configure -width $width -height $height
    } else {
        $window.fl.files configure -geometry $geometry
    }

    scrollbar $window.fl.scroll -command	"$window.fl.files yview"	\
				-borderwidth    $borderwidth

    pack $window.headline 	-side top	-fill both
    pack $window.headline.ok	-side left
    pack $window.headline.abort	-side right
    pack $window.headline.tit	-side left	-fill both -expand yes
    pack $window.op		-side top	-fill both

    pack $window.op.home	-side top	-fill x    -expand yes
    pack $window.op.home.l	-side left	-fill both
    pack $window.op.home.r	-side right	-fill x    -expand yes

    pack $window.op.cwd		-side top	-fill x    -expand yes
    pack $window.op.cwd.l	-side left	-fill both
    pack $window.op.cwd.r	-side right	-fill x    -expand yes

    pack $window.op.dir		-side top	-fill x    -expand yes
    pack $window.op.dir.l	-side left	-fill both
    pack $window.op.dir.r	-side right	-fill x    -expand yes

    pack $window.op.file	-side top	-fill x    -expand yes
    pack $window.op.file.l	-side left	-fill both
    pack $window.op.file.r	-side right	-fill x    -expand yes

    pack $window.fl		-side top	-fill both -expand yes
    pack $window.fl.scroll	-side right	-fill y
    pack $window.fl.files	-side left	-fill both -expand yes

    fillFileSelectBox   $window $path
    fillFileSelectEntry $window $file

    bind $window.op.home.r <Double-Button-1> "fillFileSelectEntry $window \$env(HOME)"

    bind $window.op.cwd.r  <Double-Button-1> "fillFileSelectEntry $window \[pwd\]"
    bind $window.op.file.r <Return> "returnFileSelectEntry $window \[$window.op.file.r get\] $answer"
    bind $window.op.file.r <Delete> "$window.op.file.r delete 0 end"

    bind $window.op.file.r <ButtonRelease-2> "fillFileSelectEntryWithSelection $window"
    bind $window.fl.files  <Double-Button-1> "fillFileSelectEntryWithSelection $window"
}

#-------------------------------------------------------------------------------

proc setResultFileSelect {window answer} {
    global $answer

    set dir  [$window.op.dir.r  get]
    set file [$window.op.file.r get]

    if [string compare $file ""]==0 {
        ErrorMessage "Please, select a file or abort"
        return
    }
    set $answer $dir/$file
}

#-------------------------------------------------------------------------------

proc setEntryValue {entry value} {
    $entry configure -state normal
    $entry delete 0 end
    $entry insert end $value
    $entry configure -state disabled
}

proc fillFileSelectBox {window dir} {
    if [file isdirectory $dir]==0 {
        ErrorMessage "no such directory:

$dir
"
        return
    }

    set cwd [pwd]
    cd $dir


    global env
    setEntryValue $window.op.home.r $env(HOME)
    setEntryValue $window.op.cwd.r $cwd
    setEntryValue $window.op.dir.r [pwd]

    $window.op.file.r delete 0 end
    $window.fl.files delete 0 end
     
    foreach i [lsort [glob -nocomplain ".*" "*"]] {
        set j [filterDotFiles $i]
        if {"$j" == ""} continue
        if [file isdirectory $j] {set j "$j/" }
        $window.fl.files insert end $j
    }
    cd $cwd
}

proc filterDotFiles name {
    if {"$name" == ".."} {return $name}
    if {"[string index $name 0]" == "."} {return ""}
    return $name
}

proc removeTrailingSlash name {
    set namelength [string length $name]
    if {$namelength == 0} {return $name}
    incr namelength -1
    if {"[string index $name $namelength]"!="/"} {return $name}
    return [string range $name 0 $namelength]
}

proc fillFileSelectEntry {window entry} {
    set dir  [$window.op.dir.r get]
    set entry [removeTrailingSlash $entry]
    set path "$dir/$entry"

    if {"[string index $entry 0]" == "/"} {
       set path $entry
    } elseif {"[string range $entry 0 1]" == "./"} {
        set path "[pwd]/$entry"
    }

    if [file isdirectory $path] {
        fillFileSelectBox $window $path
    } else {
        set path1 [file dirname $path]
        set file1 [file tail $path]
        if [string compare $file1 $entry]==0 {
            $window.op.file.r delete 0 end
            $window.op.file.r insert end $entry
        } else {
            fillFileSelectBox   $window $path1
            fillFileSelectEntry $window $file1
        }
    }
}

proc fillFileSelectEntryWithSelection window {
    set str ""
    catch {set str [selection get]}
    set str [string trim $str]
    if {"$str" == "" } return
    fillFileSelectEntry $window $str
}

proc returnFileSelectEntry {window entry answer} {
    global $answer

    set dir  [$window.op.dir.r  get]
    set file [$window.op.file.r get]

    fillFileSelectEntry $window $entry

    if {"$dir" == "[$window.op.dir.r  get]" && "$file" == "[$window.op.file.r get]" } {
        setResultFileSelect $window $answer
    }
}
#-------------------------------------------------------------------------------

proc fileAccessable file {
    if {"$file" == ""} {return ""}
    if { ! [file isfile $file] ||
         ! [file readable $file] } {
        ErrorMessage "Cannot access file $file"
        return ""
    }
    return $file
}

#-------------------------------------------------------------------------------

proc fileWritable file {
    if {"$file" == ""} {return ""}
    if { [file isfile $file] && 
         [file writable $file] } {return $file}
    if { [file exists $file] } {
        ErrorMessage "Cannot write to existing file $file"
        return ""
    }
    if [catch {exec echo > $file}] {
        ErrorMessage "Cannot create file $file, permission denied"
        return ""
    }
    return $file
}

#-------------------------------------------------------------------------------

proc fileExists file {
    if {"$file" == ""} {return ""}
    if { [file isfile $file] &&
         [file writable $file] } {
        if {[EntrySelectBox -title "overwrite"	\
		-text "overwrite file"		\
		-entryname file			\
		-entrytext $file		\
		-state disabled] == ""} {return ""}

    }
    return $file
}

#-------------------------------------------------------------------------------

