
# $Id: textwindow.tcl,v 1.6 1996/09/05 07:03:46 vdm Exp $
# create a top level text window with scrollbar
# initially invisible, for collecting text
# quit button does not destroy window, but makes it invisible

proc textWindow {w args} {
    setopt width	80
    setopt height	24
    setopt minwidth	$width
    setopt minheight	$height
    setopt maxwidth	$minwidth
    setopt maxheight	55
    setopt borderwidth	1
    setopt labelrelief	flat
    setopt buttonwidth	8

    toplevel $w
    eval wminit	$w $args
    wm withdraw $w

    frame	$w.buttons
    button	$w.buttons.ok				\
		-text		dismiss			\
		-command	"wm iconify $w"		\
		-borderwidth	$borderwidth		\
		-width		$buttonwidth

    button	$w.buttons.clear			\
		-text		clear			\
		-command	"cleardocu"		\
		-borderwidth	$borderwidth		\
		-width		$buttonwidth

    button	$w.buttons.back				\
		-text		back			\
		-command	"backdocu"		\
		-borderwidth	$borderwidth		\
		-width		$buttonwidth

    button	$w.buttons.forward			\
		-text		forward			\
		-command	"forwarddocu"		\
		-borderwidth	$borderwidth		\
		-width		$buttonwidth

    label	$w.buttons.title			\
		-relief		$labelrelief		\
		-borderwidth	$borderwidth

    pack   $w.buttons.ok	-side left
    pack   $w.buttons.clear	-side left
    pack   $w.buttons.back	-side left
    pack   $w.buttons.forward	-side left

    pack   $w.buttons.title -fill x

    text $w.t	-relief		raised			\
		-borderwidth	$borderwidth		\
		-yscrollcommand	"$w.s set"		\
		-setgrid	true			\
		-width	$width	-height	$height 

    scrollbar $w.s					\
		-relief		flat			\
		-command	"$w.t yview"		\
		-borderwidth	$borderwidth

    pack $w.buttons -side top   -fill x
    pack $w.s       -side right -fill y
    pack $w.t       -expand yes -fill both

    wm minsize $w $minwidth $minheight
    wm maxsize $w $maxwidth $maxheight

    createTags $w.t
}

# tags are bold big blue red
proc createTags w {
    $w tag configure big  -font "-Adobe-Helvetica-Bold-R-Normal-*-120-*" 

    global tk_version
    if {$tk_version >= 4.0} {
	set color [expr [winfo depth $w]>1?"color":"mono"] 
    } else {
 	set color [tk colormodel $w]
    }  
    if [string compare $color "color"]==0 {
        $w tag configure blue -foreground blue
        $w tag configure red  -foreground red
    } else {
        $w tag configure blue -foreground white -background black
        $w tag configure red  -foreground white -background black
    }
}

# insert tagged text into text widget w
proc insertWithTags {w text args} {
    set start [$w index insert]
    $w insert insert $text
    foreach tag [$w tag names $start] {
        $w tag remove $tag $start insert
    }
    foreach i $args {
        $w tag add $i $start insert
    }
}

proc insertSeparator w {
    insertWithTags $w "\n________________________________________________________________________________\n" blue
}

proc insertNewline w {
    insertWithTags $w "\n" 
}

proc insertBlue    {w text} {insertWithTags $w $text     blue}
proc insertRed     {w text} {insertWithTags $w $text     red }
proc insertBigBlue {w text} {insertWithTags $w $text big blue}
proc insertBigRed  {w text} {insertWithTags $w $text big red }
proc insertText    {w text} {insertWithTags $w $text         }

