#!/usr/X11/bin/wish -f
#
# $Id: vdmdb.tcl.sh,v 1.9 1994/06/07 18:59:25 vdm Exp $

set listgeo "20x10"

set dbtype "XXXX"

# -------------------------------------------------------------------------------

proc initdir { } {
    global editor dbname dbtype env argc argv sysdir infoprog cppprog dir fd tmpfile

    if [ info exists env(VDMROOT) ] {
        set sysdir $env(VDMROOT)
    } else {
        set sysdir "/usr/local"
    }

    set processid [ pid ]
    set tmpfile "/tmp/dbm-$processid"

    if [ file isdirectory $sysdir ] {
       if $argc>0 {
           set dir [lindex $argv 0]
           set tmp [split $dir "/"]
           set dbname [lindex $tmp [expr [llength $tmp] - 1]]
       } else {
           set dir $sysdir/lib/vdmdb-c
           set dbname vdmdb-c
       }
       set infoprog "$sysdir/bin/vdmdbm"
       set cppprog  "$sysdir/lib/vdmpp"
    } else {
       DoesNotExist "VDM Class Library installation dir \"$sysdir\" "
    }
                           # flat file database
    if {"$dbtype" == "FLAT"} {
       if [ file exists "$dir.s" ] { } else {
          DoesNotExist "VDM Class Library database \"$dir.s\" "
       }
    } else {
       if [ file isdirectory "$dir" ] { } else {
          DoesNotExist "VDM Class Library database \"$dir\" "
       }
    }

    if [ file executable $infoprog ] {
       set fd [ open "| $infoprog -c" w+ ]
       question "loaddb $dir"
       question "output $tmpfile"
    } else {
       DoesNotExist "VDM Class Library database manager \"$infoprog\" "
    }

    if [info exists env(EDITOR)] {
        set editor $env(EDITOR)
    } else {
        set editor xedit
    }

}

# -------------------------------------------------------------------------------

set lastquestion ""

proc question { cmd } {
    global fd lastquestion

    set lastquestion $cmd
    puts $fd $cmd
    flush $fd
}

proc answer { } {
    global fd

    gets $fd ans
    if {"$ans" == "illegal command"} {
       commError
    } else {
       return $ans
    }
}

proc call { cmd } {
    question $cmd
    answer
}

proc commError { } {
    global infoprog lastquestion

    MessageBox "Communication error with  $infoprog 

last command was
\"$lastquestion\"

internal error of VDM Class Library system" "quit" "error" "500"
    finishapplication
}

proc finishapplication { } {
    global fd tmpfile editedcomponents

    if {"$editedcomponents" != ""} storecomp

    question "exit"
    close $fd
    catch {exec /bin/rm -f $tmpfile}
    destroy .
}

proc tkerror { err } {
    puts $err
    finishapplication
}
# -------------------------------------------------------------------------------


set MessageBoxAns 0
proc MessageBox {Message ButtonNames {Title "Question"} {Aspect "120"}} {
    global MessageBoxAns

    toplevel .question
    wm title .question $Title
    message .question.mess -text $Message -aspect $Aspect
    pack append .question .question.mess  bottom

    foreach ButtonName $ButtonNames {
	button .question.x$ButtonName -text $ButtonName -command "set MessageBoxAns [lsearch $ButtonNames $ButtonName]; destroy .question"
	pack append .question .question.x$ButtonName left
    }

    tkwait visibility .question
    grab set .question
    tkwait window .question
    return $MessageBoxAns
}

# -------------------------------------------------------------------------------

proc DoesNotExist { something } {
       MessageBox "$something not found, \
please check VDM Class Library installation and \
enviroment variable VDMROOT" "quit" "installation error" "400"
       exit
}

# -------------------------------------------------------------------------------

label     .title -text "VDM Database Browser" -relief raised
pack      .title -side top -fill x 

frame     .topbuttons
button    .topbuttons.ok        -text "quit"        -command "finishapplication"
label     .topbuttons.title -text "" -relief raised
button    .topbuttons.help      -text "help"        -command "vdmhelp"
button    .topbuttons.update    -text "store comp " -command "storecomp"
button    .topbuttons.editcomp  -text "edit comp"   -command "editcomp"
button    .topbuttons.exportop  -text "export op"   -command "exportop"
button    .topbuttons.exportimp -text "export impl" -command "exportimp"
button    .topbuttons.exportadt -text "export adt"  -command "exportadt"
button    .topbuttons.exportdb  -text "export db"   -command "exportdb"
pack      .topbuttons.ok        -side left  -fill x
pack      .topbuttons.help      -side right -fill x
pack      .topbuttons.update    -side right -fill x
pack      .topbuttons.editcomp  -side right -fill x
pack      .topbuttons.exportop  -side right -fill x
pack      .topbuttons.exportimp -side right -fill x
pack      .topbuttons.exportadt -side right -fill x
pack      .topbuttons.exportdb  -side right -fill x
pack      .topbuttons.title                 -fill x 

frame     .database
label     .database.dbt -text "database : "
label     .database.dbn -textvariable dir
pack      .database.dbt -fill x
pack      .database.dbn -fill x

frame     .adts
label     .adts.label  -text "ADT"
label     .adts.select -textvariable curradt -relief raised
scrollbar .adts.scroll -command ".adts.list yview"
listbox   .adts.list   -yscroll ".adts.scroll set" -relief raised -geometry $listgeo -setgrid yes
pack      .adts.label  -side top  -fill both
pack      .adts.select -side top  -fill both
pack      .adts.list   -side left -fill both
pack      .adts.scroll -side left -fill y

frame     .impl
label     .impl.label  -text "Implementation"
label     .impl.select -textvariable currimp -relief raised
scrollbar .impl.scroll -command ".impl.list yview"
listbox   .impl.list   -yscroll ".impl.scroll set" -relief raised -geometry $listgeo -setgrid yes
pack      .impl.label  -side top
pack      .impl.select -side top  -fill both
pack      .impl.list   -side left -fill both
pack      .impl.scroll -side left -fill y

frame     .oper
label     .oper.label  -text "Operation"
label     .oper.select -textvariable curropr -relief raised
scrollbar .oper.scroll -command ".oper.list yview"
listbox   .oper.list   -yscroll ".oper.scroll set" -relief raised -geometry $listgeo -setgrid yes
pack      .oper.label  -side top
pack      .oper.select -side top  -fill both
pack      .oper.list   -side left -fill both
pack      .oper.scroll -side left -fill y

frame     .comp
label     .comp.label  -text "Component"
label     .comp.select -textvariable currcmp -relief raised
scrollbar .comp.scroll -command ".comp.list yview"
listbox   .comp.list   -yscroll ".comp.scroll set" -relief raised -geometry $listgeo -setgrid yes
pack      .comp.label  -side top
pack      .comp.select -side top  -fill both
pack      .comp.list   -side left -fill both
pack      .comp.scroll -side left -fill y

pack      .topbuttons  -side top -fill x
pack      .database    -side top -fill x
pack      .adts        -side left
pack      .impl        -side left
pack      .oper        -side left
pack      .comp        -side left

# -------------------------------------------------------------------------------

wm title    . "VDM DB"
wm iconname . "VDM DB"
wm minsize  . 1 1

set curradt ""
set currimp ""
set curropr ""
set currcmp ""

# -------------------------------------------------------------------------------

proc clearlist { list } {
    while { [ $list size ] != 0 } { $list delete end }
}

# -------------------------------------------------------------------------------

proc browseimpl { } {
    global curradt

    clearlist .impl.list
    clearlist .oper.list
    clearlist .comp.list
    set currimp ""
    set curropr ""
    set currcmp ""

    foreach i [ call "implementations $curradt" ] {
           .impl.list insert end $i
    }
}

proc browseoper { } {
    global curradt
    global currimp

    clearlist .oper.list
    clearlist .comp.list
    set curropr ""
    set currcmp ""

    foreach i [call "specialops $curradt $currimp" ] {
        .oper.list insert end $i
    }
}

proc browsecomp { } {
    global curradt currimp curropr

    clearlist .comp.list
    set currcmp ""

    foreach i [ call "components $curradt $currimp $curropr" ] {
        .comp.list insert end $i
    }
}

# export database into flat file
proc exportdb { } {
    global dbname
    question "output $dbname.f"
    call     "exportdb"
    question "output"
}

proc exportparts { adt imp opr cmp file } {
    question "selectadts $adt"
    question "selectimpl $imp"
    question "selectoper $opr"
    question "selectcomp $cmp"
    question "output $file"
    call     "exportparts"
    question "output"
}

proc exportadt { } {
    global curradt

    if {"$curradt" == ""} {
        MessageBox "Please select an ADT first" "quit" "error" "1000"
    } else {
        exportparts "$curradt" "" "" "" "$curradt.-.-.-.f"
    }
}

proc exportimp { } {
    global curradt currimp

    if {"$currimp" == ""} {
        MessageBox "Please select an implementation first" "quit" "error" "1000"
    } else {
        exportparts "$curradt" "$currimp" "" "" "$curradt.$currimp.-.-.f"
    }
}

proc exportop { } {
    global curradt currimp curropr

    if {"$curropr" == ""} {
        MessageBox "Please select an operation first" "quit" "error" "1000"
    } else {
        exportparts "$curradt" "$currimp sign" "$curropr" "" "$curradt.$currimp.$curropr.-.f"
    }
}

set editedcomponents ""

proc editcomp { } {
    global tmpfile editedcomponents editor
    global curradt currimp curropr currcmp

    if {"$currcmp" == ""} {
        MessageBox "Please select a component first" "quit" "error" "1000"
    } else {
        set localtmpfile "$curradt.$currimp.$curropr.$currcmp"

        if [ lsearch -exact $editedcomponents $localtmpfile ]<0 { 
            lappend editedcomponents $localtmpfile

            question "output $localtmpfile"
            call "selcomp $curradt $currimp $curropr $currcmp"
            question "output $tmpfile"
        }
        catch {exec $editor $localtmpfile &}
    }
}

proc storecomp { } {
    global editedcomponents

    if {"$editedcomponents" != ""} {
       set components $editedcomponents
       set editedcomponents ""

       while {"$components" != ""} {
         set compfile [ lindex $components 0 ]
         set comps [ split $compfile . ]

         if [ file exists $compfile ] {
            call "newcomponent $comps $compfile"
            catch {exec /bin/rm -f $compfile}
         } else {
           MessageBox "changed component file \"$compfile\" \
not found, component changes for ADT \"[ lindex $comps 0 ]\", \
implementation \"[ lindex $comps 1 ]\", \
operation \"[ lindex $comps 2 ]\" \
and component \"[ lindex $comps 3 ]\" lost" "quit" "error" "500"
         } 
         set components [ lrange $components 1 end ]
       }
       question "storedb"
    }
}

# -------------------------------------------------------------------------------

proc browseadts { } {
    clearlist .adts.list

    foreach i [ call "adts" ] {
        .adts.list insert end $i
    }
}

# -------------------------------------------------------------------------------

proc vdmhelp { } {
    MessageBox "
VDM Class Library Database Browser

The left list titled \"ADT\" shows the names of the \
abstract datatype directory. Select one of the entries by a left button \
double click of the mouse. The selected \"ADT\" is shown in the label field \
at the top of the list.

The list in the second place titled \"Implementations\" will then \
show the implementations directory for the selected ADT. Select \
one of these Implementations by a left button double click \
of the mouse. The selected implementation is shown in the label field \
at the top of the list.

The third list titled \"Operations\" shows the operations \
directory for the selected implementation. When selecting one of them \
by a left button double click, the list of components will \
appear in the right row. The selected operation is shown in the label field \
at the top of the list.

The right list with the components of an operation can be again \
selected by a double click. The selected component is shown in the label field \
at the top of the list.

The \"edit comp\"  button starts an editor with the \
selected component. Here first the environment variable EDITOR \
is tested, if not set the xedit is used. Edited components can be \
stored back into the database by activating the \"store comp\" button.

The \"export\" buttons export parts of, or the whole database into \
a flat file in shell script format (\"import\" commands are not yet implemented). 

To work on another database than the default, call this program with a \
commandline parameter for the database.

To leave the browser or any subwindow, \
click at the \"quit\" button.
" "quit" "VDM Help" "120"
}

# Set up bindings for the browser -----------------------------------------------

initdir

browseadts

bind  . <Control-q> {finishapplication"}
bind  . <Control-c> {finishapplication"}
focus .

bind  .adts.list <Double-Button-1> {
    foreach i [selection get] {
        set curradt $i
        set currimp ""
        set curropr ""
        set currcmp ""
        browseimpl 
    }
}

bind  .impl.list <Double-Button-1> {
    foreach i [selection get] {
        set currimp $i
        set curropr ""
        set currcmp ""
        browseoper 
    }
}

bind  .oper.list <Double-Button-1> {
    foreach i [selection get] {
        set curropr $i
        set currcmp ""
        browsecomp
    }
}

bind  .comp.list <Double-Button-1> {
    foreach i [selection get] {
        set currcmp $i
    }
}
# -------------------------------------------------------------------------------
