// $Id: vdmdbm.cc,v 1.36 1996/09/07 17:09:24 vdm Exp $

#include "dbm.hpp"

//------------------------------------------------------------------------------

#ifndef NDEBUG
#define o(x)x
#else
#define o(x)
#endif
#define t(x) o(cerr << x << endl;)

//------------------------------------------------------------------------------
//
// string constants
//
//------------------------------------------------------------------------------

ADTNAME		common_adt;
ADTNAME		aux_adt	;
ADTNAME		stdlib_adt;

IMPLNAME	common_impl;
IMPLNAME	sign_impl;

OPNAME		basic_op;
OPNAME		setup_op;

COMPNAME	adt_comp;
COMPNAME	vsig_comp;
COMPNAME	gsig_comp;
COMPNAME	csig_comp;
COMPNAME	psig_comp;
COMPNAME	cpp_comp;
COMPNAME 	vdm_comp;
COMPNAME	docu_comp;
COMPNAME	def_comp;
COMPNAME	defcpp_comp;
COMPNAME	param_comp;

STR special_op;
STR common_op;
STR general_op;
STR unknown_op;

STR c_ops;
STR cpp_ops;
STR app_ops;

void init1 () {
 common_adt	= "com";
 aux_adt	= "aux";
 stdlib_adt	= "stdlib";

 common_impl	= "com";
 sign_impl	= "sign";

 basic_op	= "basic";
 setup_op	= "setup";

 adt_comp	= "adttype";
 vsig_comp	= "vsig";
 gsig_comp	= "gsig";
 csig_comp	= "csig";
 psig_comp	= "psig";
 cpp_comp	= "cpp";
 vdm_comp	= "vdm";
 docu_comp	= "docu";
 def_comp	= "defp";
 defcpp_comp	= "defcpp";
 param_comp     = "param";

 special_op = "SPECIAL";
 common_op  = "COMMON";
 general_op = "GENERAL";
 unknown_op = "UNKNOWN";

 c_ops      = "C";
 cpp_ops    = "C++";
 app_ops    = "A++";
}

//------------------------------------------------------------------------------


STR newline;
STR docu_header;
STR c_header;
STR cpp_header;

void init2() {
  newline =
    "\n";

  docu_header =
    "Description\n"
    "===========\n"
    "\n";

  c_header =
    "\n"
    "C Interface\n"
    "===========\n"
    "\n"
    "#undef GEN_CPLUS\n";

  cpp_header =
    "\n"
    "C++ Interface\n"
    "=============\n"
    "\n"
    "#define GEN_CPLUS\n";
}

// strings for vdminfo ---------------------------------------------------------

STR adt_header;
STR con_header;
STR imp_header;
STR iml_header;
STR chr_header;
STR opr_header;
STR gop_header;
STR gen_header;
STR op1_header;
STR op2_header;
STR op3_header;
STR op4_header;
STR op5_header;

void init3() {
  adt_header = "constructors (abstract datatypes) in VDM ADT database ";
  con_header = "the constructor ";
  imp_header = "implementations for constructor ";
  iml_header = "implementation ";
  chr_header = "implementation characteristics of ";
  opr_header = "operations of implementation ";
  gop_header = "general operations for constructor ";
  gen_header = "general operations common for all constructors";
  op1_header = "operation ";
  op2_header = " of implementation ";
  op3_header = " for constructor ";
  op4_header = "general operation ";
  op5_header = " common for all constructors";
}

// strings for export and import into shell script format ----------------------

STR outdir0;
STR outdir;
STR exportheader;
STR cat;
STR catoutdir;

STR mask;
STR masknl;
STR nlmasknl;
STR catmask;

// g++ --> segmentation fault without explicit initialization routine

void initstrings () {
  init1();
  init2();
  init3();

  outdir0	= "OUT_DIR";
  outdir	= STR("$") + outdir0;
  exportheader  = STR("#!/bin/sh") + newline
			+ outdir0 + "=$VDMROOT/lib/vdmdb-c;"
			+ "[ \"$1\" != \"\" ] && " + outdir0 + "=$1" + newline;
  cat		= "cat >";
  catoutdir	= cat + outdir + '/';

  mask		= "-----";
  masknl	= mask + newline;
  nlmasknl	= newline + masknl;
  catmask       = STR(" <<'") + mask + '\'' + newline;
}

//------------------------------------------------------------------------------
//
// own STRINGS member functions
//
//------------------------------------------------------------------------------

inline
BOOL STRINGS :: is_empty_or_is_in ( const STR & s ) const {
    return 
      is_empty()
      || is_in(s);
}

//------------------------------------------------------------------------------
//
// own STRINGL member functions
//
//------------------------------------------------------------------------------

inline
STR STRINGL :: concat ( const STR & separator ) const {
    STR r = empty;
    NAT0 l = len();

    if ( l != 0 ) {
        r = (*this)[1];
        for (NAT0 i = 2; i<=l; ++i) {
            r += separator;
            r += (*this)[i];
        }
    }
    
    return r;
}

//------------------------------------------------------------------------------
//
// own STR member functions
//
//------------------------------------------------------------------------------

// split a string into a list of strings

STRINGL STR :: split ( const CHAR & delimiter ) const {
  STR     s = *this;
  NAT0    i = s.index(delimiter);
  STRINGL l = empty;

  while ( i != 0 ) {
    l += STR(s,1,i-1);
    ++i;
    s.sub_tup(i,s.len());
    i = s.index(delimiter);
  }

  l += s;
  return l;
}

//------------------------------------------------------------------------------

inline
STR STR :: mkdir () const {
  STR
  line  = "[ -d ";
  line += *this;
  line += " ] || mkdir ";
  line += *this;
  line += newline;

  return line;
}

inline
STR STR :: quote () const {
  STR
  qs  = CHAR('\'');
  qs += *this;
  qs += CHAR('\'');

  return qs;
}

STR STR :: headline () const {
  STR
  hl  = newline;
  hl += *this;
  hl += newline;
  hl += STR('=',this->len());
  hl += newline;
  hl += newline;

  return hl;
} 

//------------------------------------------------------------------------------
//
// VDMDB member functions 
//
//------------------------------------------------------------------------------

inline
BOOL VDMDB :: exist_adt ( const ADTNAME & an
                        ) const {
  return is_in(an);
}


//------------------------------------------------------------------------------

inline
BOOL VDMDB :: exist_impl ( const ADTNAME & an
                         , const IMPLNAME & in
                         ) const {
  ADT adt;

  return
    is_in_dom(an,adt)
    && adt.is_in(in);
}

//------------------------------------------------------------------------------

inline
BOOL VDMDB :: exist_operation ( const ADTNAME  & an
                              , const IMPLNAME & in
                              , const OPNAME   & on
                              ) const {
  ADT  adt;
  IMPL impl;

  return
    is_in_dom(an,adt)
    && adt.is_in_dom(in,impl)
    && impl.is_in(on);
}

inline
BOOL VDMDB :: search_operation ( const ADTNAME  & an
			       , const IMPLNAME & in
			       , const OPNAME   & on
                               ) const {
  ADT  adt;
  IMPL impl;

  return
    exist_operation(an,in,on)
    || exist_operation(an,common_impl,on)
    || exist_operation(common_adt,common_impl,on);
}

//------------------------------------------------------------------------------

inline
BOOL VDMDB :: comp_is_in ( const ADTNAME   & an
                         , const IMPLNAME  & in
                         , const OPNAME    & on
                         , const COMPNAME  & cn
                         ,       COMPONENT & co		// output parameter
                         ) const {
                        t("comp_is_in " << an << in << on << cn)
  ADT		adt;
  IMPL		impl;
  OPERATION	operation;

  return
    is_in_dom(an,adt)
    && adt.is_in_dom(in,impl)
    && impl.is_in_dom(on,operation)
    && operation.is_in_dom(cn,co);
}

//------------------------------------------------------------------------------

inline
BOOL VDMDB :: exist_component ( const ADTNAME  & an
                              , const IMPLNAME & in
                              , const OPNAME   & on
                              , const COMPNAME & cn
                              ) const {
                        t("exist_component " << an << in << on << cn)
    COMPONENT co;
    return
      comp_is_in(an,in,on,cn,co);
  }

//------------------------------------------------------------------------------

inline
COMPONENT VDMDB :: sel_component ( const ADTNAME  & an
                                 , const IMPLNAME & in
                                 , const OPNAME   & on
                                 , const COMPNAME & cn
                                 ) const {
                        t("sel_component " << an << in << on << cn)
  COMPONENT co;

  if ( comp_is_in(an,in,on,cn,co) )
    return co;
  else
    return COMPONENT(empty);
  }

//------------------------------------------------------------------------------

inline
COMPONENT VDMDB :: search_component ( const ADTNAME  & an
                                    , const IMPLNAME & in
                                    , const OPNAME   & on
                                    , const COMPNAME & cn
                                    ) const {
                        t("search_component " << an << in << on << cn)
  COMPONENT co;

  if ( comp_is_in(an,in,on,cn,co)
       || ( ! exist_operation(an,in,on)
	    && ( comp_is_in(an,common_impl,on,cn,co) 
		 || ( ! exist_operation(an,common_impl,on)
		      && comp_is_in(common_adt,common_impl,on,cn,co)
		      )
		 )
	    )
       )
    return co;
  else
    return COMPONENT(empty);
}

//------------------------------------------------------------------------------

inline
COMPONENT VDMDB :: search_sign_component ( const ADTNAME  & an
                                         , const OPNAME   & on
                                         , const COMPNAME & cn
                                         ) const {
                        t("search_sign_component " << an << on << cn)
  COMPONENT co;

  if ( comp_is_in(an,sign_impl,on,cn,co)
    || ( ( cn == csig_comp || cn == psig_comp )
        && comp_is_in(an,sign_impl,on,gsig_comp,co)
       )
    || comp_is_in(common_adt,sign_impl,on,cn,co)
    || ( ( cn == csig_comp || cn == psig_comp )
        && comp_is_in(common_adt,sign_impl,on,gsig_comp,co)
       )
     )
    return co;
  else
    return COMPONENT(empty);
}

//------------------------------------------------------------------------------

inline
BOOL VDMDB :: new_adt ( const ADTNAME  & an ) {
  BOOL changed = false;

  if ( ! exist_adt(an) ) {
    *this |= PAIR_VDMDB(an,empty);
    changed = true;
  }

  return changed;
}

//------------------------------------------------------------------------------

inline
BOOL VDMDB :: new_impl ( const ADTNAME  & an
                       , const IMPLNAME & in
                       ) {
  BOOL changed = false;

  if ( ! exist_impl(an,in) ) {
    changed = new_adt(an);
    ADT adt = (*this)[an]; *this -= an;

    if ( ! adt.is_in(in) ) adt |= PAIR_ADT(in,empty);

    *this |= PAIR_VDMDB(an,adt);
    changed = true;
  }

  return changed;
}

//------------------------------------------------------------------------------

inline
BOOL VDMDB :: new_operation ( const ADTNAME  & an
                            , const IMPLNAME & in
                            , const OPNAME   & on
                            ) {
  BOOL changed = false;

  if ( ! exist_operation(an,in,on) ) {
    changed = new_impl(an,in);
    ADT		adt  = (*this)[an]; *this  -= an;
    IMPL	impl = adt[in];        adt -= in;

    if ( ! impl.is_in(on) ) impl |= PAIR_IMPL(on,empty);

    adt   |= PAIR_ADT  (in,impl);
    *this |= PAIR_VDMDB(an,adt );
    changed = true;
  }

  return changed;
}

//------------------------------------------------------------------------------

inline
BOOL VDMDB :: new_component ( const ADTNAME   & an
                            , const IMPLNAME  & in
                            , const OPNAME    & on
                            , const COMPNAME  & cn
                            , const COMPONENT & co
                            ) {
			t("new_component " << an << in << on << cn)
				// nothing to change if old entry equals new one
  BOOL changed = false;
  COMPONENT cold;
  if ( ! ( comp_is_in(an,in,on,cn,cold) && cold == co ) ) {
				// assure adt impl and opr maps exist
    changed = new_operation(an,in,on);

				// maps have to be modified
				// modification is done in 3 steps
				// selection, deletion, extension
				// this optimizes automatic memory management
				// the reference count of these "large" values is 1
				// so no copy action is neccessary

				// maps are restricted
    ADT        adt  = (*this)[an]; *this -= an;
    IMPL       impl = adt [in];    adt   -= in;
    OPERATION  opr  = impl[on];    impl  -= on;

				// component is inserted
    opr  += PAIR_OPERATION(cn,co);

				// maps are extended again
    impl  |= PAIR_IMPL    (on,opr );
    adt   |= PAIR_ADT     (in,impl);
    *this |= PAIR_VDMDB   (an,adt );
    changed = true;
  }
  return changed;
}

//------------------------------------------------------------------------------

inline
BOOL VDMDB :: rm_adt ( const ADTNAME  & an ) {
  BOOL changed = false;

  if ( exist_adt(an) ) {
    *this -= an;
    changed = true;
  }

  return changed;
}

//------------------------------------------------------------------------------

inline
BOOL VDMDB :: rm_impl ( const ADTNAME  & an
                      , const IMPLNAME & in
                      ) {
  BOOL changed = false;

  if ( exist_impl(an,in) ) {
    ADT adt = (*this)[an]; *this -= an;

    adt -= in;

    *this |= PAIR_VDMDB(an,adt);
    changed = true;
  }

  return changed;
}

//------------------------------------------------------------------------------

inline
BOOL VDMDB :: rm_operation ( const ADTNAME  & an
                           , const IMPLNAME & in
                           , const OPNAME   & on
                           ) {
  BOOL changed = false;

  if ( exist_operation(an,in,on) ) {
    ADT		adt  = (*this)[an]; *this  -= an;
    IMPL	impl = adt[in];        adt -= in;

    impl  -= on;

    adt   |= PAIR_ADT  (in,impl);
    *this |= PAIR_VDMDB(an,adt );
    changed = true;
  }

  return changed;
}

//------------------------------------------------------------------------------

inline
BOOL VDMDB :: rm_component ( const ADTNAME   & an
                           , const IMPLNAME  & in
                           , const OPNAME    & on
                           , const COMPNAME  & cn
                           ) {
  BOOL changed = false;
  if ( exist_component(an,in,on,cn) ) {

				// maps are restricted
    ADT        adt  = (*this)[an]; *this -= an;
    IMPL       impl = adt [in];    adt   -= in;
    OPERATION  opr  = impl[on];    impl  -= on;

				// component is removed
    opr   -= cn;

				// maps are extended again
    impl  |= PAIR_IMPL    (on,opr );
    adt   |= PAIR_ADT     (in,impl);
    *this |= PAIR_VDMDB   (an,adt );
    changed = true;
  }
  return changed;
}

//------------------------------------------------------------------------------

STR VDMDB :: insert_include ( const STR & s ) const {
				t("insert_include " << s)
  NAT0	ix1 = s.index('<');
	++ix1;
  NAT0	ix2 = s.index('>');
	ix2--;
  STR		s1 = STR(s,ix1,ix2);
  STRINGL	l  = s1.split('/');
  STR		rs = empty;

  if ( l.len() == 4 )
    rs = sel_component(l[1],l[2],l[3],l[4]);

  return rs;
}

//------------------------------------------------------------------------------

STR VDMDB :: solve_include ( const STR & s ) const {
				t("solve_include ")
  NAT0 start_include = s.index("#include <");

  if ( ! start_include ) {
    STR  prefix        = STR(s,1,start_include - 1);
    STR  rest          = STR(s,start_include,s.len());
    NAT0 end_include   = rest.index('\n');
    STR  subst_include = solve_include(insert_include(STR(rest,1,end_include)));
    STR  postfix       = solve_include(STR(rest,end_include+1,rest.len()));

    prefix += subst_include;
    prefix += postfix;

    return prefix;
  }
  else {
    return s;
  }
}

//------------------------------------------------------------------------------

inline
COMPONENT VDMDB :: compute_component ( const ADTNAME  & an
                                     , const IMPLNAME & in
                                     , const OPNAME   & on
                                     , const COMPNAME & cn
                                     ) const {
  return
    solve_include(search_component(an,in,on,cn));
}

//------------------------------------------------------------------------------

inline
COMPONENT VDMDB :: compute_sign_component ( const ADTNAME  & an
                                          , const OPNAME   & on
                                          , const COMPNAME & cn
                                          ) const {
  return
    solve_include(search_sign_component(an,on,cn));
}

// search docu info ------------------------------------------------------------

STR VDMDB :: setup_docu () const {
  return
    compute_sign_component(stdlib_adt,setup_op,docu_comp);
}

STR VDMDB :: search_docu_components ( const ADTNAME  & an
                                    , const IMPLNAME & in
                                    , const OPNAME   & on
                                    , const STRINGS  & sign_components
                                    ) const {
                        t("search_docu_components " << an << in << on)
  STR
  docu  = compute_sign_component(an,on,vsig_comp );
  docu += newline;
  docu += compute_component(an,in,on,docu_comp);

  if ( sign_components.is_in(csig_comp) ) {
    STR csig = compute_sign_component(an,on,csig_comp);
    if ( ! csig ) {
      docu += c_header;
      docu += csig;
    }
  }

  if ( sign_components.is_in(cpp_comp) ) {
    STR cpp = compute_sign_component(an,on,cpp_comp);
    if ( ! cpp ) {
      docu += cpp_header;
      docu += compute_sign_component(stdlib_adt,setup_op,defcpp_comp);
      docu += cpp;
      docu += compute_sign_component(stdlib_adt,setup_op,def_comp);
    }
  }

  return docu;
}

// selector functions for special components -----------------------------------

STR VDMDB :: sel_adttype ( const ADTNAME  & an
                         ) const {
			t("sel_adttype " << an)
  STR co = sel_component(an,common_impl,basic_op,adt_comp);
  NAT0 i = co.index('\n');

  if ( i != 0 )
    i--,co.sub_tup(1,i);

  return co;
}

//------------------------------------------------------------------------------

inline
STRINGS VDMDB :: sel_adts () const {
  return dom();
}

//------------------------------------------------------------------------------

inline
STRINGS VDMDB :: sel_impl ( const ADTNAME & an
                          ) const {
  ADT adt;

  if ( is_in_dom(an,adt) ) {
    return adt.dom();
  }
  else {
    return STRINGS(empty);
  }
}

//------------------------------------------------------------------------------

inline
STRINGS VDMDB :: sel_op ( const ADTNAME  & an
                        , const IMPLNAME & in
                        ) const {
  ADT	adt;
  IMPL	impl;

  if ( is_in_dom(an,adt)
    && adt.is_in_dom(in,impl)
     )
    return
      impl.dom();
  else
    return
      STRINGS(empty); 
  }

//------------------------------------------------------------------------------

inline
STRINGS VDMDB :: sel_oper ( const ADTNAME  & an
                          , const IMPLNAME & in
                          ) const {
    STRINGS
    oset  = sel_op(an,in);
    oset |= sel_op(an,common_impl);
    oset |= sel_op(common_adt,common_impl);

    return
      oset;
  }

STRINGS VDMDB :: sel_opr1 ( const ADTNAME  & an
                          , const IMPLNAME & in
			  , const STR      & optype
                          ) const {
    STRINGS oset = sel_oper(an,in);

    for ( ITER_STRINGS i = oset; ! i; ++i ) {
	OPNAME opr = i.sel();

        BOOL remove_op = false;
        if ( ! optype && optype != c_ops ) {
           if (  ! exist_component(an,sign_impl,opr,cpp_comp)
              && ! exist_component(common_adt,sign_impl,opr,cpp_comp)
              )
              remove_op = true;
        }

        if ( remove_op ) {
          oset -= opr;
        } 
    }

    return
      oset;
}

//------------------------------------------------------------------------------

inline
STRINGS VDMDB :: sel_comp ( const ADTNAME  & an
                          , const IMPLNAME & in
                          , const OPNAME   & on
                          ) const {
  ADT        adt;
  IMPL       impl;
  OPERATION  opr;

  if ( is_in_dom(an,adt)
    && adt.is_in_dom(in,impl)
    && impl.is_in_dom(on,opr)
     )
    return opr.dom();
  else
    return empty; 
}

// parameter selection ---------------------------------------------------------

STRINGS VDMDB :: sel_oper_params ( const ADTNAME  & an
                                 , const IMPLNAME & in
                                 , const OPNAME   & on
                                 ) const {
  STR   	comp = sel_component(an,in,on,param_comp);
  STRINGS	res  = empty;
  NAT0		i    = comp.index('\n');

  while ( i > 1 ) {
    res |= STR(comp,1,i-1);
    comp.sub_tup(i+1,comp.len());
    i   = comp.index('\n');
  }

  return res;
}

inline
STRINGS VDMDB :: sel_impl_params ( const ADTNAME  & an
                                 , const IMPLNAME & in
                                 ) const {
  return sel_oper_params(an,in,basic_op);
}

// determine operation classification ------------------------------------------

inline
STR VDMDB :: classify_op ( const ADTNAME  & an
                         , const IMPLNAME & in
                         , const OPNAME   & on
                         ) const {
  if ( exist_operation(an,        in,         on) ) return special_op;
  if ( exist_operation(an,        common_impl,on) ) return common_op ;
  if ( exist_operation(common_adt,common_impl,on) ) return general_op;
                                                    return unknown_op;
}

// export into shell scrip -----------------------------------------------------

STR VDMDB :: export () const {
  STR r = exportheader;
  r += outdir.mkdir();

				// 4 nested loops
				// the iterators run through a sorted list of the keys
				// to ensure alphabetic ordering of the components
  STRINGL anl = dom().sort();
  for ( ITER_STRINGL i1 = anl; ! i1; ++i1 ) {
    ADTNAME an     = i1.sel();
    ADT     adt    = (*this)[an];
    STR     adtdir = outdir + '/' + an;
      r += adtdir.mkdir();

    STRINGL inl = adt.dom().sort();
    for ( ITER_STRINGL i2 = inl; ! i2; ++i2 ) {
      IMPLNAME in      = i2.sel();
      IMPL     impl    = adt[in];
      STR      impldir = adtdir + '/' + in;
        r += impldir.mkdir();

      STRINGL onl = impl.dom().sort();
      for ( ITER_STRINGL i3 = onl; ! i3; ++i3 ) {
        OPNAME     on     = i3.sel();
        OPERATION  opr    = impl[on];
        STR        oprdir = impldir + '/' + on;
          r += oprdir.mkdir();

        STRINGL cnl = opr.dom().sort();
        for ( ITER_STRINGL i4 = cnl; ! i4; ++i4 ) {
          COMPNAME  cn = i4.sel();
          COMPONENT co = opr[cn];
            r += cat + oprdir + '/' + cn + catmask;
            r += co;
            r += masknl;
        }
      }
    }
  }
  return r;
}

// select parts of the database into a new database ----------------------------

VDMDB VDMDB :: sel_db ( const STRINGS & adts
                      , const STRINGS & impl
                      , const STRINGS & oprs
                      , const STRINGS & cmps
                      ) const {
  VDMDB db1 = empty;
				// all parts of the 4 level map,
				// were the keys are in the 4 parameters,
				// are selected from the database
				// and put into the local database db1
				// an empty set as parameter indicates no selection
				// in this case all pairs are processed
  for ( ITER_VDMDB i1 = *this; ! i1; ++i1 ) {
    ADT    adt;
    ADTNAME an = i1.sel(adt);

    if ( adts.is_empty_or_is_in(an) ) {
      for ( ITER_ADT i2 = adt; ! i2; ++i2 ) {
        IMPL     imp;
        IMPLNAME in = i2.sel(imp);

        if ( impl.is_empty_or_is_in(in) ) {
          for ( ITER_IMPL i3 = imp; ! i3; ++i3 ) {
            OPERATION  opr;
            OPNAME     on = i3.sel(opr);

            if ( oprs.is_empty_or_is_in(on) ) {
              for ( ITER_OPERATION i4 = opr; ! i4; ++i4 ) {
                COMPONENT co;
                COMPNAME  cn = i4.sel(co);

                if ( cmps.is_empty_or_is_in(cn) ) {
                  (void)db1.new_component(an,in,on,cn,co);
                }
              }
            }
          }
        }
      }
    }
  }

  return db1;
}

// import from shell script source ---------------------------------------------

BOOL VDMDB :: import ( const STR & s ) {
			t("import " << s)
  NAT0 i = 1;
  BOOL changed = false;

  while ( i <= s.len() ) {
    NAT0 startcat = s.index(catoutdir,i);
    NAT0 startrest = s.len().succ();
			t("import: startcat = " << startcat.data << " len = " << s.len().data)

    if ( startcat != 0 ) {		// start of component found
           startcat += catoutdir.len();
      NAT0 endcat    = s.index(catmask,startcat);
      NAT0 starttxt  = endcat + catmask.len();
      NAT0 endtxt    = s.index(nlmasknl,starttxt-1);
      if ( endtxt != 0 ) {		// end of component found
           startrest= endtxt + nlmasknl.len();

        STR     p = STR(s,startcat,endcat-1);
        STR     c = STR(s,starttxt,endtxt);
        STRINGL l = p.split('/');
			t("p = " << p << "c = " << c)
        if ( l.len() == 4 ) {		// component names correct
          changed |= new_component(l[1],l[2],l[3],l[4],c);
        }
      }
    }
    i = startrest;
  }

  return changed;
}

//------------------------------------------------------------------------------
//
// dbcontext class (is used by dbmanager class)
//
//------------------------------------------------------------------------------

// Sun Apr 14 23:35:45 MET DST 1996 enhancement for for multisession feature

class DBCONTEXT {  
  INTG contextcntr,currentcontext;  // implizit index grows neg., current index
  CONTEXT defaultcontext;   	// Predefined defaultcontext
  CONTEXTMAP contextmap;	// Context MAP (see DSL spec. dbm.vdm)

  STRINGS&	adttypes;	// context specifiers
  STRINGS&	special_impl;   // ...
  STRINGS&	special_ops;
  STRINGS&	filter_adts;
  STRINGS&	filter_impl;
  STRINGS&	filter_oper;
  STRINGS&	filter_comp;
  STRINGS&	filter_params;
  STRINGS&	sign_comps;
  STR&		operationtype;

  CONTEXT emptycontext() { // Predefined new context
    return CONTEXT(empty,  // adttypes
                   empty,  // special_impl
    	           empty,  // special_ops
	           empty,  // filter_adts
	           empty,  // filter_impl
	           empty,  // filter_oper
	           empty,  // filter_comp
	           empty,  // filter_params
 	           empty,  // sign_comps
	           c_ops); // operationtype
  }

  void Update()  // update the context map by DBMANAGER context-specifiers
  {		 
    defaultcontext.chg_adttypes(adttypes);
    defaultcontext.chg_special_impl(special_impl);
    defaultcontext.chg_special_ops(special_ops);
    defaultcontext.chg_filter_adts(filter_adts);
    defaultcontext.chg_filter_impl(filter_impl);
    defaultcontext.chg_filter_oper(filter_oper);
    defaultcontext.chg_filter_comp(filter_comp);
    defaultcontext.chg_filter_params(filter_params);
    defaultcontext.chg_sign_comps(sign_comps);
    defaultcontext.chg_operationtype(operationtype);
    contextmap-=currentcontext;  // remove old current context
    contextmap|=CONTEXTMAP(currentcontext,defaultcontext); // update context
  }

  void Reload()  // modifies context-specifiers in DBMANAGER by reference
  {		 // to the current context
    adttypes=defaultcontext.adttypes();
    special_impl=defaultcontext.special_impl();
    special_ops=defaultcontext.special_ops();
    filter_adts=defaultcontext.filter_adts();
    filter_impl=defaultcontext.filter_impl();
    filter_oper=defaultcontext.filter_oper();
    filter_comp=defaultcontext.filter_comp();
    filter_params=defaultcontext.filter_params();
    sign_comps=defaultcontext.sign_comps();
    operationtype=defaultcontext.operationtype();
  }
public:  
  DBCONTEXT(STRINGS& ref_adttypes,          // create context and
            STRINGS& ref_special_impl,	    // mind the context-specifiers
 	    STRINGS& ref_special_ops,	    // of DBMANAGER
	    STRINGS& ref_filter_adts,
	    STRINGS& ref_filter_impl,
	    STRINGS& ref_filter_oper,
	    STRINGS& ref_filter_comp,
            STRINGS& ref_filter_params,
	    STRINGS& ref_sign_comps,
	    STR& ref_operationtype
  ):
    contextcntr(INTG(0)),         // init context counter 
    currentcontext(contextcntr),  // default context 0 is current
    defaultcontext(emptycontext()),     // create default context
    contextmap(INTG(0),defaultcontext), // defaultcontext is first in map
    adttypes(ref_adttypes),
    special_impl(ref_special_impl),
    special_ops(ref_special_ops),
    filter_adts(ref_filter_adts),
    filter_impl(ref_filter_impl),
    filter_oper(ref_filter_oper),
    filter_comp(ref_filter_comp),
    filter_params(ref_filter_params),
    sign_comps(ref_sign_comps),
    operationtype(ref_operationtype)
  {
     Reload();
  }

  BOOL Exists(const INTG& cid) { return contextmap.is_in(cid); }

  BOOL Switch(const INTG& cid) 
  {
    if (currentcontext == cid) return true; // already set
    if (Exists(cid) == false) return false;
    Update();  // update map
    defaultcontext=contextmap.sel(currentcontext=cid); // get new context
    Reload();
    return true;
  }

  INTG Create()
  {
    if (--contextcntr >= INTG(0)) contextcntr=INTG(-1); // never in live
    contextmap |= CONTEXTMAP(contextcntr,emptycontext()); // new context
    return contextcntr;
  }

  INTG Create(const NAT0& cid)
  { 
    INTG cntr(cid);
    if (Exists(cid)) return INTG(0);
    contextmap |= CONTEXTMAP(cntr,emptycontext()); // new context
    return cntr;
  }
  
  BOOL Release(const INTG& cid)
  {
    if (Exists(cid) == false) return false;
    if (cid == INTG(0)) return false;
    if (cid == currentcontext) { 
	defaultcontext=contextmap.sel(currentcontext=INTG(0)); // get new context
	Reload();
    }
    contextmap -= cid; // remove context
    return true;
  }

  STR List() // create a space separated list of all sessions in use
  {                    
    CONTEXT x;
    ITER_CONTEXTMAP im=contextmap;
    if (!!im) return STR("");        // normally def. context must be there
    STR s(im.sel(x).cvstr());
    for (im++;!im;im++) {
      s += STR(" ") + im.sel(x).cvstr();
    }
    return s;
  }

  INTG CID() { return currentcontext; }  // Context ID
  const CONTEXT Current() const { return defaultcontext; }
  CONTEXT& CurrentChangeable() { return defaultcontext; }
};

//------------------------------------------------------------------------------
//
// vdmdb manager class
//
//------------------------------------------------------------------------------

// Sun Apr 14 23:35:45 MET DST 1996 enhanced VDMDB class for multisession 
// feature

class DBMANAGER {
public:
  VDMDB		&db;
  STR		name;

  STRINGS	adttypes;
  STRINGS	special_impl;
  STRINGS	special_ops;
  STRINGS	filter_adts;
  STRINGS	filter_impl;
  STRINGS	filter_oper;
  STRINGS	filter_comp;
  STRINGS	filter_params;
  STRINGS	sign_comps;
  STR		operationtype;

private:
  BOOL		changed;
  DBCONTEXT     context;  // enhancement for multisession feature

//------------------------------------------------------------------------------

public:

  BOOL db_changed () { return changed; }

//------------------------------------------------------------------------------
// Sun Apr 14 23:35:45 MET DST 1996 added by mt for multisession feature
//
// operationtype is changed in dbmtcl.cc AS A REFERENCE, so 
// ALL valid VDM-Strings MUST be defined in DBMANAGER because otherwise
// changes are discarded or pointer-to-a-pointer-to-a-var datatypes
// are required

public:

  // session_?? methods are enhancing the DBMANAGER class for multisession
  // feature by changing the context. The DBCONTEXT class is constructed
  // using pointers to the context-identifiers in DBMANAGER and uses them
  // to copy the contents of selected context by value

  STR session_create() { return context.Create().cvstr(); }
  STR session_create(const NAT0& cid) { return context.Create(cid).cvstr(); }
  BOOL session_exists(const INTG& cid) { return context.Exists(cid); }
  BOOL session_release(const INTG& cid) { return context.Release(cid); }
  BOOL session_switch(const INTG& cid) { return context.Switch(cid); }
  STR session_list() { return context.List(); }
  STR session_current() { return context.CID().cvstr(); }

//------------------------------------------------------------------------------
// Sun Apr 14 23:35:45 MET DST 1996 added by mt for multisession feature

public:

  DBMANAGER(class VDMDB &vdmdb):db(vdmdb),
    context(adttypes,         // gives the context-specifiers to 
            special_impl,     // the context by reference
            special_ops,
            filter_adts,
            filter_impl,
            filter_oper,
            filter_comp,
            filter_params,
            sign_comps,
            operationtype)
  { 
    db = empty;
    name=empty;
    changed=false;
  }

//------------------------------------------------------------------------------

private:

  typedef BOOL (DBMANAGER::*PREDICATE)(const STR & str) const;

  STRINGS filter ( const STRINGS & sset
                 , PREDICATE p
                 ) const {
    STRINGS sset1(empty);
    ITER_STRINGS i(sset);

    while ( ! i ) {
      STR str = i.sel();
      if ( (this->*p)(str) )
        sset1 |= str;
      ++i;
    }
    return sset1; 
  }

//------------------------------------------------------------------------------
public:

  void clear_db () {
    if ( ! db.is_empty() ) {
      db      = empty;
      changed = true;
    }
  }

//------------------------------------------------------------------------------

  void load_db ( const STR & dbname ) {
    STR dbn = dbname;
			t("reading VDM Class Library database " << dbn)
#ifdef flat
    INBYTES inbytes;
    inbytes.fread(dbn);
    if ( ! inbytes )
      db.cvfromtupl(inbytes);
#else
    db.fread(dbn);
#endif
			t("loading finished") 
    
    if ( ! db.is_empty() )
      name = dbn;
    changed = false;
  }

//------------------------------------------------------------------------------

  void store_db ( const STR & newname ) /* all except flag "changed" is const */ {
    if ( ! newname && name != newname )
      changed = true, name = newname;

    if ( changed && ! name ) {
#ifdef flat
      VDMDB dbn = db;
      OUTBYTES outbytes = dbn.cvtotupl();
      outbytes.fwrite(name);
#else
      db.fwrite(name);
#endif
      changed = false;
    }
  }

//------------------------------------------------------------------------------

  BOOL adt_of_interest ( const STR & an ) const {
    return
      adttypes.is_empty()
      || adttypes.is_in(db.sel_adttype(an));
  }

//------------------------------------------------------------------------------

  BOOL impl_of_interest ( const IMPLNAME & in ) const {
    return
      ! special_impl.is_in(in);
  }

//------------------------------------------------------------------------------

  BOOL oper_of_interest ( const OPNAME & on ) const {
    return
      ! special_ops.is_in(on);
  }

// select available adts -------------------------------------------------------

  STRINGL sel_adts () const {
    return
      filter(db.sel_adts(),&DBMANAGER::adt_of_interest).sort();
  }

// select available implementations --------------------------------------------

  STRINGL sel_impl ( const ADTNAME & an ) const {
    return
      filter(db.sel_impl(an),&DBMANAGER::impl_of_interest).sort();
  }

// select all available operations ---------------------------------------------

  STRINGS sel_oper ( const ADTNAME & an
                   , const IMPLNAME & in
                   ) const {
    return
      filter(db.sel_opr1(an,in,operationtype),
             &DBMANAGER::oper_of_interest);
  }

// select documentation for adts -----------------------------------------------

  STR adtdocu () const {
    STRINGL adts = sel_adts();
    STR     r    = db.setup_docu();

    for ( ITER_STRINGL i = adts; !i; ++i ) {
      STR adt = i.sel();
      r += ( STR("The ") + adt.quote() + STR(" Constructor (ADT)")
           ).headline();
      r += db.search_docu_components(adt,common_impl,basic_op,empty);
    }
    return r;
  }

  STR imp1docu ( const ADTNAME  & an 
               , const IMPLNAME & in
               ) const {
    STR
    r  = ( STR("Implementation ") + in.quote() + " for ADT " + an.quote()
         ).headline();
    r += db.search_docu_components(an,in,basic_op,empty);

    return r;
  }
 
  STR impldocu ( const ADTNAME & an ) const {
    STRINGL impls = sel_impl(an);
    STR     r     = db.setup_docu();

    for ( ITER_STRINGL i = impls; !i; ++i ) {
      STR in = i.sel();
      r += imp1docu(an,in);
    }

    return r;
  }

  STR op1docu ( const ADTNAME  & an
              , const IMPLNAME & in
              , const OPNAME   & on
              ) const {
    STR opclass = db.classify_op(an,in,on);

    STR header  = empty;;
    if ( opclass == special_op )
      header = STR(" ( special realisation for ") + in.quote() + STR(" )");
    else
    if ( opclass == common_op )
      header = STR(" ( common realisation of ADT ") + an.quote() + STR(" )");
    else
    if ( opclass == general_op )
      header = STR(" ( general realisation for all ADTs )");

    STR
    r  = ( STR("Operation ") + on.quote() + header
         ).headline();
    r += db.search_docu_components(an,in,on,sign_comps);

    return r;
  }

  STR oprdocu ( const ADTNAME  & an
              , const IMPLNAME & in
              ) const {
    STR     r          = imp1docu(an,in);
    STRINGL opl        = sel_oper(an,in).sort();

    for ( ITER_STRINGL i = opl; !i; ++i ) {
      OPNAME on = i.sel();
      r += op1docu(an,in,on);
    }

    return r;
  }
  
// export database -------------------------------------------------------------

  STR export () const {
    return db.export();
  }

  STR export_selected () const {
    return db.sel_db(filter_adts,filter_impl,
                     filter_oper,filter_comp).export();
  }

 // mt added here at 06/23/96 to create vdm output format from pure text

  STR exportformat (const ADTNAME &an,
		    const IMPLNAME &in,
		    const OPNAME &on,
		    const COMPNAME &cn,
		    const COMPONENT &co) {
     VDMDB db=empty;
     db.new_component(an,in,on,cn,co+newline);
     return (db.export());
  }
  

// create component-------------------------------------------------------------

  void new_component ( const ADTNAME   & an
                     , const IMPLNAME  & in
                     , const OPNAME    & on
                     , const COMPNAME  & cn
                     , const COMPONENT & co
                                 ) {
    changed |= db.new_component(an,in,on,cn,co);
  }

// import components from flat file --------------------------------------------

  BOOL import ( const STR & s ) {
    BOOL rc=db.import(s);  // mt changed here at 06/23/96 to get a BOOL result 
    changed |= rc;	   // instead of void
    return rc;
  }
// remove components from database ---------------------------------------------

  void rm_adt ( const ADTNAME   & an
              ) {
    changed |= db.rm_adt(an);
  }

  void rm_impl ( const ADTNAME   & an
               , const IMPLNAME  & in
               ) {
    changed |= db.rm_impl(an,in);
  }

  void rm_operation ( const ADTNAME   & an
                    , const IMPLNAME  & in
                    , const OPNAME    & on
                                 ) {
    changed |= db.rm_operation(an,in,on);
  }

  void rm_component ( const ADTNAME   & an
                    , const IMPLNAME  & in
                    , const OPNAME    & on
                    , const COMPNAME  & cn
                                 ) {
    changed |= db.rm_component(an,in,on,cn);
  }

  void new_adt ( const ADTNAME	& an
               ) {
    changed |= db.new_adt(an);
  }

  void new_impl ( const ADTNAME   & an
                , const IMPLNAME  & in
                ) {
    changed |= db.new_impl(an,in);
  }

  void new_operation ( const ADTNAME   & an
                     , const IMPLNAME  & in
                     , const OPNAME    & on
                                 ) {
    changed |= db.new_operation(an,in,on);
  }

  void new_component ( const ADTNAME   & an
                     , const IMPLNAME  & in
                     , const OPNAME    & on
                     , const COMPNAME  & cn
                                 ) {
    if ( ! db.exist_component(an,in,on,cn) )
      changed |= db.new_component(an,in,on,cn,STR(empty));
  }

}; // end of class DBMANAGER 

// the database object ---------------------------------------------------------
// Sun Apr 14 23:35:45 MET DST 1996 changed by mt for multisession feature

class INITSTRINGS {
public:
  INITSTRINGS() { initstrings(); }
} AUTOINITSTRINGS;

VDMDB db;
DBMANAGER dbm(db);

void initdbm() {
  
// Sun Apr 14 23:35:45 MET DST 1996 commented out by mt for multisession feature
//  initstrings();
//  dbm.init();
}

