#!/usr/X11/bin/wish -f
#
# $Id: vdminfo.tcl.sh,v 1.7 1994/06/07 18:59:26 vdm Exp $

set listgeo "20x10"

set dbtype "XXXX"

# -------------------------------------------------------------------------------

proc initdir { } {
    global dbtype env argc argv sysdir infoprog cppprog dir fd tmpfile

    if [ info exists env(VDMROOT) ] {
        set sysdir $env(VDMROOT)
    } else {
        set sysdir "/usr/local"
    }

    set processid [ pid ]
    set tmpfile "/tmp/vdminfo-$processid"

    if [ file isdirectory $sysdir ] {
       if $argc>0 {
           set dir [lindex $argv 0]
       } else {
           set dir $sysdir/lib/vdmdb-c
       }
       set infoprog "$sysdir/bin/vdmdbm"
       set cppprog  "$sysdir/lib/vdmpp"
    } else {
       DoesNotExist "VDM Class Library installation dir \"$sysdir\" "
    }
                           # flat file database
    if {"$dbtype" == "FLAT"} {
       if [ file exists "$dir.s" ] { } else {
          DoesNotExists "VDM Class Library database \"$dir.s\" "
       }
    } else {
       if [ file isdirectory "$dir" ] { } else {
          DoesNotExist "VDM Class Library database \"$dir\" "
       }
    }

    if [ file executable "$cppprog" ] { } else {
       DoesNotExist "VDM Class Library macro processor \"$cppprog\" "
    }

    if [ file executable $infoprog ] {
       set fd [ open "| $infoprog -c" w+ ]
       question "loaddb $dir"
       question "adttypes REAL"
       question "specialimpl com sign aux"
       question "specialoper basic rename signature gcinfo cvfromtup1 cvtotup1 mk_one mkn_empty mfold"
       question "output $tmpfile"
    } else {
       DoesNotExist "VDM Class Library database manager \"$infoprog\" "
    }
}

proc question { cmd } {
    global fd

    puts $fd $cmd
    flush $fd
}

proc answer { } {
    global fd

    gets $fd
}

proc call { cmd } {
    question $cmd
    answer
}

proc finishapplication { } {
    global fd tmpfile

    close $fd
    catch {exec /bin/rm -f $tmpfile}
    destroy .
}

# -------------------------------------------------------------------------------

set MessageBoxAns 0
proc MessageBox {Message ButtonNames {Title "Question"} {Aspect "120"}} {
    global MessageBoxAns

    toplevel .question
    wm title .question $Title
    message .question.mess -text $Message -aspect $Aspect
    pack append .question .question.mess  bottom

    foreach ButtonName $ButtonNames {
	button .question.x$ButtonName -text $ButtonName -command "set MessageBoxAns [lsearch $ButtonNames $ButtonName]; destroy .question"
	pack append .question .question.x$ButtonName left
    }

    grab set .question
    tkwait window .question
    return $MessageBoxAns
}

# -------------------------------------------------------------------------------

proc DoesNotExist { something } {
       MessageBox "$something not found, \
please check VDM Class Library installation and \
enviroment variable VDMROOT" "quit" "installation error" "400"
       exit
}

# -------------------------------------------------------------------------------

label     .title -text "VDM Class Library Browser"

frame     .topbuttons
button    .topbuttons.ok    -text "quit" -command "finishapplication"
label     .topbuttons.title -text "VDM Class Library Browser" -relief raised
button    .topbuttons.help  -text " help " -command "vdmhelp"
pack      .topbuttons.ok    -side left 
pack      .topbuttons.help  -side right 
pack      .topbuttons.title -fill x 

frame     .database
label     .database.dbt -text "database : "
label     .database.dbn -textvariable dir 
pack      .database.dbt -fill x
pack      .database.dbn -fill x

frame     .adts
label     .adts.label  -text "ADTs"
label     .adts.select -textvariable curradt -relief raised
button    .adts.predef -text "predefined" -command "showpredefined"
scrollbar .adts.scroll -command ".adts.list yview"
listbox   .adts.list   -yscroll ".adts.scroll set" -relief raised -geometry $listgeo -setgrid yes
pack      .adts.label  -side top  -fill both 
pack      .adts.select -side top  -fill both
pack      .adts.predef -side bottom  -fill both
pack      .adts.list   -side left -fill both
pack      .adts.scroll -side left -fill y

frame     .impl
label     .impl.label  -text "Implementations"
label     .impl.select -textvariable currimp -relief raised
button    .impl.common -text "common" -command "showcommon"
scrollbar .impl.scroll -command ".impl.list yview"
listbox   .impl.list   -yscroll ".impl.scroll set" -relief raised -geometry $listgeo -setgrid yes
pack      .impl.label  -side top
pack      .impl.select -side top  -fill both
pack      .impl.common -side bottom -fill both
pack      .impl.list   -side left -fill both
pack      .impl.scroll -side left -fill y

frame     .oper
label     .oper.label  -text "Operations"
label     .oper.select -textvariable curropr -relief raised
button    .oper.feature -text "features" -command "showfeatures"
scrollbar .oper.scroll -command ".oper.list yview"
listbox   .oper.list   -yscroll ".oper.scroll set" -relief raised -geometry $listgeo -setgrid yes
pack      .oper.label  -side top
pack      .oper.select -side top  -fill both
pack      .oper.feature -side bottom -fill both
pack      .oper.list   -side left -fill both
pack      .oper.scroll -side left -fill y

pack      .topbuttons  -side top -fill x
#pack      .title       -side top
pack      .database    -side top -fill x
pack      .adts        -side left
pack      .impl        -side left
pack      .oper        -side left

# -------------------------------------------------------------------------------

wm title    . "VDM Info"
wm iconname . "VDM Info"
wm minsize  . 1 1

set curradt ""
set currimp ""
set curropr ""

# -------------------------------------------------------------------------------

proc clearlist { list } {
    while { [ $list size ] != 0 } { $list delete end }
}

# -------------------------------------------------------------------------------

proc showpredefined { } {
    showoper "--predefined--" "--common--" "--features--"
}

proc showcommon { } {
    global curradt

    if {"$curradt" == ""} {
        MessageBox "Please select an ADT first" "quit" "error" "500"
    } else {
        showoper $curradt "--common--" "--features--"
    }
}

proc showfeatures { } {
    global curradt
    global currimp

    if {"$curradt" == ""} {
        MessageBox "Please select an ADT and an implementation first" "quit" "error" "500"
    } else {
    if {"$currimp" == ""} {
        MessageBox "Please select an implementation first" "quit" "error" "500"
    } else {
        showoper $curradt $currimp "--features--"
    }
    }
}

proc browseimpl { } {
    global curradt

    clearlist .impl.list
    clearlist .oper.list
    set currimp ""
    set curropr ""

    foreach i [ call "implementations $curradt" ] {
       .impl.list insert end $i
   }
}

proc browseoper { } {
    global curradt
    global currimp

    clearlist .oper.list
    set curropr ""

    foreach i [ call "operations $curradt $currimp" ] {
        .oper.list insert end $i
    }
}

proc title { adt imp opr } {
    global dir
    global infoprog

    if {"$adt" == "--predefined--"} {
        return "predefined datatypes"
    }
    if {"$imp" == "--common--"} {
        return "general information about ADT \"$adt\""
    }
    if {"$opr" == "--features--"} {
        return "features of implementation \"$imp\" for ADT \"$adt\""
    }

    set type [ call "opertype $adt $imp $opr" ]

    if {"$type" == "SPECIAL"} {
        return "operation \"$opr\" of implementation \"$imp\" for ADT \"$adt\""
    } 
    if {"$type" == "COMMON"} {
        return "general operation \"$opr\" for ADT \"$adt\""
    } 
    return "common operation \"$opr\" for all ADT's"
}

proc getdocu { adt imp opr } {
    global tmpfile cppprog

    question "signcomps"
    if {"$adt" == "--predefined--"} {
        set quest "predefined"
    } elseif {"$imp" == "--common--"} {
        set quest "docu $adt com basic"
    } elseif {"$opr" == "--features--"} {
        set quest "docu $adt $imp basic"
    } else {
        set quest "docu $adt $imp $opr"
        question "signcomps csig cpp"
    }
    set waitanswer [ call "$quest" ]

    set docu ""
    catch {set docu [exec $cppprog -E0 -L -R -DGEN_DOCU -DGEN_C -DGEN_ANSIC $tmpfile]}
    return $docu
}

proc showoper { adt imp opr {w .window} } {

    catch {destroy $w}
    toplevel $w
    set windowtitle [ title $adt $imp $opr ]

    wm geometry $w +300+300
    wm title    $w "VDM Class Library Documentation"
    wm iconname $w "$opr"

    frame  $w.buttons
    button $w.buttons.ok    -text quit -command "destroy $w"
    label  $w.buttons.title -text "$windowtitle" -relief raised
    pack   $w.buttons.ok    -side left
    pack   $w.buttons.title -fill x

    text $w.t -relief raised -bd 2 \
              -yscrollcommand "$w.s set" \
              -setgrid true \
	      -width 80 -height 16 
    scrollbar $w.s -relief flat -command "$w.t yview"

    pack $w.buttons -side top -fill x
    pack $w.s       -side right -fill y
    pack $w.t       -expand yes -fill both

    $w.t insert 0.0 [ getdocu $adt $imp $opr ] 
}

# -------------------------------------------------------------------------------

proc browseadts { } {
    clearlist .adts.list

    foreach i [ call "adts" ] {
        .adts.list insert end $i
    }
}

# -------------------------------------------------------------------------------

proc vdmhelp { } {
    MessageBox "
VDM Class Library Browser

The left list titled \"ADTs\" shows the names of the \
abstract datatypes available. Select one of them by a left button \
double click of the mouse. The currently selected ADT is shown \
at the top of the list. To get \
information about the predefined datatypes, click on the \"predefined\" \
button in the bottom line.

The list in the middle titled \"Implementations\" will then \
show the implementations available for the selected ADT. Select \
one of these Implementations by a left button double click \
of the mouse. The currently selected implementation is shown \
at the top of the list. To get \
information common to all implementations of the selected ADT, \
click on the \"common\" button in the bottom line.

The right list titled \"Operations\" shows the operations \
available for the selected implementation. When selecting one of them \
by a left button double click, a new text window will give \
detailed information about that operation. \
The currently selected operation is shown at the top of the list. \
To get information about the \
features \
of the selected implementation, like time and \
space characteristics and implementation parameters, \
press the \"features\" button in the bottom line.

To leave the browser or any subwindow, \
click at the \"quit\" button.
" "quit" "VDM Help" "120"
}

# Set up bindings for the browser -----------------------------------------------

initdir

browseadts

bind  . <Control-q> {finishapplication}
bind  . <Control-c> {finishapplication}
focus .

bind  .adts.list <Double-Button-1> {
    foreach i [selection get] {
        set curradt $i
        set currimp ""
        set curropr ""
        browseimpl 
    }
}

bind  .impl.list <Double-Button-1> {
    foreach i [selection get] {
        set currimp $i
        set curropr ""
        browseoper 
    }
}

bind  .oper.list <Double-Button-1> {
    global curradt
    global currimp
    global curropr

    foreach i [selection get] {
        set curropr $i
        showoper $curradt $currimp $curropr
    }
}

# -------------------------------------------------------------------------------
