% $Id: expression.pl,v 1.32 1996/07/02 14:34:03 vdm Exp $
%
% expression parsing ---------------------------------------------------------------------

expr(Expr)  -->
	expr(0,Expr).

expr(MinPrio,Expr) -->
	a_primary(Primary),
	{ max_priority(MaxPrio) },
	rest_expr(Primary, MaxPrio, MinPrio, Expr).

a_primary(Primary) -->
	[Token:Pos],
	primary(Token,Pos,Primary),
	{ ! }.

a_primary(X) -->
	syntax_error('primary expression expected', primary, X).

primary(identifier(Name), Pos, Expr) -->
	a_name_or_fct_type_instantiation(Name, Pos, Expr).

primary(Token,Pos,Expr) -->
	symbolic_literal(Token, Pos, Expr).

primary(undefined, Pos, [undefined:Pos]) -->
	[].

primary('(', Pos, [parentexpr:Pos, Expr]) -->
	an_expr(Expr, expr_par),
	test_symbol(')').

primary(let, Pos, Expr) -->
	let_expr(Pos, Expr).

primary(def, Pos, Expr) -->
	def_expr(Pos, Expr).

primary(if, Pos, ['if':Pos,Condition,ThenExpr,ElseExpr]) -->
	cond_expr(Condition,ThenExpr,ElseExpr).

primary(cases, Pos, Expr) -->
	cases_expr(Pos, Expr).

primary( is_(Type), Pos, [ 'is_':Pos,  Type, Expr]) --> bracket_expr(Expr).
primary(inv_(Type), Pos, [ 'inv_':Pos, Type, Expr]) --> bracket_expr(Expr).
primary( pre_(Fct), Pos, [ 'pre_':Pos, Fct | Expr]) --> test_symbol('('), expr_list(')',Expr).
primary(post_(Fct), Pos, ['post_':Pos, Fct | Expr]) --> test_symbol('('), expr_list(')',Expr).

primary(mk_(''), Pos, [mktuple:Pos | ExprList]) -->
	test_symbol('('),
	expr_list(')', ExprList).

primary(mk_(Name), Pos, [mktree:Pos, Name | ExprList]) -->
	{ Name \= '' },
	test_symbol('('),
	empty_expr_list(')', ExprList).

primary(mu, Pos, [mu:Pos, Expr | ModifierList]) -->
	test_symbol('('),
	an_expr(Expr),
	test_symbol(','),
	record_mod_list(')', ModifierList).

primary(forall  ,Pos, Expr) --> quantified_expr_list('forall' ,Pos,Expr).
primary(exists  ,Pos, Expr) --> quantified_expr_list('exists' ,Pos,Expr).
primary(exists1 ,Pos, Expr) --> quantified_expr1('exists1',Pos,Expr).
primary(iota    ,Pos, Expr) --> quantified_expr1('iota'   ,Pos,Expr).

primary(lambda ,Pos,Expr) --> lambda_expr(Pos, Expr).

primary('{', Pos, Expr) --> set_or_map_expr(Pos, Expr).
primary('[', Pos, Expr) --> seq_expr(Pos, Expr).

primary(Op, Pos, Expr) -->
	{ prefix_op(Op, Prio, Op1) },
	expr(Prio, Expr1),
	{trans_unexpr([Op1:Pos, Expr1], Expr)}.

bracket_expr(Expr) -->
	test_symbol('('),
	an_expr(Expr),
	test_symbol(')').

% literals -------------------------------------------------------------------------------

symbolic_literal(number(Num),          Pos,[         number:Pos,Num] ) --> [].
symbolic_literal('quote literal'(Str), Pos,['quote literal':Pos,Str] ) --> [].
symbolic_literal('text literal'(Str),  Pos,[ 'text literal':Pos,Str] ) --> [].
symbolic_literal('char literal'(Char), Pos,[ 'char literal':Pos,Char]) --> [].

symbolic_literal(true,  Pos, [ true:Pos]) --> [].
symbolic_literal(false, Pos, [false:Pos]) --> [].
symbolic_literal(nil,   Pos, [  nil:Pos]) --> [].

% identifiers ----------------------------------------------------------------------------

an_ident([identifier:Pos, Name]) -->
	[ identifier(Name):Pos ],
	{ ! }.

an_ident(X) -->
	syntax_error('identifier expected', expr, X).

a_name(Name, Pos, [old:Pos1, [identifier:Pos, Name]]) -->
	[ old:Pos1 ],
	{ ! }.

a_name(Name, Pos, [moduleid:Pos1, [identifier:Pos, Name], Id2]) -->
	[ '`':Pos1 ],
	an_ident(Id2).

a_name(Name, Pos, [identifier:Pos, Name]) -->
	[].

a_name_or_fct_type_instantiation(Name, Pos, Expr) -->
	a_name(Name, Pos, Expr1),
	fct_type_instantiation(Expr1, Expr).

fct_type_instantiation(Expr, [fcttypeinst:Pos, Expr | TypeList]) -->
	[ '[':Pos ],
	a_type_list(TypeList),
	test_symbol(']').

fct_type_instantiation(Expr, Expr) -->
	[].

% local binding expressions --------------------------------------------------------------

let_expr(Pos, [letexpr:Pos, [letlocaldefs:Pos | ExprList], Expr2]) -->
	local_definitions((','), letlocaldef, ExprList),
	let_rest(Expr2).

let_expr(Pos,Expr) -->
	let_bind_expr('let be':Pos,Expr,Expr2),
	let_rest(Expr2).

let_expr(_,Expr) -->
	syntax_error('syntax error in let expression',expr_in),
	let_rest(Expr).

let_rest(Expr) -->
	test_symbol('in'),
	an_expr(Expr).

let_bind_expr(Op,[Op,Bind,BeST,Expr2],Expr2) -->
	bind(Bind),
	be_such_that_expr(BeST).

be_such_that_expr(Expr) -->
	[ 'be':_ ],
	test_symbol('st'),
	an_expr(Expr).

be_such_that_expr(X) -->
	truesy(X).

local_definitions(Del, Op, [Expr | ExprList]) -->
	local_definition(Op, Expr),
	{ ! },
	local_definition_rest(Del, Op, ExprList).

local_definition_rest(Del, Op, ExprList) -->
	[ Del:_ ],
	local_definitions(Del, Op, ExprList).

local_definition_rest(_, _, []) -->
	[].
	
local_definition(Op, [Op:Pos, Id, Type, Expr1]) -->
	pattern(Id),
	optional_type(Type),
	[ '=':Pos ],
	an_expr(Expr1).

optional_type(Type) -->
	[ ':':_ ],
	type(Type).

optional_type(X) -->
	anysy(X).

def_expr(Pos, [defexpr:Pos, [deflocaldefs:Pos | ExprList], Expr2]) -->
	local_definitions(';', deflocaldef, ExprList),
	let_rest(Expr2).

% conditional expressions ----------------------------------------------------------------

cond_expr(Condition,ThenExpr,ElseExpr) -->
	an_expr(Condition,expr_then),
	test_symbol('then'),
	an_expr(ThenExpr,expr_else),
	else_part(ElseExpr).

else_part([elseif:Pos,Condition,ThenExpr,ElseExpr]) -->
	['elseif':Pos ],
	{ ! },
	cond_expr(Condition,ThenExpr,ElseExpr).	

else_part(Expr) -->
	test_symbol('else'),
	an_expr(Expr).

cases_expr(Pos, [cases:Pos, Cond, [alternatives:Pos1 | AltList]]) -->
	an_expr(Cond),
	test_sympos(':':Pos1),
	cases_expr_alternatives(AltList),
	test_symbol(end).

cases_expr_alternatives([Alt | Rest]) -->
	cases_expr_alternative(Alt),
	cases_expr_alt_rest(Rest).

cases_expr_alt_rest(AltList) -->
	[ ',':_ ],
	cases_expr_alternatives1(AltList).

cases_expr_alt_rest([]) -->
	[].

cases_expr_alternatives1([[defaultcase:Pos, Alt]]) -->
	[ others:_ ],
	test_sympos('->':Pos),
	an_expr(Alt).

cases_expr_alternatives1(AltList) -->
	cases_expr_alternatives(AltList).

cases_expr_alternative([altcase:Pos, [altpatternlist:Pos | CondList], Expr]) -->
	pattern_list1(CondList),
	test_sympos('->':Pos),
	an_expr(Expr).

% quantified and iota expressions --------------------------------------------------------

quantified_expr_list(Op, Pos, Expr) -->
	bind_list(Op:Pos, Expr, Expr2),
	quantified_rest(Expr2).

quantified_expr1(Op, Pos, Expr) -->
	bind(Op:Pos, Expr, Expr2),
	quantified_rest(Expr2).

quantified_rest(Expr) -->
	{ ! },
	test_symbol('&'),
	an_expr(Expr).

% lambda expressions ---------------------------------------------------------------------

lambda_expr(Pos, [lambda:Pos, [params:Pos | Params], Expr1]) -->
	type_bind_list('&', Params),
	an_expr(Expr1).

% set or map construction ----------------------------------------------------------------

set_or_map_expr(Pos, [emptyset:Pos]) -->
	[ '}':_ ],
	{ ! }.
	
set_or_map_expr(_, [emptymap:Pos]) -->
	[ '|->':Pos ],
	{ ! },
	test_symbol('}').

set_or_map_expr(Pos, Expr) -->
	an_expr(Expr1),
	set_or_map_rest(Pos,Expr1,Expr).

set_or_map_rest(StartPos, Expr1, Expr) -->
	[ ',':_ ],
	{ ! },
	set_range_or_enum(StartPos, Expr1, Expr).

							% { e | i in set s & p(i) }
set_or_map_rest(StartPos, Expr1, [setcomprehension:StartPos, Expr1, BindExpr ]) -->
	[ '|':Pos ],
	{ ! },
	bind_filter(bindings:Pos, BindExpr),
	test_symbol('}').

							% { e1 |-> e2 ... }
set_or_map_rest(StartPos,Expr1,Expr) -->
	[ '|->':Pos ],
	{ ! },
	an_expr(Expr2),
	map_rest(StartPos,['|->':Pos, Expr1, Expr2],Expr).

							% { e1 }
set_or_map_rest(StartPos,Expr1,[setenum:StartPos,Expr1]) -->
	[ '}':_ ],
	{ ! }.
							% syntax error
set_or_map_rest(_, _, X) -->
	syntax_error('illegal symbol in set or map expression', setexpr, X),
	test_symbol('}').

bind_filter(Op, BindExpr) -->
	bind_list(Op, BindExpr, Expr3),
	filter_expr(Expr3).

filter_expr(Expr) -->
	[ '&':_ ],
	{ ! },
	an_expr(Expr).

filter_expr(X) -->
	truesy(X).
							% { e1, ..., en }
set_range_or_enum(_StartPos,Expr1,[setrange:Pos,Expr1,Expr2]) -->
	[ '...':Pos ],
	!,
	test_symbol(','),
	an_expr(Expr2),
	test_symbol('}').

							% { e1, e2, e3 }
set_range_or_enum(StartPos,Expr1,[setenum:StartPos,Expr1 | Rest ]) -->
	expr_list('}',Rest).

							% { e1 |-> e2 }
map_rest(StartPos, Expr1, [mapcomprehension:StartPos, Expr1, BindExpr]) -->
	[ '|':Pos ],
	bind_filter(bindings:Pos,BindExpr),
	test_symbol('}').
	
map_rest(StartPos, Expr1, [mapenum:StartPos, Expr1 | MapLetList]) -->
	map_let_list_rest('}', MapLetList).

a_map_let(['|->':Pos,Expr1,Expr2]) -->
	an_expr(Expr1),
	test_sympos('|->':Pos),
	an_expr(Expr2).

% record modification --------------------------------------------------------------------

a_record_mod([recordmod:Pos, Ident, Expr]) -->
	an_ident(Ident),
	test_sympos('|->':Pos),
	an_expr(Expr).

% sequence expressions -------------------------------------------------------------------

seq_expr(Pos, [emptyseq:Pos]) -->
	[ ']':_ ],
	{ ! }.

seq_expr(Pos, Expr ) -->
	an_expr(Expr1),
	seq_expr_rest(Pos, Expr1, Expr).

seq_expr_rest(StartPos, Expr1, [seqcomprehension:StartPos, Expr1, 
				[bindings:Pos, BindExpr, FilterExpr]]) -->
	[ '|':Pos ],
	a_set_bind(BindExpr),
	filter_expr(FilterExpr),
	test_symbol(']').

seq_expr_rest(StartPos, Expr1, [seqenum:StartPos, Expr1 | Rest]) -->
	expr_list_rest(']', Rest).

% expression lists ( cut and paste for different lists !!!??? ) --------------------------

empty_expr_list(FollowSy, []) -->
	[ FollowSy:_ ],
	{ ! }.

empty_expr_list(FollowSy, ExprList) -->
	expr_list(FollowSy, ExprList).

expr_list(FollowSy, [Expr | Rest]) -->
	an_expr(Expr),
	expr_list_rest(FollowSy, Rest).

expr_list_rest(FollowSy,ExprList) -->
	[ ',':_ ],
	{ ! },
	expr_list(FollowSy, ExprList).

expr_list_rest(FollowSy, []) -->
	test_symbol(FollowSy).

% maplet list ----------------------------------------------------------------------------


map_let_list(FollowSy, [Expr | Rest]) -->
        a_map_let(Expr),
        map_let_list_rest(FollowSy, Rest).

map_let_list_rest(FollowSy, ExprList) -->
        [ ',':_ ],
        { ! },
        map_let_list(FollowSy, ExprList).

map_let_list_rest(FollowSy, []) -->
        test_symbol(FollowSy).


% record modifier list -------------------------------------------------------------------

record_mod_list(FollowSy, [Expr | Rest]) -->
        a_record_mod(Expr),
        record_mod_list_rest(FollowSy, Rest).

record_mod_list_rest(FollowSy, ExprList) -->
        [ ',':_ ],
        { ! },
        record_mod_list(FollowSy, ExprList).

record_mod_list_rest(FollowSy, []) -->
        test_symbol(FollowSy).


% type bind list -------------------------------------------------------------------------

type_bind_list(FollowSy, [Expr | Rest]) -->
	a_type_bind(Expr),
	type_bind_list_rest(FollowSy, Rest).

type_bind_list_rest(FollowSy, ExprList) -->
	[ ',':_ ],
	{ ! },
	type_bind_list(FollowSy, ExprList).

type_bind_list_rest(FollowSy, []) -->
	test_symbol(FollowSy).

% def list list --------------------------------------------------------------------------

local_def_list([Def | DefRestList]) -->
	a_definition(Def),
	local_def_list_rest(DefRestList).

local_def_list_rest(DefList) -->
	[ ';':_ ],
	{ ! },
	local_def_list(DefList).

local_def_list_rest([]) -->
	[].

a_definition([defdef:Pos, Id, Expr]) -->
	a_pattern_bind(Id),
	test_sympos('=':Pos),
	an_expr(Expr).

% binary expressions ---------------------------------------------------------------------
%
% some operators consist of 2 or 3 symbols: "in set" "not in set"
% so they are combined here

infix_op(in, 'in set', OpPrio) -->
	[ 'set':_ ],
	{ infix_op('in set',OpPrio) }.

infix_op((not), 'not in set', OpPrio) -->
	[ 'in':_ , 'set':_ ],
	{ infix_op('not in set',OpPrio) }.

infix_op(Op,Op,OpPrio) -->
	{ infix_op(Op,OpPrio) },
	[].

rest_expr(Lhs,MaxPrio,MinPrio,Expr) -->
	[ Op1:Pos ],
	infix_op(Op1, Op, OpPrio),
	{ OpPrio =< MaxPrio, OpPrio > MinPrio },
	!,
	expr(Op, Pos, OpPrio, Lhs, Expr0),
	rest_expr(Expr0, OpPrio, MinPrio, Expr1),
	{ trans_binexpr(Expr1, Expr) }.

rest_expr(Expr,_,_,Expr) -->
	[].

expr('(', Pos, _, Function, [Op:Pos, Function | ParamList]) -->
	{ ! },
	param_list(Op, ParamList).

expr('.', Pos, _, Record, [select:Pos, Record, Id]) -->
	{ ! },
	an_ident(Id).

expr(Op, Pos, OpPrio, Lhs, [Op:Pos, Lhs, Rhs]) -->
	expr(OpPrio, Rhs).

						% f()
param_list(call, []) -->
	[ ')':_ ],
	{ ! }.
						% f(e1 ..... 
param_list(Op, [Expr1 | Rest]) -->
	an_expr(Expr1),
	param_list1(Op, Rest).
						% f(e1, ..... 
param_list1(Op, ParamList) -->
	[ ',':_ ],
	{ ! },
	param_list2(Op, ParamList).
						% f(e1)
param_list1(call, []) -->
	test_symbol(')').
						% f(i,...,j) substring
param_list2(substring, [Expr2]) -->
	[ '...':_ ],
	{ ! },
	test_symbol(','),
	an_expr(Expr2),
	test_symbol(')').
						% f(e1,e2,.....)
param_list2(call, ParamList) -->
	expr_list(')',ParamList).

trans_unexpr(['+u':_,Expr],Expr) :- !.
trans_unexpr(Expr,Expr).

trans_binexpr(Expr,Expr).
%trans_binexpr(Expr,Result) :-
%	trans_assoc_expr(Expr,Result).

% sytactically wrong expression ----------------------------------------------------------

an_expr(Expr) -->
	an_expr(Expr,expr)
	.

an_expr(Expr,_) -->
	expr(Expr),
	{!}
	.

an_expr(X, Nonterminal) -->
	syntax_error('illegal syntax in expression', Nonterminal, X).

% pattern match --------------------------------------------------------------------------
%
% currently only identifiers are patterns

pattern(Pattern) -->
	[ Op:Pos ],
	pattern(Op,Pos,Pattern1),
	pattern_rest(Pattern1,Pattern).

pattern_rest(Pattern1, [ setunionpattern:Pos, Pattern1, Pattern2 ]) -->
	[ 'union':Pos ],
	pattern(Pattern2).

pattern_rest(Pattern1, [ seqconcpattern:Pos, Pattern1, Pattern2 ]) -->
	[ '^':Pos ],
	pattern(Pattern2).

pattern_rest(Pattern,Pattern) -->
	[].

pattern(identifier(Name), Pos, [ patternid:Pos, Name  ]) --> [] .
pattern('-',              Pos, [ dontcare:Pos ]        ) --> [] .
pattern(Op,               Pos, [ matchvalue:Pos, Expr ]) -->
							symbolic_literal(Op,Pos,Expr) .
pattern('(',              Pos, [ matchvalue:Pos, Expr ]) -->
							an_expr(Expr,expr_par),
							test_symbol(')') .
pattern(mk_(''),          Pos, [ mktuple:Pos | PatternList ]) -->
							test_symbol('('),
							pattern_list2(PatternList),
							test_symbol(')') .
pattern(mk_(TreeName),    Pos, [ mktree:Pos, TreeName | PatternList ]) -->
							{ TreeName \= '' },
							test_symbol('('),
							pattern_list0(PatternList),
							test_symbol(')').
pattern('{',              Pos, [ setenumpattern:Pos | PatternList ]) -->
	pattern_list1(PatternList),
	test_symbol('}').

pattern('[',              Pos, [ seqenumpattern:Pos | PatternList ]) -->
	pattern_list1(PatternList),
	test_symbol(']').

pattern_list2([Pattern1|Pattern]) -->
	a_pattern(Pattern1),
	test_symbol(','),
	pattern_list1(Pattern).

pattern_list1([Pattern1|Pattern]) -->
	a_pattern(Pattern1),
	pattern_list_rest(Pattern).

pattern_list0([ Pattern1 | Pattern ]) -->
	pattern(Pattern1),
	{ ! },
	pattern_list_rest(Pattern).

pattern_list0([]) -->
	[].
	
pattern_list_rest(Pattern) -->
	[ ',':_ ],
	{ ! },
	pattern_list1(Pattern).

pattern_list_rest([]) -->
	[].

% this predicate succeeds always, maybe with an error message ----------------------------

a_pattern(Pattern) -->
	[ Op:Pos ],
	pattern(Op,Pos,Pattern),
	{ ! }.

a_pattern(_) -->			% the error message
	[ X:Pos ],
	{ message(Pos,['identifier or pattern expected but "',X,'" found'])
	, fail
	}.

a_pattern(X) -->		% the result in case of an error
	anysy(X).

% bindings -------------------------------------------------------------------------------

bind(Expr) -->
	pattern(Pattern),
	set_or_type_bind([Pattern],Expr)
	.

bind(Op,[Op,Bind,Expr],Expr) -->
	bind(Bind)
	.

bind_list(Op,[Op,Bind,Expr0],Expr2) -->
	multiple_bind(Bind),
	multiple_bind_rest(Op,Expr0,Expr2)
	.

multiple_bind_rest(Op:_,Expr,Expr2) -->
	[ ',':Pos ],
	bind_list(Op:Pos,Expr,Expr2)
	.

multiple_bind_rest(_,Expr,Expr) -->
	[]
	.

multiple_bind(Expr) -->
	pattern_list1(PatternList),
	set_or_type_bind(PatternList, Expr)
	.

set_or_type_bind(PatternList, Bind ) -->
	[ ':':Pos ],
	{ ! },
	type_bind(Pos, PatternList, Bind).

set_or_type_bind(PatternList, Bind) -->
	[ 'in':Pos ],
	{ ! },
	set_bind(Pos, PatternList, Bind).

set_or_type_bind(_, X) -->
	syntax_error('syntax error in binding', expr, X).

a_type_bind(Expr) -->
	a_pattern(Pattern),
	test_sympos(':':Pos),
	type_bind(Pos,[Pattern], Expr).

type_bind(Pos, PatternList, [ typebind:Pos, Type | PatternList ] ) -->
	a_type(Type, bind).

a_set_bind(Expr) -->
	a_pattern(Pattern),
	test_sympos('in':Pos),
	set_bind(Pos, [Pattern], Expr).

set_bind(Pos, PatternList, [ setbind:Pos, Expr | PatternList ] ) -->
	test_symbol('set'),
	an_expr(Expr).

a_pattern_bind(Expr) -->
	a_pattern(Pattern),
	a_pattern_rest(Pattern, Expr).

a_pattern_rest(Pattern, Expr) -->
	[ ':':Pos ],
	type_bind(Pos,[Pattern], Expr).

a_pattern_rest(Pattern, Expr) -->
	[ 'in':Pos ],
	set_bind(Pos, [Pattern], Expr).

a_pattern_rest(Expr, Expr) -->
	[].

% operator description -------------------------------------------------------------------

max_priority(1000).

% combinators ----------------------------------------------------------------------------

infix_op('**',	710).
infix_op(comp,	720).

% applicators ----------------------------------------------------------------------------

infix_op('(',	600).		% function call
infix_op('.',	600).		% record selector
infix_op('[',	600).		% type instantiation

% arithmetic operators -------------------------------------------------------------------

infix_op('+',	510).
infix_op('-',	510).

infix_op('*',	520).
infix_op('/',	520).
infix_op(div,	520).
infix_op(mod,	520).
infix_op(rem,	520).

% set operators --------------------------------------------------------------------------

infix_op(union,	510).
infix_op('\\',	510).
infix_op(inter,	510).

% map operators --------------------------------------------------------------------------

						% VDM-SL takes merge as map union op
						% IFAD allows munion, we allow both
infix_op(munion,	510).
infix_op(merge,         510).
infix_op('++',		510).

infix_op('<:',		540).
infix_op('<-:',		540).
infix_op(':>',		550).
infix_op(':->',		550).

% seq operators --------------------------------------------------------------------------

infix_op('^',		510).

% relational operators -------------------------------------------------------------------

infix_op('=',		410).
infix_op('<>',		410).
infix_op('<',		410).
infix_op('<=',		410).
infix_op('>',		410).
infix_op('>=',		410).
infix_op(subset,	410).
infix_op(psubset,	410).
infix_op('in set',	410).
infix_op('not in set',	410).

% logical operators ----------------------------------------------------------------------

infix_op('<=>',		310).
infix_op('=>',		320).
infix_op('or',		330).
infix_op('and',		340).

prefix_op(Op, Prio, Op1) :-
	once(prefix_op1(Op, Prio, Op1)).

						% overloding of unary and binary + and - resolved
prefix_op1('+',		560, '+u').
prefix_op1('-',		560, '-u').
						% overloding of unary and binary map merge
prefix_op1(merge,       560, dmerge).
prefix_op1(Op, Prio, Op) :-
	prefix_op(Op, Prio).

% arithmetic operators -------------------------------------------------------------------

prefix_op('+',		560).
prefix_op('+u',		560).
prefix_op('-',		560).
prefix_op('-u',		560).
prefix_op(abs,		560).
prefix_op(floor,	560).

% set operators --------------------------------------------------------------------------

prefix_op(card,		560).
prefix_op(power,	560).
prefix_op(dinter,	560).
prefix_op(dunion,	560).

% map operators --------------------------------------------------------------------------

prefix_op(inverse,	530).

prefix_op(dom,		560).
prefix_op(rng,		560).

% seq operators --------------------------------------------------------------------------

prefix_op(len,		560).
prefix_op(elems,	560).
prefix_op(hd,		560).
prefix_op(tl,		560).
prefix_op(conc,		560).
prefix_op(inds,		560).

% logical operators ----------------------------------------------------------------------

prefix_op('not',	350).

