
% function part syntax -------------------------------------------------------------------

function_definitions([functions:Pos | FuncList]) -->
	[ functions:Pos ],
	function_def_list(FuncList).

function_def_list([Fct | Functions]) -->
	function_def(Fct),
	function_def_rest(Functions).

function_def_list(Functions) -->
	skip_until('syntax error in function heading',function_def_rest),
	function_def_rest(Functions).

function_def_rest(Functions) -->
	[ ';':_ ],
	function_def_list(Functions).

function_def_rest(Functions) -->
	skip_until('syntax error in function definition', function_def_rest),
	function_def_rest(Functions).

function_def_rest([]) -->
	[].

%-----------------------------------------------------------------------------------------

function_def(FuncDef) -->
	[ identifier(Name):Pos ],
	empty_type_var_list(TypeParams),
	function_def_rest(Name, Pos, TypeParams, FuncDef).

empty_type_var_list([typevars:Pos | ParamTypes]) -->
	[ '[':Pos ],
	type_var_id_list(']',ParamTypes).

empty_type_var_list(X) -->
	nonesy(X).

a_type_var_id([typevar:Pos, Id]) -->
	test_sympos('@':Pos),
	an_ident(Id).

function_def_rest(Name, Pos, TypeParams,
	      [ explicitfct:Pos, Name, FctType1, FctBody, Pre, Post]) -->
	[ ':':_ ],
	{ ! },                               % explicit function definition
	a_function_type(FctType),
	{ TypeParams = [ typevars:PosT | _ ]
            -> FctType1 = [ generictype:PosT, TypeParams, FctType]
	;
	  FctType1 = FctType
        },
	an_ident([ _:Pos1, Name1]),
	{ Name = Name1
            -> true
	;
	    message(Pos1, ['function names "', Name, '" and "', Name1, '" differ'])
        },
	parameters_list(Body, FctBody),
	test_symbol('=='),
	an_expr(Body),
	a_cond(pre, Pre),
	a_cond(post, Post).

function_def_rest(Name, Pos, TypeParams,
	[ implicitfct:Pos, Name, TypeParams,
	  Params, Res, Pre, [post:PostPos, PostCond]]) -->
	parameter_types(Params),
	an_ident_type_pair(Res),
	a_cond(pre, Pre),
	test_sympos(post:PostPos),
	an_expr(PostCond).

parameter_types([paramtypes:Pos | ParamTypes]) -->
	test_sympos('(':Pos),
	param_type_pair_list0(ParamTypes).

param_type_pair_list0([]) -->
	[ ')':_ ].

param_type_pair_list0(ParamTypes) -->
	param_type_pair_list(')',ParamTypes).

a_patter_list_type_pair([patterntypepair:Pos, Type | PatternList]) -->
	pattern_list1(PatternList),
	test_sympos(':':Pos),
	a_type(Type).

an_ident_type_pair([idtypepair:Pos,Id,Type]) -->
	an_ident(Id),
	test_sympos(':':Pos),
	a_type(Type).

parameters_list(Body, [fctdef:Pos, Params, Rest]) -->
	test_sympos('(':Pos),
	parameters(Pos, Params),
	parameters_list_rest(Body, Rest).

parameters_list_rest(Body, [fctdef:Pos, Params, Rest]) -->
	[ '(':Pos ],
	parameters(Pos, Params),
	parameters_list_rest(Body,Rest).

parameters_list_rest(Body, Body) -->
	[].

parameters(Pos, [params:Pos | ParamList]) -->
	pattern_list0(ParamList),
	test_symbol(')').
	
a_cond(Op, [Op:Pos, Expr]) -->
	[ Op:Pos ],
	an_expr(Expr).

a_cond(_, X) -->
	truesy(X).

% lists ----------------------------------------------------------------------------------

type_var_id_list(FollowSy, [Expr | Rest]) -->
        a_type_var_id(Expr),
        type_var_id_list_rest(FollowSy, Rest).

type_var_id_list_rest(FollowSy, ExprList) -->
        [ ',':_ ],
        { ! },
        type_var_id_list(FollowSy, ExprList).

type_var_id_list_rest(FollowSy, []) -->
        test_symbol(FollowSy).


param_type_pair_list(FollowSy, [Expr | Rest]) -->
        a_patter_list_type_pair(Expr),
        param_type_pair_list_rest(FollowSy, Rest).

param_type_pair_list_rest(FollowSy, ExprList) -->
        [ ',':_ ],
        { ! },
        param_type_pair_list(FollowSy, ExprList).

param_type_pair_list_rest(FollowSy, []) -->
        test_symbol(FollowSy).


% value definition -----------------------------------------------------------------------

value_definitions([values:Pos | ValueList]) -->
	[ values:Pos ],
	value_def_list(ValueList).

value_def_list([Value | Values]) -->
	value_def(Value),
	value_def_rest(Values).

value_def_rest(Values) -->
	[ ';':_ ],
	value_def_list(Values).

value_def_rest(Values) -->
	skip_until('syntax error in value definition', value_def_rest),
	value_def_rest(Values).

value_def_rest([]) -->
	[].

value_def(['value':Pos, Id, Type, Expr]) -->
	a_pattern(Id),
	optional_type(Type),
	test_sympos('=':Pos),
	an_expr(Expr).

% annotations part -----------------------------------------------------------------------

annotations_part([annotations:Pos, Text]) -->
	[ annotations(Text):Pos ].

% ----------------------------------------------------------------------------------------