% $Id: message.pl,v 1.8 1995/08/26 17:35:25 vdm Exp $

% position handling ----------------------------------------------------------------------

newposition(Line0/Col0,Line1/_,Line0/Col0) :-
	Line0 < Line1,
	!.
newposition(Line0/Col0,Line0/Col1,Line0/Col0) :-
	Col0 =< Col1,
	!.
newposition(_,Pos1,Pos1).


% error messages -------------------------------------------------------------------------

errors_issued(ET) :-
	read_variable(errors,Val),
	(
	  Val = true
		-> nl,writelist([ET,' errors found'])
	;
	  true
	).
	

message(L/C/F, List) :-
	!,
	set_variable(errors,true),
	writelist([F, ':', L,': (',C,')	'|List]).

writelist(L) :-
	writelist1(L),
	nl.

writelist1([]).
writelist1([H|R]) :-
	writehead(H),
	writelist1(R).  

writehead(L/C/F) :-
	!,
	write(F),write('.'),write(L),write('.'),write(C).

writehead(H) :-
	is_list(H),
	!,
	writelist1(H).

writehead(H) :-
	atomic(H),
	!,
	write(H).

writehead(H) :-
	H =.. [Functor , Name],
	member(Functor, [is_, inv_, mk_, pre_, post]),
	!,
	write(Functor), write(Name).

writehead(annotations(_)) :-
	write(annotations).

writehead(H) :-
	H =.. [_Functor|Args],
	writehead(Args).

% trace output ---------------------------------------------------------------------------

:- set_variable(writesymbols,false).

write_symbols(L) :-
	(
	  read_variable(writesymbols,true)
		-> write_symbols0(L)
	;
	  true
	).

write_symbols0([]) :-
	nl.
write_symbols0([H | R]) :-
	write_symbol(H),
	write_symbols0(R).

write_symbol(S:L/C/F) :-
	writelist([F, '.', L, '.', C, ':	', S]).

%-----------------------------------------------------------------------------------------

write_list([]) :-
	write([]), nl.

write_list([H | T]) :-
	write_list([H | T], '[').

write_list([], _) :-
	write(']'), nl.

write_list([H | T], Sy) :-
	write(Sy),
	write(H),
	nl,
	write_list(T, ',').

%-----------------------------------------------------------------------------------------

:- set_variable(writetree,false).
:- set_variable(tokenpos,false).

write_tree(T) :-
	(
	  read_variable(writetree,true)
		-> print_tree(T),nl
	;
	  true
	).

print_tree(T) :-
	print_tree(T,0).

print_tree([Op | Args], Indent) :-
	tab(Indent),
	write('['),
	print_op(Op),
	print_args(Args,Indent).

print_op(Op:Line/Col/File) :-
	!,
	write(Op),
	( read_variable(tokenpos, true)
          ->
	  write(':'), write(File),
	  write('.'), write(Line),
	  write('.'), write(Col)
        ; true
        ).

print_op(Op) :-
	write(Op).

print_args([],_) :-
	write(']').

print_args([Arg | Args], Indent) :-
	atomic(Arg),
	!,
	write(', '),
	write(Arg),
	print_args(Args, Indent).

print_args([Arg | Args], Indent) :-
	is_list(Arg),
	!,
	write(','), nl,
	Indent1 is Indent + 2,
	print_tree(Arg, Indent1),
	print_args1(Args, Indent).

print_args1([],Indent) :-
	print_args([], Indent).

print_args1([Arg | Args], Indent) :-
	atomic(Arg),
	!,
	write(','),nl,
	tab(Indent),
	write(Arg),
	print_args(Args, Indent).

print_args1([Arg | Args], Indent) :-
	is_list(Arg),
	!,
	write(','),nl,
	Indent1 is Indent + 2,
	print_tree(Arg, Indent1),
	print_args1(Args, Indent).

:- set_variable(writeTeX,true).
:- set_variable(umlaute,true).
