% $Id: parseaux.pl,v 1.11 1995/06/12 08:48:54 vdm Exp $

% ----------------------------------------------------------------------------------------
% 
% parser: auxiliary and error predicates -------------------------------------------------
%
% syntax_error optionally overreads symbols, but emits always an error message
%
% skip_until(+Error,+Symbols,_,_)
%	if next symbol is not in set Symbols, Error is reported
%	and symbols are read until next symbol in Symbols

syntax_error(Error, Nonterminal, Dummy) -->
	anysy(Dummy),
	syntax_error(Error, Nonterminal).

syntax_error(Error,Nonterminal) -->
	skip_until(Error,Nonterminal),
	{ ! }
	.
syntax_error(Error,_) -->
	[_:Pos],
		{ message(Pos,[Error])
		, fail
		}
	.

syntax_error(_,_) -->
	[]
	.

skip_until(Error,Nonterminal) -->
	  [X:Pos],
		{ (is_list(Nonterminal) -> Symbols = Nonterminal
		  ; follow_set(Nonterminal,Symbols)
		  ),
		  not(member(X,Symbols)),
		  !,
		  message(Pos,[Error]),
		  message(Pos,['illegal symbol: "',X,'", skipping symbols'])
		},
		skip_until(Symbols)
	.

skip_until(Symbols) -->
	  [X:_],{not(member(X,Symbols)),!},skip_until(Symbols)
	| []
	.

% test_symbol(+Symbol,_,_)
%	test, whether next symbol is Symbol
%	else report error and "insert" Symbol, 
%	that means don't read a symbol from input stream

test_sympos(SyPos) -->
	[ SyPos ],
	{ ! }.

test_sympos(Sy:_) -->
	[ _:Pos ],
	{ message(Pos,['"',Sy,'" inserted']), fail }.

test_sympos(_:X) -->
	{ anypos(X) },
	[].

test_symbol(Sy) -->
	test_sympos(Sy:_Pos).

% dummy symbols and dummy positions -----------------------------------------------------

anysy([any:P]) -->
	{ anypos(P) },
	[].

truesy([true:P]) -->
	{ anypos(P) },
	[].

nonesy([none:P]) -->
	{ anypos(P) },
	[].

anypos(0/0/F) :-
	global_state(filename,F).

% ----------------------------------------------------------------------------------------

follow_sy(specification,        [],              [end_of_file]).
follow_sy(spec_rest,            [specification], [types,functions,state,values,annotations(_)]).

follow_sy(function_definitions, [spec_rest],            []).
follow_sy(function_def_rest,    [function_definitions], [';']).

follow_sy(value_definitions,    [spec_rest],            []).
follow_sy(value_def_rest,       [value_definitions],    [';']).

follow_sy(type_definitions,     [spec_rest],            []).
follow_sy(type_definition_rest, [type_definitions],     [';']).
follow_sy(type_definition,      [type_definition_rest], []).

follow_sy(right_side,           [type_definition],      [inv]).
follow_sy(type,                 [right_side],           []).
follow_sy(type0,                [type],                 [producttype]).
follow_sy(type2,                [type],                 ['->','+>']).
follow_sy(type_par,             [type],                 [')']).
follow_sy(type_bra,             [type],                 [']']).
follow_sy(type_to,              [type],                 ['to']).
follow_sy(expr,                 [type_definition],      [',']).
follow_sy(expr_par,             [expr],                 [')']).
follow_sy(expr_bra,             [expr],                 [']']).
follow_sy(expr_then,            [expr],                 [then]).
follow_sy(expr_else,            [expr],                 [else,elseif]).
follow_sy(expr_in,              [expr],                 [in]).
follow_sy(bind,                 [expr],                 [',','be','&']).
follow_sy(setexpr,              [expr],                 ['}']).
follow_sy(seqexpr,              [expr],                 [']']).
follow_sy(primary,              [expr],                 []).

follow_set(Prod,Follow) :-
	follow_sy(Prod,Prods,Follow0),
	follow_set_list(Prods,Follow1),
	union(Follow0,Follow1,Follow),
	!.

follow_set(Prod,[end_of_file]) :-
	anypos(Pos),
	message(Pos,['internal error when computing follow set, unknown nonterminal "',
		     Prod,'"']).

follow_set_list([],[]).
follow_set_list([H | T],Follow) :-
	follow_set(H,Follow0),
	follow_set_list(T,Follow1),
	union(Follow0,Follow1,Follow).
