
:- ensure_loaded(parseaux).
:- ensure_loaded(type_definitions).
:- ensure_loaded(function_part).

% parse specification --------------------------------------------------------------------

specification([spec:Pos | Defs]) -->
	{ anypos(Pos) },
	definition_list(Defs).
	
definition_list([]) -->
	[ end_of_file:_ ],
	{ ! }.

definition_list([Def | Rest]) -->
	definition_block(Def),
	{ ! },
	definition_list(Rest).

definition_list(List) -->
	syntax_error('error in definition block',specification),
	definition_list(List).

definition_block(Def) -->
	  type_definitions(Def) 
	| state_definition(Def)
	| function_definitions(Def)
        | value_definitions(Def)
        | annotations_part(Def)
%	| impl_spec
	.

state_definition([state:Pos, [record:Pos1, Id | Fields], Inv, Init]) -->
	[ state:Pos ],
	an_ident([_:Pos1, Id]),
	test_symbol('of'),
	field_list(Fields),
	invariant(Inv),
	init(Init),
	test_symbol('end').

% global symbol classification -----------------------------------------------------------

simple_type(bool).
simple_type(char).
simple_type(nat).
simple_type(nat1).
simple_type(int).
simple_type(rat).
simple_type(real).
simple_type(char).
simple_type(token).

numeric_type(nat).
numeric_type(nat1).
numeric_type(int).
numeric_type(rat).
numeric_type(real).

empty_collection_type(set).
empty_collection_type(seq).
empty_collection_type(map).
empty_collection_type(inmap).
empty_collection_type(bag).

collection1_type(set).
collection1_type(seq).
collection1_type(seq1).
collection1_type(bag).

collection2_type(map).
collection2_type(inmap).

structured_type(producttype).
structured_type('->').
structured_type('+>').

function_type_op('->').
function_type_op('+>').

cyclic_type(uniontype).
cyclic_type(optionaltype).
cyclic_type(producttype).

% associativity of operators -------------------------------------------------------------

left_assoc_op('-').
left_assoc_op('/').
left_assoc_op(div).
left_assoc_op(mod).
left_assoc_op(rem).
left_assoc_op('\\').

right_assoc_op('->').
right_assoc_op('+>').
right_assoc_op('=>').

assoc_op(uniontype).
assoc_op(producttype).

assoc_op('+').
assoc_op('*').
assoc_op(union).
assoc_op(inter).
assoc_op('^').
assoc_op('<=>').
assoc_op('or').
assoc_op('and').



%-----------------------------------------------------------------------------------------
