% $Id: scanner.pl,v 1.16 1996/07/02 14:34:03 vdm Exp $

% scanner --------------------------------------------------------------------------------
% symbol//1 the symbol list
% a symbol has the structure: Symbol:Line/Column/File
% while scanning only Line/Column pairs are computed
% File is inserted in "symbols" predicate
% so every symbol is tagget with the starting position in the file
% Line is counted from 1, Column from 0


symbols(L) --> symbols(L,1/0).	% the token list

symbols([T:(P/F) | L], Pos) -->
	symbol(T:P, Pos, NPos),
	{ read_variable(filename, F) },
	symbols(L,NPos).

symbols([],_) -->
	[].

% scanner messages -----------------------------------------------------------------------

scanmessage(Pos, List) :-
	read_variable(filename, F),
	message(Pos/F, List).

% the "complicated" symbols --------------------------------------------------------------

symbol(T    ,Pos,NPos) -->
	" ",		% blank: forget it
	{ !,
	  incrpos(Pos,1,Pos1)
	},
	symbol(T,Pos1,NPos).

symbol(T    ,Pos,NPos) -->
	[10],		% newline: forget it
	{ !,
	  incrlin(Pos,Pos1)
	},
	symbol(T,Pos1,NPos).

symbol(T    ,Pos,NPos) -->
	[9],		% tab: forget it
	{ !,
	  incrtab(Pos,Pos1)
	},
	symbol(T,Pos1,NPos).

symbol(T:Pos, Pos, NPos) -->
	[X],		% identifier
	{ ( isLetter(X) ; X = 0'$ ; X = 0'# ),
	  !
	},
	ident(X,T,Pos,NPos).

symbol(T:Pos, Pos, NPos) -->
	[X],		% number
	{ isDigit(X),
	  !
	},
	number(X,T,L),
	{ incrpos(Pos, L, NPos) }.
	
symbol(T, Pos, NPos) -->
	"--",	% comment
	{ ! },
	comment,
	{ incrlin(Pos, Pos1) },
	symbol(T, Pos1, NPos).

symbol('quote literal'(N):Pos, Pos, NPos) -->
	"<",
	id1(Xs0),
	">",
	{ Xs0 = [_|_],		% otherwise <> symbol
	  !,
	  append("<",Xs0,Xs1),append(Xs1,">",Xs),
	  name(N,Xs),
	  length(Xs,L),
	  incrpos(Pos,L,NPos)
	}.

symbol('text literal'(N):Pos, Pos, NPos) -->
	textliteral(Text, Pos, NPos),
	{ !,
	  name(N,Text) 
	}.

symbol('char literal'(N):Pos,Pos,Npos) -->
	charliteral(Char,Pos,Npos),
	{ !,
	  name(N,Char)
	}.

symbol(T:Pos, Pos, NPos) -->
	symbol(T, L),	% simple symbol
	{ !,
	  incrpos(Pos,L,NPos)
	}.

symbol(T    ,Pos, NPos) -->
	[ X ],		% unknown symbol
	{ !,		% issue message and forget symbol
	  name(N, [ X ]),
	  scanmessage(Pos,['illegal character "', N, '"']),
	  incrpos(Pos, 1, Pos1)
	},
	symbol(T, Pos1, NPos).

% the simple symbols ---------------------------------------------------------------------

symbol('->', 2)    --> "->",     {!}.
symbol('-',  1)    --> "-",      {!}.

symbol('+>', 2)    --> "+>",     {!}.
symbol('++', 2)    --> "++",     {!}.
symbol('+',  1)    --> "+",      {!}.

symbol('==>',3)    --> "==>",    {!}.
symbol('==', 2)    --> "==",     {!}.
symbol('=>', 2)    --> "=>",     {!}.
symbol('=',  1)    --> "=",      {!}.

symbol('<=>',3)    --> "<=>",    {!}.
symbol('<=:',3)    --> "<=:",    {!}.
symbol('<=', 2)    --> "<=",     {!}.
symbol('<>', 2)    --> "<>",     {!}.
symbol('<-:',3)    --> "<-:",    {!}.
symbol('<:', 2)    --> "<:",     {!}.
symbol('<',  1)    --> "<",      {!}.

symbol('>=', 2)    --> ">=",     {!}.
symbol('>',  1)    --> ">",      {!}.

symbol(':->',3)    --> ":->",    {!}.
symbol(':>', 2)    --> ":>",     {!}.
symbol('::', 2)    --> "::",     {!}.
symbol(':',  1)    --> ":",      {!}.

symbol(',',  1)    --> ",",      {!}.

symbol('^',  1)    --> "^",      {!}.

symbol('(',  1)    --> "(",      {!}.
symbol(')',  1)    --> ")",      {!}.
symbol('{',  1)    --> "{",      {!}.
symbol('}',  1)    --> "}",      {!}.
symbol('[',  1)    --> "[",      {!}.
symbol('[',  1)    --> "[",      {!}.
symbol('[',  1)    --> "[",      {!}.
symbol(']',  1)    --> "]",      {!}.

symbol('\\', 1)    --> "\\",     {!}.
symbol('/',  1)    --> "/",      {!}.

symbol('|->',3)    --> "|->",    {!}.
symbol('|',  1)    --> "|",      {!}.
symbol('**', 2)    --> "**",     {!}.
symbol('*',  1)    --> "*",      {!}.
symbol(';',  1)    --> ";",      {!}.
symbol('&',  1)    --> "&",      {!}.
symbol('@',  1)    --> "@",      {!}.
symbol(old,  1)    --> "~",      {!}.
symbol('`',  1)    --> "`",      {!}.
symbol('...',3)    --> "...",    {!}.
symbol('.',  1)    --> ".",      {!}.

symbol(end_of_file,0) --> [-1],  {!}.

% the auxiliary predicates ---------------------------------------------------------------

ident(X,T,Pos,NPos)  -->
	id2(X,Xs),
	{ transident(Xs, T1), length(Xs, L), incrpos(Pos,L,NPos1) },
	scan_annotation(T1,NPos1,T,NPos).

scan_annotation(annotations,Pos,annotations(Text),NPos) -->
	scan_text(Pos, List, NPos),
	{ !, name(Text, List) }.

scan_annotation(T,Pos,T,Pos) -->
	[].

scan_text(Pos, [], NPos) -->    % preliminary: overread everything until "end annotations" appears
	"end annotations",
	{ !, incrpos(Pos, 15, NPos) }.

scan_text(Pos, [10 | Text], NPos) -->
	[10],		
	{ !, incrlin(Pos,Pos1) },
	scan_text(Pos1, Text, NPos).

scan_text(Pos, [9 | Text], NPos) -->
	[9],
	{ !, incrtab(Pos, Pos1)	},
	scan_text(Pos1, Text, NPos).

scan_text(Pos, [X | Text], NPos) -->
	[ X ],
	{ X \= -1, !, incrpos(Pos, 1, Pos1) },
	scan_text(Pos1, Text, NPos).

scan_text(Pos, [], Pos) -->
	[ -1 ],
	{ scanmessage(Pos, ['"annotations" not terminated by "end annotations"']) }.

number(X,number(N),L) -->
	digits(X,L1),
	fract(L2),
	expo(L3),
	{ append(L1,L2,L12), append(L12,L3,L0), name(N,L0), length(L0,L) }.

id1([X | Xs])    --> [X], { isId1char(X), ! }, id1(Xs).
id1([])          --> [].
id1(X, [X | Xs]) --> id1(Xs).

id2([0'#, X1 | Xs]) --> [0'#, X1], { isLetter(X1), ! }, id2(Xs).
id2(     [X  | Xs]) --> [X],       { isId2char(X), ! }, id2(Xs).
id2(            []) --> [].

id2(0'#, [0'#, X1 | Xs]) --> [ X1 ], { isLetter(X1), ! }, id2(Xs).
id2(X,   [     X  | Xs]) --> id2(Xs).

comment --> [10], {!}.	% stop reading when \n reached
comment --> [_Char],comment.

digits([X|Xs]) --> [X],{isDigit(X),!},digits(Xs).
digits([])     --> [].
digits(X,[X|Xs]) --> digits(Xs).

sdigits(Xs)       --> [0'+,X],{isDigit(X),!},digits(X,Xs).
sdigits([0'-|Xs]) --> [0'-,X],{isDigit(X),!},digits(X,Xs).
sdigits(Xs)       -->     [X],{isDigit(X),!},digits(X,Xs).

expo([0'E|Xs]) --> "E",sdigits(Xs),{!}.
expo([0'E|Xs]) --> "e",sdigits(Xs),{!}.
expo([])       --> [].

fract([0'.|Xs]) --> [0'.,X],{isDigit(X),!},digits(X,Xs).
fract([])       --> [].

textliteral(Text,Pos,Npos) -->
	[ 0'" ],
	{ !, incrpos(Pos,1,Pos1) },
	textliteralpart(Text1,Pos1,Pos2),
	textliteralrest(Text2,Pos2,Npos),
	{ append(Text1,Text2,Text) }
	.
	
textliteralrest([0'" | Text],Pos,NPos) -->
	textliteral(Text,Pos,NPos)
	.
textliteralrest([],Pos,Pos) -->
	[]
	.

textliteralpart([],Pos,Pos1) -->
	[ 0'" ],
	{ incrpos(Pos,1,Pos1) }
	.	
textliteralpart([],Pos,Pos1) -->
	[ 10 ],
	{ incrlin(Pos,Pos1)
	, scanmessage(Pos,['newline in text literal'])
	}
	.
textliteralpart([X|Xs],Pos,Pos1) -->
	[ X ],
	{ X =\= 0'", X =\= 10, incrpos(Pos,1,Pos0) },
	textliteralpart(Xs,Pos0,Pos1)
	.

charliteral([Char],Pos,Pos1) -->
	[ 0'' , Char, 0'' ],
	{ incrpos(Pos,3,Pos1) }
	.

% the keywords ---------------------------------------------------------------------------

keyword("and").
keyword("annotations").
keyword("as").
keyword("abs").
keyword("always").
keyword("be").
keyword("bool").
keyword("by").
keyword("card").
keyword("cases").
keyword("char").
keyword("comp").
keyword("compose").
keyword("conc").
keyword("dcl").
keyword("def").
keyword("defined").
keyword("definitions").
keyword("dinter").
keyword("div").
keyword("do").
keyword("dom").
keyword("dunion").
keyword("elems").
keyword("else").
keyword("elseif").
keyword("end").
keyword("error").
keyword("errs").
keyword("exists").
keyword("exists1").
keyword("exit").
keyword("exports").
keyword("ext").
keyword("false").
keyword("floor").
keyword("for").
keyword("forall").
keyword("from").
keyword("functions").
keyword("get_value").
keyword("hd").
keyword("if").
keyword("implementations").
keyword("imports").
keyword("in").
keyword("inds").
keyword("init").
keyword("inmap").
keyword("instantiation").
keyword("int").
keyword("inter").
keyword("inv").
keyword("inverse").
keyword("iota").
keyword("lambda").
keyword("len").
keyword("let").
keyword("map").
keyword("merge").
keyword("mod").
keyword("module").
keyword("munion").
keyword("mu").
keyword("nat").
keyword("nat1").
keyword("nil").
keyword("not").
keyword("of").
keyword("operations").
keyword("or").
keyword("others").
keyword("parameters").
keyword("post").
keyword("power").
keyword("pre").
keyword("psubset").
keyword("put_value").
keyword("rat").
keyword("rd").
keyword("real").
keyword("rem").
keyword("return").
keyword("reverse").
keyword("rng").
keyword("seq").
keyword("seq1").
keyword("set").
keyword("skip").
keyword("st").
keyword("state").
keyword("subset").
keyword("then").
keyword("tixe").
keyword("token").
keyword("tl").
keyword("to").
keyword("trap").
keyword("true").
keyword("types").
keyword("undefined").
keyword("union").
keyword("using").
keyword("values").
keyword("yet").
keyword("while").
keyword("wr").
keyword("RESULT").

keyprefix("is_").
keyprefix("inv_").
keyprefix("mk_").
keyprefix("pre_").
keyprefix("post_").

% identifiers --> keywords ---------------------------------------------------------------

transident(L,N) :-		% keywords
	keyword(L),
	!,
	name(N,L).

transident(L,N) :-		% keyprefixes is_..., mk_..., ...
	keyprefix(P),append(P,L1,L),
	!,
	name(N0,P),name(N1,L1),N =..[N0,N1].

transident(L,identifier(N)) :-		% ordinary identifiers
	name(N,L).

% simple character classification --------------------------------------------------------

isUppercase(X) :- X >= 0'A, X =< 0'Z.

isLowercase(X) :- X >= 0'a, X =< 0'z.

isLetter(X) :-
	( isUppercase(X) -> true
	; isLowercase(X) -> true
        ; isUmlaut(X)
	).

isUmlaut(X) :-
	read_variable(umlaute,true),
	member(X, "").

isDigit(X) :- X >= 0'0, X =< 0'9.

isWhitespace(X) :-
	( X == 32  -> true	% space or
	; X == 9   -> true	% \t or
	; X == 10		% \n
	).

isLetterOrDigit(X) :-
	( isLetter(X) -> true
        ; isDigit(X)
        ).

isId1char(X) :-
	( isLetter(X) -> true
	; isDigit(X)  -> true
	; X = 0'_
	).

isId2char(X) :-
	( isId1char(X) -> true
	; X = 0''
	).

% symbol positions -----------------------------------------------------------------------

incrpos(L/C,N,L/C1) :- C1 is C + N.
incrlin(L/_,L1/0) :- L1 is L + 1.
incrtab(L/C,L/C1) :- C1 is (C + 8) // 8 * 8.

%-----------------------------------------------------------------------------------------
