% ------------------------------------------------------------------------------

% $Id: totex.pl,v 1.24 1996/07/02 17:28:38 vdm Exp $

% toTeX(+Spec, -TeX)
% translates a Spec into a nestet List of LaTeX text

						% empty list
listToTeX([], _Op, _Del, []).

						% 1 element list
listToTeX([Arg], _Op, _Del, [TeXArg]) :-
	toTeX(Arg, TeXArg).

						% 2 or more element list
listToTeX([Arg | Rest], Op, Del, [TeXArg, NLinfront, Del, NLbehind | TeXRest]) :-
	Rest = [Arg2 | _],
	toTeX(Arg, TeXArg),
	lineBreakInList(Arg, Arg2, Op, NLinfront, NLbehind),
	listToTeX(Rest, Op, Del, TeXRest).

						% start predicate
toTeX([Op:Pos | Args], TeXExpr) :-
	toTeX(Op, Args, Pos, TeXExpr).
						% separate Op and Args

						% keyword constants 0-ary ops
toTeX(Op, [], _, [ TeXOp ]) :-
	texOp(Op, TeXOp).

						% identifier
toTeX(Op, [Id], _, [TeXId, ' ']) :-
	idOp(Op),
	idToTeX(Id, TeXId).

toTeX(number, [Num], _, [TeXNum, ' ']) :-
	trNrTeX(Num, TeXNum).

toTeX(Op, [Arg1], _, [Pre, TeXArg1, Post]) :-
	texOp1(Op, Pre, Post),
	toTeX(Arg1, TeXArg1).

						% binary expressions
toTeX(Op, [Arg1, Arg2], Pos, [Pre, TeXArg1, NLinfront, In, NLbehind, TeXArg2, Post]) :-
	texOp2(Op, Pre, In, Post),
	lineBreakInBinexpr(Arg1, Arg2, Op, Pos, NLinfront, NLbehind),
	toTeX(Arg1, TeXArg1),
	toTeX(Arg2, TeXArg2).

						% list ops
toTeX(Op, Args, _, [Pre, TeXArgs, Post]) :-
	texOpl(Op, Pre, Del, Post),
	listToTeX(Args, Op, Del, TeXArgs).

toTeX(Op, [Expr, Bindings], _, TeXExpr) :-
	texCompOp(Op, TeXCmd),
	comprehensionToTeX(TeXCmd, Expr, Bindings, TeXExpr).

toTeX(annotations, [Text], _, TeXText) :-
	annotationsToTeX(Text, TeXText).

toTeX(call, [Fct | Params], _, ['\\fnapply{', TeXFct, '}{',TeXParams, '}']) :-
	toTeX(Fct, TeXFct),
	commaListToTeX(Params, TeXParams).

toTeX(fcttypeinst, [FctName | TypeList], _, ['{', TeXFctName, '[', TeXTypeList, ']}']) :-
	toTeX(FctName, TeXFctName),
	commaListToTeX(TypeList, TeXTypeList).

toTeX(explicitfct, [FctName, FctType, FctDef, Pre, Post], _,
	['\\begin{fn}[e]{', TeXFctName, '}\\annlab[o]{', AnnLabName, '}', '\n',
	'\\signature', TeXFctType, '\n',
	'\\parms*{', TeXFctDef,
	TeXPre,
	TeXPost,
	'\\end{fn}', '\n']
    ) :-
	idToTeX(FctName, TeXFctName),
	idToAnnLab(FctName, AnnLabName),
	fctTypeToTeX(FctType, TeXFctType),
	fctDefToTeX(FctDef, TeXFctDef),
	preOrPostToTeX(Pre,  TeXPre),
	preOrPostToTeX(Post, TeXPost).

toTeX(field, [FieldName, Type], _, [TeXFieldName, ':', TeXType, '\\\\', '\n']) :-
	fieldIdToTeX(FieldName, TeXFieldName),
	toTeX(Type, TeXType).

toTeX(if, [Cond, ThenExpr, ElseExpr], _,
	['\\If ', TeXCond, '\n',
	'\\Then\\\\ \\mbox{}\\\\', '\n', TeXThenExpr, '\n',
	TeXElseExpr, '\n',
	'\\Fi']
    ) :-
	toTeX(Cond, TeXCond),
	toTeX(ThenExpr, TeXThenExpr),
	elseToTeX(ElseExpr, TeXElseExpr).

toTeX(elseif, [Cond, ThenExpr, ElseExpr], _,
	['\\Elseif ', TeXCond, '\n',
	'\\Then\\\\ \\mbox{}\\\\', '\n', TeXThenExpr, '\n',
	TeXElseExpr ]
    ) :-
	toTeX(Cond, TeXCond),
	toTeX(ThenExpr, TeXThenExpr),
	elseToTeX(ElseExpr, TeXElseExpr).

toTeX(implicitfct, [FctName, TypeParams, ParamTypes, ResultType, Pre, Post], _,
	['\\begin{fn}[i]{', TeXFctName, '}\\annlab[o]{', AnnLabName, '}', '\n',
	'\\parms*{', TeXTypeParams, TeXParamTypes, '}\\ ', TeXResultType, '\n',
	TeXPre,
	TeXPost,
	'\\end{fn}', '\n']
    ) :-
	idToTeX(FctName, TeXFctName),
	idToAnnLab(FctName, AnnLabName),
	toTeX(TypeParams, TeXTypeParams),
	toTeX(ParamTypes, TeXParamTypes),
	toTeX(ResultType, TeXResultType),
	preOrPostToTeX(Pre,  TeXPre),
	preOrPostToTeX(Post, TeXPost).

toTeX('is_', [TypeName, Arg], _, ['{\\kw{is-}' ,TeXTypeName, '\\Lp ', TeXArg, '\\Rp}']) :-
	typeIdToTeX(TypeName, TeXTypeName),
	toTeX(Arg, TeXArg).

toTeX(Op, Args, _, TeXExpr) :-
	texLetOp(Op, TeXEnv),
	localDefToTeX(TeXEnv, Args, TeXExpr).

toTeX('let be', [Bind, Filter, Expr], _,
	['\\begin{letbeexpr}', '\n',
	'\\such{', TeXBind, '}', TeXFilter, '\n',
	'\\end{letbeexpr}\\\\', '\n',
	TeXExpr]
    ) :-
	toTeX(Bind, TeXBind),
	letFilterToTeX(Filter, Bind, TeXFilter),
	toTeX(Expr, TeXExpr).

toTeX(Op, [Pattern, Type, Expr], _,
	[ TeXCmd, '{', TeXPattern, TeXType, '}{', TeXExpr, '}', '\n']
    ) :-
	texLocalDef(Op, TeXCmd),
	toTeX(Pattern, TeXPattern),
	anyTypeToTeX(Type, TeXType),
	toTeX(Expr, TeXExpr).

toTeX(mapenum, Args, _, ['\\map{', TeXArgs, '}']) :-
	commaListToTeX(Args, TeXArgs).

toTeX(matchvalue, [ Arg ], _, TeXExpr) :-
	Arg = [ Op:_ | _ ],
	member(Op, [number, 'quote literal', 'text literal', 'char literal', true, false, nil]),
	toTeX(Arg, TeXExpr).

toTeX(matchvalue, Args, _, TeXExpr) :-
	toTeX(matchvalue1, Args, _, TeXExpr).

toTeX(mktree, [Name | Components], _,
	['\\reccons{\\kw{mk-}', TeXName, '}{', TeXComponents, '}']
    ) :-
	idToTeX(Name, TeXName),
	commaListToTeX(Components, TeXComponents).

toTeX(mktuple, Components, _,
	['\\reccons{\\kw{', Keyw, '}}{', TeXComponents, '}']
    ) :-
	tupleKeyWord(Keyw),
	commaListToTeX(Components, TeXComponents).

toTeX(mu, [RecExpr | RecMod], _,
	['\\recmod{', TeXRecExpr, '}', LBr, '{', TeXRecMod, '}']
    ) :-
	toTeX(RecExpr, TeXRecExpr),
	RecMod = [ RecMod1 | _ ],
	lineBreak(RecExpr, RecMod1, LBr),
	commaListToTeX(RecMod, TeXRecMod).

toTeX(patterntypepair, [Type | PatternList], _,
	[TeXPatternList, ': ', TeXType]
    ) :-
	toTeX(Type, TeXType),
	commaListToTeX(PatternList, TeXPatternList).

toTeX('quote literal', [Id], _, ['\\const{', TeXId, '}']) :-
	idToTeX(Id, TeXId).

						% records
						% empty records
toTeX(record, [Name], _,
	[ '\\compose{', TeXName, '}{\\mbox{}}', '\n' ]
    ) :-
	idToTeX(Name, TeXName).

						% field with one anonymous component
toTeX(record, [Name, [field:_, '_', Type]], _,
	[ '\\compose{', TeXName, '}{\\\\', TeXType, '\\\\}', '\n' ]
    ) :-
	idToTeX(Name, TeXName),
	toTeX(Type, TeXType).

						% records with 2 or more components
						% or with a named component
toTeX(record, [Name | Fields], _,
	['\n', '\\begin{composite}[', LongestId, ']{', TeXName, '}\\annlab[o]{', AnnLabName, '}', '\n',
	FieldsTeX,
	'\\end{composite}', '\n']
    ) :-
	idToTeX(Name, TeXName),
	idToAnnLab(Name, AnnLabName),
	findall(FieldId, member([field:_, FieldId | _], Fields), FieldIds),
	maxId(FieldIds, LongestId),
	lineListToTeX(Fields, FieldsTeX).

toTeX(setbind, Args, _, TeXArgs) :-
	bindToTeX(Args, '\\In ', TeXArgs).

toTeX(substring, [Expr, LbExpr, UbExpr], _, 
	['\\subseq{', TeXExpr, '}{', TeXLbExpr, '}', LBr, '{', TeXUbExpr, '}' ]
    ) :-
	toTeX(Expr, TeXExpr),
	toTeX(LbExpr, TeXLbExpr),
	lineBreak(LbExpr, UbExpr, LBr),
	toTeX(UbExpr, TeXUbExpr).

toTeX(altcase, [Cond, Expr], _, ['\\alt{', TeXCond, LBr, '}{', TeXExpr, '}', '\n'] ) :-
	toTeX(Cond, TeXCond),
	lineBreak(Cond, Expr, LBr),
	toTeX(Expr, TeXExpr).

toTeX(state,
	[ [record:_, StateId | Fields], Inv, Init],
	_,
	['\n', '\\begin{statedef}[', LongestId, ']{', StateId, '}', '\n',
	TeXFields, TeXInv, TeXInit,
	'\\end{statedef}', '\n']
    ):-
	findall(FieldId, member([field:_, FieldId | _], Fields), FieldIds),
	maxId(FieldIds, LongestId),
	lineListToTeX(Fields, TeXFields),
	invToTeX(Inv, TeXInv),
	initToTeX(Init, TeXInit).

toTeX('text literal', [Str], _, ['\\mbox{\\tt\\symbol{127}', TeXStr, '\\symbol{127}}']) :-
	textLitToTeX(Str, TeXStr).

toTeX('char literal', [Chr], _, ['\\Quote{}', Chr, '\\Quote{}']).

toTeX(typebind, Args, _, TeXArgs) :-
	bindToTeX(Args, ': ', TeXArgs).

toTeX(types, TypeList, Op0,
	['\\recidwidth{', LongestId, '}%',
	TeX]) :-
						% indentation computation
						% all type ids are computed and
						% length of longest is stored in global state
						% for records a new indent variable is generated
						% initialized and the name is stored in global_state
	findall(TypeId, member([typedef:_, TypeId | _], TypeList), TypeIds),
	maxId(TypeIds, LongestId),
	change_variable(typedefIndent, LongestId, OldId),
						% recursive call with an auxiliary list op types0
	toTeX(types0, TypeList, Op0, TeX),
	set_variable(typedefIndent, OldId).

						% record types
						% with 2 or more components
						% or with 1 named component
toTeX(typedef, Args, _, ArgsTeX) :-
	Args = [TypeName, [record:_, TypeName | Fields] | _],
	Fields \= [],
	Fields \= [[field:_, '_', _ ]],
	recordToTeX(Args,ArgsTeX).

						% other types
toTeX(typedef, [TypeName, TypeExpr, Inv], _,
	['\\type[', Indent, ']{', TeXTypeName, '}{', TeXTypeExpr, '}\\annlab[o]{', AnnLabName, '}', '\n', TeXInv]) :-
	read_variable(typedefIndent, Indent),
	idToTeX(TypeName, TeXTypeName),
	idToAnnLab(TypeName, AnnLabName),
	toTeX(TypeExpr, TeXTypeExpr),
	invToTeX(Inv, TeXInv).

toTeX(values, ValueList, _,
	['\\begin{valuedef}', '\n', TeXValueList, '\\end{valuedef}']) :-
	lineListToTeX(ValueList,TeXValueList).


toTeX(value, [ValueId, ValueType, ValueExpr], _,
	['\\val{', TeXValueId, '}', TeXValueType, '{', '\\\\', TeXValueExpr, '}', '\n']) :-
	toTeX(ValueId,   TeXValueId),
	anyValTypeToTeX(ValueType, TeXValueType),
	toTeX(ValueExpr, TeXValueExpr).

anyValTypeToTeX([any:_], []).
anyValTypeToTeX(Type, ['[', TeXType, ']']) :-
	toTeX(Type, TeXType).

						% the default rule
toTeX(Op, _, _, ['% ', Op, ' not yet translated', '\n', '<???>']).

% some auxiliary predicates ---------------------------

anyValTypeToTeX([any:_], []).
anyValTypeToTeX(Type, ['[', TeXType, ']']) :-
	toTeX(Type, TeXType).


						% VDM-SL uses mk-(_,...,_) for tuples
						% this flag enables shorter syntax (_,...,_)
tupleKeyWord(Keyw) :-
	(
	    read_variable(mktuple,true) -> Keyw = ''
	;
	    Keyw = 'mk-'
	).


% ------------------------------------------------------------------------------

idOp('type identifier').
idOp('identifier').
idOp('patternid').

idToTeX(Id, TeXId) :-
	name(Id, List),
	trTeX(List, TeXList),
	name(TeXId, TeXList).

typeIdToTeX(Id, TeXId) :-
	simple_type(Id),
	texOp(Id, TeXId).

typeIdToTeX(Id, TeXId) :-
	idToTeX(Id, TeXId).

fieldIdToTeX('_', '').
fieldIdToTeX(Id, TeXId) :-
	idToTeX(Id, TeXId).

trTeX([], []).
						% keywords as identifier: remove $
trTeX([0'$ | L], L).
						% quote literals: 60 = <, 62 = >
trTeX([0'< | L1], L) :-
	append(L0, [ 0'> ], L1),
	trTeX(L0, L).

trTeX([0'_ | L1], [0'- | L2]) :-
	trTeX(L1, L2).
						% Umlaute translation
trTeX([C | L1], L2) :-
	charToTeX(C, Atom),
	trTeX(L1, Rest),
	name(Atom, First),
	append(First, Rest, L2).

trTeX(L, TrL) :-
	isNumberSuffix(L),
	trNumberSuffix(L,TrL).
						% else (preliminary): do nothing
trTeX([C | L1], [C | L2]) :-
	trTeX(L1,L2).

trNrTeX(Num, TeXNum) :-
	name(Num, List),
	trNrListTeX(List, TeXNum).

						% real numbers
trNrListTeX(List, [N1, '\\Exponent ', N2]) :-
	append(L1, [0'E | L2], List),
	name(N1, L1),
	name(N2, L2).
						% other numbers
trNrListTeX(List, Num) :-
	name(Num, List).

% no characters exept letters an digits are allowed in annotation labels
idToAnnLab(Id, TeXId) :-
	name(Id, List),
	once(trAnnLab(List, TeXList)),
	name(TeXId, TeXList).

trAnnLab([], []).

trAnnLab([C | L1], L2) :-
	trUmlaut(C, L21),
	trAnLab(L1, L22),
	append(L21, L22, L2).

trAnnLab([C | L1], [C | L2]) :-
	isLetterOrDigit(C),
	trAnnLab(L1, L2).

trAnnLab([ _ | L1], L2) :-
	trAnnLab(L1,L2).

trUmlaut(0', "Ae").
trUmlaut(0', "Oe").
trUmlaut(0', "Ue").
trUmlaut(0', "ae").
trUmlaut(0', "oe").
trUmlaut(0', "ue").
trUmlaut(0', "ss").

% ------------------------------------------------------------------------------
						% constant ops

texOp(any,       '\\mbox{???} ' ).
texOp(bool,      '\\Bool '      ).
texOp(char,      '\\Char '      ).
texOp(dontcare,  ' - '         ).
texOp(emptymap,  '\\Emptymap '  ).
texOp(emptyseq,  '\\Emptyseq '  ).
texOp(emptyset,  '\\Emptyset '  ).
texOp(false,     '\\False '     ).
texOp(int,       '\\Int '       ).
texOp(nat,       '\\Nat '       ).
texOp(nat1,      '\\Nati '      ).
texOp(nil,       '\\Nil '       ).
texOp(none,      ''            ).
texOp(producttype0, '\\Lp \\Rp ' ).
texOp(rat,       '\\Rat '       ).
texOp(real,      '\\Real '      ).
texOp(token,     '\\Token '     ).
texOp(true,      '\\True '      ).
texOp(undefined, '\\Undefined ' ).

% ------------------------------------------------------------------------------
						% unary ops

texOp1(seq,          '\\seqof*{',                   '}').
texOp1(seq1,         '\\seqof+{',                   '}').
texOp1(set,          '\\setof{',                    '}').
texOp1(optionaltype, '\\Opt{',                      '}').
texOp1(pre,          ['\\begin{precond}\\\\',  '\n'], ['\n', '\\end{precond}',  '\n']).
texOp1(post,         ['\\begin{postcond}\\\\', '\n'], ['\n', '\\end{postcond}', '\n']).
texOp1(inverse,      '{',                          '}\\Inverse ').
texOp1(matchvalue1,  '\\Lp ',                       '\\Rp ').
texOp1(defaultcase,  '\\others{',                   ['}', '\n']).
texOp1(parentexpr,   '\\Lp ',                       '\\Rp ').
texOp1(parenttype,   '\\Lp ',                       '\\Rp ').
texOp1(old,          '\\old{',                      '}').

texOp1(Op, Pre, '') :-
	texPreOp1(Op, Pre).

texPreOp1('-u',   '\\Uminus ').
texPreOp1(abs,    '\\Abs ').
texPreOp1(card,   '\\Card ').
texPreOp1(conc,   '\\Conc ').
texPreOp1(dom,    '\\Dom ').
texPreOp1(dinter, '\\Dinter ').
texPreOp1(dunion, '\\Dunion ').
texPreOp1(elems,  '\\Elems ').
texPreOp1(floor,  '\\Floor ').
texPreOp1(hd,     '\\Hd ').
texPreOp1(inds,   '\\Inds ').
texPreOp1(len,    '\\Len ').
texPreOp1(dmerge, '\\Merge ').
texPreOp1('not',  '\\Not ').
texPreOp1(power,  '\\Power ').
texPreOp1(rng,    '\\Rng ').
texPreOp1(tl,     '\\Tl ').
texPreOp1(typevar,'@').

% ------------------------------------------------------------------------------
						% binary ops

texOp2(map,      '{',          '\\Gmap ',    '}').
texOp2(inmap,    '{',          '\\Bmap ',    '}').
texOp2(forall,   '\\all{',      '}\\\\{',      '}').
texOp2(exists,   '\\exists{',   '}\\\\{',      '}').
texOp2(exists1,  '\\unique{',   '}\\\\{',      '}').
texOp2(iota,     '\\iota{',     '}\\\\{',      '}').
texOp2(setrange, '\\setrange{', '}{',        '}').
texOp2(cases,    '\\begin{Cases}[xxxxxxx]{',
	                       ['}', '\n'], '\\end{Cases}').
texOp2(lambda,   '\\lambda{',   '}\\\\{',      '}').

texOp2(Op, '', In, '') :-
	texInOp2(Op, In).

texInOp2('and',  '\\And ').
texInOp2('or',   '\\Or ').
texInOp2('=>',   '\\Implies ').
texInOp2('<=>',  '\\Equiv ').

texInOp2('=',   '= ').
texInOp2('<>',  '\\Neq ').

texInOp2('<',   '< ').
texInOp2('<=',  '\\Le ').
texInOp2('>',   '> ').
texInOp2('>=',   '\\Ge ').

texInOp2('+',   '+ ').
texInOp2('-',   '\\Minus ').
texInOp2('*',   '* ').
texInOp2('/',   '/ ').
texInOp2(div,   '\\Div ').
texInOp2(mod,   '\\Mod ').
texInOp2(rem,   '\\Rem ').

texInOp2('^',   '\\Sconc ').

texInOp2(union,   '\\Union ').
texInOp2(inter,   '\\Inter ').
texInOp2('\\',     '\\Setdiff ').
texInOp2(subset,  '\\Subset ').
texInOp2(psubset, '\\Psubset ').

texInOp2('in set',     '\\In '      ).
texInOp2('not in set', '\\Notin '   ).

texInOp2(setunionpattern, '\\Union ').
texInOp2(seqconcpattern,  '\\Sconc ').

texInOp2(munion,       '\\Mapmerge ').
texInOp2(merge,        '\\Mapmerge ').
texInOp2('++',         '\\Override ').
texInOp2('<:',         '\\Dto '     ).
texInOp2('<-:',        '\\Dby '     ).
texInOp2(':>',         '\\Rto '     ).
texInOp2(':->',        '\\Rby '     ).
texInOp2('**',         '\\Iterate ' ).
texInOp2('comp',       '\\Compose ' ).
texInOp2('|->',        '\\Mapsto '  ).
texInOp2(recordmod,    '\\Recmod '  ).
texInOp2(select,       '.'         ).

texInOp2('->',  ' \\To ').
texInOp2('+>',  ' \\Pto ').

texInOp2(idtypepair, ': ').
texInOp2(moduleid,   '` ').

% ------------------------------------------------------------------------------
						% list ops

texOpl(Op, ['\n', '\\begin{', Env, '}', '\n'], '', ['\\end{', Env, '}', '\n', '\n']) :-
	texEnvOpl(Op, Env).

texOpl(Op, '', Del, '') :-
	texInOpl(Op, Del).

texOpl(setenumpattern,    '\\set{', ',', '}'   ).
texOpl(setenum,           '\\set{', ',', '}'   ).
texOpl(seqenum,           '\\seq{', ',', '}'   ).
texOpl(seqenumpattern,    '\\seq{', ',', '}'   ).
texOpl(paramtypes,        '\\Lp ',  ',', '\\Rp ').
texOpl(typevars,          '[',     ',', ']'   ).
texOpl(params,            '(',     ',', ')'   ).
texOpl(letlocaldefs,      '',      '',  ''    ).
texOpl(deflocaldefs,      '',      '',  ''    ).
texOpl(alternatives,      '',      '',  ''    ).
texOpl(altpatternlist,    '',      ',', ''    ).

texEnvOpl(spec,      vdm).
texEnvOpl(types0,    typedef).
texEnvOpl(functions, funcdef).
%texEnvOpl(pre,       precond).
%texEnvOpl(post,      postcond).

texInOpl(producttype, '\\X ').
texInOpl(uniontype,   '| ' ).

texCompOp(setcomprehension, '\\setcomp').
texCompOp(mapcomprehension, '\\mapcomp').
texCompOp(seqcomprehension, '\\seqcomp').

texLocalDef(letlocaldef, '\\patdef').
texLocalDef(deflocaldef, '\\eqdef').

recordToTeX([TypeName,
	[record:_, _ | Fields], Inv],
	['\\begin{record}[', LongestId, ']{\\lineup[t]{\\RecIndent}{', TeXTypeName, '}{}}\\annlab[o]{', AnnLabName, '}', '\n',
	FieldsTeX,
	'\\end{record}', '\n',
	InvTeX ]
    ) :-
	idToTeX(TypeName, TeXTypeName),
	idToAnnLab(TypeName, AnnLabName),
						% read_variable(recordIndent, RecIndent),
	findall(FieldId, member([field:_, FieldId | _], Fields), FieldIds),
	maxId(FieldIds, LongestId),
	lineListToTeX(Fields, FieldsTeX),
	invToTeX(Inv, InvTeX).

commaListToTeX(List, TeXList) :-
	listToTeX(List, comma, ',', TeXList).

lineListToTeX(List, TeXList) :-
	listToTeX(List, any, '', TeXList).

anyTypeToTeX([any:_], []).
anyTypeToTeX(Type, [': ', TeXType]) :-
	toTeX(Type, TeXType).

bindToTeX([Expr | PatternList], Sep, [ TeXPatternList, Sep, TeXExpr]) :-
	commaListToTeX(PatternList, TeXPatternList),
	toTeX(Expr, TeXExpr).

comprehensionToTeX(Compreh, Expr, Bindings,
	[Compreh, '{', TeXExpr, '}', LBr1, '{', TeXBindings, '}', TeXFilter]
    ) :-
	toTeX(Expr, TeXExpr),
	lineBreak(Expr, Bindings, LBr1),
	selBindListAndFilter(Bindings, BindList, Filter),
	commaListToTeX(BindList, TeXBindings),
	filterToTeX(Filter, Expr, TeXFilter).

selBindListAndFilter([bindings:_, Bind, Expr], [Bind | Rest], Filter) :-
	selBindListAndFilter(Expr, Rest, Filter).

selBindListAndFilter(Filter, [], Filter).

filterToTeX([true:_], _, '').

filterToTeX(Filter, Expr0, [LBr2, '[', TeXFilter, ']']) :-
	lineBreak(Expr0, Filter, LBr2),
	toTeX(Filter, TeXFilter).

elseToTeX(Expr, TeXExpr) :-
	Expr = [elseif:_ | _],
	toTeX(Expr, TeXExpr).

elseToTeX(Expr, ['\\Else\\\\ \\mbox{}\\\\', '\n', TeXExpr]) :-
	toTeX(Expr, TeXExpr).

fctDefToTeX([fctdef:_, Params, Rest], [TeXParams, TeXRest]) :-
	toTeX(Params, TeXParams),
	fctDefRestToTeX(Rest, TeXRest).

fctDefRestToTeX(Rest, TeXRest) :-
	Rest = [fctdef:_ | _],
	fctDefToTeX(Rest, TeXRest).

fctDefRestToTeX(Rest, ['}', '\n', TeXRest, '\n']) :-
	toTeX(Rest, TeXRest).

fctTypeToTeX([generictype:_, TypeVars, Type],
	[TeXTypeVars, TeXType]
    ) :-
	toTeX(TypeVars, TeXTypeVars),
	fctTypeToTeX(Type, TeXType).

fctTypeToTeX(FctType, ['{', TeXFctType, '}']) :-
	toTeX(FctType, TeXFctType).

invToTeX([true:_], ['% no inv', '\n']).
invToTeX([inv:_, Pattern, Expr], TeXInv ) :-
	invOrInitToTeX(invfn, Pattern, Expr, TeXInv).

initToTeX([true:_], ['% no init', '\n']).
initToTeX([init:_, Pattern, Expr], TeXInit) :-
	invOrInitToTeX(initfn, Pattern, Expr, TeXInit).

invOrInitToTeX(Env, Pattern, Expr,
	['\\begin{', Env, '}{', TeXPattern, '}', '\\\\', '\n',
	TeXExpr, '\n',
	'\\end{', Env, '}', '\n']
    ) :-
	toTeX(Pattern, TeXPattern),
	toTeX(Expr, TeXExpr).

letFilterToTeX([true:_], _, []).
letFilterToTeX(Filter, Bind, [LB, '[', TeXFilter, ']']) :-
	toTeX(Filter, TeXFilter),
	lineBreak(Filter, Bind, LB).

preOrPostToTeX([true:_], ['% no pre or post', '\n']).
preOrPostToTeX(Pre, TeXPre) :-
	toTeX(Pre, TeXPre).

localDefToTeX(Env, [LocalDefs, Expr],
	['\\begin{', Env, '}', '\n',
	TeXLocalDefs,
	'\\end{', Env, '}\\\\', '\n',
	TeXExpr]
    ) :-
	toTeX(LocalDefs, TeXLocalDefs),
	toTeX(Expr, TeXExpr).

texLetOp(letexpr, letexpr).
texLetOp(defexpr, defexpr).

% ------------------------------------------------------------------------------

% default style files, defined during loading

:- set_variable(texDocument,true).
:- set_variable(texStyles,'a4,umlaut,vdmsl,locvdmsl').

read_style_files(Styles) :-
	read_variable(texStyles,Styles).

toTeXDoc(Spec,
	['\\documentstyle[', Styles, ']{article}','\n',
         '\\begin{document}','\n',
	 TeX,
	 '\\end{document}','\n']
     ) :-
	read_variable(texDocument,true),
	read_style_files(Styles),
	toTeX(Spec, TeX).

toTeXDoc(Spec, TeX) :-
	toTeX(Spec, TeX).

outTeX(Spec,TeXFilename) :-
	once(toTeXDoc(Spec, TeX)),
	( read_variable(writeTeX, true) ->
	    once(writeTeX(TeX))
	; true
        ),
	tell(TeXFilename),
	once(writeTeX(TeX)),
	told.

writeTeX([]).

writeTeX([H | R]) :-
	writeTeX(H),
	writeTeX(R).

writeTeX('\\n') :-
	nl.

writeTeX(X) :-
	write(X).

% ------------------------------------------------------------------------------
						% longest name in a list
maxId(IdList, Id) :-
	maxId1(IdList, Id1),
	idToTeX(Id1, Id2),
						% 2 extra mm's for making names a bit longer
	name(Id2, L2),
	append("mm", L2, L3),
	name(Id, L3).

maxId1([],'M').
maxId1([Id], Id).
maxId1([Id1 | Rest], Id) :-
	maxId1(Rest, Id2),
	maxId1(Id1,Id2,Id).

maxId1(Id1, Id2, Id) :-
	name(Id1, L1),
	name(Id2, L2),
	length(L1, Len1),
	length(L2, Len2),
	(
	    Len1 > Len2 -> Id = Id1
	;
	    Id = Id2
	).

% ------------------------------------------------------------------------------

						% line breaks in lists
%lbr(Line1, Line2, LBr) :-
%	Line1 > Line2,
%	blr(Line2, Line1, LBr).

lbr0(Line1, Line2, ['\\\\', '\n']) :-
	Line1 < Line2 - 1.

lbr0(_, _, '').

lbr(Line1, Line2, ['\\\\', '\n']) :-
	Line2 \= Line1.

lbr(_, _, '').

lbr(Line, Line, '', '').

lbr(Line1, Line2, Pre, Post) :-
	Line1 > Line2,
	lpr(Line2, Line1, Pre, Post).

lbr(Line1, Line2, '', ['\\\\', '\n']) :-
	Line2 is Line1 + 1.

lbr(Line1, Line2, ['\\\\', '\n'], ['\\\\', '\n']) :-
	Line2 > Line1 + 1.
	
lineBreakInList([_:Line1/_/_ | _], [_:Line2/_/_ | _], Op, Pre, Post) :-
	lineBreakOp(Op),
	lbr(Line1, Line2, Pre, Post).

lineBreakInList(_, _, _, '', '').

lineBreakInBinexpr([_:Line1/_/_ | _], [_:Line2/_/_ | _], Op, Line/_/_, Pre, Post) :-
	lineBreakOp(Op),
	lbr0(Line1, Line, Pre),
	lbr(Line, Line2, Post).
						% default rule
lineBreakInBinexpr(_, _, _, _, '', '').


lineBreak([_:Line1/_/_ | _], [_:Line2/_/_ | _], ['\\\\', '\n']) :-
	Line1 \= 0,
	Line2 \= 0,
	Line1 \= Line2.
						% default rule
lineBreak(_, _, '').

lineBreakOp(altpatternlist).
lineBreakOp(mapenum).
lineBreakOp(params).
lineBreakOp(paramtypes).
lineBreakOp(producttype).
lineBreakOp(seqenum).
lineBreakOp(seqenumpattern).
lineBreakOp(setenum).
lineBreakOp(setenumpattern).
lineBreakOp(typevars).
lineBreakOp(uniontype).

lineBreakOp('->').
lineBreakOp('+>').
lineBreakOp('and').
lineBreakOp('or').
lineBreakOp('=>').
lineBreakOp('<=>').
lineBreakOp('=').
lineBreakOp('<>').
lineBreakOp('+').
lineBreakOp('-').
lineBreakOp('*').
lineBreakOp('/').
lineBreakOp(div).
lineBreakOp(mod).
lineBreakOp(rem).
lineBreakOp('^').
lineBreakOp(union).
lineBreakOp(inter).
lineBreakOp('\\').
lineBreakOp(munion).
lineBreakOp('<:').
lineBreakOp('<-:').
lineBreakOp('++').
lineBreakOp('|->').
lineBreakOp(comma).

%lineBreakOp(forall).
%lineBreakOp(exists).
%lineBreakOp(exists1).
%linebreakOp(iota).

% ------------------------------------------------------------------------------

% annotations, that are not formated as annotations but as plain LaTeX commands
% outside any vdm annotations environment

annotationsToTeX(Text, TeXText) :-
	name(Text, List),nl,
	annotationListToTeX(List, TeXText).

annotationListToTeX(Text,
	['\\end{vdm}', '\n',
	 TeXText,
	 '\\begin{vdm}', '\n']
     ) :-
	append(" TeX:", Text1, Text),
	annotationTextToTeX(Text1,TeXText).

% normal vdm annotations
annotationListToTeX(Text,
	['\\begin{annotations}{}', '\n',
	 '\\ann ', TeXText,
         '\\end{annotations}', '\n']
     ) :-
	annotationTextToTeX(Text, TeXText).

annotationTextToTeX(List, TeXText) :-
	annoToTeX(List, List1),
	name(TeXText, List1).

annoToTeX([], []).
annoToTeX([H | R], L) :-
	charToTeX(H, TrH0),
	name(TrH0, TrH),
	annoToTeX(R, L1),
	append(TrH, L1, L).

annoToTeX([H | R], [H | TrR]) :-
	annoToTeX(R, TrR).

charToTeX(0', '\\AUmlaut{}').
charToTeX(0', '\\OUmlaut{}').
charToTeX(0', '\\UUmlaut{}').
charToTeX(0', '\\aUmlaut{}').
charToTeX(0', '\\oUmlaut{}').
charToTeX(0', '\\uUmlaut{}').
charToTeX(0', '\\EsZett{}').
charToTeX(0'", '\\symbol(127)').

isNumberSuffix([C | R]) :-
	isDigit(C),
	( R = [] ->
	    true
        ; isNumberSuffix(R)
        ).
trNumberSuffix(L, R) :-
	append(L, "}", R1),
	append("\\SubscR{", R1, R).


textLitToTeX(Lit, TeXLit) :-
	name(Lit, List),
	once(textToTeX(List, List1)),
	name(TeXLit, List1).
						% not yet complete
textToTeX([], []).

textToTeX([0'_ | Rest], [ 0'\, 0'_ | Rest1]) :-
	textToTeX(Rest, Rest1).

textToTeX([0'$ | Rest], [ 0'\, 0'$ | Rest1]) :-
	textToTeX(Rest, Rest1).

textToTeX([0'& | Rest], [ 0'\, 0'& | Rest1]) :-
	textToTeX(Rest, Rest1).

textToTeX([X | Rest], Result) :-
	charToTeX(X, Atom),
	name(Atom, Rest0),
	textToTeX(Rest, Rest1),
	append(Rest0, Rest1, Result).

textToTeX([H | R], [H | R1]) :-
	textToTeX(R, R1).
