
% type definition syntax -----------------------------------------------------------------

type_definitions([types:Pos|Types]) -->
	  [types:Pos], type_definition_list(Types).

type_definition_list(Types) -->
	  type_definition(Type),
		type_definition_rest(Types1),
		{append(Type,Types1,Types)}.

type_definition_rest(Types) -->
	  [';':_],
		type_definition_list(Types)
	| skip_until('syntax error in type definition', type_definition_rest),
		type_definition_rest(Types)
	| [],
		{Types = []}
	.

type_definition([ [typedef:Pos ,TypeName, TypeExpr, Inv] ]) -->
	[ identifier(TypeName):Pos ],
	right_side(TypeName,TypeExpr),
	invariant(Inv),
	{ ! }
	.

type_definition([]) -->
	syntax_error('illegal syntax in type definition', type_definition).

			% TypeName is inherited
			% TypeExpr is derived
		
right_side(_TypeName, TypeExpr) -->
	['=':_],
	{!},
	a_type(TypeExpr,type).

right_side(TypeName, [record:RecPos, TypeName | Fields]) -->
	['::':RecPos],
	{!},
	field_list(Fields).

right_side(_,_) -->
	[X:Pos],
		{message(Pos,['"=" or "::" expected instead of "',X,'"']),fail}
	.

invariant(Expr) -->
	inv_or_init(inv, Expr).

init(Expr) -->
	inv_or_init(init, Expr).

inv_or_init(Op, [Op:Pos, Id, Expr]) -->
	[ Op:Pos ],
	{ ! },
	a_pattern(Id),
	test_symbol('=='),
	an_expr(Expr).

inv_or_init(_, X) -->
	truesy(X).

			% a_type succeeds always, even if type is not syntactically 
			% correct, type may fail without any error message
a_type(ResultType) -->
	a_type(ResultType, type).

a_type(ResultType,_) -->
	type(ResultType),
	{!}.

a_type(X, Nonterminal) -->
	syntax_error('syntax error in type expression', Nonterminal, X).

a_type0(ResultType) -->
	type0(ResultType),
	{ ! }.
a_type0(X) -->
	syntax_error('syntax error in product type', type0, X).

a_type2(ResultType) -->
	type2(ResultType),
	{ ! }.

a_type2(X) -->
	syntax_error('syntax error in function type', type2, X).

a_function_type(Type) -->
	a_type2(Type1),
	a_fct_op(Op, Pos),
	function_type(Op, Pos, Type1, Type).

a_fct_op(Op, Pos) -->
	[ Op:Pos ],
	{ function_type_op(Op), !}.

a_fct_op('->', Pos) -->
	test_sympos('->':Pos).

type(ResultType) -->
	type2(TypeExpr1),
	function_list(TypeExpr1,ResultType).

function_list(TypeExpr1,ResultType) -->
	[ Op:Pos ],
	{ function_type_op(Op), !},
	function_type(Op, Pos, TypeExpr1, ResultType).

function_list(TypeExpr,TypeExpr) -->
	[].

function_type(Op, Pos, TypeExpr1, [Op:Pos, TypeExpr1, TypeExpr2]) -->
	a_type(TypeExpr2).
						% { trans_paramlist(TypeExpr1,TypeExpr11) }.

type2([producttype0:Pos]) -->
	['(':Pos, ')':_],			% empty paramlist
	{ ! }.

type2(ResultType) -->
	type1(TypeExpr1),
	union_list(TypeExpr1,ResultType).

union_list(TypeExpr1, ResultType) -->
	[ '|':Pos ],
	{ ! },
	type1(TypeExpr2),
	union_list([uniontype:Pos, TypeExpr1, TypeExpr2], ResultType0),
	{ trans_assoc(ResultType0,ResultType) }.

union_list(TypeExpr1, TypeExpr1) -->
	[].
	
type1(ResultType) -->
	type0(TypeExpr1),
	prod_list(TypeExpr1,ResultType).

prod_list(TypeExpr1, ResultType) -->
	[ '*':Pos ],
	a_type0(TypeExpr2),
	prod_list([producttype:Pos, TypeExpr1, TypeExpr2], ResultType0),
	{ trans_assoc(ResultType0, ResultType) }.

prod_list(TypeExpr1, TypeExpr1) -->  
	[].

type0(ResultType) -->
	  basic_type(ResultType).

type0(['type identifier':Pos, TypeId]) -->
	[ identifier(TypeId):Pos ].

type0(['quote literal':Pos, Id]) -->
	[ 'quote literal'(Id):Pos ].

type0([typevar:Pos, TypeId]) -->
	[ '@':Pos ],
	an_ident(TypeId).

type0([parenttype:Pos, Type]) -->
	[ '(':Pos ],
	a_type(Type,type_par),
	test_symbol(')').

type0([record:Pos, TypeId | FieldList]) -->
	[ compose:Pos, identifier(TypeId):_ ],
	test_symbol(of),
	field_list(FieldList),
	test_symbol(end).

type0([optionaltype:Pos, TypeExpr]) -->
	[ '[':Pos ],
	a_type(TypeExpr, type_bra),
	test_symbol(']').

type0([X, ElementType]) -->
	[X],
	{ X = Op:_, collection1_type(Op), ! },
	test_symbol(of),
	type0(ElementType).

type0([X, DomType, RngType]) -->
	[X],
	{ X = Op:_, collection2_type(Op), ! },
	a_type(DomType,type_to),
	test_symbol(to),
	type0(RngType).

basic_type([X]) -->
	[ X ],
	{ X = Op:_, simple_type(Op), ! }.

field_list([Field1 | Fields2]) -->
	field(Field1),
	{ ! },
	field_list(Fields2).

field_list([]) -->
	[].

field([field:FieldPos, Fieldid, TypeExpr]) -->
	[identifier(Fieldid):FieldPos, (':'):_],
	type(TypeExpr).

field([field:Pos, '_', TypeExpr]) -->
	type(TypeExpr),
	{ TypeExpr = [_:Pos | _] }.

%-----------------------------------------------------------------------------------------

a_type_list([Type | Rest]) -->
	a_type(Type),
	a_type_list_rest(Rest).

a_type_list_rest(TypeList) -->
	[ ',':_ ],
	a_type_list(TypeList).

a_type_list_rest([]) -->
	[].

%-----------------------------------------------------------------------------------------

type_op('->',         100).
type_op('+>',         100).
type_op(map,          100).
type_op(inmap,        100).
type_op(uniontype,    110).
type_op(producttype,  120).
type_op(set,          130).
type_op(seq,          130).
type_op(seq1,         130).

%-----------------------------------------------------------------------------------------