#include "semaux.hpp"

bool TRUEvalue = bool(1);

void check_types(void)
{
  TC.proof_ids();   /* Alle Typen deklariert ? */
  TC.make_flat();   /* Unions flachmachen */
  TC.proof_base();  /* Rekursionen ueberpruefen */
}

void eval_union(t_typ expression,
		const t_typ & filter,
		nat0seq & anzahl,
		bool & warning)
{
  t_typ kind = TC.get_type(expression);
  if (VDM_TRACE)
    cout << "\nin eval_union mit: " << (char*)TC.get_typename(expression);

  while((kind == T_ID) || (kind == T_FIELD))
    {
      expression = TC.get_subtype(expression,1);  // Typ darf dafr nicht UNKNOWN sein
      kind = TC.get_type(expression);
    }

  if(!TC.is_marked_union(expression))  // Ende-Bedingung fuer Rekursions-Abstieg
    {
      if(kind == T_UNION)
	{
	  TC.mark_union(expression);
	  for(nat0 i = TC.len_subtypes(expression); i > 0; i--)
	    {
	      eval_union(TC.get_subtype(expression, i), filter, anzahl, warning);
	    }
	  TC.unmark_union(expression);
	}
      else
	{
	  if (VDM_TRACE)
	    cout << "\nin eval_union-Auswertung mit: " << (char*)TC.get_typename(expression);

	  if(kind != filter)
	    {
	      if(kind == T_VAR)
		{
		  t_typ new_typ = filter;       // neuen Typ fuer T_VAR erstellen
		  if(filter >= OWN_TYPES)
		    {
		      nat0seq tmp;
		      tmp.add(T_VAR);
		      if ( (filter == T_MAP)
			  || (filter == T_INMAP)
			  || (filter == T_FUNC)
			  )
			tmp.add(T_VAR);
		      new_typ = TC.add_type(tmp, filter, filter, (node*)0, KNOWN);
		    }
		  anzahl.add(new_typ);
		}
	      else if ( IS_NUM(filter)
		       && IS_NUM(kind)
		       )
		{
		  if ( (filter != T_REAL)
		      && !IS_IN_NUM(kind, filter)
		      )
		    {
		      /*warning = bool(1);*/
		      expression = TC.add_type(nat0seq(), filter, filter, (node*)0, KNOWN);
		    }
		  anzahl.add(expression);
		}
	      else if ( (filter == T_SET)
		       && (kind == T_EMPTYSET)
		       )
		anzahl.add(expression);
	      else if( ((filter == T_SEQ) && (kind == T_SEQ1))
		      ||
		      ((filter == T_SEQ1) && (kind == T_SEQ))
		      ||
		      ((filter == T_SEQ) && (kind == T_EMPTYSEQ))
		      )
		anzahl.add(expression);
	      else if (((filter == T_MAP) && (kind == T_INMAP))
		       ||
		       ((filter == T_MAP) && (kind == T_EMPTYMAP))
		       ||
		       ((filter == T_INMAP) && (kind == T_EMPTYMAP))
		       )
		anzahl.add(expression);
	      else if(filter == UNKNOWN) // mache Unions flach
		anzahl.add(expression);
	      else
		warning = bool(1);
	    }
	  else
	    {
	      //if(!anzahl.isin(expression))
	      anzahl.add(expression);
	    }
	}
    }
}

bool unary_ops(const t_typ & expression,
	       const t_typ & filter,
	       const String & error_message,
	       const nat0 & linenr,
	       nat0seq & anzahl,
	       const bool & output,
	       bool & warn)
{
  if (VDM_TRACE)
    cout << "\nin unary_ops mit: " << (char*)TC.get_typename(expression);

  t_typ kind = TC.get_type(expression);
  t_typ exp_z = expression;

  bool warning = bool(0);
  bool erg = bool(1);
  eval_union(expression, filter, anzahl, warning);

  if (VDM_TRACE)
    cout << "\nin unary_ops nach eval; anzahl: " << anzahl.len() << "; warning: " << warning;

  if(anzahl.len() == 0)
    {
      while((kind == T_ID) || (kind == T_FIELD))
	{
	  exp_z = TC.get_subtype(exp_z,1);  // Typ darf dafr nicht UNKNOWN sein
	  kind = TC.get_type(exp_z);
	}
      if (output)
	VDM_ERROR(36, linenr, error_message, LEXEMS.get(filter), LEXEMS.get(kind));
      else erg = bool(0);
    }

  if(warning)
    if(warn)
      VDM_WARNING(1, linenr, error_message);
    else
      warn = bool(1); // Warn-Ereignis zurueckgeben
  return erg;
}

void make_layer(void)
{   // noch regeln, wie Stack bei Reparsen behandelt werden soll
  if (VDM_TRACE) cout << "\n\n~~~~~~~~~~~~~~~~~~~~~~~~~\nneuer Level\n\n";
  TC.open_layer();
}

void kill_layer(bool nur_id)
{   // noch regeln, wie Stack bei Reparsen behandelt werden soll
  TC.close_layer(nur_id);
}

bool eval_equal(t_typ t1,
		t_typ t2,
		warn_typ & warning)
{
  if (VDM_TRACE)
    cout << "eval_equal mit t1 = " << t1 << " und t2 = " << t2;
  if (t1 == t2)
    return bool(1);   // gleicher Typschluessel: selber Typ

  nat0 i,j;
  t_typ kind1 = TC.get_type(t1);
  t_typ kind2 = TC.get_type(t2);

  if (VDM_TRACE)
    cout << "\nin eval_equal mit: "
      << (char*)TC.get_typename(t1)
	<< " und: " << (char*)TC.get_typename(t2);

  if(kind1 == DONTCARE)
    return bool(1);               // kein Binding, Typ2 matcht immer

  if(kind1 == VAR)               // 4.2.1.5.
    {
      if (VDM_TRACE)
	cout << "\nvar+++++++++++++++";

      //TC.replace_subtypes(t1, TC.get_subtypes(t2));
      //TC.put_type(t1, TC.get_type(t2));

      TC.put_state(t1, KNOWN);
      TC.add_subtype(t1, t2);
      TC.put_type(t1, T_ID);        // Variablen fuer Typcheck wie Typnamen

      return bool(1);
    }

  if(kind1 == T_VARSET)          // Typvar. einer polymorphen Funktion
    {                              // wurde wie Set-Typ benutzt:
      nat0seq tmp;                  // alle Sets aus Typ2 suchen
      unary_ops(t2, T_SET, "set-parameter", TC.get_adress(t1)->linenr, tmp);
      if(tmp.len() == 1)            // wird nur ein Set-Typ gefunden:
	{                             // Binden dieses Typs
	  TC.replace_subtypes(t1, TC.get_subtypes(tmp.get(1)));
	  TC.put_type(t1, T_SET);
	}
      else                          // sonst Union aller gefundenen Set-Typen
	{
	  TC.put_type(t1, T_UNION);
	  TC.replace_subtypes(t1, tmp);
	}

      return bool(1);
    }
  
  if(kind1 == T_VARSEQ)          // s.o. fr Seq-Typvariablen
    {
      nat0seq tmp;
      unary_ops(t2, T_SEQ, "seq-parameter", TC.get_adress(t1)->linenr, tmp);

      if(tmp.len() == 1)
	{
	  TC.replace_subtypes(t1, TC.get_subtypes(tmp.get(1)));
	  TC.put_type(t1, T_SEQ);
	}
      else
	{
	  TC.put_type(t1, T_UNION);
	  TC.replace_subtypes(t1, tmp);
	}

      return bool(1);
    }
  
  if ((kind1 == T_VAR)
      ||                          // Typvar. beliebigen Typs
      (kind2 == T_VAR))           // zunaechst wird alles gematcht, erst bei
    {                             // Instantiation und konkreter Typbelegung
      /* if(kind1 == T_VAR)
	 {
	 if(TC.len_subtypes(t1) == 0)
	 TC.add_subtype(t1, t2);
	 else
	 {
	 if(TC.get_type(TC.get_subtype(t1, 1)) != T_UNION)
	 {
	 if(t2 != TC.get_subtype(t1, 1))
	 {
	 nat0seq tmp;
	 tmp.add(TC.get_subtype(t1, 1));
	 tmp.add(t2);
	 TC.put_subtype(t1, 1,
	 (TC.add_type(tmp, T_UNION, "UNION", adr, KNOWN));
	 }
	 }
	 else
	 {
	 if(!(TC.get_subtypes(TC.get_subtype(t1, 1)).isin(t2)))
	 TC.add_subtype(TC.get_subtype(t1, 1), t2);
	 }
	 }
	 }
	 if(kind2 == T_VAR)
	 {
	 if(TC.len_subtypes(t2) == 0)
	 TC.add_subtype(t2, t1);
	 else
	 {
	 if(TC.get_type(TC.get_subtype(t2, 1)) != T_UNION)
	 {
	 if(t1 != TC.get_subtype(t2, 1))
	 {
	 nat0seq tmp;
	 tmp.add(TC.get_subtype(t2, 1));
	 tmp.add(t1);
	 TC.put_subtype(t2, 1,
	 (TC.add_type(tmp, T_UNION, "UNION", adr, KNOWN));
	 }
	 }
	 else
	 {
	 if(!(TC.get_subtypes(TC.get_subtype(t2, 1)).isin(t1)))
	 TC.add_subtype(TC.get_subtype(t2, 1), t1);
	 }
	 }
	 }*/

      return bool(1);               // wird der Check ausgefuehrt
    }

  // Typids uebergehen:
  while( (kind1 == T_ID)
	|| (kind1 == T_FIELD))
    {
      t1 = TC.get_subtype(t1,1);    // Typ darf dafr nicht UNKNOWN sein
      kind1 = TC.get_type(t1);
    }

  while((kind2 == T_ID)
	|| (kind2 == T_FIELD))
    {
      t2 = TC.get_subtype(t2,1);    // Typ darf dafr nicht UNKNOWN sein
      kind2 = TC.get_type(t2);
    }

  // Union-Behandlung, s. 4.2.1.5
  if(!TC.is_marked_union(t1))  // Ende-Bedingung fuer Rekursions-Abstieg
    {                            // wenn t2 bis dahin gleich, dann true
      bool ok = bool(0);

      // entspricht Typ_Matchall1/2(), s. 4.2.1
      if ((kind1 == T_UNION)
	  &&
	  (kind2 == T_UNION))
	{
	  for(i = TC.len_subtypes(t1); i > 0; i--)
	    for(j = TC.len_subtypes(t2); j > 0; j--)
	      {
		if(eval_equal(TC.get_subtype(t1, i), TC.get_subtype(t2, j), warning))
		  ok = bool(1);
		else warning = (warning == DIFFTYP?warning:INCOMP);
	      }
	  if((ok) && (warning == DIFFTYP)) // DIFF-Typ-Warnung nur bei einziger
	    warning = INCOMP;              // Loesung
	  return ok;
	}

      // entspricht Typ_Matchall1(), s. 4.2.1
      else if(kind1 == T_UNION)
	{
	  for(i = TC.len_subtypes(t1); i > 0; i--)
	    {
	      if(eval_equal(TC.get_subtype(t1, i), t2, warning))
		ok = bool(1);
	      else
		warning = (warning == DIFFTYP?warning:INCOMP);
	    }
	  if((ok) && (warning == DIFFTYP)) // DIFF-Typ-Warnung nur bei einziger
	    warning = INCOMP;              // Loesung

	  return ok;
	}

      // entspricht Typ_Matchall2(), s. 4.2.1
      else if (kind2 == T_UNION)
	{
	  for(i = TC.len_subtypes(t2); i > 0; i--)
	    {
	      if(eval_equal(t1, TC.get_subtype(t2, i), warning))
		ok = bool(1);
	      else
		warning = (warning == DIFFTYP?warning:INCOMP);
	    }
	  if ((ok) && (warning == DIFFTYP)) // DIFF-Typ-Warnung nur bei einziger
	    warning = INCOMP;              // Loesung
	  return ok;
	}

      else    // keine Unions mehr, Auswertung beginnt
	{
	  if (VDM_TRACE)
	    cout << "\nin eval_equal-Auswertung mit: "
	      << (char*)TC.get_typename(t1)
		<<  " und: " 
		  << (char*)TC.get_typename(t2);

	  if(kind1 != kind2)               // Behandlung der Faelle 4.2.1.6.
	    {
	      if(IS_NUM(kind1) && IS_NUM(kind2))
		{
		  /*if((kind2 != T_REAL) && !IS_IN_NUM(kind1, kind2))
		    warning = (warning == DIFFTYP?warning:INCOMP);*/
		  return bool(1);
		}
	      else if(((kind1 == T_SET) && (kind2 == T_EMPTYSET))
		      ||
		      ((kind1 == T_EMPTYSET) && (kind2 == T_SET)))
		{ /* weiter mit subtypes*/
		}
	      else if(((kind1 == T_SEQ) &&((kind2 == T_SEQ1)
					   ||
					   (kind2 == T_EMPTYSEQ)
					   )
		       )
		      ||
		      ((kind1 == T_SEQ1) && (kind2 == T_SEQ))
		      ||
		      ((kind1 == T_EMPTYSEQ) && (kind2 == T_SEQ)))
		{ /* weiter mit subtypes*/
		}
	      else if (((kind1 == T_MAP) && ((kind2 == T_INMAP)
					     ||
					     (kind2 == T_EMPTYMAP)))
		       ||
		       ((kind1 == T_INMAP) && ((kind2 == T_MAP)
					       ||
					       (kind2 == T_EMPTYMAP)))
		       ||
		       ((kind1 == T_EMPTYMAP) && ((kind2 == T_MAP)
						  ||
						  (kind2 == T_INMAP))))
		{ /* weiter mit subtypes*/
		}
	      // UNKNOWN und undefined-Ausdruck matchen alles
	      else if((kind1 == UNKNOWN) || (kind2 == UNKNOWN))
		return bool(1);
	      else if((kind1 == T_UNDEFINED) || (kind2 == T_UNDEFINED))
		return bool(1);
	      else
		return bool(0);    // keine anderen Gruende fuer versch. Typart
	    }

	  // Konstruktoren matchen, weitere Prfungen:
	  // Composenamen berprfen
	  if ((kind1 == T_COMP)
	      &&
	      (TC.get_typename(t1) != TC.get_typename(t2))
	      )
	    return bool(0);

	  bool multi_pat = bool(0);
	  // Multi-Pattern von Aufzaehlungen (z.B. {1, 2, 3}) finden
	  if ((TC.len_subtypes(t1) > 0)
	      &&
	      (TC.get_type(TC.get_subtype(t1, 1)) == MULTI_PAT)
	      )
	    {
	      t1 = TC.get_subtype(t1, 1);
	      kind1 = TC.get_type(t1);
	      multi_pat = bool(1);
	    }

	  // Komponentenanzahl muss uebereinstimmen
	  if((TC.len_subtypes(t1) != TC.len_subtypes(t2)) && !multi_pat)
	    return bool(0);

	  // entspricht Typ_Matchall1und2() und Pattern_Match(), s. 4.2.1
	  TC.mark_union(t1);

	  if (VDM_TRACE && multi_pat)
	    cout << "\nMULTIPAT: ";

	  for(i = TC.len_subtypes(t1); i > 0; i--)
	    {
	      if(VDM_TRACE && multi_pat)
		cout << "\n" << i << ". Typname: " << (char*)TC.get_typename(TC.get_subtype(t1, i));

	      if(!eval_equal(TC.get_subtype(t1, i),
			     TC.get_subtype(t2, (multi_pat?1:i)), warning))
		{// bei multipatt mit [] und seq1 als t2 Schwierigkeiten, s. input
		  TC.unmark_union(t1);

		  if (VDM_TRACE && multi_pat)
		    cout << " Erg. ist false";

		  if (MATH
		      && 
		      ((kind1 == T_SET)
		       ||
		       (kind1 == T_SEQ)
		       ||
		       (kind1 == T_SEQ1)
		       ||
		       (kind1 == T_MAP)
		       ||
		       (kind1 == T_INMAP)
		       )
		      )
		    {
		      warning = DIFFTYP;
		      return bool(1);
		    }
		  else
		    {
		      if (VDM_TRACE)
			cout << "\ntats_chlich";
		      return bool(0);
		    }
		}
	      if(VDM_TRACE && multi_pat)
		cout << " Erg. ist true";

	    }

	  TC.unmark_union(t1);

	  if (VDM_TRACE)
	    cout << "\nalles ok";

	  return bool(1);
	}
    }
  else
    {
      if (VDM_TRACE)
	cout << "\nmarkiert*********************";

      return bool(1);
    }
}

nat0 make_text(const nodeseq & terms,
	       node * adr,
	       const parse_typ & t_akt)
{
  if (t_akt == UNKNOWN)
    {
      nat0seq tmp(nat0(1), nat0(0));
      tmp.add(T_CHAR);
      return TC.add_value(tmp, T_SEQ, T_SEQ, adr, CONST);
    }
  else return t_akt;
}

nat0 make_quote(const nodeseq & terms,
		node * adr,
		const parse_typ & t_akt)
{
  nat0 quote_lit = GET(1);
  t_typ typ = TC.seek_type(LEXEMS.get(quote_lit));
  if(typ == UNKNOWN) // noch nicht deklariert
    {
      VDM_WARNING(9, adr->linenr, LEXEMS.get(quote_lit));
      nat0seq tmp(nat0(1), nat0(0));
      tmp.add(T_BASIC);

      return TC.add_type(tmp, T_QUOTE, quote_lit, adr, KNOWN);
    }

  return typ;
}
