#include "semfunc.hpp"

nat0 explicit_function_definition(const nodeseq & terms, node * adr, const parse_typ & t_akt)
{ // zweiter Teil der expliziten Funktion, s. 4.3
 nat0 function_begin      = GET(1);
 nat0 expression          = GET(2);
 nat0 pre_expression_opt  = GET(3);
 nat0 post_expression_opt = GET(4);

 // Return-Typ isolieren:
 t_typ rt = TC.get_subtype(function_begin, 1); // Funktions-Typ
 t_typ rk = TC.get_type(rt);
 t_typ params;
 if(rk == T_POLY)   // bei polym. Funk. steht Instantiation am Ende
 {
  params = TC.get_subtype(TC.get_subtype(rt, TC.len_subtypes(rt)), 1);
  rt = TC.get_subtype(TC.get_subtype(rt, TC.len_subtypes(rt)), 2);
 }
 else               // bei norm. Funk. ist es hier der zweite Subtyp
 {
  params = TC.get_subtype(rt, 1);
  rt = TC.get_subtype(rt, 2);
 }
 rk = TC.get_type(rt);

 if (PARSE == FIRST) // beim 1. Lauf pre- u. post-Typen bilden
 {
  if(pre_expression_opt != NOPAT)
  {
   nat0seq tmp;
   tmp.add(params);
   tmp.add(T_BOOL);
   TC.add_type(tmp,
	       T_FUNC,
	       LEXEMS.add("pre_" + TC.get_typename(function_begin)),
	       adr, KNOWN);
  }
  if(post_expression_opt != NOPAT)
  {
   nat0seq tmp;
   tmp.add(params);
   tmp.add(T_BOOL);
   TC.add_type(tmp,
	       T_FUNC,
	       LEXEMS.add("post_" + TC.get_typename(function_begin)),
	       adr, KNOWN);
  }
  return UNKNOWN;
 }
 else
 {
  // s. 4.3.5
  warn_typ warn = NOWARN;
  bool oldmath = MATH;
  MATH = bool(0);
  TC.output();
  if(expression != NOPAT)    // nicht is_not_yet_defined
  {
   if(!eval_equal(expression, rt, warn))
    VDM_ERROR(18, adr->linenr, TC.get_typename(function_begin));
   if (warn == INCOMP)
    VDM_WARNING(1, adr->linenr, "return-type of function " +
		TC.get_typename(function_begin));
  }

  // s. 4.3.6
  if((pre_expression_opt != NOPAT) &&
     (TC.get_type(pre_expression_opt) != T_BOOL))
    VDM_ERROR(19, TC.get_adress(pre_expression_opt)->linenr, TC.get_typename(function_begin));
  if((post_expression_opt != NOPAT) &&
     (TC.get_type(post_expression_opt) != T_BOOL))
   VDM_ERROR(20, TC.get_adress(post_expression_opt)->linenr, TC.get_typename(function_begin));

  // Nachbearbeitung und Polymorph-Behandlung
  MATH = oldmath;
  POLY = ""; // Ruecksetzen der Polymorph-Typvar-Kennzeichnung
  if (VDM_TRACE) cout << "\n\nfunction-close\n\n";
  TC.close_layer(bool(0));
  TC.output();
  return function_begin;
 }
 return t_akt;
}

nat0 function_begin1(const nodeseq & terms, node * adr, const parse_typ & t_akt)
{ // erster Teil der expliziten (polymorphen) Funktion, s. 4.3
 nat0 type_variable_list = GET(1);
 nat0 function_type      = GET(2);
 nat0 identifier2        = GET(3);
 nat0 parameters         = GET(4);

 if (t_akt == NEW_LAYER_ACT)
 { // erster Durchgang: Funktionstyp eintragen
  if (POLY != LEXEMS.get(identifier2))                // s. 4.3.1
   VDM_ERROR(9, adr->linenr, POLY);
  if(TC.seek_type(POLY) != UNKNOWN)                   // s. 4.3.2
   VDM_ERROR(25, adr->linenr, POLY);
  nat0seq tmp(nat0(1), nat0(0));                      // Funk-Typ am Ende
  TC.add_subtype(type_variable_list, function_type);  // der Typ-Var-Liste
  tmp.add(type_variable_list);                        // s. 4.3.7
  return TC.add_type(tmp, T_ID, identifier2, adr, KNOWN);
 }

 else
 {// Zweiter Durchlauf: Pattern ueberpruefen mit den jetzt bekannten Typen
  //TC.output();
/*  t_typ ft = TC.get_subtype(function_type, 1);
  if (TC.len_subtypes(parameters) != TC.len_subtypes(ft))
   VDM_ERROR(10, POLY, POLY);
  for(nat0 i = 1; i <= TC.len_subtypes(parameters); i++)
  {
   match_parameters(TC.get_subtype(parameters, i),
		    TC.get_subtype(ft, i), bool(1));
  }
  match_parameters(parameters, TC.get_subtype(function_type, 1), bool(1));*/
  bool oldmath = MATH;
  MATH = bool(0);
  warn_typ warning = NOWARN;                          // s. 4.3.3
  if(!eval_equal(parameters, TC.get_subtype(function_type, 1), warning))
   VDM_ERROR(24, TC.get_adress(parameters)->linenr, "parameters of function " + POLY);
  if(warning)
   VDM_WARNING(1, TC.get_adress(parameters)->linenr, "parameters of function " + POLY);
  MATH = oldmath;
  return t_akt;
 }
}

nat0 function_begin2(const nodeseq & terms, node * adr, const parse_typ & t_akt)
{ // erster Teil der expliziten Funktion
 nat0 identifier1     = GET(1);
 nat0 function_type   = GET(2);
 nat0 identifier2     = GET(3);
 nat0 parameters      = GET(4);

 if (t_akt == NEW_LAYER_ACT)
 { // erster Durchgang: Funktions-Typ eintragen
  if (LEXEMS.get(identifier1) != LEXEMS.get(identifier2)) // s. 4.3.1
   VDM_ERROR(9, adr->linenr, LEXEMS.get(identifier2));
  if(TC.seek_type(LEXEMS.get(identifier1)) != UNKNOWN)    // s. 4.3.2
   VDM_ERROR(25, adr->linenr, LEXEMS.get(identifier1));
  nat0seq tmp(nat0(1), nat0(0));
  tmp.add(function_type);
  return TC.add_type(tmp, T_ID, identifier1, adr, KNOWN); // s. 4.3.7
 }

 else
 { // zweiter Durchgang: Pattern ueberpruefen mit den jetzt bekannten Typen
  //TC.output();
/*  t_typ ft = TC.get_subtype(function_type, 1);
  if (TC.len_subtypes(parameters) != TC.len_subtypes(ft))
   VDM_ERROR(10, LINENR.get(identifier2), LEXEMS.get(identifier2));
  for(nat0 i = 1; i <= TC.len_subtypes(parameters); i++)
  {
   match_parameters(TC.get_subtype(parameters, i),
		    TC.get_subtype(ft, i), bool(1));
  }
  match_parameters(parameters, TC.get_subtype(function_type, 1), bool(1));*/
  bool oldmath = MATH;
  MATH = bool(0);
  warn_typ warning = NOWARN;                 // s. 4.3.3
  if(!eval_equal(parameters, TC.get_subtype(function_type, 1), warning))
   VDM_ERROR(24, TC.get_adress(parameters)->linenr, "parameters of function " + LEXEMS.get(identifier1));
  if(warning)
   VDM_WARNING(1, TC.get_adress(parameters)->linenr, "parameters of function " + LEXEMS.get(identifier1));
  MATH = oldmath;
  return t_akt;
 }
}

nat0 impl_func(const nodeseq & terms,
	       node * adr, const parse_typ & t_akt)
{ // implizite Funktion
 nat0 identifier   = GET(1);
 nat0 param_types  = GET(2);
 nat0 id_type_pair = GET(3);
 nat0 pre_expr_opt = GET(4);
 nat0 post_expr    = GET(5);
 warn_typ warn = NOWARN;
 if(TC.seek_type(LEXEMS.get(identifier)) != UNKNOWN)    // s. 4.3.1
  VDM_ERROR(25, adr->linenr, LEXEMS.get(identifier));

 if(t_akt != NEW_LAYER_ACT)
 { // erst beim zweiten Durchgang expressions pruefen
  if(!eval_equal(post_expr, T_BOOL, warn))              // s. 4.3.6
   VDM_ERROR(24, TC.get_adress(post_expr)->linenr, "post-condition");
  if(warn == INCOMP)
   VDM_WARNING(1, TC.get_adress(post_expr)->linenr, "post-condition");
  if(TC.get_type(pre_expr_opt) != NOPAT)
  {
   warn = NOWARN;
   if(!eval_equal(pre_expr_opt, T_BOOL, warn))          // s. 4.3.6
    VDM_ERROR(24, TC.get_adress(pre_expr_opt)->linenr, "pre-condition");
   if(warn == INCOMP)
    VDM_WARNING(1, TC.get_adress(pre_expr_opt)->linenr, "pre-condition");
  }
 }

 if(TC.get_type(param_types) == T_PARAMS)
  TC.put_type(param_types, T_PROD); //normale Funktions-Behandlung
 nat0seq tmp(nat0(2), nat0(0));
 tmp.add(param_types);
 tmp.add(id_type_pair);             //ohne identifier uebergeben
 nat0seq tmp2(nat0(1), nat0(0));
 tmp2.add(TC.add_type(tmp, T_FUNC, T_FUNC, adr, KNOWN)); // Funk-Typ
 if (VDM_TRACE) cout << "\n\nimplicit-function-close\n\n";
 TC.close_layer(bool(0));
 return TC.add_type(tmp2, T_ID, identifier, adr, KNOWN); // s. 4.3.7
}

nat0 type_variable_list(const nodeseq & terms, node * adr, const parse_typ & t_akt)
{ // zum Erstellen der Typvar-Liste, falls noch nicht geschehen (s.u.)
 if (t_akt == UNKNOWN)
 {
  nat0 type_variable_list = GET(2);
  t_typ t = type_variable_list;
  if (TC.get_type(type_variable_list) == T_VAR)
  { // Liste der polymorphen Typen noch nicht angelegt
   nat0seq tmp;
   tmp.add(type_variable_list);
   t = TC.add_type(tmp, T_POLY, T_POLY, adr, KNOWN); // erste polym. Typ-Variable einspeichern
  }
  return t;
 }
 else return t_akt;
}

nat0 poly_identifier(const nodeseq & terms, node * adr, const parse_typ & t_akt)
{ // zum globalen Eintrag der gerade definierten poly. Funktion
  // (wichtig  fuer eindeutige Kennzeichnung der Typvar-Identifier)
 nat0 identifier = GET(1);
 POLY = LEXEMS.get(identifier);
 return identifier;
}

nat0 type_variable_identifier_list1(const nodeseq & terms, node * adr, const parse_typ & t_akt)
{ // Aufbau einer und Anhaengen an eine Typvar-Liste
 if (t_akt == UNKNOWN)
 {
  nat0 type_variable_identifier_list = GET(1);
  nat0 type_variable_identifier      = GET(2);
  t_typ t = type_variable_identifier_list;

  if (TC.get_type(type_variable_identifier_list) == T_VAR)
  { // Liste der polymorphen Typen noch nicht angelegt
   nat0seq tmp;
   tmp.add(type_variable_identifier_list);
   t = TC.add_type(tmp, T_POLY, T_POLY, adr, KNOWN); // erste polym. Typ-Variable einspeichern
  }

  // wenn typ-variable schon vorhanden -> Fehler
  for (nat0 i = 1; i <= TC.len_subtypes(t); i++)
   if (TC.get_typename(TC.get_subtype(t, i)) ==
       LEXEMS.get(type_variable_identifier))
    VDM_ERROR(7, adr->linenr,
		 LEXEMS.get(type_variable_identifier));

  // neue typ-var eintragen
//  nat0seq tmp(nat0(1), nat0(0));
//  tmp.add(T_VAR);
  TC.add_subtype(t,
     TC.add_type(nat0seq(), T_VAR,
	LEXEMS.add(POLY + "." + LEXEMS.get(type_variable_identifier)),
	adr, KNOWN));
  return t;
 }
 else return t_akt;
}

nat0 type_variable_identifier_list2(const nodeseq & terms, node * adr, const parse_typ & t_akt)
{ // erste Typvar eintragen
 if (t_akt == UNKNOWN)
 {
  nat0 type_variable_identifier = GET(1);
//  nat0seq tmp(nat0(1), nat0(0));         /*noch nicht deklariert*/
//  tmp.add(T_VAR);
  return TC.add_type(nat0seq(), T_VAR,
	    LEXEMS.add(POLY + "." + LEXEMS.get(type_variable_identifier)),
	    adr, KNOWN);
 }
 else return t_akt;
}

nat0 id_type_pair(const nodeseq & terms,
		  node * adr, const parse_typ & t_akt)
{
 nat0 identifier = GET(1);
 nat0 type       = GET(2);
 if(t_akt != UNKNOWN)
 {
  nat0seq tmp(nat0(1), nat0(0));
  tmp.add(type);
  //return TC.add(tmp, T_ID, LEXEMS.get(identifier), adr, KNOWN);
  //identifier muss wieder geloescht werden (nur in post_con bekannt), also:
  TC.add_type(tmp, T_ID, identifier, adr, KNOWN);
 }
 return type;
}

nat0 p_type_pair_l1(const nodeseq & terms,
		    node * adr, const parse_typ & t_akt)
{ // Anlegen weiterer Parameter impliziter Funktionen
 nat0 p_type_list  = GET(1);
 nat0 pattern_list = GET(2);
 nat0 type         = GET(3);

 // Parameter-Liste anlegen, falls nicht schon weiter unten geschehen
 t_typ params;
 if(TC.get_type(p_type_list) != T_PARAMS)
 {
  nat0seq tmp;
  tmp.add(p_type_list);
  params = TC.add_type(tmp, T_PARAMS, T_PARAMS, adr, KNOWN);
 }
 else params = p_type_list;

 // neue Parameter anhaengen
 for(nat0 i = 1; i <= TC.len_subtypes(pattern_list); i++)
 {
  if(t_akt != UNKNOWN)
  {
   warn_typ warn = NOWARN;
   if(!eval_equal(TC.get_subtype(pattern_list, i), type, warn))
    VDM_ERROR(24, adr->linenr, "function-parameter");
   if(warn == INCOMP)
    VDM_WARNING(1, adr->linenr, "function-parameter");
  }
  TC.add_subtype(params, type);
 }
 return params;
}

nat0 p_type_pair_l2(const nodeseq & terms,
		    node * adr, const parse_typ & t_akt)
{ // Anlegen der ersten Parameter einer impliziten Funktion
 nat0 pattern_list = GET(1);
 nat0 type         = GET(2);
 nat0seq tmp;

 // Parameter aufbauen (alle mit gleichem Typ)
 for(nat0 i = 1; i <= TC.len_subtypes(pattern_list); i++)
 {
  if(t_akt != UNKNOWN)
  { // erst im zweiten Durchgang Pattern matchen
   warn_typ warn = NOWARN;
   if(!eval_equal(TC.get_subtype(pattern_list, i), type, warn))
    VDM_ERROR(24, adr->linenr, "function-parameter");
   if(warn == INCOMP)
    VDM_WARNING(1, adr->linenr, "function-parameter");
  }
  tmp.add(type);   // weiteren Parameter anlegen
 }

 if(tmp.len() == 1)
  return type;
 else
  return TC.add_type(tmp, T_PARAMS, T_PARAMS, adr, KNOWN);
}

nat0 parameters(const nodeseq & terms, node * adr, const parse_typ & t_akt)
{ // Umwandeln mehrerer Patterns in ein Produkt-Typ
 if (t_akt == UNKNOWN)
 {
  nat0 pattern_list_opt = GET(1);
  if(TC.get_type(pattern_list_opt) == PAT_LIST)
  {
   if(TC.len_subtypes(pattern_list_opt) > 1)
   {
    TC.put_type(pattern_list_opt, T_PROD);
    TC.put_typename(pattern_list_opt, T_PROD);
   }
   else return TC.get_subtype(pattern_list_opt, 1);
  }
  return pattern_list_opt;
 }
 else return t_akt;
}