#include "semoth.hpp"

nat0 inv_init(const nodeseq & terms,
	      node * adr, const parse_typ & t_akt)
{
 nat0 inv_pattern = GET(1);
 nat0 expression  = GET(2);

 // s. 4.2.1.3
 unary_ops(expression, T_BOOL,"invariant-condition",
	   adr->linenr, nat0seq());
 TC.close_layer(bool(0));
 return inv_pattern;
}

nat0 inv_pattern(const nodeseq & terms,
	      node * adr, const parse_typ & t_akt)
{
 nat0 pattern    = GET(1);

 // s. 4.2.1.2
 warn_typ warn = NOWARN;
 if(!eval_equal(pattern, t_akt, warn))
  VDM_ERROR(24, adr->linenr, "pattern");
 if(warn == INCOMP)
  VDM_WARNING(1, adr->linenr, "pattern");
 return t_akt;
}

nat0 val_def(const nodeseq & terms,
	     node * adr, const parse_typ & t_akt)
{
 nat0 pattern    = GET(1);
 nat0 type_opt   = GET(2);
 nat0 expression = GET(3);
 if(TC.get_type(type_opt) != NOPAT)
 {
  warn_typ warn = NOWARN;
  if(!eval_equal(pattern, type_opt, warn))
   VDM_ERROR(24, TC.get_adress(pattern)->linenr, "type-bind");
  if(warn == INCOMP)
   VDM_WARNING(1, TC.get_adress(pattern)->linenr, "type-bind");
 }
 warn_typ warn = NOWARN;
 if(!eval_equal(pattern, expression, warn))
  VDM_ERROR(24, TC.get_adress(pattern)->linenr, "value-definition");
 if(warn == INCOMP)
  VDM_WARNING(1, TC.get_adress(pattern)->linenr, "value-definition");
 return expression;
}

nat0 expression_list1(const nodeseq & terms, node * adr, const parse_typ & t_akt)
{
 // if (t_akt== UNKNOWN)
 {
  nat0 expression_list = GET(1);
  nat0 expression      = GET(2);
  if (VDM_TRACE) cout << "\nweiteres Pattern: " << expression << ":" << (char*)TC.get_typename(expression);
  TC.add_subtype(expression_list, expression);
  if (VDM_TRACE) cout << "\nnach expression";
  return expression_list;
 }
 // else return t_akt;
}

nat0 expression_list2(const nodeseq & terms, node * adr, const parse_typ & t_akt)
{
 // if (t_akt== UNKNOWN)
 {
  nat0 expression      = GET(1);
  nat0seq tmp(nat0(5), nat0(3));
  tmp.add(expression);
  if (VDM_TRACE) cout << "\nerstes Pattern = " << expression << ":" << (char*)TC.get_typename(expression);
  return TC.add_value(tmp, T_UNION, T_UNION, adr); // erstes Pattern einspeichern
 }
 // else return t_akt;
}

nat0 expression(const nodeseq & terms,         // expression = name
		node * adr, const parse_typ & t_akt)
{
 // if (t_akt== UNKNOWN)
 {
  nat0 name = GET(1);
  t_typ typ = TC.seek_identifier(LEXEMS.get(name));
  if(typ == UNKNOWN) // noch nicht deklariert
   VDM_ERROR(26, adr->linenr, LEXEMS.get(name));
  else
  {
   if(TC.get_type(typ) == T_COMP)
    VDM_ERROR(5, adr->linenr, LEXEMS.get(name));  /*typname = tagid*/
   return typ;
  }
 }
 return t_akt;
}

nat0 let_expression(const nodeseq & terms,
		    node * adr, const parse_typ & t_akt)
{
 nat0 expression = GET(2);
 if (VDM_TRACE) cout << "\n\nlet-close\n\n";
 TC.close_layer();
 return expression;
}

nat0 let_be_expression(const nodeseq & terms,
		    node * adr, const parse_typ & t_akt)
{
 nat0 be_opt = GET(2);
 nat0 expression = GET(3);
 if(TC.get_type(be_opt) != NOPAT)
 {
  warn_typ warn = NOWARN;
  if(!eval_equal(be_opt, T_BOOL, warn))
   VDM_ERROR(24, TC.get_adress(be_opt)->linenr, "be-such-that expression");
  if(warn == INCOMP)
   VDM_WARNING(1, TC.get_adress(be_opt)->linenr, "be-such-that expression");
 }
 if (VDM_TRACE) cout << "\n\nlet-be-close\n\n";
 TC.close_layer();
 return expression;
}

nat0 loc_def_list(const nodeseq & terms,
		  node * adr, const parse_typ & t_akt)
{
 return 0;
}

nat0 if_expression(const nodeseq & terms,
		   node * adr, const parse_typ & t_akt)
{
  nat0 expression1 = GET(1);
  nat0 expression2 = GET(2);
  nat0 expr_list   = GET(3);
  nat0 expression3 = GET(4);
  unary_ops(expression1, T_BOOL, "if-condition",
	    adr->linenr, nat0seq());
  t_typ t = expr_list;
  if(TC.get_type(expr_list) != NOPAT)
  {
   TC.add_subtype(expr_list, expression2);
   TC.add_subtype(expr_list, expression3);
   TC.put_type(expr_list, T_UNION);
   TC.put_typename(expr_list, T_UNION);
  }
  else
  {
   nat0seq tmp;
   tmp.add(expression2);
   tmp.add(expression3);
   t = TC.add_type(tmp, T_UNION, T_UNION, adr, KNOWN);
  }
  return t;
}

nat0 elseif_list(const nodeseq & terms,
		 node * adr, const parse_typ & t_akt)
{
  nat0 elseif_list = GET(1);
  nat0 elseif_expr = GET(2);
  t_typ t = elseif_list;
  if(TC.get_type(elseif_list) != NOPAT)
   TC.add_subtype(elseif_list, elseif_expr);
  else
  {
   nat0seq tmp;
   tmp.add(elseif_expr);
   t = TC.add_type(tmp, PAT_LIST, PAT_LIST, adr, KNOWN);
  }
  return t;
}

nat0 elseif_expression(const nodeseq & terms,
		       node * adr, const parse_typ & t_akt)
{
  nat0 expression1 = GET(1);
  nat0 expression2 = GET(2);
  unary_ops(expression1, T_BOOL,"elseif-condition",adr->linenr, nat0seq());
  return expression2;
}

nat0 cases_expression(const nodeseq & terms,
		      node * adr, const parse_typ & t_akt)
{
  nat0 expression   = GET(1);
  nat0 alternatives = GET(2);
  nat0 other_opt    = GET(3);
  nat0seq tmp;
  TC.output();
  for(nat0 i = 1; i <= TC.len_subtypes(alternatives); i++)
  {
   warn_typ warn = NOWARN;
   t_typ pattern = TC.get_subtype(TC.get_subtype(alternatives, i), 1);
   for(nat0 j = 1; j <= TC.len_subtypes(pattern); j++)
    if(!eval_equal(TC.get_subtype(pattern, j),
		   expression, warn))
     VDM_ERROR(24, TC.get_adress(TC.get_subtype(pattern, j))->linenr, "cases-alternative-pattern");
   if(warn == INCOMP)
    VDM_WARNING(1, TC.get_adress(TC.get_subtype(pattern, j))->linenr, "cases-alternative-pattern");
   tmp.add(TC.get_subtype(TC.get_subtype(alternatives, i), 2));
  }
  if(TC.get_type(other_opt) != NOPAT)
   tmp.add(other_opt);
  return TC.add_type(tmp, T_UNION, T_UNION, adr, KNOWN);
}

nat0 alternatives(const nodeseq & terms,
		  node * adr, const parse_typ & t_akt)
{
 return 0;
}

nat0 alternative(const nodeseq & terms,
		 node * adr, const parse_typ & t_akt)
{
  nat0 pat_list   = GET(1); //bindings sind schon abgeklaert
  nat0 expression = GET(2); //bindings sind schon abgeklaert
  nat0seq tmp(nat0(2), nat0(0));
  tmp.add(pat_list);
  tmp.add(expression);
  TC.close_layer();
  return TC.add_type(tmp, PAT_LIST, PAT_LIST, adr, KNOWN);
}

nat0 all_expression(const nodeseq & terms,
		    node * adr, const parse_typ & t_akt)
{
 // if (t_akt== UNKNOWN)
 {
  nat0 expression = GET(2); //bindings sind schon abgeklaert
  nat0seq dummy;
  unary_ops(expression, T_BOOL,"quantified expression filter",adr->linenr, dummy);
  if (VDM_TRACE) cout << "\n\nquantify-close\n\n";
  TC.close_layer();
  return T_BOOL;
 }
 // else return t_akt;
}

nat0 exist_expression(const nodeseq & terms,
		      node * adr, const parse_typ & t_akt)
{
 return 0;
}

nat0 exist1_expression(const nodeseq & terms,
		       node * adr, const parse_typ & t_akt)
{
 return 0;
}

nat0 iota_expression(const nodeseq & terms,
		     node * adr, const parse_typ & t_akt)
{
 // if (t_akt== UNKNOWN)
 {
  nat0 bind = GET(1);
  nat0 expression = GET(2);
  nat0seq dummy;
  unary_ops(expression, T_BOOL,"quantified expression filter",adr->linenr, dummy);
  if (VDM_TRACE) cout << "\n\niota-close\n\n";
  TC.close_layer();
  return bind;
 }
 // else return t_akt;
}

nat0 set_enum(const nodeseq & terms,
	      node * adr, const parse_typ & t_akt)
{
 // if (t_akt== UNKNOWN)
 {
  nat0 expression_list_opt = GET(1);
  t_typ t = expression_list_opt;
  if(TC.get_type(expression_list_opt) == NOPAT)
   t = T_EMPTYSET;
  else
  {
   nat0seq un;
   un.append(TC.get_subtypes(expression_list_opt));
   TC.replace_subtypes(expression_list_opt, un);
   if(TC.len_subtypes(expression_list_opt) > 1)
   {
    TC.put_type(expression_list_opt, T_UNION);
    TC.put_typename(expression_list_opt, T_UNION);
    nat0seq tmp(nat0(1), nat0(0));
    tmp.add(expression_list_opt);
    t = TC.add_type(tmp, T_SET, T_SET, adr, CONST);
   }
   else
   {
    TC.put_type(expression_list_opt, T_SET);
    TC.put_typename(expression_list_opt, T_SET);
    t = expression_list_opt;
   }
  }
  return t;
 }
 // else return t_akt;
}

nat0 set_comp(const nodeseq & terms,
	      node * adr, const parse_typ & t_akt)
{
 // if (t_akt== UNKNOWN)
 {// bind_list ist schon abgeklaert, id's sind gebunden
  nat0 expression          = GET(2);
  nat0 bind_expression_opt = GET(3);
  nat0seq dummy;
  if(TC.get_type(bind_expression_opt) != NOPAT)
   unary_ops(bind_expression_opt, T_BOOL,"set comprehension filter",adr->linenr, dummy);
  nat0seq tmp(nat0(1), nat0(0));
  tmp.add(expression);
  if (VDM_TRACE) cout << "\n\nset-comp-close\n\n";
  TC.close_layer();
  return TC.add_type(tmp, T_SET, T_SET, adr, CONST);
 }
 // else return t_akt;
}

nat0 set_range(const nodeseq & terms,
	       node * adr, const parse_typ & t_akt)
{
 // if (t_akt== UNKNOWN)
 {
  nat0 expression1 = GET(1);
  nat0 expression2 = GET(2);
  nat0seq dummy;
  unary_ops(expression1, T_INT,"set range expression, 1. number",adr->linenr, dummy);
  unary_ops(expression2, T_INT,"set range expression, 2. number",adr->linenr, dummy);
  nat0seq tmp(nat0(1), nat0(0));
  tmp.add(T_INT);
  return TC.add_type(tmp, T_SET, T_SET, adr, CONST);
 }
 // else return t_akt;
}

nat0 seq_enum(const nodeseq & terms,
	      node * adr, const parse_typ & t_akt)
{
 // if (t_akt== UNKNOWN)
 {
  nat0 expression_list_opt = GET(1);
  t_typ t = expression_list_opt;
  if(TC.get_type(expression_list_opt) == NOPAT)
   t = T_EMPTYSEQ;
  else
  {
   nat0seq un;
   un.append(TC.get_subtypes(expression_list_opt));
   TC.replace_subtypes(expression_list_opt, un);
   if(TC.len_subtypes(expression_list_opt) > 1)
   {
    TC.put_type(expression_list_opt, T_UNION);
    TC.put_typename(expression_list_opt, T_UNION);
    nat0seq tmp(nat0(1), nat0(0));
    tmp.add(expression_list_opt);
    t = TC.add_type(tmp, T_SEQ, T_SEQ, adr, CONST);
   }
   else
   {
    TC.put_type(expression_list_opt, T_SEQ);
    TC.put_typename(expression_list_opt, T_SEQ);
    t = expression_list_opt;
   }
  }
  return t;
 }
 // else return t_akt;
}

nat0 seq_comp(const nodeseq & terms,
	      node * adr, const parse_typ & t_akt)
{
 // if (t_akt== UNKNOWN)
 {// bind_list ist schon abgeklaert, id's sind gebunden
  nat0 expression          = GET(2);
  nat0 bind_expression_opt = GET(3);
  nat0seq dummy;
  if(TC.get_type(bind_expression_opt) != NOPAT)
   unary_ops(bind_expression_opt, T_BOOL,"seq comprehension filter",adr->linenr, dummy);
  nat0seq tmp(nat0(1), nat0(0));
  tmp.add(expression);
  if (VDM_TRACE) cout << "\n\nseq-comp-close\n\n";
  TC.close_layer();
  return TC.add_type(tmp, T_SEQ, T_SEQ, adr, CONST);
 }
 // else return t_akt;
}

nat0 subseq(const nodeseq & terms,
	    node * adr, const parse_typ & t_akt)
{
 // if (t_akt== UNKNOWN)
 {
  nat0 expression1 = GET(1);
  nat0 expression2 = GET(2);
  nat0 expression3 = GET(3);
  nat0seq tmp(nat0(1), nat0(0));
  nat0seq dummy;
  unary_ops(expression1, T_SEQ,"subsequence expression, sequence-type",adr->linenr, tmp);
  unary_ops(expression2, T_NAT1,"set range expression, 1. number",adr->linenr, dummy);
  unary_ops(expression3, T_NAT1,"set range expression, 2. number",adr->linenr, dummy);
  if(tmp.len() == 1)
   return tmp.get(1);
  else
   return TC.add_type(tmp, T_UNION, T_UNION, adr, KNOWN);
 }
 // else return t_akt;
}

nat0 map_enum(const nodeseq & terms,
	      node * adr, const parse_typ & t_akt)
{
 // if (t_akt== UNKNOWN)
 {
  nat0 maplet_list = GET(1);
  nat0seq dom, rng;
  for(nat0 i = 1; i <= TC.len_subtypes(maplet_list); i++)
  {
   if(!dom.isin(TC.get_subtype(TC.get_subtype(maplet_list, i), 1)))
    dom.add(TC.get_subtype(TC.get_subtype(maplet_list, i), 1));
   if(!rng.isin(TC.get_subtype(TC.get_subtype(maplet_list, i), 2)))
    rng.add(TC.get_subtype(TC.get_subtype(maplet_list, i), 2));
  }
  nat0seq map(nat0(2), nat0(0));
  if(dom.len() == 1)
   map.add(dom.get(1));
  else
   map.add(TC.add_type(dom, T_UNION, T_UNION, adr, KNOWN));
  if(rng.len() == 1)
   map.add(rng.get(1));
  else
   map.add(TC.add_type(rng, T_UNION, T_UNION, adr, KNOWN));
  return TC.add_type(map, T_MAP, T_MAP, adr, CONST);
 }
 // else return t_akt;
}

nat0 maplet_list(const nodeseq & terms,
		 node * adr, const parse_typ & t_akt)
{
 return 0;
}

nat0 maplet(const nodeseq & terms,
	    node * adr, const parse_typ & t_akt)
{
 // if (t_akt== UNKNOWN)
 {
  nat0 expression1 = GET(1);
  nat0 expression2 = GET(2);
  nat0seq tmp(nat0(2), nat0(0));
  tmp.add(expression1);
  tmp.add(expression2);
  return TC.add_type(tmp, T_MAP, T_MAP, adr, CONST);
 }
 // else return t_akt;
}

nat0 map_comp(const nodeseq & terms,
	      node * adr, const parse_typ & t_akt)
{
 // if (t_akt== UNKNOWN)
 {// bind_list ist schon abgeklaert, id's sind gebunden
  nat0 maplet              = GET(2);
  nat0 bind_expression_opt = GET(3);
  nat0seq dummy;
  if(TC.get_type(bind_expression_opt) != NOPAT)
   unary_ops(bind_expression_opt, T_BOOL,"seq comprehension filter",adr->linenr, dummy);
  if (VDM_TRACE) cout << "\n\nmap-comp-close\n\n";
  TC.close_layer();
  return maplet;
 }
 // else return t_akt;
}

nat0 tuple(const nodeseq & terms,
	   node * adr, const parse_typ & t_akt)
{
 return 0;
}

nat0 record(const nodeseq & terms,
	   node * adr, const parse_typ & t_akt)
{
 // if (t_akt== UNKNOWN)
 {
  nat0 name             = GET(1);
  nat0 expression_list_opt = GET(2);
  t_typ t = TC.seek_tagfield(LEXEMS.get(name));
  if (TC.get_type(t) != T_COMP)
   VDM_ERROR(17, adr->linenr, LEXEMS.get(name));
  TC.put_type(expression_list_opt, T_COMP);
  TC.put_typename(expression_list_opt, name);
  bool oldmath = MATH;
  MATH = bool(0);
  warn_typ warn = NOWARN;
  if(!eval_equal(expression_list_opt, t, warn))
   VDM_ERROR(24, adr->linenr, "record-construction '" + LEXEMS.get(name) + "'");
  if(warn == INCOMP)
   VDM_WARNING(1, adr->linenr, "record-construction '" + LEXEMS.get(name) + "'");
  MATH = oldmath;
  for(nat0 i = 1; i <= TC.len_subtypes(expression_list_opt); i++)
   if(TC.get_type(TC.get_subtype(t, i)) == T_FIELD)
   {
    nat0seq tmp(nat0(1), nat0(0));
    tmp.add(TC.get_subtype(expression_list_opt, i));
    TC.put_subtype(expression_list_opt, i, TC.add_type(tmp, T_FIELD,
		   TC.get_subtype(t, i), adr, KNOWN));
   }
  return expression_list_opt;
 }
 // else return t_akt;
}

nat0 record_mod(const nodeseq & terms,
		node * adr, const parse_typ & t_akt)
{
 // if (t_akt == UNKNOWN)
 {
  nat0 expression  = GET(1);
  nat0 rec_list    = GET(2);
  nat0seq anzahl;
  unary_ops(expression, T_COMP,"record modification, record expression",adr->linenr, anzahl);
  bool oldmath = MATH;
  MATH = bool(0);
  nat0seq tmp;
  bool is_incomp = bool(0);
  for(nat0 h = 1; h <= anzahl.len(); h++)
  {
   nat0 test = 0;
   bool pos_isincomp = bool(0);
   for(nat0 i = 1; i <= TC.len_subtypes(anzahl.get(h)); i++)
   {// eindeutige Liste zuerst
    bool isin = bool(0);
    bool is_break = 0;
    for(nat0 j = 1; j <= TC.len_subtypes(rec_list); j++)
    {
     if (VDM_TRACE) cout << "\n" << (char*)TC.get_typename(TC.get_subtype(anzahl.get(h), i));
     if (VDM_TRACE) cout << "\n" << (char*)TC.get_typename(TC.get_subtype(rec_list, j));
     if(TC.get_type(TC.get_subtype(anzahl.get(h), i)) == T_FIELD)
     {
      if (VDM_TRACE) cout << "\n" << (char*)TC.get_typename(TC.get_subtype(anzahl.get(h), i));
      if (VDM_TRACE) cout << "\n" << (char*)TC.get_typename(TC.get_subtype(rec_list, j));
      if(TC.get_typename(TC.get_subtype(anzahl.get(h), i)) ==
	 TC.get_typename(TC.get_subtype(rec_list, j)))
      {
       if(isin)
	VDM_ERROR(31, adr->linenr, TC.get_typename(TC.get_subtype(anzahl.get(h), i)));
       else
       {
	isin = bool(1);
	test++;
	warn_typ warn = NOWARN;
	if(!eval_equal(TC.get_subtype(anzahl.get(h), i),
		       TC.get_subtype(rec_list, j), warn))
	{
	 is_break = bool(1); // zum Abbrechen der Comp-Komponenten-Liste
	 break;              // zum Abbrechen der Map-Liste
	}
	if(warn == INCOMP)
	 pos_isincomp = bool(1);
       }
      }
     }
    }
    if(is_break) break; // naechster Comp-Typ
   }
   if(test == TC.len_subtypes(rec_list))
   {
    tmp.add(anzahl.get(h));
    is_incomp = (is_incomp?is_incomp:pos_isincomp);
   }
  }
  MATH = oldmath;
  if (VDM_TRACE) cout << "\nrec_mod an : " << adr->linenr;
  TC.output();
  if(tmp.len() < 1)
   VDM_ERROR(24, adr->linenr, "record-modification '" +
	     TC.get_typename(anzahl.get(h)) + "'");
  /*if(test < TC.len_subtypes(rec_list))
   VDM_ERROR(32, yylino, TC.get_typename(expression));*/
  if(is_incomp)
   VDM_WARNING(1, adr->linenr, "record-modification '" +
	       TC.get_typename(anzahl.get(h)) + "'");
  if(tmp.len() == 1)
   return tmp.get(1);
  else
   return TC.add_type(tmp, T_UNION, T_UNION, adr, KNOWN);
 }
 // else return t_akt;
}

nat0 recmod_list(const nodeseq & terms,
		 node * adr, const parse_typ & t_akt)
{
 return 0;
}

nat0 rec_mod(const nodeseq & terms,
	     node * adr, const parse_typ & t_akt)
{
 // if (t_akt== UNKNOWN)
 {
  nat0 identifier = GET(1);
  nat0 expression = GET(2);
  nat0seq tmp(nat0(2), nat0(0));
  /*tmp.add(identifier);*/
  tmp.add(expression);
  /*return TC.add_type(tmp, T_MAP, T_MAP, adr, CONST);*/

  return TC.add_type(tmp, T_FIELD, identifier, adr, CONST);
 }
 // else return t_akt;
}

nat0 apply(const nodeseq & terms,
	   node * adr, const parse_typ & t_akt)
{
 // if (t_akt== UNKNOWN)
 {
  nat0 expression = GET(1);
  nat0 expression_list_opt = GET(2);
  nat0seq anzahl1, anzahl2;
  bool warning1 = bool(0), warning2 = bool(0);
  bool is_incomp = bool(0)/*, is_difftyp = bool(0)*/;
  bool old_math = MATH;
  MATH = bool(0);
  nat0 i;
  nat0seq tmp;
  if(TC.len_subtypes(expression_list_opt) == 1) // NOPAT hat 0
  {
   unary_ops(expression, T_SEQ, "", adr->linenr, anzahl1, bool(0), warning1);
   if(anzahl1.len() > 0)
    unary_ops(TC.get_subtype(expression_list_opt, 1),
	      T_NAT1, "", adr->linenr, anzahl2, bool(0), warning2);
   if(anzahl2.len() > 0)
   {
    for(i = 1; i <= anzahl1.len(); i++)
    {
     if (VDM_TRACE) cout << "\nseq-apply ok";
     tmp.add(TC.get_subtype(anzahl1.get(i), 1));
    }
    is_incomp = (warning1?warning1:warning2);
   }
   anzahl1 = nat0seq();
   warning1 = bool(0);
   unary_ops(expression, T_MAP,"",adr->linenr, anzahl1, bool(0), warning1);
   warn_typ warn = NOWARN;
   bool isin = bool(0);
   for(i = 1; i <= anzahl1.len(); i++)
    if(eval_equal(TC.get_subtype(anzahl1.get(i), 1),
		  TC.get_subtype(expression_list_opt, 1), warn))
    {
     if (VDM_TRACE) cout << "\nmap-apply ok";
     tmp.add(TC.get_subtype(anzahl1.get(i), 2));
     isin = bool(1);
     if(warn == INCOMP) is_incomp = bool(1);
    }
   is_incomp = (is_incomp?is_incomp:isin?warning1:is_incomp);
  }
  anzahl1 = nat0seq();
  warning1 = bool(0);
  unary_ops(expression, T_FUNC,"",adr->linenr, anzahl1, bool(0), warning1);
  if(anzahl1.len() > 0)
  {
   warn_typ warn = NOWARN;
   bool isin = bool(0);
   nat0 i;
   if(TC.len_subtypes(expression_list_opt) == 1)
    expression_list_opt = TC.get_subtype(expression_list_opt, 1);
   if(TC.len_subtypes(expression_list_opt) > 1)
    TC.put_type(expression_list_opt, T_PROD);
   for(i = 1; i <= anzahl1.len(); i++)
    if(eval_equal(expression_list_opt,
		  TC.get_subtype(anzahl1.get(i), 1), warn))
    {
     tmp.add(TC.add_type(TC.get_all(TC.get_subtype(anzahl1.get(i), 2))));
     isin = bool(1);
     if(warn == INCOMP) is_incomp = bool(1);
//     if(warn == DIFFTYP) is_difftyp = bool(1);
    }
  }
  MATH = old_math;
  if(tmp.len() == 0)
   VDM_ERROR(24, adr->linenr, "apply operator '.(.)'");
  if(is_incomp)  // Set enthaelt noch andere Union-Typen
  {
   VDM_WARNING(1, adr->linenr, "apply operator '.(.)'");
  }
  if(tmp.len() == 1)
   return tmp.get(1);
  else
   return TC.add_type(tmp, T_UNION, T_UNION, adr, KNOWN);
 }
 // else return t_akt;
}

nat0 field_sel(const nodeseq & terms,
	       node * adr, const parse_typ & t_akt)
{
 // if (t_akt == UNKNOWN)
 {
  nat0 expression  = GET(1);
  nat0 identifier    = GET(2);
  nat0seq anzahl, tmp;
  unary_ops(expression, T_COMP,"field-selection, record expression",adr->linenr, anzahl);
  bool var_typ = bool(0);
  for(nat0 h = 1; h <= anzahl.len(); h++)
   for(nat0 i = 1; i <= TC.len_subtypes(anzahl.get(h)); i++)
    if((TC.get_type(TC.get_subtype(anzahl.get(h), i)) == T_FIELD) &&
       (TC.get_typename(TC.get_subtype(anzahl.get(h), i)) ==
	LEXEMS.get(identifier)))
     tmp.add(TC.get_subtype(anzahl.get(h), i));
    else if(TC.get_type(TC.get_subtype(anzahl.get(h), i)) == T_VAR)
     var_typ = bool(1);
  if(tmp.len() == 0)
   if(var_typ)
   {
//    nat0seq tmp;
    tmp.add(T_VAR);
//    TC.insert_subtype(var_typ, 1,
//     TC.add_type(tmp, T_FIELD, LEXEMS.get(identifier), adr, KNOWN));
//    tmp.add(T_VAR);
   }
   else
    VDM_ERROR(32, adr->linenr, TC.get_typename(expression));
  if(tmp.len() == 1)
   return tmp.get(1);
  else
   return TC.add_type(tmp, T_UNION, T_UNION, adr, KNOWN);
 }
 // else return t_akt;
}

nat0 type_inst(const nodeseq & terms,
	       node * adr, const parse_typ & t_akt)
{
 // if (t_akt == UNKNOWN)
 {
  nat0 name      = GET(1);
  nat0 type_list = GET(2);
  t_typ typ = TC.seek_identifier(LEXEMS.get(name));
  if(typ == UNKNOWN) // noch nicht deklariert
   VDM_ERROR(26, adr->linenr, LEXEMS.get(name));
  t_typ t = TC.get_subtype(typ, 1);
  if(TC.get_type(t) != T_POLY) // kein identifier einer poly-func
   VDM_ERROR(24, adr->linenr, LEXEMS.get(name));
  if((TC.len_subtypes(t) - 1) != TC.len_subtypes(type_list))
   VDM_ERROR(33, adr->linenr, LEXEMS.get(name));
  if(!TC.is_marked_union(typ))
  {
   TC.mark_union(typ);
   for(nat0 i = 1; i <= TC.len_subtypes(type_list); i++)
   {
    if (VDM_TRACE) cout << "\n" << (char*)TC.get_typename(TC.get_subtype(t, i)) << " und " << (char*)TC.get_typename(TC.get_subtype(type_list, i));
    if(TC.get_type(TC.get_subtype(t, i)) == T_VAR)
     TC.add_subtype(TC.get_subtype(t, i), TC.get_subtype(type_list, i));
    else
     TC.put_subtype(TC.get_subtype(t, i), 1, TC.get_subtype(type_list, i));
    TC.put_type(TC.get_subtype(t, i), T_ID); // Typ-Variable wird Typ-Name
   }
   TC.get_prec(typ)->parse();
   TC.unmark_union(typ);
  }
  nat0seq tmp(nat0(2), nat0(0));
  // Funktions-Parameter:
  tmp.add(TC.get_subtype(TC.get_subtype(t, TC.len_subtypes(t)), 1));
  // Kopie des Return-Typs (wird weiterverwendet, darf nicht berschr. werden)
  tmp.add(TC.copy_type(TC.get_subtype(TC.get_subtype(t, TC.len_subtypes(t)), 2)));
  return TC.add_type(tmp, T_FUNC, T_FUNC, adr, KNOWN);
 }
 // else return t_akt;
}

nat0 type_list(const nodeseq & terms,
	       node * adr, const parse_typ & t_akt)
{
 return 0;
}

nat0 lambda_expression(const nodeseq & terms,
		       node * adr, const parse_typ & t_akt)
{
  nat0 expression = GET(2);
  if (VDM_TRACE) cout << "\n\nlambda-close\n\n";
  TC.close_layer();
  return expression;
}

nat0 is_expression1(const nodeseq & terms,
		    node * adr, const parse_typ & t_akt)
{
  nat0 name       = GET(1);
  nat0 expression = GET(2);
  t_typ t = TC.seek_identifier(LEXEMS.get(name)); // Standard: typenames
  if(t == UNKNOWN)
   VDM_ERROR(26, adr->linenr, LEXEMS.get(name));
  warn_typ warn = NOWARN;
  if(!eval_equal(t, expression, warn))
   VDM_WARNING(8, adr->linenr, LEXEMS.get(name));
  return T_BOOL;
}

nat0 is_expression2(const nodeseq & terms,
		    node * adr, const parse_typ & t_akt)
{
  nat0 is_basic_type = GET(1);
  nat0 expression    = GET(2);
  warn_typ warn = NOWARN;
  if(!eval_equal(is_basic_type, expression, warn))
   VDM_WARNING(8, TC.get_adress(is_basic_type)->linenr, TC.get_typename(is_basic_type));
  return T_BOOL;
}