#include "semtypes.hpp"

nat0 type_definition1(const nodeseq & terms, node * adr, const parse_typ & t_akt)
{
 if (t_akt == UNKNOWN)
 {
  nat0 identifier    = GET(1);
  nat0 type          = GET(2);
  nat0 invariant_opt = GET(3);

  // s. 4.2.1.1
  t_typ typ = TC.seek_identifier(LEXEMS.get(identifier));
  if (TC.seek_tagfield(LEXEMS.get(identifier)) != UNKNOWN)
  {
   VDM_ERROR(1, adr->linenr, LEXEMS.get(identifier));
  }

  t_typ ret_typ = typ;
  if (typ == UNKNOWN) // typname neu: anlegen
  {
   nat0seq tmp(nat0(1), nat0(0));
   tmp.add(type);
   ret_typ = TC.add_type(tmp, T_ID, identifier, adr, KNOWN);
  }

  // typname schon als typname eingetragen (spaetere Deklaration)
  else
  {
   if (TC.len_subtypes(typ) == 0)
   {      // wenn vorhandener Typ nicht deklariert (bisher nur verwendet)
    TC.add_subtype(typ, type);  // trage Deklaration ein
    TC.put_state(typ, KNOWN);
    ret_typ = typ;
   }      // sonst Redeklaration -> Fehler
   else VDM_ERROR(2, adr->linenr, LEXEMS.get(identifier));
  }

  if(TC.get_type(invariant_opt) != NOPAT)
  {
  // Pattern-Typ fuer Invariante weiterreichen (erst hier moeglich)
   terms.get(3)->terms.get(1)->typ = ret_typ; // typ fuer invariant-pattern
  // inv-Funktion bereitstellen
   nat0seq tmp2(nat0(2), nat0(0));
   tmp2.add(ret_typ);
   tmp2.add(T_BOOL);
   TC.add_type(tmp2,
	       T_FUNC,
	       LEXEMS.add("inv_" + LEXEMS.get(identifier)),
	       adr, KNOWN);
  }

  return ret_typ;
 }
 else return t_akt;
}

nat0 type_definition2 (const nodeseq & terms, node * adr, const parse_typ & t_akt)
{
 if (t_akt == UNKNOWN)
 {
  nat0 identifier    = GET(1);
  nat0 field_list    = GET(2);
  nat0 invariant_opt = GET(3);
  t_typ typ = TC.seek_identifier(LEXEMS.get(identifier));
  if(typ != UNKNOWN)
   if(TC.len_subtypes(typ) != 0)
   {
    if (VDM_TRACE) cout << "\nTyp " << typ << ":" << (char*)TC.get_typename(typ) << "; Art: " << TC.get_type(typ) << ":" << (char*)TC.get_kindname(typ);
    VDM_ERROR(3, adr->linenr, LEXEMS.get(identifier));
   }
  t_typ tag_typ = TC.seek_tagfield(LEXEMS.get(identifier));// s. 4.2.2.2.3.
  if(tag_typ != UNKNOWN)
  {
   warn_typ warn = NOWARN;
   if(!eval_equal(tag_typ, field_list, warn))
    VDM_ERROR(34, adr->linenr, LEXEMS.get(identifier));
  }
  TC.put_typename(field_list, identifier);
  nat0seq tmp(nat0(1), nat0(0));
  tmp.add(field_list);
  t_typ ret_typ;
  if(typ != UNKNOWN)
  {
   ret_typ = typ;
   TC.replace_subtypes(typ, tmp);
   TC.put_state(typ, KNOWN);
  }
  else ret_typ = TC.add_type(tmp, T_ID, identifier, adr, KNOWN);
  if(TC.get_type(invariant_opt) != NOPAT)
  {
   (terms.get(3))->terms.get(1)->typ = ret_typ; // typ fuer invariant-pattern
   nat0seq tmp2(nat0(2), nat0(0));
   tmp2.add(ret_typ);
   tmp2.add(T_BOOL);
   TC.add_type(tmp2,
	       T_FUNC,
	       LEXEMS.add("inv_" + LEXEMS.get(identifier)),
	       adr, KNOWN);
  }
  return ret_typ;
 }
 else return t_akt;
}

nat0 quote_type(const nodeseq & terms, node * adr, const parse_typ & t_akt)
{ // s. 4.2.2.1
 if (t_akt == UNKNOWN)
 {
  nat0 quotelit = GET(1);
  nat0seq tmp(nat0(1), nat0(0));
  tmp.add(T_BASIC);
  return TC.add_type(tmp, T_QUOTE, quotelit, adr, KNOWN);
 }
 else return t_akt;
}

nat0 composite_type(const nodeseq & terms, node * adr, const parse_typ & t_akt)
{ // s. 4.2.2.2
 if (t_akt == UNKNOWN)
 {
  nat0 identifier = GET(1);
  nat0 field_list = GET(2);
  t_typ typ = TC.seek_identifier(LEXEMS.get(identifier));  // s. 4.2.2.2.1.
  t_typ tag_typ = TC.seek_tagfield(LEXEMS.get(identifier));// s. 4.2.2.2.3.
  if(tag_typ != UNKNOWN)
  {
   warn_typ warn = NOWARN;
   if(!eval_equal(tag_typ, field_list, warn))
    VDM_ERROR(34, adr->linenr, LEXEMS.get(identifier));
  }
  if(typ != UNKNOWN)
  {
   warn_typ warn = NOWARN;
   if(!eval_equal(typ, field_list, warn))
    VDM_ERROR(3, adr->linenr, LEXEMS.get(identifier));
  }
  else TC.put_typename(field_list, identifier);           // s. 4.2.2.2.4.
  return field_list;
 }
 else return t_akt;
}

nat0 field_list1(const nodeseq & terms, node * adr, const parse_typ & t_akt)
{ // sammelt die einzelnen Compose-Felder
 if (t_akt == UNKNOWN)
 {
  nat0 field_list = GET(1);
  nat0 field      = GET(2);
  for (nat0 i = 1; i <= TC.len_subtypes(field_list); i++)
  if (TC.get_typename(TC.get_subtype(field_list, i)) ==
      TC.get_typename(field))                             // s. 4.2.2.2.2.
   VDM_ERROR(4, adr->linenr, TC.get_typename(field));
  TC.add_subtype(field_list, field);
  return field_list;
 }
 else return t_akt;
}

nat0 field_list2(const nodeseq & terms, node * adr, const parse_typ & t_akt)
{ // nimmt unbenannte Compose-Felder auf
 if (t_akt == UNKNOWN)
 {
  nat0 field_list = GET(1);
  nat0 type       = GET(2);
  TC.add_subtype(field_list, type);
  return field_list;
 }
 else return t_akt;
}

nat0 field_list3(const nodeseq & terms, node * adr, const parse_typ & t_akt)
{ // legt den Compose-Typ an
 if (t_akt == UNKNOWN)
 {
  return TC.add_type(nat0seq(), T_COMP, T_COMP, adr, KNOWN);
 }
 else return t_akt;
}

nat0 field(const nodeseq & terms, node * adr, const parse_typ & t_akt)
{ // nimmt das erste Compose-Feld auf, falls vorhanden
 if (t_akt == UNKNOWN)
 {
  nat0 identifier = GET(1);
  nat0 type       = GET(2);
  nat0seq tmp(nat0(1), nat0(0));
  tmp.add(type);
  return TC.add_type(tmp, T_FIELD, identifier, adr, KNOWN);
 }
 else return t_akt;
}

nat0 union_type1(const nodeseq & terms, node * adr, const parse_typ & t_akt)
{ // Konstruiert die ersten beiden Union-Komponenten, s 4.2.2.3
 if (t_akt == UNKNOWN)
 {
  nat0 type1 = GET(1);
  nat0 type2 = GET(2);
  nat0seq tmp;
  tmp.add(type1);
  tmp.add(type2);
  return TC.add_type(tmp, T_UNION, T_UNION, adr, KNOWN);
 }
 else return t_akt;
}

nat0 union_type2(const nodeseq & terms, node * adr, const parse_typ & t_akt)
{ // haengt eine weitere Union-Komponente an, s. 4.2.2.3
 if (t_akt == UNKNOWN)
 {
  nat0 union_type = GET(1);
  nat0 type       = GET(2);
  TC.add_subtype(union_type, type);
  return union_type;
 }
 else return t_akt;
}

nat0 product_type1(const nodeseq & terms, node * adr, const parse_typ & t_akt)
{ // Konstruiert die ersten beiden Produkt-Komponenten, s 4.2.2.4
 if (t_akt == UNKNOWN)
 {
  nat0 type1 = GET(1);
  nat0 type2 = GET(2);
  nat0seq tmp;
  tmp.add(type1);
  tmp.add(type2);
  return TC.add_type(tmp, T_PROD, T_PROD, adr, KNOWN);
 }
 else return t_akt;
}

nat0 product_type2(const nodeseq & terms, node * adr, const parse_typ & t_akt)
{ // haengt eine weitere Produkt-Komponente an, s. 4.2.2.4
 if (t_akt == UNKNOWN)
 {
  nat0 product_type = GET(1);
  nat0 type         = GET(2);
  TC.add_subtype(product_type, type);
  return product_type;
 }
 else return t_akt;
}

nat0 optional_type(const nodeseq & terms, node * adr, const parse_typ & t_akt)
{ // s. 4.2.2.5
 if (t_akt == UNKNOWN)
 {
  nat0 type = GET(1);
  nat0seq tmp(nat0(2), nat0(0));
  tmp.add(type);
  tmp.add(T_NIL);
  return TC.add_type(tmp, T_UNION, T_UNION, adr, KNOWN);
 }
 else return t_akt;
}

nat0 set_type(const nodeseq & terms, node * adr, const parse_typ & t_akt)
{ // s. 4.2.2.6
 if (t_akt == UNKNOWN)
 {
  nat0 type = GET(1);
  nat0seq tmp(nat0(1), nat0(0));
  tmp.add(type);
  return TC.add_type(tmp, T_SET, T_SET, adr, KNOWN);
 }
 else return t_akt;
}

nat0 seq_type1(const nodeseq & terms, node * adr, const parse_typ & t_akt)
{ // s. 4.2.2.7
 if (t_akt == UNKNOWN)
 {
  nat0 type = GET(1);
  nat0seq tmp(nat0(1), nat0(0));
  tmp.add(type);
  /*nat0seq tmp2;
  tmp2.add(TC.add_type(tmp, T_SEQ1, T_SEQ1, adr, KNOWN));
  tmp2.add(T_EMPTYSEQ);
  return TC.add_type(tmp2, T_UNION, T_UNION, adr, KNOWN);*/
  return TC.add_type(tmp, T_SEQ, T_SEQ, adr, KNOWN);
 }
 else return t_akt;
}

nat0 seq_type2(const nodeseq & terms, node * adr, const parse_typ & t_akt)
{ // s.4.2.2.7
 if (t_akt == UNKNOWN)
 {
  nat0 type = GET(1);
  nat0seq tmp(nat0(1), nat0(0));
  tmp.add(type);
  return TC.add_type(tmp, T_SEQ1, T_SEQ1, adr, KNOWN);
 }
 else return t_akt;
}

nat0 general_map_type(const nodeseq & terms, node * adr, const parse_typ & t_akt)
{ // s. 4.2.2.8
 if (t_akt == UNKNOWN)
 {
  nat0 type1 = GET(1);
  nat0 type2 = GET(2);
  nat0seq tmp(nat0(2), nat0(0));
  tmp.add(type1);
  tmp.add(type2);
  return TC.add_type(tmp, T_MAP, T_MAP, adr, KNOWN);
 }
 else return t_akt;
}

nat0 injective_map_type(const nodeseq & terms, node * adr, const parse_typ & t_akt)
{ // s. 4.2.2.8
 if (t_akt == UNKNOWN)
 {
  nat0 type1 = GET(1);
  nat0 type2 = GET(2);
  nat0seq tmp(nat0(2), nat0(0));
  tmp.add(type1);
  tmp.add(type2);
  return TC.add_type(tmp, T_INMAP, T_INMAP, adr, KNOWN);
 }
 else return t_akt;
}

nat0 partial_function_type1(const nodeseq & terms, node * adr, const parse_typ & t_akt)
{ // s. 4.2.2.9
 if (t_akt == UNKNOWN)
 {
  nat0 type1 = GET(1);
  nat0 type2 = GET(2);
  t_typ t = type1;
/*  {
   if(TC.get_type(type1) == T_PROD)
   {
    TC.put_type(type1, T_PARAMS);
    TC.put_typename(type1, "PARAMETER");
   }
   else
   {
    nat0seq tmp(nat0(1), nat0(0));
    tmp.add(type1);
    t = TC.add_type(tmp, T_PARAMS, String("PARAMETER"), adr, KNOWN);
   }
  } */
  {
   nat0seq tmp(nat0(2), nat0(0));
   tmp.add(t);
   tmp.add(type2);
   return TC.add_type(tmp, T_FUNC, T_FUNC, adr, KNOWN);
  }
 }
 else return t_akt;
}

nat0 partial_function_type2(const nodeseq & terms, node * adr, const parse_typ & t_akt)
{ // s. 4.2.2.9
 if (t_akt == UNKNOWN)
 {
  nat0 type = GET(1);
  /*t_typ t = type;
  {
   nat0seq tmp(nat0(1), nat0(0));
   tmp.add(type);
   t = TC.add_type(tmp, T_PARAMS, String("PARAMETER"), adr, KNOWN);
  }*/
  nat0seq tmp(nat0(2), nat0(0));
  tmp.add(NOPAT);
  tmp.add(type);
  return TC.add_type(tmp, T_FUNC, T_FUNC, adr, KNOWN);
 }
 else return t_akt;
}

nat0 total_function_type1(const nodeseq & terms, node * adr, const parse_typ & t_akt)
{ // s. 4.2.2.9
 if (t_akt == UNKNOWN)
 {
  nat0 type1 = GET(1);
  nat0 type2 = GET(2);
  t_typ t = type1;
/*  {
   if(TC.get_type(type1) == T_PROD)
   {
    TC.put_type(type1, T_PARAMS);
    TC.put_typename(type1, "PARAMETER");
   }
   else
   {
    nat0seq tmp(nat0(1), nat0(0));
    tmp.add(type1);
    t = TC.add_type(tmp, T_PARAMS, String("PARAMETER"), adr, KNOWN);
   }
  } */
  nat0seq tmp(nat0(2), nat0(0));
  tmp.add(t);
  tmp.add(type2);
  return TC.add_type(tmp, T_FUNC, T_FUNC, adr, KNOWN);
 }
 else return t_akt;
}

nat0 total_function_type2(const nodeseq & terms, node * adr, const parse_typ & t_akt)
{ // s. 4.2.2.9
 if (t_akt == UNKNOWN)
 {
  nat0 type = GET(1);
  /*t_typ t = type;
  {
   nat0seq tmp(nat0(1), nat0(0));
   tmp.add(type);
   t = TC.add_type(tmp, T_PARAMS, String("PARAMETER"), adr, KNOWN);
  } */
  nat0seq tmp(nat0(2), nat0(0));
  tmp.add(NOPAT);
  tmp.add(type);
  return TC.add_type(tmp, T_FUNC, T_FUNC, adr, KNOWN);
 }
 else return t_akt;
}

nat0 type_name(const nodeseq & terms, node * adr, const parse_typ & t_akt)
{ // s. 4.2.2.10
 if (t_akt == UNKNOWN)
 {
  nat0 name = GET(1);
  t_typ typ = TC.seek_identifier(LEXEMS.get(name));
  if(typ == UNKNOWN) // noch nicht deklariert
   return TC.add_type(nat0seq(), T_ID, name, adr, UNKNOWN);
  else
  {
   if(TC.get_type(typ) == T_COMP)
    VDM_ERROR(5, adr->linenr, LEXEMS.get(name));  /*typname = tagid*/
   return typ;
  }
 }
 else return t_akt;
}

nat0 type_variable(const nodeseq & terms, node * adr, const parse_typ & t_akt)
{ // s. 4.2.2.11
  nat0 type_variable_identifier = GET(1);
  if (VDM_TRACE) cout << "\ngesucht: " << (char*)POLY << "." << (char*)LEXEMS.get(type_variable_identifier) << "|";
  t_typ typ = TC.seek_type(POLY + "." + LEXEMS.get(type_variable_identifier));
  if (typ == UNKNOWN)
     VDM_ERROR(6, adr->linenr,
		LEXEMS.get(type_variable_identifier));
  return typ;
}