#include "typcheck.hpp"

node * NULL_NODE = (node*)0;

void VDM_ERROR(const nat0 & errnr, const nat0 & lnr,
	       const String &str1,
	       const String &str2,
	       const String &str3,
	       const String &str4)
{
  char line[128];

  sprintf(line, (char*)(ERRORS.get(errnr)),
	  (char*)str1, (char*)str2,(char*)str3, (char*)str4);

  cerr << (char*)DATNAM << ":" << lnr << ": " << line << endl;

  exit(1);
}

void VDM_WARNING(const nat0 & errnr, const nat0 & lnr,
		 const String &str1,
		 const String &str2,
		 const String &str3,
		 const String &str4)
{
  if (WUNION || (errnr != 1)) // keine union-Warnungen bei entspr. Schalter
    {
      char line[128];
      
      sprintf(line, (char*)(WARNINGS.get(errnr)),
	      (char*)str1, (char*)str2,(char*)str3, (char*)str4);
      
      cerr << (char*)DATNAM << ":" << lnr << ": " << line << endl;
    }
}

typCheck::typCheck()
{
  open_layer(); // globale Typen und values als erste auf den Stack
  {
    nat0seq tmp(nat0(0), nat0(0));
    add_type(tmp, UNKNOWN, UNKNOWN);
    add_type(tmp, T_BASIC, T_BASIC);
    add_type(tmp, T_NAT1, T_NAT1);
    add_type(tmp, T_NAT, T_NAT);
    add_type(tmp, T_INT, T_INT);
    add_type(tmp, T_RAT, T_RAT);
    add_type(tmp, T_REAL, T_REAL);
    add_type(tmp, T_BOOL, T_BOOL);
    add_type(tmp, T_CHAR, T_CHAR);
    add_type(tmp, T_TOKEN, T_TOKEN);
    add_type(tmp, T_NIL, T_NIL);
    add_type(tmp, DONTCARE, DONTCARE);
    add_type(tmp, NOPAT, NOPAT);
    add_type(tmp, T_UNDEFINED, T_UNDEFINED);
    tmp = nat0seq(nat0(1), nat0(0));
    tmp.add(UNKNOWN);
    nat0seq tmp2 = nat0seq(nat0(2), nat0(0));
    tmp2.add(UNKNOWN);
    tmp2.add(UNKNOWN);
    add_type(tmp, T_EMPTYSET, T_EMPTYSET);
    add_type(tmp, T_EMPTYSEQ, T_EMPTYSEQ);
    add_type(tmp2, T_EMPTYMAP, T_EMPTYMAP);
    /*  add_type(tmp, T_PROD, String("PRODUCT"));
	add_type(tmp, T_SET, String("SET"));
	add_type(tmp, T_QUOTE, String("QUOTE"));
	add_type(tmp, T_TAG, String("TAG"));
	add_type(tmp, T_COMP, String("COMPOSE"));
	add_type(tmp, T_ID, String("TYPID"));
	add_type(tmp, T_FIELD, String("FIELD"));
	add_type(tmp, T_UNION, String("UNION"));
	add_type(tmp, T_NIL, String("NIL"));
	add_type(tmp, T_SEQ, String("SEQ"));
	add_type(tmp, T_SEQ1, String("SEQ1"));
	add_type(tmp, T_MAP, String("MAP"));
	add_type(tmp, T_INMAP, String("INMAP"));
	add_type(tmp, T_PFUNC, String("PART_FUNC"));
	add_type(tmp, T_TFUNC, String("TOT_FUNC"));
	add_type(tmp, T_VAR, String("TYP_VAR"));
	add_type(tmp, T_POLY, String("POLY_FUNC"));
	add_type(tmp, VAR, String("VARIABLE"));
	add_type(tmp, DONTCARE, String("DONTCARE"));
	add_type(tmp, VALUE, String("VALUE"));
	add_type(tmp, PAT_LIST, String("PAT_LIST"));
	add_type(tmp, NOPAT, String("NO_PATTERN"));
	add_type(tmp, T_PARAMS, String("PARAMETER"));
	add_type(tmp, MULTI_PAT, String("MULTI_PATTERN"));*/
  } // weitere type-defs durch Anwender in type-definitions
  ERRORS.add(String("Typename '%s' overloads tag-field"));
  ERRORS.add(String("Multiple declaration of typename '%s'"));
  ERRORS.add(String("Tag-field '%s' overloads typename or other tag-field"));
  ERRORS.add(String("Multiple declaration of field '%s' in compose-type"));
  ERRORS.add(String("Incorrect use of tag-field '%s'"));
  ERRORS.add(String("Type-variable '%s' is not declared or not in polym. function"));
  ERRORS.add(String("Multiple declaration of type-variable '%s'"));
  ERRORS.add(String("Operations are not available in this version"));
  ERRORS.add(String("Function-identifier '%s' not the same as declared"));
  /*10*/ERRORS.add(String("Function '%s': Number of parameters differs from\n number of declared types"));
  ERRORS.add(String("Typename '%s' not expectet"));
  ERRORS.add(String("Too many subparts of Parameter '%s'"));
  ERRORS.add(String("Not enough subparts given of Parameter '%s'"));
  ERRORS.add(String("Parameter '%s' doesn't match declared type"));
  ERRORS.add(String("Tag-field '%s' isn't same as declared"));
  ERRORS.add(String("Parameter '%s' isn't in declared union-type"));
  ERRORS.add(String("Tag-field '%s' isn't declared"));
  ERRORS.add(String("Function '%s': body-return differs from declared return-type"));
  ERRORS.add(String("Function '%s': pre-condition not boolean"));
  /*20*/ERRORS.add(String("Function '%s': post-condition not boolean"));
  ERRORS.add(String("No set-union with type '%s'"));
  ERRORS.add(String("No seq-concat with type '%s'"));
  ERRORS.add(String("Only one subpart of '%s' specified"));
  ERRORS.add(String("Type-mismatch in %s"));
  ERRORS.add(String("Multiple declaration of function-name '%s'"));
  ERRORS.add(String("'%s' not defined"));
  ERRORS.add(String("No hd-operation on empty sequences"));
  ERRORS.add(String("No tl-operation on empty sequences"));
  ERRORS.add(String("Non-disjunct recursive union-type"));
  /*30*/ERRORS.add(String("No base-case for recursivly defined type"));
  ERRORS.add(String("Non-unique field-identifier '%s' in record-modification"));
  ERRORS.add(String("Unknown field-identifier in record-modification '%s'"));
  ERRORS.add(String("Polymorphic function instantiation '%s': Number of types differs from nunber of declared type-variables"));
  ERRORS.add(String("Compose-typ with tag '%s' differs from prev. declared compose-typ"));
  ERRORS.add(String("Parse error before '%s'"));
  ERRORS.add(String("Type mismatch in %s\nexpected type kind       : '%s'\nreceived (inspect unions): '%s'"));

  WARNINGS.add(String("Non-compatible union-types possible in %s"));
  WARNINGS.add(String("Assume correct type in %s"));
  WARNINGS.add(String("Use of empty seq where seq1-typ expected"));
  WARNINGS.add(String("Possibly non-compatible types of numbers used%s"));
  WARNINGS.add(String("hd may fail due to possible empty sequence"));
  WARNINGS.add(String("tl may fail due to possible empty sequence"));
  WARNINGS.add(String("Different operand-types in %s"));
  WARNINGS.add(String("No '%s'-type in is_expression, result will be 'false'"));
  WARNINGS.add(String("Quote-literal '%s' not used in type-definitions"));

  LEXEMS.add(String("UNKNOWN"));
  LEXEMS.add(String("T_BASIC"));
  LEXEMS.add(String("nat1"));
  LEXEMS.add(String("nat"));
  LEXEMS.add(String("int"));
  LEXEMS.add(String("rat"));
  LEXEMS.add(String("real"));
  LEXEMS.add(String("bool"));
  LEXEMS.add(String("char"));
  LEXEMS.add(String("token"));
  LEXEMS.add(String("nil"));
  LEXEMS.add(String("dontcare(-)"));
  LEXEMS.add(String("no pattern"));
  LEXEMS.add(String("undefined"));
  LEXEMS.add(String("empty-set"));
  LEXEMS.add(String("empty-seq"));
  LEXEMS.add(String("empty-map"));
  LEXEMS.add(String("product"));
  LEXEMS.add(String("set"));
  LEXEMS.add(String("quote"));
  LEXEMS.add(String("tag-field"));
  LEXEMS.add(String("compose"));
  LEXEMS.add(String("typ-id"));
  LEXEMS.add(String("compose-field-id"));
  LEXEMS.add(String("union"));
  LEXEMS.add(String("seq"));
  LEXEMS.add(String("seq1"));
  LEXEMS.add(String("map"));
  LEXEMS.add(String("inmap"));
  LEXEMS.add(String("partial-function"));
  LEXEMS.add(String("total-function"));
  LEXEMS.add(String("typ-variable"));
  LEXEMS.add(String("polymorph-function"));
  LEXEMS.add(String("variable"));
  LEXEMS.add(String("value"));
  LEXEMS.add(String("pattern-list"));
  LEXEMS.add(String("parameters"));
  LEXEMS.add(String("multi-pattern"));
  LEXEMS.add(String("total-function"));
  LEXEMS.add(String("set-typ-variable"));
  LEXEMS.add(String("seq-typ-variable"));
  
}

typCheck::typCheck(const typCheck & t)
{
  operator = (t);
}

nat0 typCheck::add_type(const typerec & tr)
{
  return TS.add(tr);
  /*if(PARSE != FIRST) // Rekursivitaet bei Typen des 2. Durchgangs pruefen
    {
    if(get_type(t) == T_UNION)
    flat_union(t);
    base_case(t);
    mark_type(t, BASE);
    } nicht mehr noetig, da alle Typen global im ersten Durchgang erstellt sind*/
}

t_typ typCheck::copy_type(const t_typ & t)
{
  if(!is_marked_union(t))
    {
      mark_union(t);
      nat0seq tmp;
      for(nat0 i = 1; i <= len_subtypes(t); i++)
	tmp.add(copy_type(get_subtype(t, i)));
      unmark_union(t);
      t_typ typ = add_type(get_all(t));
      replace_subtypes(typ, tmp);
      return typ;
    }
  else return t;
}

void typCheck::proof_ids(void)  /* Typnamen-Deklaration ueberpruefen */
{
  for(nat0 i = OWN_TYPES; i <= TS.len(); i++)
    if((get_type(i) == T_ID) && (len_subtypes(i) == 0))
      VDM_ERROR(26, get_adress(i)->linenr, get_typename(i));
}

bool typCheck::flat_union(t_typ expression)
{
  t_typ kind = get_type(expression);

  if (VDM_TRACE)
    cout << "\nin flat_union mit: " << (char*)get_typename(expression);

  while((kind == T_ID) || (kind == T_FIELD))
    {
      expression = get_subtype(expression,1);  // Typ darf dafr nicht UNKNOWN sein
      kind = get_type(expression);
    }

  if(!is_marked_union(expression))  // Ende-Bedingung fuer Rekursions-Abstieg
    {
      if(kind == T_UNION)
	{
	  mark_union(expression);
	  for(nat0 i = len_subtypes(expression); i > 0; i--)
	    {
	      if(!flat_union(get_subtype(expression, i)))
		{
		  unmark_union(expression);
		  return bool(0);
		}
	    }
	  unmark_union(expression);
	  return bool(1);
	}
      else return bool(1);
    }
  else return bool(0);
}

void typCheck::make_flat(void)   /* Unions flachmachen */
{
  for(nat0 i = OWN_TYPES; i <= TS.len(); i++)
    if(get_type(i) == T_UNION)
      {
	if (VDM_TRACE)
	  cout << "\ninitial make_flat mit: " << i << "-> " << (char*)get_typename(i);
	if(!flat_union(i))
	  VDM_ERROR(29, get_adress(i)->linenr, "");
      }
}

bool typCheck::base_case(t_typ expression)
{
  t_typ kind = get_type(expression);

  if (VDM_TRACE)
    cout << "\nin base_case mit " << expression << ":" << (char*)get_typename(expression);

  while((kind == T_ID) || (kind == T_FIELD))
    {
      expression = get_subtype(expression,1);  // Typ darf dafr nicht UNKNOWN sein
      kind = get_type(expression);
    }

  if(!is_marked_union(expression))  // Ende-Bedingung fuer Rekursions-Abstieg
    {
      if(kind == T_UNION)
	{
	  for(nat0 i = len_subtypes(expression); i > 0; i--)
	    {
	      if(is_marked_type(get_subtype(expression, i)) == NEW) // neuer Typ
		{
		  if(base_case(get_subtype(expression, i)))
		    {
		      mark_type(get_subtype(expression, i), BASE);
		      return bool(1);
		    }
		  //else mark_type(get_subtype(expression, i), NOBASE);
		}
	      else if(is_marked_type(get_subtype(expression, i)) == BASE)
		return bool(1); // schon geprfter Typ hat base_case
	    }
	  return bool(0);
	}
      else
	{
	  if((kind != T_SET) && (kind != T_SEQ) && (kind != T_MAP) && (kind != T_INMAP))
	    {
	      mark_union(expression);
	      for(nat0 i = len_subtypes(expression); i > 0; i--)
		{
		  if(is_marked_type(get_subtype(expression, i)) == NEW) // neuer Typ
		    {
		      if(!base_case(get_subtype(expression, i)))
			{
			  unmark_union(expression);
			  //mark_type(get_subtype(expression, i), NOBASE);
			  return bool(0);
			}
		    }
		  else if(is_marked_type(get_subtype(expression, i)) == NOBASE)
		    {
		      unmark_union(expression);
		      return bool(0);
		    }
		}
	      unmark_union(expression);
	    }
	  return bool(1);
	}
    }
  else return bool(0);
}

void typCheck::proof_base(void)  /* Rekursionen ueberpruefen */
{
  output();
  for(nat0 i = OWN_TYPES; i <= TS.len(); i++)
    {
      if (VDM_TRACE)
	cout << "\ninitial proof_base mit: " << i << "-> " << (char*)get_typename(i);

      if (i == 59) output();

      if(/*(get_type(i) != T_ID) && (get_type(i) != T_FIELD) &&*/ !base_case(i))
	VDM_ERROR(30, get_adress(i)->linenr, "");
      mark_type(i, BASE);
    }
}

void typCheck::output(void)
{
  nat0 i,j;
  if (VDM_TRACE)
    {
      for (i = 1; i <= TS.len(); i++)
	{
	  cout << "\nTypnr " << i << ":  " << (char*)get_typename(i) << " == ";
	  for (j = 1; j <= TS.get(i).tg.len(); j++)
	    {
	      cout << (TS.get(i).tg).get(j) << ":" << (char*)get_typename((TS.get(i).tg).get(j)) << "  ";
	    }
	  cout << "  Typart " << get_type(i) << ":" << (char*)get_kindname(i);
	  //cout << "  Adr.: " << TS.get(i).ta;
	  cout << "  Mark: " << TS.get(i).type_mark;
	}
      cout << "\nSymbols:";
      for (i = 1; i <= LEXEMS.len(); i++)
	cout << "\n" << (char*)LEXEMS.get(i);
      cout << "\n";
      cout << "\nyylino: " << yylino;
    }
}

typCheck & typCheck::operator = (const typCheck & t)
{
  TS = t.TS;
  // IDS = t.IDS;
  return *this;
}

bool typCheck::operator == (const typCheck & t)
{
  return bool((TS == t.TS) /*&& (IDS == t.IDS)*/);
}

bool typCheck::operator != (const typCheck & t)
{
  return bool(!(operator == (t)));
}
