#ifndef TYPCHECK_H
#define TYPCHECK_H

#include <iostream.h>
#include "parse.hpp"
#include "types.hpp"
#include "typesrec.hpp"
#include "stringba.hpp"
//#include "ids.hpp"

#define TS_LAST (TS.get(TS.len()))

extern nat0seq LINENR/*(nat0(100), nat0(25))*/;
extern Basicarray_String ERRORS/*(nat0(100))*/;
extern Basicarray_String WARNINGS/*(nat0(100))*/;
extern String DATNAM;
extern int PARSE;
extern String POLY;

extern bool MATH;
extern bool WUNION;
extern bool VDM_TRACE;

extern node * NULL_NODE; // verhindert Intitallisierungsprobleme

extern void VDM_ERROR(const nat0 & errnr, const nat0 & lnr, const String &str1 = "",
		      const String & str2 = "", const String & str3 = "", const String & str4 = "");

extern void VDM_WARNING(const nat0 & errnr, const nat0 & lnr, const String &str1 = "",
		 const String & str2 = "", const String & str3 = "", const String & str4 = "");

class typCheck
{
 types TS;
public:
 typCheck();
 typCheck(const typCheck & t);

 void open_layer(void)
 {
  TS.push();   // Das Ende der letzten Typ-Ebene markieren
//  IDS.push();  // ebenso fr ID-Stack
 }

 void close_layer(bool nur_id = 1)
 {
  TS.pop(nur_id);   // letzte Typ-Ebene vom Stack
//  IDS.pop();  // ebenso fr ID-Stack
 }

 typerec & get_all(const nat0 & t)
 {
  return TS.get(t);
 }

 nat0 add_type(const typerec & tr);

 t_typ copy_type(const t_typ & t);

 nat0 add_type(const nat0seq & ns,
	       const t_typ & t = UNKNOWN,
	       const nat0 & s = UNKNOWN,
	       node * adr = NULL_NODE,
	       const short & st = UNKNOWN)
 {
  return TS.add(ns, t, s, adr, st);
 }

 t_typ seek_type(const String & s)
 {
  return TS.seek(s); // Typen haben nur bei type-defs eine ID
 }                   // es ist noch unklar, ob composite-type
		     // Ausnahme bilden mit ihrem tag-field

 t_typ seek_tagfield(const String & s)
 {
  return TS.seektag(s); // Typen haben nur bei type-defs eine ID
 }                   // es ist noch unklar, ob composite-type
		     // Ausnahme bilden mit ihrem tag-field

 t_typ seek_identifier(const String & s)
 {
  return TS.seekid(s); // Typen haben nur bei type-defs eine ID
 }                   // es ist noch unklar, ob composite-type
		     // Ausnahme bilden mit ihrem tag-field

 t_typ layer_seek_type(const String & s)
 {
  return TS.layer_seek(s); // Typen haben nur bei type-defs eine ID
 }                   // es ist noch unklar, ob composite-type
		     // Ausnahme bilden mit ihrem tag-field

 t_typ get_type(const nat0 & t)
 {
  return TS.get(t).tk;
 }

 void put_type(const t_typ & t, const t_typ & st)
 {
  TS.get(t).tk = st;
 }

 void del_type(const t_typ & t)
 {
  TS.del(t);
 }

 nat0 add_subtype(const t_typ & t, const t_typ & st)
 {
  return TS.get(t).tg.add(st);
 }

 nat0 un_subtype(const t_typ & t, const t_typ & st)
 {
  if(!TS.get(t).tg.isin(st))
   return TS.get(t).tg.add(st);
  return 0;
 }

 t_typ get_subtype(const nat0 & t, const nat0 & st)
 {
  return TS.get(t).tg.get(st);
 }

 void put_subtype(const nat0 & t, const nat0 & st, const t_typ & typ)
 {
  TS.get(t).tg.put(st, typ);
 }

 void insert_subtype(const nat0 & t, const nat0 & st, const t_typ & typ)
 {
  TS.get(t).tg.insert(st, typ);
 }

 void del_subtype(const t_typ & t, const t_typ & st)
 {
  TS.get(t).tg.del(st);
 }

 nat0seq & get_subtypes(const nat0 & t)
 {
  return TS.get(t).tg;
 }

 void replace_subtypes(const nat0 & t, const nat0seq & rt)
 {
  TS.get(t).tg = rt;
 }

 void add_subtypes(const nat0 & t, const nat0seq & at)
 {
  for(nat0 i = 1; i <= at.len(); i++)
   TS.get(t).tg.add(at.get(i));
 }

 nat0 len_subtypes(const t_typ & t)
 {
  return TS.get(t).tg.len();
 }

 String & get_typename(const t_typ & t)
 {
  return LEXEMS.get(TS.get(t).tn);
 }

 nat0 & get_typenamenr(const t_typ & t)
 {
  return TS.get(t).tn;
 }

 String & get_kindname(const t_typ & t)
 {
  return LEXEMS.get(TS.get(t).tk);
 }

 void put_typename(const nat0 & t, const nat0 & s)
 {
  TS.get(t).tn = s;
 }

 void mark_union(const t_typ t)
 {
  TS.get(t).union_mark = bool(1);
 }

 void unmark_union(const t_typ t)
 {
  TS.get(t).union_mark = bool(0);
 }

 bool is_marked_union(const t_typ t)
 {
  return TS.get(t).union_mark;
 }

 void mark_type(const t_typ t, const int & k)
 {
  TS.get(t).type_mark = k;
 }

 int is_marked_type(const t_typ t)
 {
  return TS.get(t).type_mark;
 }

 nat0 add_id(const nat0seq & ns,
	     const t_typ & t = UNKNOWN,
	     const nat0 & s = UNKNOWN,
	     node * adr = NULL_NODE,
	     const short & st = UNKNOWN)
 {
  return add_type(ns, t, s, adr, st);
 }

 nat0 add_value(const nat0seq & ns,
		const t_typ & t = UNKNOWN,
		const nat0 & s = UNKNOWN,
		node * adr = NULL_NODE,
		const short & st = UNKNOWN)
 {
  return add_type(ns, t, s, adr, st);
 }

 short get_state(const t_typ & t)
 {
  return TS.get(t).state;
 }

 void put_state(const t_typ & t, const short & st)
 {
  TS.get(t).state = st;
 }

 short get_substate(const nat0 & t, const nat0 & st)
 {
  return TS.get(TS.get(t).tg.get(st)).state;
 }

 node * get_adress(const t_typ & t)
 {
  return TS.get(t).ta;
 }

 node * get_prec(const t_typ & t)
 {
  return TS.get(t).ta->prec;
 }

 bool base_case(t_typ expression);
 bool flat_union(t_typ expression);

 void proof_ids(void);   /* Typnamen-Deklaration ueberpruefen */
 void make_flat(void);   /* Unions flachmachen */
 void proof_base(void);  /* Rekursionen ueberpruefen */

 void output(void);

 typCheck & operator = (const typCheck & t);
 bool operator == (const typCheck & t);
 bool operator != (const typCheck & t);
};


extern typCheck TC;

#endif

