/*****************************************************************************
* File:     types.hpp                                                        *
* Funktion: Sammlung von allgemein wichtigen consts, defines und typedefs.   *
*           Wichtige Funktions-defines sind:                                 *
*           - assertions                                                     *
*           - trace                                                          *
*                                                                            *
*****************************************************************************/

#ifndef TYPES_H
 #define TYPES_H

 #include <stdio.h>

 // #define WITHTRACE  // Schalter fr Traceausgabe nach stderr
 // #define NDEBUG     // Schalter fr das bergehen von assertions; ist hier
		       // noch nicht optimal gelst, da auch Eingabefehler
		       // und Datei-Fehler ber assert() abgefangen werden,
		       // nicht nur Programmierfehler. Man sollte sie also
		       // nicht ausschalten!
 // #define WITHDOSIO  // Schalter fr DOS-Eingabe/Ausgabe. Aufrufe von
		       // Editor, Listdlg, Message und Window sollten
		       // abhngig von diesem Schalter compiliert werden.

 #ifdef WITHDOSIO
  #include <strstream.h>        // fr Serialisierung von Ausgaben
  #define LISTWIN 41, 1, 79, 25 // Liste-Window-Koordinaten
  #define ABBWIN 1, 1, 39, 7    // Abbildung-Window-Koordinaten
  #define WERTWIN 1, 8, 39, 20  // Werteeingabe-Window-Koordinaten
  #define MESSWIN 1, 21, 39, 25 // Message-Window-Koordinaten
 #endif

 // Tastatur - CHARs - Synonyme
 const CTRL = 0,            // erstes char bei Steuerzeichen
       CUP  = 72,           // Cursor up
       CDN  = 80,           // Cursor down
       CLT  = 75,           // Cursor left
       CRT  = 77,           // Cursor right
       PUP  = 73,           // Page up
       PDN  = 81,           // Page down
       CTRLPUP = -124,      // Ctrl Page up
       CTRLPDN =  118,      // Ctrl Page down
       ESC  = 27,           // Escape
       RETURN_AH = 13,      // Enter
       F3     = 61,         // Funktionstaste F3
       HOME   = 71,         // Pos1
       END_AH = 79,         // Ende
       BS     = 8,          // Backspace
       DEL    = 83,         // Entf
       LEER   = 32,         // Leertaste
       INS    = 82,         // Einf
       CTRLY  = 25,         // Strg-Y
       TAB    = 9,          // Tabulator
       SHIFTTAB = 15;       // Shift-Tabulator

 // Typen
 const INT_AH           = 1,
       LONGINT          = 2,
       SHORTINT         = 3,
       UNSIGNEDINT      = 4,
       UNSIGNEDLONGINT  = 5,
       UNSIGNEDSHORTINT = 6,
       FLOAT            = 7,
       DOUBLE           = 8,
       LONGDOUBLE       = 9,
       STRING           = 10,
       SIGNED           = 11,
       UNSIGNED         = 12;


 #ifndef NDEBUG
  #include <stdlib.h>        // fr exit() in assert()
  #include <iostream.h>      // fr cerr in assert

  /* Funktionsname: assert
   * Funktion: Ausgabe von Programmzeile nach stderr und Rckgabe einer
   *           Fehlernr. bei Auftreten eines Fehlers.
   * Parameter   Typ         Bedeutung
   * expr      | Bedingung | Ein Bedingungsausdruck, der Fehler ausgrenzt.
   *           |           | Wird er zu false ausgewertet, starten die
   *           |           | Fehlermanahmen (s.o.).
   * excode    | int       | Fehlernummer (z.B. zum Auffinden einer Fehler-
   *           |           | meldung in einer entsprechenden Datei.
   * Return:   | int       | Fehlernummer (s.o.)
   * Bemerkung: noch keine schne Fehlerbehandlung (exit). */

  #define assert(expr, excode)\
   if (!(expr))\
   {\
    cerr << "\nAssertion failed: (" << #expr << ") in "\
    << __FILE__ << " Zeile " << __LINE__;\
    exit(excode);\
   }

 #else   // ist NDEBUG-Schalter gesetzt, assert() zu "nichts" auswerten.
  #define assert(expr, excode)
 #endif



 #ifdef WITHTRACE
  #include <stdio.h>               // fr sprintf()
  #include <iostream.h>            // fr cerr
  const  adr__count = 10,  // max. Anzahl Adressen im Ringspeicher
	 adr__size  = 10;  // Adressen-Zeichenkettenlnge
  // 10 stimmt nur bei Adressen, die hexa-mig bis 8 Ziffern betragen (far),
  // da sie im Format FFFF:FFFF plus Endekennung ausgegeben werden.
  const  adr__mem = adr__count * adr__size;
  static char typ__adrbuffer[adr__mem]; // Adressen-char-Array fr trace-Ausgabe
  static int typ__adrptr = adr__mem - adr__size; // Anfangspos. im Adressen-Array

  /* Funktionsname: filladr
   * Funktion: Implementiert einen Ringspeicher fr Adressen-Zeichenketten.
   *           Der ist ntig, weil cout keine Adressen ausgibt (nur Inhalte)
   *           und pro trace-Aufruf eine unbekannte Anzahl von Adressen
   *           interessieren knnte (hier beschrnkt auf zehn, da der Ring-
   *           speicher lteste Adressen jeweils berschreibt).
   * Parameter   Typ         Bedeutung
   * a         | void *    | Die Adresse, die ausgegeben werden soll. Sie
   *           |           | wird zunchst in den Ringspeicher ausgegeben.
   * Return:   | char *    | Zeiger auf Adressen-Zeichenkette fr cout
   * Bemerkung: Ich wei nicht mehr, ob man nicht auch mit dem Speicherplatz
   *            einer Adresse ausgekommen wre, es kommt auf die Reihenfolge
   *            der Abarbeitung an (ausprobieren).*/

  static char * filladr(void * a)
  {
   typ__adrptr = (typ__adrptr + adr__size) % adr__mem; // Bewegung im Ringsp.
   sprintf((typ__adrbuffer + typ__adrptr), "%p", a);   // Speichern der Adr.
   return (typ__adrbuffer + typ__adrptr);              // Anfangsadresse
  }

  // verschiedene einfache trace-Ausgabefunktionen;
  // 1. fr member-Funktionen:
  #define TRACE(x)\
   cerr << "\n" << "// " << this << ": " << x;
  // 2. fr friend und normale Funktionen:
  #define OUT(x)\
   cerr << "\n" << "//           " << x;
  // 3. zum optischen Eingrenzen der Ausgaben:
  #define GG(x)\
   "\"" << x << "\" "
  // 4. zum Ausgeben von Adressen mit cout:
  #define ADR(x)\
   filladr((void*)&x)  // Funktion erwartet einen typlosen Zeiger auf den Wert

 #else   // ist WITHTRACE nicht gesetzt, alle Funktionen zu "nichts" auswerten
  #define TRACE(x)
  #define OUT(x)
  #define GG(x)
  #define ADR(x)
 #endif

 #define bool int                // nur fr optische Zwecke, kein Typcheck
 typedef unsigned long int nat0; // natrliche Zahlen mit Null
 typedef unsigned int pixadr;    // Pixel-Adressen-Typ
 typedef float num;              // Fliekommazahlen

 typedef void (*fpointer) (void *, void *); // Funktionszeiger-Typ und
 typedef void * argtype;         // Argumentzeiger-Typ, beide fr Adressen-
 // bergabe einer Funktion als Parameter (leider nicht fr member-Funktion)

#endif