/*****************************************************************************
* template: Basicarray                                                       *
* Funktion: Diese template implementiert ein dynamisches Array.              *
*           Die zu speichernden Objekte knnen, so wie sie sind, angehngt   *
*           oder berschrieben werden, ihr Speicher wird automatisch am Ende *
*           freigegeben (im Unterschied zu Typlosen-Zeiger-Varianten).       *
*                                                                            *
*           Es werden jeweils Kopien der zu speichernden Objekte angelegt,   *
*           die Originale werden nicht angefat. Ist dieser Schutz unntig,  *
*           empfiehlt sich beim Aufrufen von put() oder add(), die Objekt-   *
*           konstruktion quasi als Parameter dieser Funktionen vorzunehmen,  *
*           damit der Speicher des Originals gleich nach dem Aufruf wieder   *
*           frei wird.                                                       *
*                                                                            *
*           Einige typische Operationen einer VDM-Liste sind schon hier im-  *
*           plementiert, da sie auch fr Arrays sinnvoll sein knnen.        *
*                                                                            *
*           Nur der direkte "C"-Zugriff auf das Feld arbeitet mit der Null   *
*           als Index-Basis, alle anderen Funktionen setzen VDM-mig eine   *
*           Eins voraus.                                                     *
*                                                                            *
*****************************************************************************/

#ifndef BASICARRAY_TYPES
 #define BASICARRAY_TYPES

 #include <stdio.h>
 #include <stdlib.h>
 #include "types.hpp" // wichtige const, define und typedef

 #include "string.hpp"
 #include "parse.hpp"
 #include "nat0ba.hpp"
 #include "node.hpp"

 #define nat0seq Basicarray_nat0
 const NEW = -1,
       NOBASE = 0,
       BASE = 1;

 class typerec
 {
  public:
  nat0seq tg;
  t_typ tk;
  nat0 tn;
  node * ta;
  short state;
  bool union_mark;
  int type_mark;

  typerec(const nat0seq & ns, const t_typ & t,
	    const nat0 & s, node * adr, const short & st)
  {
   tg = ns;
   tk = t;
   tn = s;
   ta = adr;
   state = st;
   union_mark = bool(0);
   type_mark = -1;
  }

  typerec(const typerec & t)
  {
   operator = (t);
  }

  typerec & operator = (const typerec & t)
  {
   tg = t.tg;
   tk = t.tk;
   tn = t.tn;
   ta = t.ta;
   state = t.state;
   union_mark = t.union_mark;
   type_mark = t.type_mark;
   return *this;
  }
  bool operator == (const typerec & t)
  {
   return bool((tg == t.tg) && (tk == t.tk) &&
	       (tn == t.tn) && (ta == t.ta) &&
	       (state == t.state) && (union_mark == t.union_mark) &&
	       (type_mark == t.type_mark));
  }
  bool operator != (const typerec & t)
  {
   return bool(!(operator == (t)));
  }
 };

 class Basicarray_types {
 protected:
  typerec ** feld;          // ein Feld fr Objekt-Zeiger
  nat0 feldsize;      // Gesamt-Gre
  nat0 feldcount;     // aktuelle Belegungs-Gre
  nat0 delta;         // Zuwachs bei Vergrerung

  typerec ** getmem(nat0 size); // zum Speicheranlegen frs Objektzeigerfeld
			  // size: Gesamt-Gre
			  // return: Zeiger auf den neuen Speicherbereich

  void releasemem(typerec ** tf); // Zum Freigeben des Feld-Speichers
			    // tf: freizugebendes Feld

  void copyliste(const Basicarray_types & a); // Zum Kopieren des Felds a in feld
					   // a: zu kopierendes Feld

  nat0 indexcheck(nat0 i) const; // berprft Feld-Zugriff auf Gltigkeit
				 // i: Zu prfender Index
				 // return: berprfter Index


 public:
  Basicarray_types (nat0 l = 10, // Konstruktor (inkl. default). l: Gesamt-Gre
	      nat0 d = 5); // d: Zuwachsrate bei Grenzereichung

  Basicarray_types (const Basicarray_types &a); // Copy - Konstruktor
				       // a: zu bernehmendes Feld

  ~Basicarray_types (); // Destruktor

  nat0 add (const typerec & Daten); // Einfgeroutine (nicht VDM!) am Ende des Felds
			      // Daten: anzuhngender Eintrag
			      // return: zugehriger Index-Wert

  nat0 del (const typerec & Daten, // Lsche Eintrag (nicht VDM)
		    int deltyp);     // Daten: zu lschender Eintrag
				     // deltyp: noch nicht belegt, dient z.Z.
				     // der Unterscheidung vom folgenden del
				     // mit T = nat0
				     // return: wenn vorhanden, Pos.; sonst 0

  nat0 del (nat0 i); // Lschen des Eintrags an der Position i (nicht VDM!)
		     // i: an dieser Position soll gelscht werden (ab 1)
		     // return: z.Z. i
		     // Anmerkung: beide del()-Funktionen verkleinern die
		     // Gesamt-Gre (size) des Feldes nicht

  void put ( nat0 i, const typerec & Daten); // berschreibroutine ( l+[i->e] )
				       // i: zu berschreibende Pos. (ab 1)
				       // Daten: zu speichernder Eintrag

  typerec & get (nat0 i) const; // Zugriffsroutine ( l(i) )
			  // i: Diese Position soll ausgelesen werden (ab 1)
			  // return: gefundener Eintrag (hier noch als Refe-
			  // renz aus Performance-Grnden, sollte Kopie sein!)

  void insert(nat0 j, const typerec & Daten); // Einfgen von Items
					// j: hier wird eingefgt (ab 1)
					// Daten: das wird eingefgt

  typerec & operator [] (nat0 i); // Fr Array-Fans ( z.B.: feld[i] = e;)
			    // i: Index (ab 0 !)
			    // return: Referenz auf Eintrag (fr Updates)

  Basicarray_types & operator = (const Basicarray_types & a); // Zuweisung
							// a: zugewiesen
							// return: Referenz
							// auf sich selbst

  nat0 len() const; // Arraynlnge ( len l )
		    // return: Lnge des Felds

  friend bool operator == (const Basicarray_types & a1, // Vergleich ( l1 = l2 )
			   const Basicarray_types & a2); // a1, a2: zu vergleichen
						     // return: gleich / ungl.
  friend bool operator != (const Basicarray_types & a1, // Ungleich ( l1 /= l2 )
			   const Basicarray_types & a2); // s.o.

  bool isin(const typerec &e) const; // Test auf Vorhandensein
				       // e: zu testendes Element
				       // return: vorhanden / nicht vorhanden

  bool isempty() const         // Test auf leeres Feld
  { return bool(len() == 0); } // return: leer / nicht leer

  void foreach(fpointer f, argtype arg); // All-Quantor (so nicht VDM)
					 // f: Funktionsname, leider n. global
					 // arg: void-Zeiger auf Argumente

  nat0 getsize() const; // ermittelt Array-Gre (fr interne Zwecke)
			// return: Gesamt-Gre

  nat0 getdelta() const; // ermittelt Zuwachs-Rate
			 // return: eingestellte Zuwachsrate

  void setdelta(nat0 d); // ndern der Zuwachsrate
			 // d: neue Zuwachsrate

  void setsize(nat0 l); // ndern der Feld-Gesamt-Gre
			// Achtung, Arrayinhalt geht z.Z. immer verloren!
			// l: neue Gesamt-Gre
 };

extern bool operator == (const Basicarray_types & a1, // Vergleich ( l1 = l2 )
			   const Basicarray_types & a2); // a1, a2: zu vergleichen
						     // return: gleich / ungl.
extern bool operator != (const Basicarray_types & a1, // Ungleich ( l1 /= l2 )
			   const Basicarray_types & a2); // s.o.

#endif
