%{
//#define YY_DECL int yylex YY_PROTO(( typeStack & tp ))
#include "typcheck.hpp"

#define YYSTYPE node *
#include "y.tab.h"

extern int yywrap();
extern int yyerror(char * error);

node * install_resid();
node * install_id();
node * install_distid();
node * install_char();
node * install_text();
node * install_num();
node * install_typevar();

void newline_count(void);

%}
comment_vdm    "--".*\n
delim          [ \n\t]
ws             {delim}+
letter         [A-Za-z]
characters     [\,\:\;\=\(\)\|\-\[\]\{\}\+\/\<\>\.\*\&\^\_\'\@\~\ ]
all_char       [^\"]
digit          [1-9]
zero           0
id             {letter}({letter}|{digit}|{zero}|"_"|"'")*
quotelit       "<"{letter}({letter}|{digit}|"_")*">"
charlit        "'"({digit}|{zero}|{letter}|{characters}|\")"'"
/*textlit        \"({digit}|{zero}|{letter}|{characters}|\"\")*\"*/
textlit        \"({all_char}|\"\")*\"

nat1_lit       {digit}({digit}|{zero})*
digits         ({digit}|{zero})+
numlit         ({digits}(\.{digits}?)([Ee][\+\-]?{digits})?)|({digits}[Ee][\+\-]?{digits})
typevarid      @{id}
seq_of         seq{ws}of
seq1_of        seq1{ws}of
set_of         set{ws}of
in_set         in{ws}set
not_in_set     not{ws}in{ws}set
is_not_yet_defined     is{ws}not{ws}yet{ws}defined
etc            \,{delim}*\.\.\.{delim}*\,
else           .

%%


"/*"           {register int c;
		for ( ; ; )
		{
		 while ( ((c = yyinput()) != '*') &&
			 (c != EOF) &&
			 (c != '\n') );
		 if ( c == '*' )
		 {
                  while ( (c = yyinput()) == '*' );
		  if ( c == '/' ) break;    /* found the end */
                 }
                 if ( c == EOF )
                 {
                  yyerror( "EOF in comment, \"*/\" not found" );
                  break;
                 }
                 if (c == '\n') yylino++;
                }
	       }

"annotations"  { int c;
                 c = yyinput();
                 while (1) {
                   if (c == EOF) {
                     yyerror( "EOF in annotations, \"end annotations\" not found" );
                     break;
                   }
                   if (c == '\n') ++yylino;
                   if (c != 'e') { c = yyinput(); continue; } c = yyinput();
                   if (c != 'n') { c = yyinput(); continue; } c = yyinput();
                   if (c != 'd') { c = yyinput(); continue; } c = yyinput();
                   if (c != ' ') { c = yyinput(); continue; } c = yyinput();
                   if (c != 'a') { c = yyinput(); continue; } c = yyinput();
                   if (c != 'n') { c = yyinput(); continue; } c = yyinput();
                   if (c != 'n') { c = yyinput(); continue; } c = yyinput();
                   if (c != 'o') { c = yyinput(); continue; } c = yyinput();
                   if (c != 't') { c = yyinput(); continue; } c = yyinput();
                   if (c != 'a') { c = yyinput(); continue; } c = yyinput();
                   if (c != 't') { c = yyinput(); continue; } c = yyinput();
                   if (c != 'i') { c = yyinput(); continue; } c = yyinput();
                   if (c != 'o') { c = yyinput(); continue; } c = yyinput();
                   if (c != 'n') { c = yyinput(); continue; } c = yyinput();
                   if (c != 's') { c = yyinput(); continue; }
                   break;
                 }
               }
{comment_vdm}  {yylino++;}
{seq_of}       {newline_count(); return(SEQ_OF);}
{seq1_of}      {newline_count(); return(SEQ1_OF);}
{set_of}       {newline_count(); return(SET_OF);}
{in_set}       {newline_count(); return(IN_SET);}
{not_in_set}   {newline_count(); return(NOT_IN_SET);}
{is_not_yet_defined}   {newline_count(); return(IS_NOT_YET_DEFINED);}
{etc}          {return(ETC);}
{ws}           {newline_count();}
and            {return(AND);}
as             {return(AS);}
abs            {return(ABS);}
always         {return(ALWAYS);}
be             {return(BE);}
bool           {return(BOOL);}
by             {return(BY);}
card           {return(CARD);}
cases          {return(CASES);}
char           {return(CHAR);}
comp           {return(COMP);}
compose        {return(COMPOSE);}
conc           {return(CONC);}
dcl            {return(DCL);}
def            {return(DEF);}
definitions    {return(DEFINITIONS);}
dinter         {return(DINTER);}
div            {return(DIV);}
do             {return(DO);}
dom            {return(DOM);}
dunion         {return(DUNION);}
elems          {return(ELEMS);}
else           {return(ELSE);}
elseif         {return(ELSEIF);}
end            {return(END_VDM);}
error          {return(ERROR);}
errs           {return(ERRS);}
exists         {return(EXISTS);}
exists1        {return(EXISTS1);}
exit           {return(EXIT);}
exports        {return(EXPORTS);}
ext            {return(EXT);}
false          {return(FALSE);}
floor          {return(FLOOR);}
for            {return(FOR);}
forall         {return(FORALL);}
all            {return(FORALL);}
from           {return(FROM);}
functions      {return(FUNCTIONS);}
get_value      {return(GET_VALUE);}
hd             {return(HD);}
if             {return(IF);}
imports        {return(IMPORTS);}
inset          {return(IN_SET);}
in             {return(IN);}
inds           {return(INDS);}
init           {return(INIT);}
inmap          {return(INMAP);}
instantiation  {return(INSTANTIATION);}
int            {return(INT_VDM);}
inter          {return(INTER);}
inv            {return(INV);}
inverse        {return(INVERSE);}
iota           {return(IOTA);}
lambda         {return(LAMBDA);}
len            {return(LEN);}
let            {return(LET);}
map            {return(MAP);}
merge          {return(MERGE);}
mk_            {return(MK_);}
mod            {return(MOD);}
module         {return(MODULE);}
munion         {return(MUNION);}
mu             {return(MU);}
nat            {return(NAT);}
nat1           {return(NAT1);}
nil            {return(NIL);}
notinset       {return(NOT_IN_SET);}
not            {return(NOT);}
of             {return(OF);}
operations     {return(OPERATIONS);}
or             {return(OR);}
others         {return(OTHERS);}
parameters     {return(PARAMETERS);}
post           {return(POST);}
power          {return(POWER);}
pre            {return(PRE);}
psubset        {return(PSUBSET);}
put_value      {return(PUT_VALUE);}
rat            {return(RAT);}
rd             {return(RD);}
real           {return(REAL);}
rem            {return(REM);}
return         {return(RETURN_VDM);}
reverse        {return(REVERSE);}
rng            {return(RNG);}
seq            {return(SEQ);}
seq1           {return(SEQ1);}
set            {return(SET);}
skip           {return(SKIP);}
st             {return(ST);}
state          {return(STATE);}
subset         {return(SUBSET);}
then           {return(THEN);}
tixe           {return(TIXE);}
token          {return(TOKEN);}
tl             {return(TL);}
to             {return(TO);}
trap           {return(TRAP);}
true           {return(TRUE);}
types          {return(TYPES);}
undefined      {return(UNDEFINED);}
union          {return(UNION);}
using          {return(USING);}
values         {return(VALUES);}
while          {return(WHILE);}
wr             {return(WR);}
RESULT         {return(RESULT);}

is_bool        {return(IS_BOOL);}
is_char        {return(IS_CHAR);}
is_nat         {return(IS_NAT);}
is_nat1        {return(IS_NAT1);}
is_int         {return(IS_INT);}
is_rat         {return(IS_RAT);}
is_real        {return(IS_REAL);}
is_token       {return(IS_TOKEN);}

init_{id}      {yylval = install_resid(); return(INIT_);}
inv_{id}       {yylval = install_resid(); return(INV_);}
is_{id}        {yyless(3); return(IS_NAME);}
mk_{id}        {yyless(3); return(MK_NAME);}
post_{id}      {yylval = install_resid(); return(POST_);}
pre_{id}       {yylval = install_resid(); return(PRE_);}

{id}           {yylval = install_id(); return(ID);}
{quotelit}     {yylval = install_distid(); return(QUOTELIT);}
{charlit}      {yylval = install_char(); return(CHARLIT);}
{textlit}      {yylval = install_text(); return(TEXTLIT);}
{zero}         {yylval = install_num(); return(NAT_LIT);}
{nat1_lit}     {yylval = install_num(); return(NAT1_LIT);}
{numlit}       {yylval = install_num(); return(DEC_LIT);}
{typevarid}    {yylval = install_typevar(); return(TYPEVARID);}

"<="           {return(LESS_EQUAL);}
">="           {return(GREATER_EQUAL);}
"<>"           {return(NOT_EQUAL);}
"->"           {return(PART_FUNC);}
"+>"           {return(TOTAL_FUNC);}
"==>"          {return(OPERATION);}
"::"           {return(TAG);}
"||"           {return(NONDET_STATE);}
"=>"           {return(IMPLY);}
"<=>"          {return(LOGIC_EQUAL);}
"|->"          {return(MAPS);}
"<:"           {return(DOM_RSTR_TO);}
":>"           {return(RNG_RSTR_TO);}
"<-:"          {return(DOM_RSTR_BY);}
":->"          {return(RNG_RSTR_BY);}
"=="           {return(FUNCTION);}
"**"           {return(ITERATE);}
"++"           {return(UPDATE);}
{else}         {return(*yytext);}

%%

void newline_count(void)
{
 int i;
 for (i =0; i < yyleng; i++)
  if (yytext[i] == '\n') yylino++;
}

node * install_resid()
{
 return install_id();
}

node * install_id()
{
 node * tmp = NULL;
 LINENR.add(yylino);
 tmp = new node(NULL, nodeseq(), LEXEMS.add(yytext), NULL);
 assert(tmp > (node*)NULL, 99);
 return tmp;
}

node * install_distid()
{
 return install_id();
}

node * install_char()
{
 return install_id();
}

node * install_text()
{
 return install_id();
}

node * install_num()
{
 return install_id();
}

node * install_typevar()
{
 return install_id();
}







