#!/usr/bin/wish -f
#
# $Id: vdmdb.tcl,v 1.1 1994/01/12 21:57:37 vdm Exp $

set listgeo "20x10"

# -------------------------------------------------------------------------------

proc initdir { } {
    global env
    global sysdir
    global editor

    if [ info exists env(VDMROOT) ] {
        set sysdir $env(VDMROOT)
    } else {
        set sysdir "/usr/lib"
    }

    if [info exists env(EDITOR)] {
        set editor $env(EDITOR)
    } else {
        set editor xedit
    }
}

proc checkdir { } {
    global sysdir
    global dir

    if [ file isdirectory $sysdir ] {
       set dir      "$sysdir/lib/vdmdb-c"
    } else {
       set dir ""
    }
}

# -------------------------------------------------------------------------------

set MessageBoxAns 0
proc MessageBox {Message ButtonNames {Title "Question"} {Aspect "120"}} {
    global MessageBoxAns

    toplevel .question
    wm title .question $Title
    message .question.mess -text $Message -aspect $Aspect
    pack append .question .question.mess  bottom

    foreach ButtonName $ButtonNames {
	button .question.x$ButtonName -text $ButtonName -command "set MessageBoxAns [lsearch $ButtonNames $ButtonName]; destroy .question"
	pack append .question .question.x$ButtonName left
    }

    grab set .question
    tkwait window .question
    return $MessageBoxAns
}

# -------------------------------------------------------------------------------

proc dirDoesNotExists { } {
       MessageBox "VDM database not found, \
please check VDM Class Library installation and \
enviroment variable VDMROOT" "quit" "error" "200"
       exit
}

# -------------------------------------------------------------------------------

initdir
checkdir

label     .title -text "VDM Database Browser"

frame     .topbuttons
button    .topbuttons.ok    -text "quit" -command "destroy ."
label     .topbuttons.title -text "VDM Database Browser" -relief raised
button    .topbuttons.help  -text " help " -command "vdmhelp"
pack      .topbuttons.ok    -side left 
pack      .topbuttons.help  -side right 
pack      .topbuttons.title -fill x 

frame     .database
label     .database.dbt -text "database : "
label     .database.dbn -text "$dir"
pack      .database.dbt -fill x
pack      .database.dbn -fill x

frame     .adts
label     .adts.label  -text "ADT"
scrollbar .adts.scroll -command ".adts.list yview"
listbox   .adts.list   -yscroll ".adts.scroll set" -relief raised -geometry $listgeo -setgrid yes
pack      .adts.label  -side top  -fill both
pack      .adts.list   -side left -fill both
pack      .adts.scroll -side left -fill y

frame     .impl
label     .impl.label  -text "Implementation"
scrollbar .impl.scroll -command ".impl.list yview"
listbox   .impl.list   -yscroll ".impl.scroll set" -relief raised -geometry $listgeo -setgrid yes
pack      .impl.label  -side top
pack      .impl.list   -side left -fill both
pack      .impl.scroll -side left -fill y

frame     .oper
label     .oper.label  -text "Operation"
scrollbar .oper.scroll -command ".oper.list yview"
listbox   .oper.list   -yscroll ".oper.scroll set" -relief raised -geometry $listgeo -setgrid yes
pack      .oper.label  -side top
pack      .oper.list   -side left -fill both
pack      .oper.scroll -side left -fill y

frame     .comp
label     .comp.label  -text "Component"
scrollbar .comp.scroll -command ".comp.list yview"
listbox   .comp.list   -yscroll ".comp.scroll set" -relief raised -geometry $listgeo -setgrid yes
pack      .comp.label  -side top
pack      .comp.list   -side left -fill both
pack      .comp.scroll -side left -fill y

pack      .topbuttons  -side top -fill x
pack      .database    -side top -fill x
pack      .adts        -side left
pack      .impl        -side left
pack      .oper        -side left
pack      .comp        -side left

# -------------------------------------------------------------------------------

wm title    . "VDM DB"
wm iconname . "VDM DB"
wm minsize  . 1 1

set curradt "unknown"
set currimp "unknown"
set curropr "unknown"
set currcmp "unknown"

# -------------------------------------------------------------------------------

proc clearlist { list } {
    while { [ $list size ] != 0 } { $list delete end }
}

# -------------------------------------------------------------------------------

proc browseimpl { } {
    global dir
    global curradt

    clearlist .impl.list
    clearlist .oper.list
    clearlist .comp.list

    foreach i [ exec ls $dir/$curradt ] {
           .impl.list insert end $i
    }
}

proc browseoper { } {
    global dir
    global curradt
    global currimp

    clearlist .oper.list
    clearlist .comp.list

    foreach i [exec ls $dir/$curradt/$currimp ] {
        .oper.list insert end $i
    }
}

proc browsecomp { } {
    global dir
    global curradt
    global currimp
    global curropr

    clearlist .comp.list

    foreach i [exec ls $dir/$curradt/$currimp/$curropr ] {
        .comp.list insert end $i
    }
}

proc editcomp { } {
    global dir
    global curradt
    global currimp
    global curropr
    global currcmp
    global editor

    set comp $curradt/$currimp/$curropr/$currcmp
    set file $dir/$comp

    if [file isfile $file] {
        exec $editor $file &
    } else {
        MessageBox "Component $comp not found in \
VDM Class Library database $dir" "quit" "error" "200"
    }
}

# -------------------------------------------------------------------------------

proc browseadts { } {
    global dir

    clearlist .adts.list

    if {"$dir" == ""} {
        dirDoesNotExists
    }
    foreach i [ exec ls $dir ] {
        .adts.list insert end $i
    }
}

# -------------------------------------------------------------------------------

proc vdmhelp { } {
    MessageBox "
VDM Class Library Database Browser

The left list titled \"ADT\" shows the names of the \
abstract datatype directory. Select one of the entries by a left button \
double click of the mouse. 

The list in the second place titled \"Implementations\" will then \
show the implementations directory for the selected ADT. Select \
one of these Implementations by a left button double click \
of the mouse.

The third list titled \"Operations\" shows the operations \
directory for the selected implementation. When selecting one of them \
by a left button double click, the list of components will \
appear in the right row.

The right list with the components of an operation can be again \
selected by a double click. An editor is then started with this \
component file. Here first the environment variable EDITOR \
is tested, if not set the xedit is used.

To leave the browser or any subwindow, \
click at the \"quit\" button.
" "quit" "VDM Help" "120"
}

# Set up bindings for the browser -----------------------------------------------

browseadts

bind  . <Control-q> {destroy .}
bind  . <Control-c> {destroy .}
focus .

bind  .adts.list <Double-Button-1> {
    foreach i [selection get] {
        set curradt $i
        browseimpl 
    }
}

bind  .impl.list <Double-Button-1> {
    foreach i [selection get] {
        set currimp $i
        browseoper 
    }
}

bind  .oper.list <Double-Button-1> {
    foreach i [selection get] {
        set curropr $i
        browsecomp
    }
}

bind  .comp.list <Double-Button-1> {
    foreach i [selection get] {
        set currcmp $i
        editcomp
    }
}
# -------------------------------------------------------------------------------
