#!/usr/bin/wish -f
#
# $Id: vdminfo.tcl,v 1.4 1994/01/29 13:55:53 vdm Exp $

set listgeo "20x10"

# -------------------------------------------------------------------------------

proc initdir { } {
    global env
    global sysdir

    if [ info exists env(VDMROOT) ] {
        set sysdir $env(VDMROOT)
    } else {
        set sysdir "/usr/lib"
    }
}

proc checkdir { } {
    global sysdir
    global dir
    global infoprog

    if [ file isdirectory $sysdir ] {
       set dir      "$sysdir/lib/vdmdb-c"
       set infoprog "$sysdir/lib/vdminfo.sh"
    } else {
       set dir ""
       set infoprog ""
    }
}

set getadt  "getadt"
set getimpl "getimpl"
set getopr  "getop"
set getdocu "getdocus"
set gettype "getoptype"

# -------------------------------------------------------------------------------

set MessageBoxAns 0
proc MessageBox {Message ButtonNames {Title "Question"} {Aspect "120"}} {
    global MessageBoxAns

    toplevel .question
    wm title .question $Title
    message .question.mess -text $Message -aspect $Aspect
    pack append .question .question.mess  bottom

    foreach ButtonName $ButtonNames {
	button .question.x$ButtonName -text $ButtonName -command "set MessageBoxAns [lsearch $ButtonNames $ButtonName]; destroy .question"
	pack append .question .question.x$ButtonName left
    }

    grab set .question
    tkwait window .question
    return $MessageBoxAns
}

# -------------------------------------------------------------------------------

proc dirDoesNotExists { } {
       MessageBox "VDM database not found, \
please check VDM Class Library installation and \
enviroment variable VDMROOT" "quit" "error" "200"
       exit
}

# -------------------------------------------------------------------------------

initdir
checkdir

label     .title -text "VDM Class Library Browser"

frame     .topbuttons
button    .topbuttons.ok    -text "quit" -command "destroy ."
label     .topbuttons.title -text "VDM Class Library Browser" -relief raised
button    .topbuttons.help  -text " help " -command "vdmhelp"
pack      .topbuttons.ok    -side left 
pack      .topbuttons.help  -side right 
pack      .topbuttons.title -fill x 

frame     .database
label     .database.dbt -text "database : "
label     .database.dbn -text "$dir"
pack      .database.dbt -fill x
pack      .database.dbn -fill x

frame     .adts
label     .adts.label  -text "ADTs"
button    .adts.predef -text "predefined" -command "showpredefined"
scrollbar .adts.scroll -command ".adts.list yview"
listbox   .adts.list   -yscroll ".adts.scroll set" -relief raised -geometry $listgeo -setgrid yes
pack      .adts.label  -side top  -fill both
pack      .adts.predef -side bottom  -fill both
pack      .adts.list   -side left -fill both
pack      .adts.scroll -side left -fill y

frame     .impl
label     .impl.label  -text "Implementations"
button    .impl.common -text "common" -command "showcommon"
scrollbar .impl.scroll -command ".impl.list yview"
listbox   .impl.list   -yscroll ".impl.scroll set" -relief raised -geometry $listgeo -setgrid yes
pack      .impl.label  -side top
pack      .impl.common -side bottom -fill both
pack      .impl.list   -side left -fill both
pack      .impl.scroll -side left -fill y

frame     .oper
label     .oper.label  -text "Operations"
button    .oper.feature -text "features" -command "showfeatures"
scrollbar .oper.scroll -command ".oper.list yview"
listbox   .oper.list   -yscroll ".oper.scroll set" -relief raised -geometry $listgeo -setgrid yes
pack      .oper.label  -side top
pack      .oper.feature -side bottom -fill both
pack      .oper.list   -side left -fill both
pack      .oper.scroll -side left -fill y

pack      .topbuttons  -side top -fill x
#pack      .title       -side top
pack      .database    -side top -fill x
pack      .adts        -side left
pack      .impl        -side left
pack      .oper        -side left

# -------------------------------------------------------------------------------

wm title    . "VDM Info"
wm iconname . "VDM Info"
wm minsize  . 1 1

set curradt "unknown"
set currimp "unknown"
set curropr "unknown"

# -------------------------------------------------------------------------------

proc clearlist { list } {
    while { [ $list size ] != 0 } { $list delete end }
}

# -------------------------------------------------------------------------------

proc showpredefined { } {
    showoper "--predefined--" "--common--" "--features--"
}

proc showcommon { } {
    global curradt

    if {"$curradt" == "unknown"} {
        MessageBox "Please select an ADT first" "quit" "error" "500"
    } else {
        showoper $curradt "--common--" "--features--"
    }
}

proc showfeatures { } {
    global curradt
    global currimp

    if {"$curradt" == "unknown"} {
        MessageBox "Please select an ADT and an implementation first" "quit" "error" "500"
    } else {
    if {"$currimp" == "unknown"} {
        MessageBox "Please select an implementation first" "quit" "error" "500"
    } else {
        showoper $curradt $currimp "--features--"
    }
    }
}

proc browseimpl { } {
    global dir
    global curradt
    global currimp
    global curropr
    global infoprog
    global getimpl

    clearlist .impl.list
    clearlist .oper.list
    set currimp "unknown"

    foreach i [ exec $infoprog $getimpl $dir $curradt ] {
       .impl.list insert end $i
   }
}

proc browseoper { } {
    global dir
    global curradt
    global currimp
    global infoprog
    global getopr

    clearlist .oper.list

    foreach i [exec $infoprog $getopr $dir $curradt $currimp ] {
        .oper.list insert end $i
    }
}

proc title { adt imp opr } {
    global dir
    global infoprog
    global gettype

    if {"$adt" == "--predefined--"} {
        return "predefined datatypes"
    }
    if {"$imp" == "--common--"} {
        return "general information about ADT \"$adt\""
    }
    if {"$opr" == "--features--"} {
        return "features of implementation \"$imp\" for ADT \"$adt\""
    }

    set type [ exec $infoprog $gettype $dir $adt $imp $opr ]

    if {"$type" == "SPECIAL"} {
        return "operation \"$opr\" of implementation \"$imp\" for ADT \"$adt\""
    } 
    if {"$type" == "COMMON"} {
        return "general operation \"$opr\" for ADT \"$adt\""
    } 
    return "common operation \"$opr\" for all ADT's"
}

proc showoper { adt imp opr {w .window} } {
    global dir
    global infoprog
    global getdocu

    catch {destroy $w}
    toplevel $w
    set windowtitle [ title $adt $imp $opr ]

    wm geometry $w +300+300
    wm title    $w "VDM Documentation"
    wm iconname $w "$opr"

    frame  $w.buttons
    button $w.buttons.ok    -text quit -command "destroy $w"
    label  $w.buttons.title -text "$windowtitle" -relief raised
    pack   $w.buttons.ok    -side left
    pack   $w.buttons.title -fill x

    text $w.t -relief raised -bd 2 \
              -yscrollcommand "$w.s set" \
              -setgrid true \
	      -width 80 -height 16 
    scrollbar $w.s -relief flat -command "$w.t yview"

    pack $w.buttons -side top -fill x
    pack $w.s       -side right -fill y
    pack $w.t       -expand yes -fill both

    $w.t insert 0.0 [ exec $infoprog $getdocu $dir $adt $imp $opr gsig ] 
}

# -------------------------------------------------------------------------------

proc browseadts { } {
    global dir
    global infoprog
    global getadt

    clearlist .adts.list

    if {"$dir" == ""} {
        dirDoesNotExists
    }

    foreach i [ exec "$infoprog" $getadt $dir ] {
        .adts.list insert end $i
    }
}

# -------------------------------------------------------------------------------

proc vdmhelp { } {
    MessageBox "
VDM Class Library Browser

The left list titled \"ADTs\" shows the names of the \
abstract datatypes available. Select one of them by a left button \
double click of the mouse. To get \
information about the predefined datatypes, click on the \"predefined\" \
button in the bottom line.

The list in the middle titled \"Implementations\" will then \
show the implementations available for the selected ADT. Select \
one of these Implementations by a left button double click \
of the mouse. To get \
information common to all implementations of the selected ADT, \
click on the \"common\" button in the bottom line.

The right list titled \"Operations\" shows the operations \
available for the selected implementation. When selecting one of them \
by a left button double click, a new text window will give \
detailed information about that operation. \
To get information about the \
features \
of the selected implementation, like time and \
space characteristics and implementation parameters, \
press the \"features\" button in the bottom line.

To leave the browser or any subwindow, \
click at the \"quit\" button.
" "quit" "VDM Help" "120"
}

# Set up bindings for the browser -----------------------------------------------

browseadts

bind  . <Control-q> {destroy .}
bind  . <Control-c> {destroy .}
focus .

bind  .adts.list <Double-Button-1> {
    foreach i [selection get] {
        set curradt $i
        set adtdir    $dir/$curradt
        set comadtdir $adtdir/com
        browseimpl 
    }
}

bind  .impl.list <Double-Button-1> {
    foreach i [selection get] {
        set currimp $i
        set impadtdir $adtdir/$currimp
        browseoper 
    }
}

bind  .oper.list <Double-Button-1> {
    global curradt
    global currimp
    global curropr

    foreach i [selection get] {
        set curropr $i
        showoper $curradt $currimp $curropr
    }
}

# -------------------------------------------------------------------------------
