/***************************************************************************/
/*                                                                         */
/* IDENTITY:         cppmain                                               */
/*                                                                         */
/* TYPE:             Main Module (project: cpp)                            */
/*                                                                         */
/* HISTORY:          DESIGN           Schlotfeldt  10.04.91                */
/*                   CODE             Schlotfeldt  02.05.91                */
/*                                                                         */
/* FUNCTION:         Complete ANSI C-preprocessor with additional          */
/*                   features as:                                          */
/*                     - global stringization operator (#)                 */
/*                       (use ""#... to start a line with this operator,   */
/*                        use \# to get normal # character)                */
/*                     - global merge operator (##)                        */
/*                     - removal of double empty lines                     */
/*                     - automatic string concatenation                    */
/*                   See external project documentation for details!       */
/*                                                                         */
/* ------------------------------------------------------------------------

   %W% %D%

   Change History:
   ===============
   Mar-07-92:	'main' renamed to 'cppmain' if cond. comp. 'INCLUDE_CPP' is
   .		set. This allows linking the cpp to the Class Library Generator.

   ------------------------------------------------------------------------*/
/*                                                                         */
/* DATA/TABLES:      -                                                     */
/*                                                                         */
/* PROCEDURES:       for external use: 'main'                              */
/*                                                                         */
/***************************************************************************/
#include "types.h"
#include "parser.h"
#include "cmdline.h"

#include "errmsg.h"
#include "myalloc.h"

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         main                                                  */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         See module header                                     */
/*                                                                         */
/* PARAMETERS:       1) Number of command line parameters                  */
/*                   2) List of parameter strings                          */
/*                   (including the program call name as 1st parameter)    */
/*                                                                         */
/* RETURN VALUE:     -                                                     */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   Calls 'parse_cmdl_opt' to parse the parameters on     */
/*                   the commandline and initialize all modules.           */
/*                   If 'parse_cmdl_opt' returns a valid 'stream' (not     */
/*                   NULL) to be used by the parser for output, the        */
/*                   'parse' module is called and the selected input       */
/*                   will be processed (and produce output to file or      */
/*                   stdout)                                               */
/*                   All error and help messages will be written to stderr */
/*                   (see 'errmsg' module)                                 */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  -                                                     */
/*                                                                         */
/* ERROR CONDITIONS: Generates a error message by calling 'errmsg' if      */
/*                   'parse_cmdl_opt' returns no valid 'stream'            */
/*                                                                         */
/***************************************************************************/

#ifdef INCLUDE_CPP
int cppmain S2(int,argc,char **,argv)
#else
int main S2(int,argc,char **,argv)
#endif
{
  stream *dest;

  if( (dest=parse_cmdl_opt(argc,argv)) )
    parse(dest);
  else
    errmsg(0,15,(dest->name)?dest->name:"stdin",NULL);

  return 0;
}
