/***************************************************************************/
/*                                                                         */
/* IDENTITY:         error                                                 */
/*                                                                         */
/* TYPE:             Module (project: cpp, sub-module to: parser)          */
/*                                                                         */
/* HISTORY:          DESIGN           Schlotfeldt  03.04.91                */
/*                   CODE             Schlotfeldt  27.04.91                */
/*                   VERSION(CURRENT) Schlotfeldt  29.05.91  23:46         */
/*                                                                         */
/* FUNCTION:         Reads a stream of 'symbols' generated by 'next_lex_e' */
/*                   (module 'expandm') and terminated (for this module)   */
/*                   by the ENDOFFILE or NEWLINE 'symbol'.                 */
/*                   The generated 'symbol' list will be passed to the     */
/*                   'errmsg' module.                                      */
/*                                                                         */
/* DATA/TABLES:      -                                                     */
/*                                                                         */
/* PROCEDURES:       for external use: 'ppERROR'                           */
/*                                                                         */
/***************************************************************************/
#include "error.h"

#include "expandm.h"

#include "myalloc.h"
#include "errmsg.h"

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         ppERROR                                               */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         #error-line "sub-parser" (see function of module)     */
/*                                                                         */
/* PARAMETERS:       -                                                     */
/*                                                                         */
/* RETURN VALUE:     -                                                     */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   See inline documentation                              */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  Last 'symbol' read by the 'parser' module before      */
/*                   calling this function is assumed to be the ERROR      */
/*                   'symbol'                                              */
/*                                                                         */
/* ERROR CONDITIONS: Calls 'errmsg' module/function in any case            */
/*                                                                         */
/***************************************************************************/
void ppERROR S0(void)
{
  symbol *pa;                          /* list of 'symbols' buidling the   */
				       /* user defined error message       */

  pa=next_line(NO_EXP);                /* get remaining line as 'symbol'   */
				       /* list and pass this list to the   */
  errmsg(1,18,NULL,pa);                /* 'errmsg' module/function         */

  free(pa);                            /* set now useless memory block     */
}                                      /* 'free'                           */
