/***************************************************************************/
/*                                                                         */
/* IDENTITY:         include                                               */
/*                                                                         */
/* TYPE:             Module (project: cpp, sub-module to: parser)          */
/*                                                                         */
/* HISTORY:          DESIGN           Schlotfeldt  04.04.91                */
/*                   CODE             Schlotfeldt  28.04.91                */
/*                   VERSION(CURRENT) Schlotfeldt  30.05.91  12:06         */
/*                                                                         */
/* FUNCTION:         Parses a stream of 'symbols' generated by 'next_lex_e'*/
/*                   (module 'expandm') and terminated (for this module)   */
/*                   by the ENDOFFILE or NEWLINE 'symbol'.                 */
/*                   'include' assumes the stream to describe a pp-command */
/*                   line, which was already identified as #include-line   */
/*                   by reading the first 'symbol' of this line by the     */
/*                   'parser' module.                                      */
/*                                                                         */
/* DATA/TABLES:      -                                                     */
/*                                                                         */
/* PROCEDURES:       for external use: 'ppINCLUDE'                         */
/*                                                                         */
/***************************************************************************/
#include "include.h"

#include "input.h"
#include "expandm.h"
#include "stringu.h"
#include "srcstack.h"

#include "myalloc.h"
#include "errmsg.h"

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         ppINCLUDE                                             */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         #include-line sub-parser (see function of module).    */
/*                                                                         */
/* PARAMETERS:       -                                                     */
/*                                                                         */
/* RETURN VALUE:     -                                                     */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   Looks for a STRING 'symbol' specifying the name of a  */
/*                   FILE to include (push attached 'stream' to the 'src-  */
/*                   stack'). Uses 'conv2name' to convert the STRING to a  */
/*                   filename.                                             */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  Last 'symbol' read by the 'parser' module before      */
/*                   calling this function is assumed to be the INCLUDE    */
/*                   'symbol'                                              */
/*                   Depending on the type of STRING (start character may  */
/*                   be ", L or >) 'ppINCLUDE' tries to open a 'stream'    */
/*                   searching for the attached FILE in the global and/or  */
/*                   local directory(ies). Which will be passed to 'push-  */
/*                   stream' function of module 'srcstack'.                */
/*                                                                         */
/* ERROR CONDITIONS: Calls 'errmsg' module/function on illegal syntax of   */
/*                   the pp-command line or failure on searching the spec. */
/*                   file or pushing the attached 'stream' to the 'src-    */
/*                   stack' (mail fail on recursive #includes)             */
/*                                                                         */
/***************************************************************************/
void ppINCLUDE S0(void)
{
# define reterr(n,x)                   /* reterr:                          */\
  {                                    /*   generate error message no. 'n',*/\
    errmsg(1,n,x,NULL);                /*   ignore remaining 'symbols' on  */\
    skip2eol(p,NO_EXP)                 /*   the actual line, 'free' 's'    */\
    free(s);                           /*   and return                     */\
    return;                            \
  }

  symbol p;                            /* temporary storage for a 'symbol' */
  char   loc,                          /* 1st character of the STRING spec.*/
				       /* the name of the file to include  */
	 *s;                           /* name of file                     */
  stream *nfp;                         /* 'stream' generated by 'open_r'   */

  p=next_lex_e(LINE_EXP);              /* skip optional WHITESPace         */
  skipsp(p,LINE_EXP)
  if(!(s=conv2name(p)))                /* get filename                     */
    reterr(26,NULL);
  loc=p->text[0];                      /* save first character of string   */
  p=next_lex_e(LINE_EXP);
  skipsp(p,LINE_EXP);                  /* next non-WHITESPace 'symbol' must*/
  if(!endofline(p))                    /* terminate the line               */
    reterr(23,NULL);

  nfp=open_r(s,0);                     /* try to open a 'stream' searching */
				       /* for the FILE in the global in-   */
				       /* clude directories                */
  if(!nfp && loc=='"')                 /* try again if opening failed and  */
    nfp=open_r(s,1);		       /* STRING started with " (for in-   */
				       /* cluding local FILE search)       */
  if(nfp)                              /* success on opening the 'stream' ?*/
  {                                    /* YES: try to push 'stream' to the */
    if(push_stream(nfp))               /*      'srcstack' or generate a    */
    {                                  /*      error for recursive #include*/
      free(nfp->name);                 /*      files                       */
      free(nfp);
      reterr(32,s)
    }
  }
  else                                 /* NO : generate error for "open    */
    reterr(15,s)                       /*      file failed"                */

  free(s);                             /* 'free' allocated memory space    */
  return;                              /* and return                       */
}

