/***************************************************************************/
/*                                                                         */
/* IDENTITY:         input                                                 */
/*                                                                         */
/* TYPE:             Module (project: cpp)                                 */
/*                                                                         */
/* HISTORY:          DESIGN           Schlotfeldt  14.03.91                */
/*                   CODE             Schlotfeldt  11.04.91                */
/*                   VERSION(CURRENT) Schlotfeldt  18.06.91  15:11         */
/*                                                                         */
/* FUNCTION:         Should be used instead of stdio for opening 'streams' */
/*		     for input and to read characters from this 'streams'. */
/*                   The type 'stream' should now be used instead of FILE, */
/*                   'get_char' instead of fgetc and 'open_r' instead of   */
/*                   fopen(...,'r').                                       */
/*                   'open_r' offers together with 'add_incl_dir' the      */
/*                   feature of non-local directory search.                */
/*                   The first call to this module should be 'init_input'! */
/*                                                                         */
/* DATA/TABLES:      Contains static data (hidden in the not global used   */
/*                   function 'open_r_op')                                 */
/*                                                                         */
/* PROCEDURES:       for external use: 'open_r', 'add_incl_dir',           */
/*                       	       'init_input', 'get_char'            */
/*		     for local use   : 'open_r_op'                         */
/*                                                                         */
/***************************************************************************/
#include "input.h"

#include "myalloc.h"

enum cmd_t { open_,add_dir_,init_ };   /* local type for calls to          */
				       /* 'open_r_op'                      */

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         open_r_op                                             */
/*                                                                         */
/* TYPE:             Local function                                        */
/*                                                                         */
/* FUNCTION:         See 'open_r', 'add_incl_dir' and 'init_input'         */
/*                                                                         */
/* PARAMETERS:       1) Command to specify one of the 3 actions            */
/*                   2) Name of 'stream' to build or directory to add to   */
/*                      the local list of search directories               */
/*		     3) Flag for global or (if set) local file search      */
/*                                                                         */
/* RETURN VALUE:     Builded 'stream' object for 'open_r' function or      */
/*                   NULL = error on 'open_r' function or dummy for other  */
/*                          functions                                      */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   'open_' function:                                     */
/*                     Creates initialized 'stream' object and tries to    */
/*                     open the file via fopen in one of the search        */
/*                     (if any defined by 'add_incl_dir'). If 'local'      */
/*                     flag is set and the file couldn't be opened yet,    */
/*                     'open_r_op' tries to open the file in the current   */
/*                     (local) directory.                                  */
/*                     If the name (2nd parameter) was NULL no file is     */
/*                     opened. stdin is used as file pointer.              */
/*                   'add_dir_' function:                                  */
/*                     Adds one entry to the dynamic allocated list of     */
/*                     directory names.                                    */
/*                   'init_' function:                                     */
/*                     Initializes the static variables ndirs and dirs.    */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  See 'open_r', 'add_incl_dir' and 'init_input'         */
/*                                                                         */
/* ERROR CONDITIONS: See 'open_r'                                          */
/*                                                                         */
/***************************************************************************/
stream *open_r_op S3(enum cmd_t,cmd,char *,s,char,local)
{
  static char   **dirs;                /* dynamic allocated array of       */
				       /* directory names                  */
  static size_t ndirs;                 /* actual size of dirs array        */
  auto   size_t i;                     /* loop variable                    */
  auto   stream *f;                    /* 'stream' for sub-function 'open' */

#if __MSDOS__
  auto   size_t j;                     /* extra loop variable for DOS      */
				       /* systems (transl. '/' -> '\')     */
#define TRANSL(x,y)                    \
 for(j=1;j<strlen(f->name);j++)        /* translation macro defined for    */\
   if(f->name[j]==x) f->name[j]=y      /* DOS ...                          */
#else
#define TRANSL(x,y)                    /* ... and ignored by other OSs     */
#endif

  switch(cmd)
  {
    case init_:                        /* 'init_input' function            */
	 dirs=NULL;
	 ndirs=0;
	 return(NULL);
    case open_:                        /* 'open_r' function                */
	 f=(stream *)malloc(sizeof(stream));
	 f->fp=NULL;
	 f->line=0;
	 f->name=NULL;
	 time(&f->time);
	 if(!s)                             /* create 'stream' for 'stdin' */
	 {                                  /* if filename was NULL        */
	   f->fp=stdin;
	   return(f);
	 }
	 if(!local)
	   for(i=0;i<ndirs;i++)             /* search in global directories*/
	   {
	     f->name
	       =(char *)malloc(strlen(dirs[i])+strlen(s)+2);
	     strcpy(f->name,dirs[i]);
	     strcat(f->name,"/");
	     strcat(f->name,s);
	     TRANSL('/','\\');
	     if( (f->fp=fopen(f->name,"r")) ) break;
	     free(f->name);
	   }
	 else                               /* search in local directory   */
	 {
	   f->name=strdup(s);
	   TRANSL('/','\\');
	   if(!(f->fp=fopen(s,"r"))) free(f->name);
	 }
	 if(!f->fp)
	 {
	   free(f);
	   return(NULL);
	 }
	 TRANSL('\\','/');
	 return(f);
    case add_dir_:                     /* 'add_incl_dir' function          */
	 dirs=
	   (char **)realloc(dirs,++ndirs*sizeof(char *));
	 dirs[ndirs-1]=s;
	 return(NULL);
  }
  return(NULL);
}

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         add_incl_dir                                          */
/*                                                                         */
/* TYPE:             Procedure                                             */
/*                                                                         */
/* FUNCTION:         Adds one directory name to the local list of search   */
/*                   directories for include files in 'open_r_op'.         */
/*                                                                         */
/* PARAMETERS:       Name of the directory. Subdirectories should be       */
/*                   separeted by '/' (translation to '\' on DOS systems   */
/*                   is done automatic)                                    */
/*                                                                         */
/* RETURN VALUE:     -                                                     */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   Calls 'open_r_op' to add the name to the local list   */
/*                   of search directories for include files.              */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  Name must not be NULL.                                */
/*                                                                         */
/* ERROR CONDITIONS: -                                                     */
/*                                                                         */
/***************************************************************************/
void add_incl_dir S1(char *,d)
{
  open_r_op(add_dir_,d,0);
}

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         open_r                                                */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         Opens a 'stream' for reading.                         */
/*                                                                         */
/* PARAMETERS:       1) Name of 'stream' to be opened                      */
/*                      (NULL builds a 'stream' object for stdin)          */
/*                   2) Flag showing wether 'stream' may be local          */
/*                                                                         */
/* RETURN VALUE:     initialized 'stream' object as defined in 'types.h'   */
/*                   or NULL (error)                                       */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   Gets character from the open 'stream'. If the         */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  2nd parameter makes just sense if previously          */
/*                   'add_incl_dir' was called                             */
/*                                                                         */
/* ERROR CONDITIONS: Returns NULL if 'stream' can't by opened.             */
/*                                                                         */
/***************************************************************************/
stream *open_r S2(char *,s,char,local)
{
  return(open_r_op(open_,s,local));
}

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         get_char                                              */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         Gets one character from a 'stream'.                   */
/*                                                                         */
/* PARAMETERS:       'stream' object as defined in 'types.h'               */
/*                                                                         */
/* RETURN VALUE:     Character read and changed 'stream'                   */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   Gets character from the open 'stream'. If the         */
/*		     'stream' is closed, get_char returns EOF. The         */
/*                   'stream' is automatically closed if the 'stream'      */
/*                   returns EOF himself.                                  */
/*                   Closed 'streams' are marked by the NULL pointer as    */
/*                   file pointer (defined in stdio.h).                    */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  'stream' should be previously opened by 'open_r'.     */
/*                                                                         */
/* ERROR CONDITIONS: -							   */
/*                                                                         */
/***************************************************************************/
char get_char S1(stream *,f)
{
  char c;

  if(f->fp)
  {
    if((c=getc(f->fp))==(char)EOF)
    {
      fclose(f->fp);
      f->fp=NULL;
    }
  }
  else c=(char)EOF;
  return(c);
}

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         init_input                                            */
/*                                                                         */
/* TYPE:             Procedure                                             */
/*                                                                         */
/* FUNCTION:         Initialization of module 'input'.                     */
/*                                                                         */
/* PARAMETERS:       -                                                     */
/*                                                                         */
/* RETURN VALUE:     -                                                     */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   Calls 'open_r_op' to clear the local list of search   */
/*		     directories for include files.                        */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  Must be called as first procedure of this module.     */
/*                                                                         */
/* ERROR CONDITIONS: -							   */
/*                                                                         */
/***************************************************************************/
void init_input S0(void)
{
  open_r_op(init_,"",0);
}
