/***************************************************************************/
/*                                                                         */
/* IDENTITY:         line                                                  */
/*                                                                         */
/* TYPE:             Module (project: cpp, sub-module to: parser)          */
/*                                                                         */
/* HISTORY:          DESIGN           Schlotfeldt  05.04.91                */
/*                   CODE             Schlotfeldt  01.05.91                */
/*                   VERSION(CURRENT) Schlotfeldt  01.06.91  11:27         */
/*                                                                         */
/* FUNCTION:         Parses a stream of 'symbols' generated by 'next_lex_e'*/
/*                   (module 'expandm') and terminated (for this module)   */
/*                   by the ENDOFFILE or NEWLINE 'symbol'.                 */
/*                   'line' assumes the stream to describe a pp-command    */
/*                   line, which was already identified as #line-line by   */
/*                   reading the first 'symbol' of this line by the        */
/*                   'parser' module.                                      */
/*                                                                         */
/* DATA/TABLES:      -                                                     */
/*                                                                         */
/* PROCEDURES:       for external use: 'ppLINE'                            */
/*                                                                         */
/***************************************************************************/
#include "line.h"

#include "evalexpr.h"
#include "expandm.h"
#include "stringu.h"
#include "srcstack.h"

#include "myalloc.h"
#include "errmsg.h"

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         ppLINE                                                */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         #line-line sub-parser (see function of module)        */
/*                                                                         */
/* PARAMETERS:       -                                                     */
/*                                                                         */
/* RETURN VALUE:     -                                                     */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   Tries to read a legal line number (>0) by calling     */
/*                   the 'const_expr' parser. The next 'symbols' on        */
/*                   the pp-command line must be a (convertable by 'conv-  */
/*                   2name') or a line termination.                        */
/*                   'line' uses 'curr_stream' function to modify the      */
/*                   line number (+ opt.filename) of the current 'stream'  */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  Last 'symbol' read by the 'parser' module before      */
/*                   calling this function is assumed to be the LINE       */
/*                   'symbol'                                              */
/*                                                                         */
/* ERROR CONDITIONS: Calls 'errmsg' module/function on illegal line number */
/*                   (may be caused by error on expression syntax) or      */
/*                   illegal filename ('conv2name' returned NULL).         */
/*                                                                         */
/***************************************************************************/
void ppLINE S0(void)
{
# define reterr(n)                     /* reterr:                          */\
  {                                    /*   generate error message no. 'n',*/\
    errmsg(1,n,NULL,NULL);             /*   ignore remaining 'symbols' on  */\
    skip2eol(p,NO_EXP)                 /*   the actual line and return to  */\
    return;                            /*   calling function               */\
  }

  unsigned long l;                     /* line number                      */
  symbol        p;                     /* temporary storage for a 'symbol' */
  stream        *currstr;              /* return value from 'curr_stream'  */

  currstr=curr_stream();               /* get 'stream' for 'srcstack'      */

  l=const_expr();                      /* parser line number               */
  p=next_lex_e(LINE_EXP);
  if(!l) reterr(34)                    /* check for valid number (1, 2,...)*/
  if(!endofline(p))                    /* next 'symbol' is does not termi- */
  {                                    /* nate the current line:           */
    char *s;                                /* filename                    */

    s=conv2name(p);                         /* try to convert to c-string  */
    if(s && p->text[0]!='"')                /* STRING 'symbol' must start  */
    {                                       /* with double quotes          */
      free(s);                              /* else: modify 's' as error   */
      s=NULL;                               /*       return value NULL     */
    }
    p=next_lex_e(LINE_EXP);                 /* get next non-WHITESPace     */
    skipsp(p,LINE_EXP);                     /* 'symbol'                    */
    if(!s)            reterr(36)            /* return error on illegal     */
					    /* string                      */
    if(!endofline(p)) reterr(23)            /* return error if 'p' does not*/
					    /* terminate the line          */
    if(currstr)
    {
      free(currstr->name);                  /* set new name for 'currstr'  */
      currstr->name=s;
    }
  }

  if(currstr) currstr->line=l-1;            /* set new line number for     */
					    /* 'currstr'                   */
}

