/***************************************************************************/
/*                                                                         */
/* IDENTITY:         macro                                                 */
/*                                                                         */
/* TYPE:             Module (project: cpp)                                 */
/*                                                                         */
/* HISTORY:          DESIGN           Schlotfeldt  26.03.91                */
/*                   CODE             Schlotfeldt  19.04.91                */
/*                   VERSION(CURRENT) Schlotfeldt  26.05.91  15:06         */
/*                                                                         */
/* FUNCTION:         Manages the 'mcr' (macro) part of all 'symbols',      */
/*		     which can be defined, searched and undefined by       */
/*                   function calls to this module.                        */
/*                   On Initialization of this module there'll be a few    */
/*                   system 'macros' defined containing eg. the actual     */
/*                   input filename.                                       */
/*                   The actual value of system 'macros' is build by       */
/*                   calling module 'strinfo'.                             */
/*                   The first call to this module should be 'init_macros' */
/*                   at this time module 'symbol' must already be initia-  */
/*                   lized!                                                */
/*                                                                         */
/* DATA/TABLES:      -                                                     */
/*                                                                         */
/* PROCEDURES:       for external use: 'init_macros', 'definem',           */
/*                       	       'undefm', 'searchm'                 */
/*                   for local use   : 'symlcmp'                           */
/*                                                                         */
/***************************************************************************/
#include "macro.h"

#include "strinfo.h"
#include "stringu.h"
#include "symbol.h"

#include "myalloc.h"

static char ignoreredefinition = 0;

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         searchm                                               */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         Looks if there is a 'macro' defined for a given       */
/*                   'symbol'.                                             */
/*                                                                         */
/* PARAMETERS:       'symbol' specifying the name of the searched 'macro'  */
/*                                                                         */
/* RETURN VALUE:     Searched 'macro' or NULL if there was no 'macro'      */
/*                   found.                                                */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   -                                                     */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  Module 'macro' must be previously initialized by a    */
/*                   call to 'init_macros'                                 */
/*                                                                         */
/* ERROR CONDITIONS: Offending against restriction may cause chaos.        */
/*                                                                         */
/***************************************************************************/
macro searchm S1(symbol,s)
{
# define quoted(x)                     /* quoted:                          */\
  {                                    /*   Stores in the 1st element of   */\
    sysrplc=NULL;                      /*   the 'rpl' part of the local    */\
    app_string(&sysrplc,"\"");         /*   variable 'm' a STRING 'sym-    */\
    app_string(&sysrplc,x);            /*   bol' (quoted string 'x')       */\
    app_string(&sysrplc,"\"");                                               \
    m->rpl[0]=mk_symbol(sysrplc,STRING,0l);                                  \
  }

  macro m;
  char  *sysrplc;                      /* replace text for sytem 'macro'   */

  if((m=s->mcr) && m->syst)
  {
    if(s==STDC_MCR)                    /* create new replacement for       */
      m->rpl[0]=ONE;                   /* system 'macros':                 */
    if(s==DATE_MCR)                    /* __STDC__ (if defined by 'init-   */
      quoted(get_date())               /*          macros')                */
    if(s==TIME_MCR)                    /* __DATE__ (quoted strings)        */
      quoted(get_time())               /* __TIME__        "                */
    if(s==FILE_MCR)                    /* __FILE__        "                */
      quoted(get_name())               /* __LINE__                         */
    if(s==LINE_MCR)
      m->rpl[0]=mk_symbol(strdup(get_line(0)),ICONST,0l);
    m->rpl[1]=NULL;
  }
  return(m);
}

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         symlcmp                                               */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         Compares two list of 'symbols'                        */
/*                                                                         */
/* PARAMETERS:       'symbol' lists to compare (NULL terminated or NULL)   */
/*                                                                         */
/* RETURN VALUE:     0 : for lists are identical (NULL == NULL !)          */
/*                   1 : for lists are different                           */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   -                                                     */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  -                                                     */
/*                                                                         */
/* ERROR CONDITIONS: -                                                     */
/*                                                                         */
/***************************************************************************/
int symlcmp S2(symbol *,sl1,symbol *,sl2)
{
  if(sl1 && sl2)
  {
    while(*sl1==*sl2++)
      if(!*sl1++) return(0);
    return(1);
  }
  return(sl1 || sl2);
}

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         definem                                               */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         Defines (creates) a 'macro' with spec.name,           */
/*                   parameters and replace text.                          */
/*                                                                         */
/* PARAMETERS:       1) 'symbol' specifying the name of the new 'macro'    */
/*                   2) List of 'symbols' specifying the names of the      */
/*                      parameters (NULL terminated list) or NULL for:     */
/*                      'macro' has no parameters.                         */
/*                   3) List of 'symbols' specifying the replace text      */
/*                      containing parameter names.                        */
/*                      (NULL or NULL terminated list)                     */
/*                   4) Flag for system attribute for 'macro' to define    */
/*                      (see 'init_macros', normal use: 0)                 */
/*                                                                         */
/* RETURN VALUE:     Defined 'macro' or NULL if error occured (illegal     */
/*                   redefinition of a system 'macro' or a non-system      */
/*                   'macro' with an uniue replace text)                   */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   -                                                     */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  Don't call 'definem' from outside this module with    */
/*                   4th parameter !=0.                                    */
/*                                                                         */
/* ERROR CONDITIONS: -                                                     */
/*                                                                         */
/***************************************************************************/
macro definem S4(symbol,s,symbol *,par,symbol *,replc,char,syst)
{
  macro m=NULL;

  if(s && s->type==IDENT)              /* check 1st parameter to represent */
  {                                    /* an identifier                    */
    if(ignoreredefinition && s->mcr)
      undefm(s);
    if(!s->mcr)                        /* 'macro' already defined for this */
    {                                  /* 'symbol' ?                       */
      m=(macro)malloc(sizeof(struct mac_str));
      m->pc  =0;                       /* initialize 'pc' part of the      */
				       /* 'macro' for parameter counting   */
      m->par =par;                     /* store parameter names            */
      if(par)                          /* count parameters (if any)        */
	while(par[m->pc]) m->pc++;
      m->rpl =replc;                   /* store replacement                */
      m->syst=syst;                    /* set/clear system-attribute       */
      m->lock=0;                       /* clear lock flag                  */
      return(s->mcr=m);
    }
    else                               /* check for legal redefinition     */
      m=(!symlcmp(replc,s->mcr->rpl) && !symlcmp(par,s->mcr->par))?
	s->mcr:NULL;
  }
  free(replc);
  free(par);
  return(m);
}

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         undefm                                                */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         Undefines a 'macro'                                   */
/*                                                                         */
/* PARAMETERS:       'symbol' specifying the name of the 'macro' to delete */
/*                                                                         */
/* RETURN VALUE:     -                                                     */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   -                                                     */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  'undefm' doesn't check if the 'macro' was defined by  */
/*                   'definem'. Deletion of undefined 'macros' is simply   */
/*                   ignored.                                              */
/*                                                                         */
/* ERROR CONDITIONS: -                                                     */
/*                                                                         */
/***************************************************************************/
void undefm S1(symbol,s)
{
  if(s && s->mcr)
  {
    free(s->mcr->rpl);
    free(s->mcr->par);
    free(s->mcr);
    s->mcr=NULL;
  }
}

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         init_macros                                           */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         Initialization of this module (defines "empty bodies" */
/*                   for the system 'macros' __DATE__, __TIME__, __FILE__, */
/*                   __LINE__ and optional __STDC__)                       */
/*                                                                         */
/* PARAMETERS:       Flag: 1=__STDC__ 'macro' will be defined as system    */
/*                           'macro' with value "1"                        */
/*                         0=__STDC__ 'macro' won't be defined             */
/*                 2.Flag: 1= ignore redefinition of macros                */
/*                                                                         */
/*                                                                         */
/* RETURN VALUE:     -                                                     */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   -                                                     */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  Module 'symbol' must be previously initialized by a   */
/*                   call to 'init_symbols'                                */
/*                   'init_macros' should be after that the first call to  */
/*                   this module.                                          */
/*                                                                         */
/* ERROR CONDITIONS: -                                                     */
/*                                                                         */
/***************************************************************************/
void init_macros S2(char,stdc,char,ignoreredef)
{
  symbol *rpl;                         /* 'macro' replacement for the syst.*/
				       /* 'macros'                         */
  ignoreredefinition=ignoreredef;
  if(stdc)
  {
    rpl=(symbol *)malloc(2*sizeof(symbol));
    definem(STDC_MCR,NULL,rpl,1);
  }
  rpl=(symbol *)malloc(2*sizeof(symbol));
  definem(DATE_MCR,NULL,rpl,1);
  rpl=(symbol *)malloc(2*sizeof(symbol));
  definem(TIME_MCR,NULL,rpl,1);
  rpl=(symbol *)malloc(2*sizeof(symbol));
  definem(FILE_MCR,NULL,rpl,1);
  rpl=(symbol *)malloc(2*sizeof(symbol));
  definem(LINE_MCR,NULL,rpl,1);
}
