/***************************************************************************/
/*                                                                         */
/* IDENTITY:         myctype                                               */
/*                                                                         */
/* TYPE:             Module (project: cpp)                                 */
/*                                                                         */
/* HISTORY:          DESIGN           Schlotfeldt  11.03.91                */
/*                   CODE             Schlotfeldt  13.04.91                */
/*                   VERSION(CURRENT) Schlotfeldt  20.06.91  15:38         */
/*                                                                         */
/* FUNCTION:         'myctype.h' should be used instead of 'ctype.h'.      */
/*                   'myctype.h' defines macros calling 'chk_cclass'       */
/*		     This macros have similar/equal names as macros de-    */
/*                   fined by 'ctype.h'. Function implementation ('chk-    */
/*                   cclass') wont have side effects on the parameters !   */
/*                                                                         */
/* DATA/TABLES:      -                                                     */
/*                                                                         */
/* PROCEDURES:       for external use: 'chk_cclass'                        */
/*                                                                         */
/***************************************************************************/
#include "myctype.h"

#undef isdigit
#undef isxdigit
#undef iscntrl
#undef tolower

#include <ctype.h>

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         chk_cclass                                            */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         check wether given parameter to belong to a specified */
/*                   class of characters                                   */
/*                                                                         */
/* PARAMETERS:       1) Character to analyze                               */
/*                   2) Class (0..9) :                                     */
/*                      0 : a decimal digit                                */
/*                      1 : a hexadecimal digit (non case sensitive)       */
/*                      2 : legal as first character of an indentifier     */
/*                      3 : legal as 2nd,3rd,4th,... character of an iden- */
/*                          tifier                                         */
/*                      4 : a control character                            */
/*                      5 : a upper case letter (-> special return value)  */
/*                                                                         */
/* RETURN VALUE:     1 or 0 for character belongs / belong not to the spe- */
/*                   cific class                                           */
/*                   or for class 5: the 'tolower' character               */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   Uses macros defined in 'ctype.h'                      */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  Character may have any value (even non-ASCII)         */
/*                                                                         */
/* ERROR CONDITIONS: -                                                     */
/*                                                                         */
/***************************************************************************/
char chk_cclass S2(char,c,char,cl)
{
  if(isascii(c))
  switch(cl)
  {
    case 0 : return(isdigit(c));
    case 1 : return(isdigit(c) || (c<='f' && c>='a') || (c<='F' && c>='A'));
    case 2 : return(isalpha(c) || c=='_' || c=='$');
    case 3 : return(isalnum(c) || c=='_' || c=='$');
    case 4 : return(iscntrl(c));
    case 5 : return(isupper(c)?tolower(c):c);
  }
  return(cl==5?c:0);
}

