/***************************************************************************/
/*                                                                         */
/* IDENTITY:         undef                                                 */
/*                                                                         */
/* TYPE:             Module (project: cpp, sub-module to: parser)          */
/*                                                                         */
/* HISTORY:          DESIGN           Schlotfeldt  03.04.91                */
/*                   CODE             Schlotfeldt  27.04.91                */
/*                   VERSION(CURRENT) Schlotfeldt  29.05.91  23:25         */
/*                                                                         */
/* FUNCTION:         Parses a stream of 'symbols' generated by 'next_lex_e'*/
/*                   (module 'expandm') and terminated (for this module)   */
/*                   by the ENDOFFILE or NEWLINE 'symbol'.                 */
/*                   'undef' assumes the stream to describe a pp-command   */
/*                   line, which was already identified as #undef-line by  */
/*                   reading the first 'symbol' of this line by the        */
/*                   'parser' module.                                      */
/*                                                                         */
/* DATA/TABLES:      -                                                     */
/*                                                                         */
/* PROCEDURES:       for external use: 'ppUNDEF'                           */
/*                                                                         */
/***************************************************************************/
#include "undef.h"

#include "expandm.h"
#include "macro.h"

#include "errmsg.h"

/***************************************************************************/
/*                                                                         */
/* IDENTITY:         ppUNDEF                                               */
/*                                                                         */
/* TYPE:             Function                                              */
/*                                                                         */
/* FUNCTION:         #undef-line sub-parser (see function of module)       */
/*                                                                         */
/* PARAMETERS:       -                                                     */
/*                                                                         */
/* RETURN VALUE:     -                                                     */
/*                                                                         */
/* PROCESSING                                                              */
/*    INFORMATION:   See inline documentation                              */
/*                                                                         */
/* ASSUMPTIONS /                                                           */
/*    RESTRICTIONS:  Last 'symbol' read by the 'parser' module before      */
/*                   calling this function is assumed to be the UNDEF      */
/*                   'symbol'                                              */
/*                                                                         */
/* ERROR CONDITIONS: Calls 'errmsg' module/function on illegal syntax of   */
/*                   the pp-command line                                   */
/*                                                                         */
/***************************************************************************/
void ppUNDEF S0(void)
{
# define reterr(n)                     /* reterr:                          */\
  {                                    /*   generates the error message 'n'*/\
    errmsg(1,n,NULL,NULL);             /*   and jumps to the error exit of */\
    goto error;                        /*   this function                  */\
  }

  symbol name,                         /* name of 'macro' to be un-defined */
	 p;                            /* temporary storage for a 'symbol' */

  p=next_lex_e(NO_EXP);                /* skip optional WHITESPace         */
  skipsp(p,NO_EXP)
  if((name=p)->type!=IDENT)            /* check 'macro' name to be an      */
    reterr( 7)                         /* identifier -> else: error        */
  p=next_lex_e(NO_EXP);
  if(!endofline(p))                    /* next 'symbol' must terminate the */
    reterr(23)                         /* #undef-line                      */
  undefm(name);                        /* call 'undef' to remove the spec. */
  return;                              /* 'macro' from the 'macro' table   */
				       /* (see 'macro' module)             */

  error:                               /* error exit:                      */
  skip2eol(p,NO_EXP);                  /*   ignore the remaining part of   */
				       /*   the line and return            */
}
